C***************************************************************************
C***************************************************************************
C.....Interface to the MD simulation program EGO.
C.....by M. Eichinger and J. Hutter
C.....Functions in this file:
C.....INTERFACE(...)	The main loop for the interface mode.
C.....INTERFACE_WAIT(.)	Waits for the QMCONTINUE file created by EGO.
C.....INTERFACE_READ(.) Reads the new data from EGO.
C.....RDEGOF(..)	Reads the external field.
C.....EXT_FIELD		Calculates the external field at grid points.
C.....CLCPOT(..)	Calculates the electrostatic field at any point.
C.....INTERFACE_WRITE() Writes the forces on the QM atoms for EGO.
C.....EGOSTOP		Creates a file for stopping EGO in case of problems.
C.....MYPROPPT(...)     Calculates the dipole moment
C..... Check function calls for new CPMD versions:
C..... DYNIT(...)

C***************************************************************************
C***************************************************************************
C     ==================================================================
      SUBROUTINE INTERFACE(C0,C2,SC0,PME,GDE,VPP,EIGV,SCR,LSCR)
C     ==--------------------------------------------------------------==
      IMPLICIT REAL*8 (A-H,O-Z)
      INCLUDE 'system.h'
      INCLUDE 'ener.inc' 
      INCLUDE 'elct.inc' 
      INCLUDE 'tpar.inc' 
      INCLUDE 'pslo.inc' 
      INCLUDE 'norm.inc' 
      INCLUDE 'ropt.inc' 
      INCLUDE 'cppt.inc' 
      INCLUDE 'sfac.inc' 
      INCLUDE 'coor.inc' 
      INCLUDE 'nlcc.inc' 
      INCLUDE 'ions.inc' 
      INCLUDE 'epot.inc'
      COMPLEX*16 C0(*),C2(*),SC0(*),PME(*),GDE(*)
      LOGICAL EOFRUN,FNOWF
      CHARACTER INTFILEN*80,LINE*80
      DIMENSION VPP(*),EIGV(*),SCR(*)
      DIMENSION ACC(NACX),IREC(100)
      DATA IREC /1,1,1,1,1,1,1,1,1,0,90*0/
C     ==--------------------------------------------------------------==
      CLOSE(5)
      TIME1 = TIMEC()
C..+1 for esp charges
      CALL MEMORY(IP_TAUP,3*NAX*NSX+1,'TAUP')
      CALL MEMORY(IP_FION,3*NAX*NSX,'FION')
      CALL MEMORY(IP_VELP,3*NAX*NSX,'VELP')
      CALL AZZERO(TAUP,3*NAX*NSX)
      CALL AZZERO(VELP,3*NAX*NSX)
      CALL AZZERO(BOXQM,6)
      CALL AZZERO(BOXDUM,6)
      CALL AZZERO(KOLOC,3*MAXIND) 
      CALL AZZERO(LOCEXP,10*MAXIND)
      CALL AZZERO(DATNEAR,5*MAXNEAR)
      CALL AZZERO(KONEAR,3*MAXNEAR)
      CALL IAZZERO(LOCID,MAXIND)
      CALL IAZZERO(NINCL,MAXIND)
      CALL IAZZERO(IDNEAR,MAXNEAR)
      CALL IAZZERO(LINCL,MAXIND*MAXNEAR)       
      DO I=1,100
         MYRAG(I)=1.2
      ENDDO 
      MYRAG(1)=0.85 
      MYRAG(8)=1.23    
         
      NNEAR = 0
      NLOC  = 0

      RESTART=.FALSE.  
      RWF    =.TRUE. 
      EOFRUN=.FALSE.
      NACC  = 7
      NFI   = 0
      ECNSTR= 0.0D0
      ETOTO = 0.0D0
C     
      IF(PARENT) WRITE(6,'(A)') ' INTERFACE| ENTER EGO-INTERFACE'
      CALL DYNIT(EKINCP,EKIN1,EKIN2,TEMP1,TEMP2,EKINH1,EKINH2)
Cold  CALL DYNIT(EKINCP,EKIN1,EKIN2,TEMP1,TEMP2)
C     ==--------------------------------------------------------------==
C     ==      THE BASIC MOLECULAR DYNAMICS LOOP                       ==
C     ==--------------------------------------------------------------==
    1 CONTINUE
        NFI=NFI+1
        CONVWF=.FALSE.
        SDIIS =.TRUE. 
        SPCG  =.TRUE. 
        MODENS=.FALSE.
        ENGPRI=.FALSE.
        GAMDIA=.FALSE.
        CALSTE=.FALSE.
C
C.......wait for the interface file
C
        CALL INTERFACE_WAIT(INTFILEN)
        CALL TESTEX(EOFRUN)
        IF(EOFRUN) GOTO 2
C
C.......read and process interface file
C
        CALL INTERFACE_READ(TAU0)
        CALL PHFAC(TAU0)
        CALL MY_SYNC(ALLGRP)
C
C.......Randomize Wavefunction
C
        IF(TRANE) THEN
           WRITE(6,'(A,F12.8)') 
     *   ' RANDOMIZE INITIAL WAVEFUNCTION,  AMPLITUDE=',AMPRE
           CALL RRANE(AMPRE,C0,C2,N,NGW)
        ENDIF
C       
        FNOWF=.FALSE.
        DO INFI=1,NOMORE
          TIME1=TIMEC()
          IF(PARENT) THEN
            ENGPRI=PPRN.AND.MOD(INFI-1,IPRINT).EQ.0
          ELSE
            ENGPRI=.FALSE.
          ENDIF
C
C.......update the wavefunctions 
C
           CALL UPDWF(C0,C2,SC0,TAU0,FION,PME,GDE,VPP,
     *               EIGV,SCR,LSCR,N,FNOWF)
C
C.......printout evolution of the progress
C
          IF(PARENT) THEN
            DETOT=ETOT+EEXT-ETOTO
            IF(INFI.EQ.1) DETOT=0.0D0
            TIME2=TIMEC()
            TCPU=(TIME2-TIME1)*0.001D0
            IF(ENGPRI) CALL WRENER
            IF(ENGPRI) WRITE(6,'(A)')
     *' NFI     GEMAX     CNORM          ETOT         DETOT        TCPU'
            WRITE(6,'(I4,F10.6,F10.6,F14.5,F14.6,F12.2) ')
     *       INFI,GEMAX,CNORM,ETOT+EEXT,DETOT,TCPU
            ETOTO=ETOT+EEXT
          ENDIF
          IF((CONVWF).AND.(FNOWF)) GOTO 100
          IF(CONVWF) THEN
             CONVWF=.FALSE.
             FNOWF=.TRUE.
          ENDIF          
        ENDDO
  100   CONTINUE
        IF (PARENT) THEN
         PRINT*,'EXTERNAL ENERGY    =',EEXT,' AU'
         PRINT*,'REAL TOTAL ENERGY  =',EEXT+ETOT,' AU'
        ENDIF
C
C..Calculate the dipole moment
C 
        IF(PROPER) THEN
            CALL GETARG(1,LINE)
            OPEN(UNIT=5,FILE=LINE,STATUS='OLD',ERR=2)
            CALL MYPROPPT(C0,TAU0,SCR)
            CLOSE(5)
        ENDIF
C
C..periodically write the restart file
C
C        IF(MOD(NFI/10,ISTORE).EQ.0) THEN  
         CALL ZHWWF(2,IREC,C0,C2,N,ACC,NACC,TAU0,VELP,TAUP,NFI)
C        ENDIF
C
C..process data and write the interface file
C
        CALL INTERFACE_WRITE(INTFILEN,TAU0,FION,C0,TAUP,SCR,LSCR)

C
C..test for exit-file
C        
        IF(EOFRUN) GOTO 2
        CALL TESTEX(EOFRUN)
        IF(EOFRUN) GOTO 2
      GOTO 1
    2 CONTINUE
C
      CALL ZHWWF(2,IREC,C0,C2,N,ACC,NACC,TAU0,VELP,TAUP,NFI)
      IF(PARENT) CALL FINALP(TAU0,FION)
      CALL FREEM(IP_TAUP)
      CALL FREEM(IP_FION)
      CALL FREEM(IP_VELP)
C     ==--------------------------------------------------------------==
      RETURN
      END
C     ==================================================================
C
C
C
C     ==================================================================
      SUBROUTINE INTERFACE_READ(TAU0)
C     ==--------------------------------------------------------------==
      IMPLICIT REAL*8 (A-H,O-Z)
      INCLUDE 'system.h'
      INCLUDE 'ions.inc'
      INCLUDE 'efld.inc'
      INCLUDE 'epot.inc'
      INCLUDE 'irat.inc'
      CHARACTER LINE*80
      DIMENSION TAU0(3,NAX,*)
C     ==--------------------------------------------------------------==
      IF(PARENT) THEN
        IUNIT=5
        IMDUM=IARGC()
        IF(IMDUM.LT.1) THEN
           WRITE(6,*) ' INTERFACE| NO INPUT FILE NAME SPECIFIED '
           CALL STOPGM('INTERFACE',' ')
        ENDIF        
        CALL GETARG(1,LINE)
        WRITE(6,*) ' INTERFACE| READING NEW COORDINATES FROM FILE ',LINE
        OPEN(UNIT=IUNIT,FILE=LINE,STATUS='OLD',ERR=300)
C
C.......read atomic coordinates
C
        IERR=INSCAN(IUNIT,'&ATOM')
        ISP=0
   10   CONTINUE
           READ(IUNIT,END=200,ERR=300,FMT='(A)') LINE
           IF(INDEX(LINE,'&END').NE.0) GOTO 30
           IF(LINE(1:1).NE.'*') GOTO 10
             ISP=ISP+1
             READ(IUNIT,*)
             READ(IUNIT,*) NAIT
             DO I=1,NAIT
               READ(IUNIT,*) TAU0(1,I,ISP),TAU0(2,I,ISP),TAU0(3,I,ISP)
             ENDDO
           GOTO 10
   30   CONTINUE
C
C.......read external field
C
        IERR=INSCAN(IUNIT,'&EXTE')
        IF(IERR.EQ.0) THEN
          CALL RDEGOF(IUNIT)
          TEXTFLD=.TRUE.
        ELSE
          TEXTFLD=.FALSE.
        ENDIF      
        CLOSE(IUNIT)
C
C.......terminate reading normal
C
  200   CONTINUE
        WRITE(6,*) 'INTERFACE| READING NEW COORDINATES ... DONE'
        GOTO 400
  300   CONTINUE
        WRITE(6,*) 'INTERFACE| CAN NOT READ FROM FILE ',LINE
        CALL STOPGM('INTERFACE',' ')
  400 CONTINUE
      ENDIF
C.....broadcast information to all processors
      CALL MY_SYNC(ALLGRP)
c.....tau0
      MSGLEN = 3*NAX*NSX * 8
      CALL MY_BCAST(TAU0,MSGLEN,SOURCE,ALLGRP)
c.....textfld
      MSGLEN = 1 * 8/IRAT
      CALL MY_BCAST(TEXTFLD,MSGLEN,SOURCE,ALLGRP)
      CALL MY_SYNC(ALLGRP)
      IF(TEXTFLD) THEN
c.....epot1
        MSGLEN = (2+2*MAXIND+MAXNEAR+MAXIND*MAXNEAR)*8/IRAT
        CALL MY_BCAST(NNEAR,MSGLEN,SOURCE,ALLGRP)
c.....epot2
        MSGLEN = (12+13*MAXIND+8*MAXNEAR) * 8
        CALL MY_BCAST(BOXQM,MSGLEN,SOURCE,ALLGRP)
      ENDIF
C.....calculate external field at grid points
      IF(TEXTFLD) CALL EXT_FIELD
      CALL MY_SYNC(ALLGRP)
C     ==--------------------------------------------------------------==
      RETURN
      END
C     ==================================================================
C
C
C
C##############################################
      SUBROUTINE RDEGOF(IUNIT)
C##############################################
      IMPLICIT NONE
      INCLUDE 'epot.inc'
      INTEGER IUNIT,ILINES,K,JJ
C
      CHARACTER STRING*100
      CHARACTER SUBSTRING*7
      INTEGER STEP,I,J
C***** Read bounding box **********************************
      READ(IUNIT,'(A)') STRING
      READ(IUNIT,*) BOXDUM(1),BOXDUM(2),
     @              BOXDUM(3),BOXDUM(4),
     @              BOXDUM(5),BOXDUM(6)
      PRINT'(A10,F11.5,F11.5,F11.5,F11.5,F11.5,F11.5)',STRING,
     @              BOXDUM(1),BOXDUM(2),
     @              BOXDUM(3),BOXDUM(4),
     @              BOXDUM(5),BOXDUM(6)
C
C***** Read QM-bounding box ************************************
      READ(IUNIT,'(A)') STRING
      READ(IUNIT,*) BOXQM(1),BOXQM(2),
     @              BOXQM(3),BOXQM(4),
     @              BOXQM(5),BOXQM(6)
      PRINT'(A10,F11.5,F11.5,F11.5,F11.5,F11.5,F11.5)',STRING,
     @              BOXQM(1),BOXQM(2),
     @              BOXQM(3),BOXQM(4),
     @              BOXQM(5),BOXQM(6)
C
C.....Read local expansions ************************************
C
      READ(IUNIT,'(A)') STRING
      READ(IUNIT,*) NLOC
      PRINT'(A11,I8)',STRING,NLOC
      IF (NLOC.GE.MAXIND) THEN
         WRITE(6,*) ' INTERFACE| Error: MAXIND not big enough!'
         WRITE(6,*) '            Increase this value in epot.inc'     
         CALL STOPGM('INTERFACE','To many QM-atoms.')
      ENDIF
C.....
C.....Loop over all local expansion      
C
      DO 30 I=1,NLOC
C...........Read position of local expansion
            READ(IUNIT,*) LOCID(I),KOLOC(1,I),KOLOC(2,I),KOLOC(3,I)
            READ(IUNIT,*) LOCEXP(1,I)
            READ(IUNIT,*) LOCEXP(2,I),LOCEXP(3,I),LOCEXP(4,I)
            READ(IUNIT,*) LOCEXP(5,I),LOCEXP(6,I),LOCEXP(7,I),
     @                    LOCEXP(8,I),LOCEXP(9,I),LOCEXP(10,I)
            READ(IUNIT,*) NINCL(I)
C.....
            IF (I.LE.3) THEN
               PRINT'(I8,TR2,F11.5,F11.5,F11.5,TR2,A,I3)',LOCID(I),
     @               KOLOC(1,I),KOLOC(2,I),KOLOC(3,I),'INCL: ',NINCL(I)
            ENDIF
            IF (NINCL(I).GE.MAXNEAR) THEN
               WRITE(6,*) ' INTERFACE| Error: MAXNEAR not big enough!'
               WRITE(6,*) '            Increase this value in epot.inc'     
               CALL STOPGM('INTERFACE','To many near atoms.')
            ENDIF
C
C...........Read the near inclusion list *********
C
            ILINES = INT(NINCL(I)/10) + 1
            DO 35 K=0,ILINES-1
               READ(IUNIT,'(A)') STRING
               DO 40 J=0,9
                  JJ = 10*K + J
                  IF (JJ.GE.NINCL(I)) GOTO 30 
                  SUBSTRING = STRING(1+J*7:J*7+7)
                  READ(SUBSTRING,*) LINCL(I,JJ+1)
                  LINCL(I,JJ+1) = LINCL(I,JJ+1) + 1
40             CONTINUE
35          CONTINUE
30    CONTINUE
C
C.....Read list of near atoms ***************************
C
      READ(IUNIT,'(A)') STRING
      READ(IUNIT,*) NNEAR
      PRINT'(A11,I5)',STRING,NNEAR
      IF (NNEAR.GE.MAXNEAR) THEN
            WRITE(6,*) ' INTERFACE| Error: MAXNEAR not big enough!'
            WRITE(6,*) '            Increase this value in epot.inc'     
            CALL STOPGM('INTERFACE','To many near atoms.')
      ENDIF
      DO 50 I=1,NNEAR
            READ(IUNIT,*) IDNEAR(I),
C....................Partial Charge
     @               DATNEAR(1,I),
C....................CutOffWidth
     @               DATNEAR(4,I),
C....................CutOffValue
     @               DATNEAR(5,I),
C....................Core charge
     @               DATNEAR(2,I),
C....................Sigma
     @               DATNEAR(3,I),
C....................the position of the atoms
     @               KONEAR(1,I),KONEAR(2,I),KONEAR(3,I)
C
            IF (DATNEAR(3,I) .LE. 0.0) THEN
               DATNEAR(2,I) = 0.0
               DATNEAR(3,I) = -DATNEAR(3,I)
            ENDIF
            IF (I.LE.3) THEN
                PRINT'(I8,TR2,F11.5,F11.5,F11.5,F11.5,F11.5,F11.5)',
     @               IDNEAR(I),
     @               DATNEAR(1,I),
     @               DATNEAR(2,I),DATNEAR(3,I),
     @               KONEAR(1,I),KONEAR(2,I),KONEAR(3,I)            
            ENDIF
C...........Multiply conversion factor from kcal/mol to Hartree
            DATNEAR(1,I) = DATNEAR(1,I)*0.001594153*332.0636*1.88972688
            DATNEAR(2,I) = DATNEAR(2,I)*0.001594153*332.0636*1.88972688
50    CONTINUE
C
C*****
      END
C*****END of SUBROUTINE RDEGOF(FNAME)
C
C
C
C     ==================================================================
      SUBROUTINE EXT_FIELD
C     ==--------------------------------------------------------------==
      IMPLICIT REAL*8 (A-H,O-Z)
      INCLUDE 'system.h'
      INCLUDE 'epot.inc'
      INCLUDE 'efld.inc'
      DIMENSION EXTF(KR1,KR2S,KR3S)
      REAL*8 K0(0:2)
      SAVE IFIRST
      DATA IFIRST /0/
C     ==--------------------------------------------------------------==
      CALL TISET(' EXT_FIELD',ISUB)
      IF(IFIRST.EQ.0) THEN
        IFIRST=1
        CALL MEMORY(IP_EXTF,KR1*KR2S*KR3S,'EXTF')
      ENDIF

      CALL AZZERO(EXTF,KR1*KR2S*KR3S)
      
      DK=(BOXDUM(6)-BOXDUM(5))/DFLOAT(NR3S)
      DJ=(BOXDUM(4)-BOXDUM(3))/DFLOAT(NR2S)
      DI=(BOXDUM(2)-BOXDUM(1))/DFLOAT(NR1S)

      DO K=1,KR3S
        K0(2)=BOXDUM(5)+DFLOAT(K-1)*DK
        DO J=1,KR2S
          K0(1)=BOXDUM(3)+DFLOAT(J-1)*DJ
          DO I=NRXPL(MEPOS,1),NRXPL(MEPOS,2)
            K0(0)=BOXDUM(1)+DFLOAT(I-1)*DI
            EXTF(I-NRXPL(MEPOS,1)+1,J,K)=CLCPOT(K0)
          ENDDO
        ENDDO
      ENDDO
      IF(PARENT) THEN
          WRITE (6,'(A)') ' INTERFACE| EXTERNAL FIELD CALCULATED '
      ENDIF
      CALL TIHALT(' EXT_FIELD',ISUB)
C     ==--------------------------------------------------------------==
      RETURN
      END
C     ==================================================================
C
C
C
C##############################################
      REAL*8 FUNCTION CLCPOT(KO)
C##############################################
      IMPLICIT REAL*8 (A-H,O-Z)
      INCLUDE 'epot.inc'
      REAL*8 KO(0:2)
C
      INTEGER I,J,LI,IT
      REAL*8 DX,DY,DZ,ADDEPOT
      REAL*8 DMIN,D,DKO(0:2)        
C
C.... find the nearest local expansion ********
C
      CLCPOT = 0.0
      DMIN   = 999999.0
      LI     = -1
C
      IF (NLOC .EQ. 0) GOTO 150
C
      DO 100 I=1,NLOC
             DX=KO(0)-KOLOC(1,I)
             DY=KO(1)-KOLOC(2,I)
             DZ=KO(2)-KOLOC(3,I)
             D=DX*DX+DY*DY+DZ*DZ
             IF (D .LT. DMIN) THEN
                DMIN=D
                LI=I
             END IF
100   CONTINUE
C
C.... evaluate potential from local expansion *****
C
      DX=KO(0)-KOLOC(1,LI)
      DY=KO(1)-KOLOC(2,LI)
      DZ=KO(2)-KOLOC(3,LI)
C
      CLCPOT = -( LOCEXP(1,LI) +	
     @         DX*LOCEXP(2,LI) + 
     @         DY*LOCEXP(3,LI) + 
     @         DZ*LOCEXP(4,LI) +
     @  0.5*DX*DX*LOCEXP(5,LI) +
     @  0.5*DY*DY*LOCEXP(6,LI) +
     @  0.5*DZ*DZ*LOCEXP(7,LI) +
     @      DX*DY*LOCEXP(8,LI) +
     @      DX*DZ*LOCEXP(9,LI) +
     @      DY*DZ*LOCEXP(10,LI) )
C
150   CONTINUE
C
      IF (NINCL(LI) .EQ. 0) GOTO 300
C
C.....add El.Pot from all near atoms      
C
      DO 200 I=1,NINCL(LI)
             IT = LINCL(LI,I)
C
             DX=KO(0)-KONEAR(1,IT)
             DY=KO(1)-KONEAR(2,IT)
             DZ=KO(2)-KONEAR(3,IT)

C............Distance in Bohr
             d = SQRT(DX*DX + DY*DY + DZ*DZ)
             IF (d .LT. 0.1) d = 0.1
C
             IF (d .GT. 6.0) THEN
C               If we a far away take simple Coulumb-Potential
C               Phi = q / d
                ADDEPOT = DATNEAR(1,IT) / d
             ELSE IF (DATNEAR(2,IT) .GT. 0.5) THEN
C               If core-charges are provided use the more
C               advanced atom-potential.
                DY=MYRAG(NINT(DATNEAR(2,IT)))               	
                DX = (DATNEAR(2,IT) - DATNEAR(1,IT))*
     *               (1.0-exp(-d/DATNEAR(3,IT)))                
                ADDEPOT = DERF(d/DY)*(DATNEAR(2,IT)-DX)/d
             ELSE
C               Use the simple atom-potential given 
C               just by partial charges.            
C.              Phi = q * erf(d/sigma) / d
                DX = DERF(d/DATNEAR(3,IT))
                ADDEPOT = DATNEAR(1,IT) * DX / d                          
             ENDIF
                                     
C............we have an excluded atom here
             IF (DATNEAR(4,IT) .GT. 0.01) THEN
                IF (d .GT. (DATNEAR(5,IT)+DATNEAR(4,IT))) THEN
C                   do nothing in that case
C                   because we are far away from it 
                ELSE IF (d .LT. DATNEAR(5,IT)) THEN
C                   we are to close, dont add any contribution from it
                    ADDEPOT = 0.0
                ELSE            
C                   Use a switching function: No typo here!
                    DX = (d - DATNEAR(5,IT)) / DATNEAR(4,IT)
                    DX = 1.0 - DX*DX                                    
                    ADDEPOT = (1.0 - DX*DX) * ADDEPOT
                ENDIF
             ENDIF
                         
             CLCPOT = CLCPOT - ADDEPOT
200   CONTINUE
      IF (CLCPOT .LT. -0.3) CLCPOT = -0.3 - log(0.7-CLCPOT)
C
300   CONTINUE
C
      END

C
C
C
C     ==================================================================
      SUBROUTINE INTERFACE_WAIT(INTFILEN)
C     ==--------------------------------------------------------------==
      IMPLICIT REAL*8 (A-H,O-Z)
      INCLUDE 'system.h'
      LOGICAL EOFRUN,TEST
      CHARACTER FILEN*80,INTFILEN*(*)
C     ==--------------------------------------------------------------==
      CALL TISET('INTER_WAIT',ISUB)
      IF(PARENT) THEN
        WRITE(6,'(/,1X,63(1H=))')
        FILEN=FPATH(IAPATH:IEPATH)//'QMCONTINUE'
        WRITE(6,*) ' INTERFACE| WAIT FOR CONTINUE-FILE ', FILEN
    1   CONTINUE
        INQUIRE(FILE=FILEN,EXIST=EOFRUN)
        IF(EOFRUN) THEN
          INTFILEN=FPATH(IAPATH:IEPATH)//'qmoutput.out'
C.......Now delete the QMCONTINUE file
          OPEN(5,FILE=FILEN,STATUS='UNKNOWN')
          CLOSE(5,STATUS='DELETE')
          WRITE(6,*) ' INTERFACE| CONTINUE QM CALCULATION ',INTFILEN
          GOTO 3
        ENDIF
        INQUIRE(FILE='EXIT',EXIST=TEST)
        IF(TEST) GOTO 3
#ifdef IBM
        CALL SLEEP_(1)
#else
        CALL SLEEP(1)
#endif
        GOTO 1
    3   CONTINUE
      ENDIF
      CALL MY_SYNC(ALLGRP)
      CALL TIHALT('INTER_WAIT',ISUB)
C     ==--------------------------------------------------------------==
      RETURN
      END
C     ==================================================================
C
C
C
C     ==================================================================
      SUBROUTINE INTERFACE_WRITE(INTFILEN,TAU0,FION,C0,ACHRG,SCR,LSCR)
C     ==--------------------------------------------------------------==
      IMPLICIT REAL*8 (A-H,O-Z)
      INCLUDE 'system.h'
      INCLUDE 'elct.inc'
      INCLUDE 'epot.inc'
      INCLUDE 'efld.inc'
      INCLUDE 'ions.inc'
      INCLUDE 'ener.inc'
      INCLUDE 'prop.inc'
      INCLUDE 'dipo.inc'
      INCLUDE 'lodp.inc'
      INCLUDE 'atwf.inc'
      INCLUDE 'geq0.inc'
      INCLUDE 'isos.inc'
      INCLUDE 'cppt.inc'
      INCLUDE 'irat.inc'
      INCLUDE 'fft.inc'
      INCLUDE 'fpoint.inc'
      CHARACTER LINE*80,STR*80,FILEN*80,INTFILEN*(*)
      REAL*8 K0(0:2),K1(0:2),DX,DY,DZ,EF(0:2),EEF
      COMPLEX*16 C0(*),EIVPS(*),EIROP(*),V(*),VTEMP(*)
      DIMENSION TAU0(3,NAX,*),FION(3,NAX,*),SCR(*),ACHRG(*),RHOE(*),
     *          QPHI(*)
      DIMENSION ISEL(*),EFIELD(*)
      POINTER (IP_ISEL,ISEL),(IP_EFIELD,EFIELD)
C     ==--------------------------------------------------------------==
      CALL AZZERO(ACHRG,NAT)
C
C..calculate atomic charges
C
      IF(TEXTFLD) THEN
        IF(ICMET.EQ.1) THEN
C..Mulliken Charges
          CALL SETBASIS
          NATTOT=0
          DO IS=1,NSP
            NUMAOR(IS)=0
            DO ISH=1,NSHELL(IS)
              L=LSHELL(ISH,IS)
              NATTOT=NATTOT+NA(IS)*(2*L+1)
              NUMAOR(IS)=NUMAOR(IS)+(2*L+1)
            ENDDO
          ENDDO
          CALL MULLIKEN(1,C0,NSTATE,TAU0,ACHRG,SCR,LSCR)
        ELSEIF(ICMET.EQ.4) THEN
C..Lowdin Charges
          CALL SETBASIS
          NATTOT=0
          DO IS=1,NSP
            NUMAOR(IS)=0
            DO ISH=1,NSHELL(IS)
              L=LSHELL(ISH,IS)
              NATTOT=NATTOT+NA(IS)*(2*L+1)
              NUMAOR(IS)=NUMAOR(IS)+(2*L+1)
            ENDDO
          ENDDO
          CALL MULLIKEN(2,C0,NSTATE,TAU0,ACHRG,SCR,LSCR)
        ELSEIF(ICMET.EQ.2) THEN
C..Hirshfeld Charges
          CALL SETFPT(SCR)
          CALL RHOOFR(C0,TAU0,N)
          CALL ATOMC(TAU0,ACHRG)
        ELSEIF(ICMET.EQ.3) THEN
C..ESP Charges
          CALL SETFPT(SCR)
          CALL RHOOFR(C0,TAU0,N)
          CALL EICALC(EIVPS,EIROP)
          DO I=1,NNR1
            V(I)=DCMPLX(RHOE(I),0.D0)
          ENDDO
          CALL FWFFT(V)
          DO IG=1,NHG
            VTEMP(IG) = V(NZH(IG))
          ENDDO
          CALL ELSTPO(VTEMP,EIROP,EIVPS,V,QPHI,TAU0)
          CALL DCOPY(2*NHG,V,1,VTEMP,1)
          CALL AZZERO(V,2*MAXFFT)
          DO IG=1,NHG
            V(INDZ(IG)) = DCONJG(VTEMP(IG))
            V(NZH(IG))  = VTEMP(IG)
          ENDDO
          IF(GEQ0.AND.TCLUST) THEN
            V(NZH(1)) = VTEMP(1)
          ELSEIF(GEQ0) THEN
            V(NZH(1)) = DCMPLX(0.D0,0.D0)
          ENDIF
          CALL INVFFT(V)
          DO I=1,NNR1
            RHOE(I)=DREAL(V(I))
          ENDDO
          LEN=NNR1/IRAT + 1
          CALL MEMORY(IP_ISEL,LEN,'ISEL')
          CALL SELECTP(ISEL,TAU0,IPPC)
          LEN=(NAT+1)*IPPC
          CALL MEMORY(IP_EFIELD,LEN,'EFIELD')
          DO I=1,IPPC
            EFIELD(I)=RHOE(ISEL(I))
          ENDDO
          CALL ATFIELD(EFIELD,V,VTEMP,EIROP,QPHI,ISEL,IPPC)
          CALL ESPSOLV(EFIELD,ACHRG,IPPC)
          CALL FREEM(IP_ISEL)
          CALL FREEM(IP_EFIELD)
        ELSE
          CALL STOPGM('INTERFACE_WRITE','INVALID ICMET')
        ENDIF
      ENDIF
C
C..writeout all data to interface-file
C
      IF(PARENT) THEN
        OPEN(5,FILE=INTFILEN,STATUS='UNKNOWN')
        REWIND 5
        IAT=0
        WRITE(5,'(G15.6,TR1,G15.6)') ETOT,EEXT
        DO IS=1,NSP
          DO IA=1,NA(IS)
            IAT=IAT+1
            WRITE(5,'(G15.6,TR1,G15.6,TR1,G15.6,TR1,G15.6)')
     *      FION(1,IA,IS),FION(2,IA,IS),FION(3,IA,IS),ACHRG(IAT)
          ENDDO
        ENDDO
C
C..Calculate mean E-field and writeout dipole-moment
C
      EF(0)=0.0
      EF(1)=0.0
      EF(2)=0.0
      EEF=0.0
      IF(TEXTFLD) THEN
        DZ=BOXDUM(6)-BOXDUM(5)
        DY=BOXDUM(4)-BOXDUM(3)
        DX=BOXDUM(2)-BOXDUM(1)
C..E-Field X-Komponente
        K0(0) = BOXDUM(1) + DX/3.0
        K0(1) = BOXDUM(3) + DY/2.0
        K0(2) = BOXDUM(5) + DZ/2.0
        K1(0) = BOXDUM(1) + 2.0*DX/3.0
        K1(1) = BOXDUM(3) + DY/2.0
        K1(2) = BOXDUM(5) + DZ/2.0
        EF(0) = 3.0*(CLCPOT(K1) - CLCPOT(K0))/DX
C..E-Field Y-Komponente
        K0(0) = BOXDUM(1) + DX/2.0
        K0(1) = BOXDUM(3) + DY/3.0
        K0(2) = BOXDUM(5) + DZ/2.0
        K1(0) = BOXDUM(1) + DX/2.0
        K1(1) = BOXDUM(3) + 2.0*DY/3.0
        K1(2) = BOXDUM(5) + DZ/2.0
        EF(1) = 3.0*(CLCPOT(K1) - CLCPOT(K0))/DY
C..E-Field Z-Komponente
        K0(0) = BOXDUM(1) + DX/2.0
        K0(1) = BOXDUM(3) + DY/2.0
        K0(2) = BOXDUM(5) + DZ/3.0
        K1(0) = BOXDUM(1) + DX/2.0
        K1(1) = BOXDUM(3) + DY/2.0
        K1(2) = BOXDUM(5) + 2.0*DZ/3.0
        EF(2) = 3.0*(CLCPOT(K1) - CLCPOT(K0))/DZ
        EEF   = sqrt(EF(0)*EF(0) + EF(1)*EF(1) + EF(2)*EF(2))                    
        ENDIF
C..Dipole Moment
        IF(LOCD) THEN
           DEB=2.541765D0
           D1=-DEB*DMOMLO(1,1)
           D2=-DEB*DMOMLO(2,1)
           D3=-DEB*DMOMLO(3,1)
           DD=SQRT(D1*D1+D2*D2+D3*D3)
        ELSE IF(LDIP) THEN
           DEB=2.541765D0
           D1=-DEB*DMOM(1)
           D2=-DEB*DMOM(2)
           D3=-DEB*DMOM(3)
           DD=SQRT(D1*D1+D2*D2+D3*D3)
        ELSE
           D1=-99999.9
           D2=-99999.9
           D3=-99999.9
           DD=-99999.9        
        ENDIF
C..Write it to file
        WRITE(5,'(F12.5,TR1,F12.5,TR1,F12.5,TR1,F12.5,A)')
     @            EF(0),EF(1),EF(2),EEF,' E-Field'
        WRITE(5,'(F12.5,TR1,F12.5,TR1,F12.5,TR1,F12.5,A)')
     @            D1,D2,D3,DD,' Debye'
        CLOSE(5)
C
C..Create the EGOCONTINUE file
C
        IF(IFTYPE.EQ.1) THEN
          STR=FPATH(IAPATH:IEPATH)//'EGOCONTINUE'
          CALL XSTRING(STR,IA,IE)
          LINE=STR(IA:IE)
          OPEN(5,FILE=LINE,STATUS='UNKNOWN')
          REWIND 5
          WRITE(5,'(A)') 'EGOCONTINUE'
          CLOSE(5)
        ELSE
          CALL STOPGM('INTERFACE','NOT PROGRAMMED')
        ENDIF
        CALL WRGEOF(TAU0,FION)
        WRITE(6,*) ' INTERFACE| FORCES WRITTEN TO FILE '
      ENDIF
      CALL MY_SYNC(ALLGRP)
C     ==--------------------------------------------------------------==
      RETURN
      END
C     ==================================================================
C
C
C
C     ==================================================================
      SUBROUTINE EGOSTOP
C     ==--------------------------------------------------------------==
      IMPLICIT REAL*8 (A-H,O-Z)
      INCLUDE 'system.h'
      
      CHARACTER LINE*80,STR*80

      IF(IFTYPE.EQ.1) THEN
          WRITE(6,'(//,A)') ' Creating EGOEXIT file'
          STR=FPATH(IAPATH:IEPATH)//'EGOEXIT'
          CALL XSTRING(STR,IA,IE)
          LINE=STR(IA:IE)
          OPEN(5,FILE=LINE,STATUS='UNKNOWN')
          REWIND 5
          WRITE(5,'(A)') 'EGOEXIT'
          CLOSE(5)
      ENDIF
      
      RETURN
      END


C     ==================================================================
      SUBROUTINE MYPROPPT(C0,TAU0,SCR)
C     ==--------------------------------------------------------------==
      IMPLICIT REAL*8 (A-H,O-Z)
      INCLUDE 'system.h'
      INCLUDE 'spin.inc'
      INCLUDE 'elct.inc'
      INCLUDE 'prop.inc'
      INCLUDE 'soft.inc'
      INCLUDE 'sfac.inc'
      INCLUDE 'ions.inc'
      INCLUDE 'pslo.inc'
      INCLUDE 'lodp.inc'
      INCLUDE 'atwf.inc'
      INCLUDE 'cppt.inc'
      INCLUDE 'geq0.inc'
      INCLUDE 'irat.inc'
      INCLUDE 'isos.inc'
      INCLUDE 'fft.inc'
      INCLUDE 'fpoint.inc'
      INCLUDE 'epot.inc'
      COMPLEX*16 C0(NGW,*),EIVPS(*),EIROP(*),V(*)
      DIMENSION TAU0(*),SCR(*)
      PARAMETER (NACC=7)
      DIMENSION ACC(NACC),RHOE(*),FOCC(*),IREC(100)
      DATA IREC /1,1,1,1,0,0,1,1,1,0,90*0/
      POINTER (IP_FBACK,FBACK(*))
C     ==--------------------------------------------------------------==
      IF(PARENT) THEN
        WRITE(6,'(/,1X,63(1H*))')
        WRITE(6,'(" *",9X,A,19X,"*")') 
     @                  ' INTERFACE| PROPERTY CALCULATIONS'
        WRITE(6,'(1X,63(1H*),/)')
      ENDIF
      CALL PROPIN(INSYS)
C     ==--------------------------------------------------------------==
      NUMORB = MAX(N,NUMORB)
C     ==--------------------------------------------------------------==
C     ==  CALCULATE DIPOLE MOMENTS                                    ==
C     ==--------------------------------------------------------------==
      IF(LDIP.OR.LOCD.OR.LEXT) THEN
        CALL FORCEP(NUMORB,IPTOP)
        CALL PHFAC(TAU0)
        CALL RNLSM1(C0,SCR,NUMORB,.FALSE.)
        CALL SETFPT(SCR)
        IF(LEXT) THEN
          CALL MEMORY(IP_FBACK,NUMORB,'FBACK')
          CALL DCOPY(NUMORB,F,1,FBACK,1)
          DO I=1,NEXDIP
            DO J=1,NUMORB
              JI=(I-1)*NUMORB+J
              F(J)=FBACK(J)-FOCC(JI)
            ENDDO
            CALL DIFRHO(C0,TAU0,NUMORB)
            CALL AZZERO(EIROP,2*NHG)
            DO IR=1,NNR1
              V(IR)=DCMPLX(RHOE(IR),0.D0)
            ENDDO
            CALL FWFFT(V)
            CALL EXDIPO(I,TAU0,V)
          ENDDO
          CALL DCOPY(NUMORB,FBACK,1,F,1)
        ENDIF
        IF(LDIP.OR.LOCD) THEN
          CALL RHOOFR(C0,TAU0,N)
          CALL EICALC(EIVPS,EIROP)
          DO IR=1,NNR1
            V(IR)=DCMPLX(RHOE(IR),0.D0)
          ENDDO
          CALL FWFFT(V)
          IF(LDIP) CALL DIPO(TAU0,EIROP,V)
          IF(LOCD) CALL LODIPO(EIROP,V)
        ENDIF
        IF(PARENT) THEN
          CALL PRDIP
        ENDIF
      ENDIF
C     ==--------------------------------------------------------------==
      IF(TMEMCHK) CALL PRINT_MEMORY
C     ==--------------------------------------------------------------==
      RETURN
      END
      
