/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

#define DEBUG		0
#define INFO		0
#define REALLOCINFO	1
#define SUMINFO		1

#include <stdio.h>
#include <string.h>
#if defined(PARIX) || defined(PARIXHAL) || defined(MPI) || defined(SEQ)

#else
#include "pvm3.h"
#endif

#include "ego.h"
#include "egalloc.h"
#include "egbolib.h"
#include "wodecls.h"
#include "util_lib.h"
#include "dbmalloc.h"	/*** include always as last include file ***/

static void	oom(void *ptr,int size);
static int	my_number;
static long	now_allocated=0L;	
static char	allocinfostr[200];
	
/***********************************************************************/
/***********************************************************************/
void   malloc_init (int a)
{
#if DEBUGMALLOC
	ini_mallocdebug(1,2,0,8,8,0,a); /*** we use debuglevel 1 here     ***/
					/*** this does not take much time ***/
					/*** as in EGO there are no mallocs**/
					/*** in inner loops!		  ***/
#endif

#ifdef NEXT
	malloc_debug(0);	/* for debugging use 16, normal is 0 */
#endif
	my_number = a;		/* so that we have it for debugging purposes */
	strcpy(allocinfostr,"XXX");
	now_allocated=0L;
	return;
}

int _PX_CheckMem(int a)
{	a=0;
	return 0;
}

void	set_allocinfo(char *str)
{
	strcpy(allocinfostr,str);
}

void	print_meminfo()
{
#if SUMINFO
	txtout("Node(ALC) %2d: until now allocated %ld\n",
	my_number,now_allocated);
	
#endif
}

/***********************************************************************/
/***********************************************************************/
/***********************************************************************/
void  *merealloc(void *oldp, int news, int olds)
{
	void  *p;
#if INFO
	txtout("Node(ALC) %2d: realloc [%-10s] oldp %p  olds %d  news %d ...",
		my_number,allocinfostr,oldp,olds,news);
	
#endif

	if ((news<=0) || (olds<=0)) {
		errout("Node(ALC) %2d: error in merealloc news %d olds %d\n",
			my_number,news,olds);
		pex(-1);
	}

	p = (void*) realloc(oldp,(size_t)news);
	oom(p,news);
	now_allocated += news;
	now_allocated -= olds;
/********
	p = (void *)malloc((size_t)news);
	oom(p,news);
	memcpy (p, oldp, olds);
	myfree(oldp);
*********/

#if INFO
	txtout(" newp %p\n",p);
#endif
	return (void*) p;
}

/***********************************************************************/
static void oom (void *ptr,int news)
{				/* Out Of Memory */
	void   pex(int);

	if (ptr==NULL) {
		errout("Node(ALC) %2d: Out of memory (%d)! EGO stops.\n",
			my_number,news);
		errout("Node(ALC) %2d: Current info-string [%s]\n",
			my_number,allocinfostr);	
		errout("Node(ALC) %2d: until now allocated %ld\n",
			my_number,now_allocated);
		pex(-1);
	}
	return;
}

/***********************************************************************/
TRAVELLING_PACKETS *alloc_packets(int len)
{
	TRAVELLING_PACKETS *ptr;
	size_t	nobytes;
 
	if (len<1)	len = 1;
	nobytes = (size_t) (len * sizeof(TRAVELLING_PACKETS));
#if INFO
	txtout("Node(ALC) %2d: alloc_packets [%-10s] for %d x %d = %d ...",
		my_number,allocinfostr,len,sizeof(TRAVELLING_PACKETS),nobytes);
	
#endif
	ptr = (TRAVELLING_PACKETS *) malloc(nobytes);
	oom(ptr,nobytes);
	now_allocated += nobytes;
#if INFO
	txtout("  newp %p\n",ptr);
#endif
	return (ptr);
}
/***********************************************************************/
struct_cluster *alloc_cluster(int len)
{
	struct_cluster   *ptr;
	size_t	nobytes;

	if (len < 1)	len = 1;
	nobytes = (size_t) (len * sizeof(struct_cluster));
#if INFO
	txtout("Node(ALC) %2d: alloc_cluster [%-10s] for %d x %d = %d ...",
		my_number,allocinfostr,len,sizeof(struct_cluster),nobytes);
	
#endif

	ptr = (struct_cluster *) malloc(nobytes);
	oom(ptr,nobytes);
	now_allocated += nobytes;
#if INFO
	txtout("  newp %p\n",ptr);
#endif
	return (ptr);
}

/***********************************************************************/
struct_cluster_reduced *alloc_cluster_reduced(int len)
{
	struct_cluster_reduced   *ptr;
	size_t	nobytes;

	if (len<1)	len = 1;
	nobytes = (size_t) (len * sizeof(struct_cluster_reduced));
#if INFO
	txtout("Node(ALC) %2d: alloc_cluster_reduced [%-10s] for %d x %d = %d ...",
		my_number,allocinfostr,len,
		sizeof(struct_cluster_reduced),nobytes);
#endif

	ptr = (struct_cluster_reduced *) malloc(nobytes);
	oom(ptr,nobytes);
	now_allocated += nobytes;
#if INFO
	txtout("  newp %p\n",ptr);
#endif
	return (ptr);
}

/***********************************************************************/
STRUCT_ATOMDATA *alloc_atomdata (int len)
{
	STRUCT_ATOMDATA   *ptr;
	size_t	nobytes;

	if (len < 1)	len = 1;
	nobytes = (size_t) (len * sizeof(STRUCT_ATOMDATA));
#if INFO
	txtout("Node(ALC) %2d: alloc_atomdata [%-10s] for %d x %d = %d ...",
		my_number,allocinfostr,len,sizeof(STRUCT_ATOMDATA),nobytes);
	
#endif

	ptr = (STRUCT_ATOMDATA *)malloc(nobytes);
	oom(ptr,nobytes);
	now_allocated += nobytes;
#if INFO
	txtout("  newp %p\n",ptr);
#endif
	return (ptr);
}

/***********************************************************************/
int   *alloc_ints(int len)
{
	int   *ptr;
	size_t	nobytes;

	if (len < 1)	len = 1;
	nobytes = (size_t) (len * sizeof(int));
#if INFO
	txtout("Node(ALC) %2d: alloc_ints [%-10s] for %d x %d = %d ...",
		my_number,allocinfostr,len,sizeof(int),nobytes);
	
#endif
	ptr = (int *)malloc(nobytes);
	oom(ptr,nobytes);
	now_allocated += nobytes;
#if INFO
	txtout("  newp %p\n",ptr);
#endif
	return (ptr);
}


/***********************************************************************/
void   *alloc_blk(size_t nobytes)
{
	void   *ptr;

#if INFO
	txtout("Node(ALC) %2d: alloc_ints [%-10s] for %d x %d = %d ...",
		my_number,allocinfostr,len,sizeof(int),nobytes);
	
#endif
	ptr = (void *)malloc(nobytes);
	oom(ptr,nobytes);
	now_allocated += nobytes;
#if INFO
	txtout("  newp %p\n",ptr);
#endif
	return (ptr);
}


/***********************************************************************/
int   **alloc_ptrs(int len)
{
	int   **ptr;
	size_t	nobytes;

	if (len < 1)	len = 1;
	nobytes = (size_t) (len * sizeof(int*));
#if INFO
	txtout("Node(ALC) %2d: alloc_ptrs [%-10s] for %d x %d = %d ...",
		my_number,allocinfostr,len,sizeof(int*),nobytes);
	
#endif
	ptr = (int **)malloc(nobytes);
	oom(ptr,nobytes);
	now_allocated += nobytes;
#if INFO
	txtout("  newp %p\n",ptr);
#endif
	return (ptr);
}


/***********************************************************************/
void   realloc_ints(int **old_ptr, int len, int old_len)
{

	int   *ptr;

	/* we ADD memory of len integers to the old block */
	if (len < 1)	len = 1;

#if REALLOCINFO > 100
	txtout ("%d ints: len=%d\n", my_number, len + old_len);
#endif

	ptr = (int *)merealloc ((void *)*old_ptr,
			  	((size_t) (sizeof (int) * (len + old_len))),
			  	((size_t) (sizeof (int) * (old_len))));
	*old_ptr = ptr;
	return;
}


/***********************************************************************/
char  *alloc_chars (int len)
{
	char  *ptr;
	size_t	nobytes;

	if (len<1) len = 1;
	nobytes = (size_t) (len * sizeof(char));
#if INFO
	txtout("Node(ALC) %2d: alloc_chars [%-10s] for %d x %d = %d ...",
		my_number,allocinfostr,len,sizeof(char),nobytes);
	
#endif

	ptr = (char *)malloc(nobytes);
	oom (ptr,nobytes);
	now_allocated += nobytes;
#if INFO
	txtout("  newp %p\n",ptr);
#endif
	return (ptr);
}


/***********************************************************************/
void   realloc_chars(char **old_ptr, int len, int old_len)
{

	char  *ptr;

	/* we ADD memory of len integers to the old block */
	if (len < 1) len = 1;

#if REALLOCINFO > 100
	txtout ("Node %d: in realloc chars %p %d chars: len=%d\n",
		my_number,*old_ptr,len + old_len,len);
#endif

	ptr = (char *)merealloc ((void *)(*old_ptr),
			   	((size_t) (sizeof(char) * (len + old_len))),
			   	((size_t) (sizeof(char) * old_len)));

	oom(ptr,sizeof(char) * (len + old_len));
	*old_ptr = ptr;

	return;
}


/***********************************************************************/
double *alloc_doubles(int len)
{
	double *ptr;
	size_t	nobytes;

	if (len<1) len = 1;
	nobytes = (size_t) (len * sizeof(double));
#if INFO
	txtout("Node(ALC) %2d: alloc_doubles [%-10s] for %d x %d = %d ...",
		my_number,allocinfostr,len,sizeof(double),nobytes);
	
#endif
	ptr = (double *)malloc(nobytes);
	oom (ptr,nobytes);
	now_allocated += nobytes;
#if INFO
	txtout("  newp %p\n",ptr);
#endif
	return (ptr);
}

/***********************************************************************/
float *alloc_floats(int len)
{
	float *ptr;
	size_t	nobytes;

	if (len<1) len = 1;
	nobytes = (size_t) (len * sizeof(float));
#if INFO
	txtout("Node(ALC) %2d: alloc_floats [%-10s] for %d x %d = %d ...",
		my_number,allocinfostr,len,sizeof(float),nobytes);
	
#endif
	ptr = (float *)malloc(nobytes);
	oom(ptr,nobytes);
	now_allocated += nobytes;
#if INFO
	txtout("  newp %p\n",ptr);
#endif
	return (ptr);
}


/***********************************************************************/
void   realloc_doubles(double **old_ptr, int increase_by, int old_len)
{

	double *ptr;

	/* we ADD memory of increase_by doubles to the old block */
	if (increase_by < 1) increase_by = 1;

#if REALLOCINFO > 100
	txtout ("%d doubles: increase_by=%d\n", my_number, increase_by + old_len);
#endif

	ptr = (double *)merealloc ((void *)*old_ptr,
			 	    (size_t) (sizeof (double) * (increase_by + old_len)),
			   	    (size_t) (sizeof (double) * (old_len)));

	oom (ptr,sizeof (double) * (increase_by + old_len));
	*old_ptr = ptr;
	return;
}


/***********************************************************************/
EXT_3_ATOM *alloc_EXT_3_ATOMs (int len)
{
	EXT_3_ATOM *ptr;
	size_t	nobytes;

	if (len<1) len = 1;
	nobytes = (size_t) (len * sizeof(EXT_3_ATOM));
#if INFO
	txtout("Node(ALC) %2d: alloc_EXT_3_ATOMs [%-10s] for %d x %d = %d ...",
		my_number,allocinfostr,len,sizeof(EXT_3_ATOM),nobytes);
	
#endif
	ptr = (EXT_3_ATOM *)malloc(nobytes);
	oom(ptr,nobytes);
	now_allocated += nobytes;
#if INFO
	txtout("  newp %p\n",ptr);
#endif
	return (ptr);
}

/***********************************************************************/
EXT_4_ATOM *alloc_EXT_4_ATOMs (int len)
{
	EXT_4_ATOM *ptr;
	size_t	nobytes;
	if (len<1) len = 1;
	nobytes = (size_t) (len * sizeof(EXT_4_ATOM));

#if INFO
	txtout("Node(ALC) %2d: alloc_EXT_4_ATOMs [%-10s] for %d x %d = %d ...",
		my_number,allocinfostr,len,sizeof(EXT_4_ATOM),nobytes);
	
#endif
	ptr = (EXT_4_ATOM *) malloc(nobytes);
	oom (ptr,nobytes);
	now_allocated += nobytes;
#if INFO
	txtout("  newp %p\n",ptr);
#endif
	return (ptr);
}

/***********************************************************************/
char **alloc_strings(int len)
{
  	char **ptr;
  	int    i,nobytes;

  	/* First we get an array of pointers to chars */
	if (len < 1) len = 1;
	nobytes = (size_t) (len * sizeof(char*));
#if INFO
	txtout("Node(ALC) %2d: alloc_strings [%-10s] for %d x (%d+%d) = %d ...",
		my_number,allocinfostr,len,sizeof(char*),MAX_STRING_SIZE,
		nobytes+len*MAX_STRING_SIZE);
	
#endif
	ptr = (char **)malloc(nobytes);
	oom(ptr,nobytes);
	now_allocated += nobytes;

  	/* now we can allocate the real memory */
	for (i = 0; i < len; i++) {
		ptr[i] = (char *)malloc ((size_t) (sizeof(char) * MAX_STRING_SIZE));
		oom(ptr,sizeof(char) * MAX_STRING_SIZE);
	}
#if INFO
	txtout("  newp %p\n",ptr);
#endif
	return (ptr);
}

#if defined(PARIX) || defined(PARIXHAL) || defined(MPI) || defined(SEQ)

#else
/***********************************************************************/
struct pvmhostinfo **alloc_pvmhostinfo (int len)
{
  	struct pvmhostinfo **ptr;
  	int    i,nobytes;

  	if (len<1) len = 1;
	nobytes = (size_t) (sizeof(struct pvmhostinfo *) * len);
#if INFO
	txtout("Node(ALC) %2d: alloc_pvmhostinfo [%-10s] for %d x (%d + %d) = %d ...",
		my_number,allocinfostr,len,sizeof(struct pvmhostinfo *),
		sizeof(struct pvmhostinfo),nobytes+len*sizeof(struct pvmhostinfo));
	
#endif
  	/* First we get an array of pointers to struct pvmhostinfo: */
 	ptr = (struct pvmhostinfo **) malloc(nobytes);
	oom(ptr,nobytes);

  	/* now we can allocate the real memory */
  	for (i = 0; i < len; i++) {
    		ptr[i] = (struct pvmhostinfo *) malloc((size_t) (sizeof(struct pvmhostinfo)));
		oom(ptr[i],sizeof(struct pvmhostinfo));
  	}
#if INFO
	txtout("  newp %p\n",ptr);
#endif
  	return (ptr);
}

/***********************************************************************/
void	free_pvmhostinfo(struct pvmhostinfo **ptr,int len)
{	
	int i;
	if (len < 1) len = 1;
	for (i=0;i<len;i++) {
		myfree(ptr[i]);	
	}	
	myfree(ptr);
}

/***********************************************************************/
struct pvmtaskinfo **alloc_pvmtaskinfo (int len)
{
  	struct pvmtaskinfo **ptr;
  	int    i,nobytes;

  	if (len<1) len = 1;
	nobytes = (size_t) (sizeof(struct pvmtaskinfo *) * len);
#if INFO
	txtout("Node(ALC) %2d: alloc_pvmtaskinfo [%-10s] for %d x (%d + %d) = %d ...",
		my_number,allocinfostr,len,sizeof(struct pvmtaskinfo *),
		sizeof(struct pvmtaskinfo),nobytes+len*sizeof(struct pvmtaskinfo));
	
#endif
  	/* First we get an array of pointers to struct pvmtaskinfo: */
  	ptr = (struct pvmtaskinfo **) malloc(nobytes);
	oom(ptr,nobytes);

  	/* now we can allocate the real memory */
  	for (i = 0; i < len; i++) {
    		ptr[i] = (struct pvmtaskinfo *) malloc ((size_t) (sizeof (struct pvmtaskinfo)));
		oom (ptr[i],sizeof (struct pvmtaskinfo));
  	}
#if INFO
	txtout("  newp %p\n",ptr);
#endif
  	return (ptr);
}

/***********************************************************************/
void	free_pvmtaskinfo(struct pvmtaskinfo **ptr,int len)
{	
	int i;
	if (len < 1) len = 1;
	for (i=0;i<len;i++) {
		myfree(ptr[i]);	
	}	
	myfree(ptr);
}
#endif

/***********************************************************************/
int   *alloc_listsize(int len)
{
	int   *ptr;

	ptr = alloc_ints(len);
	return (ptr);
}

/***********************************************************************/
int   *alloc_list (int *len)
{
	int   *ptr;

	ptr = alloc_ints(CHUNK_BOND_EXCL);
	*len = CHUNK_BOND_EXCL;
	return (ptr);
}

/***********************************************************************/
void   realloc_list(int **old_ptr, int *len)
{
	realloc_ints(old_ptr, CHUNK_BOND_EXCL, *len);
	*len += CHUNK_BOND_EXCL;
	return;
}

/***********************************************************************/
void   alloc_TRAVELLING_FORCES (TRAVELLING_FORCES *force)
{

	force->Fx = alloc_doubles(force->no_atoms);
	force->Fy = alloc_doubles(force->no_atoms);
	force->Fz = alloc_doubles(force->no_atoms);
	force->len= force->no_atoms;
	return;
}

/***********************************************************************/
void   realloc_TRAVELLING_FORCES (TRAVELLING_FORCES *forces, int new_len)
{

	if (new_len>forces->len) {
		realloc_doubles (&(forces->Fx), (new_len - forces->len),forces->len);
		realloc_doubles (&(forces->Fy), (new_len - forces->len),forces->len);
		realloc_doubles (&(forces->Fz), (new_len - forces->len),forces->len);
		forces->len = new_len;
	}
	return;
}

/***********************************************************************/
void   free_TRAVELLING_FORCES (TRAVELLING_FORCES *force)
{
	myfree(force->Fx);
	myfree(force->Fy);
	myfree(force->Fz);
	force->len = 0;
}

/***********************************************************************/
void   alloc_TRAVELLING_COORDINATES (TRAVELLING_COORDINATES *coord)
{
	coord->x = alloc_doubles(coord->no_atoms);
	coord->y = alloc_doubles(coord->no_atoms);
	coord->z = alloc_doubles(coord->no_atoms);
	coord->q = alloc_doubles(coord->no_atoms);
	coord->typ 	      = alloc_ints(coord->no_atoms);
	coord->atom_id	      = alloc_ints(coord->no_atoms);
	coord->atom_sele_flag1 = alloc_ints(coord->no_atoms);
	coord->atom_sele_flag2 = alloc_ints(coord->no_atoms);
	coord->len = coord->no_atoms;
#if DEBUG
	errout("Node %d: in alloc_TRAVELLING_COORDINATES before return\n",
	        my_number);
#endif
	return;
}

/************************************************************************/
void   realloc_TRAVELLING_COORDINATES(TRAVELLING_COORDINATES *coord,int new_len)
{
	if (new_len>coord->len) {
		realloc_doubles(&(coord->x),(new_len - coord->len),coord->len);
		realloc_doubles(&(coord->y),(new_len - coord->len),coord->len);
		realloc_doubles(&(coord->z),(new_len - coord->len),coord->len);
		realloc_doubles(&(coord->q),(new_len - coord->len),coord->len);
		realloc_ints(&(coord->typ), (new_len - coord->len),coord->len);
		realloc_ints(&(coord->atom_id), (new_len - coord->len),coord->len);
		realloc_ints(&(coord->atom_sele_flag1), (new_len - coord->len),coord->len);
		realloc_ints(&(coord->atom_sele_flag2), (new_len - coord->len),coord->len);
		coord->len = new_len;
	}
	return;
}

/************************************************************************/
void   free_TRAVELLING_COORDINATES(TRAVELLING_COORDINATES *coord)
{
	myfree(coord->x);
	myfree(coord->y);
	myfree(coord->z);
	myfree(coord->q);
	myfree(coord->typ);
	myfree(coord->atom_id);
	myfree(coord->atom_sele_flag1);
	myfree(coord->atom_sele_flag2);
	coord->len = coord->no_atoms = 0;
	return;
}

/***********************************************************************/
void   alloc_TRAVELLING_BRANCH(TRAVELLING_BRANCHES *branch)
{
	int	level,nr_cl;

#if DEBUG
	errout("Node %d: in alloc_TRAVELLING_BRANCH, levels=%d, alloc=%d\n",
	        my_number,branch->levels, branch->alloc_per_level[0]);
#endif
	for (level=0;level<branch->levels;level++) {
		nr_cl = branch->alloc_per_level[level];
		branch->cluster[level]=alloc_cluster(nr_cl);
	}
	return;
}

/************************************************************************/
void   realloc_TRAVELLING_BRANCH(TRAVELLING_BRANCHES *branch)
{
	int	level;
	int	new_nr_cl;

	for (level=0;level<branch->levels;level++) {

		new_nr_cl = branch->cl_per_level[level];
		if (new_nr_cl<=branch->alloc_per_level[level])	continue;
		branch->cluster[level]=(struct_cluster*)
				merealloc (
					branch->cluster[level],
					sizeof(struct_cluster)*new_nr_cl,
					sizeof(struct_cluster)*branch->alloc_per_level[level]
				);

		branch->alloc_per_level[level] = new_nr_cl;
	}
	return;
}

/************************************************************************/
void   free_TRAVELLING_BRANCH(TRAVELLING_BRANCHES *branch)
{
	int	level;

	for (level=0;level<branch->levels;level++) {
		myfree(branch->cluster[level]);
		branch->alloc_per_level[level] = branch->cl_per_level[level] = 0;
	}
	return;
}

/***********************************************************************/
void   alloc_TRAVELLING_PACKETS(TRAVELLING_PACKETS *pack)
{
	alloc_TRAVELLING_COORDINATES(&(pack->coordinates));
#if DEBUG
	errout("Node %d: in alloc_TRAVELLING_PACKETS after alloc_TRAVELLING_COORDINATES\n",
	        my_number);
#endif
	alloc_TRAVELLING_BRANCH(&(pack->branch));
	return;
}

/***********************************************************************/
void   free_TRAVELLING_PACKETS(TRAVELLING_PACKETS *pack)
{
	free_TRAVELLING_COORDINATES(&(pack->coordinates));
	free_TRAVELLING_BRANCH(&(pack->branch));
	return;
}

/***********************************************************************/
void   myfree (void *ptr)
{
	free(ptr);
	return;
}

/***********************************************************************/
void   alloc_LOAD_MSG_ONE_NODE (LOAD_MSG_ONE_NODE *msg)
{
	msg->x_new = alloc_doubles(msg->no_more_atoms);
	msg->y_new = alloc_doubles(msg->no_more_atoms);
	msg->z_new = alloc_doubles(msg->no_more_atoms);

	msg->x_old = alloc_doubles(msg->no_more_atoms);
	msg->y_old = alloc_doubles(msg->no_more_atoms);
	msg->z_old = alloc_doubles(msg->no_more_atoms);

	msg->atom_id     = alloc_ints(msg->no_more_atoms);
	msg->typ         = alloc_ints(msg->no_more_atoms);
	msg->atom_string = (char (*) [LEN_ATOM_TYPE+1]) alloc_chars(msg->no_more_atoms*(LEN_ATOM_TYPE+1)+10); 
	msg->res_id      = alloc_ints(msg->no_more_atoms);
	msg->res_string  = (char (*) [LEN_RES_TYPE+1]) alloc_chars(msg->no_more_atoms *(LEN_RES_TYPE+1)+10);
	
	msg->q   = alloc_doubles(msg->no_more_atoms);
	msg->beta= alloc_doubles(msg->no_more_atoms);

	msg->harmon_x = alloc_doubles(msg->no_more_atoms);
	msg->harmon_y = alloc_doubles(msg->no_more_atoms);
	msg->harmon_z = alloc_doubles(msg->no_more_atoms);
	msg->harmon_k = alloc_doubles(msg->no_more_atoms);

	return;
}

/***********************************************************************/
void   realloc_LOAD_MSG_ONE_NODE (LOAD_MSG_ONE_NODE *msg)
{	/* the old allocation size is in msg->no_more_atoms and the new size in msg->no_atoms */
	/* we allocate an extra 40 atoms space */
	int increase = msg->no_atoms - msg->no_more_atoms + 40;

#if REALLOCINFO > 0
	txtout("Node %2d: realloc_LOAD_MSG_ONE_NODE [%-10s] old len: %d, increase by: %d\n",
		my_number,allocinfostr,msg->no_more_atoms, increase);
#endif
	
	realloc_doubles(&msg->x_new, increase, msg->no_more_atoms);
	realloc_doubles(&msg->y_new, increase, msg->no_more_atoms);
	realloc_doubles(&msg->z_new, increase, msg->no_more_atoms);

	realloc_doubles(&msg->x_old, increase, msg->no_more_atoms);
	realloc_doubles(&msg->y_old, increase, msg->no_more_atoms);
	realloc_doubles(&msg->z_old, increase, msg->no_more_atoms);

	realloc_ints(&msg->atom_id, increase, msg->no_more_atoms);
	realloc_ints(&msg->typ,     increase, msg->no_more_atoms);
	realloc_chars((char**)(&msg->atom_string), increase*(LEN_ATOM_TYPE+1), msg->no_more_atoms*(LEN_ATOM_TYPE+1)+10); 
	realloc_ints(&msg->res_id, increase, msg->no_more_atoms);
	realloc_chars((char **)(&msg->res_string), increase*(LEN_RES_TYPE+1), msg->no_more_atoms *(LEN_RES_TYPE+1)+10);
	
	realloc_doubles(&msg->q,    increase, msg->no_more_atoms);
	realloc_doubles(&msg->beta, increase, msg->no_more_atoms);

	realloc_doubles(&msg->harmon_x, increase, msg->no_more_atoms);
	realloc_doubles(&msg->harmon_y, increase, msg->no_more_atoms);
	realloc_doubles(&msg->harmon_z, increase, msg->no_more_atoms);
	realloc_doubles(&msg->harmon_k, increase, msg->no_more_atoms);
	
	msg->no_more_atoms += increase;
	return;
}

/***********************************************************************/
void   free_LOAD_MSG_ONE_NODE(LOAD_MSG_ONE_NODE *msg)
{
	int level;

	myfree(msg->x_new);
	myfree(msg->y_new);
	myfree(msg->z_new);

	myfree(msg->x_old);
	myfree(msg->y_old);
	myfree(msg->z_old);

	myfree(msg->atom_id);		/* 24.8.95 mcb */
	myfree(msg->typ);
	myfree(msg->atom_string);
	myfree(msg->res_id);
	myfree(msg->res_string);
	
	myfree(msg->q);
	myfree(msg->beta);

	myfree(msg->pickbond.atom_ids);
	myfree(msg->pickbond.force);

	myfree(msg->pickelec.atom_ids);
	myfree(msg->pickelec.force);

	myfree(msg->pickvdw.atom_ids);
	myfree(msg->pickvdw.force);

	myfree(msg->pickangle.atom_ids);
	myfree(msg->pickangle.force);

	myfree(msg->pickdihe.atom_ids);
	myfree(msg->pickdihe.force);

	myfree(msg->pickimpr.atom_ids);
	myfree(msg->pickimpr.force);

	myfree(msg->harmon_x);
	myfree(msg->harmon_y);
	myfree(msg->harmon_z);
	myfree(msg->harmon_k);

	myfree(msg->ex_list);
	myfree(msg->ex14_list);

	myfree(msg->int_bond_list);
	myfree(msg->ext_bond_list);

	myfree(msg->shake_list);

	myfree(msg->int_angle_Nr1);
	myfree(msg->int_angle_Nr2);
	myfree(msg->int_angle_Nr3);
	myfree(msg->int_angle_index);

	myfree(msg->ext_angle_Nr1);
	myfree(msg->ext_angle_Nr2);
	myfree(msg->ext_angle_Nr3);
	myfree(msg->ext_angle_index);

	myfree(msg->int_dihe_Nr1);
	myfree(msg->int_dihe_Nr2);
	myfree(msg->int_dihe_Nr3);
	myfree(msg->int_dihe_Nr4);
	myfree(msg->int_dihe_index);

	myfree(msg->ext_dihe_Nr1);
	myfree(msg->ext_dihe_Nr2);
	myfree(msg->ext_dihe_Nr3);
	myfree(msg->ext_dihe_Nr4);
	myfree(msg->ext_dihe_index);

	myfree(msg->int_impr_Nr1);
	myfree(msg->int_impr_Nr2);
	myfree(msg->int_impr_Nr3);
	myfree(msg->int_impr_Nr4);
	myfree(msg->int_impr_index);

	myfree(msg->ext_impr_Nr1);
	myfree(msg->ext_impr_Nr2);
	myfree(msg->ext_impr_Nr3);
	myfree(msg->ext_impr_Nr4);
	myfree(msg->ext_impr_index);

	for (level=0;level<MAX_NO_LEVELS;level++) {
		if (msg->cluster[level]) myfree(msg->cluster[level]);
		msg->cluster[level] = NULL;
	}
	return;
}

/***********************************************************************/
void	free_global_data(LOAD_MSG_ALL_NODE *global_data)
{
	int i;

	myfree(global_data->first_atom);
	myfree(global_data->writeout_atoms);
	myfree(global_data->atom_sele_flag1);
	myfree(global_data->atom_sele_flag2);

	if (global_data->rotation_mode) {
	  myfree(global_data->rot_x);
	  myfree(global_data->rot_y);
	  myfree(global_data->rot_z);
	  myfree(global_data->rot1_x);
	  myfree(global_data->rot1_y);
	  myfree(global_data->rot1_z);
	  myfree(global_data->rot2_x);
	  myfree(global_data->rot2_y);
	  myfree(global_data->rot2_z);
	  myfree(global_data->rotation_torque);
	  myfree(global_data->rotation_map);
	}

	for (i=0;i<global_data->treeinfo.levels;i++) {
		if (global_data->levelinfo[i].first_cluster!=NULL)
			myfree(global_data->levelinfo[i].first_cluster);

		if (global_data->levelinfo[i].last_cluster!=NULL)
			myfree(global_data->levelinfo[i].last_cluster);
	}

	if ((global_data->flooding) || (global_data->rot_correction)) {
		myfree(global_data->flood_used_atoms);
		for (i=0;i<global_data->nr_of_flooding_gaussians;i++) {
			myfree(global_data->flood_matrix[i]);
			myfree(global_data->flood_cm[i]);
		}
	}
}

/***********************************************************************/
void 	alloc_extrapolation(	int len,
				double **new_fx_elecvdw, double **new_fy_elecvdw, double **new_fz_elecvdw, double **new_E_elecvdw,
				double **old_fx_elecvdw, double **old_fy_elecvdw, double **old_fz_elecvdw, double **old_E_elecvdw,
				double **new_fx_elec,    double **new_fy_elec,    double **new_fz_elec,    double **new_E_elec,
				double **old_fx_elec,    double **old_fy_elec,    double **old_fz_elec,    double **old_E_elec)
{
#if INFO
	txtout("Node %2d: alloc_extrapolation [%-10s] len %d\n",
		my_number,allocinfostr,len);
#endif

  /***** following buffers we need for extrapolation ******/
  *new_fx_elecvdw = alloc_doubles(len);
  *new_fy_elecvdw = alloc_doubles(len);
  *new_fz_elecvdw = alloc_doubles(len);
  *new_E_elecvdw  = alloc_doubles(len);
  
  *old_fx_elecvdw = alloc_doubles(len);
  *old_fy_elecvdw = alloc_doubles(len);
  *old_fz_elecvdw = alloc_doubles(len);
  *old_E_elecvdw  = alloc_doubles(len);
  
  /***** the following buffers we need for extrapolation ******/
  *new_fx_elec = alloc_doubles(len);
  *new_fy_elec = alloc_doubles(len);
  *new_fz_elec = alloc_doubles(len);
  *new_E_elec  = alloc_doubles(len);
  
  *old_fx_elec = alloc_doubles(len);
  *old_fy_elec = alloc_doubles(len);
  *old_fz_elec = alloc_doubles(len);
  *old_E_elec  = alloc_doubles(len);
}


/***********************************************************************/
void 	realloc_extrapolation(	int *len, int increase,
				double **new_fx_elecvdw, double **new_fy_elecvdw, double **new_fz_elecvdw, double **new_E_elecvdw,
				double **old_fx_elecvdw, double **old_fy_elecvdw, double **old_fz_elecvdw, double **old_E_elecvdw,
				double **new_fx_elec,    double **new_fy_elec,    double **new_fz_elec,    double **new_E_elec,
				double **old_fx_elec,    double **old_fy_elec,    double **old_fz_elec,    double **old_E_elec)
{ /* *len is the old len on entry and the new len on exit */
#if REALLOCINFO > 0
	txtout("Node %2d: realloc_extrapolation [%-10s] old len %d, adding %d\n",
		my_number,allocinfostr,*len, increase);
#endif
  
  /***** following buffers we need for extrapolation ******/
  realloc_doubles(new_fx_elecvdw, increase, *len);
  realloc_doubles(new_fy_elecvdw, increase, *len);
  realloc_doubles(new_fz_elecvdw, increase, *len);
  realloc_doubles(new_E_elecvdw, increase, *len);
  
  realloc_doubles(old_fx_elecvdw, increase, *len);
  realloc_doubles(old_fy_elecvdw, increase, *len);
  realloc_doubles(old_fz_elecvdw, increase, *len);
  realloc_doubles(old_E_elecvdw, increase, *len);
  
  /***** the following buffers we need for extrapolation ******/
  realloc_doubles(new_fx_elec, increase, *len);
  realloc_doubles(new_fy_elec, increase, *len);
  realloc_doubles(new_fz_elec, increase, *len);
  realloc_doubles(new_E_elec, increase, *len);
  
  realloc_doubles(old_fx_elec, increase, *len);
  realloc_doubles(old_fy_elec, increase, *len);
  realloc_doubles(old_fz_elec, increase, *len);
  realloc_doubles(old_E_elec, increase, *len);
  
  *len += increase;
  return;
}

/***********************************************************************/
void 	free_extrapolation(	int *len,
				double *new_fx_elecvdw, double *new_fy_elecvdw, double *new_fz_elecvdw, double *new_E_elecvdw,
				double *old_fx_elecvdw, double *old_fy_elecvdw, double *old_fz_elecvdw, double *old_E_elecvdw,
				double *new_fx_elec,    double *new_fy_elec,    double *new_fz_elec,    double *new_E_elec,
				double *old_fx_elec,    double *old_fy_elec,    double *old_fz_elec,    double *old_E_elec)
{
#if INFO
	txtout("Node %2d: free_extrapolation [%-10s] len %d\n",
		my_number,allocinfostr,*len);
#endif

  myfree(new_fx_elecvdw);
  myfree(new_fy_elecvdw);
  myfree(new_fz_elecvdw);
  myfree(new_E_elecvdw);

  myfree(old_fx_elecvdw);
  myfree(old_fy_elecvdw);
  myfree(old_fz_elecvdw);
  myfree(old_E_elecvdw);

  myfree(new_fx_elec);
  myfree(new_fy_elec);
  myfree(new_fz_elec);
  myfree(new_E_elec);

  myfree(old_fx_elec);
  myfree(old_fy_elec);
  myfree(old_fz_elec);
  myfree(old_E_elec);

  *len = 0;
  return;
}



void free_potgrid(	STRUCT_ADD_POTENTIAL potgrid ) 
{
  int	i,j,k;

  for (i=0; i<potgrid.nx; i++) {
    for (j=0; j<potgrid.ny; j++) {
      for (k=0; k<potgrid.ny; k++) {
	if (potgrid.pt[i][j][k]!=NULL) myfree(potgrid.pt[i][j][k]);
      }
      myfree(potgrid.pt[i][j]);
    }
    myfree(potgrid.pt[i]);
  }
  myfree(potgrid.pt);

  return;
}
