/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/* only for backward compatibility to PVM3.2 */
/*
#define pvmhostinfo hostinfo
#define pvmtaskinfo taskinfo
*/

#include "wodecls.h"

#if defined(NEXT) || defined(RS6K) || defined(SUN4) || defined(MAC)
#include <stdlib.h>
#elif defined(PARIX) || defined(PARIXPPC)
#include <sys/memory.h>
#elif defined(PARIXHAL)
#include </usr/include/memory.h>
#else
#include <malloc.h>
#endif

void	malloc_init (int my_number);
void	set_allocinfo(char *str);
void	print_meminfo(void);

void	*merealloc(void *oldp, int news, int olds);

void	*alloc_blk(size_t no_bytes);
int	*alloc_ints(int len);
int	**alloc_ptrs(int len);
void	realloc_ints(int **old_ptr, int len, int old_len);
char	*alloc_chars(int len);
void	realloc_chars(char **old_ptr, int len, int old_len);

float	*alloc_floats(int len);

double	*alloc_doubles(int len);
void	realloc_doubles(double **old_ptr, int len, int old_len);

EXT_3_ATOM	*alloc_EXT_3_ATOMs(int len);
void		realloc_EXT_3_ATOMs(EXT_3_ATOM **old_ptr,int len,int old_len);
EXT_4_ATOM	*alloc_EXT_4_ATOMs(int len);
void		realloc_EXT_4_ATOMs(EXT_4_ATOM ** old_ptr,int len,int old_len);

char **alloc_strings(int len);

#if defined(PARIX)

#include <sys/link.h>
#include <sys/select.h>
Option_t   *alloc_Option_t (int len);
LinkCB_t   **alloc_LinkCB_t (int len);

#elif defined(PARIXHAL)
#include </export/epx/epx/include/epx/link.h>
#include </export/epx/epx/include/epx/select.h>
Option_t   *alloc_Option_t (int len);
LinkCB_t   **alloc_LinkCB_t (int len);

#elif defined(MPI)

#else

struct pvmhostinfo	**alloc_pvmhostinfo(int len);
struct pvmtaskinfo	**alloc_pvmtaskinfo(int len);
void	free_pvmtaskinfo(struct pvmtaskinfo **ptr,int len);
void	free_pvmhostinfo(struct pvmhostinfo **ptr,int len);

#endif

int	*alloc_listsize(int len);
int	*alloc_list (int *len);
void	realloc_list (int **old_ptr, int *len);

void	alloc_TRAVELLING_FORCES(TRAVELLING_FORCES *force);
void	realloc_TRAVELLING_FORCES(TRAVELLING_FORCES *forces,int new_len);
void	free_TRAVELLING_FORCES(TRAVELLING_FORCES *force);

void	alloc_TRAVELLING_COORDINATES(TRAVELLING_COORDINATES *coord);
void	realloc_TRAVELLING_COORDINATES(TRAVELLING_COORDINATES *coord,int new_len);
void	free_TRAVELLING_COORDINATES(TRAVELLING_COORDINATES *coord);

void	alloc_TRAVELLING_PACKETS(TRAVELLING_PACKETS *pack);
void	free_TRAVELLING_PACKETS(TRAVELLING_PACKETS *pack);

void	alloc_TRAVELLING_BRANCH(TRAVELLING_BRANCHES *branch);
void	realloc_TRAVELLING_BRANCH(TRAVELLING_BRANCHES *branch);
void	free_TRAVELLING_BRANCH(TRAVELLING_BRANCHES *branch);

void	myfree(void *ptr);

void	alloc_LOAD_MSG_ONE_NODE (LOAD_MSG_ONE_NODE *msg);
void    realloc_LOAD_MSG_ONE_NODE (LOAD_MSG_ONE_NODE *msg);
void	free_LOAD_MSG_ONE_NODE (LOAD_MSG_ONE_NODE *msg);
void	free_global_data(LOAD_MSG_ALL_NODE *global_data);

struct_cluster		*alloc_cluster(int len);
struct_cluster_reduced	*alloc_cluster_reduced(int len);

STRUCT_ATOMDATA		*alloc_atomdata(int len);
TRAVELLING_PACKETS	*alloc_packets(int len);

void 	alloc_extrapolation(	int len,
				double **new_fx_elecvdw, double **new_fy_elecvdw, double **new_fz_elecvdw, double **new_E_elecvdw,
				double **old_fx_elecvdw, double **old_fy_elecvdw, double **old_fz_elecvdw, double **old_E_elecvdw,
				double **new_fx_elec,    double **new_fy_elec,    double **new_fz_elec,    double **new_E_elec,
				double **old_fx_elec,    double **old_fy_elec,    double **old_fz_elec,    double **old_E_elec);

void 	realloc_extrapolation(	int *len, int increase, 
				double **new_fx_elecvdw, double **new_fy_elecvdw, double **new_fz_elecvdw, double **new_E_elecvdw,
				double **old_fx_elecvdw, double **old_fy_elecvdw, double **old_fz_elecvdw, double **old_E_elecvdw,
				double **new_fx_elec,    double **new_fy_elec,    double **new_fz_elec,    double **new_E_elec,
				double **old_fx_elec,    double **old_fy_elec,    double **old_fz_elec,    double **old_E_elec);

void 	free_extrapolation(	int *len,
				double *new_fx_elecvdw, double *new_fy_elecvdw, double *new_fz_elecvdw, double *new_E_elecvdw,
				double *old_fx_elecvdw, double *old_fy_elecvdw, double *old_fz_elecvdw, double *old_E_elecvdw,
				double *new_fx_elec,    double *new_fy_elec,    double *new_fz_elec,    double *new_E_elec,
				double *old_fx_elec,    double *old_fy_elec,    double *old_fz_elec,    double *old_E_elec);

void	free_potgrid(		STRUCT_ADD_POTENTIAL potgrid );

