/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: egbolib.c	by M.Eichinger,Munich			**
**	Belongs2: EGO		The parallel MD-Simulation programm	**
**	Contents: 							**
**									**									**
**  	Modif's:							**
**									**		
**									**
*************************************************************************/

#define DEBUG_PCK		0
#define DEBUG_TRAVBRANCH	0
#define DEBUG_NODESEND		0
#define DEBUG_GLOBALSEND	0
#define DEBUG			0

#include <math.h>
#include <stdio.h>
#include <string.h>

#if defined(SEQ)

#elif defined(PARIX) || defined(PARIXHAL) || defined(MPI)
#include "egmepvm.h"
#else
#include "pvm3.h"
#endif

#include "ego.h"
#include "egalloc.h"
#include "egbolib.h"
#include "egtree.h"
#include "util_lib.h"
#include "egqminter.h"

void adjust_info(int *info)
{
#ifdef SP2 
if (*info == -2) /* special PVMe code for packing of zero items,to be ignored */ 
*info = 0; 
return;
#else
int a=*info;
a++; 
return; 
#endif 
}

void set_working_flags(	LOAD_MSG_ALL_NODE *global_data,
			int *do_analyze,
			int *restartlevel,
			int *miss_first,
			int *coords_are_coming)
{
	if ((global_data->cycles_passed % global_data->Analyse_nth)==0) {
		*do_analyze = 1;
 	} else	*do_analyze = 0;

	*miss_first = 0;

      /*-- restart levels:
       *-- restartlevel == 0: do nothing,no restart file writing in progress 
       *-- restartlevel == 1: writing OLD coordinates
       *-- restartlevel == 2: writing NEW coordinates */

	/* treat second passthrough extra! */
	if (global_data->cycles_passed != 1) {
		/* the next cycle will be a restart cycle.
		** prepare things (like: write the current
		** coord set as the OLD coords)! */
		if (	((global_data->cycles_passed==global_data->final_cycle_no-1) ||
			(((global_data->cycles_passed+1)%global_data->Restart_nth)==0)) &&
			 (*restartlevel == 0) ) {
			*restartlevel = 1;	
		} else
		/* AND necessary,since in case of a
		** DIFFERENT  analysis stepsize the second
		** step would be performed without the first */
		if (	((global_data->cycles_passed==global_data->final_cycle_no) ||
			(((global_data->cycles_passed)%global_data->Restart_nth)==0)) &&
			(*restartlevel == 1) ){	
			*restartlevel = 2;
		} else
		/* AND necessary,since at a restart it
		** would otherwise be attempted to close
		** the restart fold which was not yet opened */
		if (	(((global_data->cycles_passed - 1) % global_data->Restart_nth) == 0) &&
			(*restartlevel == 2)) {	
			*restartlevel = 0;
		}
	}

	if (global_data->schwing_mode!=0) *restartlevel=0;

	if (	(*do_analyze) ||
		(*restartlevel==1) ||
		(*restartlevel==2) ) 
		*coords_are_coming=1;
	else	*coords_are_coming=0;

	return;
}

/************************************************************************/

static	int	effort_for_coords=COMEFFORT_ALL,
		effort_for_branches=COMEFFORT_ALL;

int
set_communication_effort(int for_coords,int for_branches)
{
	effort_for_coords   = for_coords;
	effort_for_branches = for_branches;
	return 0;
}

int
get_communication_effort(int *for_coords,int *for_branches)
{
	*for_coords   = effort_for_coords;
	*for_branches = effort_for_branches;
	return 0;
}

/*** here I sort forces comming from the nodes by a bubblesort	***/
/*** to increasing numbers. Note that the forces are float on	***/
/*** the master.						***/	
int
sort_to_myforces(int atom_id,STRUCT_FORCE_INFO *finfo,double *force)
{
	int	i=0,offs=0;

	if (finfo->cur_len==finfo->len) {
		errout("Fatal error: Inconsistent number (%d) of forces in sort_to_myforces!\n",
			finfo->cur_len);
		pex(-1);
	}

	/*** search were to insert the new force-data:	***/
	/*** a kind of bubble sort!			***/
	if (finfo->cur_len==0) i = 0;
	else 
	for (i=0;i<finfo->cur_len;i++) 
		if (atom_id<finfo->atom_id[i]) break;

	if (i<finfo->cur_len) {	/*** make a gap ***/
		int j;
		for (j=finfo->cur_len;j>=i+1;j--) {
			finfo->atom_id[j] = finfo->atom_id[j-1];
			memcpy(	finfo->force+FORCEINFOLEN*j,
				finfo->force+FORCEINFOLEN*(j-1),
				FORCEINFOLEN*sizeof(float));
		}
	} else {		/*** add it to the end ***/
		i=finfo->cur_len;
	}
	finfo->atom_id[i]   = atom_id;
	offs = FORCEINFOLEN*i;
	finfo->force[offs  ] = (float) force[0]; /*** E  ***/
	finfo->force[offs+1] = (float) force[1]; /*** Fx ***/
	finfo->force[offs+2] = (float) force[2]; /*** Fy ***/
	finfo->force[offs+3] = (float) force[3]; /*** Fz ***/
	finfo->force[offs+4] = (float) force[4]; /*** Kx ***/
	finfo->force[offs+5] = (float) force[5]; /*** Ky ***/
	finfo->force[offs+6] = (float) force[6]; /*** Kz ***/	
	finfo->cur_len++;
	return 0;
}

#if defined(SEQ)

#else

/****** Dirty Trick *************************************************/
/*** In the loader start_msg is allocted to hold all PICK_? data ****/
/*** We keep the this buffer via its pointers for the unpackbuffer **/
/*** in upk_FORCE_INFO						  ***/
/********************************************************************/
static int 	*pickbond_atomids_buf;
static double 	*pickbond_force_buf;

static int 	*pickelec_atomids_buf;
static double 	*pickelec_force_buf;

static int 	*pickvdw_atomids_buf;
static double 	*pickvdw_force_buf;

static int 	*pickangle_atomids_buf;
static double 	*pickangle_force_buf;

static int 	*pickdihe_atomids_buf;
static double 	*pickdihe_force_buf;

static int 	*pickimpr_atomids_buf;
static double 	*pickimpr_force_buf;

/************************************************************************/
int    pk_LOAD_MSG_ONE_NODE(LOAD_MSG_ONE_NODE *msg)
{
	int    info=0;

	info|=pvm_pkint(&(msg->no_atoms),1,1);
	info|=pvm_pkint(&(msg->no_more_atoms),1,1);

	info|=pvm_pkdouble(msg->x_new,msg->no_atoms,1);
	info|=pvm_pkdouble(msg->y_new,msg->no_atoms,1);
	info|=pvm_pkdouble(msg->z_new,msg->no_atoms,1);

	info|=pvm_pkdouble(msg->x_old,msg->no_atoms,1);
	info|=pvm_pkdouble(msg->y_old,msg->no_atoms,1);
	info|=pvm_pkdouble(msg->z_old,msg->no_atoms,1);

	info|=pvm_pkint(msg->atom_id,msg->no_atoms,1);
	info|=pvm_pkint(msg->typ,msg->no_atoms,1);
	info|=pvm_pkbyte ((char *) msg->atom_string,msg->no_atoms * (LEN_ATOM_TYPE+1),1);
	info|=pvm_pkint(msg->res_id,msg->no_atoms,1);
	info|=pvm_pkbyte ((char *) msg->res_string,msg->no_atoms * (LEN_RES_TYPE+1),1);

	info|=pvm_pkdouble(msg->q,msg->no_atoms,1);
	info|=pvm_pkdouble(msg->beta,msg->no_atoms,1);

	info|=pvm_pkdouble(msg->harmon_x,msg->no_atoms,1);
	info|=pvm_pkdouble(msg->harmon_y,msg->no_atoms,1);
	info|=pvm_pkdouble(msg->harmon_z,msg->no_atoms,1);
	info|=pvm_pkdouble(msg->harmon_k,msg->no_atoms,1);

	info|=pvm_pkint(&(msg->ex_ptr),1,1);
	info|=pvm_pkint(msg->ex_list,msg->ex_ptr,1);

	info|=pvm_pkint(&(msg->ex14_ptr),1,1);
	info|=pvm_pkint(msg->ex14_list,msg->ex14_ptr,1);

	info|=pvm_pkint(&(msg->int_bond_ptr),1,1);
	info|=pvm_pkint(&(msg->ext_bond_ptr),1,1);
	info|=pvm_pkint(msg->int_bond_list,msg->int_bond_ptr,1);
	info|=pvm_pkint(msg->ext_bond_list,msg->ext_bond_ptr,1);

	info|=pvm_pkint(&(msg->shake_number),1,1);
	info|=pvm_pkint(msg->shake_list,msg->shake_number,1);

	info|=pvm_pkint(&(msg->int_angle_ptr),1,1);
	info|=pvm_pkint(&(msg->ext_angle_ptr),1,1);
	info|=pvm_pkint(msg->int_angle_Nr1,msg->int_angle_ptr,1);
	info|=pvm_pkint(msg->int_angle_Nr2,msg->int_angle_ptr,1);
	info|=pvm_pkint(msg->int_angle_Nr3,msg->int_angle_ptr,1);
	info|=pvm_pkint(msg->int_angle_index,msg->int_angle_ptr,1);
	info|=pvm_pkint(msg->ext_angle_Nr1,msg->ext_angle_ptr,1);
	info|=pvm_pkint(msg->ext_angle_Nr2,msg->ext_angle_ptr,1);
	info|=pvm_pkint(msg->ext_angle_Nr3,msg->ext_angle_ptr,1);
	info|=pvm_pkint(msg->ext_angle_index,msg->ext_angle_ptr,1);

	info|=pvm_pkint(&(msg->int_dihe_ptr),1,1);
	info|=pvm_pkint(&(msg->ext_dihe_ptr),1,1);
	info|=pvm_pkint(msg->int_dihe_Nr1,msg->int_dihe_ptr,1);
	info|=pvm_pkint(msg->int_dihe_Nr2,msg->int_dihe_ptr,1);
	info|=pvm_pkint(msg->int_dihe_Nr3,msg->int_dihe_ptr,1);
	info|=pvm_pkint(msg->int_dihe_Nr4,msg->int_dihe_ptr,1);
	info|=pvm_pkint(msg->int_dihe_index,msg->int_dihe_ptr,1);
	info|=pvm_pkint(msg->ext_dihe_Nr1,msg->ext_dihe_ptr,1);
	info|=pvm_pkint(msg->ext_dihe_Nr2,msg->ext_dihe_ptr,1);
	info|=pvm_pkint(msg->ext_dihe_Nr3,msg->ext_dihe_ptr,1);
	info|=pvm_pkint(msg->ext_dihe_Nr4,msg->ext_dihe_ptr,1);
	info|=pvm_pkint(msg->ext_dihe_index,msg->ext_dihe_ptr,1);

	info|=pvm_pkint(&(msg->int_impr_ptr),1,1);
	info|=pvm_pkint(&(msg->ext_impr_ptr),1,1);
	info|=pvm_pkint(msg->int_impr_Nr1,msg->int_impr_ptr,1);
	info|=pvm_pkint(msg->int_impr_Nr2,msg->int_impr_ptr,1);
	info|=pvm_pkint(msg->int_impr_Nr3,msg->int_impr_ptr,1);
	info|=pvm_pkint(msg->int_impr_Nr4,msg->int_impr_ptr,1);
	info|=pvm_pkint(msg->int_impr_index,msg->int_impr_ptr,1);
	info|=pvm_pkint(msg->ext_impr_Nr1,msg->ext_impr_ptr,1);
	info|=pvm_pkint(msg->ext_impr_Nr2,msg->ext_impr_ptr,1);
	info|=pvm_pkint(msg->ext_impr_Nr3,msg->ext_impr_ptr,1);
	info|=pvm_pkint(msg->ext_impr_Nr4,msg->ext_impr_ptr,1);
	info|=pvm_pkint(msg->ext_impr_index,msg->ext_impr_ptr,1);

  	{
  	int level,cl;

	info|=pvm_pkint(&(msg->levels),1,1);
	info|=pvm_pkint(msg->cl_per_level,MAX_NO_LEVELS,1);
	info|=pvm_pkint(msg->cl_more_per_level,MAX_NO_LEVELS,1);

	for (level=0;level<msg->levels;level++) {
		for (cl=0;cl<msg->cl_per_level[level];cl++) {
			info|=pvm_pkint(&(msg->cluster[level][cl].nr_childs),1,1);
			info|=pvm_pkint(&(msg->cluster[level][cl].child_pos),1,1);
			info|=pvm_pkint(&(msg->cluster[level][cl].typ),1,1);

			info|=pvm_pkdouble(msg->cluster[level][cl].bbox,sizeof(msg->cluster[level][cl].bbox)/sizeof(double),1);
			info|=pvm_pkdouble(msg->cluster[level][cl].moment,sizeof(msg->cluster[level][cl].moment)/sizeof(double),1);
			info|=pvm_pkdouble(msg->cluster[level][cl].local,sizeof(msg->cluster[level][cl].local)/sizeof(double),1);
			info|=pvm_pkdouble(msg->cluster[level][cl].cen_geom,sizeof(msg->cluster[level][cl].cen_geom)/sizeof(double),1);
			info|=pvm_pkdouble(msg->cluster[level][cl].cen_refp,sizeof(msg->cluster[level][cl].cen_refp)/sizeof(double),1);
#if DEBUG_NODESEND
txtout("Send cluster: cur-level %d  childs %d  %d  geom %f %f %f\n",
	level,
	msg->cluster[level][cl].nr_childs,
	msg->cluster[level][cl].child_pos,
	msg->cluster[level][cl].cen_geom[0],
	msg->cluster[level][cl].cen_geom[1],
	msg->cluster[level][cl].cen_geom[2]);
#endif
		}
  	}
	}
	info|=pvm_pkint(&(msg->pickbond.len),1,1);
	info|=pvm_pkint(msg->pickbond.atom_ids,2*msg->pickbond.len,1);
	pickbond_atomids_buf = msg->pickbond.atom_ids;
	pickbond_force_buf   = msg->pickbond.force;

	info|=pvm_pkint(&(msg->pickelec.len),1,1);
	info|=pvm_pkint(msg->pickelec.atom_ids,2*msg->pickelec.len,1);
	pickelec_atomids_buf = msg->pickelec.atom_ids;
	pickelec_force_buf   = msg->pickelec.force;

	info|=pvm_pkint(&(msg->pickvdw.len),1,1);
	info|=pvm_pkint(msg->pickvdw.atom_ids,2*msg->pickvdw.len,1);
	pickvdw_atomids_buf = msg->pickvdw.atom_ids;
	pickvdw_force_buf   = msg->pickvdw.force;

	info|=pvm_pkint(&(msg->pickangle.len),1,1);
	info|=pvm_pkint(msg->pickangle.atom_ids,3*msg->pickangle.len,1);
	pickangle_atomids_buf = msg->pickangle.atom_ids;
	pickangle_force_buf   = msg->pickangle.force;

	info|=pvm_pkint(&(msg->pickdihe.len),1,1);
	info|=pvm_pkint(msg->pickdihe.atom_ids,4*msg->pickdihe.len,1);
	pickdihe_atomids_buf = msg->pickdihe.atom_ids;
	pickdihe_force_buf   = msg->pickdihe.force;

	info|=pvm_pkint(&(msg->pickimpr.len),1,1);
	info|=pvm_pkint(msg->pickimpr.atom_ids,4*msg->pickimpr.len,1);
	pickimpr_atomids_buf = msg->pickimpr.atom_ids;
	pickimpr_force_buf   = msg->pickimpr.force;

	if (info!=0) {
		errout("Packing failed in pk_LOAD_MSG_ONE_NODE with %d!\n",info);
		pex(-1);

	}
	return (info);
}

/************************************************************************/
int    pk_CLUSTER_DATA (LOAD_MSG_ONE_NODE *msg)
{
	int info=0;
	int level,cl;

	info|=pvm_pkint(&(msg->levels),1,1);
	info|=pvm_pkint(msg->cl_per_level,MAX_NO_LEVELS,1);

	for (level=0;level<msg->levels;level++) {
		for (cl=0;cl<msg->cl_per_level[level];cl++) {
			info|=pvm_pkint(&(msg->cluster[level][cl].nr_childs),1,1);
			info|=pvm_pkint(&(msg->cluster[level][cl].child_pos),1,1);
			info|=pvm_pkint(&(msg->cluster[level][cl].typ),1,1);
			info|=pvm_pkdouble(msg->cluster[level][cl].bbox,sizeof(msg->cluster[level][cl].bbox)/sizeof(double),1);
			info|=pvm_pkdouble(msg->cluster[level][cl].moment,sizeof(msg->cluster[level][cl].moment)/sizeof(double),1);
			info|=pvm_pkdouble(msg->cluster[level][cl].local,sizeof(msg->cluster[level][cl].local)/sizeof(double),1);
			info|=pvm_pkdouble(msg->cluster[level][cl].cen_geom,sizeof(msg->cluster[level][cl].cen_geom)/sizeof(double),1);
			info|=pvm_pkdouble(msg->cluster[level][cl].cen_refp,sizeof(msg->cluster[level][cl].cen_refp)/sizeof(double),1);
		}
 	}
	if (info!=0) {
		errout("Packing failed in  pk_CLUSTER_DATA !\n");
		pex(-1);
	}
	return (info);
}

/************************************************************************/
int    upk_CLUSTER_DATA (LOAD_MSG_ONE_NODE *msg,int *alloc_per_level)
{
	int level,cl,info=0;
	set_allocinfo("upk_CLUSTER_DATA");

	ERROR_PCK("upk_CLUSTER_DATA",info,pvm_upkint(&(msg->levels),1,1),0)
	ERROR_PCK("upk_CLUSTER_DATA",info,pvm_upkint(msg->cl_per_level,MAX_NO_LEVELS,1),1)

	/**** alloc or realloc all stuff ******/
	for (level=0;level<MAX_NO_LEVELS;level++) {
		if (level>=msg->levels) msg->cl_per_level[level] = 10;

		if (alloc_per_level[level]<=0) {
			msg->cluster[level]=alloc_cluster(msg->cl_per_level[level]);
			alloc_per_level[level] = msg->cl_per_level[level];
		} else
		if (alloc_per_level[level]<msg->cl_per_level[level]) {
			msg->cluster[level]= (struct_cluster*) 
					merealloc(msg->cluster[level],
						sizeof(struct_cluster)*msg->cl_per_level[level],
						sizeof(struct_cluster)*alloc_per_level[level]);
			alloc_per_level[level] = msg->cl_per_level[level];
		}
	}

	/****** unpack it ********/
	for (level=0;level<msg->levels;level++) {
		for (cl=0;cl<msg->cl_per_level[level];cl++) {
			ERROR_PCK("upk_CLUSTER_DATA",info,pvm_upkint(&(msg->cluster[level][cl].nr_childs),1,1),2)
			ERROR_PCK("upk_CLUSTER_DATA",info,pvm_upkint(&(msg->cluster[level][cl].child_pos),1,1),3)
			ERROR_PCK("upk_CLUSTER_DATA",info,pvm_upkint(&(msg->cluster[level][cl].typ),1,1),4)

			ERROR_PCK("upk_CLUSTER_DATA",info,pvm_upkdouble(msg->cluster[level][cl].bbox,sizeof(msg->cluster[level][cl].bbox)/sizeof(double),1), 5  )
			ERROR_PCK("upk_CLUSTER_DATA",info,pvm_upkdouble(msg->cluster[level][cl].moment,sizeof(msg->cluster[level][cl].moment)/sizeof(double),1),6)
			ERROR_PCK("upk_CLUSTER_DATA",info,pvm_upkdouble(msg->cluster[level][cl].local,sizeof(msg->cluster[level][cl].local)/sizeof(double),1),7)
			ERROR_PCK("upk_CLUSTER_DATA",info,pvm_upkdouble(msg->cluster[level][cl].cen_geom,sizeof(msg->cluster[level][cl].cen_geom)/sizeof(double),1),8)
			ERROR_PCK("upk_CLUSTER_DATA",info,pvm_upkdouble(msg->cluster[level][cl].cen_refp,sizeof(msg->cluster[level][cl].cen_refp)/sizeof(double),1),9)
		}
	}
	if (info!=0) {
		errout("Unpacking failed in upk_CLUSTER_DATA !\n");
		pex(-1);

	}
	set_allocinfo("XXX");
	return (info);	
}


/************************************************************************/
int    pk_ALL_NODE_UPDATE(LOAD_MSG_ALL_NODE *msg)
{
  	int    info = 0,i;

	info|=pvm_pkint(msg->first_atom,msg->no_nodes,1);
	info|=pvm_pkint(msg->new2old_map,msg->total_no_atoms,1);	
	info|=pvm_pkdouble(msg->pullspringpos,7,1);

	/******* paking of levelinfo: basic data of each level of the (total) tree *****/
	for (i=0;i<msg->treeinfo.levels;i++) {
		info|=pvm_pkint(&(msg->levelinfo[i].nr_of_clusters),1,1);
		info|=pvm_pkint(&(msg->levelinfo[i].nr_of_childs),1,1);
		info|=pvm_pkint(msg->levelinfo[i].first_cluster,msg->treeinfo.branches,1);
		info|=pvm_pkint(msg->levelinfo[i].last_cluster,msg->treeinfo.branches,1);

#if DEBUG_GLOBALSEND
txtout("send cur-lev %d  nr-clusters %d  nr-childs %d\n(first last):(%d %d) (%d %d)\n",
i,
msg->levelinfo[i].nr_of_clusters,
msg->levelinfo[i].nr_of_childs,
msg->levelinfo[i].first_cluster[0],msg->levelinfo[i].last_cluster[0],
msg->levelinfo[i].first_cluster[1],msg->levelinfo[i].last_cluster[1]);
#endif
 	}

	if (info!=0) {
		errout("Packing failed in pk_ALL_NODE_UPDATE !\n");
		pex(-1);
	}

	return (info);
}


/************************************************************************/
int    pk_LOAD_MSG_ALL_NODE(LOAD_MSG_ALL_NODE *msg)
{
	int    info=0;
	int    i;

	info|=pvm_pkint(&(msg->nr_qmatoms),1,1);
	info|=pvm_pkint(&(msg->qmstart_step),1,1);
	info|=pvm_pkint(&(msg->qmcontin_step),1,1);
	info|=pvm_pkdouble(msg->qmbox,6,1);
	info|=pvm_pkdouble(msg->qmboxext,6,1);

	info|=pvm_pkint(&(msg->restarting),1,1);
	info|=pvm_pkdouble(&(msg->elec14),1,1);
	info|=pvm_pkdouble(&(msg->f_elstat),1,1);
	info|=pvm_pkint(&(msg->special_14_excl),1,1);
	info|=pvm_pkint(&(msg->stochastic),1,1);
	info|=pvm_pkint(&(msg->harmonic),1,1);
	info|=pvm_pkint(&(msg->sbound_range),1,1);
	info|=pvm_pkint(&(msg->shake),1,1);
	info|=pvm_pkint(&(msg->equilibration),1,1);
	info|=pvm_pkint(&(msg->minimisation),1,1);
	info|=pvm_pkdouble(&(msg->fric_coeff),1,1);
	info|=pvm_pkdouble(&(msg->max_move),1,1);
	info|=pvm_pkdouble(&(msg->target_temp),1,1);
	info|=pvm_pkdouble(&(msg->time_step),1,1);
	info|=pvm_pkdouble(&(msg->total_mass),1,1);
	info|=pvm_pkdouble(msg->selimob_mass,3,1);

	info|=pvm_pkint(&(msg->no_nodes),1,1);
	info|=pvm_pkint(&(msg->total_no_atoms),1,1);
	info|=pvm_pkint(&(msg->maxatoms_per_node),1,1);

	info|=pvm_pkint(msg->first_atom,msg->no_nodes,1);
	info|=pvm_pkint(msg->new2old_map,msg->total_no_atoms,1);
	info|=pvm_pkint(&(msg->nr_of_atom_typs),1,1);

	info|=pvm_pkdouble(msg->AB[0],MXNO_ATOM_TYPES * MXNO_ATOM_TYPES,1);
	info|=pvm_pkdouble(msg->AB14[0],MXNO_ATOM_TYPES * MXNO_ATOM_TYPES,1);
	info|=pvm_pkdouble(msg->k_bond[0],MXNO_ATOM_TYPES * MXNO_ATOM_TYPES,1);

	info|=pvm_pkdouble(msg->k_angle,MXNO_ANGLE_PARAM,1);
	info|=pvm_pkdouble(msg->theta0, MXNO_ANGLE_PARAM,1);
	info|=pvm_pkdouble(msg->ubk,    MXNO_ANGLE_PARAM,1);
	info|=pvm_pkdouble(msg->ubd,    MXNO_ANGLE_PARAM,1);

	info|=pvm_pkdouble(msg->k_dihe,MXNO_DIHED_PARAM,1);
	info|=pvm_pkdouble(msg->phi0,MXNO_DIHED_PARAM,1);
	info|=pvm_pkint(msg->n_dihe,MXNO_DIHED_PARAM,1);

	info|=pvm_pkdouble(msg->k_impr,MXNO_IMPROP_PARAM,1);
	info|=pvm_pkdouble(msg->omega0,MXNO_IMPROP_PARAM,1);

	info|=pvm_pkdouble(msg->sbound_droff,sizeof(msg->sbound_droff)/sizeof(double),1);
	info|=pvm_pkdouble(&(msg->sbound_layerthick),1,1);
	info|=pvm_pkdouble(&(msg->sbound_fbetagradient),1,1);
	info|=pvm_pkint(&(msg->keep_press_const),1,1);

	info|=pvm_pkint(&(msg->Analyse_nth),1,1);
	info|=pvm_pkint(&(msg->Restart_nth),1,1);

	info|=pvm_pkint(&(msg->cycles_passed),1,1);
	info|=pvm_pkint(&(msg->final_cycle_no),1,1);

	info|=pvm_pkint(&(msg->nr_distclasses),1,1);
	info|=pvm_pkdouble(msg->outer_classrad,MAX_NO_LEVELS,1);

  	/****** paking of treeinfo: basic data of (total) tree we have ********/
  	info|=pvm_pkint(&(msg->treeinfo.algo),1,1);
  	info|=pvm_pkint(&(msg->treeinfo.levels),1,1);
  	info|=pvm_pkint(&(msg->treeinfo.branches),1,1);
  	info|=pvm_pkint(msg->treeinfo.clusterseq,MAX_NO_LEVELS,1);
  	info|=pvm_pkint(&(msg->treeinfo.steps),1,1);

#if DEBUG_GLOBALSEND
txtout("send used-alg %d   nr-lev %d   nr-branch %d\n",
	msg->treeinfo.algo,
	msg->treeinfo.levels,
	msg->treeinfo.branches);
#endif

  	/******* paking of levelinfo: basic data of each level of the (total) tree *****/
  	for (i=0;i<msg->treeinfo.levels;i++) {
		info|=pvm_pkint(&(msg->levelinfo[i].nr_of_clusters),1,1);
		info|=pvm_pkint(&(msg->levelinfo[i].nr_of_childs),1,1);
		info|=pvm_pkint(msg->levelinfo[i].first_cluster,msg->treeinfo.branches,1);
		info|=pvm_pkint(msg->levelinfo[i].last_cluster,msg->treeinfo.branches,1);
  	}

	/******************************** packing of flooding ***************/
    	info|=pvm_pkint(&(msg->flooding),1,1);
	info|=pvm_pkint(&(msg->rot_correction),1,1);
    	info|=pvm_pkint(&(msg->flag_adaptive_flood),1,1);

    	info|=pvm_pkdouble(&(msg->flood_energy_ini),1,1);
    	info|=pvm_pkdouble(&(msg->flood_energy_final),1,1);
    	info|=pvm_pkdouble(&(msg->flood_energy_inc),1,1);
    	info|=pvm_pkdouble(&(msg->flood_time_const),1,1);
    	info|=pvm_pkdouble(&(msg->flood_energy_average),1,1);
    	info|=pvm_pkdouble(&(msg->flood_energy_strength),1,1);
    	info|=pvm_pkdouble(&(msg->flood_scale),1,1);
    	info|=pvm_pkdouble(&(msg->flood_distance),1,1);
    	info|=pvm_pkdouble(&(msg->flood_time_distance),1,1);
    	
    	info|=pvm_pkint(&(msg->flood_nr_atoms),1,1);
    	info|=pvm_pkint(&(msg->nr_of_flooding_gaussians),1,1);
     
    	for (i=0;i<msg->nr_of_flooding_gaussians;i++) { 
		info|=pvm_pkdouble(msg->flood_cm[i],3*msg->flood_nr_atoms,1);      
    		info|=pvm_pkdouble(msg->flood_matrix[i],3*3*msg->flood_nr_atoms*msg->flood_nr_atoms,1);
    	}
	if ((msg->flooding) ||(msg->rot_correction)) {
 		info|=pvm_pkint(msg->flood_used_atoms,msg->total_no_atoms,1);
	}

	info|=pvm_pkint(&(msg->newtransrotcor),1,1);
	info|=pvm_pkint(msg->atom_sele_flag1,msg->total_no_atoms,1);
	info|=pvm_pkint(msg->atom_sele_flag2,msg->total_no_atoms,1);
	info|=pvm_pkint(&(msg->forceout_start_step),1,1);
	info|=pvm_pkint(&(msg->forceout_stride_step),1,1);

	/******************** vars for arbitrarily shaped sbound: ***************/
	info|=pvm_pkdouble(&(msg->arbs_water_shell_thickness),1,1);
	info|=pvm_pkdouble(&(msg->arbs_surface_level),1,1);
	info|=pvm_pkdouble(&(msg->arbs_gauss_height[0]),MXNO_ARBOUND_GAUSSIANS,1);
	info|=pvm_pkdouble(&(msg->arbs_gauss_y[0][0]),3*MXNO_ARBOUND_GAUSSIANS,1);
	info|=pvm_pkdouble(&(msg->arbs_gauss_w[0][0][0]),3*3*MXNO_ARBOUND_GAUSSIANS,1);
	info|=pvm_pkdouble(&(msg->arbs_gauss_rho2[0][0]),3*MXNO_ARBOUND_GAUSSIANS,1);
	info|=pvm_pkint(&(msg->arbs_ncodebook),1,1);

        /******************** vars for additional potential grid ***************/
        info|=pvm_pkdouble(&(msg->potgrid.stepsize),1,1);
        info|=pvm_pkdouble(&(msg->potgrid.xmin),1,1);
        info|=pvm_pkdouble(&(msg->potgrid.xmax),1,1);
        info|=pvm_pkdouble(&(msg->potgrid.ymin),1,1);
        info|=pvm_pkdouble(&(msg->potgrid.ymax),1,1);
        info|=pvm_pkdouble(&(msg->potgrid.zmin),1,1);
        info|=pvm_pkdouble(&(msg->potgrid.zmax),1,1);
	info|=pvm_pkint(&(msg->potgrid.nx),1,1);
	info|=pvm_pkint(&(msg->potgrid.ny),1,1);
	info|=pvm_pkint(&(msg->potgrid.nz),1,1);
	if (msg->potgrid.stepsize>0.0) {
	  double *buffer;
	  long int i,j,k;
	  long int nos,l;

	  nos = msg->potgrid.ny * msg->potgrid.nz;
	  buffer = alloc_doubles( nos );
	  for (i=0;i<msg->potgrid.nx;i++) {
	    l = 0;
	    for (j=0;j<msg->potgrid.ny;j++) {
	      for (k=0;k<msg->potgrid.nz;k++) {
	        if (msg->potgrid.pt[i][j][k] != NULL) {
		  buffer[l] = msg->potgrid.pt[i][j][k]->sbound;
	        } else {
		  buffer[l] = 0.0;
	        }
	        l++;
	      }
	    }
	    info|=pvm_pkdouble(buffer,nos,1);
	  }
	  myfree( buffer );
	}
	info|=pvm_pkint(&(msg->no_sbound_res_string),1,1);
	info|=pvm_pkbyte ((char *) msg->sbound_res_string,(msg->no_sbound_res_string+1)*(LEN_RES_TYPE+1),1);

	/******************************** user defined vars ***************/
	info|=pvm_pkint(&(msg->num_userints),1,1);
	info|=pvm_pkint(&(msg->num_userdoubles),1,1);
	info|=pvm_pkint(msg->userint,MXNO_USERINTS,1);
	info|=pvm_pkdouble(msg->userdouble,MXNO_USERDOUBLES,1);

	/**************** now we pack some debug-flags ********************/
    	info|=pvm_pkint(msg->debug_flags,sizeof(msg->debug_flags)/sizeof(int),1);

	/**************** now we all for pulling ********************/
	info|=pvm_pkint(&(msg->flag_selimob),1,1);
	info|=pvm_pkint(&(msg->selimob_start),1,1);
	info|=pvm_pkint(&(msg->flag_pull),1,1);

	info|=pvm_pkdouble(&(msg->selimob_const),1,1);
	info|=pvm_pkdouble(&(msg->pull_vel),1,1);
	info|=pvm_pkdouble(msg->pull_dir,3,1);
	info|=pvm_pkdouble(msg->pull_const,2,1);
	info|=pvm_pkdouble(msg->pullspringpos,7,1);
	info|=pvm_pkdouble(msg->mass_pull_tot,2,1);
	info|=pvm_pkdouble(msg->xcom_pull,2,1);
	info|=pvm_pkdouble(msg->ycom_pull,2,1);
	info|=pvm_pkdouble(msg->zcom_pull,2,1);
	info|=pvm_pkint(msg->nr_of_pull_atoms,2,1);
        for (i=0;i<2;i++) {
                info|=pvm_pkint(msg->indpull[i],MXNO_PULLING_ATOMS,1);
                info|=pvm_pkdouble(msg->mass_pull[i],MXNO_PULLING_ATOMS,1);
        }
	/*** all for rotation ***/
	info|=pvm_pkint(&(msg->nr_of_rot_atoms),1,1);
	info|=pvm_pkint(msg->rotation_map,msg->nr_of_rot_atoms,1);
	info|=pvm_pkint(&(msg->rotation_mode),1,1);
	info|=pvm_pkdouble(&(msg->angular_vel),1,1);
	info|=pvm_pkdouble(msg->rotation_dir,3,1);
	info|=pvm_pkdouble(msg->rotation_const,3,1);
	info|=pvm_pkdouble(msg->rotation_point,3,1);

	info|=pvm_pkdouble(msg->rot_x,msg->nr_of_rot_atoms,1);
	info|=pvm_pkdouble(msg->rot_y,msg->nr_of_rot_atoms,1);
	info|=pvm_pkdouble(msg->rot_z,msg->nr_of_rot_atoms,1);

	info|=pvm_pkdouble(msg->rot1_x,msg->nr_of_rot_atoms,1);
	info|=pvm_pkdouble(msg->rot1_y,msg->nr_of_rot_atoms,1);
	info|=pvm_pkdouble(msg->rot1_z,msg->nr_of_rot_atoms,1);
	info|=pvm_pkdouble(msg->rot2_x,msg->nr_of_rot_atoms,1);
	info|=pvm_pkdouble(msg->rot2_y,msg->nr_of_rot_atoms,1);
	info|=pvm_pkdouble(msg->rot2_z,msg->nr_of_rot_atoms,1);

	info|=pvm_pkdouble(msg->rot_t1_x,msg->nr_of_rot_atoms,1);
	info|=pvm_pkdouble(msg->rot_t1_y,msg->nr_of_rot_atoms,1);
	info|=pvm_pkdouble(msg->rot_t1_z,msg->nr_of_rot_atoms,1);
	info|=pvm_pkdouble(msg->rot_t2_x,msg->nr_of_rot_atoms,1);
	info|=pvm_pkdouble(msg->rot_t2_y,msg->nr_of_rot_atoms,1);
	info|=pvm_pkdouble(msg->rot_t2_z,msg->nr_of_rot_atoms,1);

	info|=pvm_pkdouble(msg->rotation_torque,msg->nr_of_rot_atoms,1);
	info|=pvm_pkdouble(msg->rotation_t_torque,msg->nr_of_rot_atoms,1);


	if (info!=0) {
		errout("Packing failed in pk_LOAD_MSG_ALL_NODE !\n");
		pex(-1);
	}

  return (info);
}

/************************************************************************/
int    upk_LOAD_MSG_ONE_NODE(LOAD_MSG_ONE_NODE *msg)
{
  	int	info=0,do_alloc_now=0, dummy;

	if (msg->no_more_atoms == 0) do_alloc_now=1; /* first call, nothing allocated so far */

  	info|=pvm_upkint(&(msg->no_atoms),1,1); /* how many atoms we actually have */
  	if (do_alloc_now)
	  info|=pvm_upkint(&(msg->no_more_atoms),1,1); /* space for how many atoms we should allocate */
	else
	  info|=pvm_upkint(&dummy,1,1); /* don't overwrite the local allocation! */
	  
	set_allocinfo("upk_LOAD_MSG_ONE_NODE");

	if (msg->no_atoms >= msg->no_more_atoms) {
		realloc_LOAD_MSG_ONE_NODE(msg);
	}

	if (do_alloc_now) {
		alloc_LOAD_MSG_ONE_NODE(msg);

	} else { /* we free and alloc the lists every time.
		  * It would be better if we only realloced those lists that really have
		  * outgrown their space. */
		myfree(msg->ex14_list);
		myfree(msg->ex_list);
		myfree(msg->ext_bond_list);
		myfree(msg->int_bond_list);
		myfree(msg->shake_list);
		myfree(msg->int_angle_Nr1);
		myfree(msg->int_angle_Nr2);
		myfree(msg->int_angle_Nr3);
		myfree(msg->int_angle_index);
		myfree(msg->ext_angle_Nr1);
		myfree(msg->ext_angle_Nr2);
		myfree(msg->ext_angle_Nr3);
		myfree(msg->ext_angle_index);
		myfree(msg->int_dihe_Nr1);
		myfree(msg->int_dihe_Nr2);
		myfree(msg->int_dihe_Nr3);
		myfree(msg->int_dihe_Nr4);
		myfree(msg->int_dihe_index);
		myfree(msg->ext_dihe_Nr1);
		myfree(msg->ext_dihe_Nr2);
		myfree(msg->ext_dihe_Nr3);
		myfree(msg->ext_dihe_Nr4);
		myfree(msg->ext_dihe_index);
		myfree(msg->int_impr_Nr1);
		myfree(msg->int_impr_Nr2);
		myfree(msg->int_impr_Nr3);
		myfree(msg->int_impr_Nr4);
		myfree(msg->int_impr_index);
		myfree(msg->ext_impr_Nr1);
		myfree(msg->ext_impr_Nr2);
		myfree(msg->ext_impr_Nr3);
		myfree(msg->ext_impr_Nr4);
		myfree(msg->ext_impr_index);

	}

	info|=pvm_upkdouble(msg->x_new,msg->no_atoms,1);
	info|=pvm_upkdouble(msg->y_new,msg->no_atoms,1);
	info|=pvm_upkdouble(msg->z_new,msg->no_atoms,1);

	info|=pvm_upkdouble(msg->x_old,msg->no_atoms,1);
	info|=pvm_upkdouble(msg->y_old,msg->no_atoms,1);
	info|=pvm_upkdouble(msg->z_old,msg->no_atoms,1);

	info|=pvm_upkint(msg->atom_id,msg->no_atoms,1);
	info|=pvm_upkint(msg->typ,msg->no_atoms,1);
	info|=pvm_upkbyte((char *)msg->atom_string,msg->no_atoms*(LEN_ATOM_TYPE+1),1);
	info|=pvm_upkint(msg->res_id,msg->no_atoms,1); 
	info|=pvm_upkbyte((char *)msg->res_string,msg->no_atoms*(LEN_RES_TYPE+1),1);


	info|=pvm_upkdouble(msg->q,msg->no_atoms,1);
	info|=pvm_upkdouble(msg->beta,msg->no_atoms,1);

	info|=pvm_upkdouble(msg->harmon_x,msg->no_atoms,1);
	info|=pvm_upkdouble(msg->harmon_y,msg->no_atoms,1);
	info|=pvm_upkdouble(msg->harmon_z,msg->no_atoms,1);
	info|=pvm_upkdouble(msg->harmon_k,msg->no_atoms,1);

	info|=pvm_upkint(&(msg->ex_ptr),1,1);
	msg->ex_list = alloc_ints(msg->ex_ptr);
 	info|=pvm_upkint(msg->ex_list,msg->ex_ptr,1);

	info|=pvm_upkint(&(msg->ex14_ptr),1,1);
	msg->ex14_list = alloc_ints(msg->ex14_ptr);
	info|=pvm_upkint(msg->ex14_list,msg->ex14_ptr,1);

	info|=pvm_upkint(&(msg->int_bond_ptr),1,1);
	info|=pvm_upkint(&(msg->ext_bond_ptr),1,1);

	msg->int_bond_list = alloc_ints(msg->int_bond_ptr);
	info|=pvm_upkint(msg->int_bond_list,msg->int_bond_ptr,1);

	msg->ext_bond_list = alloc_ints(msg->ext_bond_ptr);
	info|=pvm_upkint(msg->ext_bond_list,msg->ext_bond_ptr,1);

	info|=pvm_upkint(&(msg->shake_number),1,1);
	msg->shake_list = alloc_ints(msg->shake_number);
	info|=pvm_upkint(msg->shake_list,msg->shake_number,1);

	info|=pvm_upkint(&(msg->int_angle_ptr),1,1);
 	msg->int_angle_Nr1   = alloc_ints(msg->int_angle_ptr);
  	msg->int_angle_Nr2   = alloc_ints(msg->int_angle_ptr);
  	msg->int_angle_Nr3   = alloc_ints(msg->int_angle_ptr);
  	msg->int_angle_index = alloc_ints(msg->int_angle_ptr);

	info|=pvm_upkint(&(msg->ext_angle_ptr),1,1);
	msg->ext_angle_Nr1   = alloc_ints(msg->ext_angle_ptr);
  	msg->ext_angle_Nr2   = alloc_ints(msg->ext_angle_ptr);
  	msg->ext_angle_Nr3   = alloc_ints(msg->ext_angle_ptr);
  	msg->ext_angle_index = alloc_ints(msg->ext_angle_ptr);

	info|=pvm_upkint(msg->int_angle_Nr1,msg->int_angle_ptr,1);
	info|=pvm_upkint(msg->int_angle_Nr2,msg->int_angle_ptr,1);
	info|=pvm_upkint(msg->int_angle_Nr3,msg->int_angle_ptr,1);
	info|=pvm_upkint(msg->int_angle_index,msg->int_angle_ptr,1);
	info|=pvm_upkint(msg->ext_angle_Nr1,msg->ext_angle_ptr,1);
	info|=pvm_upkint(msg->ext_angle_Nr2,msg->ext_angle_ptr,1);
	info|=pvm_upkint(msg->ext_angle_Nr3,msg->ext_angle_ptr,1);
	info|=pvm_upkint(msg->ext_angle_index,msg->ext_angle_ptr,1);

	info|=pvm_upkint(&(msg->int_dihe_ptr),1,1);
  	msg->int_dihe_Nr1   = alloc_ints(msg->int_dihe_ptr);
  	msg->int_dihe_Nr2   = alloc_ints(msg->int_dihe_ptr);
  	msg->int_dihe_Nr3   = alloc_ints(msg->int_dihe_ptr);
  	msg->int_dihe_Nr4   = alloc_ints(msg->int_dihe_ptr);
  	msg->int_dihe_index = alloc_ints(msg->int_dihe_ptr);

 	info|=pvm_upkint(&(msg->ext_dihe_ptr),1,1);
	msg->ext_dihe_Nr1   = alloc_ints(msg->ext_dihe_ptr);
	msg->ext_dihe_Nr2   = alloc_ints(msg->ext_dihe_ptr);
	msg->ext_dihe_Nr3   = alloc_ints(msg->ext_dihe_ptr);
	msg->ext_dihe_Nr4   = alloc_ints(msg->ext_dihe_ptr);
	msg->ext_dihe_index = alloc_ints(msg->ext_dihe_ptr);

	info|=pvm_upkint(msg->int_dihe_Nr1,msg->int_dihe_ptr,1);
	info|=pvm_upkint(msg->int_dihe_Nr2,msg->int_dihe_ptr,1);
	info|=pvm_upkint(msg->int_dihe_Nr3,msg->int_dihe_ptr,1);
	info|=pvm_upkint(msg->int_dihe_Nr4,msg->int_dihe_ptr,1);
	info|=pvm_upkint(msg->int_dihe_index,msg->int_dihe_ptr,1);
	info|=pvm_upkint(msg->ext_dihe_Nr1,msg->ext_dihe_ptr,1);
	info|=pvm_upkint(msg->ext_dihe_Nr2,msg->ext_dihe_ptr,1);
	info|=pvm_upkint(msg->ext_dihe_Nr3,msg->ext_dihe_ptr,1);
	info|=pvm_upkint(msg->ext_dihe_Nr4,msg->ext_dihe_ptr,1);
	info|=pvm_upkint(msg->ext_dihe_index,msg->ext_dihe_ptr,1);

	info|=pvm_upkint(&(msg->int_impr_ptr),1,1);
  	msg->int_impr_Nr1 = alloc_ints(msg->int_impr_ptr);
  	msg->int_impr_Nr2 = alloc_ints(msg->int_impr_ptr);
  	msg->int_impr_Nr3 = alloc_ints(msg->int_impr_ptr);
  	msg->int_impr_Nr4 = alloc_ints(msg->int_impr_ptr);
  	msg->int_impr_index = alloc_ints(msg->int_impr_ptr);

  	info|=pvm_upkint(&(msg->ext_impr_ptr),1,1);
  	msg->ext_impr_Nr1   = alloc_ints(msg->ext_impr_ptr);
 	msg->ext_impr_Nr2   = alloc_ints(msg->ext_impr_ptr);
  	msg->ext_impr_Nr3   = alloc_ints(msg->ext_impr_ptr);
  	msg->ext_impr_Nr4   = alloc_ints(msg->ext_impr_ptr);
  	msg->ext_impr_index = alloc_ints(msg->ext_impr_ptr);

	info|=pvm_upkint(msg->int_impr_Nr1,msg->int_impr_ptr,1);
	info|=pvm_upkint(msg->int_impr_Nr2,msg->int_impr_ptr,1);
	info|=pvm_upkint(msg->int_impr_Nr3,msg->int_impr_ptr,1);
	info|=pvm_upkint(msg->int_impr_Nr4,msg->int_impr_ptr,1);
	info|=pvm_upkint(msg->int_impr_index,msg->int_impr_ptr,1);
	info|=pvm_upkint(msg->ext_impr_Nr1,msg->ext_impr_ptr,1);
	info|=pvm_upkint(msg->ext_impr_Nr2,msg->ext_impr_ptr,1);
	info|=pvm_upkint(msg->ext_impr_Nr3,msg->ext_impr_ptr,1);
	info|=pvm_upkint(msg->ext_impr_Nr4,msg->ext_impr_ptr,1);
	info|=pvm_upkint(msg->ext_impr_index,msg->ext_impr_ptr,1);

  	{
  	int level,cl;
	info|=pvm_upkint(&(msg->levels),1,1);
	info|=pvm_upkint(msg->cl_per_level,MAX_NO_LEVELS,1);
	info|=pvm_upkint(msg->cl_more_per_level,MAX_NO_LEVELS,1);

#if DEBUG_NODESEND
txtout("\nRecv levels %d    cl per lev %d %d\n",msg->levels,msg->cl_per_level[0],
msg->cl_per_level[1]);
#endif

	for (level=0;level<msg->levels;level++) {
		if (do_alloc_now) msg->cluster[level]=alloc_cluster(msg->cl_more_per_level[level]);
		for (cl=0;cl<msg->cl_per_level[level];cl++) {
			info|=pvm_upkint(&(msg->cluster[level][cl].nr_childs),1,1);
			info|=pvm_upkint(&(msg->cluster[level][cl].child_pos),1,1);
			info|=pvm_upkint(&(msg->cluster[level][cl].typ),1,1);

			info|=pvm_upkdouble(msg->cluster[level][cl].bbox,sizeof(msg->cluster[level][cl].bbox)/sizeof(double),1);
			info|=pvm_upkdouble(msg->cluster[level][cl].moment,sizeof(msg->cluster[level][cl].moment)/sizeof(double),1);
			info|=pvm_upkdouble(msg->cluster[level][cl].local,sizeof(msg->cluster[level][cl].local)/sizeof(double),1);
			info|=pvm_upkdouble(msg->cluster[level][cl].cen_geom,sizeof(msg->cluster[level][cl].cen_geom)/sizeof(double),1);
			info|=pvm_upkdouble(msg->cluster[level][cl].cen_refp,sizeof(msg->cluster[level][cl].cen_refp)/sizeof(double),1);
		}
	}
	for (level=msg->levels;level<MAX_NO_LEVELS;level++) {
		if (do_alloc_now) msg->cluster[level]=alloc_cluster(10);
	}
  	}

	if (!do_alloc_now) {
		myfree(msg->pickbond.atom_ids);
		myfree(msg->pickbond.force);
		myfree(msg->pickelec.atom_ids);
		myfree(msg->pickelec.force);
		myfree(msg->pickvdw.atom_ids);
		myfree(msg->pickvdw.force);
		myfree(msg->pickangle.atom_ids);
		myfree(msg->pickangle.force);
		myfree(msg->pickdihe.atom_ids);
		myfree(msg->pickdihe.force);
		myfree(msg->pickimpr.atom_ids);
		myfree(msg->pickimpr.force);
	}

	info|=pvm_upkint(&(msg->pickbond.len),1,1);
	msg->pickbond.atom_ids = alloc_ints(2*msg->pickbond.len);
	msg->pickbond.force    = alloc_doubles(5*msg->pickbond.len);
	info|=pvm_upkint(msg->pickbond.atom_ids,2*msg->pickbond.len,1);

	info|=pvm_upkint(&(msg->pickelec.len),1,1);
	msg->pickelec.atom_ids = alloc_ints(2*msg->pickelec.len);
	msg->pickelec.force    = alloc_doubles(5*msg->pickelec.len);
	info|=pvm_upkint(msg->pickelec.atom_ids,2*msg->pickelec.len,1);

	info|=pvm_upkint(&(msg->pickvdw.len),1,1);
	msg->pickvdw.atom_ids = alloc_ints(2*msg->pickvdw.len);
	msg->pickvdw.force    = alloc_doubles(5*msg->pickvdw.len);
	info|=pvm_upkint(msg->pickvdw.atom_ids,2*msg->pickvdw.len,1);

	info|=pvm_upkint(&(msg->pickangle.len),1,1);
	msg->pickangle.atom_ids = alloc_ints(3*msg->pickangle.len);
	msg->pickangle.force    = alloc_doubles(8*msg->pickangle.len);
	info|=pvm_upkint(msg->pickangle.atom_ids,3*msg->pickangle.len,1);

	info|=pvm_upkint(&(msg->pickdihe.len),1,1);
	msg->pickdihe.atom_ids = alloc_ints(4*msg->pickdihe.len);
	msg->pickdihe.force    = alloc_doubles(14*msg->pickdihe.len);
	info|=pvm_upkint(msg->pickdihe.atom_ids,4*msg->pickdihe.len,1);

	info|=pvm_upkint(&(msg->pickimpr.len),1,1);
	msg->pickimpr.atom_ids = alloc_ints(4*msg->pickimpr.len);
	msg->pickimpr.force    = alloc_doubles(14*msg->pickimpr.len);
	info|=pvm_upkint(msg->pickimpr.atom_ids,4*msg->pickimpr.len,1);

	if (info!=0) {
		errout("Unpacking failed in upk_LOAD_MSG_ONE_NODE !\n");
		pex(-1);

	}
	set_allocinfo("XXX");

	return (info);
}

/************************************************************************/
int    upk_ALL_NODE_UPDATE(LOAD_MSG_ALL_NODE *msg)
{
	int    info=0,i;
	set_allocinfo("upk_ALL_NODE_UPDATE");

	/** no alloc needed here ***/
	info|=pvm_upkint(msg->first_atom,msg->no_nodes,1);
	info|=pvm_upkint(msg->new2old_map,msg->total_no_atoms,1);
	info|=pvm_upkdouble(msg->pullspringpos,7,1);

/**
txtout("in upk_ALL_NODE_UPDATE %d %d %d %d %d\n",
msg->first_atom[0],msg->first_atom[1],msg->first_atom[2],msg->first_atom[3],msg->first_atom[4]);
**/
	for (i=0;i<msg->treeinfo.levels;i++) {
		info|=pvm_upkint(&(msg->levelinfo[i].nr_of_clusters),1,1);
		info|=pvm_upkint(&(msg->levelinfo[i].nr_of_childs),1,1);

		myfree(msg->levelinfo[i].first_cluster);
		myfree(msg->levelinfo[i].last_cluster);

		msg->levelinfo[i].first_cluster = alloc_ints(msg->treeinfo.branches);
		msg->levelinfo[i].last_cluster  = alloc_ints(msg->treeinfo.branches);

		info|=pvm_upkint(msg->levelinfo[i].first_cluster,msg->treeinfo.branches,1);
		info|=pvm_upkint(msg->levelinfo[i].last_cluster ,msg->treeinfo.branches,1);
	}
	if (info!=0) {
		errout("unpacking failed in upk_ALL_NODE_UPDATE %d\n",info);
		pex(-1);
	}	
	set_allocinfo("XXX");
	return 0;
}

/************************************************************************/
int    upk_LOAD_MSG_ALL_NODE (LOAD_MSG_ALL_NODE *msg)
{
	int    info=0;
	int    i;

	set_allocinfo("upk_LOAD_MSG_ALL_NODE");
	info|=pvm_upkint(&(msg->nr_qmatoms),1,1);
	info|=pvm_upkint(&(msg->qmstart_step),1,1);
	info|=pvm_upkint(&(msg->qmcontin_step),1,1);
	info|=pvm_upkdouble(msg->qmbox,6,1);
	info|=pvm_upkdouble(msg->qmboxext,6,1);

	info|=pvm_upkint(&(msg->restarting),1,1);
	info|=pvm_upkdouble(&(msg->elec14),1,1);
	info|=pvm_upkdouble(&(msg->f_elstat),1,1);
	info|=pvm_upkint(&(msg->special_14_excl),1,1);
	info|=pvm_upkint(&(msg->stochastic),1,1);
	info|=pvm_upkint(&(msg->harmonic),1,1);
	info|=pvm_upkint(&(msg->sbound_range),1,1);
	info|=pvm_upkint(&(msg->shake),1,1);
	info|=pvm_upkint(&(msg->equilibration),1,1);
	info|=pvm_upkint(&(msg->minimisation),1,1);
	info|=pvm_upkdouble(&(msg->fric_coeff),1,1);
	info|=pvm_upkdouble(&(msg->max_move),1,1);
	info|=pvm_upkdouble(&(msg->target_temp),1,1);
	info|=pvm_upkdouble(&(msg->time_step),1,1);
	info|=pvm_upkdouble(&(msg->total_mass),1,1);
	info|=pvm_upkdouble(msg->selimob_mass,3,1);

	info|=pvm_upkint(&(msg->no_nodes),1,1);
	msg->first_atom = alloc_ints(msg->no_nodes);
	
	info|=pvm_upkint(&(msg->total_no_atoms),1,1);
  	msg->new2old_map= alloc_ints(msg->total_no_atoms); 
	info|=pvm_upkint(&(msg->maxatoms_per_node),1,1);
	info|=pvm_upkint(msg->first_atom,msg->no_nodes,1);
	info|=pvm_upkint(msg->new2old_map,msg->total_no_atoms,1);
 	
	info|=pvm_upkint(&(msg->nr_of_atom_typs),1,1);
	info|=pvm_upkdouble(msg->AB[0],    MXNO_ATOM_TYPES * MXNO_ATOM_TYPES,1);
	info|=pvm_upkdouble(msg->AB14[0],  MXNO_ATOM_TYPES * MXNO_ATOM_TYPES,1);
	info|=pvm_upkdouble(msg->k_bond[0],MXNO_ATOM_TYPES * MXNO_ATOM_TYPES,1);
	
	info|=pvm_upkdouble(msg->k_angle,MXNO_ANGLE_PARAM,1);
	info|=pvm_upkdouble(msg->theta0, MXNO_ANGLE_PARAM,1);
	info|=pvm_upkdouble(msg->ubk,    MXNO_ANGLE_PARAM,1);
	info|=pvm_upkdouble(msg->ubd,    MXNO_ANGLE_PARAM,1);	
	
	info|=pvm_upkdouble(msg->k_dihe,MXNO_DIHED_PARAM,1);
	info|=pvm_upkdouble(msg->phi0,  MXNO_DIHED_PARAM,1);
	info|=pvm_upkint(msg->n_dihe,   MXNO_DIHED_PARAM,1);
	info|=pvm_upkdouble(msg->k_impr,MXNO_IMPROP_PARAM,1);
	info|=pvm_upkdouble(msg->omega0,MXNO_IMPROP_PARAM,1);
	info|=pvm_upkdouble(msg->sbound_droff,sizeof(msg->sbound_droff)/sizeof(double),1);
	info|=pvm_upkdouble(&(msg->sbound_layerthick),1,1);
	info|=pvm_upkdouble(&(msg->sbound_fbetagradient),1,1);
	info|=pvm_upkint(&(msg->keep_press_const),1,1);
 	
	info|=pvm_upkint(&(msg->Analyse_nth),1,1);
	info|=pvm_upkint(&(msg->Restart_nth),1,1);
	
	info|=pvm_upkint(&(msg->cycles_passed),1,1);
	info|=pvm_upkint(&(msg->final_cycle_no),1,1);
	
	info|=pvm_upkint(&(msg->nr_distclasses),1,1);
	info|=pvm_upkdouble(msg->outer_classrad,MAX_NO_LEVELS,1);

  	info|=pvm_upkint(&(msg->treeinfo.algo),1,1);
  	info|=pvm_upkint(&(msg->treeinfo.levels),1,1);
  	info|=pvm_upkint(&(msg->treeinfo.branches),1,1);
  	info|=pvm_upkint(msg->treeinfo.clusterseq,MAX_NO_LEVELS,1);
  	info|=pvm_upkint(&(msg->treeinfo.steps),1,1);

#if DEBUG_GLOBALSEND
txtout("recv used-alg %d   nr-lev %d   nr-branch %d\n",
	msg->treeinfo.algo,
	msg->treeinfo.levels,
	msg->treeinfo.branches);
#endif
  	for (i=0;i<msg->treeinfo.levels;i++) {
		info|=pvm_upkint(&(msg->levelinfo[i].nr_of_clusters),1,1);
		info|=pvm_upkint(&(msg->levelinfo[i].nr_of_childs),1,1);

		msg->levelinfo[i].first_cluster = alloc_ints(msg->treeinfo.branches);
		msg->levelinfo[i].last_cluster  = alloc_ints(msg->treeinfo.branches);

		info|=pvm_upkint(msg->levelinfo[i].first_cluster,msg->treeinfo.branches,1);
		info|=pvm_upkint(msg->levelinfo[i].last_cluster ,msg->treeinfo.branches,1);
  	}

  
   	/******************************** unpacking of flooding *************/  
    	info|=pvm_upkint(&(msg->flooding),1,1);
	info|=pvm_upkint(&(msg->rot_correction),1,1);
    	info|=pvm_upkint(&(msg->flag_adaptive_flood),1,1);

    	info|=pvm_upkdouble(&(msg->flood_energy_ini),1,1);
    	info|=pvm_upkdouble(&(msg->flood_energy_final),1,1);
    	info|=pvm_upkdouble(&(msg->flood_energy_inc),1,1);
    	info|=pvm_upkdouble(&(msg->flood_time_const),1,1);
    	info|=pvm_upkdouble(&(msg->flood_energy_average),1,1);
    	info|=pvm_upkdouble(&(msg->flood_energy_strength),1,1);
       	info|=pvm_upkdouble(&(msg->flood_scale),1,1);
    	info|=pvm_upkdouble(&(msg->flood_distance),1,1);
    	info|=pvm_upkdouble(&(msg->flood_time_distance),1,1);
    	
    	info|=pvm_upkint(&(msg->flood_nr_atoms),1,1);
    	info|=pvm_upkint(&(msg->nr_of_flooding_gaussians),1,1);
     
    	for (i=0;i<msg->nr_of_flooding_gaussians;i++) {
		msg->flood_cm[i]     = alloc_doubles(3*msg->flood_nr_atoms);
		msg->flood_matrix[i] = alloc_doubles(3*3*msg->flood_nr_atoms*msg->flood_nr_atoms);

		info|=pvm_upkdouble(msg->flood_cm[i],3*msg->flood_nr_atoms,1);      
    		info|=pvm_upkdouble(msg->flood_matrix[i],3*3*msg->flood_nr_atoms*msg->flood_nr_atoms,1);
    	}
	if ((msg->flooding) ||(msg->rot_correction)) {
		msg->flood_used_atoms = alloc_ints(msg->total_no_atoms);
  		info|=pvm_upkint(msg->flood_used_atoms,msg->total_no_atoms,1);
	} 

	info|=pvm_upkint(&(msg->newtransrotcor),1,1);
	msg->atom_sele_flag1 = alloc_ints(msg->total_no_atoms);
	msg->atom_sele_flag2 = alloc_ints(msg->total_no_atoms);
	info|=pvm_upkint(msg->atom_sele_flag1,msg->total_no_atoms,1);
	info|=pvm_upkint(msg->atom_sele_flag2,msg->total_no_atoms,1);
	info|=pvm_upkint(&(msg->forceout_start_step),1,1);
	info|=pvm_upkint(&(msg->forceout_stride_step),1,1);

	/******************** vars for arbitrarily shaped sbound: ***************/
	info|=pvm_upkdouble(&(msg->arbs_water_shell_thickness),1,1);
	info|=pvm_upkdouble(&(msg->arbs_surface_level),1,1);
	info|=pvm_upkdouble(&(msg->arbs_gauss_height[0]),MXNO_ARBOUND_GAUSSIANS,1);
	info|=pvm_upkdouble(&(msg->arbs_gauss_y[0][0]),3*MXNO_ARBOUND_GAUSSIANS,1);
	info|=pvm_upkdouble(&(msg->arbs_gauss_w[0][0][0]),3*3*MXNO_ARBOUND_GAUSSIANS,1);
	info|=pvm_upkdouble(&(msg->arbs_gauss_rho2[0][0]),3*MXNO_ARBOUND_GAUSSIANS,1);
	info|=pvm_upkint(&(msg->arbs_ncodebook),1,1);

        /******************** vars for additional potential grid ***************/
        info|=pvm_upkdouble(&(msg->potgrid.stepsize),1,1);
        info|=pvm_upkdouble(&(msg->potgrid.xmin),1,1);
        info|=pvm_upkdouble(&(msg->potgrid.xmax),1,1);
        info|=pvm_upkdouble(&(msg->potgrid.ymin),1,1);
        info|=pvm_upkdouble(&(msg->potgrid.ymax),1,1);
        info|=pvm_upkdouble(&(msg->potgrid.zmin),1,1);
        info|=pvm_upkdouble(&(msg->potgrid.zmax),1,1);
	info|=pvm_upkint(&(msg->potgrid.nx),1,1);
	info|=pvm_upkint(&(msg->potgrid.ny),1,1);
	info|=pvm_upkint(&(msg->potgrid.nz),1,1);
	if (msg->potgrid.stepsize>0.0) {
	  double *buffer;
	  long int i,j,k;
	  long int nos,l;

	  nos = msg->potgrid.ny * msg->potgrid.nz;
	  buffer = alloc_doubles( nos );

          /****** allocate and fill grid ******/
          msg->potgrid.pt= (STRUCT_POTGRID_ELEMENT ****) alloc_ptrs (msg->potgrid.nx);
          for (i=0;i<msg->potgrid.nx;i++) {
            info|=pvm_upkdouble(buffer,nos,1);
	    l = 0;
            msg->potgrid.pt[i]= (STRUCT_POTGRID_ELEMENT ***) alloc_ptrs (msg->potgrid.ny);
            for (j=0;j<msg->potgrid.ny;j++) {
              msg->potgrid.pt[i][j]= (STRUCT_POTGRID_ELEMENT **) alloc_ptrs (msg->potgrid.nz);
              for (k=0;k<msg->potgrid.nz;k++) {
                if (buffer[l] != 0.0) {
                  msg->potgrid.pt[i][j][k] =
                                        (STRUCT_POTGRID_ELEMENT *) alloc_blk(sizeof(STRUCT_POTGRID_ELEMENT));
                  msg->potgrid.pt[i][j][k]->sbound = buffer[l];
                } else {
                  msg->potgrid.pt[i][j][k] = NULL;
                }
		l++;
              }
            }
          }
	  myfree( buffer );
	}
	info|=pvm_upkint(&(msg->no_sbound_res_string),1,1);
	msg->sbound_res_string = (char (*) [LEN_RES_TYPE+1]) alloc_chars((msg->no_sbound_res_string+1) * (LEN_RES_TYPE+1));
	info|=pvm_upkbyte((char *)msg->sbound_res_string,(msg->no_sbound_res_string+1)*(LEN_RES_TYPE+1),1);

	/******************************** user defined vars ***************/
	info|=pvm_upkint(&(msg->num_userints),1,1);
	info|=pvm_upkint(&(msg->num_userdoubles),1,1);
	info|=pvm_upkint(msg->userint,MXNO_USERINTS,1);
	info|=pvm_upkdouble(msg->userdouble,MXNO_USERDOUBLES,1);

    	info|=pvm_upkint(msg->debug_flags,sizeof(msg->debug_flags)/sizeof(int),1);

	/**************** now we unpack all for pulling *****************/
	info|=pvm_upkint(&(msg->flag_selimob),1,1);
	info|=pvm_upkint(&(msg->selimob_start),1,1);
	info|=pvm_upkint(&(msg->flag_pull),1,1);

	info|=pvm_upkdouble(&(msg->selimob_const),1,1);
	info|=pvm_upkdouble(&(msg->pull_vel),1,1);
	info|=pvm_upkdouble(msg->pull_dir,3,1);
	info|=pvm_upkdouble(msg->pull_const,2,1);
	info|=pvm_upkdouble(msg->pullspringpos,7,1);
	info|=pvm_upkdouble(msg->mass_pull_tot,2,1);
	info|=pvm_upkdouble(msg->xcom_pull,2,1);
	info|=pvm_upkdouble(msg->ycom_pull,2,1);
	info|=pvm_upkdouble(msg->zcom_pull,2,1);
	info|=pvm_upkint(msg->nr_of_pull_atoms,2,1);

        for (i=0;i<2;i++) {
                /* msg->indpull[i]     = alloc_ints(MXNO_PULLING_ATOMS);
                msg->mass_pull[i]   = alloc_doubles(MXNO_PULLING_ATOMS); */

                info|=pvm_upkint(msg->indpull[i],MXNO_PULLING_ATOMS,1);
                info|=pvm_upkdouble(msg->mass_pull[i],MXNO_PULLING_ATOMS,1);
        }

	info|=pvm_upkint(&(msg->nr_of_rot_atoms),1,1);
	msg->rotation_map = alloc_ints(msg->nr_of_rot_atoms);
	info|=pvm_upkint(msg->rotation_map,msg->nr_of_rot_atoms,1);
	info|=pvm_upkint(&(msg->rotation_mode),1,1);
	info|=pvm_upkdouble(&(msg->angular_vel),1,1);
	info|=pvm_upkdouble(msg->rotation_dir,3,1);
	info|=pvm_upkdouble(msg->rotation_const,3,1);
	info|=pvm_upkdouble(msg->rotation_point,3,1);

	msg->rot_x   = alloc_doubles(msg->nr_of_rot_atoms);
	msg->rot_y   = alloc_doubles(msg->nr_of_rot_atoms);
	msg->rot_z   = alloc_doubles(msg->nr_of_rot_atoms);
	info|=pvm_upkdouble(msg->rot_x,msg->nr_of_rot_atoms,1);
	info|=pvm_upkdouble(msg->rot_y,msg->nr_of_rot_atoms,1);
	info|=pvm_upkdouble(msg->rot_z,msg->nr_of_rot_atoms,1);

	msg->rot1_x   = alloc_doubles(msg->nr_of_rot_atoms);
	msg->rot1_y   = alloc_doubles(msg->nr_of_rot_atoms);
	msg->rot1_z   = alloc_doubles(msg->nr_of_rot_atoms);
	info|=pvm_upkdouble(msg->rot1_x,msg->nr_of_rot_atoms,1);
	info|=pvm_upkdouble(msg->rot1_y,msg->nr_of_rot_atoms,1);
	info|=pvm_upkdouble(msg->rot1_z,msg->nr_of_rot_atoms,1);

	msg->rot2_x   = alloc_doubles(msg->nr_of_rot_atoms);
	msg->rot2_y   = alloc_doubles(msg->nr_of_rot_atoms);
	msg->rot2_z   = alloc_doubles(msg->nr_of_rot_atoms);
	info|=pvm_upkdouble(msg->rot2_x,msg->nr_of_rot_atoms,1);
	info|=pvm_upkdouble(msg->rot2_y,msg->nr_of_rot_atoms,1);
	info|=pvm_upkdouble(msg->rot2_z,msg->nr_of_rot_atoms,1);

	msg->rot_t1_x   = alloc_doubles(msg->nr_of_rot_atoms);
	msg->rot_t1_y   = alloc_doubles(msg->nr_of_rot_atoms);
	msg->rot_t1_z   = alloc_doubles(msg->nr_of_rot_atoms);
	info|=pvm_upkdouble(msg->rot_t1_x,msg->nr_of_rot_atoms,1);
	info|=pvm_upkdouble(msg->rot_t1_y,msg->nr_of_rot_atoms,1);
	info|=pvm_upkdouble(msg->rot_t1_z,msg->nr_of_rot_atoms,1);

	msg->rot_t2_x   = alloc_doubles(msg->nr_of_rot_atoms);
	msg->rot_t2_y   = alloc_doubles(msg->nr_of_rot_atoms);
	msg->rot_t2_z   = alloc_doubles(msg->nr_of_rot_atoms);
	info|=pvm_upkdouble(msg->rot_t2_x,msg->nr_of_rot_atoms,1);
	info|=pvm_upkdouble(msg->rot_t2_y,msg->nr_of_rot_atoms,1);
	info|=pvm_upkdouble(msg->rot_t2_z,msg->nr_of_rot_atoms,1);

	msg->rotation_torque    = alloc_doubles(msg->nr_of_rot_atoms);
	info|=pvm_upkdouble(msg->rotation_torque,msg->nr_of_rot_atoms,1);
	msg->rotation_t_torque    = alloc_doubles(msg->nr_of_rot_atoms);
	info|=pvm_upkdouble(msg->rotation_t_torque,msg->nr_of_rot_atoms,1);


	if (info!=0) {
		errout("Unpacking failed in upk_LOAD_MSG_ALL_NODE ! %d\n",info);
		pex(-1);
	}
	set_allocinfo("XXX");
	return (info);
}

/************************************************************************/
int    pk_PRESSURE(PRESSURE *p)
{
	int    info=0;
	info|=pvm_pkdouble(&(p->sphere),1,1);
	info|=pvm_pkdouble(&(p->cyl_rad),1,1);
	info|=pvm_pkdouble(p->plane_planar,6,1);
	info|=pvm_pkdouble(p->plane_edge,12,1);
	info|=pvm_pkdouble(p->plane_corner,8,1);
	info|=pvm_pkdouble(&(p->virial[0][0]),9,1);
	info|=pvm_pkdouble(&(p->virial_ext[0][0]),9,1);
	info|=pvm_pkdouble(&(p->virial_ekin[0][0]),9,1);
        /*rok241096 Packing total virial */
        info|=pvm_pkdouble(&(p->virial_tot),1,1);
        info|=pvm_pkdouble(&(p->virial_ext_tot),1,1);
        info|=pvm_pkdouble(&(p->virial_ekin_tot),1,1);
	if (info!=0) {
		errout("Packing failed in pk_PRESSURE !\n");
		pex(-1);
	}
	return(info);
}

/************************************************************************/
int    upk_PRESSURE(PRESSURE *p)
{
	int    info=0;
	info|=pvm_upkdouble(&(p->sphere),1,1);
	info|=pvm_upkdouble(&(p->cyl_rad),1,1);
	info|=pvm_upkdouble(p->plane_planar,6,1);
	info|=pvm_upkdouble(p->plane_edge,12,1);
	info|=pvm_upkdouble(p->plane_corner,8,1);
	info|=pvm_upkdouble(&(p->virial[0][0]),9,1);
	info|=pvm_upkdouble(&(p->virial_ext[0][0]),9,1);
	info|=pvm_upkdouble(&(p->virial_ekin[0][0]),9,1);
        /*rok241096 UnPacking total virial */
        info|=pvm_upkdouble(&(p->virial_tot),1,1);
        info|=pvm_upkdouble(&(p->virial_ext_tot),1,1);
        info|=pvm_upkdouble(&(p->virial_ekin_tot),1,1);
	if (info!=0) {
		errout("Unpacking failed in upk_PRESSURE !\n");
		pex(-1);
	}
	return(info);
}

/************************************************************************/
int    pk_ENERGY_INFO(ENERGY_INFO *p,int no_nodes)
{
	int    info=0;
	info|=pvm_pkdouble(&(p->elec),1,1);
	info|=pvm_pkdouble(&(p->vdw),1,1);
	info|=pvm_pkdouble(&(p->bond),1,1);
	info|=pvm_pkdouble(&(p->angle),1,1);
	info|=pvm_pkdouble(&(p->dihe),1,1);
	info|=pvm_pkdouble(&(p->impr),1,1);
	info|=pvm_pkdouble(&(p->kinetic),1,1);
	info|=pvm_pkdouble(&(p->kinetic_h2o),1,1);
	info|=pvm_pkdouble(&(p->kinetic_protein),1,1);
	info|=pvm_pkdouble(&(p->restraint),1,1);
	info|=pvm_pkdouble(&(p->hbond),1,1);
	info|=pvm_pkdouble(&(p->flooding),1,1);
	info|=pvm_pkdouble(&(p->flood_energy_average),1,1);
	info|=pvm_pkdouble(&(p->flood_energy_strength),1,1);
	info|=pvm_pkdouble(&(p->flooding_distance),1,1);
	info|=pvm_pkdouble(&(p->sbound),1,1);
	info|=pvm_pkdouble(&(p->time[0]),NUM_TIMERS,1);
	info|=pvm_pkdouble(&(p->C[0]),3,1);
	info|=pvm_pkdouble(&(p->P[0]),3,1);
	info|=pvm_pkdouble(&(p->L[0]),3,1);
	info|=pvm_pkdouble(&(p->imobC[0]),9,1);
	info|=pvm_pkdouble(&(p->pullspringpos[0]),7,1);
	info|=pvm_pkdouble(&(p->pull_indir[0]),2,1);
	info|=pvm_pkdouble(&(p->pull_outdir[0]),2,1);
       	info|=pvm_pkdouble(&(p->rotation_energy),1,1);
	info|=pvm_pkdouble(&(p->rot_force[0]),3,1);

	info|=pvm_pkdouble(&(p->gradinfo[0]),2,1);

	info|=pvm_pkdouble(&(p->qmftot[0]),3,1);	
	info|=pvm_pkdouble(&(p->mmftot[0]),3,1);
	info|=pvm_pkdouble(&(p->qmang[0]) ,3,1);		
	info|=pvm_pkdouble(&(p->mmang[0]) ,3,1);
        /*rok030796 Packing of Force differences and EkinTensor */
        info|=pvm_pkdouble(&(p->dFx),1,1);
        info|=pvm_pkdouble(&(p->dFy),1,1);
        info|=pvm_pkdouble(&(p->dFz),1,1);
        info|=pvm_pkdouble(&(p->ekin_tensor[0][0]),9,1);
        info|=pvm_pkdouble(&(p->ekin_trace),1,1);

        info|=pvm_pkdouble(&(p->special_YP_pos),1,1); /* position of special P43 group */
        info|=pvm_pkdouble(&(p->special_YM_pos),1,1);
        info|=pvm_pkint(&(p->special_YP_count),1,1); /* count of special P43 group */
        info|=pvm_pkint(&(p->special_YM_count),1,1);

	info|=pk_PRESSURE(&(p->pressure));
	info|=pk_PRESSURE(&(p->density));
	if (info!=0) {
		errout("Packing failed in pk_ENERGY_INFO !\n");
		pex(-1);
	}
	return (info);
}


/************************************************************************/
int    upk_ENERGY_INFO(ENERGY_INFO *p,int no_nodes)
{
	int    info=0;
	info|=pvm_upkdouble(&(p->elec),1,1);
	info|=pvm_upkdouble(&(p->vdw),1,1);
	info|=pvm_upkdouble(&(p->bond),1,1);
	info|=pvm_upkdouble(&(p->angle),1,1);
	info|=pvm_upkdouble(&(p->dihe),1,1);
	info|=pvm_upkdouble(&(p->impr),1,1);
	info|=pvm_upkdouble(&(p->kinetic),1,1);
	info|=pvm_upkdouble(&(p->kinetic_h2o),1,1);
	info|=pvm_upkdouble(&(p->kinetic_protein),1,1);
	info|=pvm_upkdouble(&(p->restraint),1,1);
	info|=pvm_upkdouble(&(p->hbond),1,1);
	info|=pvm_upkdouble(&(p->flooding),1,1);
	info|=pvm_upkdouble(&(p->flood_energy_average),1,1);
	info|=pvm_upkdouble(&(p->flood_energy_strength),1,1);
	info|=pvm_upkdouble(&(p->flooding_distance),1,1);
	info|=pvm_upkdouble(&(p->sbound),1,1);
	info|=pvm_upkdouble(&(p->time[0]),NUM_TIMERS,1);
	info|=pvm_upkdouble(&(p->C[0]),3,1);
	info|=pvm_upkdouble(&(p->P[0]),3,1);
	info|=pvm_upkdouble(&(p->L[0]),3,1);
	info|=pvm_upkdouble(&(p->imobC[0]),9,1);
	info|=pvm_upkdouble(&(p->pullspringpos[0]),7,1);
	info|=pvm_upkdouble(&(p->pull_indir[0]),2,1);
	info|=pvm_upkdouble(&(p->pull_outdir[0]),2,1);
       	info|=pvm_upkdouble(&(p->rotation_energy),1,1);
	info|=pvm_upkdouble(&(p->rot_force[0]),3,1);

	info|=pvm_upkdouble(&(p->gradinfo[0]),2,1);

	info|=pvm_upkdouble(&(p->qmftot[0]),3,1);	
	info|=pvm_upkdouble(&(p->mmftot[0]),3,1);
	info|=pvm_upkdouble(&(p->qmang[0]) ,3,1);		
	info|=pvm_upkdouble(&(p->mmang[0]) ,3,1);
        /*rok030796 UnPacking of Forcedifferences and EkinTensor */
        info|=pvm_upkdouble(&(p->dFx),1,1);
        info|=pvm_upkdouble(&(p->dFy),1,1);
        info|=pvm_upkdouble(&(p->dFz),1,1);
        info|=pvm_upkdouble(&(p->ekin_tensor[0][0]),9,1);
        info|=pvm_upkdouble(&(p->ekin_trace),1,1);

        info|=pvm_upkdouble(&(p->special_YP_pos),1,1);
        info|=pvm_upkdouble(&(p->special_YM_pos),1,1);
        info|=pvm_upkint(&(p->special_YP_count),1,1);
        info|=pvm_upkint(&(p->special_YM_count),1,1);

	info|=upk_PRESSURE(&(p->pressure));
	info|=upk_PRESSURE(&(p->density));
	if (info!=0) {
		errout("Unpacking failed in upk_ENERGY_INFO !\n");
		pex(-1);
	}
	return(info);
}

/************************************************************************/
int    pk_POSITIONS_INFO(LOAD_MSG_ALL_NODE *msg)
{
  int info=0;
  info|=pvm_pkdouble(msg->rot_x,msg->nr_of_rot_atoms,1);
  info|=pvm_pkdouble(msg->rot_y,msg->nr_of_rot_atoms,1);
  info|=pvm_pkdouble(msg->rot_z,msg->nr_of_rot_atoms,1);
  info|=pvm_pkdouble(msg->rot1_x,msg->nr_of_rot_atoms,1);
  info|=pvm_pkdouble(msg->rot1_y,msg->nr_of_rot_atoms,1);
  info|=pvm_pkdouble(msg->rot1_z,msg->nr_of_rot_atoms,1);
  info|=pvm_pkdouble(msg->rot2_x,msg->nr_of_rot_atoms,1);
  info|=pvm_pkdouble(msg->rot2_y,msg->nr_of_rot_atoms,1);
  info|=pvm_pkdouble(msg->rot2_z,msg->nr_of_rot_atoms,1);
  info|=pvm_pkdouble(msg->rotation_torque,msg->nr_of_rot_atoms,1);

  if (info!=0) {
    errout("Packing failed in pk_POSITIONS_INFO !\n");
    pex(-1);
  }
  return (info);
}
/************************************************************************/
int    upk_POSITIONS_INFO(LOAD_MSG_ALL_NODE *msg)
{
  int    info=0;
  info|=pvm_upkdouble(msg->rot_x,msg->nr_of_rot_atoms,1);
  info|=pvm_upkdouble(msg->rot_y,msg->nr_of_rot_atoms,1);
  info|=pvm_upkdouble(msg->rot_z,msg->nr_of_rot_atoms,1);
  info|=pvm_upkdouble(msg->rot1_x,msg->nr_of_rot_atoms,1);
  info|=pvm_upkdouble(msg->rot1_y,msg->nr_of_rot_atoms,1);
  info|=pvm_upkdouble(msg->rot1_z,msg->nr_of_rot_atoms,1);
  info|=pvm_upkdouble(msg->rot2_x,msg->nr_of_rot_atoms,1);
  info|=pvm_upkdouble(msg->rot2_y,msg->nr_of_rot_atoms,1);
  info|=pvm_upkdouble(msg->rot2_z,msg->nr_of_rot_atoms,1);
  info|=pvm_upkdouble(msg->rotation_torque,msg->nr_of_rot_atoms,1);

  if (info!=0) {
    errout("Unpacking failed in pk_POSITIONS_INFO !\n");
    pex(-1);
  }
  return (info);
}

/************************************************************************/
int	pk_FORCE_INFO(	int my_number,
			LOAD_MSG_ONE_NODE *my_data,
			LOAD_MSG_ALL_NODE *global_data)
{
	int i,my_count=0;
	int difstep=global_data->cycles_passed-global_data->forceout_start_step;
	
	pvm_pkint(&my_number,1,1);

	if ( (difstep<0)||
	     (global_data->forceout_stride_step<=0)||
	     ((difstep%global_data->forceout_stride_step)!=0)) {
		pvm_pkint(&my_count,1,1);
		return 0;
	} else {
		pvm_pkint(&(my_data->forceout_nr_atoms),1,1);
		if (my_data->forceout_nr_atoms==0) return 0;
	}

	for (i=0,my_count=0;i<my_data->no_atoms;i++) {
		double	coords[3];
		if (my_data->atom_sele_flag1[i]==0) continue;
		
		/*** now we have to pack forces for atom i ***/
		my_count++;

		pvm_pkint(&(my_data->atom_sele_flag1[i]),1,1);
		pvm_pkint(&(my_data->atom_id[i]),1,1);

		coords[0] = my_data->x_new[i];
		coords[1] = my_data->y_new[i];
		coords[2] = my_data->z_new[i];
				
		if (my_data->atom_sele_flag1[i]&FORCE_TOTAL) {   
			pvm_pkdouble(my_data->force_total + 4*i,4,1);
			pvm_pkdouble(coords,3,1);
		}
		
		if (my_data->atom_sele_flag1[i]&FORCE_ELECVDW) {
			pvm_pkdouble(my_data->force_elecvdw + 4*i,4,1);
			pvm_pkdouble(coords,3,1);
		}
		
		if (my_data->atom_sele_flag1[i]&FORCE_BOND) {
			pvm_pkdouble(my_data->force_bond + 4*i,4,1);
			pvm_pkdouble(coords,3,1);
		}

		if (my_data->atom_sele_flag1[i]&FORCE_ELEC) {
			pvm_pkdouble(my_data->force_elec + 4*i,4,1);
			pvm_pkdouble(coords,3,1);
		}

		if (my_data->atom_sele_flag1[i]&FORCE_VDW) {
			pvm_pkdouble(my_data->force_vdw + 4*i,4,1);
			pvm_pkdouble(coords,3,1);
		}
		
		if (my_data->atom_sele_flag1[i]&FORCE_ANGLE) {   
			pvm_pkdouble(my_data->force_angle + 4*i,4,1);
			pvm_pkdouble(coords,3,1);
		}
	
		if (my_data->atom_sele_flag1[i]&FORCE_DIHE) {
			pvm_pkdouble(my_data->force_dihe + 4*i,4,1);
			pvm_pkdouble(coords,3,1);
		}

		if (my_data->atom_sele_flag1[i]&FORCE_IMPR) {
			pvm_pkdouble(my_data->force_impr + 4*i,4,1);
			pvm_pkdouble(coords,3,1);
		}

		if (my_data->atom_sele_flag1[i]&FORCE_RESTR) {
			pvm_pkdouble(my_data->force_restr + 4*i,4,1);
			pvm_pkdouble(coords,3,1);
		}

		if (my_data->atom_sele_flag1[i]&FORCE_SBOUND) {
			pvm_pkdouble(my_data->force_sbound + 4*i,4,1);
			pvm_pkdouble(coords,3,1);
		}

		if (my_data->atom_sele_flag1[i]&FORCE_INTERN) {
			pvm_pkdouble(my_data->force_intern + 4*i,4,1);
			pvm_pkdouble(coords,3,1);
		}

		if (my_data->atom_sele_flag1[i]&FORCE_FLOOD) {
			pvm_pkdouble(my_data->force_flood + 4*i,4,1);
			pvm_pkdouble(coords,3,1);
		}

		if (my_data->atom_sele_flag1[i]&FORCE_KIN) {
			pvm_pkdouble(my_data->force_kin + 4*i,4,1);
			pvm_pkdouble(coords,3,1);
		}

		/**********************************************************/
		/*** here we pack all stuff for the QM-interface 	***/
		/**********************************************************/
		if (my_data->atom_sele_flag1[i]&FLAG_QMATOM) {
			int atom_id = my_data->atom_id[i];
			int index   = get_qmnode_index(atom_id,my_data);

			/*** send the index of the atom at this node	***/
			pvm_pkint(&i,1,1);

			/*** and the coordinates of the qm-atoms	***/
			pvm_pkdouble(coords,3,1);

			/*** elec forces ***/			
			pvm_pkdouble(my_data->force_elec + 4*i + 1,3,1);
			pvm_pkdouble(my_data->q + i,1,1);
			
			/*** the local expansion at the atom position	***/
			pvm_pkdouble(my_data->qmlocal.local+10*index,10,1);
			
			/*** bonding force with MD-Fragment		***/
			pvm_pkdouble(my_data->qmlocal.bondf+4*index,4,1);
			pvm_pkdouble(my_data->qmlocal.anglef+4*index,4,1);
			pvm_pkdouble(my_data->qmlocal.dihef+4*index,4,1);
			pvm_pkdouble(my_data->qmlocal.imprf+4*index,4,1);
									
			/*** the number and ids of dummy atoms to be added***/
			pvm_pkint(&(my_data->qmlocal.incllen[index]),1,1);
			if (my_data->qmlocal.incllen[index]>0)
				pvm_pkint(my_data->qmlocal.incllist+MAXDUMMYSPERQMATOM*index,
					  my_data->qmlocal.incllen[index],1);
			
		}
	}

	/*****************************************************/
	/*** here we pack all pick-forces (=forces between ***/
	/*** atoms, like bonding forces, etc		   ***/
	/*****************************************************/
	pvm_pkint(&(my_data->pickbond.len),1,1);
	if (my_data->pickbond.len>0) {
		pvm_pkint(my_data->pickbond.atom_ids,2*my_data->pickbond.len,1);
		pvm_pkdouble(my_data->pickbond.force,5*my_data->pickbond.len,1);
	}

	pvm_pkint(&(my_data->pickelec.len),1,1);
	if (my_data->pickelec.len>0) {
		pvm_pkint(my_data->pickelec.atom_ids,2*my_data->pickelec.len,1);
		pvm_pkdouble(my_data->pickelec.force,5*my_data->pickelec.len,1);
	}

	pvm_pkint(&(my_data->pickvdw.len),1,1);
	if (my_data->pickvdw.len>0) {
		pvm_pkint(my_data->pickvdw.atom_ids,2*my_data->pickvdw.len,1);
		pvm_pkdouble(my_data->pickvdw.force,5*my_data->pickvdw.len,1);
	}

	pvm_pkint(&(my_data->pickangle.len),1,1);
	if (	(my_data->pickangle.len>0) &&
		(my_data->pickangle.cur_len==my_data->pickangle.len) ) {
		pvm_pkint(my_data->pickangle.atom_ids,3*my_data->pickangle.len,1);
		pvm_pkdouble(my_data->pickangle.force,8*my_data->pickangle.len,1);
	} else
	if (my_data->pickangle.len>0) {
		errout("Fatal error: Inconsistent PICKANGLE data to send %d %d\n",
			my_data->pickangle.cur_len,my_data->pickangle.len);
		pex(-1);	
	}

	pvm_pkint(&(my_data->pickdihe.len),1,1);
	if (	(my_data->pickdihe.len>0) &&
		(my_data->pickdihe.cur_len==my_data->pickdihe.len) ) {
		pvm_pkint(my_data->pickdihe.atom_ids,4*my_data->pickdihe.len,1);
		pvm_pkdouble(my_data->pickdihe.force,14*my_data->pickdihe.len,1);
	} else
	if (my_data->pickdihe.len>0) {
		errout("Fatal error: Inconsistent PICKDIHE data to send %d %d\n",
			my_data->pickdihe.cur_len,my_data->pickdihe.len);
		pex(-1);	
	}

	pvm_pkint(&(my_data->pickimpr.len),1,1);
	if (	(my_data->pickimpr.len>0) &&
		(my_data->pickimpr.cur_len==my_data->pickimpr.len) ) {
		pvm_pkint(my_data->pickimpr.atom_ids,4*my_data->pickimpr.len,1);
		pvm_pkdouble(my_data->pickimpr.force,14*my_data->pickimpr.len,1);
	} else
	if (my_data->pickimpr.len>0) {
		errout("Fatal error: Inconsistent PICKIMPR data to send %d %d\n",
			my_data->pickimpr.cur_len,my_data->pickimpr.len);
		pex(-1);	
	}

	/******************************************/
	/*** send all dummy-atoms to the master ***/
	/*** which are needed by the qm-interface**/
	/******************************************/
	if (global_data->nr_qmatoms>0) {
		int i,nr_dummys=0;
		
		for (i=0;i<my_data->no_atoms;i++)
			if (my_data->atom_sele_flag1[i]&FLAG_QMDUMMY) nr_dummys++;
			
		pvm_pkint(&nr_dummys,1,1);
	
		for (i=0;i<my_data->no_atoms;i++) {
			double ko[3];
			if (my_data->atom_sele_flag1[i]&FLAG_QMDUMMY) {
				ko[0] = my_data->x_new[i];
				ko[1] = my_data->y_new[i];
				ko[2] = my_data->z_new[i];
				
				pvm_pkint(my_data->atom_id+i,1,1);
				pvm_pkdouble(my_data->q+i,1,1);
				pvm_pkdouble(ko,3,1);				
			}
		}	
	} else {
		int nr_dummys=0;
		pvm_pkint(&nr_dummys,1,1);
	}

	/*********************************************************/
	if (my_count!=my_data->forceout_nr_atoms) {
		errout("Fatal error: Inconsistent number of forces to pack %d!=%d\n",
			my_count,my_data->forceout_nr_atoms);
		pex(-1);
	}
	return 0;
}

int	upk_FORCE_INFO(STRUCT_FORCES *my_force)
{
	int	i,my_number=-1,atom_id,force_id,len;
	int	forceout_nr_atoms=0;

	pvm_upkint(&my_number,1,1);
	pvm_upkint(&forceout_nr_atoms,1,1);
#if DEBUG
txtout("upk_FORCE_INFO from node %d num atoms %d\n",my_number,forceout_nr_atoms);
#endif
	if (forceout_nr_atoms==0) return 0;

	for (i=0;i<forceout_nr_atoms;i++) {
		double	force[FORCEINFOLEN];

		pvm_upkint(&force_id,1,1);
		pvm_upkint(&atom_id,1,1);	/*** EGO-atom-id ***/

		if (force_id&FORCE_TOTAL) {
			pvm_upkdouble(force,FORCEINFOLEN,1);
			sort_to_myforces(atom_id,&(my_force->total),force);		
		}	

		if (force_id&FORCE_ELECVDW) {
			pvm_upkdouble(force,FORCEINFOLEN,1);
			sort_to_myforces(atom_id,&(my_force->elecvdw),force);
		}

		if (force_id&FORCE_BOND) {
			pvm_upkdouble(force,FORCEINFOLEN,1);
			sort_to_myforces(atom_id,&(my_force->bond),force);
		}

		if (force_id&FORCE_ELEC) {
			pvm_upkdouble(force,FORCEINFOLEN,1);
			sort_to_myforces(atom_id,&(my_force->elec),force);
		}

		if (force_id&FORCE_VDW) {
			pvm_upkdouble(force,FORCEINFOLEN,1);
			sort_to_myforces(atom_id,&(my_force->vdw),force);
		}

		if (force_id&FORCE_ANGLE) {
			pvm_upkdouble(force,FORCEINFOLEN,1);
			sort_to_myforces(atom_id,&(my_force->angle),force);
		}

		if (force_id&FORCE_DIHE) {
			pvm_upkdouble(force,FORCEINFOLEN,1);
			sort_to_myforces(atom_id,&(my_force->dihe),force);
		}

		if (force_id&FORCE_IMPR) {
			pvm_upkdouble(force,FORCEINFOLEN,1);
			sort_to_myforces(atom_id,&(my_force->impr),force);
		}

		if (force_id&FORCE_RESTR) {
			pvm_upkdouble(force,FORCEINFOLEN,1);
			sort_to_myforces(atom_id,&(my_force->restr),force);
		}

		if (force_id&FORCE_SBOUND) {
			pvm_upkdouble(force,FORCEINFOLEN,1);
			sort_to_myforces(atom_id,&(my_force->sbound),force);
		}
	
		if (force_id&FORCE_INTERN) {
			pvm_upkdouble(force,FORCEINFOLEN,1);
			sort_to_myforces(atom_id,&(my_force->intern),force);
		}

		if (force_id&FORCE_FLOOD) {
			pvm_upkdouble(force,FORCEINFOLEN,1);
			sort_to_myforces(atom_id,&(my_force->flood),force);
		}

		if (force_id&FORCE_KIN) {
			pvm_upkdouble(force,FORCEINFOLEN,1);
			sort_to_myforces(atom_id,&(my_force->kin),force);
		}

		if (force_id&FLAG_QMATOM) {
			int	nodeatomindex,incllen;
			int	incllist[MAXDUMMYSPERQMATOM];
			double	data[33]; /*** 3*ko, 3*vdw, 1*q, 10*local,   ***/
					  /*** 4*bondf,4*angle,4*dihe,4*impr ***/
			
			pvm_upkint(&nodeatomindex,1,1);
			pvm_upkdouble(data  ,3 ,1); /*** "old" coords      ***/
			pvm_upkdouble(data+3,4 ,1); /*** elec-forces  + q  ***/

			/*** unpack local expansions ***/
			pvm_upkdouble(data+7,10,1); 

			/*** unpack bonding force with MD-Fragment ***/
			pvm_upkdouble(data+17,4,1);
			pvm_upkdouble(data+21,4,1);
			pvm_upkdouble(data+25,4,1);
			pvm_upkdouble(data+29,4,1);
									
			pvm_upkint(&incllen,1,1);
			if (incllen>0) pvm_upkint(incllist,incllen,1);
				
			sort_to_qminfo(	atom_id,
					nodeatomindex,my_number,
					&(my_force->qm),
					data,incllen,incllist);
		}
	}

	/******************************************/
	/*** unpack picked bond data 		***/
	/******************************************/
	pvm_upkint(&len,1,1);
	if (len>0) {
		int i,j,k,l;
		pvm_upkint(pickbond_atomids_buf ,2*len,1);
		pvm_upkdouble(pickbond_force_buf,5*len,1);
		for (i=0,l=0;i<2*len;i+=2,l+=5) {
			int bp1 = pickbond_atomids_buf[i  ];
			int bp2 = pickbond_atomids_buf[i+1];
			for (j=0,k=0;j<2*my_force->pickbond.len;j+=2,k+=5) {
				if (	(bp1==my_force->pickbond.atom_ids[j  ]) &&
					(bp2==my_force->pickbond.atom_ids[j+1]) ) {
					memcpy(	my_force->pickbond.force+k,
						pickbond_force_buf+l,
						5*sizeof(double));
				}	
			}	
		}	
	}

	/******************************************/
	/*** unpack picked elec data 		***/
	/******************************************/
	pvm_upkint(&len,1,1);
	if (len>0) {
		int i,j,k,l;
		pvm_upkint(pickelec_atomids_buf ,2*len,1);
		pvm_upkdouble(pickelec_force_buf,5*len,1);
		for (i=0,l=0;i<2*len;i+=2,l+=5) {
			int bp1 = pickelec_atomids_buf[i  ];
			int bp2 = pickelec_atomids_buf[i+1];
			for (j=0,k=0;j<2*my_force->pickelec.len;j+=2,k+=5) {
				if (	(bp1==my_force->pickelec.atom_ids[j  ]) &&
					(bp2==my_force->pickelec.atom_ids[j+1]) ) {
					memcpy(	my_force->pickelec.force+k,
						pickelec_force_buf+l,
						5*sizeof(double));
				}	
			}	
		}	
	}

	/******************************************/
	/*** unpack picked vdw data 		***/
	/******************************************/
	pvm_upkint(&len,1,1);
	if (len>0) {
		int i,j,k,l;
		pvm_upkint(pickvdw_atomids_buf ,2*len,1);
		pvm_upkdouble(pickvdw_force_buf,5*len,1);
		for (i=0,l=0;i<2*len;i+=2,l+=5) {
			int bp1 = pickvdw_atomids_buf[i  ];
			int bp2 = pickvdw_atomids_buf[i+1];
			for (j=0,k=0;j<2*my_force->pickvdw.len;j+=2,k+=5) {
				if (	(bp1==my_force->pickvdw.atom_ids[j  ]) &&
					(bp2==my_force->pickvdw.atom_ids[j+1]) ) {
					memcpy(	my_force->pickvdw.force+k,
						pickvdw_force_buf+l,
						5*sizeof(double));
				}	
			}	
		}	
	}

	/******************************************/
	/*** unpack picked angle data 		***/
	/******************************************/
	pvm_upkint(&len,1,1);
	if (len>0) {
		int i,j,k,l;
		pvm_upkint(pickangle_atomids_buf,3*len,1);
		pvm_upkdouble(pickangle_force_buf,8*len,1);
		for (i=0,l=0;i<3*len;i+=3,l+=8) {
			int bp1 = pickangle_atomids_buf[i  ];
			int bp2 = pickangle_atomids_buf[i+1];
			int bp3 = pickangle_atomids_buf[i+2];
			for (j=0,k=0;j<3*my_force->pickangle.len;j+=3,k+=8) {
				if (	(bp1==my_force->pickangle.atom_ids[j  ]) &&
					(bp2==my_force->pickangle.atom_ids[j+1]) &&
					(bp3==my_force->pickangle.atom_ids[j+2])) {
					memcpy(	my_force->pickangle.force+k,
						pickangle_force_buf+l,
						8*sizeof(double));
				}	
			}	
		}	
	}

	/******************************************/
	/*** unpack picked dihedral data 	***/
	/******************************************/
	pvm_upkint(&len,1,1);
	if (len>0) {
		int i,j,k,l;
		pvm_upkint(pickdihe_atomids_buf,4*len,1);
		pvm_upkdouble(pickdihe_force_buf,14*len,1);
		for (i=0,l=0;i<4*len;i+=4,l+=14) {
			int bp1 = pickdihe_atomids_buf[i  ];
			int bp2 = pickdihe_atomids_buf[i+1];
			int bp3 = pickdihe_atomids_buf[i+2];
			int bp4 = pickdihe_atomids_buf[i+3];
			for (j=0,k=0;j<4*my_force->pickdihe.len;j+=4,k+=14) {
				if (	(bp1==my_force->pickdihe.atom_ids[j  ]) &&
					(bp2==my_force->pickdihe.atom_ids[j+1]) &&
					(bp3==my_force->pickdihe.atom_ids[j+2]) &&
					(bp4==my_force->pickdihe.atom_ids[j+3])) {
					memcpy(	my_force->pickdihe.force+k,
						pickdihe_force_buf+l,
						14*sizeof(double));
				}	
			}	
		}	
	}

	/******************************************/
	/*** unpack picked improper data 	***/
	/******************************************/
	pvm_upkint(&len,1,1);
	if (len>0) {
		int i,j,k,l;
		pvm_upkint(pickimpr_atomids_buf,4*len,1);
		pvm_upkdouble(pickimpr_force_buf,14*len,1);
		for (i=0,l=0;i<4*len;i+=4,l+=14) {
			int bp1 = pickimpr_atomids_buf[i  ];
			int bp2 = pickimpr_atomids_buf[i+1];
			int bp3 = pickimpr_atomids_buf[i+2];
			int bp4 = pickimpr_atomids_buf[i+3];
			for (j=0,k=0;j<4*my_force->pickimpr.len;j+=4,k+=14) {
				if (	(bp1==my_force->pickimpr.atom_ids[j  ]) &&
					(bp2==my_force->pickimpr.atom_ids[j+1]) &&
					(bp3==my_force->pickimpr.atom_ids[j+2]) &&
					(bp4==my_force->pickimpr.atom_ids[j+3])) {
					memcpy(	my_force->pickimpr.force+k,
						pickimpr_force_buf+l,
						14*sizeof(double));
				}	
			}	
		}	
	}

	/**************************************************/
	/*** unpack dummy atoms for QM-interface 	***/
	/**************************************************/
	{int i;
	pvm_upkint(&my_force->qm.cur_nr_dummys,1,1);
	if (my_force->qm.cur_nr_dummys>0) {
		if (my_force->qm.cur_nr_dummys>my_force->qm.len_dummys) {
			errout("Fatal error: Not enough space allocated for\n");
			errout("QM-dummy atoms (%d > %d).\n",
				my_force->qm.cur_nr_dummys,
				my_force->qm.len_dummys);
			pex(-1);
		}
		for (i=0;i<my_force->qm.cur_nr_dummys;i++) {
			pvm_upkint   (my_force->qm.dummy_id+  i,1,1);
			pvm_upkdouble(my_force->qm.dummy_q +  i,1,1);
			pvm_upkdouble(my_force->qm.dummy_ko+3*i,3,1);
		}	
	}
	}
	return 0;
}

/************************************************************************/
int    pk_TRAVELLING_COORDINATES(TRAVELLING_COORDINATES *p)
{
	int    info=0;

	info|=pvm_pkint(&effort_for_coords,1,1);
	info|=pvm_pkint(&(p->no_atoms),1,1);
	info|=pvm_pkint(&(p->node),1,1);
	info|=pvm_pkdouble(p->x,p->no_atoms,1);
	info|=pvm_pkdouble(p->y,p->no_atoms,1);
	info|=pvm_pkdouble(p->z,p->no_atoms,1);

	if (effort_for_coords==COMEFFORT_ALL) {
		info|=pvm_pkdouble(p->q,p->no_atoms,1);
		info|=pvm_pkint(p->typ,p->no_atoms,1);
		info|=pvm_pkint(p->atom_id,p->no_atoms,1);
		info|=pvm_pkint(p->atom_sele_flag1,p->no_atoms,1);
		info|=pvm_pkint(p->atom_sele_flag2,p->no_atoms,1);
	}

	if (info!=0) {
		errout("Packing failed in pk_TRAVELLING_COORDINATES !\n");
		pex(-1);
	}
	return(info);
}

/************************************************************************/
int    upk_TRAVELLING_COORDINATES(TRAVELLING_COORDINATES *p)
{
	int	info=0,no_atoms;
	int	cur_effort;

	set_allocinfo("upk_TRAVELLING_COORDINATES");

	info|=pvm_upkint(&cur_effort,1,1);
	info|=pvm_upkint(&no_atoms,1,1);

	realloc_TRAVELLING_COORDINATES(p,no_atoms);
	p->no_atoms = no_atoms;

	info|=pvm_upkint  (&(p->node),1,1);
	info|=pvm_upkdouble(p->x,p->no_atoms,1);
	info|=pvm_upkdouble(p->y,p->no_atoms,1);
	info|=pvm_upkdouble(p->z,p->no_atoms,1);

	if (cur_effort==COMEFFORT_ALL) {
		info|=pvm_upkdouble(p->q,p->no_atoms,1);
		info|=pvm_upkint(p->typ,p->no_atoms,1);
		info|=pvm_upkint(p->atom_id,p->no_atoms,1);
		info|=pvm_upkint(p->atom_sele_flag1,p->no_atoms,1);
		info|=pvm_upkint(p->atom_sele_flag2,p->no_atoms,1);
	}

	if (info!=0) {
		errout("Unpacking failed in upk_TRAVELLING_COORDINATES !\n");
		pex(-1);
	}
	set_allocinfo("XXX");
	return (info);
}


/************************************************************************/
int    pk_TRAVELLING_BRANCHES (TRAVELLING_BRANCHES *p)
{
	int	info=0;
	int	level,cl;

	ERROR_PCK("pk_TRAVELLING_BRANCHES",info,pvm_pkint(&effort_for_branches,1,1),0 )
	ERROR_PCK("pk_TRAVELLING_BRANCHES",info,pvm_pkint(&(p->node),1,1),1 )
	ERROR_PCK("pk_TRAVELLING_BRANCHES",info,pvm_pkint(&(p->levels),1,1),2 )
	ERROR_PCK("pk_TRAVELLING_BRANCHES",info,pvm_pkint(p->cl_per_level,MAX_NO_LEVELS,1),3 )

	if (effort_for_branches==COMEFFORT_SKIP) return info;

	for (level=0;level<p->levels;level++) {
		for (cl=0;cl<p->cl_per_level[level];cl++) {

			if (effort_for_branches==COMEFFORT_ALL) {
				ERROR_PCK("pk_TRAVELLING_BRANCHES",info,pvm_pkint(&(p->cluster[level][cl].nr_childs),1,1),4 )
				ERROR_PCK("pk_TRAVELLING_BRANCHES",info,pvm_pkint(&(p->cluster[level][cl].child_pos),1,1),5 )
				ERROR_PCK("pk_TRAVELLING_BRANCHES",info,pvm_pkint(&(p->cluster[level][cl].typ),1,1),6 )
			}

			ERROR_PCK("pk_TRAVELLING_BRANCHES",info,pvm_pkdouble(p->cluster[level][cl].cen_refp  ,3,1),7 )
			ERROR_PCK("pk_TRAVELLING_BRANCHES",info,pvm_pkdouble(p->cluster[level][cl].cen_geom+3,1,1),8 )

			if (p->cluster[level][cl].typ==CHARGED) {
				ERROR_PCK("pk_TRAVELLING_BRANCHES",info,pvm_pkdouble (p->cluster[level][cl].moment  ,1,1),9 )
			} else 
			if (p->cluster[level][cl].typ==POLAR) {
				ERROR_PCK("pk_TRAVELLING_BRANCHES",info,pvm_pkdouble (p->cluster[level][cl].moment+1,3,1),10 )
			} else
			if (p->cluster[level][cl].typ==NEUTRAL) {
				/*** do need nothing ***/
			} else {
				errout("Fatal error in pk_TRAVELLING_BRANCHES");
				pex(-1);
			}

			/*** we do not need the boundry-boxes and the local expansion ***/
			/*** in an case ***/
 		}
	}

	if (info!=0) {
		errout("Packing failed in pk_TRAVELLING_BRANCHES !\n");
		pex(-1);
	}
	return info;
}


/************************************************************************/
int    upk_TRAVELLING_BRANCHES (TRAVELLING_BRANCHES *p)
{
	int	level,cl,info=0,cur_effort;

	ERROR_PCK("upk_TRAVELLING_BRANCHES",info,pvm_upkint(&cur_effort,1,1),0)
	ERROR_PCK("upk_TRAVELLING_BRANCHES",info,pvm_upkint(&(p->node),1,1),1)
	ERROR_PCK("upk_TRAVELLING_BRANCHES",info,pvm_upkint(&(p->levels),1,1),2)
	ERROR_PCK("upk_TRAVELLING_BRANCHES",info,pvm_upkint(p->cl_per_level,MAX_NO_LEVELS,1),3)

	if (cur_effort==COMEFFORT_SKIP) return info;

	/*** do not read in here alloc_per_level==specific to this node !! **/

#if DEBUG_TRAVBRANCH
txtout("\nRecv %d trav branch levels %d    cl per lev %d %d\n",
	p->node,p->levels,p->cl_per_level[0],p->cl_per_level[1]);
#endif
	set_allocinfo("upk_TRAVELLING_BRANCHES");
	realloc_TRAVELLING_BRANCH(p);

	for (level=0;level<p->levels;level++) {
	
		for (cl=0;cl<p->cl_per_level[level];cl++) {

			if (cur_effort==COMEFFORT_ALL) {
				ERROR_PCK("upk_TRAVELLING_BRANCHES",info,pvm_upkint(&(p->cluster[level][cl].nr_childs),1,1),4)
				ERROR_PCK("upk_TRAVELLING_BRANCHES",info,pvm_upkint(&(p->cluster[level][cl].child_pos),1,1),5)
				ERROR_PCK("upk_TRAVELLING_BRANCHES",info,pvm_upkint(&(p->cluster[level][cl].typ),1,1),6)
			}
		
			ERROR_PCK("upk_TRAVELLING_BRANCHES",info,pvm_upkdouble(p->cluster[level][cl].cen_refp  ,3,1),7)
			ERROR_PCK("upk_TRAVELLING_BRANCHES",info,pvm_upkdouble(p->cluster[level][cl].cen_geom+3,1,1),8)

			if (p->cluster[level][cl].typ==CHARGED) {
				ERROR_PCK("upk_TRAVELLING_BRANCHES",info,pvm_upkdouble (p->cluster[level][cl].moment  ,1,1),9)
			} else 
			if (p->cluster[level][cl].typ==POLAR) {
				ERROR_PCK("upk_TRAVELLING_BRANCHES",info,pvm_upkdouble (p->cluster[level][cl].moment+1,3,1),10)
			} else
			if (p->cluster[level][cl].typ==NEUTRAL) {
				/*** do need nothing ***/
			} else {
				errout("Fatal error in upk_TRAVELLING_BRANCHES");
				pex(-1);
			}

			/*** we do not need the boundry-boxes and the local expansion ***/
			/*** in any case ***/

#if DEBUG_TRAVBRANCH
txtout("Recv cluster trav branch:cur-lev %d  childs %d  %d  geom %f %f %f %f\n",
	level,
	p->cluster[level][cl].nr_childs,
	p->cluster[level][cl].child_pos,
	p->cluster[level][cl].cen_geom[0],
	p->cluster[level][cl].cen_geom[1],
	p->cluster[level][cl].cen_geom[2],
	p->cluster[level][cl].cen_geom[3]);
#endif
		}
	}

	if (info!=0) {
		errout("Unpacking failed in upk_TRAVELLING_BRANCHES !\n");
		pex(-1);
	}
	set_allocinfo("XXX");
	return info;
}



/************************************************************************/
/*** only needed in egmach.c for analysis of forces *********************/
/************************************************************************/
int    pk_TRAVELLING_FORCES (TRAVELLING_FORCES *p)
{
	int    info=0;

	ERROR_PCK("pk_TRAVELLING_FORCES",info,pvm_pkint  (&(p->no_atoms),1,1),0 )
	ERROR_PCK("pk_TRAVELLING_FORCES",info,pvm_pkint  (&(p->node),1,1),1 )
	ERROR_PCK("pk_TRAVELLING_FORCES",info,pvm_pkdouble(&(p->Fx[0]),p->no_atoms,1),2 )
	ERROR_PCK("pk_TRAVELLING_FORCES",info,pvm_pkdouble(&(p->Fy[0]),p->no_atoms,1),3 )
	ERROR_PCK("pk_TRAVELLING_FORCES",info,pvm_pkdouble(&(p->Fz[0]),p->no_atoms,1),4 )

	if (info!=0) {
		errout("Packing failed in pk_TRAVELLING_FORCES !\n");
		pex(-1);
	}

	return (info);
}


int    upk_TRAVELLING_FORCES(TRAVELLING_FORCES *p)
{
	int    info=0,no_atoms;
	set_allocinfo("upk_TRAVELLING_FORCES");
	ERROR_PCK("upk_TRAVELLING_FORCES",info,pvm_upkint(&no_atoms,1,1),0)
	realloc_TRAVELLING_FORCES(p,no_atoms);
	p->no_atoms = no_atoms;

	ERROR_PCK("upk_TRAVELLING_FORCES",info,pvm_upkint   (&(p->node),1,1),1)
 	ERROR_PCK("upk_TRAVELLING_FORCES",info,pvm_upkdouble(&(p->Fx[0]),p->no_atoms,1),2)
	ERROR_PCK("upk_TRAVELLING_FORCES",info,pvm_upkdouble(&(p->Fy[0]),p->no_atoms,1),3)
	ERROR_PCK("upk_TRAVELLING_FORCES",info,pvm_upkdouble(&(p->Fz[0]),p->no_atoms,1),4)

	if (info!=0) {
		errout("Unpacking failed in upk_TRAVELLING_FORCES !\n");
		pex(-1);
	}
	set_allocinfo("XXX");
	return (info);
}

#endif
