/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

#define SBOUND_FILLING_H2O	1.8	/* offset from minimum to how far water fills
					   the SBOUND under normal conditions */

#define SBOUND_FILLING		0.0	/* offset from minimum to how far s.th. fills
					   the SBOUND under normal conditions */

#define XM		0
#define XP		1
#define YM		2
#define YP		3
#define ZM		4
#define ZP		5
#define SP_ORI_X	6
#define SP_ORI_Y	7
#define SP_ORI_Z	8
#define R_SPHERE	9
#define R_CURVE		10
#define ORI_X_CYL	11
#define ORI_Y_CYL	12
#define R_CYL		13

#define VOLUME_MODE	0	/* keep volume constant */
#define PRESSURE_MODE	1	/* keep outside pressure constant */
#define DENSITY_MODE	2	/* keep density in SBOUND constant */
#define SPECIAL_MODE	3	/* special mode: keep P43 distance constant */

/* The wall mass of the pistons should be choosen so that the pressure
 * fluctuations are on a timescale of speed_of_sound/typical_system_length
 * where typical_system_length can be assumed as volume^(1/3).
 * For the membrane system a timescale of 10ps seems adequate. 
 * \cite{ANDE80,FELL95} */
#define PISTON_MASS	225.0 /* in amu */
#define PISTON_FRICTION 20.0  /* in 1/ps */

/* structure used for ranged SBOUND. This structure and the
 * functions which operate on it:
 * define_range()	to set the range properly
 * ini_range()		to reset (initialize) the range
 * can be changed to fit the particular needs. 
 * So far it has been tailored to lipid/water system: 
 * for atoms between the average glycerol positions the lipid-type
 * weak and short ranged SBOUND will be used, while for all heavy atoms
 * outside the stronger water-type SBOUND params will be used.
 */
typedef struct {
	double YP_pos, YM_pos;             /* average position of (glycerol or s.th. else) */
	int    YP_count, YM_count; /* how many atoms already contributed to the above average */
	int    range_set;          /* bool to identify if range has been set up already */
	int    range_used;         /* bool set from forceout.lis if ranged SBOUND
				      should be used at all */
	        } RANGE ;

typedef struct {
	int O;	/* The water oxygen index */
	int H1;	/* the water hydrogen indices */
	int H2;
	} WATER;
	
typedef struct {
	double x, y, z;
	} VECTOR;
	
typedef struct {
	double energy;
	VECTOR pos;
	} EVECTOR;

void ini_range (int range_used);
void define_range (int no_atoms, double *y, int *atom_id);
void transfer_special (ENERGY_INFO *energy);

void   boundary ( LOAD_MSG_ONE_NODE * my_data,
		  LOAD_MSG_ALL_NODE * global_data,
		  double *rnd_force_std_deviation,
		  double *Gamma, /***changes every step, if any sbound set! ***/
		  PRESSURE *pressure,
		  PRESSURE *density
);

int any_planar_sbound_set (double *sbound_droff);

int any_sbound_set (double *sbound_droff);

void pressure_update(LOAD_MSG_ALL_NODE *global_data, ENERGY_INFO *energy);
void zero_pressure (PRESSURE *pressure);
void print_pressure(PRESSURE *pressure, double *sbound_droff, int mode);
int  fprint_pressure(FILE *to_file, PRESSURE *pressure);

void   compute_Gamma(	LOAD_MSG_ALL_NODE *global_data,
			LOAD_MSG_ONE_NODE *my_data,
			double *Gamma,
			double *rnd_force_std_deviation,
			int my_number);

void ini_constant_pressure(double *sbound_droff, int mode);
void ini_special (int range_used);




