/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: egcolib.c	by M.Eichinger, H.Grubmueller, Munich	**
**	Belongs2: EGO		The parallel MD-Simulation programm	**
**                              This library file contains routines used**
**                              on the CONTROLLER (master).             **
**	Contents: 	enter_ext(...)					**
**									**
**			read_ctl_file(...)				**
**									**
**			read_restart_file(...)				**
**			readin_bondfile(...)				**
**			readin_exclfile(...)				**
**			readin_anglefile(...)				**
**			readin_dihefile(...)				**				
**			read_flood_file(...)				**
**			read_boundary_file(...)				**
**			read_sbound_file(...)				**
**  	Modif's:	See RCS log (wishful thinking of HH)!!		**
**									**		
*************************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#if defined(NEXT)
#include <sys/file.h>
#include <libc.h>
#endif

#include "ego.h"
#include "egcolib.h"
#include "egbolib.h"
#include "egalloc.h"
#include "util_lib.h"
#include "egclust.h"
#include "egloader.h"
#include "egbound.h"
#include "egqminter.h"
#include "timing.h"

#define DEBUG		0
#define DEBUG_FORCEOUT	0

static	int truncate_forcefile(	char *fname,int step,
				LOAD_MSG_ALL_NODE *global_data);

static  int parse_sboundmode(	char *fname,
				char *line,
				LOAD_MSG_ALL_NODE *global_data);
static	int parse_pickbond(	char *selstr,
				STRUCT_ATOMDATA *atom_data,
				int *atom_sele_flag1,
				LOAD_MSG_ALL_NODE *global_data,
				STRUCT_PICK_BOND *pinfo,
				int mask);
static	int parse_pickangle(	char *selstr,
				STRUCT_ATOMDATA *atom_data,
				int *atom_sele_flag1,
				LOAD_MSG_ALL_NODE *global_data,
				STRUCT_PICK_BOND *pinfo,
				int mask);
static	int parse_pickdihe(	char *selstr,
				STRUCT_ATOMDATA *atom_data,
				int *atom_sele_flag1,
				LOAD_MSG_ALL_NODE *global_data,
				STRUCT_PICK_BOND *pinfo,
				int mask);

static	void read_sbound_droff(	FILE *cfd, double *sbound_droff,
				int call_from_restart);

static	void read_sbound_droff_new(	FILE *cfd, double *sbound_droff,
				int call_from_restart);

static int parse_schwing(	char *keystr,char *selstr,
				STRUCT_ATOMDATA *atom_data,
				LOAD_MSG_ALL_NODE *global_data);

static	char	sdum[MAX_STRING_SIZE];

/************************************************************************/
/*** here comes a terribly long function,... but don't worry,... it   ***/
/*** is very straight forward 					      ***/
/************************************************************************/
int	read_ctl_file(	char *control_filename,
			LOAD_MSG_ALL_NODE *global_data)
{
        int    new;
	char   fn[MAX_STRING_SIZE];
	char   end_string[MAX_STRING_SIZE];
	char  *slash;
	int    number_fd,i,k,len;
	FILE  *cfd;

  	/*** now we open and read in the control file ***/
	if ((cfd = fopen(control_filename,"r")) == NULL) {
		errout("Fatal error: Cannot open CONTROL file <%s>.\n",
			control_filename);
		pex(-1);
	}

	txtout("Reading control file      : <%s>\n",control_filename);

	
/*************** New Format ? *********************/
        fscanf(cfd,"%s",fn);
#if (DEBUG == 2)
printf("output  =%s!\n\n",fn);
#endif
        if (strcmp(fn,"CTL2000")==0)
        {
          new = 1;
	  txtout("Switch to CTL2000 Version!!\n");
        }
        else
        {
          new = 0;
          fseek(cfd,0L,SEEK_SET);
        }
/**************************************************/

	/*** read number of lis-files which have to be read later ***/
	number_fd = read_int_skip(cfd);

	if (number_fd>=MAX_FILES) {
		errout("Fatal error: Too many input files (%d). Increase MAX_FILES (%d).\n",
			number_fd,MAX_FILES);
		errout("             This may be the result of a version conflict!");
		fclose(cfd);	
		pex(-1);
	}	
	if (number_fd!=CUR_LISFILES) {
		errout("Fatal error: Expecting %d lis-files in the control file, but got %d.\n",
			CUR_LISFILES,number_fd);
		errout("             Check your control file !\n");	
		errout("             This may be the result of a version conflict!");
		fclose(cfd);	
		pex(-1);
	}
	
  	/*** get the current path in global_data->workpath***/
	global_data->lispath[0] = 0; 
	fn[0] = 0;
	slash = strrchr(control_filename,'/');
	if (slash != NULL) {
		strncat(global_data->lispath,control_filename,(size_t)(slash - control_filename));
		strcat(global_data->lispath,"/");
	}
#if defined(DOS)
	strcpy(global_data->workpath,".");
#else
	strcpy(global_data->workpath,getenv("PWD"));
#endif
	len = strlen(global_data->workpath);
	if (len<1) {
		errout("Fatal error: Could not get PWD.\n");
		goto ctllis_error;	
	}	
	if (global_data->workpath[len-1]!='/') {
		global_data->workpath[len]='/';
		global_data->workpath[len+1]=0;
	} 
		
	/*** read the lis-file names from the control file and ***/
	/*** append the current path                           ***/
	for (i=0;i<number_fd;i++) {
		fscanf(cfd,"%s",fn);
		strcpy(global_data->filenames[i],global_data->lispath);
		strcat(global_data->filenames[i],fn);
	}
	strcpy(global_data->restartfile,global_data->filenames[RESTART_FILE_INDEX]);
	strcpy(global_data->ctlfile,control_filename);

	/*************************************************************/
	/******** now we read parameters *****************************/
	/*************************************************************/
	global_data->no_nodes	        = read_int_skip(cfd);
	global_data->total_no_atoms     = read_int_skip(cfd);
#if DEBUG
	txtout("In read_ctl_file(...): atoms=%d\n",global_data->total_no_atoms);
#endif
	global_data->Analyse_nth	= read_int_skip(cfd);
	read_string(cfd,global_data->writeout_atom_selection);
	global_data->Restart_nth	= read_int_skip(cfd)*global_data->Analyse_nth;
	global_data->syscallstride 	= read_int_skip(cfd);
	read_string(cfd,global_data->syscallstring);
	global_data->step_of_energy	= read_int_skip(cfd);
	global_data->final_cycle_no	= read_int_skip(cfd);
	global_data->time_step		= read_double_skip(cfd)/TIMEDIV;

	global_data->ex_number 		= read_int_skip(cfd);
	global_data->special_14_excl    = ((global_data->ex_number== 5) || 
					   (global_data->ex_number==-5));
	global_data->minimisation 	= read_bool_skip(cfd);
	global_data->fric_coeff    	= read_double_skip(cfd);
	global_data->max_move 	   	= read_double_skip(cfd);
	global_data->equilibration 	= read_bool_skip(cfd);
	global_data->target_temp	= read_double_skip(cfd);
	global_data->tau_T 		= read_double_skip(cfd);
#if DEBUG
	txtout("In read_ctl_file(...): tau_T=%lf\n",global_data->tau_T);
#endif
	/*************************************************************/
	/***** some things for fast multipole ************************/
	/*************************************************************/
	global_data->nr_distclasses	= read_int_skip(cfd);
	for (i=0;i<global_data->nr_distclasses-1;i++) { 
		global_data->outer_classrad[i] = read_double(cfd);
		global_data->treeinfo.clusterseq[i]        = read_int(cfd);

		if (global_data->outer_classrad[i]<=3.0) {
			errout("Illegal values for distance classes !\n");
			goto ctllis_error;
		}	

		if (	(global_data->treeinfo.clusterseq[i]<0) ||
			(global_data->treeinfo.clusterseq[i]>10) ) {
			fclose(cfd);
			errout("Illegal values for cluster-sequence !\n");
			goto ctllis_error;
		}
	}

	global_data->treeinfo.algo		= read_int_skip(cfd);
	global_data->treeinfo.levels	= read_int_skip(cfd);
	global_data->treeinfo.branches	= read_int_skip(cfd);
	global_data->treeinfo.steps		= read_int_skip(cfd);
#if DEBUG
	txtout("In read_ctl_file(...): ReclustSteps=%d\n",global_data->treeinfo.steps);
#endif
	/************************************************/
	/*** read the output path for the EGO files   ***/
	/*** and create the path if it does not exist ***/
	/************************************************/
	fscanf(cfd,"%s",fn);
	strcpy(global_data->outpath,global_data->lispath);
	strcat(global_data->outpath,fn);
#if DEBUG
	txtout("In read_ctl_file(...): outpath=<%s>\n",fn);
#endif
	create_outpath(global_data);

	skip_line(cfd,1);	/*** skip rest of the line ***/

	global_data->elec14 	= read_double_skip(cfd);
#if DEBUG
	txtout("In read_ctl_file(...): elec14=%lf\n",global_data->elec14);
#endif
	global_data->stochastic = read_bool_skip(cfd);
	global_data->harmonic 	= read_bool_skip(cfd);
	global_data->shake 	= read_bool_skip(cfd);
#if DEBUG
	txtout("In read_ctl_file(...): SHAKE=%d\n",global_data->shake);
#endif
	/*************************************************************/
	/************ some things for SBOUND *************************/
	/*************************************************************/
/***********************New Format?****************************************/
if (new == 1)
        read_sbound_droff_new(cfd,global_data->sbound_droff,0);
else
	read_sbound_droff(cfd,global_data->sbound_droff,0);
/**************************************************************************/	
	
	/*** constant pressure mode data ***/
	global_data->keep_press_const	= read_int_skip(cfd);
	/*** data for planar SBOUND = a box ***/
	/* X_min and X_max: */
	global_data->outside_pressure.plane_planar[0] = read_double(cfd);
	global_data->outside_pressure.plane_planar[1] = global_data->outside_pressure.plane_planar[0];
	/* Y_min and Y_max: */
	global_data->outside_pressure.plane_planar[2] = read_double(cfd);
	global_data->outside_pressure.plane_planar[3] = global_data->outside_pressure.plane_planar[2];
	/* Z_min and Z_max: */
	global_data->outside_pressure.plane_planar[4] = read_double_skip(cfd);
	global_data->outside_pressure.plane_planar[5] = global_data->outside_pressure.plane_planar[4];

	global_data->outside_pressure.sphere  = global_data->outside_pressure.plane_planar[0];
	global_data->outside_pressure.cyl_rad = global_data->outside_pressure.plane_planar[0];

	global_data->sbound_layerthick	 = read_double_skip(cfd);
	global_data->sbound_fbetagradient= read_double_skip(cfd);
#if DEBUG
	txtout("In read_ctl_file(...): Beta=%lf\n",global_data->sbound_fbetagradient);
#endif
	if ((global_data->keep_press_const != 0) &&
	    (global_data->keep_press_const != 1) &&
	    (global_data->keep_press_const != 2) &&
	    (global_data->keep_press_const != 3)) {
		errout("Constant pressure/density mode can only be 0, 1, 2, or 3 but not %d\n",
	  		global_data->keep_press_const);
	 	goto ctllis_error;
	}

	/*************************************************************/
	/***** all things for flooding *******************************/
	/*************************************************************/
	global_data->flooding		= read_bool_skip(cfd);
	global_data->rot_correction	= read_bool_skip(cfd);
	global_data->flag_adaptive_flood= read_bool_skip(cfd);
	global_data->flood_energy_ini	= read_double_skip(cfd);
	global_data->flood_energy_final	= read_double_skip(cfd);
	global_data->flood_energy_inc	= read_double_skip(cfd);
	global_data->flood_time_const	= read_double_skip(cfd);
	global_data->flood_time_distance= global_data->flood_time_const;
#if DEBUG
	txtout("In read_ctl_file(...): flood_time=%lf\n",global_data->flood_time_const);
#endif	
	/*************************************************************/
	/****** user defined variables, is usefull for development ***/
	/*************************************************************/
	for (i=0;i<MXNO_USERINTS;i++) {
		global_data->userint[i]       = 0;
		global_data->str_userint[i][0]= 0;
	}
	for (i=0;i<MXNO_USERDOUBLES;i++) {
		global_data->userdouble[i]       = 0.0;
		global_data->str_userdouble[i][0]= 0;
	}

/**************New Format ?*************************************************/
	if (new == 0)
	{
          fscanf(cfd,"%d %d",	&global_data->num_userints,
				&global_data->num_userdoubles);
	  read_string(cfd,sdum);

	  if  ( 	(global_data->num_userints<0) ||
		(global_data->num_userints>MXNO_USERINTS) ) {
		errout("Illegal number (%d) of user defined integers.\n",
			global_data->num_userints);
		goto ctllis_error;
	  }	
	  if  ( 	(global_data->num_userdoubles<0) ||
		(global_data->num_userdoubles>MXNO_USERINTS) ) {
		errout("Illegal number (%d) of user defined doubles.\n",
			global_data->num_userdoubles);
		goto ctllis_error;
	  }
	  for (i=0;i<global_data->num_userints;i++) {
		fscanf(cfd,"%d",&(global_data->userint[i]));
		read_string(cfd,sdum);
		sdum[LEN_CTLDESCR-1]=0;
		strcpy(global_data->str_userint[i],sdum);
	  }
	  for (i=0;i<global_data->num_userdoubles;i++) {
#if PRECISION == DOUBLE
		fscanf(cfd,"%le",&(global_data->userdouble[i]));
#else
		fscanf(cfd,"%e",&(global_data->userdouble[i]));
#endif
		read_string(cfd,sdum);
		sdum[LEN_CTLDESCR-1]=0;
		strcpy(global_data->str_userdouble[i],sdum);
	  }
	}
/*****************New format ? end *********************************/
	/*******************************************************************/
	/*** reading parameters for pulling 				 ***/
	/*******************************************************************/
	if (new == 0) 
	  {
	    global_data->flag_selimob = read_bool_skip(cfd);
	    read_string(cfd,global_data->selimob_str[0]);
	  }
	else 
	  {
	    global_data->flag_selimob = read_int_skip(cfd);
	    for (i=0;i<global_data->flag_selimob;i++) {
	      read_string(cfd,global_data->selimob_str[i]);
	    }
	  }
	global_data->selimob_start= read_int_skip(cfd);
	global_data->selimob_const= read_double_skip(cfd);

	global_data->flag_pull    = read_int_skip(cfd);
	if ((new == 0) || ((new == 1) && (global_data->flag_pull > 0)))
	{
	  read_string(cfd,global_data->pull_str[0]);
	  if (global_data->flag_pull==3) read_string(cfd,global_data->pull_str[1]);
	  global_data->pull_vel	  = read_double_skip(cfd);
#if DEBUG
	  txtout("In read_ctl_file(...): pull_vel=%lf\n",global_data->pull_vel);
#endif	
	/*** we expect now the direction given by a vector ***/
	  if (	(global_data->flag_pull==0) ||
		(global_data->flag_pull==1) ||
		(global_data->flag_pull==2) ||
		(global_data->flag_pull==3) ) {
		global_data->pull_dir[0]  = read_double(cfd);
		global_data->pull_dir[1]  = read_double(cfd);		
		global_data->pull_dir[2]  = read_double_skip(cfd);
		global_data->pull_atoms[0]= -1;
		global_data->pull_atoms[1]= -1;
	  } else
	    /*** we expect two atom numbers whose positions define     ***/
	    /*** the direction. the pulling vector is later calculated ***/
	    /*** in the function ini_imob_and_pull(...)	           ***/
	    if (	(global_data->flag_pull==4) ||
			(global_data->flag_pull==5) ||
			(global_data->flag_pull==6)) {
	        global_data->pull_atoms[0] = read_int(cfd);
		global_data->pull_atoms[1] = read_int_skip(cfd);
		if (	(global_data->pull_atoms[0]<1) ||
			(global_data->pull_atoms[1]<1) ||
			(global_data->pull_atoms[0]>global_data->total_no_atoms) ||
			(global_data->pull_atoms[1]>global_data->total_no_atoms)) {
			errout("Wrong atom selection for pulling direction\n");
			goto ctllis_error;			
		}
	    }
	
	  global_data->pull_const[0]= read_double(cfd);		
	  global_data->pull_const[1]= read_double_skip(cfd);
	}
	else
	{
	  global_data->pull_str[0][0] = 0;
	  global_data->pull_vel = 0.00;
	  global_data->pull_dir[0] = 1.000000;
	  global_data->pull_dir[1] = 0.000000;
	  global_data->pull_dir[2] = 0.000000;
	  global_data->pull_atoms[0]= -1;
	  global_data->pull_atoms[1]= -1;
	}
	/*********************Rotation modi**************************************/
	if (new == 1)
	{
	  global_data->rotation_mode = read_int_skip(cfd);
	  if (global_data->rotation_mode > 0)
	  {
	    read_string(cfd,global_data->rotation_str);
	    global_data->angular_vel = read_double_skip(cfd); /*Angular velocity */
	    global_data->rotation_dir[0] = read_double(cfd);  /* Rotation direction X */
	    global_data->rotation_dir[1] = read_double(cfd);  /* Rotation direction Y */
	    global_data->rotation_dir[2] = read_double_skip(cfd);  /* Rotation direction Z */
	    global_data->rotation_const[0] = read_double(cfd);/*Spring constants for rotation (||,T) */
	    global_data->rotation_const[1] = read_double(cfd);
	    global_data->rotation_const[2] = read_double_skip(cfd);
	    if (global_data->rotation_mode ==2) global_data->rotation_sigma = read_double_skip(cfd);
	    if (global_data->rotation_mode ==1)
	      {
		global_data->rotation_point[0] = read_double(cfd);  /* point on rotation axis X */
		global_data->rotation_point[1] = read_double(cfd);  /* point on rotation axis Y */
		global_data->rotation_point[2] = read_double_skip(cfd);  /* point on rotation axis Z */
	      }
	  }
	}
	/************************Rotation input end *****************************/
/**************User defined variables have new position in new ctl.lis format ? *************************************************/
	if (new == 1)
	{
          fscanf(cfd,"%d %d",	&global_data->num_userints,
				&global_data->num_userdoubles);
	  read_string(cfd,sdum);

	  if  ( 	(global_data->num_userints<0) ||
		(global_data->num_userints>MXNO_USERINTS) ) {
		errout("Illegal number (%d) of user defined integers.\n",
			global_data->num_userints);
		goto ctllis_error;
	  }	
	  if  ( 	(global_data->num_userdoubles<0) ||
		(global_data->num_userdoubles>MXNO_USERINTS) ) {
		errout("Illegal number (%d) of user defined doubles.\n",
			global_data->num_userdoubles);
		goto ctllis_error;
	  }
	  for (i=0;i<global_data->num_userints;i++) {
		fscanf(cfd,"%d",&(global_data->userint[i]));
		read_string(cfd,sdum);
		sdum[LEN_CTLDESCR-1]=0;
		strcpy(global_data->str_userint[i],sdum);
	  }
	  for (i=0;i<global_data->num_userdoubles;i++) {
#if PRECISION == DOUBLE
		fscanf(cfd,"%le",&(global_data->userdouble[i]));
#else
		fscanf(cfd,"%e",&(global_data->userdouble[i]));
#endif
		read_string(cfd,sdum);
		sdum[LEN_CTLDESCR-1]=0;
		strcpy(global_data->str_userdouble[i],sdum);
	  }
	}
/*****************User defined variables have new position in new ctl.lis format ? end *********************************/

    	fscanf(cfd,"%s",end_string);
	if (strcmp("END",end_string) != 0) {
		errout("Too few input data in your ctl.lis file.\n");
		goto ctllis_error;
	}

                for (k=0;k<2;k++) {
                /*       global_data->indpull[k] = alloc_ints(MXNO_PULLING_ATOMS);
                      global_data->mass_pull[k] = alloc_doubles(MXNO_PULLING_ATOMS); */

                    for (i=0;i<MXNO_PULLING_ATOMS;i++) {
                        global_data->indpull[k][i]=0;
                        global_data->mass_pull[k][i]=0.0;
                    }
                }

	
	/****** now we scan in some debug-flags, if some are there **********/
	for (i=0;i<sizeof(global_data->debug_flags)/sizeof(int);i++)
		global_data->debug_flags[i] = 0;
	global_data->debug_flags[0] = read_int_skip(cfd);/** exact force: 0 off, 1 on **/
	global_data->debug_flags[1] = read_int_skip(cfd);/** output of force-file **/

	/*************************************************************/
	/*********** some checks whether values are ok ***************/
	/*************************************************************/
	if (	(global_data->stochastic<0) 	||
		(global_data->harmonic<0)   	||
 		(global_data->shake<0)		||
		(global_data->equilibration<0)	||
		(global_data->minimisation<0)	||
		(global_data->flooding<0)	||
		(global_data->rot_correction<0)	||
		(global_data->flag_selimob<0)	||
		(global_data->flag_adaptive_flood<0) ) {
		errout("Fatal error: Some of the BOOL vars are incorrect !\n");
		goto ctllis_error;
	} else
	if ((global_data->flag_pull<0)||(global_data->flag_pull>5)) {
		errout("Fatal error: Pulling modus can be only 0,1,2,3,4,5!\n");
		goto ctllis_error;		
	} else
	if ((global_data->rotation_mode<0) || (global_data->rotation_mode>2)) {
	  errout("FATAL error: Rotation Mode is restricted to 0..2!\n");
	  goto ctllis_error;
        } else
	if (global_data->selimob_start<0) {
		errout("Fatal error: Start integration step for imobilisation/pulling is illegal!\n");
		goto ctllis_error;
	} else
	if (global_data->pull_vel<0.0) {
		errout("Fatal error: Pullvelocity must be positive!\n");
		goto ctllis_error;
	} else
	if (	(global_data->pull_const[0]<0.0) ||
		(global_data->pull_const[1]<0.0) ) {
		errout("Fatal error: Harmonic constants for pulling must be positive!\n");
		goto ctllis_error;
	} else
	if (global_data->no_nodes<1) {
		errout("Number of nodes must be greater than 1 !\n");
		goto ctllis_error;
	} else
    	if (global_data->total_no_atoms<=0) {
		errout("Fatal error: Impossible number of atoms in molecule !\n");
		goto ctllis_error;
    	} else
    	if (global_data->Analyse_nth<0) {
		errout("Wrong number for analysis dutycycle ! Must be greater than 0.\n");
		goto ctllis_error;
    	} else
    	if (global_data->Restart_nth<5) {
		errout("Wrong number for restart dutycycle ! Must be greater than 4.\n");
		goto ctllis_error;
    	} else
	if (global_data->step_of_energy<0) {
		errout("Negative value (%d) for 'Freq. of energy printout' is illegal.\n",
			global_data->step_of_energy);
		errout("You probably use a old ctl.lis-file !\n");
		goto ctllis_error;
	}
    	if ((global_data->time_step * TIMEDIV) < MIN_TIME_STEP) {
		errout("Timestep (%f<%f) too small!\n",
		        global_data->time_step,MIN_TIME_STEP);
		goto ctllis_error;
    	} else
    	if (	(global_data->ex_number<-5) || 
		(global_data->ex_number> 5) || 
		(global_data->ex_number==0) ) {
		errout("Exclusion order (%d) out of range!\n",global_data->ex_number);
		goto ctllis_error;
    	} else
    	if (global_data->equilibration && (global_data->target_temp < ZERO)) {
		errout("Target temperature can\'t be negative!\n");
		goto ctllis_error;
    	} else
    	if (global_data->equilibration && global_data->minimisation) 
		errout("Equilibration AND minimization -- I don't see the point in it, but\nif you want it, you can have it\n");
    	else
    	if (	(global_data->equilibration) && 
		(global_data->tau_T < global_data->time_step*TIMEDIV)) {
		errout("Temperature coupling time constant too small\n");
		goto ctllis_error;
    	} else
    	if (strlen(global_data->outpath) <= 0) {
		errout("No output path given!\n");
		goto ctllis_error;
    	} else
	if ((global_data->treeinfo.algo>2)||(global_data->treeinfo.algo<0)) {
		errout("You can only choose 0,1 or 2 for cluster algorithm.\n");
		goto ctllis_error;
	} else
	if (	(global_data->treeinfo.levels>MAX_NO_LEVELS) ||
		(global_data->treeinfo.levels<-MAX_NO_LEVELS)||
		(global_data->treeinfo.levels==0)) {
		errout("Illegal number of levels.");
		goto ctllis_error;
	} else
	if (global_data->treeinfo.steps<5) {
		errout("Step for reclustering to small. Must be greater than 5.\n");
		errout("Reasonable values range between 300 and 600.\n");		
		goto ctllis_error;
	}
	else {
		int i;	
		for (i=0;i<MXNO_SBOUND_PARAM;i++) {
    			if (global_data->sbound_droff[i]==-99999.0) {
      				errout("Problem reading SBOUND value!\n");
				goto ctllis_error;
			}
			if (fabs(global_data->sbound_droff[i])<=SMALL) 
				global_data->sbound_droff[i]=0.0;
		}
    	}
 
    	fclose(cfd); /*** now the reading of the control file is done. ***/
    	return 0;

ctllis_error:
	fclose(cfd);
	errout("Check parameters in your control file (ctl.lis).\n");	
	errout("Maybe you erased or inserted a line in the control file.\n");
	pex(-1);
	return -1;
}


/*** here we read in the last section of ctl file (the section	***/
/*** which is in the free format 				***/
int
read_ctl_file_free_format_sec(	int nr_of_atoms,
			char *fname,
			STRUCT_ATOMDATA *atom_data,
			int *atom_sele_flag1,
			LOAD_MSG_ALL_NODE *global_data,
			STRUCT_FORCES *my_force)
{
	int	i,j,k,len,mask=0,ccoment=0;
	int	*used_atoms;
	FILE	*h_force;

	if ((h_force=fopen(fname,"r"))==NULL) {
		errout("Fatal error: Can not open control file <%s>.\n",fname);
		pex(-1);
	}
	
	while (!feof(h_force)) {	/*** scan the file up to the keyword END ***/
		read_string(h_force,sdum);
		len = strlen(sdum);
		if (strcmp(sdum,"END")==0) {
			read_string(h_force,sdum);
			read_string(h_force,sdum);		
			break;
		}
	}
	my_force->qm.asize     = alloc_doubles(global_data->total_no_atoms);
	my_force->qm.qcore     = alloc_doubles(global_data->total_no_atoms);
	
#if (DEBUG == 3)
printf("Before QM stuff ini, no_atoms=%d\n",global_data->total_no_atoms);
#endif
	for (i=0;i<global_data->total_no_atoms;i++) {
		char atomchar = global_data->atomdata[i].atomstr[0];

#if (DEBUG == 3)
printf("In QM stuff, atomchar=%c\n", atomchar);
#endif
		switch(atomchar) {
			case 'H': my_force->qm.qcore[i] = 1.0; break;
			case 'C': my_force->qm.qcore[i] = 6.0; break;
			case 'O': my_force->qm.qcore[i] = 8.0; break;	
			case 'N': my_force->qm.qcore[i] = 7.0; break;			
			case 'P': my_force->qm.qcore[i] =15.0; break;			
			default : my_force->qm.qcore[i] = 1.0; break;
		}
		
		switch(atomchar) {
			case 'H': my_force->qm.asize[i] = -0.8; break;
			case 'C': my_force->qm.asize[i] = -0.8;  break;
			case 'O': my_force->qm.asize[i] = -0.8;  break;	
			case 'N': my_force->qm.asize[i] = -0.8;  break;			
			case 'P': my_force->qm.asize[i] = -0.8;  break;			
			default : my_force->qm.asize[i] = -0.8;  break;
		}
	}	
#if (DEBUG == 3)
printf("QM stuff ini done\n");
#endif

	used_atoms = alloc_ints(nr_of_atoms);
	/*************************************************/
	/*** read each line in the free format section ***/
	/*************************************************/
	while (!feof(h_force)) { 
		read_string(h_force,sdum);
		len = strlen(sdum);
#if DEBUG
txtout("In read_ctl_file_free_format_sec(...): <%s>\n",sdum);
#endif
		for (i=0;i<len;i++)	/*** scan to first space or tab ***/
			if ((sdum[i]!=' ')&&(sdum[i]!='\t')) break;
		for (k=i,j=0;k<len;k++,j++) /*** delete leading white spaces **/	
			sdum[j] = sdum[k];
		sdum[j] = 0;
		len = strlen(sdum);
		if (len<1)	continue; /*** line of white spaces, read next line ***/

		/*** first check for comment lines with ! and # ***/
		if ((sdum[0]=='!')||(sdum[0]=='#')) continue; /*** read next line ***/
		
		/*** check for C-Comment-Style ***/
		if ((ccoment==0)&&(sdum[0]=='/')&&(sdum[1]=='*')) {
			for (i=1;i<len-1;i++)
				if ((sdum[i]=='*')&&(sdum[i+1]=='/')) break;
			if ((sdum[i]=='*')&&(sdum[i+1]=='/')) continue; /*** read next line ***/
			ccoment=1;
			continue;	/*** read next line ***/
		} else
		if (ccoment==1) {
			for (i=0;i<len-1;i++)
				if ((sdum[i]=='*')&&(sdum[i+1]=='/')) break;
			if ((sdum[i]=='*')&&(sdum[i+1]=='/')) {
				ccoment=0;
			}
			continue;	/*** read next line ***/
		}
		
		/*** now we have something to analyze	***/
		/*** seperate the keyword from the rest	***/
		for (i=0;i<len;i++)  /*** scan to next space or tab ***/
			if ((sdum[i]==' ')||(sdum[i]=='\t')) {
				sdum[i]=0; /*** now we have the keyword ***/	
				break;
			}
	
		for (j=i+1;j<len;j++) 
			if ((sdum[j]!=' ')&&(sdum[j]!='\t')) break;
				
		/*** terminate line at first '!' or '#' ***/
		for (k=j;k<len;k++) 
			if ((sdum[k]=='!')||(sdum[k]=='#')) {
				sdum[k] = 0;
				break;
			}
			
		len = strlen(sdum+j);
		strup(sdum);	/*** only capital letters ***/

		/*******************************************/	
		/*** sdum now points to the keyword, and ***/
		/*** sdum+j is the rest of the line      ***/
		/*******************************************/	
#if DEBUG
txtout("In read_ctl_file_free_format_sec(...): keyword: <%s> rest-string <%s>\n",sdum,sdum+j);
#endif	

		if (parse_specforce(sdum,
				sdum+j,
				atom_data,
				atom_sele_flag1,
				global_data,
				my_force,
				used_atoms)==1) {
			continue;		
		}

		/***********************************************/
		/*** Set the ID-String for output file names ***/
		/***********************************************/
		if (strcmp(sdum,"IDSTRING")==0) {	
			strcpy(global_data->idstr,sdum+j);
			for (k=0;k<strlen(global_data->idstr);k++) 
				if (	(global_data->idstr[k]==' ') ||
					(global_data->idstr[k]=='\t') ) {
					global_data->idstr[k] = 0;
					break;
				}
			continue;			
		}

		/***********************************************/
		/*** Set stop minimisation mode              ***/
		/***********************************************/
		if (strcmp(sdum,"TRANSROTCOR")==0) {
			if (strcmp(sdum+j,"TRANS")==0) {
				global_data->newtransrotcor=1;
				txtout("Only translation correction switched on.\n");
				continue;
			}
			if (strcmp(sdum+j,"ROT")==0) {
				global_data->newtransrotcor=2;
				txtout("Only rotation correction switched on.\n");
				continue;
			}
			if (	(strcmp(sdum+j,"TRANSROT")==0) ||
				(strcmp(sdum+j,"TRUE")==0) ||
				(sdum[j]==0)	) {
				global_data->newtransrotcor=3;
				txtout("Translation and rotation correction switched on.\n");
				continue;
			}
			errout("Fatal error: Illegal TRANSROTCOR mode.\n");
			pex(-1);
		}

		/***********************************************/
		/*** Set stop minimisation mode              ***/
		/***********************************************/
		if (strcmp(sdum,"MINIMODE")==0) {
			if (strcmp(sdum+j,"GRADIENT")==0) {
				global_data->minimisation=2;
				txtout("Minimization mode is GRADIENT\n");
				continue;
			}
			if (strcmp(sdum+j,"NEWTON")==0) {
				global_data->minimisation=3;
				txtout("Minimization mode is NEWTON\n");
				continue;
			}
			if (strcmp(sdum+j,"SPECIAL")==0) {
				global_data->minimisation=4;
				txtout("Minimization mode is SPECIAL\n");
				continue;
			}
			errout("Fatal error: Illegal Minimisation mode.\n");
			pex(-1);			
		}

		/***********************************************/
		/*** Set stop criterion for minimisation     ***/
		/***********************************************/
		if (strcmp(sdum,"MINIKRIT")==0) {	
#if PRECISION == DOUBLE
			if (sscanf(sdum+j,"%lf %lf",&global_data->minikrit[0],&global_data->minikrit[1])!=2) {
#else
			if (sscanf(sdum+j,"%f %f",&global_data->minikrit[0],&global_data->minikrit[1])!=2) {
#endif
				errout("Fatal error: No value for minimisation criterion given.\n");
				pex(-1);	
			}
			txtout("Setting minimisation criterions to %g (max) and %g (ave) kcal/mol A.\n",
				global_data->minikrit[0],global_data->minikrit[1]);
			continue;
		}

		/*** Set RANGE for SBOUND ***/
		if (strcmp(sdum,"SBOUNDMODE")==0) {	
			parse_sboundmode(fname,
					 sdum+j,
					 global_data);
			continue;
		}

		/************************************************/
		/*** Set start integration step for write-out ***/
		/************************************************/
		if (strcmp(sdum,"START")==0) {	
			if (sscanf(sdum+j,"%d",&global_data->forceout_start_step)!=1) {
				errout("Fatal error: No START integration step specified.\n");
				pex(-1);	
			}
			continue;
		}

		/**************************************************/
		/*** Set stride of writing out forces 		***/
		/**************************************************/
		if (strcmp(sdum,"STRIDE")==0) {
			if (sscanf(sdum+j,"%d",&global_data->forceout_stride_step)!=1) {
				errout("Fatal error: No stride specified in <%s>.\n",
					fname);
				pex(-1);	
			}
			continue;
		}
		
		if (strcmp(sdum,"PICKBOND")==0) {
			parse_pickbond(	sdum+j,
					atom_data,
					atom_sele_flag1,
					global_data,
					&(my_force->pickbond),
					PICK_BOND);
			continue;
		}

		if (strcmp(sdum,"PICKELEC")==0) {
			parse_pickbond(	sdum+j,
					atom_data,
					atom_sele_flag1,
					global_data,
					&(my_force->pickelec),
					PICK_ELEC);
			continue;
		}

		if (strcmp(sdum,"PICKVDW")==0) {
			parse_pickbond(	sdum+j,
					atom_data,
					atom_sele_flag1,
					global_data,
					&(my_force->pickvdw),
					PICK_VDW);
			continue;
		}

		if (strcmp(sdum,"PICKANGLE")==0) {
			parse_pickangle(sdum+j,
					atom_data,
					atom_sele_flag1,
					global_data,
					&(my_force->pickangle),
					PICK_ANGLE);
			continue;
		}

		if (strcmp(sdum,"PICKDIHE")==0) {
			parse_pickdihe(	sdum+j,
					atom_data,
					atom_sele_flag1,
					global_data,
					&(my_force->pickdihe),
					PICK_DIHE);
			continue;
		}

		if (strcmp(sdum,"PICKIMPR")==0) {
			parse_pickdihe(	sdum+j,
					atom_data,
					atom_sele_flag1,
					global_data,
					&(my_force->pickimpr),
					PICK_IMPR);
			continue;
		}

		if (	(strcmp(sdum,"SCHWING")==0) ||
			(strcmp(sdum,"HESS")==0)) {
			parse_schwing(sdum,sdum+j,
					atom_data,
					global_data);
			continue;
		}

                if (strcmp(sdum,"EPSILON")==0) {
                        double myeps=1.0;
#if PRECISION == DOUBLE
			sscanf(sdum+j,"%lf",&myeps);
#else
			sscanf(sdum+j,"%f",&myeps);
#endif                   
                    
                        global_data->f_elstat=332.0636/myeps;
                        txtout("Dielectric constant set to %f.\n",myeps);
		        continue;
                }

		if (strcmp(sdum,"FLOODSCALE")==0) {
			double scale=0.0;
		
#if PRECISION == DOUBLE
			sscanf(sdum+j,"%lf",&scale);
#else
			sscanf(sdum+j,"%f",&scale);
#endif
			global_data->flood_scale=scale;
			txtout("Flood_scale=%g assigned.\n",scale);
			continue;	
		}

		if (strcmp(sdum,"FLOODTIMEDISTANCE")==0) {
			double scale=0.0;
		
#if PRECISION == DOUBLE
			sscanf(sdum+j,"%lf",&scale);
#else
			sscanf(sdum+j,"%f",&scale);
#endif
			global_data->flood_time_distance=scale;
			txtout("Flood_time_distance=%g assigned.\n",scale);
			continue;	
		}

		if (parse_qminterface(	sdum,
					sdum+j,
					atom_data,
					atom_sele_flag1,
					global_data,
					my_force,
					used_atoms)==1) continue;

		/**************************************************/
		/*** now we manage the atom selection strings	***/
		/**************************************************/
		mask = 0;
		if (strcmp(sdum,"TOTAL")==0)		mask = FORCE_TOTAL;
		if (strcmp(sdum,"ELECVDW")==0)		mask = FORCE_ELECVDW;
		if (strcmp(sdum,"BOND")==0)		mask = FORCE_BOND;
		if (strcmp(sdum,"ELEC")==0)		mask = FORCE_ELEC;
		if (strcmp(sdum,"VDW")==0)		mask = FORCE_VDW;
		if (strcmp(sdum,"ANGLE")==0)		mask = FORCE_ANGLE;
		if (strcmp(sdum,"DIHE")==0)		mask = FORCE_DIHE;
		if (strcmp(sdum,"IMPR")==0)		mask = FORCE_IMPR;
		if (strcmp(sdum,"RESTR")==0)		mask = FORCE_RESTR;
		if (strcmp(sdum,"SBOUND")==0)		mask = FORCE_SBOUND;
		if (strcmp(sdum,"INTERN")==0)		mask = FORCE_INTERN;
		if (strcmp(sdum,"FLOOD")==0)		mask = FORCE_FLOOD;
		if (strcmp(sdum,"KIN")==0)		mask = FORCE_KIN;

		if (strcmp(sdum,"QMATOM")==0)		mask = FLAG_QMATOM;
		if (strcmp(sdum,"SELSCHWING")==0)	mask = FLAG_SCHWING;
		if (strcmp(sdum,"SELHESS")==0)		mask = FLAG_SCHWING;
				
		if (strcmp(sdum,"SHAKEOFF")==0)		mask = FLAG_SHAKEOFF;
		if (strcmp(sdum,"FORCEOFF")==0)		mask = FLAG_FORCEOFF;
		if (strcmp(sdum,"SPEEDOFF")==0)		mask = FLAG_SPEEDOFF;		
		if (strcmp(sdum,"PART1")==0)		mask = FLAG_PART1;
		if (strcmp(sdum,"PART2")==0)		mask = FLAG_PART2;

		if (mask==0) {
			errout("Fatal error: Illegal line in the free format section\n");
			errout("             of the control file.\n");
			errout("The line is: <%s>\n",sdum);
			pex(-1);
		}
		
		/*** now we analyse the atom-selection string ***/
		for (i=0;i<nr_of_atoms;i++) used_atoms[i] = 0;
		if (set_used_atoms(sdum+j,nr_of_atoms,atom_data,used_atoms)<0) {
			errout("Fatal error: Correct this error in your control file <%s>\n",sdum);
			pex(-1);
		}
		
		/*** now used_atoms[...] is 1 for every atom which is selected ***/
		/*** lets set the bit-pattern in the atom_sele_flag1[...] array ***/
		for (i=0,j=0;i<nr_of_atoms;i++) {
			if (used_atoms[i]) {
				j++;
				atom_sele_flag1[i] |= mask;
			}	
		}
		
#if DEBUG_FORCEOUT
txtout("%d atoms selected for %s\n",j,sdum);
#endif
	}
	fclose(h_force);
	myfree(used_atoms);


	/*** Info for free energy calculation ***/
	for (i=0,j=0;i<nr_of_atoms;i++)
		if (atom_sele_flag1[i]&(FLAG_PART1|FLAG_PART2)) j++;
	if ((global_data->num_userdoubles>=1)&&(j!=0)) {
		txtout("Selected %d atoms for free energy calculation with LAMBDA %f.\n",
			j,global_data->userdouble[0]);
	} else 
	if (j!=0) {
		errout("Fatal error: No LAMBDA defined for free energy calculation.\n");
		pex(-1);
	}

	/*** count the atoms with switched of forces ***/
	for (i=0,j=0;i<nr_of_atoms;i++)
		if (atom_sele_flag1[i]&FLAG_FORCEOFF) j++;
	txtout("Selected %d atoms with forces switched off.\n",j);

	/*** count the atoms with switched of velocity ***/
	for (i=0,j=0;i<nr_of_atoms;i++)
		if (atom_sele_flag1[i]&FLAG_SPEEDOFF) j++;
	txtout("Selected %d atoms with velocity switched off.\n",j);

	/*** count the atoms with normal mode analysis ***/
	for (i=0,j=0;i<nr_of_atoms;i++)
		if (atom_sele_flag1[i]&FLAG_SCHWING) j++;
	txtout("Selected %d atoms for calculation of Hesse matrix.\n",j);
	
	if (	(global_data->qmchargemode==QMCHARGEMODE_VARIABLE) &&
		(global_data->qmcontin_step<=1) ) {
		global_data->qmcontin_step=2;
		txtout("Setting QMCONTINUE to 2!\n");
	}
	
#if defined(SEQ)
	if (j>0) {
		global_data->cycles_passed=0;
		global_data->shake        =FALSE;
		global_data->minimisation =FALSE;
		global_data->equilibration=FALSE;
		global_data->stochastic   =FALSE;
		global_data->flooding     =FALSE;
		global_data->rot_correction=FALSE;
		global_data->flag_pull    =0;
		global_data->flag_selimob =FALSE;
	}
#else
	if (j>0) {
		txtout("***********************************************\n");
		txtout("Fatal error: Normal-Mode analysis works only\n");
		txtout("             in the sequential version of EGO!\n");
		txtout("***********************************************\n");
		pex(-1);
	}
#endif
	if (j!=0) 
		global_data->final_cycle_no = global_data->cycles_passed + 
			3*j*abs(global_data->schwing_mode);
	
	/*** count the atoms included to the QM-part ***/
	for (i=0,global_data->nr_qmatoms=0;i<nr_of_atoms;i++)
		if (atom_sele_flag1[i]&FLAG_QMATOM) 
			global_data->nr_qmatoms++;
#if defined(SEQ)

#else
	if (global_data->nr_qmatoms>0) {
		txtout("***********************************************\n");
		txtout("Fatal error: QM-interface works only in the\n");
		txtout("             sequential version of EGO!\n");
		txtout("***********************************************\n");
		pex(-1);
	}
#endif					
	txtout("Selected %d atoms for QM calculation.\n",
		global_data->nr_qmatoms);

	set_qmboundbox(atom_sele_flag1,atom_data,global_data);

	/*** count the atoms for which we need any forces ***/
	for (i=0,j=0;i<nr_of_atoms;i++) 
		if (atom_sele_flag1[i]&(~(FLAG_PART1 |
					 FLAG_PART2 |
					 FLAG_FORCEOFF |
					 FLAG_SPEEDOFF |
					 FLAG_SHAKEOFF ))) j++;

	global_data->forceout_nr_atoms = j;
	
	txtout("Forces of %d atoms will be written every %d integration step\n",
			global_data->forceout_nr_atoms,
			global_data->forceout_stride_step);
	txtout("starting with integration step %d.\n",
			global_data->forceout_start_step);
	return 0;
}


int read_flood_file(	char *fname,
			LOAD_MSG_ALL_NODE *global_data)
{
	FILE	*h_floodfile;
	int	i,k,j;
	int	cmptr=0,num_usedatoms=0;
	int	test_nr_atoms;

	/*** we need the flooding.lis also for rot_correction, since we ***/
	/*** need global_data->flood_nr_atoms 				***/
	if ((!global_data->flooding) && (!global_data->rot_correction)) {	
		     /**** set some dummy-data *****/
		global_data->nr_of_flooding_gaussians = 0;		
		return 0;
	}

	if ((h_floodfile=fopen(fname,"r"))==NULL) {
		errout("Fatal error: Can not open flooding-file <%s>.\n",fname);
		errout("If you want to use translatation/rotation correction without any\n");
		errout("flooding, you have to create an empty flooding file via 'mkflood' !\n");
		errout("Call mkflood without parameters for further instrucions.\n");
		pex(-1);
	}	
	txtout("Reading flood/rotcor-file : <%s>.\n",fname);

	/***************************************************/
	/***** reading the header **************************/
	/***************************************************/
	skip_line(h_floodfile,1);
	global_data->nr_of_flooding_gaussians	= read_int_skip (h_floodfile);
	if (global_data->nr_of_flooding_gaussians >= 10) {
		errout("Fatal error in flooding file: Too many centers for flooding !\n");
		pex(-1);
	}

	/***** this is just a basic check *********/
	test_nr_atoms				= read_int_skip(h_floodfile);
	if (test_nr_atoms!=global_data->total_no_atoms) {
		errout("Fatal error: Different number of atoms while reading flooding data ! %d != %d\n",
			test_nr_atoms,global_data->total_no_atoms);
		errout("Maybe you use the wrong flooding-file for this simulation !?\n");
		pex(-1);
	}

	global_data->flood_nr_atoms 		= read_int_skip(h_floodfile);

	/******************************************************/
	/*** reading the list which specifies which          **/
	/*** atoms the flooding potential or the transl/rot  **/
	/*** correction has to affect			     **/
	/*****************************************************/
	skip_line(h_floodfile,1);
	global_data->flood_used_atoms = alloc_ints(test_nr_atoms);
	for (i=0,num_usedatoms=0;i<test_nr_atoms;i++) {
		fscanf(h_floodfile,"%d",global_data->flood_used_atoms+i);
		num_usedatoms += global_data->flood_used_atoms[i];
	}

	if (num_usedatoms==0) {
		errout("Fatal error: In flooding.lis are no atoms selected!\n");
		errout("Re-run mkflood again and select some atoms!\n");
		pex(-1);	
	}

	if (global_data->nr_of_flooding_gaussians==0) { /*** this may happen if only ***/
		fclose(h_floodfile);			/*** trans/rot correction ***/	
		return 0;
	}
	
	/***************************************************/
	/*** read the koordiante of the center of the    ***/
	/*** conformational substate and the flooding-   ***/
	/*** matrix. 					 ***/
	/***************************************************/
	for (k=0;k<global_data->nr_of_flooding_gaussians;k++) {
		skip_line(h_floodfile,2);

		global_data->flood_cm[k] = alloc_doubles(3*global_data->flood_nr_atoms);

		/*** now we readin center of gaussians ***/
		for (i=0,cmptr=0;i<global_data->flood_nr_atoms;i++,cmptr+=3) {
			global_data->flood_cm[k][0+cmptr]=read_double(h_floodfile);
			global_data->flood_cm[k][1+cmptr]=read_double(h_floodfile);
			global_data->flood_cm[k][2+cmptr]=read_double(h_floodfile);
		}

		skip_line(h_floodfile,2);
		global_data->flood_matrix[k] = alloc_doubles(3*3*
				global_data->flood_nr_atoms*global_data->flood_nr_atoms);

		/*** now we readin the covarianz-matrix == flooding-matrix ***/
		for (i=0;i<3*global_data->flood_nr_atoms;i++) {
			int offs=i*3*global_data->flood_nr_atoms;
			for (j=0;j<3*global_data->flood_nr_atoms;j++) {
				global_data->flood_matrix[k][offs+j]=read_double(h_floodfile);
			}
		}		
	}
	fclose(h_floodfile);

	/*************************************/
	/*** print to screen what we got *****/
	/*************************************/
	txtout("Number of flood-Gaussians: %d\n",global_data->nr_of_flooding_gaussians);
	txtout("Number of flooded atoms  : %d\n",global_data->flood_nr_atoms);
	txtout("Start of flood used-list : %d %d %d %d %d %d\n",
		global_data->flood_used_atoms[0],global_data->flood_used_atoms[1],
		global_data->flood_used_atoms[2],global_data->flood_used_atoms[3],
		global_data->flood_used_atoms[4],global_data->flood_used_atoms[5]);

	for (k=0;k<global_data->nr_of_flooding_gaussians;k++) {
		txtout("Start of flood cm  %d      : %f %f %f   %f %f %f\n",k,
			global_data->flood_cm[k][0],global_data->flood_cm[k][1],
			global_data->flood_cm[k][2],global_data->flood_cm[k][3],
			global_data->flood_cm[k][4],global_data->flood_cm[k][5]);
		txtout("Start of flood matrix  %d  : %f %f %f   %f %f %f\n",k,
			global_data->flood_matrix[k][0],
			global_data->flood_matrix[k][1],
			global_data->flood_matrix[k][2],
			global_data->flood_matrix[k][3],
			global_data->flood_matrix[k][4],
			global_data->flood_matrix[k][5]);
	}
	return 0;	
}



int read_boundary_file(	char *fname,
			LOAD_MSG_ALL_NODE *global_data)
{
	FILE	*h_boundary;
	int	i,k,j;
	double  ddummy;

	/*** set default values ***/
	global_data->arbs_ncodebook=1;
	global_data->arbs_water_shell_thickness=0.0;
	global_data->arbs_surface_level=0.0;
	for (i=0; i<global_data->arbs_ncodebook; i++) {
		global_data->arbs_gauss_height[i]=0.0;
		}
	for (i=0; i<global_data->arbs_ncodebook; i++) {
		for (j=0; j<3; j++) {
			global_data->arbs_gauss_y[j][i]=0.0;
			global_data->arbs_gauss_rho2[j][i]=0.0;
			}
		}
	for (i=0; i<global_data->arbs_ncodebook; i++) {
		for (j=0; j<3; j++) {
			for (k=0; k<3; k++) {
				global_data->arbs_gauss_w[j][k][i]=0.0;
				}
			}
		}

	/*** we need the cboundary.lis for sbound with arbitrarily shaped boundary ***/
	if (global_data->sbound_droff[R_SPHERE]>= -SMALL) return 0;  /* no arboundary needed */

	if ((h_boundary=fopen(fname,"r"))==NULL) {
		errout("Fatal error: Can not open boundary-file <%s>.\n",fname);
		errout("Use Solvate 1.2 to create a boundary-file!\n\n");
		pex(-1);
	}	
	txtout("Reading boundary-file        : <%s>.\n",fname);

	/***************************************************/
	/***** reading the header **************************/
	/***************************************************/
	read_string(h_boundary,sdum);
	if (strcmp(sdum,"[MINIMUM WATER SHELL THICKNESS]")!=0) {
		errout("<%s> is not a valid boundary-file.\n",fname);
		errout("(Token [MINIMUM WATER SHELL THICKNESS] expected)\n\n");
		pex(-1);
		}
        global_data->arbs_water_shell_thickness = read_double_skip(h_boundary);
	if (global_data->arbs_water_shell_thickness < 1.0) {
		errout("A water shell thickness smaller than 1 A (%f A)\n",global_data->arbs_water_shell_thickness);
		errout("is strange and not allowed.\n\n",fname);
		pex(-1);
		}
	if (global_data->arbs_water_shell_thickness > 1000.0) {
		errout("A water shell thickness larger than 1000 A (%f A)\n",global_data->arbs_water_shell_thickness);
		errout("is strange and not allowed.\n\n",fname);
		pex(-1);
		}

	read_string(h_boundary,sdum);
	if (strcmp(sdum,"[ENCLOSING SPHERE (X,Y,Z),R]")!=0) {
		errout("<%s> is not a valid boundary-file.\n",fname);
		errout("<%s>.\n",sdum);
		errout("(Token [ENCLOSING SPHERE (X,Y,Z),R] expected)\n\n");
		pex(-1);
		}
        ddummy = read_double_skip(h_boundary);
        ddummy = read_double_skip(h_boundary);

	read_string(h_boundary,sdum);
	if (strcmp(sdum,"[ENCLOSING BOX (X,Y,Z),(XMIN,XMAX),(YMIN,YMAX),(ZMIN,ZMAX)]")!=0) {
		errout("<%s> is not a valid boundary-file.\n",fname);
		errout("<%s>.\n",sdum);
		errout("(Token [ENCLOSING BOX (X,Y,Z),(XMIN,XMAX),(YMIN,YMAX),(ZMIN,ZMAX)] expected)\n\n");
		pex(-1);
		}
        ddummy = read_double_skip(h_boundary);
        ddummy = read_double_skip(h_boundary);
        ddummy = read_double_skip(h_boundary);
        ddummy = read_double_skip(h_boundary);

	read_string(h_boundary,sdum);
	if (strcmp(sdum,"[NR OF GAUSSIANS]")!=0) {
		errout("<%s> is not a valid boundary-file.\n",fname);
		errout("<%s>.\n",sdum);
		errout("(Token [NR OF GAUSSIANS] expected)\n\n");
		pex(-1);
		}
        global_data->arbs_ncodebook = read_int_skip(h_boundary);
	if (global_data->arbs_ncodebook < 1) {
		errout("There must be at least one gaussian in the boundary-file.\n");
		errout("Found %d.\n\n",global_data->arbs_ncodebook);
		pex(-1);
		}
	if (global_data->arbs_ncodebook > MXNO_ARBOUND_GAUSSIANS) {
		errout("There must be at most %d gaussians in the boundary-file.\n",MXNO_ARBOUND_GAUSSIANS);
		errout("Found %d.\n\n",global_data->arbs_ncodebook);
		pex(-1);
		}

	read_string(h_boundary,sdum);
	if (strcmp(sdum,"[SURFACE LEVEL]")!=0) {
		errout("<%s> is not a valid boundary-file.\n",fname);
		errout("<%s>.\n",sdum);
		errout("(Token [SURFACE LEVEL] expected)\n\n");
		pex(-1);
		}
        global_data->arbs_surface_level = read_double_skip(h_boundary);
	if (global_data->arbs_surface_level <= 0.0) {
		errout("A non-positive surface level (%f)\n",global_data->arbs_surface_level);
		errout("is strange and not allowed.\n\n",fname);
		pex(-1);
		}
	if (global_data->arbs_surface_level > 100.0) {
		errout("A surface level larger than 100 A (%f)\n",global_data->arbs_surface_level);
		errout("is strange and not allowed.\n\n",fname);
		pex(-1);
		}

	read_string(h_boundary,sdum);
	if (strcmp(sdum,"[GAUSSIAN HEIGHTS]")!=0) {
		errout("<%s> is not a valid boundary-file.\n",fname);
		errout("<%s>.\n",sdum);
		errout("(Token [GAUSSIAN HEIGHTS] expected)\n\n");
		pex(-1);
		}
	for (i=0; i<global_data->arbs_ncodebook; i++) {
          	global_data->arbs_gauss_height[i] = read_double_skip(h_boundary);
	  	if (global_data->arbs_gauss_height[i] <= 0.0) {
			errout("A non-positive height of a gaussian (%f)\n",global_data->arbs_gauss_height[i]);
			errout("is strange and not allowed.\n\n",fname);
			pex(-1);
			}
		}

	read_string(h_boundary,sdum);
	if (strcmp(sdum,"[GAUSSIAN CENTERS]")!=0) {
		errout("<%s> is not a valid boundary-file.\n",fname);
		errout("<%s>.\n",sdum);
		errout("(Token [GAUSSIAN CENTERS] expected)\n\n");
		pex(-1);
		}
	for (i=0; i<global_data->arbs_ncodebook; i++) {
          	global_data->arbs_gauss_y[0][i] = read_double(h_boundary);
          	global_data->arbs_gauss_y[1][i] = read_double(h_boundary);
          	global_data->arbs_gauss_y[2][i] = read_double_skip(h_boundary);
		}

	read_string(h_boundary,sdum);
	if (strcmp(sdum,"[GAUSSIAN VARIANCES]")!=0) {
		errout("<%s> is not a valid boundary-file.\n",fname);
		errout("<%s>.\n",sdum);
		errout("(Token [GAUSSIAN VARIANCES] expected)\n\n");
		pex(-1);
		}
	for (i=0; i<global_data->arbs_ncodebook; i++) {
          	global_data->arbs_gauss_rho2[0][i] = read_double(h_boundary);
          	global_data->arbs_gauss_rho2[1][i] = read_double(h_boundary);
          	global_data->arbs_gauss_rho2[2][i] = read_double_skip(h_boundary);
	  	if (global_data->arbs_gauss_rho2[0][i] <= 0.0) {
			errout("A non-positive variance of a gaussian (%f)\n",global_data->arbs_gauss_rho2[0][i]);
			errout("is strange and not allowed.\n\n",fname);
			pex(-1);
			}
	  	if (global_data->arbs_gauss_rho2[1][i] <= 0.0) {
			errout("A non-positive variance of a gaussian (%f)\n",global_data->arbs_gauss_rho2[1][i]);
			errout("is strange and not allowed.\n\n",fname);
			pex(-1);
			}
	  	if (global_data->arbs_gauss_rho2[2][i] <= 0.0) {
			errout("A non-positive variance of a gaussian (%f)\n",global_data->arbs_gauss_rho2[2][i]);
			errout("is strange and not allowed.\n\n",fname);
			pex(-1);
			}
		}

	read_string(h_boundary,sdum);
	if (strcmp(sdum,"[GAUSSIAN ORIENTATIONS]")!=0) {
		errout("<%s> is not a valid boundary-file.\n",fname);
		errout("<%s>.\n",sdum);
		errout("(Token [GAUSSIAN ORIENTATIONS] expected)\n\n");
		pex(-1);
		}
	for (i=0; i<global_data->arbs_ncodebook; i++) {
          	global_data->arbs_gauss_w[0][0][i] = read_double(h_boundary);
          	global_data->arbs_gauss_w[0][1][i] = read_double(h_boundary);
          	global_data->arbs_gauss_w[0][2][i] = read_double_skip(h_boundary);
          	global_data->arbs_gauss_w[1][0][i] = read_double(h_boundary);
          	global_data->arbs_gauss_w[1][1][i] = read_double(h_boundary);
          	global_data->arbs_gauss_w[1][2][i] = read_double_skip(h_boundary);
          	global_data->arbs_gauss_w[2][0][i] = read_double(h_boundary);
          	global_data->arbs_gauss_w[2][1][i] = read_double(h_boundary);
          	global_data->arbs_gauss_w[2][2][i] = read_double_skip(h_boundary);
		}
	fclose(h_boundary);

	txtout("Number of ARB.SBOUND-Gaussians: %d\n",global_data->arbs_ncodebook);
	txtout("ARB.SBOUND water shell thickness: %f\n",global_data->arbs_water_shell_thickness);
	txtout("ARB.SBOUND surface level: %f\n",global_data->arbs_surface_level);
	txtout("First ARB.SBOUND height: %f\n",global_data->arbs_gauss_height[0]);
	txtout("First ARB.SBOUND center: %f  %f  %f\n",
	      global_data->arbs_gauss_y[0][0],global_data->arbs_gauss_y[1][0],global_data->arbs_gauss_y[2][0]);
	txtout("First ARB.SBOUND variances: %f  %f  %f\n",
	      global_data->arbs_gauss_rho2[0][0],global_data->arbs_gauss_rho2[1][0],global_data->arbs_gauss_rho2[2][0]);
	txtout("First ARB.SBOUND orientations: %f  %f  %f\n",
	      global_data->arbs_gauss_w[0][0][0],global_data->arbs_gauss_w[0][1][0],global_data->arbs_gauss_w[0][2][0]);
	txtout("                               %f  %f  %f\n",
	      global_data->arbs_gauss_w[1][0][0],global_data->arbs_gauss_w[1][1][0],global_data->arbs_gauss_w[1][2][0]);
	txtout("                               %f  %f  %f\n",
	      global_data->arbs_gauss_w[2][0][0],global_data->arbs_gauss_w[2][1][0],global_data->arbs_gauss_w[2][2][0]);
	return 0;	
}

int read_sbound_file(	char *fname,
			LOAD_MSG_ALL_NODE *global_data)
{
	FILE  *fp_sbound;
        char   sdum[MAX_STRING_SIZE];
	int    ii,jj,kk;
	int    nii,njj,nkk;
	double ddummy;

	global_data->potgrid.stepsize=-1;
	global_data->potgrid.xmin=0;
	global_data->potgrid.xmax=0;
	global_data->potgrid.ymin=0;
	global_data->potgrid.ymax=0;
	global_data->potgrid.zmin=0;
	global_data->potgrid.zmax=0;

        if ((global_data->sbound_droff[R_SPHERE]< -1.0) || (global_data->sbound_droff[R_SPHERE] >= -SMALL)) return 0;
					/* no SBound grid needed */

	if ((fp_sbound=fopen(fname,"r"))==NULL) {
		errout("Fatal error: Can not open SBound-file <%s>.\n",fname);
		errout("Use Solvate 1.2 to create a SBound-file!\n\n");
		pex(-1);
	}	
	txtout("Reading SBound-file        : <%s>.\n",fname);

	read_string(fp_sbound,sdum);
	if (strcmp(sdum,"[POTGRID DIMENSIONS (XMIN,XMAX),(YMIN,YMAX),(ZMIN,ZMAX)]")!=0) {
		errout("<%s> is not a valid SBound-file.\n",fname);
		errout("<%s>.\n",sdum);
		errout("(Token [POTGRID DIMENSIONS (XMIN,XMAX),(YMIN,YMAX),(ZMIN,ZMAX)] expected)\n\n");
		pex(-1);
	}
	global_data->potgrid.xmin=read_double(fp_sbound);
	global_data->potgrid.xmax=read_double_skip(fp_sbound);
	global_data->potgrid.ymin=read_double(fp_sbound);
	global_data->potgrid.ymax=read_double_skip(fp_sbound);
	global_data->potgrid.zmin=read_double(fp_sbound);
	global_data->potgrid.zmax=read_double_skip(fp_sbound);

	read_string(fp_sbound,sdum);
	if (strcmp(sdum,"[STEPSIZE FOR POTGRID]")!=0) {
		errout("<%s> is not a valid SBound-file.\n",fname);
		errout("<%s>.\n",sdum);
		errout("(Token [STEPSIZE FOR POTGRID] expected)\n\n");
		pex(-1);
	}
	global_data->potgrid.stepsize=read_double_skip(fp_sbound);

	nii = ceil ((global_data->potgrid.xmax - global_data->potgrid.xmin + 0.1) 
		        / global_data->potgrid.stepsize);
	njj = ceil ((global_data->potgrid.ymax - global_data->potgrid.ymin + 0.1) 
		        / global_data->potgrid.stepsize);
	nkk = ceil ((global_data->potgrid.zmax - global_data->potgrid.zmin + 0.1) 
		        / global_data->potgrid.stepsize);
	global_data->potgrid.nx = nii;
	global_data->potgrid.ny = njj;
	global_data->potgrid.nz = nkk;

	read_string(fp_sbound,sdum);
	if (strcmp(sdum,"[ERROR AFTER GRADIENT DESCENT OF SBOUND]")!=0) {
		errout("<%s> is not a valid SBound-file.\n",fname);
		errout("<%s>.\n",sdum);
		errout("(Token [ERROR AFTER GRADIENT DESCENT OF SBOUND] expected)\n\n");
		pex(-1);
	}
	ddummy=read_double_skip(fp_sbound);

	read_string(fp_sbound,sdum);
	if (strcmp(sdum,"[POTGRID VALUES (SBOUND,...)]")!=0) {
		errout("<%s> is not a valid SBound-file.\n",fname);
		errout("<%s>.\n",sdum);
		errout("(Token [POTGRID VALUES (SBOUND,...)] expected)\n\n");
		pex(-1);
	}

	/* initialize grid */
	global_data->potgrid.pt= (STRUCT_POTGRID_ELEMENT ****) alloc_ptrs (nii);
	for (ii=0;ii<nii;ii++) {
	  global_data->potgrid.pt[ii]= (STRUCT_POTGRID_ELEMENT ***) alloc_ptrs (njj);
	  for (jj=0;jj<njj;jj++) {
	    global_data->potgrid.pt[ii][jj]= (STRUCT_POTGRID_ELEMENT **) alloc_ptrs (nkk);
	    for (kk=0;kk<nkk;kk++) {
	      ddummy=read_double_skip(fp_sbound);
	      if (fabs(ddummy) >= SMALL) {
	        global_data->potgrid.pt[ii][jj][kk] =
	    	  	(STRUCT_POTGRID_ELEMENT *) alloc_blk(sizeof(STRUCT_POTGRID_ELEMENT));
	        global_data->potgrid.pt[ii][jj][kk]->sbound = ddummy;
	        /**  initialize for later use (has to be corrected) **/
/*	        global_data->potgrid.pt[ii][jj][kk]->polar_fac = 0.1;perhaps later needed for depolarisation */
	      } else {
	    	global_data->potgrid.pt[ii][jj][kk] = NULL;
	      }
	    }
	  }
	}

	return 0;	
}

int read_dbound_file(	char *fname,
			LOAD_MSG_ALL_NODE *global_data)
{
	FILE	*h_boundary;
	int	i,k,j;
	double  ddummy;

	/*** set default values ***/
	global_data->arbs_ncodebook=1;
	global_data->arbs_water_shell_thickness=0.0;
	global_data->arbs_surface_level=0.0;
	for (i=0; i<global_data->arbs_ncodebook; i++) {
		global_data->arbs_gauss_height[i]=0.0;
		}
	for (i=0; i<global_data->arbs_ncodebook; i++) {
		for (j=0; j<3; j++) {
			global_data->arbs_gauss_y[j][i]=0.0;
			global_data->arbs_gauss_rho2[j][i]=0.0;
			}
		}
	for (i=0; i<global_data->arbs_ncodebook; i++) {
		for (j=0; j<3; j++) {
			for (k=0; k<3; k++) {
				global_data->arbs_gauss_w[j][k][i]=0.0;
				}
			}
		}

	/*** we need the cboundary.lis for sbound with arbitrarily shaped boundary ***/
	if (global_data->sbound_droff[R_SPHERE]>= -SMALL) return 0;  /* no arboundary needed */

	if ((h_boundary=fopen(fname,"r"))==NULL) {
		errout("Fatal error: Can not open boundary-file <%s>.\n",fname);
		errout("Use Solvate 1.0 to create a boundary-file!\n\n");
		pex(-1);
	}	
	txtout("Reading boundary-file        : <%s>.\n",fname);

	/***************************************************/
	/***** reading the header **************************/
	/***************************************************/
	read_string(h_boundary,sdum);
	if (strcmp(sdum,"[MINIMUM WATER SHELL THICKNESS]")!=0) {
		errout("<%s> is not a valid boundary-file.\n",fname);
		errout("(Token [MINIMUM WATER SHELL THICKNESS] expected)\n\n");
		pex(-1);
		}
        global_data->arbs_water_shell_thickness = read_double_skip(h_boundary);
	if (global_data->arbs_water_shell_thickness < 1.0) {
		errout("A water shell thickness smaller than 1 A (%f A)\n",global_data->arbs_water_shell_thickness);
		errout("is strange and not allowed.\n\n",fname);
		pex(-1);
		}
	if (global_data->arbs_water_shell_thickness > 1000.0) {
		errout("A water shell thickness larger than 1000 A (%f A)\n",global_data->arbs_water_shell_thickness);
		errout("is strange and not allowed.\n\n",fname);
		pex(-1);
		}
	read_string(h_boundary,sdum);
	if (strcmp(sdum,"[ENCLOSING SPHERE (X,Y,Z,R)]")!=0) {
		errout("<%s> is not a valid boundary-file.\n",fname);
		errout("<%s>.\n",sdum);
		errout("(Token [ENCLOSING SPHERE (X,Y,Z,R)] expected)\n\n");
		pex(-1);
		}
        ddummy = read_double_skip(h_boundary);
        ddummy = read_double_skip(h_boundary);
        ddummy = read_double_skip(h_boundary);
        ddummy = read_double_skip(h_boundary);
	read_string(h_boundary,sdum);
	if (strcmp(sdum,"[NR OF GAUSSIANS]")!=0) {
		errout("<%s> is not a valid boundary-file.\n",fname);
		errout("(Token [NR OF GAUSSIANS] expected)\n\n");
		pex(-1);
		}
        global_data->arbs_ncodebook = read_int_skip(h_boundary);
	if (global_data->arbs_ncodebook < 1) {
		errout("There must be at least one gaussian in the boundary-file.\n");
		errout("Found %d.\n\n",global_data->arbs_ncodebook);
		pex(-1);
		}
	if (global_data->arbs_ncodebook > MXNO_ARBOUND_GAUSSIANS) {
		errout("There must be at most %d gaussians in the boundary-file.\n",MXNO_ARBOUND_GAUSSIANS);
		errout("Found %d.\n\n",global_data->arbs_ncodebook);
		pex(-1);
		}
	read_string(h_boundary,sdum);
	if (strcmp(sdum,"[SURFACE LEVEL]")!=0) {
		errout("<%s> is not a valid boundary-file.\n",fname);
		errout("(Token [SURFACE LEVEL] expected)\n\n");
		pex(-1);
		}
        global_data->arbs_surface_level = read_double_skip(h_boundary);
	if (global_data->arbs_surface_level <= 0.0) {
		errout("A non-positive surface level (%f)\n",global_data->arbs_surface_level);
		errout("is strange and not allowed.\n\n",fname);
		pex(-1);
		}
	if (global_data->arbs_surface_level > 100.0) {
		errout("A surface level larger than 100 A (%f)\n",global_data->arbs_surface_level);
		errout("is strange and not allowed.\n\n",fname);
		pex(-1);
		}
	read_string(h_boundary,sdum);
	if (strcmp(sdum,"[GAUSSIAN HEIGHTS]")!=0) {
		errout("<%s> is not a valid boundary-file.\n",fname);
		errout("(Token [GAUSSIAN HEIGHTS] expected)\n\n");
		pex(-1);
		}
	for (i=0; i<global_data->arbs_ncodebook; i++) {
          	global_data->arbs_gauss_height[i] = read_double_skip(h_boundary);
	  	if (global_data->arbs_gauss_height[i] <= 0.0) {
			errout("A non-positive height of a gaussian (%f)\n",global_data->arbs_gauss_height[i]);
			errout("is strange and not allowed.\n\n",fname);
			pex(-1);
			}
		}
	read_string(h_boundary,sdum);
	if (strcmp(sdum,"[GAUSSIAN CENTERS]")!=0) {
		errout("<%s> is not a valid boundary-file.\n",fname);
		errout("(Token [GAUSSIAN CENTERS] expected)\n\n");
		pex(-1);
		}
	for (i=0; i<global_data->arbs_ncodebook; i++) {
          	global_data->arbs_gauss_y[0][i] = read_double(h_boundary);
          	global_data->arbs_gauss_y[1][i] = read_double(h_boundary);
          	global_data->arbs_gauss_y[2][i] = read_double_skip(h_boundary);
		}
	read_string(h_boundary,sdum);
	if (strcmp(sdum,"[GAUSSIAN VARIANCES]")!=0) {
		errout("<%s> is not a valid boundary-file.\n",fname);
		errout("(Token [GAUSSIAN VARIANCES] expected)\n\n");
		pex(-1);
		}
	for (i=0; i<global_data->arbs_ncodebook; i++) {
          	global_data->arbs_gauss_rho2[0][i] = read_double(h_boundary);
          	global_data->arbs_gauss_rho2[1][i] = read_double(h_boundary);
          	global_data->arbs_gauss_rho2[2][i] = read_double_skip(h_boundary);
	  	if (global_data->arbs_gauss_rho2[0][i] <= 0.0) {
			errout("A non-positive variance of a gaussian (%f)\n",global_data->arbs_gauss_rho2[0][i]);
			errout("is strange and not allowed.\n\n",fname);
			pex(-1);
			}
	  	if (global_data->arbs_gauss_rho2[1][i] <= 0.0) {
			errout("A non-positive variance of a gaussian (%f)\n",global_data->arbs_gauss_rho2[1][i]);
			errout("is strange and not allowed.\n\n",fname);
			pex(-1);
			}
	  	if (global_data->arbs_gauss_rho2[2][i] <= 0.0) {
			errout("A non-positive variance of a gaussian (%f)\n",global_data->arbs_gauss_rho2[2][i]);
			errout("is strange and not allowed.\n\n",fname);
			pex(-1);
			}
		}
	read_string(h_boundary,sdum);
	if (strcmp(sdum,"[GAUSSIAN ORIENTATIONS]")!=0) {
		errout("<%s> is not a valid boundary-file.\n",fname);
		errout("(Token [GAUSSIAN ORIENTATIONS] expected)\n\n");
		pex(-1);
		}
	for (i=0; i<global_data->arbs_ncodebook; i++) {
          	global_data->arbs_gauss_w[0][0][i] = read_double(h_boundary);
          	global_data->arbs_gauss_w[0][1][i] = read_double(h_boundary);
          	global_data->arbs_gauss_w[0][2][i] = read_double_skip(h_boundary);
          	global_data->arbs_gauss_w[1][0][i] = read_double(h_boundary);
          	global_data->arbs_gauss_w[1][1][i] = read_double(h_boundary);
          	global_data->arbs_gauss_w[1][2][i] = read_double_skip(h_boundary);
          	global_data->arbs_gauss_w[2][0][i] = read_double(h_boundary);
          	global_data->arbs_gauss_w[2][1][i] = read_double(h_boundary);
          	global_data->arbs_gauss_w[2][2][i] = read_double_skip(h_boundary);
		}
	fclose(h_boundary);

	txtout("Number of ARB.SBOUND-Gaussians: %d\n",global_data->arbs_ncodebook);
	txtout("ARB.SBOUND water shell thickness: %f\n",global_data->arbs_water_shell_thickness);
	txtout("ARB.SBOUND surface level: %f\n",global_data->arbs_surface_level);
	txtout("First ARB.SBOUND height: %f\n",global_data->arbs_gauss_height[0]);
	txtout("First ARB.SBOUND center: %f  %f  %f\n",
	      global_data->arbs_gauss_y[0][0],global_data->arbs_gauss_y[1][0],global_data->arbs_gauss_y[2][0]);
	txtout("First ARB.SBOUND variances: %f  %f  %f\n",
	      global_data->arbs_gauss_rho2[0][0],global_data->arbs_gauss_rho2[1][0],global_data->arbs_gauss_rho2[2][0]);
	txtout("First ARB.SBOUND orientations: %f  %f  %f\n",
	      global_data->arbs_gauss_w[0][0][0],global_data->arbs_gauss_w[0][1][0],global_data->arbs_gauss_w[0][2][0]);
	txtout("                               %f  %f  %f\n",
	      global_data->arbs_gauss_w[1][0][0],global_data->arbs_gauss_w[1][1][0],global_data->arbs_gauss_w[1][2][0]);
	txtout("                               %f  %f  %f\n",
	      global_data->arbs_gauss_w[2][0][0],global_data->arbs_gauss_w[2][1][0],global_data->arbs_gauss_w[2][2][0]);
	return 0;	
}


/****************************************************************************/
int read_restart_file(		char *filename,
				int nr_of_atoms,
				int *file_count,
				STRUCT_ATOMDATA *atom_data,
				LOAD_MSG_ALL_NODE *global_data)
{	
	FILE	*h_restfile;
	int	i=strlen(filename);
	double	old_time_step,time_factor=1.0;
#if defined(SR8000)
        /* for HITACHI use a BIG buffer */
#define BUFSIZE 4194304
        char buffer[BUFSIZE];
#endif
	
	*file_count    = 0;
	/*** open the restart-file ***/
	if (	(h_restfile=fopen(filename,"r"))==NULL) {
		txtout("Found NO restart-file     : <%s>\n",filename);
		return 0;
	}
	
#if defined(SR8000)
        /* for HITACHI use a BIG buffer */
        setvbuf(h_restfile, buffer, _IOFBF, (size_t) BUFSIZE);
#endif

	/*** do not use restart-file if last character is '@' ***/
	if (filename[i-1]=='@') {
		txtout("Found restart-file, but it will not be used because of '@'!\n");
		fclose(h_restfile);
		return 0;
	}
	
	txtout("Using restart file        : <%s>\n",filename);

	/**********************************/
	/*** read the header 		***/
	/**********************************/

	if (fscanf(h_restfile,"%s%s%s%s",sdum,sdum,sdum,sdum)!=4) {
		goto error_restart;
	}	

	/*** if we are restarting we have to take care of different ***/
	/*** time steps                                             ***/
	old_time_step = read_double_skip(h_restfile);	/* the OLD time_step */
	time_factor = (global_data->time_step*TIMEDIV)/old_time_step;

	if (fscanf(h_restfile,"%s%s%s%s",sdum,sdum,sdum,sdum)!=4) {
		goto error_restart;
	}

	global_data->cycles_passed = read_int_skip(h_restfile)-1;
	if (global_data->cycles_passed<0) global_data->cycles_passed=0;
	if (fscanf(h_restfile,"%s%s%s",sdum,sdum,sdum)!=3) {
		goto error_restart;	
	}
	*file_count = read_int_skip(h_restfile);
	skip_line(h_restfile,3); /* skip the 3 format description lines */

	/**************************************************/
	/*** no we readin the old coordinates		***/
	/**************************************************/
	for (i=0;i<nr_of_atoms;i++) {
		if (read_string(h_restfile,sdum)==-1) {
			errout("Fatal error: Not enough lines in restart-file (old-ko).\n");
			fclose(h_restfile);
			pex(-1);
		}
#if PRECISION == DOUBLE
		if (sscanf(sdum,"%lf %lf %lf",
#else
		if (sscanf(sdum,"%f %f %f",
#endif
			atom_data[i].vel+0,
			atom_data[i].vel+1,
			atom_data[i].vel+2)!=3) {
			errout("Fatal error: Illegal line in restart-file (old-ko).\n");
			errout(sdum);
			fclose(h_restfile);
			pex(-1);
		}			
	}

	/**************************************************/
	/*** no we read in the new coordinates		***/
	/**************************************************/
	for (i=0;i<nr_of_atoms;i++) {	
		if (read_string(h_restfile,sdum)==-1) {
			errout("Fatal error: Not enough lines in restart-file (new-ko).\n");
			fclose(h_restfile);
			pex(-1);
		}
#if PRECISION == DOUBLE
		if (sscanf(sdum,"%lf %lf %lf",
#else
		if (sscanf(sdum,"%f %f %f",
#endif
			atom_data[i].ko+0,
			atom_data[i].ko+1,
			atom_data[i].ko+2)!=3) {
			errout("Fatal error: Illegal line in restart-file (new-ko).\n");
			errout(sdum);
			fclose(h_restfile);
			pex(-1);
		}

	    	/* mention any difference in timestep  between restart  */
		/* file and control-file. Change the OLD positions,     */
	     	/* because the force is always calculated for the new   */
	      	/* positions  	                                        */		
		atom_data[i].vel[0]=atom_data[i].ko[0]+
				time_factor*(atom_data[i].vel[0]-atom_data[i].ko[0]);
		atom_data[i].vel[1]=atom_data[i].ko[1]+
				time_factor*(atom_data[i].vel[1]-atom_data[i].ko[1]);
		atom_data[i].vel[2]=atom_data[i].ko[2]+
				time_factor*(atom_data[i].vel[2]-atom_data[i].ko[2]);					
	}
	
	/**************************************************/
	/*** any immobilisation or pulling data here	***/
	/**************************************************/
	if (read_string(h_restfile,sdum)!=-1) {

#if PRECISION == DOUBLE
		if (sscanf(sdum,"%d %lf %lf %lf",
#else
		if (sscanf(sdum,"%d %f %f %f",
#endif
				&global_data->start_refC,
				&global_data->refC[0],
				&global_data->refC[1],
				&global_data->refC[2])!=4) {
			goto error_pulling;
		}
	     	if ((global_data->flag_selimob>=2) && (global_data->selimob_start<=global_data->start_refC)) {
		  if (read_string(h_restfile,sdum)!=-1) {

#if PRECISION == DOUBLE
		    if (sscanf(sdum,"%lf %lf %lf",
#else
	 	    if (sscanf(sdum,"%f %f %f",
#endif
			     &global_data->refC[3],
			     &global_data->refC[4],
			     &global_data->refC[5])!=3) {
		      goto error_pulling;
		    }
		  }
		if (global_data->flag_selimob==3) {
	          if (read_string(h_restfile,sdum)!=-1) {

#if PRECISION == DOUBLE
		    if (sscanf(sdum,"%lf %lf %lf",
#else
	 	    if (sscanf(sdum,"%f %f %f",
#endif
			     &global_data->refC[6],
			     &global_data->refC[7],
			     &global_data->refC[8])!=3) {
		      goto error_pulling;
		    }
		  }
		 } /** if ==3**/
		} /** if >=2**/
		
		if (read_string(h_restfile,sdum)==-1) {
			goto error_pulling;
		}
                if (global_data->flag_pull!=3) {
#if PRECISION == DOUBLE
		if (sscanf(sdum,"%lf %lf %lf",
#else
		if (sscanf(sdum,"%f %f %f",
#endif
		
		&global_data->pullspringpos[0],
				&global_data->pullspringpos[1],
				&global_data->pullspringpos[2])!=3) {
			goto error_pulling;
	  	}
             }
		if (global_data->flag_pull==3) {
#if PRECISION == DOUBLE
			if (sscanf(sdum,"%lf %lf %lf %lf %lf %lf %lf",
#else
			if (sscanf(sdum,"%f %f %f %f %f %f %f",
#endif
					&global_data->pullspringpos[0],
					&global_data->pullspringpos[1],
					&global_data->pullspringpos[2],
					&global_data->pullspringpos[3],
					&global_data->pullspringpos[4],
					&global_data->pullspringpos[5],
					&global_data->pullspringpos[6])!=7) {
				goto error_pulling;
			}
		}
	} 
	
	read_sbound_droff(h_restfile,global_data->sbound_droff,1);
	
	/**************************************************/
	/*** any QM-Boundbox data here? 		***/
	/**************************************************/
	if (read_string(h_restfile,sdum)!=-1) {
		char tstr[100];
		
#if PRECISION == DOUBLE
		if (sscanf(sdum,"%s %lf %lf",
#else
		if (sscanf(sdum,"%s %f %f",
#endif
				tstr,
				&global_data->qmboxext[0],
				&global_data->qmboxext[1])!=3) {
			goto error_qmboundbox;
		}
		if (read_string(h_restfile,sdum)==-1) {
			goto error_qmboundbox;
		}
#if PRECISION == DOUBLE
		if (sscanf(sdum,"%s %lf %lf",
#else
		if (sscanf(sdum,"%s %f %f",
#endif
				tstr,
				&global_data->qmboxext[2],
				&global_data->qmboxext[3])!=3) {
			goto error_qmboundbox;
		}
		if (read_string(h_restfile,sdum)==-1) {
			goto error_qmboundbox;
		}
#if PRECISION == DOUBLE
		if (sscanf(sdum,"%s %lf %lf",
#else
		if (sscanf(sdum,"%s %f %f",
#endif
				tstr,
				&global_data->qmboxext[4],
				&global_data->qmboxext[5])!=3) {
			goto error_qmboundbox;
		}
	} 	
	
	/**************************************************/
	/*** any FLOODING_ENERGY_AVERAGE data here? 	***/
	/**************************************************/
	if (read_string(h_restfile,sdum)!=-1) {
		char tstr[100];
#if PRECISION == DOUBLE
		if (sscanf(sdum,"%s %lf",tstr, &global_data->flood_energy_average)!=2) 
#else
		if (sscanf(sdum,"%s %f",tstr, &global_data->flood_energy_average)!=2) 
#endif
		{
			goto error_flooding_energy;
		}

		if (read_string(h_restfile,sdum)==-1) goto error_flooding_energy;
#if PRECISION == DOUBLE
		if (sscanf(sdum,"%s %lf",tstr, &global_data->flood_energy_strength)!=2) 
#else
		if (sscanf(sdum,"%s %f",tstr, &global_data->flood_energy_strength)!=2) 
#endif
		{
			goto error_flooding_energy;
		}
		if (read_string(h_restfile,sdum)==-1) {
			global_data->flood_distance=0.0;
		} else
#if PRECISION == DOUBLE
		if (sscanf(sdum,"%s %lf",tstr, &global_data->flood_distance)!=2) 
#else
		if (sscanf(sdum,"%s %f",tstr, &global_data->flood_distance)!=2) 
#endif
		{
			goto error_flooding_energy;
		}
	} 	
	if (global_data->rotation_mode) {
	  if (read_string(h_restfile,sdum)!=-1) {
	    char tstr[100];/*HH this was double, i changed it to char. not sure if this is right */
	    if (sscanf(sdum,"%s %d",
		       tstr,&global_data->nr_of_rot_atoms)!=2) {
	      goto error_rotation_1;
	    }
	  
	    /*** allocation of rot_atoms_x,y,z here necessary ***/
	    if (!(global_data->nr_of_rot_atoms == 0)) {
	      global_data->rot_x = alloc_doubles(global_data->nr_of_rot_atoms);
	      global_data->rot_y = alloc_doubles(global_data->nr_of_rot_atoms);
	      global_data->rot_z = alloc_doubles(global_data->nr_of_rot_atoms);
	      global_data->rot1_x = alloc_doubles(global_data->nr_of_rot_atoms);
	      global_data->rot1_y = alloc_doubles(global_data->nr_of_rot_atoms);
	      global_data->rot1_z = alloc_doubles(global_data->nr_of_rot_atoms);
	      global_data->rot2_x = alloc_doubles(global_data->nr_of_rot_atoms);
	      global_data->rot2_y = alloc_doubles(global_data->nr_of_rot_atoms);
	      global_data->rot2_z = alloc_doubles(global_data->nr_of_rot_atoms);
	      global_data->rotation_torque  = alloc_doubles(global_data->nr_of_rot_atoms);
	      global_data->rot_t1_x = alloc_doubles(global_data->nr_of_rot_atoms);
	      global_data->rot_t1_y = alloc_doubles(global_data->nr_of_rot_atoms);
	      global_data->rot_t1_z = alloc_doubles(global_data->nr_of_rot_atoms);
	      global_data->rot_t2_x = alloc_doubles(global_data->nr_of_rot_atoms);
	      global_data->rot_t2_y = alloc_doubles(global_data->nr_of_rot_atoms);
	      global_data->rot_t2_z = alloc_doubles(global_data->nr_of_rot_atoms);
	      
	      global_data->rotation_t_torque  = alloc_doubles(global_data->nr_of_rot_atoms);

	      global_data->rotation_map    = alloc_ints(global_data->nr_of_rot_atoms);
	    }

	    for (i=0;i<global_data->nr_of_rot_atoms;i++) {	
	      if (read_string(h_restfile,sdum)!=-1) {

#if PRECISION == DOUBLE
	      if (sscanf(sdum,"%d %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf",
#else
	      if (sscanf(sdum,"%d %f %f %f %f %f %f %f %f %f %f",
#endif
			 &global_data->rotation_map[i],
			 &global_data->rot_x[i],
			 &global_data->rot_y[i],
			 &global_data->rot_z[i],
			 &global_data->rot1_x[i],
			 &global_data->rot1_y[i],
			 &global_data->rot1_z[i],
			 &global_data->rot2_x[i],
			 &global_data->rot2_y[i],
			 &global_data->rot2_z[i],
			 &global_data->rotation_torque[i])!=11) {
		goto error_rotation_2;
	      }
	      }
	    }
	      for (i=0;i<global_data->nr_of_rot_atoms;i++) {
		global_data->rot_t1_x[i]      = 0.0;
		global_data->rot_t1_y[i]      = 0.0;
		global_data->rot_t1_z[i]      = 0.0;
		global_data->rot_t2_x[i]      = 0.0;
		global_data->rot_t2_y[i]      = 0.0;
		global_data->rot_t2_z[i]      = 0.0;
		global_data->rotation_t_torque[i] = 0.0;
	      }
	      if (read_string(h_restfile,sdum)!=-1) {
#if PRECISION == DOUBLE
		if (sscanf(sdum,"%lf",
#else
		if (sscanf(sdum,"%f",
#endif
			   &global_data->rotation_energy)!=1) {
		  goto error_rotation_2;
	        }
	      }
	    }
	}



	fclose(h_restfile);
	return 1;

error_restart:
	errout("Fatal error: Restart file is currupted!\n");
	fclose(h_restfile);
	pex(-1);
	
error_pulling:
	errout("Fatal error: Last line in restart file contains illegal\n");
	errout("pulling/immobilisation information!\n");
	errout("<%s>\n",sdum);
	fclose(h_restfile);
	pex(-1);

error_rotation_1:
	errout("Fatal error in Restart File concerning Rotation - 1\n");
	errout("there HAS to be a 0 at the end of the restart-file\n");
	errout("at rotation start or the number of rotation atoms, respectively\n");
	errout("<%s>\n",sdum);
	fclose(h_restfile);
	pex(-1);

error_rotation_2:
	errout("Fatal error in Restart File concerning Rotation - 2\n");
	errout("Could not read in coordinates of rotation atoms or the rotation energy\n");
	errout("<%s>\n",sdum);
	fclose(h_restfile);
	pex(-1);


error_qmboundbox:
        errout("Fatal error: Wrong format for QM-Boundbox data in restart-file!\n");
        errout("<%s>\n",sdum);
        fclose(h_restfile);
        pex(-1);

error_flooding_energy:
	errout("Fatal error: Wrong format for flooding energy in restart-file!\n");
        errout("<%s>\n",sdum);
	fclose(h_restfile);
	pex(-1);

return -1;
}

/************************************************************************/
static void read_sbound_droff(FILE *cfd, double *sbound_droff,int call_from_restart)
{
	int i;
	char sdum[MAX_STRING_SIZE];
	double a,b,c,d;
	
	/* We call this also from read_restart_file(), but in old restart files these
	 * data are not yet present, so we have to check if we are dealing with
	 * sufficiently up to date restart files. if not, we use the data
	 * from the control file and do NOT overwrite them. */
	
	if (read_string(cfd,sdum)!=-1) 
	{
#if PRECISION == DOUBLE
	  if (sscanf(sdum,"%lf %lf",&a,&b)==2) 
#else
	  if (sscanf(sdum,"%f %f",&a,&b)==2) 
#endif
	  {	int changed=0;
	
	    	if (!call_from_restart) 
	    		for (i=0;i<MXNO_SBOUND_PARAM;i++) sbound_droff[i] = -99999.0;
	    	
		    /*** reading planes position ***/
		if (a!=sbound_droff[0])	changed=1;
		if (b!=sbound_droff[1])	changed=1;
	    	sbound_droff[0] = a;
	    	sbound_droff[1] = b;
	    
	    	a = read_double(cfd);
	    	b = read_double_skip(cfd);
		if (a!=sbound_droff[2])	changed=1;
		if (b!=sbound_droff[3])	changed=1;	    	
	    	sbound_droff[2] = a; 
	    	sbound_droff[3] = b; 
	    	
	    	a = read_double(cfd);
	    	b = read_double_skip(cfd);
		if (a!=sbound_droff[4])	changed=1;
		if (b!=sbound_droff[5])	changed=1;    	
	    	sbound_droff[4] = a;
	    	sbound_droff[5] = b;
	    	
	    	/*** reading data of sphere (x,y,z,R) ***/
	    	a = read_double(cfd);
	    	b = read_double(cfd); 
	    	c = read_double_skip(cfd);
	    	d = read_double_skip(cfd);
	    	if (a!=sbound_droff[6])	changed=1;
		if (b!=sbound_droff[7])	changed=1; 
		if (c!=sbound_droff[8])	changed=1;
		if (d!=sbound_droff[9])	changed=1;   

	    	sbound_droff[6] = a;
	    	sbound_droff[7] = b; 
	    	sbound_droff[8] = c; 
	    	sbound_droff[9] = d;
	    	
	    	/*** reading data of cylinder (x,y,R) parallel to z-axis ***/
	    	a = read_double(cfd);
	    	b = read_double(cfd); 
	    	c = read_double_skip(cfd);
	    	if (a!=sbound_droff[11]) changed=1;
		if (b!=sbound_droff[12]) changed=1; 
		if (c!=sbound_droff[13]) changed=1;
	    	sbound_droff[11] = a;
	    	sbound_droff[12] = b;
	    	sbound_droff[13] = c;
    	
	    	/*** curvature of edges ***/	
	    	a = read_double_skip(cfd);
	    	if (a!=sbound_droff[10]) changed=1;	    	
	    	sbound_droff[10] = a;
	    	
	    	if ((call_from_restart)&&(changed)) {
	   		txtout("***************************************************\n");
	   		txtout("***************************************************\n");
	   		txtout("Info: Taking SBOUND parameters from restart file!!!\n");
	   		txtout("      If you want to change these parameters you\n");
	   		txtout("      must change them in this restart file\n");
	   		txtout("***************************************************\n");	    	
	   		txtout("***************************************************\n");
	   		SLEEP((unsigned int) 7);
	    	}
	    
	  }
	}	
	return;
}

static void read_sbound_droff_new (FILE *cfd, double *sbound_droff,int call_from_restart)
{
	int i,ch;
	char schalter;
	double wert[14];	
	int changed=0;
        for (i=0;i<14;i++) wert[i]=0.000000;
	schalter = ((char) fgetc(cfd));
        while (((ch = fgetc (cfd)) != EOF) && (ch != ((int)'\n'))) {}	
	if ((schalter != 'K') && (schalter != 'S') && (schalter != 'N') && (schalter != 'A') && (schalter != 'C'))
	{
	  fclose(cfd);
	  errout("You forgot the SBOUND-Switch !!!\n");
	  pex(-1);
	}
	if (schalter == 'K')            /* Cubic Sbound -- Read Planes */
	{
	  wert[0] = read_double(cfd);
	  wert[1] = read_double_skip(cfd);

          wert[2] = read_double(cfd);
          wert[3] = read_double_skip(cfd);

          wert[4] = read_double(cfd);
          wert[5] = read_double_skip(cfd);
	}
	    	/*** reading spring coordinates for arbitrary, numrical or spheric and radius for spheric SBound (x,y,z,R) ***/
	if ((schalter == 'S') || (schalter == 'N') || (schalter == 'A'))
	{
	  wert[6] = read_double(cfd);
	  wert[7] = read_double(cfd);
          wert[8] = read_double_skip(cfd);
	  if (schalter == 'S') wert[9] = read_double_skip(cfd); /* Radius for spheric SBound */
	}
	    	/*** reading data of cylinder (x,y,R) parallel to z-axis ***/
	if (schalter == 'C')
	{
          wert[11] = read_double(cfd);
          wert[12] = read_double(cfd);
          wert[13] = read_double_skip(cfd);
	}
	    	/*** curvature of edges ***/	
	wert[10] = read_double_skip(cfd);
	if (!call_from_restart) 
	  for (i=0;i<MXNO_SBOUND_PARAM;i++) sbound_droff[i] = -99999.0;
	
	if (schalter == 'A') wert[9] = -1.500000;
	if (schalter == 'N') wert[9] = -1.000000;
	      
/* Vergleiche neue Werte mit alten und kopiere*/
	  for (i=0;i<14;i++)
	  {
		if (wert[i]!=sbound_droff[i])	changed=1;
	    	sbound_droff[i] = wert[i];
	  }
	    	
	    	if ((call_from_restart)&&(changed)) {
	   		txtout("***************************************************\n");
	   		txtout("***************************************************\n");
	   		txtout("Info: Taking SBOUND parameters from restart file!!!\n");
	   		txtout("      If you want to change these parameters you\n");
	   		txtout("      must change them in this restart file\n");
	   		txtout("***************************************************\n");	    	
	   		txtout("***************************************************\n");
	   		SLEEP((unsigned int) 7);
	    	}
	    
	  
		
	return;
}

/*****************************************************************************/
int readin_bondfile(	int nr_of_atoms,
			char *bond_filen,
			int *bond_data,
			int *old2new_map)
{	
	FILE	*h_bondfile;
	int	i,number,aktuell,newakt;
	int	offset=0,c;
						/* open my coord-file */
	if ((h_bondfile=fopen(bond_filen,"r"))==NULL) {
		errout("Fatal error: Can not open bond-file <%s>\n",bond_filen);
		pex(-1);
	}	

	for (i=0;i<nr_of_atoms;i++) {
		number=0;
		fscanf(h_bondfile,"%s %d %s %s",sdum,&number,sdum,sdum);
		if (number != i) {
			errout("Fatal error: Found wrong atom number in BOND file ! %d %d\n",number,i);
			fclose(h_bondfile);
			pex(-1);
		}
		offset=old2new_map[i]*MXNO_FIELDS_BOND;

		c=0;
		do {
			fscanf(h_bondfile,"%d",&aktuell);

			if (aktuell==-1) {	/* do not map end_of_list */
				newakt=-1;
				bond_data[offset+c]=newakt;
			} else {		/* map and sort atom-numbers */
				int k;
				newakt=old2new_map[aktuell];
				bond_data[offset+c]=newakt;
				k=c;
				while ((k>0)&&(bond_data[offset+k-1]>newakt)) {
					bond_data[offset+k  ]=bond_data[offset+k-1];
					bond_data[offset+k-1]=newakt;
					k--;
				}
			}

			if (++c>=MXNO_FIELDS_BOND) {
				errout("Fatal error: Too many bonds for atom %d (EGO), aktuell %d.\n",
					i,aktuell);
				errout("Increase MXNO_FIELDS_BOND (%d) in ego.h !\n",
					MXNO_FIELDS_BOND);
				fclose(h_bondfile);
				pex(-1);
			}
		} while (aktuell!=-1);
	}

	fclose(h_bondfile);
	return 0;
}

/*****************************************************************************/

int readin_exclfile(	int nr_of_atoms,
			int flag,
			char *excl_filen,
			int **excl_data,
			int *old2new_map,
			int cur_maxc)
{	
	FILE	*h_exclfile;
	int	i,number,aktuell,newakt;
	int	offset=0,c,maxc=0;
						/* open my excl-file */
	if ((h_exclfile=fopen(excl_filen,"r"))==NULL) {
		errout("Fatal error: Can not open excl-file <%s>\n",excl_filen);
		pex(-1);
	}

	/*****************************************************************/
	/**** first we determine the max number of exclusions per atom  **/
	/**** and alloc the needed space 				**/		
	/*****************************************************************/
	if (*excl_data==NULL){	
		for (i=0;i<nr_of_atoms;i++) {
			if (flag==0) fscanf(h_exclfile,"%s %d %s",sdum,&number, sdum);
			else	     fscanf(h_exclfile,"%s %d %s %s",sdum,&number, sdum, sdum);

			c=0;
			do {
				fscanf(h_exclfile,"%d",&aktuell);
				c++;
			} while (aktuell!=-1);
			if (c>maxc) maxc=c;
		}
		fclose(h_exclfile);
		maxc++;
		if ((h_exclfile=fopen(excl_filen,"r"))==NULL) {
			errout("Fatal error: Can not open excl-file <%s>\n",excl_filen);
			pex(-1);
		}
		*excl_data = alloc_ints((nr_of_atoms+5)*maxc);
#if DEBUG
	txtout("found maxeclusion %d in alloc\n",maxc);
#endif
	} else {
		maxc = cur_maxc;
#if DEBUG
	txtout("found maxeclusion %d from last\n",maxc);
#endif
	}	

	/*********************************************************/
	/**** now we readin the data really **********************/
	/*********************************************************/
	for (i=0;i<nr_of_atoms;i++) {
		number=0;
		if (flag==0) {
			fscanf(h_exclfile,"%s %d %s",sdum,&number,sdum);
		} else {
			fscanf(h_exclfile,"%s %d %s %s",sdum,&number,sdum,sdum);
		}
		if (number != i) {
			errout("Fatal error: Found wrong atom number in exclusion file (file %d != count %d\n",
				number,i);
			fclose(h_exclfile);
			pex(-1);
		}
#if DEBUG
/*
txtout("read exclusion for atom %d\n",i);
*/
#endif
		offset=old2new_map[i]*maxc;
		c=0;
		do {
			fscanf(h_exclfile,"%d",&aktuell);

			if (aktuell==-1) {	/* do not map end_of_list */
				newakt=-1;
				(*excl_data)[offset+c]=newakt;
			} else {		/* map and sort atom-numbers */
				int k;
				newakt=old2new_map[aktuell];
				(*excl_data)[offset+c]=newakt;
				k=c;
				while ((k>0)&&((*excl_data)[offset+k-1]>newakt)) {
					(*excl_data)[offset+k]  =(*excl_data)[offset+k-1];
					(*excl_data)[offset+k-1]=newakt;
					k--;
				}
			}
	
			if (++c>=maxc) {
				errout("Fatal error: Too many exclusions for atom %d (EGO), aktuell %d, max %d\n",
					i,aktuell,maxc);
				fclose(h_exclfile);
				pex(-1);
			}
		} while (aktuell!=-1);
	}

	fclose(h_exclfile);
	return maxc;
}

/****************************************************************************/

int readin_anglefile(	int *size,
			char *angle_filen,
			int **angle_data,
			int *old2new_map)
{	
	FILE	*h_anglefile;
	int	number,c=0;
	int 	max_angle_list=0;
	int	cur_angle_list=0;
					/* open my angle-file */
	if ((h_anglefile=fopen(angle_filen,"r"))==NULL) {
		errout("Can not open angle-file <%s>\n",angle_filen);
		pex(-1);
	}

	max_angle_list = read_int_skip(h_anglefile);
	if (*angle_data==NULL) 	
		*angle_data    = alloc_ints((max_angle_list+5)*4);
#if DEBUG
	txtout("found angles %d\n",max_angle_list);
#endif
	skip_line(h_anglefile,2);

	while (!feof(h_anglefile)) {
		number=0;
		fscanf(h_anglefile,"%d %d %d %d %d",
			&number,
			*angle_data+c,
			*angle_data+c+1,
			*angle_data+c+2,
			*angle_data+c+3);

		if (feof(h_anglefile)) break;

		(*angle_data)[c  ] = old2new_map[(*angle_data)[c  ]];
		(*angle_data)[c+1] = old2new_map[(*angle_data)[c+1]];
		(*angle_data)[c+2] = old2new_map[(*angle_data)[c+2]];

		if (number != cur_angle_list) {
			errout("Fatal error: Found wrong atom number in angle file! (file %d != count %d )\n",
				number,cur_angle_list);
			fclose(h_anglefile);
			pex(-1);
		}

		c+=4;
		cur_angle_list++;
		if (cur_angle_list>max_angle_list) {
			errout("Fatal error: Inconsistent data in angle-file (%d). Too many angles !",
				number);
			errout("Maximum number is %d, current is %d\n",max_angle_list,cur_angle_list);
			fclose(h_anglefile);
			pex(-1);	
		}
	}
	*size=max_angle_list;
	fclose(h_anglefile);
	return cur_angle_list;
}

/**************************************************************************/
/*** also used for reading of impropers *******/

int readin_dihefile(	int *size,
			char *dihe_filen,
			int **dihe_data,
			int *old2new_map)
{	
	FILE	*h_dihefile;
	int	number,c=0;
	int 	max_dihe_list=0;
	int	cur_dihe_list=0;
					/* open my dihe-file */
	if ((h_dihefile=fopen(dihe_filen,"r"))==NULL) {
		errout("Fatal error: Can not open dihedral-file <%s>\n",dihe_filen);
		pex(-1);
	}

	max_dihe_list = read_int_skip(h_dihefile);
	if (*dihe_data==NULL)
		*dihe_data    = alloc_ints((max_dihe_list+5)*5);
#if DEBUG
	errout("found dihdrals %d\n",max_dihe_list);
#endif
	skip_line (h_dihefile, 2);

	while (!feof(h_dihefile)) {
		number=0;
		fscanf(h_dihefile, "%d %d %d %d %d %d",
			&number,
			*dihe_data+c,
			*dihe_data+c+1,
			*dihe_data+c+2,
			*dihe_data+c+3,
			*dihe_data+c+4);

		if (feof(h_dihefile)) break;

		(*dihe_data)[c  ] = old2new_map[(*dihe_data)[c  ]];
		(*dihe_data)[c+1] = old2new_map[(*dihe_data)[c+1]];
		(*dihe_data)[c+2] = old2new_map[(*dihe_data)[c+2]];
		(*dihe_data)[c+3] = old2new_map[(*dihe_data)[c+3]];

		if (number != cur_dihe_list) {
			errout("Fatal error: Found wrong atom number in dihedral file (file %d != count %d)!\n",
				number,cur_dihe_list);
			fclose(h_dihefile);
			pex(-1);
		}

		c+=5;
		cur_dihe_list++;
		if (cur_dihe_list>max_dihe_list) {
			errout("Fatal error: Inconsitent data in dihedral-file. Too many dihedrals !\n");
			errout("Maximum number is %d, current is %d\n",max_dihe_list,cur_dihe_list);
			fclose(h_dihefile);
			pex(-1);	
		}
	}
	*size=max_dihe_list;
	fclose(h_dihefile);
	return cur_dihe_list;
}
	
static int
parse_sboundmode(	char *fname,
			char *line,
			LOAD_MSG_ALL_NODE *global_data)
{
	int	i,j,nos,idx;

	if (sscanf(line,"%d",&global_data->sbound_range)!=1) {
		errout("Fatal error: No SBOUND MODE parameter specified in <%s>.\n",
			fname);
		pex(-1);	
	}

	if (global_data->sbound_range == 3) {
	  /* searching for residue selection strings */
	  i=0;
	  nos=0;
	  while (line[i]!='\0') {
	    if (line[i]=='R') {
	      i++;
	      j=i;
	      while (isalnum(line[j]) || (line[j]=='*')) j++;
	      if (j==i) {
		errout("Fatal error: No residue found after annotation\n");
		errout("             I need a residue name after R (without space)\n");
		pex(-1);
	      }
	      if (j > i+LEN_RES_TYPE) {
		errout("Fatal error: No residue found after annotation\n");
		errout("             I need a residue name after R (without space)\n");
		pex(-1);
	      }
	      nos++;
	      i=j;
	    } 
	    else i++;
	  }
          myfree(global_data->sbound_res_string);
          global_data->no_sbound_res_string = nos;
          global_data->sbound_res_string = (char (*) [LEN_RES_TYPE+1]) alloc_chars((nos+1)*(LEN_RES_TYPE+1));
	  i=0;
	  idx=0;
	  while (line[i]!='\0') {
	    if (line[i]=='R') {
	      i++;
	      j=0;
	      while (isalnum(line[i]) || (line[i]=='*')) {
	        (global_data->sbound_res_string[idx])[j]=line[i];
		i++;
		j++;
	      }
	      (global_data->sbound_res_string[idx])[j] = '\0';
	      idx++;
	    } 
	    else i++;
	  }
	  (global_data->sbound_res_string[idx])[0] = '\0';
	}

	return 0;

}

/**************************************************/
/*** fuer schwingungsanalyse			***/
/**************************************************/
static int
parse_schwing(char *keystr,char *selstr,
		STRUCT_ATOMDATA *atom_data,
		LOAD_MSG_ALL_NODE *global_data)
{
	int	i=atom_data[0].atomnr;
	int	len=strlen(selstr),res,mode=0;
	double	offs=0.0;
	
	/*** truncate trailing white spaces ***/
	for (i=len-1;i>=0;i--) {
		if ((selstr[i]==' ')||(selstr[i]=='\t')) {
			selstr[i]=0;
		} else break;
	}

	if (	(strcmp(keystr,"SCHWING")==0) ||
		(strcmp(keystr,"HESS")==0)  ) {
#if PRECISION == DOUBLE
		res = sscanf(selstr,"%d %le",
#else
		res = sscanf(selstr,"%d %e",
#endif
				&mode,
				&offs);
			
		if ((mode<-2)||(mode>2)) {
			errout("Fatal error: Illegal mode for HESS\n");
			pex(-1);	
		}	
		global_data->schwing_mode = mode;
		global_data->schwing_offs = offs;	
		txtout("Normal mode analysis mode %d with offset %f\n",
			global_data->schwing_mode,
			global_data->schwing_offs);
		
	}
	
	return 0;
}

/*** Used for PICK_BOND, PICK_ELEC and PICK_VDW	***/	
static int
parse_pickbond(	char *selstr,
		STRUCT_ATOMDATA *atom_data,
		int *atom_sele_flag1,
		LOAD_MSG_ALL_NODE *global_data,
		STRUCT_PICK_BOND *pinfo,
		int mask)
{
	int	atomid1=0,atomid2=0,i=atom_data[0].atomnr;

	/**************************************************/
	/***  Check/Scan the selection string 		***/
	/**************************************************/
	if (sscanf(selstr,"%d %d",&atomid1,&atomid2)!=2)
		goto illegal_pickbond_sel;

	if ((atomid1<=0)||(atomid1>global_data->total_no_atoms))
		goto illegal_pickbond_sel;

	if ((atomid2<=0)||(atomid2>global_data->total_no_atoms))
		goto illegal_pickbond_sel;

	txtout("PICK-statement ok for atoms: %d %d.\n",atomid1,atomid2);
	atomid1--;	/*** EGO-numbering of atoms starts with 0 ***/
	atomid2--;

	/**************************************************/
	/*** alloc/realloc space 			***/
	/**************************************************/
	if (pinfo->len<=pinfo->cur_len) {
		if (pinfo->len==0) {
			pinfo->atom_ids = alloc_ints(2*100);
			pinfo->len      = 100;
		} else {
			realloc_ints(	&(pinfo->atom_ids),
					2*100,
					2*pinfo->len);
			pinfo->len      += 100;
		}
	}

	/**************************************************/
	/*** enter selection to list 			***/
	/**************************************************/
	i = 2*pinfo->cur_len;

	atom_sele_flag1[atomid1] |= mask;
	atom_sele_flag1[atomid2] |= mask;

	pinfo->atom_ids[i  ] = atomid1;
	pinfo->atom_ids[i+1] = atomid2;
	pinfo->cur_len++;
	return 0;

illegal_pickbond_sel:
	errout("Fatal error: Illegal selection string <%s> for %d!\n",
		selstr,mask);
	pex(-1);
	return -1;
}

static int
parse_pickangle(char *selstr,
		STRUCT_ATOMDATA *atom_data,
		int *atom_sele_flag1,
		LOAD_MSG_ALL_NODE *global_data,
		STRUCT_PICK_BOND *pinfo,
		int mask)
{
	int	atomid1=0,atomid2=0,atomid3=0,i=atom_data[0].atomnr;

	/**************************************************/
	/***  Check/Scan the selection string 		***/
	/**************************************************/
	if (sscanf(selstr,"%d %d %d",&atomid1,&atomid2,&atomid3)!=3)
		goto illegal_pickangle_sel;

	if ((atomid1<=0) || (atomid1>global_data->total_no_atoms))
		goto illegal_pickangle_sel;

	if ((atomid2<=0) || (atomid2>global_data->total_no_atoms))
		goto illegal_pickangle_sel;

	if ((atomid3<=0) || (atomid3>global_data->total_no_atoms))
		goto illegal_pickangle_sel;

	txtout("PICK-statement ok for atoms: %d %d %d.\n",
		atomid1,atomid2,atomid3);
	atomid1--;	/*** EGO-numbering of atoms starts with 0 ***/
	atomid2--;
	atomid3--;

	/**************************************************/
	/*** alloc/realloc space 			***/
	/**************************************************/
	if (pinfo->len<=pinfo->cur_len) {
		if (pinfo->len==0) {
			pinfo->atom_ids = alloc_ints(3*100);
			pinfo->len      = 100;
		} else {
			realloc_ints(	&(pinfo->atom_ids),
					3*100,
					3*pinfo->len);
			pinfo->len      += 100;
		}
	}

	/**************************************************/
	/*** enter selection to list 			***/
	/**************************************************/
	i = 3*pinfo->cur_len;

	atom_sele_flag1[atomid1] |= mask;
	atom_sele_flag1[atomid2] |= mask;
	atom_sele_flag1[atomid3] |= mask;

	pinfo->atom_ids[i  ] = atomid1;
	pinfo->atom_ids[i+1] = atomid2;
	pinfo->atom_ids[i+2] = atomid3;
	pinfo->cur_len++;
	return 0;

illegal_pickangle_sel:
	errout("Fatal error: Illegal selection string <%s> for %d!\n",
		selstr,mask);
	pex(-1);
	return -1;
}

static int
parse_pickdihe(	char *selstr,
		STRUCT_ATOMDATA *atom_data,
		int *atom_sele_flag1,
		LOAD_MSG_ALL_NODE *global_data,
		STRUCT_PICK_BOND *pinfo,
		int mask)
{
	int	atomid1=0,atomid2=0,atomid3=0,atomid4=0,i=atom_data[0].atomnr;

	/**************************************************/
	/***  Check/Scan the selection string 		***/
	/**************************************************/

	if (sscanf(selstr,"%d %d %d %d",&atomid1,&atomid2,&atomid3,&atomid4)!=4)
		goto illegal_pickdihe_sel;

	if ((atomid1<=0) || (atomid1>global_data->total_no_atoms))
		goto illegal_pickdihe_sel;
	if ((atomid2<=0) || (atomid2>global_data->total_no_atoms))
		goto illegal_pickdihe_sel;
	if ((atomid3<=0) || (atomid3>global_data->total_no_atoms))
		goto illegal_pickdihe_sel;
	if ((atomid4<=0) || (atomid4>global_data->total_no_atoms))
		goto illegal_pickdihe_sel;

	txtout("PICK-statement ok for atoms: %d %d %d %d.\n",
		atomid1,atomid2,atomid3,atomid4);
	atomid1--;	/*** EGO-numbering of atoms starts with 0 ***/
	atomid2--;
	atomid3--;
	atomid4--;

	/**************************************************/
	/*** alloc/realloc space 			***/
	/**************************************************/
	if (pinfo->len<=pinfo->cur_len) {
		if (pinfo->len==0) {
			pinfo->atom_ids = alloc_ints(4*100);
			pinfo->len      = 100;
		} else {
			realloc_ints(	&(pinfo->atom_ids),
					4*100,
					4*pinfo->len);
			pinfo->len      += 100;
		}
	}

	/**************************************************/
	/*** enter selection to list 			***/
	/**************************************************/
	i = 4*pinfo->cur_len;

	atom_sele_flag1[atomid1] |= mask;
	atom_sele_flag1[atomid2] |= mask;
	atom_sele_flag1[atomid3] |= mask;
	atom_sele_flag1[atomid4] |= mask;

	pinfo->atom_ids[i  ] = atomid1;
	pinfo->atom_ids[i+1] = atomid2;
	pinfo->atom_ids[i+2] = atomid3;
	pinfo->atom_ids[i+3] = atomid4;
	pinfo->cur_len++;
	return 0;

illegal_pickdihe_sel:
	errout("Fatal error: Illegal selection string <%s> for %d!\n",
		selstr,mask);
	pex(-1);
	return -1;
}

void
ini_my_forces(STRUCT_FORCES *my_force,LOAD_MSG_ALL_NODE *global_data)
{
	int i;
	double dx,dy,dz;

	/*** allocate no memory at all, if no forces to write out ***/
	my_force->total.len 	= my_force->total.cur_len 	= 0;
 	my_force->elecvdw.len 	= my_force->elecvdw.cur_len 	= 0;
 	my_force->bond.len 	= my_force->bond.cur_len 	= 0;
	my_force->elec.len 	= my_force->elec.cur_len 	= 0;
	my_force->vdw.len 	= my_force->vdw.cur_len 	= 0;
	my_force->angle.len 	= my_force->angle.cur_len 	= 0;
	my_force->dihe.len 	= my_force->dihe.cur_len 	= 0;
	my_force->impr.len 	= my_force->impr.cur_len 	= 0;
	my_force->restr.len 	= my_force->restr.cur_len 	= 0;
	my_force->sbound.len 	= my_force->sbound.cur_len 	= 0;
	my_force->intern.len 	= my_force->intern.cur_len 	= 0;
	my_force->flood.len 	= my_force->flood.cur_len 	= 0;
	my_force->kin.len 	= my_force->kin.cur_len 	= 0;
	my_force->qm.len 	= my_force->qm.cur_len 		= 0;
#if DEBUG
txtout("in ini_my_forces : forceout_nr_atoms=%d\n",
global_data->forceout_nr_atoms);
#endif
	if (global_data->forceout_nr_atoms<=0)	return;

	/*** here we count from how many atoms we need the specific	***/
	/*** forces --> Later only this is allocated 			***/
	for (i=0;i<global_data->total_no_atoms;i++) {
		int	force_id = global_data->atom_sele_flag1[i];
/***
txtout("Test %d: force_id %d\n",i,force_id);
***/
	  	if (force_id&FORCE_TOTAL) 	my_force->total.len++;
		if (force_id&FORCE_ELECVDW)	my_force->elecvdw.len++;
	  	if (force_id&FORCE_BOND)	my_force->bond.len++;
		if (force_id&FORCE_ELEC)	my_force->elec.len++;
		if (force_id&FORCE_VDW)		my_force->vdw.len++;
		if (force_id&FORCE_ANGLE)	my_force->angle.len++;
		if (force_id&FORCE_DIHE)	my_force->dihe.len++;
		if (force_id&FORCE_IMPR)	my_force->impr.len++;
		if (force_id&FORCE_RESTR)	my_force->restr.len++;
		if (force_id&FORCE_SBOUND)	my_force->sbound.len++;
		if (force_id&FORCE_INTERN)	my_force->intern.len++;
		if (force_id&FORCE_FLOOD)	my_force->flood.len++;
		if (force_id&FORCE_KIN)		my_force->kin.len++;
		if (force_id&FLAG_QMATOM)	my_force->qm.len++;
	}
#if DEBUG
txtout("Alloc TOTAL  : %d\n",my_force->total.len);
txtout("Alloc ELECVDW: %d\n",my_force->elecvdw.len);
#endif

	/*** alloc the space ***/
	my_force->total.force	= alloc_floats(FORCEINFOLEN*my_force->total.len);
	my_force->total.atom_id	= alloc_ints(my_force->total.len);

	my_force->elecvdw.force	= alloc_floats(FORCEINFOLEN*my_force->elecvdw.len);
	my_force->elecvdw.atom_id= alloc_ints(my_force->elecvdw.len);

	my_force->bond.force	= alloc_floats(FORCEINFOLEN*my_force->bond.len);
	my_force->bond.atom_id	= alloc_ints(my_force->bond.len);

	my_force->elec.force	= alloc_floats(FORCEINFOLEN*my_force->elec.len);
	my_force->elec.atom_id	= alloc_ints(my_force->elec.len);

	my_force->vdw.force	= alloc_floats(FORCEINFOLEN*my_force->vdw.len);
	my_force->vdw.atom_id	= alloc_ints(my_force->vdw.len);

	my_force->angle.force	= alloc_floats(FORCEINFOLEN*my_force->angle.len);
	my_force->angle.atom_id	= alloc_ints(my_force->angle.len);

	my_force->dihe.force	= alloc_floats(FORCEINFOLEN*my_force->dihe.len);
	my_force->dihe.atom_id	= alloc_ints(my_force->dihe.len);

	my_force->impr.force	= alloc_floats(FORCEINFOLEN*my_force->impr.len);
	my_force->impr.atom_id	= alloc_ints(my_force->impr.len);

	my_force->restr.force	= alloc_floats(FORCEINFOLEN*my_force->restr.len);
	my_force->restr.atom_id	= alloc_ints(my_force->restr.len);

	my_force->sbound.force	= alloc_floats(FORCEINFOLEN*my_force->sbound.len);
	my_force->sbound.atom_id= alloc_ints(my_force->sbound.len);

	my_force->intern.force	= alloc_floats(FORCEINFOLEN*my_force->intern.len);
	my_force->intern.atom_id= alloc_ints(my_force->intern.len);

	my_force->flood.force	= alloc_floats(FORCEINFOLEN*my_force->flood.len);
	my_force->flood.atom_id	= alloc_ints(my_force->flood.len);

	my_force->kin.force	= alloc_floats(FORCEINFOLEN*my_force->kin.len);
	my_force->kin.atom_id	= alloc_ints(my_force->kin.len);

	dx=global_data->qmboxext[1]-global_data->qmboxext[0];
	dy=global_data->qmboxext[3]-global_data->qmboxext[2];
	dz=global_data->qmboxext[5]-global_data->qmboxext[4];

	my_force->qm.len_dummys      = 5000;
	my_force->qm.sumlinkforce[0] = 0.0;
	my_force->qm.sumlinkforce[1] = 0.0;
	my_force->qm.sumlinkforce[2] = 0.0;	
	my_force->qm.excltype	= alloc_ints(my_force->qm.len_dummys);
	my_force->qm.dummy_id	= alloc_ints(my_force->qm.len_dummys);
	my_force->qm.dummy_q	= alloc_doubles(my_force->qm.len_dummys);
	my_force->qm.dummy_ko	= alloc_doubles(3*my_force->qm.len_dummys);
	my_force->qm.ko		= alloc_doubles(3*my_force->qm.len);
	my_force->qm.kolink	= alloc_doubles(3*my_force->qm.len);
	my_force->qm.elecf	= alloc_doubles(4*my_force->qm.len); /*** forces + charge ***/
	my_force->qm.qm_force	= alloc_doubles(3*my_force->qm.len);
	my_force->qm.qm_q	= alloc_doubles(my_force->qm.len);
	my_force->qm.local	= alloc_doubles(10*my_force->qm.len);
	my_force->qm.bondf	= alloc_doubles(4*my_force->qm.len);
	my_force->qm.anglef	= alloc_doubles(4*my_force->qm.len);
	my_force->qm.dihef	= alloc_doubles(4*my_force->qm.len);
	my_force->qm.imprf	= alloc_doubles(4*my_force->qm.len);
		
	my_force->qm.nodeatomi	= alloc_ints(my_force->qm.len);
	my_force->qm.node	= alloc_ints(my_force->qm.len);
	my_force->qm.qmmap	= alloc_ints(my_force->qm.len);
	my_force->qm.atom_id	= alloc_ints(my_force->qm.len);
	my_force->qm.flag	= alloc_ints(my_force->qm.len);
	my_force->qm.incllen	= alloc_ints(my_force->qm.len);	
	my_force->qm.incllist	= alloc_ints(MAXDUMMYSPERQMATOM*my_force->qm.len);		

	/*** truncate force files in case of restart! ***/
	truncate_forcefile(FFN_TOTAL,global_data->cycles_passed,global_data);
	truncate_forcefile(FFN_ELECVDW,global_data->cycles_passed,global_data);
	truncate_forcefile(FFN_BOND,global_data->cycles_passed,global_data);
	truncate_forcefile(FFN_ELEC,global_data->cycles_passed,global_data);
	truncate_forcefile(FFN_VDW,global_data->cycles_passed,global_data);
	truncate_forcefile(FFN_ANGLE,global_data->cycles_passed,global_data);
	truncate_forcefile(FFN_DIHE,global_data->cycles_passed,global_data);
	truncate_forcefile(FFN_IMPR,global_data->cycles_passed,global_data);
	truncate_forcefile(FFN_RESTR,global_data->cycles_passed,global_data);
	truncate_forcefile(FFN_SBOUND,global_data->cycles_passed,global_data);
	truncate_forcefile(FFN_INTERN,global_data->cycles_passed,global_data);
	truncate_forcefile(FFN_FLOOD,global_data->cycles_passed,global_data);
	truncate_forcefile(FFN_KIN,global_data->cycles_passed,global_data);

	truncate_forcefile(FFN_PICKBOND,global_data->cycles_passed,global_data);
	truncate_forcefile(FFN_PICKANGLE,global_data->cycles_passed,global_data);
	truncate_forcefile(FFN_PICKELEC,global_data->cycles_passed,global_data);
	truncate_forcefile(FFN_PICKVDW,global_data->cycles_passed,global_data);
	truncate_forcefile(FFN_PICKDIHE,global_data->cycles_passed,global_data);
	truncate_forcefile(FFN_PICKIMPR,global_data->cycles_passed,global_data);

	truncate_forcefile("qmtraj.out",global_data->cycles_passed,global_data);
}

void
reset_force_buffer(STRUCT_FORCES *my_force,LOAD_MSG_ALL_NODE *global_data)
{

{
	int difstep=global_data->cycles_passed-global_data->forceout_start_step;
#if DEBUG
	txtout("Step %d atoms %d Step %d Stride %d test %d\n",
		global_data->cycles_passed,
		global_data->forceout_nr_atoms,
		global_data->forceout_start_step,
		global_data->forceout_stride_step,
		difstep%global_data->forceout_stride_step);
#endif
	if (global_data->forceout_nr_atoms<=0)			return;
	if (difstep<0) 						return;
	if (global_data->forceout_stride_step<=0)		return;
	if ((difstep%global_data->forceout_stride_step)!=0) 	return;
}

	if (	(my_force->total.len!=my_force->total.cur_len) ||
		(my_force->elecvdw.len!=my_force->elecvdw.cur_len) ||
		(my_force->bond.len!=my_force->bond.cur_len) ||
		(my_force->elec.len!=my_force->elec.cur_len) ||
		(my_force->vdw.len!=my_force->vdw.cur_len) ||
		(my_force->dihe.len!=my_force->dihe.cur_len) ||
		(my_force->impr.len!=my_force->impr.cur_len) ||
		(my_force->angle.len!=my_force->angle.cur_len) ||
		(my_force->restr.len!=my_force->restr.cur_len) ||
		(my_force->sbound.len!=my_force->sbound.cur_len) ||
		(my_force->intern.len!=my_force->intern.cur_len) ||
		(my_force->flood.len!=my_force->flood.cur_len) ||
		(my_force->kin.len!=my_force->kin.cur_len) ) {
		errout("Fatal error: Inconsistency of force data in reset_force_buffer\n");
		pex(-1);

	}
	my_force->total.cur_len  = 0;
	my_force->elecvdw.cur_len= 0;
	my_force->bond.cur_len   = 0;
	my_force->elec.cur_len   = 0;
	my_force->vdw.cur_len    = 0;
	my_force->dihe.cur_len   = 0;
	my_force->impr.cur_len   = 0;
	my_force->angle.cur_len  = 0;
	my_force->restr.cur_len  = 0;
	my_force->sbound.cur_len = 0;
	my_force->intern.cur_len = 0;
	my_force->flood.cur_len  = 0;
	my_force->kin.cur_len    = 0;
}

void
free_my_forces(STRUCT_FORCES *my_force,LOAD_MSG_ALL_NODE *global_data)
{
	if (global_data->forceout_nr_atoms<=0)	return;

	myfree(my_force->total.force);
	myfree(my_force->total.atom_id);

	myfree(my_force->elecvdw.force);
	myfree(my_force->elecvdw.atom_id);

	myfree(my_force->bond.force);
	myfree(my_force->bond.atom_id);

	myfree(my_force->elec.force);
	myfree(my_force->elec.atom_id);

	myfree(my_force->vdw.force);
	myfree(my_force->vdw.atom_id);

	myfree(my_force->angle.force);
	myfree(my_force->angle.atom_id);

	myfree(my_force->dihe.force);
	myfree(my_force->dihe.atom_id);

	myfree(my_force->impr.force);
	myfree(my_force->impr.atom_id);

	myfree(my_force->restr.force);
	myfree(my_force->restr.atom_id);

	myfree(my_force->sbound.force);
	myfree(my_force->sbound.atom_id);

	myfree(my_force->intern.force);
	myfree(my_force->intern.atom_id);

	myfree(my_force->flood.force);
	myfree(my_force->flood.atom_id);

	myfree(my_force->kin.force);
	myfree(my_force->kin.atom_id);
}

/**************************************************************************/
/*** This function looks for a force file and compares the last written	***/
/*** integration step number with the integration step number next to	***/
/*** write. If int. steps overlap, the force file is truncated in order	***/
/*** to get the forces consistent with coordinates in the *.ego files	***/
/**************************************************************************/
static
int truncate_forcefile(	char *file_name,int step,
			LOAD_MSG_ALL_NODE *global_data)
{
	FILE	*oldh,*newh;
	char	newname[MAX_STRING_SIZE];
	char	tstr[MAX_STRING_SIZE];
	int	nr_atoms,j,cur_step;
	char	fname[MAX_STRING_SIZE];

	strcpy(fname,global_data->outpath);
	strcat(fname,global_data->idstr);
	strcat(fname,file_name);
	strcpy(newname,fname);
	strcat(newname,".tmp");

	if ((oldh=fopen(fname,"r"))==NULL)	return 0;

	if ((global_data->schwing_mode==0)&&(step<=0)) { /*** simulation from step 0 ***/
			/*** --> erase total file   ***/
		txtout("Force file <%s> erased.\n",fname);	
		fclose(oldh);
		remove(fname);
		return 0;
	}	

	if ((newh=fopen(newname,"w"))==NULL) {
		errout("\nFatal error: Can not create <%s>.\n",
			newname);
		fclose(oldh);
		pex(-1);
	}

	/*** read/write until "[BEGINFORCE]" or "[QMTRAJECTORY]" ***/
	while (read_string(oldh,sdum)!=-1) {	
		strcpy(tstr,sdum);tstr[12]=0;
		fprintf(newh,"%s\n",sdum);
		if (strcmp(tstr,"[BEGINFORCE]")==0)	break;
		if (strcmp(tstr,"[QMTRAJECTOR")==0)	break;
	}

	/*** read/write number of atoms ***/
	read_string(oldh,sdum);
	fprintf(newh,"%s\n",sdum);
	sscanf(sdum,"%s %d",tstr,&nr_atoms);

	/*** read/write number atom infos ***/
	read_string(oldh,sdum);
	fprintf(newh,"%s\n",sdum);
	for (j=0;j<nr_atoms;j++) {
		read_string(oldh,sdum);
		fprintf(newh,"%s\n",sdum);
	}
	
	/*** for normal simulation read until cur_step>=step ***/
	/*** for calculation of hess matrix do not truncate  ***/
	while (read_string(oldh,sdum)!=-1) {	
		sscanf(sdum,"%s %d",tstr,&cur_step);
                if (strcmp(tstr,"Restart")==0) {
                     fprintf(newh,"%s\n",sdum);
                     continue;
                } else		
		if (strcmp(tstr,"Step")!=0) {
			errout("Fatal error: Force file <%s> corrupt!\n",
				fname);
			errout("<%s>\n",sdum);
			fclose(oldh);
			fclose(newh);
			pex(-1);
		}

		/*** reach end ***/
		if (global_data->schwing_mode==0)
			if (cur_step>=step) break;

		/*** add next integration step ***/
		fprintf(newh,"%s\n",sdum);	
		for (j=0;j<nr_atoms;j++) {
			read_string(oldh,sdum);
			fprintf(newh,"%s\n",sdum);
		}
	}

	if (global_data->schwing_mode!=0) {
		fprintf(newh,"Restart of HESS-Calculation!\n");		
	}

	fclose(oldh);
	fclose(newh);
	remove(fname);
	rename(newname,fname);
	txtout("Force file <%s> truncated.\n",fname);	
	return 0;
}

