/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: egcontro.c	by M.Eichinger, Munich			**
**	Belongs2: EGO							**
**	Contents: startup of EGO,					**	
**		  envokes the loader,					**
**		  main calculation loop for sequential EGO version.	**
**									**
*************************************************************************/

#if defined(PVM)
#include "ego.h"
double get_min_unitdist(int flag,
                int my_atom,int other_unitnum,
                LOAD_MSG_ONE_NODE *my_data,
                TRAVELLING_PACKETS *packet)
{
	/*** dummy ****/
	return 0.0;
}
#endif

#if defined(SEQ)

#define DEBUG 		0
#define DEBUGQM		0

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <unistd.h>

#include "ego.h"
#include "egbolib.h"
#include "eganaly.h"
#include "egcolib.h"
#include "egloader.h"
#include "egnolib.h"
#include "egintern.h"
#include "egintegr.h"
#include "egtree.h"
#include "egrotflood.h"
#include "egcontro.h"
#include "egalloc.h"
#include "util_lib.h"
#include "egpull.h"
#include "egbound.h"
#include "egqminter.h"
#include "egpressure.h"
#include "timing.h"

void	master_sigterm_proc(int signo);
		
static int	handle_extern_dynamic(	LOAD_MSG_ALL_NODE *global_data,
					LOAD_MSG_ONE_NODE *my_data);


/******************************************************************/
/*** for the sequential version the main calculation loop is 	***/
/*** given below in seq_master(...)				***/
/*** ini_calculation(...) and stop_calculation(...) manage 	***/
/*** initialization and clean-up.				***/
/*** do_all_calculations(...) does the force calculation for    ***/
/*** every integration step.					***/
/******************************************************************/ 


#define STEP_NEW_WWLIST		64


static	int	skipcounter=0;
static	int	timerhandle[20];
static	int	my_number=0;
static	int	do_analyze,miss_first,restartlevel,coords_are_coming;
static	int	tree_mode,steps_after_reclust,new_clustering;
static	TRAVELLING_PACKETS	*packet;
static	ENERGY_INFO 	estm_local_energy,dummy_energy;
static	TRAVELLING_FORCES my_new_forces;	/* not really travelling ! */
static	TRAVELLING_FORCES my_exact_LRforces;
static 	double 	*rnd_force_std_deviation,*Gamma;/* stochastic boundary */
static	int    	my_first_atom;
static	int	*node_status;
static	EXT_3_ATOM 	*ext_angle;
static	EXT_4_ATOM 	*ext_dihe;
static	EXT_4_ATOM 	*ext_impr;
static	LOAD_MSG_ONE_NODE *my_data;	

static int stop_calculation(LOAD_MSG_ALL_NODE *global_data);
static int ini_calculation(	LOAD_MSG_ALL_NODE *global_data,
				ENERGY_INFO *local_energy);
static int do_all_calculations(	LOAD_MSG_ALL_NODE *global_data,
				TRAVELLING_COORDINATES *coordinates,
				ENERGY_INFO *local_energy,
				STRUCT_FORCES *my_force);

static	int	calculate_time(ENERGY_INFO *energy);
static	int	mystart_timer(int timer);
static	int	mystop_timer(int timer);

static int resort_myforces(	LOAD_MSG_ONE_NODE *my_data,
				LOAD_MSG_ALL_NODE *global_data,
				STRUCT_FORCES *my_force);

static	int	do_load_output=0;

void   seq_master(LOAD_MSG_ALL_NODE *global_data,
		double time_step,
		int file_count,
		char *version_string,
		STRUCT_FORCES *my_force)
{
	double temperature = 1234567890.0;
	double temperature_h2o     = 1234567890.0;
	double temperature_protein = 1234567890.0;

	TRAVELLING_COORDINATES coordinates;
	ENERGY_INFO local_energy;
	double 	scaling_factor=1.0;
 	double	temp_factor,temp_factor_h2o=1.0,temp_factor_protein=1.0;
 	double	cycletime_user=0.0,cycletime_real=0.0;	
	int	cycletimer,l,k;
	
	set_terminate_flag(DONT_STOP);
	do_load_output=global_data->debug_flags[1]&1;
	/*********************************************************************/
  	/*** 2/(3*N*Kb) in AKMA 3(N-2) is the number of degrees of freedom ***/
   	/*** (velocities) we have to disregard velocity of the center of   ***/
   	/*** mass and the total angular momentum, therefore we take 6 DOF  ***/
   	/*** away. if shaking hydrogens, we further subtract num_shake/2 as***/
	/*** we remove one (velocity) DOF per hydrogen atom. Temperture in ***/
	/*** Xplor is calculated with 3N DOF (this is wrong !!)		   ***/
	/*********************************************************************/
	if (global_data->total_no_atoms<3) {
		temp_factor = 2.0 / (
			KBOLTZ*(3.0*global_data->total_no_atoms-global_data->num_shake/2.0));
	} else {
		temp_factor = 2.0 / (
			KBOLTZ*(3.0*(global_data->total_no_atoms-2.0)-global_data->num_shake/2.0));
		temp_factor_h2o = 2.0 / (
			KBOLTZ*(3.0*(global_data->total_h2o_atoms-2.0)-global_data->num_shake_h2o/2.0));
		temp_factor_protein = 2.0 / (
			KBOLTZ*(3.0*(global_data->total_protein_atoms-2.0)-global_data->num_shake_protein/2.0));
	}	
#if DEBUG
txtout("Master: enter master.\n");
#endif

	set_allocinfo("travelling coordinates");
	coordinates.no_atoms 	= (int) global_data->maxatoms_per_node;
	coordinates.node 	= 0;
	alloc_TRAVELLING_COORDINATES(&coordinates);

	set_allocinfo("inirestart");
	initialize_restart();
	set_allocinfo("inianalyze");
	initialize_analyze();
	ini_constant_pressure(global_data->sbound_droff, 
			      global_data->keep_press_const);
	set_allocinfo("XXX");

	miss_first        = 0;
	do_analyze	  = 1;	  
	restartlevel      = 0;
	coords_are_coming = do_analyze;
	new_clustering    = global_data->treeinfo.steps;
	cycletime_user 	  = 0.0;
	cycletime_real	  = 0.0;

	ini_calculation(global_data,&local_energy);
	print_meminfo();
	
	/*** Now we are really computing and now we manipulate the signals ***/
	set_signal_procs(master_sigterm_proc);

	cycletimer = init_timer(); /*** for the spent computer time ***/ 
	clear_timer(cycletimer);
	start_timer(cycletimer);
#if (DEBUG & 8) == 8
        testmalloc("before loop over steps in egcontro(...)",1);
#endif

	local_energy.pullspringpos[0]=global_data->pullspringpos[0];
	local_energy.pullspringpos[1]=global_data->pullspringpos[1];
	local_energy.pullspringpos[2]=global_data->pullspringpos[2];

	/******************************************************************/
	/*** Here the main loop of every integration step starts	***/	
	/******************************************************************/
	while (global_data->cycles_passed<=global_data->final_cycle_no) {

		if (get_terminate_flag()==AFTER_STEP) {
			pex(1);	
		}
		if ((global_data->rotation_mode) && (global_data->cycles_passed==global_data->selimob_start)) {
		  local_energy.rotation_energy = 0.0;
		  global_data->rotation_energy = 0.0;
		}

		/*** here all forces are calculated and the 	***/
		/*** integration step is done			***/ 
		do_all_calculations(	global_data,
					&coordinates,
					&local_energy,
					my_force);
#if DEBUG
txtout("Skipcounter after do_all_calculations is %d\n",skipcounter);
#endif
#if (DEBUG & 8) == 8
        testmalloc("after do_all_calculations in egcontro(...)",1);
#endif
		/*** get spent time of current integration step ***/
		/*** and reset timer again. 			***/
		stop_timer(cycletimer);	
		cycletime_user = (double)get_user_timer(cycletimer);
		cycletime_real = (double)get_rt_timer(cycletimer);
		clear_timer(cycletimer);
		start_timer(cycletimer);
		
		if (do_load_output) {
	        	txtout("      TREE1  TREE2  OTHER  EXACT  STUFF = SUMM\n");
			txtout("     %6.3f %6.3f %6.3f %6.3f %6.3f = %6.3f\n",
				local_energy.time[0],local_energy.time[1],
				local_energy.time[2],local_energy.time[3],
				local_energy.time[4],
				local_energy.time[0]+local_energy.time[1]+
				local_energy.time[2]+local_energy.time[3]+
				local_energy.time[4]);
		}			

#if DEBUG
txtout("Master: cycles_passed=%d,final_cycle_no=%d\n",
global_data->cycles_passed,global_data->final_cycle_no);
#endif
		/*** In case of an anlysis step, restart, or reclustering ***/
		/*** some special things have to be made		  ***/
		if (	(skipcounter==0) &&
			(	(coords_are_coming) ||
				(new_clustering<=1) ||
				((global_data->rotation_mode) && 
				 (global_data->cycles_passed==global_data->selimob_start)) ||
				((global_data->nr_qmatoms>0)&&(global_data->schwing_mode==0)) 
				)
			) {
			
		  sort_to_analybuffer(new_clustering,&coordinates,global_data);

			/*** if we are in the QM/MM interface mode we want to write a ***/
			/*** restart file in every step of the simulation in order to ***/
			/*** be synchronised with the CPMD restart file RESTART.1.    ***/
			/*** The strategy of writing the restart file in every step   ***/
			/*** works only in the sequential version! 		      ***/	
			if ((global_data->nr_qmatoms>0)&&(global_data->schwing_mode==0)) {
				restart(1,				
					global_data->cycles_passed+1,
					file_count,
					time_step,
					global_data,
					&local_energy);

				memcpy(coordinates.x,my_data->x_new,sizeof(double)*my_data->no_atoms);
				memcpy(coordinates.y,my_data->y_new,sizeof(double)*my_data->no_atoms);
				memcpy(coordinates.z,my_data->z_new,sizeof(double)*my_data->no_atoms);
				sort_to_analybuffer(new_clustering,&coordinates,global_data);

				restart(2,				
					global_data->cycles_passed+1,
					file_count,
					time_step,
					global_data,
					&local_energy);

				memcpy(coordinates.x,my_data->x_old,sizeof(double)*my_data->no_atoms);
				memcpy(coordinates.y,my_data->y_old,sizeof(double)*my_data->no_atoms);
				memcpy(coordinates.z,my_data->z_old,sizeof(double)*my_data->no_atoms);
				sort_to_analybuffer(new_clustering,&coordinates,global_data);
			} else
			if ( 	(restartlevel==1) ||  /*** write first set  ***/
				(restartlevel==2) ) { /*** write second set ***/
#if DEBUG
txtout("Master: write out restart-file level %d  filecount %d cycles_passed %d\n",
	restartlevel,file_count,global_data->cycles_passed);
#endif
				restart(restartlevel,				
					global_data->cycles_passed+1,
					file_count,
					time_step,
					global_data,
					&local_energy);
				
				if (	(restartlevel==2) &&
					(get_terminate_flag()==AFTER_FILE) ) {
					set_terminate_flag(AFTER_STEP);
				} 
			}

			/*** new ego-output file has to be created ***/
			if (do_analyze) {
#if DEBUG
txtout("Master: write out ego-file %d\n",file_count);
#endif
				writeout_coordinates(	global_data,
							time_step,
							file_count,
							version_string);
				file_count++;
			}
#if DEBUG
txtout("Master: got and anlyzed all coords\n");
#endif
		} /**** end of if coords are comming ****/
		/***********************************************************/


		if ((global_data->rotation_mode) && (global_data->cycles_passed==global_data->selimob_start)) {
		  for (l=0,k=0;l<global_data->total_no_atoms;l++) {
		    if (!(global_data->atom_sele_flag2[l] & FLAG_ROT)) continue;
		    global_data->rot_x[k] = global_data->atomdata[l].ko[0];
		    global_data->rot_y[k] = global_data->atomdata[l].ko[1];
		    global_data->rot_z[k] = global_data->atomdata[l].ko[2];
		    global_data->rot1_x[k] = global_data->atomdata[l].ko[0];
		    global_data->rot1_y[k] = global_data->atomdata[l].ko[1];
		    global_data->rot1_z[k] = global_data->atomdata[l].ko[2];

		    global_data->rot2_x[k] = global_data->atomdata[l].ko[0];
		    global_data->rot2_y[k] = global_data->atomdata[l].ko[1];
		    global_data->rot2_z[k] = global_data->atomdata[l].ko[2];

		    global_data->rotation_torque[k]  = 0.0;
		    k++;
		  }
		}
		/***********************************************************/

		calc_imob_energy(&local_energy,global_data);

		adopt_pull_frame(&local_energy,global_data);

		if (global_data->rotation_mode) {
		  global_data->rotation_energy += local_energy.rotation_energy;
		}

		temperature = local_energy.kinetic * temp_factor;
		temperature_h2o     = local_energy.kinetic_h2o * temp_factor_h2o;
		temperature_protein = local_energy.kinetic_protein * temp_factor_protein;

		/*** only energies are added to the ego output file ***/
		if ((skipcounter==0)&&(!miss_first)) {
#if DEBUG
txtout("Master: before print analyze !miss_first\n");
#endif
			writeout_energy(&local_energy,
		 			temperature,temperature_h2o,temperature_protein,
		 			cycletime_user,cycletime_real,
		 			global_data,
		 			time_step);
#if DEBUG
txtout("Master: after print analyze !miss_first\n");
#endif
			writeout_myforces(global_data,my_force);
#if DEBUG
txtout("Master: after writeout_myforces\n");
#endif
		}

		if (	(skipcounter==0) &&
			(global_data->minimisation) &&
			((global_data->minikrit[0]>0.0)||(global_data->minikrit[1]>0.0)) ) {
			int krit1 = (	(local_energy.gradinfo[0]<=global_data->minikrit[0]) &&
					(global_data->minikrit[1]<=0.0) );

			int krit2 = (	((local_energy.gradinfo[1]/global_data->total_no_atoms)<=global_data->minikrit[1]) &&
					(global_data->minikrit[0]<=0.0) );				

			int krit3 = (	((local_energy.gradinfo[1]/global_data->total_no_atoms)<=global_data->minikrit[1]) &&
					(local_energy.gradinfo[0]<=global_data->minikrit[0]) );

			if ( krit1 || krit2 || krit3 ) {
				txtout("***************************************************\n");
				txtout("Reached minimisation criterion %g  %g.\n",
					global_data->minikrit[0],global_data->minikrit[1]);
				txtout("Maximum and average gradient was %g and %g at step %d.\n",
					local_energy.gradinfo[0],(local_energy.gradinfo[1]/global_data->total_no_atoms),
					global_data->cycles_passed);
				txtout("Simulation will be finished after next restart file!\n");	
				txtout("****************************************************\n");
				set_terminate_flag(AFTER_FILE);
				global_data->minikrit[0]=-1.0;
				global_data->minikrit[1]=-1.0;
			}
		}

		/* notes on equilibration: The equilibration scheme by velocity
		 * rescaling is described in the following paper: H. J. C.
		 * Berendsen, J. P. M. Postma, W. F. van Gunsteren, A. DiNola,
		 * J. R. Haak; "Molecular dynamics with coupling to an external
		 * bath"; J. Chem. Phys. 81(8), 15. Oct. 1984, pp 3684 - 3690; 
		 * Proportional scaling gives the least squares local 
		 * disturbance satisfying a global constraint. tau_T determines
		 * how fast the rescaling reacts to temperature deviations. 
                 * Don't use Tau_T smaller than 0.1 ps. Better use 0.5 ps    */

		if (global_data->equilibration) {
			if (temperature==0.0) temperature = global_data->target_temp;
			scaling_factor=sqrt(1.0 + 
					(global_data->target_temp/temperature-1.0)*
		       			(time_step/global_data->tau_T));

			rescale(fabs(scaling_factor),global_data,my_data);
#if DEBUG
txtout("Master: rescaling done with factor %f\n",scaling_factor);
#endif
		}

#if DEBUG
txtout("Master: setting new cycle skipcounter=%d  cycles passed %d do_analyze=%d\n",
skipcounter,global_data->cycles_passed,do_analyze,do_analyze);
#endif
		if (skipcounter==0) {
			global_data->cycles_passed++;
			set_working_flags(global_data,
				&do_analyze,
				&restartlevel,
				&miss_first,
				&coords_are_coming);
		}
		
		/*** Here we trigger the reclustering ***/
		/*** we dont want reclustering between writing the restart file ***/
		if ((skipcounter==0)&&(new_clustering==0)) {
			if (restartlevel==2) {	/*** do it not yet, but later ***/
				new_clustering=2; 
			} else {		/*** lets do reclustering ***/	
				mystop_timer(cycletimer);
				do_new_clustering(	global_data->total_no_atoms,
							global_data,
							0,
							my_force);
				new_clustering = global_data->treeinfo.steps;
				mystart_timer(cycletimer);
				reinit_node(global_data,
					my_data,
					&my_new_forces,
					&my_exact_LRforces,
	    				0,
					rnd_force_std_deviation,
					Gamma,
					packet);

				TREE_update(0,global_data,my_data);
				global_data->cycles_passed--;
				miss_first = 1;
				do_analyze = 0;
			}
#if (DEBUG & 8) == 8
        		testmalloc("after clustering in egcontro(...)",1);
#endif
		}
		if (skipcounter==0) new_clustering--;

	} /******* end of main while loop ********/

	close_output_file(global_data,file_count);
	finish_qmcalculation();
	close_extern_dynamic();
#if (DEBUG & 8) == 8
        testmalloc("after run in egcontro(...)",1);
#endif	
	txtout("Master: Simulation has finished now with step %d !\n",
		global_data->cycles_passed-1);
	print_meminfo();

	stop_calculation(global_data);
	free_timer(cycletimer);
	free_TRAVELLING_COORDINATES(&coordinates);

	free_loader_buffers(global_data);
	myfree(my_data->force_total);
	myfree(my_data->force_elecvdw);
	myfree(my_data->force_bond);
	myfree(my_data->force_elec);
	myfree(my_data->force_vdw);
	myfree(my_data->force_angle);
	myfree(my_data->force_dihe);
	myfree(my_data->force_impr);
	myfree(my_data->force_restr);
	myfree(my_data->force_sbound);
	myfree(my_data->force_flood);
	myfree(my_data->force_intern);
	myfree(my_data->force_kin);
	free_my_forces(my_force,global_data);
	myfree(my_data->inv_rotation_map);

	txtout("Master: CleanUp ended !\n");
	return;
}

int set_my_data(LOAD_MSG_ONE_NODE *m_d)
{
	my_data = m_d;
	return 0;
}

static int
resort_myforces(LOAD_MSG_ONE_NODE *my_data,
		LOAD_MSG_ALL_NODE *global_data,
		STRUCT_FORCES *my_force)
{
	int i;
	int difstep=global_data->cycles_passed-global_data->forceout_start_step;

	if (my_data->forceout_nr_atoms<=0) 			return 0;
	if (difstep<0) 						return 0;
	if (global_data->forceout_stride_step<=0)		return 0;
	if ((difstep%global_data->forceout_stride_step)!=0) 	return 0;

	for (i=0;i<my_data->no_atoms;i++) {
		int	atom_id  = my_data->atom_id[i];
		int	force_id = my_data->atom_sele_flag1[i];		
		double	fdata[FORCEINFOLEN];
		
		fdata[4] = my_data->x_new[i];
		fdata[5] = my_data->y_new[i];
		fdata[6] = my_data->z_new[i];
				
		if (force_id&FORCE_TOTAL) {
			memcpy(fdata,my_data->force_total+4*i,4*sizeof(double));
			sort_to_myforces(atom_id,&(my_force->total),fdata);		
		}	

		if (force_id&FORCE_ELECVDW) {
			memcpy(fdata,my_data->force_elecvdw+4*i,4*sizeof(double));
			sort_to_myforces(atom_id,&(my_force->elecvdw),fdata);
		}

		if (force_id&FORCE_BOND) {
			memcpy(fdata,my_data->force_bond+4*i,4*sizeof(double));
			sort_to_myforces(atom_id,&(my_force->bond),fdata);
		}

		if (force_id&FORCE_ELEC) {
			memcpy(fdata,my_data->force_elec+4*i,4*sizeof(double));		
			sort_to_myforces(atom_id,&(my_force->elec),fdata);
		}

		if (force_id&FORCE_VDW) {
			memcpy(fdata,my_data->force_vdw+4*i,4*sizeof(double));		
			sort_to_myforces(atom_id,&(my_force->vdw),fdata);
		}

		if (force_id&FORCE_ANGLE) {
			memcpy(fdata,my_data->force_angle+4*i,4*sizeof(double));		
			sort_to_myforces(atom_id,&(my_force->angle),fdata);
		}

		if (force_id&FORCE_DIHE) {
			memcpy(fdata,my_data->force_dihe+4*i,4*sizeof(double));		
			sort_to_myforces(atom_id,&(my_force->dihe),fdata);
		}

		if (force_id&FORCE_IMPR) {
			memcpy(fdata,my_data->force_impr+4*i,4*sizeof(double));		
			sort_to_myforces(atom_id,&(my_force->impr),fdata);
		}

		if (force_id&FORCE_RESTR) {
			memcpy(fdata,my_data->force_restr+4*i,4*sizeof(double));		
			sort_to_myforces(atom_id,&(my_force->restr),fdata);
		}

		if (force_id&FORCE_SBOUND) {
			memcpy(fdata,my_data->force_sbound+4*i,4*sizeof(double));		
			sort_to_myforces(atom_id,&(my_force->sbound),fdata);
		}
	
		if (force_id&FORCE_INTERN) {
			memcpy(fdata,my_data->force_intern+4*i,4*sizeof(double));
			sort_to_myforces(atom_id,&(my_force->intern),fdata);
		}

		if (force_id&FORCE_FLOOD) {
			memcpy(fdata,my_data->force_flood+4*i,4*sizeof(double));			
			sort_to_myforces(atom_id,&(my_force->flood),fdata);
		}

		if (force_id&FORCE_KIN) {
			memcpy(fdata,my_data->force_kin+4*i,4*sizeof(double));
			sort_to_myforces(atom_id,&(my_force->kin),fdata);
		}
	}

	if (my_data->pickbond.len) {
		memcpy(	my_force->pickbond.force,
			my_data->pickbond.force,
			5*sizeof(double)*my_data->pickbond.len);
	}	
	if (my_data->pickelec.len) {
		memcpy(	my_force->pickelec.force,
			my_data->pickelec.force,
			5*sizeof(double)*my_data->pickelec.len);
	}
	if (my_data->pickvdw.len) {
		memcpy(	my_force->pickvdw.force,
			my_data->pickvdw.force,
			5*sizeof(double)*my_data->pickvdw.len);
	}
	if (my_data->pickangle.len) {
		memcpy(	my_force->pickangle.force,
			my_data->pickangle.force,
			8*sizeof(double)*my_data->pickangle.len);
	}	
	if (my_data->pickdihe.len) {
		memcpy(	my_force->pickdihe.force,
			my_data->pickdihe.force,
			14*sizeof(double)*my_data->pickdihe.len);
	}
	if (my_data->pickimpr.len) {
		memcpy(	my_force->pickimpr.force,
			my_data->pickimpr.force,
			14*sizeof(double)*my_data->pickimpr.len);
	}

	reset_force_buffer(my_force,global_data);
	return 0;
}

/******************************************************************/
/*** here we copy the forces from STRUCT_QM_INFO to the 	***/
/*** LOAD_MSG_ONE_NODE structure.				***/
/******************************************************************/
int set_qmforces(LOAD_MSG_ALL_NODE *global_data,STRUCT_QM_INFO *finfo)
{
	int i;
#if DEBUGQM	
	double	sumqm[3],sumbo[3],suman[3],sumdi[3],sumim[3];
	sumqm[0]=sumqm[1]=sumqm[2]=0.0;
	sumbo[0]=sumbo[1]=sumbo[2]=0.0;
	suman[0]=suman[1]=suman[2]=0.0;
	sumdi[0]=sumdi[1]=sumdi[2]=0.0;
	sumim[0]=sumim[1]=sumim[2]=0.0;
#endif

	my_data->qmlocal.sumlinkforce[0]=finfo->sumlinkforce[0];
	my_data->qmlocal.sumlinkforce[1]=finfo->sumlinkforce[1];
	my_data->qmlocal.sumlinkforce[2]=finfo->sumlinkforce[2];
	for (i=0;i<global_data->nr_qmatoms;i++) {
		int atom_id = finfo->atom_id[i];
		int j       = get_qmnode_index(atom_id,my_data);

		my_data->qmlocal.atomi[j]	= finfo->nodeatomi[i];
		my_data->qmlocal.qm_q[j]	= finfo->qm_q[i];
		
		/*** that are the forces from th QM-program ***/
		/*** they have the right sign.              ***/

		my_data->qmlocal.qm_force[3*j  ]= finfo->qm_force[3*i  ];
		my_data->qmlocal.qm_force[3*j+1]= finfo->qm_force[3*i+1];
		my_data->qmlocal.qm_force[3*j+2]= finfo->qm_force[3*i+2];

		/*** now we add the classical forces from EGO***/
		/*** for the link-atom-situation. As         ***/
		/*** my_data->qmlocal.qm_force has the right ***/
		/*** sign, we must add finfo->bondf, anglef, ***/
		/*** etc. as they have also the right sign.  ***/

		my_data->qmlocal.qm_force[3*j  ] += finfo->bondf[4*i+1];
		my_data->qmlocal.qm_force[3*j+1] += finfo->bondf[4*i+2];
		my_data->qmlocal.qm_force[3*j+2] += finfo->bondf[4*i+3];
		
		my_data->qmlocal.qm_force[3*j  ] += finfo->anglef[4*i+1];
		my_data->qmlocal.qm_force[3*j+1] += finfo->anglef[4*i+2];
		my_data->qmlocal.qm_force[3*j+2] += finfo->anglef[4*i+3];

		my_data->qmlocal.qm_force[3*j  ] += finfo->dihef[4*i+1];
		my_data->qmlocal.qm_force[3*j+1] += finfo->dihef[4*i+2];
		my_data->qmlocal.qm_force[3*j+2] += finfo->dihef[4*i+3];
				
		my_data->qmlocal.qm_force[3*j  ] += finfo->imprf[4*i+1];
		my_data->qmlocal.qm_force[3*j+1] += finfo->imprf[4*i+2];
		my_data->qmlocal.qm_force[3*j+2] += finfo->imprf[4*i+3];					

#if DEBUGQM
		sumqm[0] +=finfo->qm_force[3*i  ];
		sumqm[1] +=finfo->qm_force[3*i+1];
		sumqm[2] +=finfo->qm_force[3*i+2];

		sumbo[0]+=finfo->bondf[4*i+1];
		sumbo[1]+=finfo->bondf[4*i+2];
		sumbo[2]+=finfo->bondf[4*i+3];	
		
		suman[0]+=finfo->anglef[4*i+1];
		suman[1]+=finfo->anglef[4*i+2];
		suman[2]+=finfo->anglef[4*i+3];	
		
		sumdi[0]+=finfo->dihef[4*i+1];
		sumdi[1]+=finfo->dihef[4*i+2];
		sumdi[2]+=finfo->dihef[4*i+3];
		
		sumim[0]+=finfo->imprf[4*i+1];
		sumim[1]+=finfo->imprf[4*i+2];
		sumim[2]+=finfo->imprf[4*i+3];
#endif
	}
#if DEBUGQM
	txtout("SUM QM-Part Bond: %f %f %f\n",sumbo[0],sumbo[1],sumbo[2]);
	txtout("SUM QM-Part Angl: %f %f %f\n",suman[0],suman[1],suman[2]);
	txtout("SUM QM-Part Dihe: %f %f %f\n",sumdi[0],sumdi[1],sumdi[2]);
	txtout("SUM QM-Part Impr: %f %f %f\n",sumim[0],sumim[1],sumim[2]);
	txtout("SUM CPMD        : %f %f %f\n",sumqm[0],sumqm[1],sumqm[2]);	
#endif
	return 0;
}

/******************************************************************/
static int
stop_calculation(LOAD_MSG_ALL_NODE *global_data)
{
	int i;
	TREE_exit(global_data);
	myfree(rnd_force_std_deviation);
	myfree(Gamma);
	myfree(node_status);

	for (i=0;i<global_data->no_nodes;i++) {
		free_TRAVELLING_PACKETS(&(packet[i]));
	}
	myfree(packet);

	free_TRAVELLING_FORCES(&my_new_forces);
	free_TRAVELLING_FORCES(&my_exact_LRforces);
	myfree(ext_angle);
	myfree(ext_dihe);
	myfree(ext_impr);
	free_rot_flood();
	free_pull(); 

	for (i=0;i<NUM_TIMERS;i++) {
		free_timer(timerhandle[i]);
	}

	return 0;
}

/******************************************************************/
static int
ini_calculation(	LOAD_MSG_ALL_NODE *global_data,
			ENERGY_INFO *local_energy)
{
	int i;


	for (i=0;i<NUM_TIMERS;i++) {
		timerhandle[i]       = init_timer();
		local_energy->time[i]= 0.0;
	}
	set_allocinfo("STUFF");
	node_status 		= alloc_ints(global_data->no_nodes);
	set_allocinfo("PACKETS");
	packet			= alloc_packets(global_data->no_nodes);
	set_allocinfo("XXX");

	my_first_atom = global_data->first_atom[my_number];
#if DEBUG
txtout("Node %d: before init node my_first_atom %d\n",my_number,my_first_atom);
#endif
	init_node(	global_data,
			my_data,
			&my_new_forces,
			&my_exact_LRforces,
	    		my_number,
			&rnd_force_std_deviation,
			&Gamma,
	    		packet,
			&ext_angle,
			&ext_dihe,
			&ext_impr);

	for (i=0;i<global_data->no_nodes;i++) {
		node_status[i] = 0;
	}

#if DEBUG
txtout("Node %d: after init node\n",my_number);
#endif
	
	set_allocinfo("TREE_ini_all");
	TREE_ini_all(my_number,global_data,my_data);
	set_allocinfo("XXX");
#if DEBUG
txtout("Node %d: after TREE_ini_all\n",my_number);
#endif

	tree_mode = 3;	/*** calc ww-list and calc new local expansion ***/
	steps_after_reclust = 0;
	return 0;
}

/******************************************************************/
static int
do_all_calculations(	LOAD_MSG_ALL_NODE *global_data,
			TRAVELLING_COORDINATES *coordinates,
			ENERGY_INFO *local_energy,
			STRUCT_FORCES *my_force)
{

#if DEBUG
txtout("Calculate: cycles_passed=%d,final_cycle_no=%d\n",
global_data->cycles_passed,global_data->final_cycle_no);
#endif
	mystart_timer(timerhandle[ORGA_STUFF]);

	if (global_data->extdynstr[0]!=0) {
		handle_extern_dynamic(global_data,my_data);
	}
	
	init_for_each_step(	local_energy,
				&estm_local_energy,
				&dummy_energy,		
				global_data,
				my_data,
				&my_new_forces,
				&my_exact_LRforces,
				my_number,
				packet,
				node_status);
	mystop_timer(timerhandle[ORGA_STUFF]);

#if DEBUG
txtout("Calculate: before TREE_calc_multipol\n");
#endif
	/*** Here the Multipole moments of structural units 	***/
	/*** and clusters are calculated			***/
	mystart_timer(timerhandle[CALC_TREE1]);
	TREE_start_loop();
	TREE_calc_multipol(	my_number,
				global_data,
				&(packet[my_number]));
	mystop_timer(timerhandle[CALC_TREE1]);
	
#if DEBUG
txtout("Calculate: after TREE_calc_multipol\n");
#endif
	mystart_timer(timerhandle[ORGA_STUFF]);
	{int level,toplevel;
	toplevel = my_data->levels-1;

	for (level=0;level<=toplevel;level++) {
		int nr_cl=my_data->cl_per_level[level];	
		memcpy(	my_data->cluster[level],
			packet[my_number].branch.cluster[level],
			sizeof(struct_cluster)*nr_cl);
	}
	}
	mystop_timer(timerhandle[ORGA_STUFF]);
	/*****************************************************/
	/********* now all orga-stuff ended ******************/
	/*****************************************************/

	mystart_timer(timerhandle[CALC_EXACT]);
	if (global_data->debug_flags[0]==1) { 
		/*** here extact elec + VDW forces are calculated     ***/
		/*** I used this for comparision with extrapol forces ***/
#if DEBUG
txtout("Calculate: before CalcIntern_LongRange\n");
#endif
		CalcIntern_LongRange (	&dummy_energy,
					global_data,
					my_data,
					&my_exact_LRforces,
					my_first_atom);
	}
	mystop_timer(timerhandle[CALC_EXACT]);

#if DEBUG
txtout("Calculate: before CalcIntern_ShortRange\n");
#endif
	/*** ShortRange means Bond, Angle, SBound, etc. forces ***/
	mystart_timer(timerhandle[CALC_OTHER]);
	define_range (my_data->no_atoms, my_data->y_new, my_data->atom_id);
	CalcIntern_ShortRange(	local_energy,
	       			my_data,
	       			global_data,
	       			&my_new_forces,
	       			rnd_force_std_deviation,
				Gamma);
	transfer_special(local_energy);
	mystop_timer(timerhandle[CALC_OTHER]);
#if DEBUG
txtout("Calculate: after CalcIntern_ShortRange\n");
#endif
	/*** LongRange means electrostatic and VDW forces: 	***/
	/*** Electrs. is approximated by FMM and MTS.	   	***/
	/*** VDW is approximated by a CutOff scheme and MTS.	***/	
	mystart_timer(timerhandle[CALC_TREE2]);
	TREE_calc_LongRange(	my_number,
				global_data,
				my_data,
				&(packet[my_number]),
				tree_mode);
	mystop_timer(timerhandle[CALC_TREE2]);
#if DEBUG
txtout("Calculate: after TREE_calc_LongRange\n");
#endif
	/**** needed for flooding and/or rotation-correction ****/
	mystart_timer(timerhandle[ORGA_STUFF]);
	merge_packet_to_buffer(	my_number,
				global_data->total_no_atoms,
				&(packet[my_number].coordinates));


        /**** needed for pulling ****/ 
        merge_packet_to_buffer_pull( my_number,
                                global_data->total_no_atoms,
                                &(packet[my_number].coordinates),             
                                global_data->atomdata);

	mystop_timer(timerhandle[ORGA_STUFF]);
	/*** here we add the electrostatic forces of structural	***/
	/*** units and clusters. (Local Taylor expansion)	***/

	mystart_timer(timerhandle[CALC_OTHER]);
	TREE_add_distforces(my_number,global_data,my_data);

	if (global_data->nr_qmatoms>0) {
		TREE_calc_qmlocal(my_number,global_data,my_data);
	}

#if WITH_EXACT_FORCES	/*** Was only used during development ***/
	for (i=0; i<my_new_forces.no_atoms ;i++) {
		my_new_forces.Fx[i] += my_exact_LRforces.Fx[i];
		my_new_forces.Fy[i] += my_exact_LRforces.Fy[i];
		my_new_forces.Fz[i] += my_exact_LRforces.Fz[i];
	}
#endif	
	if (skipcounter==0)
		do_rot_correction(global_data,
			  my_data->x_new,my_data->y_new,my_data->z_new,
			  my_data->x_old,my_data->y_old,my_data->z_old);

#if DEBUG
txtout("Calculate: before add_flood_forces\n");
#endif
	/***** for grubis flooding ************************/
	add_flood_forces(my_data,
			global_data,
			my_new_forces.Fx,
			my_new_forces.Fy,
			my_new_forces.Fz,
			local_energy);
#if DEBUG
txtout("Calculate: before add_imob_forces\n");
#endif		
	/***** Immobilisation forces *********************/

	add_imob_forces(my_data,
			global_data,
			my_new_forces.Fx,
			my_new_forces.Fy,
			my_new_forces.Fz);

	if (global_data->nr_qmatoms>0) {
#if DEBUG
txtout("Calculate: before add_qm_forces\n");
#endif
		add_qm_forces(	my_data,
				global_data,
				local_energy,
				my_force,
				&skipcounter);
	}

#if DEBUG
txtout("Calculate:before add_pull_forces\n");
#endif

	add_pull_forces(0,
                        local_energy,
			my_data,
			global_data,
			my_new_forces.Fx,
			my_new_forces.Fy,
			my_new_forces.Fz);


#if DEBUG
txtout("Calculate:before add_rotation_forces\n");
#endif

        add_rotation_forces(  0,
			      local_energy,
			      my_data,
			      global_data,
			      my_new_forces.Fx,
			      my_new_forces.Fy,
			      my_new_forces.Fz);

#if DEBUG
txtout("Calculate:before addup_all_forces\n");
#endif		
	addup_all_forces(local_energy,
			my_data,
			global_data,
			&my_new_forces);

	/*** in the sequential version these two functions ***/
	/*** follow right behind each other. Note: in the  ***/
	/*** prallel version a communication-step between the**/
	/*** nodes and the master is necessary, see egnode.c**/
	do_newrottranscor_part1(my_number,
                        local_energy,
                        my_data,
                        global_data,
                        packet);
                    
	do_newrottranscor_part2(my_number,
                        local_energy,
                        my_data,
                        global_data,
                        packet,
                        &my_new_forces);                   


#if DEBUG
txtout("Calculate:before resort_myforces\n");
#endif
	resort_myforces(my_data,
			global_data,
			my_force);
		
	/*** my_new_forces.F holds now the force on each atom	***/

#if DEBUG
txtout("Calculate:before calc_virial\n");
#endif
        calc_virial(local_energy,my_data,global_data,&my_new_forces);
	
	if (skipcounter==0) {
#if DEBUG
txtout("Calculate: before do_integrate.\n");
#endif		
		do_integrate(	local_energy,
		  		my_data,
		  		global_data,
		  		&my_new_forces,
				Gamma,
		  		my_number);
	}
	
	mystop_timer(timerhandle[CALC_OTHER]);

	calculate_time(local_energy);

#if DEBUG
txtout("Calculate: before copy coords %d\n",coords_are_coming);
#endif
	/*** Lets see if we have an analysis integration step ***/
	/*** or we have to write are restart file ***/
	if (	(coords_are_coming) ||
		(new_clustering==0) || (new_clustering==1) ||
		((global_data->rotation_mode) && 
		 (global_data->cycles_passed==global_data->selimob_start)) ||
		((global_data->nr_qmatoms>0)&&(global_data->schwing_mode==0))
	    ) {
		memcpy(coordinates->x,my_data->x_old,sizeof(double)*my_data->no_atoms);
		memcpy(coordinates->y,my_data->y_old,sizeof(double)*my_data->no_atoms);
		memcpy(coordinates->z,my_data->z_old,sizeof(double)*my_data->no_atoms);
	}

	/**************************************************/
	/**** new setup for the next integration-step *****/
	/**************************************************/

#if DEBUG
txtout("Calculate: setting up new step.\n");
#endif
	steps_after_reclust++;
	if ((steps_after_reclust%STEP_NEW_WWLIST)==0) {
		tree_mode = 3;
	} else {
		tree_mode = 2;	
	}

	if (new_clustering==0) {
		if (restartlevel==2) {	/*** do it not yet, but later ***/
			new_clustering=2; 
		} else {		/*** lets do reclustering ***/	
			tree_mode           = 3;
			steps_after_reclust = 0;
			reset_merge_buffer();
		}
	}
	return 0;
}

static
int mystart_timer(int timer)
{
	if (do_load_output) {
		start_timer(timer);
	}
	return 0;
}

static
int mystop_timer(int timer)
{
	if (do_load_output) {
		stop_timer(timer);
	}
	return 0;
}

static int
calculate_time(ENERGY_INFO *energy)
{
	int i;
	if (do_load_output) {
		for (i=0;i<NUM_TIMERS;i++) {
			energy->time[i] = get_user_timer(timerhandle[i]);
			clear_timer(timerhandle[i]);
		}
	}
	return 0;
}


static	int	extdyn_between=0,extdyn_count=0;
static	int	extdyn_laststep=0;
static	int	extdyn_atoms=0;
static	FILE 	*fh_extdyn;
		
static int
handle_extern_dynamic(	LOAD_MSG_ALL_NODE *global_data,
			LOAD_MSG_ONE_NODE *my_data)
{
	int i,j,res,atomid,mapi;
	double ko[3];
	char fname[MAX_STRING_SIZE],sdum[MAX_STRING_SIZE];

	/*** no extern dynamic ***/
	if (extdyn_atoms<0) return -1;
	if (global_data->extdyn_step[0]>global_data->cycles_passed) return -1;
	
	/*** no we have to start the extern dynamic ***/
	if (extdyn_atoms==0) {
		extdyn_laststep = global_data->extdyn_step[0] + 
			   	  global_data->extdyn_step[1] +
			   	  global_data->extdyn_step[1]*global_data->extdyn_step[2] -
			   	 (1 + global_data->extdyn_step[2]);
		
		/*** first check restart situation ***/
		if (extdyn_laststep<global_data->cycles_passed) {
			extdyn_atoms=-1; /*** restart step after extdyn_laststep ***/
			return 0;	 /*** nothing more to do		 ***/
		}
			   		
		strcpy(fname,global_data->lispath);
		strcat(fname,global_data->extdynstr);		
		if ((fh_extdyn = fopen(fname,"r")) == NULL) {
			errout("Fatal error: Cannot open extern dynamic file <%s>.\n",
				fname);
			pex(-1);
		}
	    	
		while (!feof(fh_extdyn)) {  /*** scan the header ***/
			
			read_string(fh_extdyn,fname);
			if ((fname[0]=='!')||(fname[0]=='#')) continue;
			
			if (sscanf(fname,"%s %d",sdum,&extdyn_atoms)!=2) {
				errout("Fatal error: Illegal extern dynamic file <%s> %d.\n",sdum,extdyn_atoms);
				pex(-1);
			}
			extdyn_between=global_data->extdyn_step[2];
			extdyn_count=0;
			txtout("EXTERN DYNAMIC: Starting for %d atoms.\n",extdyn_atoms);
			
			if (global_data->extdyn_step[0]==global_data->cycles_passed) break;
			
			/*** we have to seek to the right step in the file ***/
			extdyn_count = (global_data->cycles_passed - global_data->extdyn_step[0])/
				     (1+global_data->extdyn_step[2]);
				    
			txtout("EXTERN DYNAMIC: Skipping %d entries because of restart.\n",extdyn_count);
			for (i=0;i<extdyn_count;i++) {
				if (read_string(fh_extdyn,fname)==-1) goto endof_extfile;
				for (j=0;j<extdyn_atoms;j++) read_string(fh_extdyn,fname);
			}
			extdyn_between = global_data->cycles_passed - 1 - 
				(global_data->extdyn_step[0]+extdyn_count*(1+global_data->extdyn_step[2]));
			if (extdyn_between<0) extdyn_between=global_data->extdyn_step[2];
			break;
		}
	}

	/*** read first/next coordinate set ***/	
	if (extdyn_atoms==0) return 0;

	if (extdyn_count>=global_data->extdyn_step[1]) {
		txtout("EXTERN DYNAMIC: Extern dynamic stops now at step %d!\n",global_data->cycles_passed);
		fclose(fh_extdyn);
		extdyn_atoms = -1;
		return 0;	
	}

	extdyn_between++;
	if (extdyn_between<=global_data->extdyn_step[2]) {
		txtout("EXTERN DYNAMIC: Free dynamic at step %d\n",global_data->cycles_passed);
		return 0;
	}
	
	while (!feof(fh_extdyn)) {  /*** scan next entry ***/
		if (read_string(fh_extdyn,fname)==-1) break;
		if ((fname[0]=='!')||(fname[0]=='#')) continue;
			
		sscanf(fname,"%s",sdum);
		sdum[4]=0;

		if (strcmp(sdum,"Step")!=0) {
			errout("Fatal error: Illegal extern dynamic file reading frame.\n");
			pex(-1);
		}
		
		txtout("EXTERN DYNAMIC: Constrained dynamic at step %d\n",global_data->cycles_passed);
		for (i=0;i<extdyn_atoms;i++) {
			read_string(fh_extdyn,fname);
#if PRECISION == DOUBLE
			res = sscanf(fname,"%d %lf %lf %lf",&atomid,ko,ko+1,ko+2);
#else
			res = sscanf(fname,"%d %f %f %f",&atomid,ko,ko+1,ko+2);
#endif
			if ((res!=4)||(atomid<0)||(atomid>global_data->total_no_atoms)) {
				errout("Fatal error: Illegal extern dynamic file reading set.\n");
				pex(-1);	
			}
				
			mapi = global_data->old2new_map[atomid];
				
			my_data->x_new[mapi] = ko[0];
			my_data->y_new[mapi] = ko[1];
			my_data->z_new[mapi] = ko[2];				

			my_data->x_old[mapi] = ko[0];
			my_data->y_old[mapi] = ko[1];
			my_data->z_old[mapi] = ko[2];
		}
		extdyn_between=0;
		extdyn_count++;
		return 0;	
	}

endof_extfile:		
	txtout("EXTERN DYNAMIC: Reached end of file at step %d! Extern dynamic finished.\n",
		global_data->cycles_passed);
	fclose(fh_extdyn);
	extdyn_atoms = -1;
		
	return 0;	
}

int
close_extern_dynamic()
{
	if (extdyn_atoms>0) {
		txtout("EXTERN DYNAMIC: File closed!\n");
		fclose(fh_extdyn);	
		extdyn_atoms = -1;
	}	
	return 0;
}

#endif

