/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: egloader.c	by M.Eichinger, Munich			**
**	Belongs2: EGO							**
**	Contents: read control file,					**
**		  prepare and distribute data to nodes			**
**  	Modif's:							**
**									**
*************************************************************************/
#define DEBUG	0

#include <stdio.h>
#include <string.h>
#include "ego.h"
#include "egalloc.h"
#include "egbolib.h"
#include "egmach.h"
#include "egcolib.h"
#include "util_lib.h"
#include "egclust.h"
#include "eganaly.h"
#include "egloader.h"
#include "egcontro.h"
#include "egpull.h"
#include "egqminter.h"
#include "timing.h"

static	int	start_msg_numalloc=0;
static	LOAD_MSG_ONE_NODE start_msg;

static	int	mxno_bond_excl,		mxno_bond_ex14;
static	int	mxno_intern_bonds,	mxno_extern_bonds;
static	int	mxno_angles,		mxno_extern_angles;
static	int	mxno_dihed,		mxno_extern_dihed;
static	int	mxno_improp,		mxno_extern_improp;
static	int	mxlen_local_shake_list;

static	int	*last_atom;
static	int	*atoms_per_node;
static	int	*bonddata;
static	int	*excldata,*excl14data;
static	int	*angledata,size_anglelist=0;
static	int	*dihedata,size_dihelist=0;
static	int	*imprdata,size_imprlist=0;
static	int	*shakedata;

static	int	alloc_per_level[MAX_NO_LEVELS];

static	int	last_maxexcl=0,last_maxexcl14=0;

/***********************************************************************/

static	struct_cluster_reduced	*cluster[MAX_NO_LEVELS];/* cluster[0] holds later a pointer */
						/* to an array of cluster-structs */

static	int		*child_nr[MAX_NO_LEVELS];/* child_nr[0] holds later a pointer*/
						/* to an array of integers with */
						/* index_nr of child_clusters */
/************************************************************************/

static
int 	ini_global_data_defaults(LOAD_MSG_ALL_NODE *global_data,
				 STRUCT_FORCES *my_force);

static
int	do_node_select(	int request_no_nodes,
			int *no_nodes,
			int nProcs,
			int host_tid,
			int Nr_of_atoms);

static					
int distri_atomdata(	int nodenum,
			int first,
			int last,
			STRUCT_ATOMDATA *atom_data,
			LOAD_MSG_ONE_NODE *start_msg,
			int *new2old_map,
			LOAD_MSG_ALL_NODE *global_data);

static					
int 	distri_harmonic(int first,
			int last,
			STRUCT_ATOMDATA *atom_data,
			LOAD_MSG_ONE_NODE *start_msg,
			int *new2old_map);

static
int	distri_shake(	int first,
			int last,
			int *shakedata,
			LOAD_MSG_ONE_NODE *start_msg,
			int *mxno_shake,
			LOAD_MSG_ALL_NODE *global_data);

static
int 	distri_bonds(	int first,
			int last,
			int *bondlist,
			LOAD_MSG_ONE_NODE *start_msg,
			int *mxno_extern_bonds,
			int *mxno_intern_bonds);

static
int    is_local_shake_hyd(int atom,int shake_number,int *shake_list);

static
void   enter_ext(	int element,
			LOAD_MSG_ONE_NODE * start_msg,
			int *mxno_extern_bonds);

static
int	read_shake_file(LOAD_MSG_ALL_NODE *global_data,
			int *shake_num,
			int *shake_num_h2o,
			int *shake_num_protein,
			char *shakefilen,
			int **shakedata,
			int *old2new_map);

static
int	distri_excl(	int first,
			int last,
			int *excldata,
			LOAD_MSG_ONE_NODE *start_msg,
			int *mxno_bond_excl,
			int maxexcl,
			LOAD_MSG_ALL_NODE *global_data);

static
int	distri_excl14(	int first,
			int last,
			int *excldata,
			LOAD_MSG_ONE_NODE *start_msg,
			int *mxno_bond_excl,
			int maxexcl,
			LOAD_MSG_ALL_NODE *global_data);

static
int	distri_angles(	int size_anglelist,
			int first,
			int last,
			int *angledata,
			LOAD_MSG_ONE_NODE *start_msg,
			int *mxno_angles,
			int *mxno_extern_angles);

static
int	distri_dihe(	int size_dihelist,
			int first,
			int last,
			int *dihedata,
			LOAD_MSG_ONE_NODE *start_msg,
			int *mxno_dihed,
			int *mxno_extern_dihed);

static
int	distri_impr(	int size_imprlist,
			int first,
			int last,
			int *imprdata,
			LOAD_MSG_ONE_NODE *start_msg,
			int *mxno_improp,
			int *mxno_extern_improp);

static int	distri_tree(	int nodenum,int no_nodes,
			int first,
			struct_treeinfo *treeinfo,
			struct_levelinfo *levelinfo,
			struct_cluster_reduced *cluster[],
			LOAD_MSG_ONE_NODE *start_msg);

static int	read_parameter_data(	LOAD_MSG_ALL_NODE *global_data,
				char filenames[MAX_FILES][MAX_STRING_SIZE]);

static int alloc_loader_buffers(	int Nr_of_Atoms, 
			int no_nodes,
			STRUCT_FORCES *my_force);

static int
distri_pickselect(	int nodenum,
			int first,int last,
			STRUCT_FORCES *my_force,
			LOAD_MSG_ONE_NODE *start_msg,
			int *old2new);

static int	make_atommap(	int nr_of_atoms,
			int nr_of_nodes,
			struct_treeinfo *treeinfo,
			struct_levelinfo *levelinfo,
			struct_cluster_reduced *cluster[],
			int *child_nr[],
			int *old2new_map,
			int *new2old_map,
			int *atoms_per_node);

/***************************************************************************/

void   loader(	LOAD_MSG_ALL_NODE *global_data,
		double *time_step,
		int *file_count,
		int host_tid,
		int nProcs,	/* The number of real nodes in the network */
		STRUCT_FORCES *my_force)
{
	int	request_no_nodes;
	int	res,i;
	char	strtime[100];
	
	bonddata  =NULL;
	excldata  =NULL;
	excl14data=NULL;
	angledata =NULL;
	dihedata  =NULL;
	imprdata  =NULL;
	shakedata =NULL;

	for (i=0;i<MAX_NO_LEVELS;i++) {	/*** these buffers are allocated in CLU_buildup_clusters***/
		cluster[i] = NULL;	/*** here we ini them with NULL, so it is clear, that	***/
		child_nr[i]= NULL;	/*** they are not allocated yet				***/
	}

	/******************************************************************/
#if DEBUG
	txtout("Before read_ctl_file\n");
#endif
  	read_ctl_file( global_data->ctlfile, global_data);
	request_no_nodes = global_data->no_nodes;
	*time_step = global_data->time_step * TIMEDIV;

	txtout("Lis-files are in directory: <%s>\n",  global_data->lispath);
	txtout("Writing data to directory : <%s>\n",  global_data->outpath);
	txtout("Current working directory : <%s>\n\n",global_data->workpath);
	txtout("========== Start of control file =============================\n");
	writeout_ctlfile(stdout,global_data);
	txtout("========== End of control file ===============================\n\n");
#if DEBUG
	txtout("Before do_node_select\n");
#endif
	do_node_select(	request_no_nodes,
			&(global_data->no_nodes),
			nProcs,
			host_tid,
			global_data->total_no_atoms);
#if DEBUG
	txtout("Before ini_global_data_defaults\n");
#endif
	ini_global_data_defaults(global_data,my_force);

#if DEBUG
	txtout("Before read_flood_file\n");
#endif
	read_flood_file(global_data->filenames[FLOOD_FILE_INDEX],global_data);
	read_boundary_file(global_data->filenames[BOUNDARY_FILE_INDEX],global_data);
	read_sbound_file(global_data->filenames[SBOUND_FILE_INDEX],global_data);
		    	
#if DEBUG
	txtout("Before read_coord_file\n");
#endif
	/*****************************************************************/
	/**** we read the coord-file. We need it in memory, because we  **/
	/**** have to resort the atoms to the nodes. if memory		**/
	/**** gets a problem for very large systems, we have to do it 	**/
	/**** in a other way, but for now we do by brute force=get me   **/ 
	/**** the memory						**/	
	/*****************************************************************/
	res=read_coord_file(	global_data,
			        global_data->total_no_atoms,
				global_data->filenames[COORD_FILE_INDEX],
				global_data->atomdata);
	if (res==-1) {
		errout("Fatal error in read_coord_file <%s> !\n",
			global_data->filenames[COORD_FILE_INDEX]);
		pex(-1);
	}

#if DEBUG
	txtout("Before read_restart_file\n");
#endif
	/*****************************************************************/
	/**** if possible we use the restart-file 		 	**/
	/*****************************************************************/
	*file_count    = 0;
	res=read_restart_file(	global_data->restartfile,
				global_data->total_no_atoms,
				file_count,
				global_data->atomdata,
				global_data);
	if (res==-1) {
		errout("Fatal error in reading restart file !\n");
		pex(-1);
	} else
	if (res==1) {
		global_data->restarting = TRUE;
		txtout("Restarting at integration step %d.\n",
			global_data->cycles_passed);
	}

#if DEBUG
	txtout("Before set_used_atoms\n");
#endif
	/******************************************************************/
	/*** Set the array for the atoms to write out to EGO-output files */	
	/******************************************************************/
	if (set_used_atoms(	global_data->writeout_atom_selection,
				global_data->total_no_atoms,
				global_data->atomdata,
				global_data->writeout_atoms)<0) {
		errout("Error in control file (ctl.lis)\n");
		errout("The selection string specifying which atoms to write out\n");
		errout("during analysis is syntactically incorrect.\n");
		pex(-1);							
	}
	for (i=0,global_data->writeout_nr_atoms=0;i<global_data->total_no_atoms;i++) 
		if (global_data->writeout_atoms[i]!=0) global_data->writeout_nr_atoms++;
	txtout("%d atoms will be written to EGO output files.\n",
		global_data->writeout_nr_atoms);

#if DEBUG
	txtout("Before read_ctl_file_free_format_sec\n");
#endif
	/******************************************************************/
	/*** now we read the free format section of the control file	***/	
	/******************************************************************/
	read_ctl_file_free_format_sec(	global_data->total_no_atoms,
				global_data->ctlfile,
				global_data->atomdata,
				global_data->atom_sele_flag1,
				global_data,
				my_force);

#if DEBUG
	txtout("Before alloc_loader_buffers\n");
#endif
	alloc_loader_buffers(	global_data->total_no_atoms,
				global_data->no_nodes,
				my_force);

	/*********************************************************************/
	/***** here the clustering of structurla units is done ***************/
	/***** very important for FAMUSAMM ***********************************/
	/*********************************************************************/
	do_new_clustering(global_data->total_no_atoms,global_data,1,my_force);
		
#if DEBUG
	txtout("Before read_parameter_data\n");
#endif
	/*********************************************************************/
	/***************** read general data and parameters ******************/
	/***************** this sets many fields of global_data **************/
	/*********************************************************************/
	read_parameter_data(global_data,global_data->filenames);

#if DEBUG
	txtout("Before ini_my_forces\n");
#endif
	ini_my_forces(my_force,global_data);

#if DEBUG
	txtout("Before ini_imob_and_pull\n");
#endif
	ini_imob_and_pull(global_data,global_data->atomdata);
#if DEBUG
	txtout("Before ini_rotation\n");
#endif
	/*** now all is done used for rotation in EGO                  ***/
        ini_rotation(global_data,global_data->atomdata);
	/*** ready with preparation for rotation in EGO                ***/
	
#if defined(SEQ)
	/*** that is a little trick: in the sequential version my_data    ***/
	/*** contains all atom-data in my_data. In start_msg we have that ***/
	/*** data already, so we just have to set the my_data-pointer to  ***/
	/*** the start_msg structure 					  ***/
	set_my_data(&start_msg);
#else
	txtout("\nSending global data to all nodes.\n");
	PCI_send_msg_all_node(global_data);
#endif

	get_time_string(strtime);
	txtout("\nSwitched over to calculation phase at %s!\n\n",strtime);

} /*** end of the loader function ***/


/****************************************************************************/

static
int ini_global_data_defaults(	LOAD_MSG_ALL_NODE *global_data,
				STRUCT_FORCES *my_force)
{
	int i,level;

	/******************************************************/
	/*** first lets allocate some buffers *****************/
	/******************************************************/
	global_data->first_atom  = alloc_ints(global_data->no_nodes);
	
	set_allocinfo("atomdata");
	global_data->atomdata = alloc_atomdata(global_data->total_no_atoms);

	set_allocinfo("bonddata");
	bonddata = alloc_ints(global_data->total_no_atoms*MXNO_FIELDS_BOND);
			
	set_allocinfo("atommaps");
	global_data->writeout_atoms = alloc_ints(global_data->total_no_atoms);
	global_data->atom_sele_flag1 = alloc_ints(global_data->total_no_atoms);
	global_data->atom_sele_flag2 = alloc_ints(global_data->total_no_atoms);
	global_data->old2new_map    = alloc_ints(global_data->total_no_atoms);
	global_data->new2old_map    = alloc_ints(global_data->total_no_atoms);

	/*** fill allocated arrays to 0 ***/
	for (i=0;i<global_data->total_no_atoms;i++) global_data->writeout_atoms[i] = 0;
	memcpy(global_data->atom_sele_flag1,global_data->writeout_atoms,global_data->total_no_atoms*sizeof(int));
	memcpy(global_data->atom_sele_flag2,global_data->writeout_atoms,global_data->total_no_atoms*sizeof(int));
	memcpy(global_data->old2new_map   ,global_data->writeout_atoms,global_data->total_no_atoms*sizeof(int));
	memcpy(global_data->new2old_map   ,global_data->writeout_atoms,global_data->total_no_atoms*sizeof(int));

	/**** set default values ******/		
	global_data->total_mass        = 0.0;	
	global_data->selimob_mass[0]   = 0.0;
	global_data->selimob_mass[1]   = 0.0;
	global_data->selimob_mass[2]   = 0.0;
	global_data->restarting        = FALSE;
	global_data->cycles_passed     = 0;
	global_data->forceout_nr_atoms = 0;
	global_data->num_shake 	       = 0;	
	global_data->num_shake_h2o     = 0;	
	global_data->num_shake_protein = 0;	
	global_data->total_h2o_atoms     = 0;
	global_data->total_protein_atoms =0;

        global_data->f_elstat		= 332.0636; /*** this is the factor for epsilon = 1.0 ****/

	global_data->start_refC    = -1;
	global_data->refC[0]       = 0.0;
	global_data->refC[1]       = 0.0;
	global_data->refC[2]       = 0.0;
	global_data->refC[3]       = 0.0;
	global_data->refC[4]       = 0.0;
	global_data->refC[5]       = 0.0;
	global_data->refC[6]       = 0.0;
	global_data->refC[7]       = 0.0;
	global_data->refC[8]       = 0.0;

	global_data->pullspringpos[0] = 0.0;
	global_data->pullspringpos[1] = 0.0;
	global_data->pullspringpos[2] = 0.0;
	global_data->pullspringpos[3] = 0.0;
	global_data->pullspringpos[4] = 0.0;
	global_data->pullspringpos[5] = 0.0;
	global_data->pullspringpos[6] = 0.0;
	
	global_data->nr_of_rot_atoms = 0;
	global_data->rotation_energy = 0.0;

	global_data->forceout_start_step  = 0;
	global_data->forceout_stride_step = 1;
	global_data->idstr[0]             = 0;
	global_data->sbound_range         = 0;
        global_data->no_sbound_res_string = 1;

        global_data->schwing_mode	  = 0;
        global_data->schwing_offs         = 0.0;

	global_data->minikrit[0]   = -1.0;
	global_data->minikrit[1]   = -1.0;
	
        global_data->sbound_res_string  = (char (*) [LEN_RES_TYPE+1]) alloc_chars((LEN_RES_TYPE+1)*2);
	strcpy(global_data->sbound_res_string[0],"*");
	strcpy(global_data->sbound_res_string[1],"");

	global_data->extdyn_step[0]= -1; 
	global_data->extdyn_step[1]= -1; 
	global_data->extdyn_step[2]= -1; 
	global_data->extdynstr[0]  = 0;

	global_data->flood_energy_average  = 0.0;
	global_data->flood_energy_strength = 0.0;		
	global_data->flood_distance        = 0.0;
	global_data->flood_scale           = 2.0;
	global_data->newtransrotcor	   = 0;

	/**** initialization of QM-interface stuff ****/
	global_data->dipol[0] = 0.0;
	global_data->dipol[1] = 0.0;
	global_data->dipol[2] = 0.0;
	global_data->qmlinkmode    = QMLINKMODE_ACTIVEPOS;
	global_data->qmchargemode  = QMCHARGEMODE_FIXED;
	global_data->qmcharge      = -999.0;	/*** take CHARGE in template file ***/
	global_data->nr_qmatoms    = 0;
	global_data->qmstart_step  = -1; 
	global_data->qmstartstr[0] = 0;
	global_data->qmcontin_step = 1; 
	strcpy(global_data->qmcontinstr,"date");
	global_data->qmnoextelec   = 0;
	global_data->qmfixbox      = 0;
	global_data->qmatomforce   = QMATOMFORCE_QM;
	global_data->qmcutoff      = -1;
	global_data->qmenergyoffset= 0.0;
	global_data->qmexcl[0]	   = 4.0;	/*** distance ***/
	global_data->qmexcl[1]     = 2.0;	/*** width    ***/
	for (i=0;i<6;i++) {
		global_data->qmbox[i]   = 0.0; 
		global_data->qmboxext[i]= 2.0;
	}

	/**** initialization of clustering stuff ****/
	for (level=MAX_NO_LEVELS-1;level>=0;level--) {
		global_data->levelinfo[level].first_cluster=NULL;
		global_data->levelinfo[level].last_cluster =NULL;
	}

	/*************************************************************/
	/*** this is also a good place to initialize the my_force  ***/
	/*** structure                                             ***/
	/*************************************************************/
	my_force->pickbond.len = my_force->pickbond.cur_len = 0;
	my_force->pickangle.len= my_force->pickangle.cur_len= 0;
	my_force->pickelec.len = my_force->pickelec.cur_len = 0;
	my_force->pickvdw.len  = my_force->pickvdw.cur_len  = 0;
	my_force->pickdihe.len = my_force->pickdihe.cur_len = 0;
	my_force->pickimpr.len = my_force->pickimpr.cur_len = 0;

	my_force->qm.qmlink_len= 0;
	my_force->qm.asize = (double*) NULL;
	my_force->qm.qcore = (double*) NULL;
	return 0;
}

/****************************************************************************/

static
int	do_node_select(	int request_no_nodes,
			int *no_nodes,
			int nProcs,
			int host_tid,
			int Nr_of_Atoms)
{
#if defined(SEQ)
	*no_nodes=1;
#else

	/*** First we check, if the requested number of nodes is resonable ***/
	if (request_no_nodes<MINNO_NODES) {
		txtout("\nWARNING: the requested No of nodes (%d) is smaller than\n",
			request_no_nodes);
		txtout("         the minimum No of nodes (%d)!\n",MINNO_NODES);
		txtout("         request_no_nodes set equal to MINNO_NODES.\n\n");
		request_no_nodes = MINNO_NODES;
	}
	
	if (nProcs<=0) {
		txtout("ERROR: The available number of nodes (%d) is too small!\n",nProcs);
		pex(-1);
	}

	*no_nodes = PCI_set_no_nodes(request_no_nodes,nProcs);
	if (*no_nodes<MINNO_NODES) {
		errout("Fatal error: Number of selected nodes smaller than MINNO_NODES (%d)!\n",
			MINNO_NODES);
		pex(-1);
	}
	
	/* Now we check, if the problem is large enough for the nodes */
	if (Nr_of_Atoms / *no_nodes < MINNO_ATOMS_PER_NODE) {
		*no_nodes = Nr_of_Atoms / MINNO_ATOMS_PER_NODE;
		if (*no_nodes < MINNO_NODES) {
			txtout("\nERROR:the number of atoms (%d) is too small, leading to\n",
					Nr_of_Atoms);
			txtout("       a number of nodes (%d) below the minimum (%d)!\n\n",
					*no_nodes,MINNO_NODES);
			pex(-1);
		}
	}

	txtout("\nNo of nodes used in calculation = %d (out of %d available nodes)\n",
		*no_nodes,nProcs);

	PCI_master_setup (	*no_nodes,
				nProcs,
				host_tid);
#endif
	return 0;
}

/***********************************************************************/
	/* input (into arrays, reordering); */
	/* arrays are used to allow for output of bulk; */
	/* note that this requires as much memory per */
	/* node as needed later anyway */

static	FILE	*distrih;

static					
int distri_atomdata(	int nodenum,
			int first,
			int last,
			STRUCT_ATOMDATA *atom_data,
			LOAD_MSG_ONE_NODE *start_msg,
			int *new2old_map,
			LOAD_MSG_ALL_NODE *global_data)
{
	int    atom_number;
	int    Atomanzahl = last - first + 1;

	if (nodenum==0) {
		if ((distrih=fopen("atomdistri.out","w"))==NULL) {
			errout("Fatal error: Could not open atom-distribution log-file\n");
			pex(-1);
 		}
		fprintf(distrih,"Atom distribution to nodes at step %d!\n",global_data->cycles_passed);
		fprintf(distrih,"Atom numbers are EGO-coord.lis numbers starting with 0!\n");
	}

	fprintf(distrih,"Node %d loaded with %d atoms\n",nodenum+1,Atomanzahl);

	start_msg->no_atoms = Atomanzahl;
	for (atom_number=0; atom_number<Atomanzahl;atom_number++) {
		int newnum=atom_number+first;
		int i=new2old_map[newnum];	/* i is the old number */
		if (atom_data[i].atomnr!=i) {
			errout("\nFound wrong atom number in coordinate data!\n");
			errout("Should be: %d, but found %d\n",
				i,atom_data[i].atomnr);
			pex(-1);
		}
		fprintf(distrih,"%d\n",atom_data[i].atomnr);

		start_msg->x_new[atom_number] = atom_data[i].ko[0];
		start_msg->y_new[atom_number] = atom_data[i].ko[1];
		start_msg->z_new[atom_number] = atom_data[i].ko[2];

		/*** the old coordinates carry the velocities (temporarily) ***/
		/*** in case of no restart file, otherwise they carry the   ***/
		/*** real old coordinates	  			    ***/
		start_msg->x_old[atom_number] = atom_data[i].vel[0];
		start_msg->y_old[atom_number] = atom_data[i].vel[1];
		start_msg->z_old[atom_number] = atom_data[i].vel[2];
	    
	    		/* strictly, one more element of the Taylor */
	    		/* series should be included:               */
	    		/* 0.5 * dt^2 / m * F                	    */
	    		/* this is not possible, because the forces */
	    		/* are not known at this place;             */
	    		/* therefore, velocities are used instead   */
	    		/* of "old coordinates"                     */

		start_msg->atom_id[atom_number] = atom_data[i].atomnr;
		start_msg->typ[atom_number]     = atom_data[i].atomtype;
		start_msg->res_id[atom_number]  = atom_data[i].residnr;
		start_msg->q[atom_number]       = atom_data[i].charge;
		strcpy(start_msg->atom_string[atom_number],atom_data[i].atomstr);
		strcpy(start_msg->res_string[atom_number],atom_data[i].residstr);

			/* friction and harmonic coeff not read in here */
			/* see in distri_harmonic */
	}

	fflush(distrih);
	if (nodenum+1==global_data->no_nodes) {
		fclose(distrih);
	}
	txtout("Node number % 3d is being loaded with % 4d atom-coordinates now.\n",
		nodenum+1,Atomanzahl);
	return 0;
}

/*****************************************************************************************/

static					
int distri_harmonic(	int first,
			int last,
			STRUCT_ATOMDATA *atom_data,
			LOAD_MSG_ONE_NODE *start_msg,
			int *new2old_map)
{
	int	atom_number;
	int	Atomanzahl = last - first + 1;

	/**** refp is initialized with data comming from the coord-file ***/
	/**** this positions never changes				***/

	for (atom_number = 0; atom_number < Atomanzahl; atom_number++) {
		int newnum=atom_number+first;
		int i=new2old_map[newnum];	/* i is the old number */
		if (atom_data[i].atomnr!= i) {
			errout("\nFound wrong atom number in distri_harmonic !\n");
			errout("Should be: %d, but found %d\n",
				i,atom_data[i].atomnr);
			pex(-1);
		}

		start_msg->harmon_x[atom_number] = atom_data[i].refp[0];
		start_msg->harmon_y[atom_number] = atom_data[i].refp[1];
		start_msg->harmon_z[atom_number] = atom_data[i].refp[2];
		start_msg->beta[atom_number]	 = atom_data[i].beta;
		start_msg->harmon_k[atom_number] = atom_data[i].harmon;
	}

	return 0;
}

/****************************************************************************/

/* format of SHAKE list: <hydrogen> <antezedent> <hydrogen> <antezedent> ***/
/* shake_number is always even and gives the number of atoms 		 ***/
/* that are local in the shake_list 					 ***/
static
int    is_local_shake_hyd(int atom,int shake_number,int *shake_list)
{
  	int    i;
	if (shake_number==0) return FALSE;
	for (i=0;i<shake_number;i+=2) if (shake_list[i]==atom) return TRUE;
	return FALSE;
}

/*** add to the end of list ***/
static
void   enter_ext(	int element,
			LOAD_MSG_ONE_NODE * start_msg,
			int *mxno_extern_bonds)
{
	start_msg->ext_bond_list[start_msg->ext_bond_ptr] = element;
	start_msg->ext_bond_ptr++;
	if (start_msg->ext_bond_ptr >= *mxno_extern_bonds)
		realloc_list (&(start_msg->ext_bond_list), mxno_extern_bonds);
}

static
int distri_bonds(int first,
		int last,
		int *bondlist,
		LOAD_MSG_ONE_NODE *start_msg,
		int *mxno_extern_bonds,
		int *mxno_intern_bonds)
{
	int	i;
	int	offset;
	int	count_extern=0;
	int	shake_number = start_msg->shake_number;
	int	*shake_list;

	shake_list = start_msg->shake_list;

	set_allocinfo("distri_bonds");
	
	start_msg->int_bond_ptr = 0;	
	start_msg->ext_bond_ptr = 0;
	for (i=first;i<=last;i++) {	/* as often as atoms per node */
	      
		int    aktuell;		/* stores the BONDs. */
	
		offset  = i * MXNO_FIELDS_BOND;
		aktuell = bondlist[offset++];

	    while (aktuell != END_OF_LIST) {

		  /* we use the shake-list of that node, so make shure we have it here; 
		   * if there is no shake, the shake_number will be = 0, such
		   * that the function is_local_shake_hyd always returns FALSE! */

		if (is_local_shake_hyd(aktuell,shake_number,shake_list) ||
		    is_local_shake_hyd(i      ,shake_number,shake_list)) {

			/*** do nothing here ***/
			/*** no bond interaction for shaken hydrogens ! ***/

		} else
		if (inrange(aktuell,first,last) && 	/* in local region */
		    (aktuell > i) ) {			/* calculate bond only once */
		     				
			start_msg->int_bond_list[start_msg->int_bond_ptr] = (int)aktuell;
			start_msg->int_bond_ptr++;
			if (start_msg->int_bond_ptr>=*mxno_intern_bonds) {
				realloc_list(&(start_msg->int_bond_list),mxno_intern_bonds);
			}
		} else
		if (!inrange(aktuell,first,last)) {
				/* put in ext_bond_list */
				/* format of the ext_bond_list: <ext_atom> <int_1>
				int_2> ... <END_OF_LIST> <ext_atom> <int_1> ...;
				the <ext_atom>s are sorted in ascending order,
				whereas the <int_1> are not ordered (can be
				accessed at random, though); <ext_atom> are checked
				for the extern region *******/

			int    ptr=0;	/* points to ext.basisatom or to a free */
			count_extern++;

			/** find place for insert of new external list-part ***/
			/** is there already an entry for <aktuell>=<ext.atom>?
			   and if so, where is it *****/

			while ( (ptr < start_msg->ext_bond_ptr) &&
				(aktuell > start_msg->ext_bond_list[ptr])) {
		      				/* move ptr to the next ext_atom */
				while (start_msg->ext_bond_list[ptr] != END_OF_LIST) ptr++;
						/* go to next internal atom in sublist */
				ptr++;		/* go to next external atom (or end of list) */
			}

			if (ptr >= start_msg->ext_bond_ptr) {
				/* there is no entry with a number higher as aktuell yet */
				/* so we add it to the end of the list *****/
				enter_ext (aktuell,	start_msg, mxno_extern_bonds);
				enter_ext (i,		start_msg, mxno_extern_bonds);
				enter_ext (END_OF_LIST,	start_msg, mxno_extern_bonds);
			} else
			if (aktuell == start_msg->ext_bond_list[ptr]) {
				/* there is already a element.insert in sub-list */
				/* shift rest of list by 1, starting at ptr inclusively; */
				/* direction: upwards. loop goes backwards! */
				int j;
				ptr++;
				for (j = 0; j < start_msg->ext_bond_ptr - ptr; j++) {
					int    ziel = start_msg->ext_bond_ptr - j;
					start_msg->ext_bond_list[ziel]=
						start_msg->ext_bond_list[ziel-1];
				}
				start_msg->ext_bond_list[ptr] = i;
				start_msg->ext_bond_ptr++;
				if (start_msg->ext_bond_ptr >= *mxno_extern_bonds)
					realloc_list(&(start_msg->ext_bond_list),mxno_extern_bonds);
			} else {
				/* there is no entry yet, but we have higher
		       		   external atom numbers as aktuell */
				/* create and insert new sublist and shift by 3 upwards */
				int    j;
				int    ziel;
		      
				if ((start_msg->ext_bond_ptr + 3) >= *mxno_extern_bonds) 
					realloc_list(&(start_msg->ext_bond_list),mxno_extern_bonds);	
			
				for (j = 0; j < (start_msg->ext_bond_ptr - ptr); j++) {
					ziel = start_msg->ext_bond_ptr - j;
					start_msg->ext_bond_list[ziel+2] = 
						start_msg->ext_bond_list[ziel-1];
				}
				start_msg->ext_bond_list[ptr++] = aktuell;
				start_msg->ext_bond_list[ptr++] = i;
				start_msg->ext_bond_list[ptr]   = END_OF_LIST;
				start_msg->ext_bond_ptr         = start_msg->ext_bond_ptr+3;
		    	}
		} 	/* end of else: put it in external bond list */

		  aktuell=bondlist[offset++];
	   } 	/* end while for each bond of atom i */

		/* always add an END_OF_LIST in the internal list! */
		start_msg->int_bond_list[start_msg->int_bond_ptr] =(int)END_OF_LIST;
		start_msg->int_bond_ptr++;
		if (start_msg->int_bond_ptr >= *mxno_intern_bonds)
			realloc_list (&(start_msg->int_bond_list), mxno_intern_bonds);

	} /* end of for loop over all local atoms */
	set_allocinfo("XXX");
	return count_extern;
}	    

/**********************************************************************/
static
int	distri_shake(	int first,int last,
			int *shakedata,
			LOAD_MSG_ONE_NODE *start_msg,
			int *mxno_shake,
			LOAD_MSG_ALL_NODE *global_data)
			
{
	int	hd,ad,offset,deleted,flag1,flag2;
	
	set_allocinfo("distri_shake");
	if (shakedata==NULL){
		errout("Fatal error #1 in distri_shake. Programm error !\n");
		pex(-1);
	}
	
	deleted = 0;
	start_msg->shake_number = 0;	/* points to the next empty position */
					/* and is the counter for the shaked */
					/* atoms on this node		     */
			
	for (offset=0;offset<global_data->num_shake;offset+=2) {
		hd = shakedata[offset  ];
		ad = shakedata[offset+1];
	
		if (	(!inrange(hd,first,last)) &&  /*** not for this node ***/
			(!inrange(ad,first,last)) ) continue;	

		if (	(!inrange(hd,first,last)) &&  /*** error-check ***/
			( inrange(ad,first,last)) ) goto not_on_samenode;

		if (	( inrange(hd,first,last)) &&  /*** error-check ***/
			(!inrange(ad,first,last)) ) goto not_on_samenode;

		/*** check if some atoms are not shaked ***/
		/*** due to some SHAKEOFF selections	***/
		flag1=global_data->atom_sele_flag1[global_data->new2old_map[hd]];
		flag2=global_data->atom_sele_flag1[global_data->new2old_map[ad]];
		if (	(flag1&FLAG_SHAKEOFF) ||
	 		(flag2&FLAG_SHAKEOFF) ) {
			deleted++;
			continue;	
		}

		start_msg->shake_list[start_msg->shake_number++] = hd;
		if (start_msg->shake_number >= *mxno_shake)
			realloc_list(&(start_msg->shake_list),mxno_shake);

		start_msg->shake_list[start_msg->shake_number++] = ad;
		if (start_msg->shake_number >= *mxno_shake)
			realloc_list(&(start_msg->shake_list),mxno_shake);
	}
	set_allocinfo("XXX");
	return deleted;

not_on_samenode:
	errout("For SHAKE hydrogens and antiacent have to be on the same node !\n");
	errout("But I found a pair which is NOT on the same node !\n");
	errout("That should not be, because of the unit-clustering, this\n");
	errout("should never happen --> Check code of clustering !\n");
	pex(-1);
	return -1;
}

static
int	distri_excl(	int first,int last,
			int *excldata,
			LOAD_MSG_ONE_NODE *start_msg,
			int *mxno_bond_excl,
			int maxexcl,
			LOAD_MSG_ALL_NODE *global_data)
			
{
	int	atom;
	set_allocinfo("distri_excl");
	start_msg->ex_ptr = 0;	/* points to the next empty position */
			
	for (atom = first; atom <= last;atom++) {
		int	aktuell;
		int	offset;
		offset = atom*maxexcl;
		
		aktuell=excldata[offset++];

		while (aktuell!=END_OF_LIST) {

			if (global_data->nr_qmatoms>0) {
				/*** For the QM-Interface: here we catch ***/
				/*** all normal atoms which are          ***/
				/*** excluded by at least one QM-atom	 ***/
			
				int atom1 = global_data->new2old_map[atom   ];
				int atom2 = global_data->new2old_map[aktuell];
				if (	 (global_data->atom_sele_flag1[atom1]&FLAG_QMATOM)   &&
				       ( (!(global_data->atom_sele_flag1[atom2]&FLAG_QMATOM))|| 
				         (global_data->atom_sele_flag1[atom2]&FLAG_QMLINK)     ) ) {
			
					set_qmexclusionatom(	atom1,atom2,
								global_data,
								1);
				}
			}
			
			start_msg->ex_list[start_msg->ex_ptr] = aktuell;
			start_msg->ex_ptr++;
			if (start_msg->ex_ptr >= *mxno_bond_excl)
				realloc_list(&(start_msg->ex_list),mxno_bond_excl);

			aktuell=excldata[offset++];
		}

		start_msg->ex_list[start_msg->ex_ptr] = aktuell;   /* aktuell is END_OF_LIST */
		start_msg->ex_ptr++;
		if (start_msg->ex_ptr >= *mxno_bond_excl)
			realloc_list(&(start_msg->ex_list),mxno_bond_excl);
	}
	set_allocinfo("XXX");
	return 0;
}

/********************************************************************************/

static
int	distri_excl14(	int first,
			int last,
			int *excl14data,
			LOAD_MSG_ONE_NODE *start_msg,
			int *mxno_bond_ex14,
			int maxexcl,
			LOAD_MSG_ALL_NODE *global_data)
			
{
	int	atom;

	set_allocinfo("distri_excl14");
	start_msg->ex14_ptr = 0;	/* points to the next empty position */
			
	for (atom = first; atom <= last;atom++) {
		int	aktuell;
		int	offset;
		offset=atom*maxexcl;
	
		aktuell=excl14data[offset++];

		while (aktuell != END_OF_LIST) {

			if (global_data->nr_qmatoms>0) {
				/*** For the QM-Interface: here we catch ***/
				/*** all normal atoms which are          ***/
				/*** excluded by at least one QM-atom	 ***/
			
				int atom1 = global_data->new2old_map[atom   ];
				int atom2 = global_data->new2old_map[aktuell];
				if (	 (global_data->atom_sele_flag1[atom1]&FLAG_QMATOM)   &&
				       ( (!(global_data->atom_sele_flag1[atom2]&FLAG_QMATOM))|| 
				         (global_data->atom_sele_flag1[atom2]&FLAG_QMLINK)     ) ) {
			
					set_qmexclusionatom(	atom1,atom2,
								global_data,
								2);
				}
			}

			start_msg->ex14_list[start_msg->ex14_ptr] = aktuell;
			start_msg->ex14_ptr++;
			if (start_msg->ex14_ptr >= *mxno_bond_ex14)
				realloc_list (&(start_msg->ex14_list),mxno_bond_ex14);

			aktuell=excl14data[offset++];
		}

		start_msg->ex14_list[start_msg->ex14_ptr] = aktuell;   /* aktuell is END_OF_LIST */
		start_msg->ex14_ptr++;
		if (start_msg->ex14_ptr >= *mxno_bond_ex14)
			realloc_list(&(start_msg->ex14_list),mxno_bond_ex14);
	}
	set_allocinfo("XXX");
	return 0;
}

/****************************************************************************/

static
int	distri_angles(	int size_anglelist,
			int first,
			int last,
			int *angledata,
			LOAD_MSG_ONE_NODE *start_msg,
			int *mxno_angles,
			int *mxno_extern_angles)
{
	int	a1,a2,a3,ip;
	int	i,c=0;
	int	count_extern=0;
	set_allocinfo("distri_angles");
	start_msg->int_angle_ptr = 0;	/* points to the next free position */
	start_msg->ext_angle_ptr = 0;	/* points to the next free position */

	for (i=0,c=0;i<size_anglelist;i++,c+=4) {  /* loop over total angle-list */
		a1=angledata[c  ];
		a2=angledata[c+1];
		a3=angledata[c+2];
		ip=angledata[c+3];

		if (	(!inrange(a2,first,last)) &&
			(!inrange(a1,first,last)) &&
			(!inrange(a3,first,last)) ) continue;

		if (	inrange(a2,first,last) &&
			inrange(a1,first,last) && 
			inrange(a3,first,last) ) {
			/**************************************/
			/*** all atoms are on the same node ***/
			/**************************************/	
			start_msg->int_angle_Nr1[start_msg->int_angle_ptr] = a1 - first;
			start_msg->int_angle_Nr2[start_msg->int_angle_ptr] = a2 - first;
			start_msg->int_angle_Nr3[start_msg->int_angle_ptr] = a3 - first;
			start_msg->int_angle_index[start_msg->int_angle_ptr] = ip;
			start_msg->int_angle_ptr++;
			if (start_msg->int_angle_ptr >= *mxno_angles) {
				int    dummy = *mxno_angles;
				realloc_list(&(start_msg->int_angle_Nr1), &dummy);
				dummy = *mxno_angles;
				realloc_list(&(start_msg->int_angle_Nr2), &dummy);
				dummy = *mxno_angles;
				realloc_list(&(start_msg->int_angle_Nr3), &dummy);
				realloc_list(&(start_msg->int_angle_index), mxno_angles);
			}
	      } else {					/* into extern list */
			/************************************************************************/
			/*** at least one atom is not same node atoms are on the same node    ***/
			/************************************************************************/
			count_extern++;
			start_msg->ext_angle_Nr1[start_msg->ext_angle_ptr] = a1;
			start_msg->ext_angle_Nr2[start_msg->ext_angle_ptr] = a2;
			start_msg->ext_angle_Nr3[start_msg->ext_angle_ptr] = a3;
			start_msg->ext_angle_index[start_msg->ext_angle_ptr] = ip;
			start_msg->ext_angle_ptr = start_msg->ext_angle_ptr + 1;
			if (start_msg->ext_angle_ptr >= *mxno_extern_angles) {
				int    dummy;

				dummy = *mxno_extern_angles;
				realloc_list (&(start_msg->ext_angle_Nr1),&dummy);
				dummy = *mxno_extern_angles;
				realloc_list (&(start_msg->ext_angle_Nr2),&dummy);
				dummy = *mxno_extern_angles;
				realloc_list (&(start_msg->ext_angle_Nr3),&dummy);
				realloc_list (&(start_msg->ext_angle_index),mxno_extern_angles);
			}
	      }
	} /* end loop over total angle-list */
	set_allocinfo("XXX");
	return count_extern;
}

/*****************************************************************************************/
	
static
int	distri_dihe(	int size_dihelist,
			int first,
			int last,
			int *dihedata,
			LOAD_MSG_ONE_NODE *start_msg,
			int *mxno_dihed,
			int *mxno_extern_dihed)
{
	int	a1,a2,a3,a4,ip;
	int	i,c;
	int	count_extern=0;
	set_allocinfo("distri_dihe");
	start_msg->int_dihe_ptr = 0;	/* points to next free position */
	start_msg->ext_dihe_ptr = 0;	/* points to next free position */

	for (i=0,c=0;i<size_dihelist;i++,c+=5) {  /* loop over total dihe-list */
		a1=dihedata[c  ];
		a2=dihedata[c+1];
		a3=dihedata[c+2];
		a4=dihedata[c+3];
		ip=dihedata[c+4];

		if (	(!inrange(a1,first,last)) &&
			(!inrange(a2,first,last)) &&
			(!inrange(a3,first,last)) &&
			(!inrange(a4,first,last)) ) continue;

		if (	inrange(a1,first,last) &&
			inrange(a2,first,last) &&
			inrange(a3,first,last) && 
			inrange(a4,first,last) ) { /* in intern list */
			/**************************************/
			/*** all atoms are on the same node ***/
			/**************************************/	
			start_msg->int_dihe_Nr1[start_msg->int_dihe_ptr] = a1 - first;
			start_msg->int_dihe_Nr2[start_msg->int_dihe_ptr] = a2 - first;
			start_msg->int_dihe_Nr3[start_msg->int_dihe_ptr] = a3 - first;
			start_msg->int_dihe_Nr4[start_msg->int_dihe_ptr] = a4 - first;
			start_msg->int_dihe_index[start_msg->int_dihe_ptr] = ip;
			start_msg->int_dihe_ptr++;
			if (start_msg->int_dihe_ptr >= *mxno_dihed) {
				int    dummy = *mxno_dihed;
				realloc_list (&(start_msg->int_dihe_Nr1), &dummy);
				dummy = *mxno_dihed;
				realloc_list (&(start_msg->int_dihe_Nr2), &dummy);
				dummy = *mxno_dihed;
				realloc_list (&(start_msg->int_dihe_Nr3), &dummy);
				dummy = *mxno_dihed;
				realloc_list (&(start_msg->int_dihe_Nr4), &dummy);
				realloc_list (&(start_msg->int_dihe_index), mxno_dihed);
			}
	      } else { /* in extern list */
			/************************************************************************/
			/*** at least one atom is not same node atoms are on the same node    ***/
			/************************************************************************/
			count_extern++;
			start_msg->ext_dihe_Nr1[start_msg->ext_dihe_ptr] = a1;
			start_msg->ext_dihe_Nr2[start_msg->ext_dihe_ptr] = a2;
			start_msg->ext_dihe_Nr3[start_msg->ext_dihe_ptr] = a3;
			start_msg->ext_dihe_Nr4[start_msg->ext_dihe_ptr] = a4;
			start_msg->ext_dihe_index[start_msg->ext_dihe_ptr]= ip;
			start_msg->ext_dihe_ptr++;
			if (start_msg->int_dihe_ptr >= *mxno_extern_dihed) {
	  			int    dummy = *mxno_extern_dihed;
				realloc_list (&(start_msg->ext_dihe_Nr1), &dummy);
				dummy = *mxno_extern_dihed;
				realloc_list (&(start_msg->ext_dihe_Nr2), &dummy);
				dummy = *mxno_extern_dihed;
				realloc_list (&(start_msg->ext_dihe_Nr3), &dummy);
				dummy = *mxno_extern_dihed;
				realloc_list (&(start_msg->ext_dihe_Nr4), &dummy);
				realloc_list (&(start_msg->ext_dihe_index), mxno_extern_dihed);
			}
	      }
	} /* end loop over all dihedrals */
	set_allocinfo("XXX");
	return count_extern;
}

/*********************************************************************************/

static
int	distri_impr(	int size_imprlist,
			int first,
			int last,
			int *imprdata,
			LOAD_MSG_ONE_NODE *start_msg,
			int *mxno_improp,
			int *mxno_extern_improp)
{
	int	a1,a2,a3,a4,ip;
	int	i,c;
	int	count_extern=0;
	set_allocinfo("distri_impr");
	start_msg->int_impr_ptr = 0;	
	start_msg->ext_impr_ptr = 0;	

	for (i=0,c=0;i<size_imprlist;i++,c+=5) {  /* loop over total impr-list */
		a1=imprdata[c  ];
		a2=imprdata[c+1];
		a3=imprdata[c+2];
		a4=imprdata[c+3];
		ip=imprdata[c+4];

		if (	(!inrange(a1,first,last)) &&
			(!inrange(a2,first,last)) &&
			(!inrange(a3,first,last)) &&
			(!inrange(a4,first,last))  ) continue;

		if (	inrange(a1,first,last) &&
			inrange(a2,first,last) &&
			inrange(a3,first,last) && 
			inrange(a4,first,last) ) { /* in intern list */
			/**************************************/
			/*** all atoms are on the same node ***/
			/**************************************/	
			start_msg->int_impr_Nr1[start_msg->int_impr_ptr] = a1 - first;
			start_msg->int_impr_Nr2[start_msg->int_impr_ptr] = a2 - first;
			start_msg->int_impr_Nr3[start_msg->int_impr_ptr] = a3 - first;
			start_msg->int_impr_Nr4[start_msg->int_impr_ptr] = a4 - first;
			start_msg->int_impr_index[start_msg->int_impr_ptr] = ip;
			start_msg->int_impr_ptr++;
			if (start_msg->int_impr_ptr >= *mxno_improp) {
				int    dummy = *mxno_improp;
				realloc_list (&(start_msg->int_impr_Nr1), &dummy);
				dummy = *mxno_improp;
				realloc_list (&(start_msg->int_impr_Nr2), &dummy);
				dummy = *mxno_improp;
				realloc_list (&(start_msg->int_impr_Nr3), &dummy);
				dummy = *mxno_improp;
				realloc_list (&(start_msg->int_impr_Nr4), &dummy);
				realloc_list (&(start_msg->int_impr_index), mxno_improp);
			}
		} else { 				/* in extern list */
			/************************************************************************/
			/*** at least one atom is not same node atoms are on the same node    ***/
			/************************************************************************/

			count_extern++;
			start_msg->ext_impr_Nr1[start_msg->ext_impr_ptr] = a1;
			start_msg->ext_impr_Nr2[start_msg->ext_impr_ptr] = a2;
			start_msg->ext_impr_Nr3[start_msg->ext_impr_ptr] = a3;
			start_msg->ext_impr_Nr4[start_msg->ext_impr_ptr] = a4;
			start_msg->ext_impr_index[start_msg->ext_impr_ptr] = ip;
			start_msg->ext_impr_ptr++;
			if (start_msg->int_impr_ptr >= *mxno_extern_improp) {
				int    dummy= *mxno_extern_improp;
				realloc_list (&(start_msg->ext_impr_Nr1), &dummy);
				dummy = *mxno_extern_improp;
				realloc_list (&(start_msg->ext_impr_Nr2), &dummy);
				dummy = *mxno_extern_improp;
				realloc_list (&(start_msg->ext_impr_Nr3), &dummy);
				dummy = *mxno_extern_improp;
				realloc_list (&(start_msg->ext_impr_Nr4), &dummy);
				realloc_list (&(start_msg->ext_impr_index), mxno_extern_improp);
			}
		}
	} /* end loop over all impropers */
	set_allocinfo("XXX");
	return count_extern;
}

/****************************************************************************************/
/***** here we also have to correct the child-position-pointers of the unit-level *******/

static
int	distri_tree(	int nodenum,int no_nodes,
			int first,
			struct_treeinfo *treeinfo,
			struct_levelinfo *levelinfo,
			struct_cluster_reduced *cluster[],
			LOAD_MSG_ONE_NODE *start_msg)
{
	int	level,toplevel=treeinfo->levels-1,i,cl,j;
	int	left_branch =  nodenum   * treeinfo->branches/no_nodes;
	int	right_branch= (nodenum+1)* treeinfo->branches/no_nodes - 1;
	set_allocinfo("distri_tree");
	start_msg->levels=treeinfo->levels;

	for (level=0; level<= toplevel; level++) {
		
		int left_cl = levelinfo[level].first_cluster[left_branch];
		int right_cl= levelinfo[level].last_cluster[right_branch];		
		int nr_cl=right_cl-left_cl+1;
		
		if (	(left_cl<0) ||
			(right_cl>levelinfo[level].nr_of_clusters) ||
			(nr_cl<1) ) {
			errout("Fatal error: Illegal tree-Structure in distri_tree.\n");
			pex(-1);
		}

		if (alloc_per_level[level]<=0) {
			alloc_per_level[level]   = (int) (nr_cl*1.3);
			start_msg->cluster[level]= alloc_cluster(alloc_per_level[level]);
		} else		
		if (nr_cl>alloc_per_level[level]) {
			start_msg->cluster[level]=(struct_cluster*)
						merealloc(start_msg->cluster[level],
						sizeof(struct_cluster)*nr_cl,
						sizeof(struct_cluster)*alloc_per_level[level]);
			alloc_per_level[level] = nr_cl;
		}

		/**** correct unit-level ******/
		if (level==0) {
			int cur_child_pos = first;
			for (i=0;i<nr_cl;i++) {
				cluster[0][left_cl+i].child_pos  = cur_child_pos;
				cur_child_pos += cluster[0][left_cl+i].nr_childs;
			}
		}
		start_msg->cl_per_level[level]     = nr_cl;
		start_msg->cl_more_per_level[level]= (int)(nr_cl*1.5);

		/*** the next loop is neccessary because start_msg->cluster has ***/
		/*** a different structure definition than cluster[level]	***/
		/*** Thus no memcpy can be used here				***/
		for (cl=left_cl,j=0;cl<left_cl+nr_cl;cl++,j++) {
			for (i=0;i<6;i++)
			start_msg->cluster[level][j].bbox[i] = cluster[level][cl].bbox[i]; 

			start_msg->cluster[level][j].moment[0] = cluster[level][cl].moment[0];
 			for (i=1;i<9;i++)
			start_msg->cluster[level][j].moment[i] = 0.0;

			for (i=0;i<10;i++)
			start_msg->cluster[level][j].local[i] = 0.0;

			for (i=0;i<4;i++)
			start_msg->cluster[level][j].cen_geom[i] = cluster[level][cl].cen_geom[i];
 
			for (i=0;i<3;i++)
			start_msg->cluster[level][j].cen_refp[i] = 0.0;

			start_msg->cluster[level][j].child_pos = cluster[level][cl].child_pos;
 			start_msg->cluster[level][j].status    = cluster[level][cl].status;
			start_msg->cluster[level][j].nr_childs = cluster[level][cl].nr_childs;
			start_msg->cluster[level][j].typ       = cluster[level][cl].typ;
		}
	}
	set_allocinfo("XXX");
	return 0;
}


/***************************************************************************
format of SHAKE list: <hydrogen> <antezedent> <hydrogen> <antezedent>
the hydrogens are in ascending order
the ptr points to the next free position in the list and its
value is equal to the No of elements in the list
this is a global shake list used in the loader only (length
num_shake); not to be confused with the single-node shake_list of
global_msg, which has shake_number atoms (shake_number/2 hydrogens)
****************************************************************************/

static	int alloc_shakelist=0;

static
int	read_shake_file(	LOAD_MSG_ALL_NODE *global_data,
			        int *shake_num,
				int *shake_num_h2o,
				int *shake_num_protein,
				char *shakefilen,
				int **shakedata,
				int *old2new_map)
{
	int	mxlen_shake_list=0;
	FILE	*fdshake;
	int	old_egonum;
	int	shakecount=-1;
	int     shakecount_h2o=-1;
	int     shakecount_protein=-1;

 	if ((fdshake = fopen(shakefilen,"r")) == NULL) {
		errout("Fatal error: Cannot open SHAKE file <%s>!\n",shakefilen);
		pex(-1);
	}

	/*** parse header ***/
	skip_line(fdshake, 2);
	fscanf(fdshake,"%d",&mxlen_shake_list);

	/*** alloc space for shake-list if not yet allocated ***/
	if (*shakedata==NULL) {
		set_allocinfo("shakedata");
		alloc_shakelist = 2*mxlen_shake_list+10;
		*shakedata = alloc_ints(alloc_shakelist);
		if (*shakedata==NULL) {
			errout("Fatal error: No space for shake-data in loader.c !\n");
			pex(-1);
		}
	}

	while (!feof(fdshake)) {
		old_egonum=-1;
		fscanf(fdshake,"%d",&old_egonum);

		if (feof(fdshake))	break;
		if (old_egonum<0)	break;

		shakecount++;
		if (strcmp(global_data->atomdata[old_egonum].residstr,"TIP3")==0) {
		  shakecount_h2o++;
		} else {
		  shakecount_protein++;
		}

		if (shakecount >= alloc_shakelist) {
			errout("Fatal error: programm error in read_shake_file %d %d\n",
				shakecount,alloc_shakelist);
			pex(-1);
	  	}

		(*shakedata)[shakecount] = old2new_map[old_egonum];
	}

	shakecount++;

	if ((shakecount%2)!=0) {
		errout("Shake-file is not valid ! Check for even number (%d) !\n",shakecount);
		fclose(fdshake);
		pex(-1);
	}

	txtout("Found %4d hydrogen-atoms in shake-list.\n",shakecount/2);
	fclose(fdshake);

	*shake_num = shakecount;
	*shake_num_h2o     = shakecount_h2o;
	*shake_num_protein = shakecount_protein;

	set_allocinfo("XXX");

	return 0;
}

static int
read_parameter_data(	LOAD_MSG_ALL_NODE *global_data,
			char filenames[MAX_FILES][MAX_STRING_SIZE])
{
	char	sdum[MAX_STRING_SIZE];
	int	nr_of_atom_typs;
	FILE	*fdata;
	int	i,j,dummy;
	int	index,nr_of_param,withub=0;
	
	/******************************************************************/
	/***** van der Waals parameters and masses for all atom types. ****/
	/***** the masses for the atom types are the diagonal elements ****/
	/******************************************************************/
	txtout("Reading van der Waals parameters.....: %s\n",filenames[VDWPARA_FILE_INDEX]);
	if ((fdata = fopen(filenames[VDWPARA_FILE_INDEX],"r")) == NULL) {
		errout("Fatal error: Cannot open van der Waals parameter file!\n");
		pex(-1);
	}

	skip_line(fdata,1);
	fscanf(fdata,"%s%s%s%s%s%d",sdum,sdum,sdum,sdum,sdum,&nr_of_atom_typs);
	global_data->nr_of_atom_typs = nr_of_atom_typs;
	if (nr_of_atom_typs+1>=MXNO_ATOM_TYPES) {
		errout("MXNO_ATOM_TYPES %d to small. You need at least %d.\n",
			MXNO_ATOM_TYPES,nr_of_atom_typs+1);
		pex(-1);	
	}
	for (i=0;i<MXNO_ATOM_TYPES;i++){
		for (j=0;j<MXNO_ATOM_TYPES;j++) {
			global_data->AB[i][j] = 0.0;
		}
	}

	for (i=0;i<=nr_of_atom_typs;i++) {
		for (j=0;j<=nr_of_atom_typs;j++) {
			if (feof(fdata)) {
				errout("Unexpected end of file while reading van der Waals data !\n");
				pex(-1);
			}
			global_data->AB[i][j] = read_double(fdata);	/* read one line */
		}
		
	}
	fclose(fdata);
	/********* end van der Waals parameters ***************************/
	
	/*** now we can calculate the total mass of the system ***/
	global_data->total_mass = 0.0;
	for (i=0;i<global_data->total_no_atoms;i++) {
		int	ty = global_data->atomdata[i].atomtype;
		double	m  = global_data->AB[ty][ty];
		global_data->total_mass += m;
	}

	/******************************************************************/
	/********** special 1-4 vdW parameter *****************************/
	/******************************************************************/
	txtout("Reading 1-4 van der Waals parameters.: %s\n",filenames[VDW14PARA_FILE_INDEX]);
	if ((fdata=fopen(filenames[VDW14PARA_FILE_INDEX], "r")) == NULL) {
		errout("Fatal error: Cannot open 1-4 van der Waals parameter file !\n");
		pex(-1);
	}

	skip_line (fdata, 1);
	fscanf(fdata,"%s%s%s%s%s%d",sdum,sdum,sdum,sdum,sdum,&dummy);

	if (dummy!=nr_of_atom_typs) {
		errout("Fatal error: Error reading atom No in 1-4 van der Waals parameter file\n");
		pex(-1);
	}
	for (i=0;i<MXNO_ATOM_TYPES;i++){
		for (j=0;j<MXNO_ATOM_TYPES;j++) {
			global_data->AB14[i][j] = 0.0;
		}
	}
	for (i=0;i<=nr_of_atom_typs;i++) {
		for (j=0;j<=nr_of_atom_typs;j++) {
			if (feof(fdata)) {
				errout("Fatal error: Unexpected EOF while reading 1-4 van der Waals data!\n");
				pex(-1);
			}
			global_data->AB14[i][j] = read_double(fdata); /* read one line */
		}
	}
	fclose (fdata);
	/********** end special 1-4 vdW ***********************************/

	/******************************************************************/
 	/***************** BONDed parameter for all nodes *****************/
	/******************************************************************/
	txtout("Reading bond parameters..............: %s\n",filenames[BONDPARA_FILE_INDEX]);
	if ((fdata = fopen(filenames[BONDPARA_FILE_INDEX], "r")) == NULL) {
		errout("Fatal error: Cannot open BOND parameter file!\n");
		pex(-1);
	}

	skip_line (fdata, 1);
	fscanf(fdata,"%s%s%s%s%s%d",sdum,sdum,sdum,sdum,sdum,&dummy);
	if (dummy != nr_of_atom_typs) {
		errout("Fatal error: Error reading atom No in the bond parameter file\n");
		pex(-1);
	} else {
		for (i=0;i<MXNO_ATOM_TYPES;i++){
			for (j=0;j<MXNO_ATOM_TYPES;j++) {
				global_data->k_bond[i][j] = 0.0;
			}
		}
		for (i=0;i<=nr_of_atom_typs;i++){
			for (j=0;j<=nr_of_atom_typs;j++) {
				if (feof(fdata)) {
					errout("Unexpected eof while reading BOND data!\n");
					pex(-1);
				}
				global_data->k_bond[i][j] = read_double(fdata);
			}
		}
 	}
 	fclose(fdata);
 	/**************** end BONDed parameter for all nodes **************/

	/******************************************************************/
	/************************** ANGLE parameter for all nodes *********/
	/******************************************************************/
	txtout("Reading angle parameters.............: %s\n",filenames[ANGLEPARA_FILE_INDEX]);
	if ((fdata = fopen(filenames[ANGLEPARA_FILE_INDEX],"r")) == NULL) {
		errout("Fatal error: Cannot open ANGLE parameter file!\n");
		pex(-1);
 	}

	skip_line(fdata,1);	/* skip first line */
	fscanf(fdata,"%s%s%s%s",sdum,sdum,sdum,sdum);	/* skip 4 words */
	nr_of_param = read_int_skip(fdata);
	skip_line(fdata,1);	/* skip 2 more text lines */
	read_string(fdata,sdum);

	if (strlen(sdum)>35) withub=1;
	else		     withub=0;
	
	if (nr_of_param>=MXNO_ANGLE_PARAM) {
		txtout("Fatal error: MXNO_ANGLE_PARAM %d to small. You need at least %d.\n",
			MXNO_ANGLE_PARAM,nr_of_param);
		pex(-1);	
	}
	for (i=0;i<MXNO_ANGLE_PARAM;i++){
		global_data->k_angle[i] = 0.0;
		global_data->theta0[i]  = 0.0;
		global_data->ubk[i]     = 0.0; /*** for Urey-Bradley-Term ***/
		global_data->ubd[i]     = 0.0;
	}
	for (i=0;i<nr_of_param;i++){
		if (feof(fdata)) {
			errout("Fatal error: Unexpected end of file while reading ANGLE data!\n");
			pex(-1);
		}
		index = read_int(fdata);	/* numbering */
		if (i==index) {
			global_data->k_angle[i] = read_double(fdata);
			global_data->theta0[i]  = read_double(fdata);
			if (withub) {
				global_data->ubk[i] = read_double(fdata);
				global_data->ubd[i] = read_double(fdata);	
			}
#if DEBUG
			txtout("ANGLEPARA: %d %f %f %f %f\n",i,
					global_data->k_angle[i],
					global_data->theta0[i],
					global_data->ubk[i],
					global_data->ubd[i]);
#endif

		} else {
			errout("Fatal error: Error reading angle potential parameters.\n");
			pex(-1);
		}
 	}
	fclose(fdata);
	/************** end ANGLE parameter for all nodes *****************/

	/******************************************************************/
 	/******** DIHEDRAL parameter for all nodes ************************/
	/******************************************************************/
	txtout("Reading dihedral parameters..........: %s\n",filenames[DIHEPARA_FILE_INDEX]);
	if ((fdata = fopen(filenames[DIHEPARA_FILE_INDEX], "r")) == NULL) {
		errout("Fatal error: Cannot open DIHEDRAL parameter file!\n");
		pex(-1);
	}

	skip_line (fdata, 1);	/* skip first line */
	fscanf(fdata,"%s%s%s%s",sdum,sdum,sdum,sdum);	/* skip 4 words */
	nr_of_param = read_int_skip(fdata);
	skip_line(fdata, 2);	/* skip 2 more text lines */
	if (nr_of_param>=MXNO_DIHED_PARAM) {
		errout("MXNO_DIHED_PARAM %d to small. You need at least %d.\n",
			MXNO_DIHED_PARAM,nr_of_param);
		pex(-1);	
	}
	for (i=0;i<MXNO_DIHED_PARAM;i++) {
		global_data->k_dihe[i] = 0.0;
		global_data->phi0[i]   = 0.0;
		global_data->n_dihe[i] = 0;
	}
	for (i=0;i<nr_of_param;i++) {
		if (feof(fdata)) {
			errout("Unexpected end of file while reading DIHEDRAL data!\n");
			pex(-1);
		}
		index = read_int (fdata);	/* numbering */
		if (i==index) {
			global_data->k_dihe[i] = read_double (fdata);
			global_data->phi0[i]   = read_double (fdata);
			global_data->n_dihe[i] = read_int (fdata);
		} else {
			errout("Error reading dihedral potential parameters\n");
			pex(-1);
		}
	}
 	fclose (fdata);
	/*********** end DIHEDRAL parameter for all nodes *****************/

	/******************************************************************/
	/*********** IMPROPER parameter for all nodes *********************/
	/******************************************************************/
	txtout("Reading improper parameters..........: %s\n",filenames[VDWPARA_FILE_INDEX]);
	if ((fdata = fopen(filenames[IMPRPARA_FILE_INDEX], "r")) == NULL) {
		errout("Cannot open IMPROPER parameter file!\n");
		pex(-1);
	}

	skip_line(fdata, 1);	/* skip first line */
	fscanf(fdata,"%s%s%s%s",sdum,sdum,sdum,sdum);	/* skip 4 words */
	nr_of_param = read_int_skip(fdata);
	skip_line(fdata, 2);	/* skip 2 more text lines */
	if (nr_of_param>=MXNO_IMPROP_PARAM) {
		errout("MXNO_IMPROP_PARAM %d to small. You need at least %d.\n",
			MXNO_IMPROP_PARAM,nr_of_param);
		pex(-1);	
	}
	for (i=0;i<MXNO_IMPROP_PARAM;i++) {
		global_data->k_impr[i] = 0.0;
		global_data->omega0[i] = 0.0;
	}
	for (i=0;i<nr_of_param;i++){
		if (feof(fdata)) {
			errout("Unexpected end of file while reading IMPROPER data!\n");
			pex(-1);
		}
		index = read_int(fdata);	/* numbering */
		if (i==index) {
			global_data->k_impr[i] = read_double(fdata);
			global_data->omega0[i] = read_double(fdata);
		} else {
			errout("Error reading improper potential parameters\n");
			pex(-1);	
		}
	}
	fclose(fdata);
	/*********** end IMPROPER parameter for all nodes *******************/

	return 0;
}

static int alloc_loader_buffers(	int Nr_of_Atoms, 
			int no_nodes,
			STRUCT_FORCES *my_force)
{
	int i;
	
	last_atom      = alloc_ints(no_nodes);
	atoms_per_node = alloc_ints(no_nodes);
	
	set_allocinfo("LOADMSGONENODE");
	if (no_nodes==1) {
		start_msg_numalloc = Nr_of_Atoms+10;
	} else {
		start_msg_numalloc = (int)(((Nr_of_Atoms+no_nodes+1)/no_nodes)*1.5);
	}
	start_msg.no_atoms      = start_msg_numalloc;
	start_msg.no_more_atoms = start_msg_numalloc;
	alloc_LOAD_MSG_ONE_NODE(&start_msg);

	set_allocinfo("startmsglisten");
	mxno_bond_excl    = (int) ((Nr_of_Atoms/no_nodes)*3.6);
	start_msg.ex_list = alloc_listsize(mxno_bond_excl);

	mxno_bond_ex14      = (int) ((Nr_of_Atoms/no_nodes)*1.6);
	start_msg.ex14_list = alloc_listsize(mxno_bond_ex14);
    
	mxlen_local_shake_list = (int) ((Nr_of_Atoms/no_nodes)*1.5);
	start_msg.shake_list   = alloc_listsize(mxlen_local_shake_list);

	mxno_intern_bonds       = (int) ((Nr_of_Atoms/no_nodes)*2);
	start_msg.int_bond_list = alloc_listsize(mxno_intern_bonds);
	start_msg.ext_bond_list = alloc_list(&mxno_extern_bonds);
    
	mxno_angles = (int) ((Nr_of_Atoms/no_nodes)*1.5);
	start_msg.int_angle_Nr1 = alloc_listsize(mxno_angles);
	start_msg.int_angle_Nr2 = alloc_listsize(mxno_angles);
	start_msg.int_angle_Nr3 = alloc_listsize(mxno_angles);
	start_msg.int_angle_index=alloc_listsize(mxno_angles);
    
	start_msg.ext_angle_Nr1 = alloc_list(&mxno_extern_angles);
	start_msg.ext_angle_Nr2 = alloc_list(&mxno_extern_angles);
	start_msg.ext_angle_Nr3 = alloc_list(&mxno_extern_angles);
	start_msg.ext_angle_index=alloc_list(&mxno_extern_angles);
    
	start_msg.int_dihe_Nr1 = alloc_list(&mxno_dihed);
	start_msg.int_dihe_Nr2 = alloc_list(&mxno_dihed);
	start_msg.int_dihe_Nr3 = alloc_list(&mxno_dihed);
	start_msg.int_dihe_Nr4 = alloc_list(&mxno_dihed);
	start_msg.int_dihe_index=alloc_list(&mxno_dihed);
    
	start_msg.ext_dihe_Nr1 = alloc_list(&mxno_extern_dihed);
	start_msg.ext_dihe_Nr2 = alloc_list(&mxno_extern_dihed);
	start_msg.ext_dihe_Nr3 = alloc_list(&mxno_extern_dihed);
	start_msg.ext_dihe_Nr4 = alloc_list(&mxno_extern_dihed);
	start_msg.ext_dihe_index=alloc_list(&mxno_extern_dihed);
    
	start_msg.int_impr_Nr1 = alloc_list(&mxno_improp);
	start_msg.int_impr_Nr2 = alloc_list(&mxno_improp);
	start_msg.int_impr_Nr3 = alloc_list(&mxno_improp);
	start_msg.int_impr_Nr4 = alloc_list(&mxno_improp);
	start_msg.int_impr_index=alloc_list(&mxno_improp);
    
	start_msg.ext_impr_Nr1 = alloc_list(&mxno_extern_improp);
	start_msg.ext_impr_Nr2 = alloc_list(&mxno_extern_improp);
	start_msg.ext_impr_Nr3 = alloc_list(&mxno_extern_improp);
	start_msg.ext_impr_Nr4 = alloc_list(&mxno_extern_improp);
	start_msg.ext_impr_index=alloc_list(&mxno_extern_improp);

	{int level;
	for (level=0;level<MAX_NO_LEVELS;level++) {
		alloc_per_level[level]  = 0;
		start_msg.cluster[level]= NULL;
	}
	}

	/*******************************************************************/
	/*** here we alloc and initialize some buffers which we need for ***/
	/*** the handling of energy terms which have been selected for   ***/
	/*** writing out to files (e.g. by the command PICKDIHE, etc.)   ***/
	/*** the my_force structure keeps the forces comming from the    ***/
	/*** nodes. The start_msg.pick structure is needed for sending   ***/
	/*** the specification which atoms are selected to the nodes     ***/
	/*******************************************************************/
	my_force->pickbond.len     = my_force->pickbond.cur_len;
	my_force->pickbond.cur_len = 0;
	my_force->pickbond.force   = alloc_doubles(5*my_force->pickbond.len);
	for (i=0;i<5*my_force->pickbond.len;i++) my_force->pickbond.force[i]=0.0;

	my_force->pickelec.len     = my_force->pickelec.cur_len;
	my_force->pickelec.cur_len = 0;
	my_force->pickelec.force   = alloc_doubles(5*my_force->pickelec.len);
	for (i=0;i<5*my_force->pickelec.len;i++) my_force->pickelec.force[i]=0.0;

	my_force->pickvdw.len     = my_force->pickvdw.cur_len;
	my_force->pickvdw.cur_len = 0;
	my_force->pickvdw.force   = alloc_doubles(5*my_force->pickvdw.len);
	for (i=0;i<5*my_force->pickvdw.len;i++) my_force->pickvdw.force[i]=0.0;

	my_force->pickangle.len     = my_force->pickangle.cur_len;
	my_force->pickangle.cur_len = 0;
	my_force->pickangle.force   = alloc_doubles(8*my_force->pickangle.len);
	for (i=0;i<8*my_force->pickangle.len;i++) my_force->pickangle.force[i]=0.0;

	my_force->pickdihe.len     = my_force->pickdihe.cur_len;
	my_force->pickdihe.cur_len = 0;
	my_force->pickdihe.force   = alloc_doubles(14*my_force->pickdihe.len);
	for (i=0;i<14*my_force->pickdihe.len;i++) my_force->pickdihe.force[i]=0.0;

	my_force->pickimpr.len     = my_force->pickimpr.cur_len;
	my_force->pickimpr.cur_len = 0;
	my_force->pickimpr.force   = alloc_doubles(14*my_force->pickimpr.len);
	for (i=0;i<14*my_force->pickimpr.len;i++) my_force->pickimpr.force[i]=0.0;

	start_msg.pickbond.atom_ids = alloc_ints(2*my_force->pickbond.len);
	start_msg.pickbond.force    = alloc_doubles(5*my_force->pickbond.len);
	for (i=0;i<5*my_force->pickbond.len;i++) start_msg.pickbond.force[i]=0.0;

	start_msg.pickelec.atom_ids = alloc_ints(2*my_force->pickelec.len);
	start_msg.pickelec.force    = alloc_doubles(5*my_force->pickelec.len);
	for (i=0;i<5*my_force->pickelec.len;i++) start_msg.pickelec.force[i]=0.0;

	start_msg.pickvdw.atom_ids = alloc_ints(2*my_force->pickvdw.len);
	start_msg.pickvdw.force    = alloc_doubles(5*my_force->pickvdw.len);
	for (i=0;i<5*my_force->pickvdw.len;i++) start_msg.pickvdw.force[i]=0.0;

	start_msg.pickangle.atom_ids = alloc_ints(3*my_force->pickangle.len);
	start_msg.pickangle.force    = alloc_doubles(8*my_force->pickangle.len);
	for (i=0;i<8*my_force->pickangle.len;i++) start_msg.pickangle.force[i]=0.0;

	start_msg.pickdihe.atom_ids = alloc_ints(4*my_force->pickdihe.len);
	start_msg.pickdihe.force    = alloc_doubles(14*my_force->pickdihe.len);
	for (i=0;i<14*my_force->pickdihe.len;i++) start_msg.pickdihe.force[i]=0.0;

	start_msg.pickimpr.atom_ids = alloc_ints(4*my_force->pickimpr.len);
	start_msg.pickimpr.force    = alloc_doubles(14*my_force->pickimpr.len);
	for (i=0;i<14*my_force->pickimpr.len;i++) start_msg.pickimpr.force[i]=0.0;

	set_allocinfo("XXX");
	return 0;
}

int
free_loader_buffers(LOAD_MSG_ALL_NODE *global_data)
{
  	myfree(last_atom);
	myfree(atoms_per_node);

	myfree(bonddata);
	if (excldata!=NULL)	myfree(excldata);
	if (excl14data!=NULL)	myfree(excl14data);

	myfree(angledata);
	myfree(dihedata);
	myfree(imprdata);

	if (shakedata!=NULL)	myfree(shakedata);

	free_LOAD_MSG_ONE_NODE(&start_msg);

	CLU_free_clusters(	&(global_data->treeinfo),
				global_data->levelinfo,
				cluster,
				child_nr);

	return 0;
}

int
do_new_clustering(	int Nr_of_Atoms,
			LOAD_MSG_ALL_NODE *global_data,
			int flag_first_time,
			STRUCT_FORCES *my_force)
{
	int	res,nodenum,i=flag_first_time; /*** no warning ***/
	int	cur_size_anglelist,cur_size_dihelist,cur_size_imprlist;
	int	myexclind=0;
		
	set_allocinfo("do_new_clustering");
	
	/*********************************************************************/
	/*** unit-file is read, and a tree of clusterlevels is build up.    **/
	/*** child_nr holds the numbers of childs for each Cluster 	    **/
	/*** child_nr[0] holds the (OLD-EGO)atom-numbers 		    **/
	/*********************************************************************/

	res=CLU_buildup_clusters(	Nr_of_Atoms,
					global_data->no_nodes,
					global_data->atomdata,
					global_data->filenames[UNIT_FILE_INDEX],
					&(global_data->treeinfo),
					global_data->levelinfo,
					cluster,
					child_nr,
					global_data);
	if (res==-1) {
		errout("Fatal error in making clusters !\n");
                errout("Change in ctl.lis 'Final number of branches per node'\n");
		pex(-1);
	}

	/*******************************************************************/
	res=make_atommap(	Nr_of_Atoms,
				global_data->no_nodes,
				&(global_data->treeinfo),
				global_data->levelinfo,
				cluster,
				child_nr,
				global_data->old2new_map,
				global_data->new2old_map,
				atoms_per_node);
		
	for (i=0,global_data->maxatoms_per_node=0;i<global_data->no_nodes;i++)
		if (global_data->maxatoms_per_node<atoms_per_node[i])
			global_data->maxatoms_per_node = atoms_per_node[i];

	/*******************************************************************/
	/******* we need the bondings in memory         ********************/
	/******* bondlist is sorted due to new ordering ********************/
	/*******************************************************************/
	txtout("Reading bondlist.............: %s\n",global_data->filenames[BONDLIST_FILE_INDEX]);
	res=readin_bondfile(	Nr_of_Atoms,
				global_data->filenames[BONDLIST_FILE_INDEX],
				bonddata,
				global_data->old2new_map);

	/*******************************************************************/
	/******* we need the exclusion-list in memory    *******************/
	/******* exclusion is sorted due to new ordering *******************/
	/*******************************************************************/
	if (global_data->ex_number<0) myexclind = 4 + abs(global_data->ex_number);
	else
	if (global_data->ex_number>0) myexclind = global_data->ex_number-1;

	if (global_data->ex_number==-5) myexclind = 7;
	else
	if (global_data->ex_number==5) myexclind = 2;
	
	txtout("Reading exclusion-list.......: %s\n",global_data->filenames[EXCL_FILE_INDEX+myexclind]);
	res=readin_exclfile(	Nr_of_Atoms,
				0, /* flag, because header of 14-exclusion is different */
				global_data->filenames[EXCL_FILE_INDEX+myexclind],
				&excldata,
				global_data->old2new_map,
				last_maxexcl);
	last_maxexcl=res;	

	if (global_data->special_14_excl) {
		txtout("Reading exclusion-14-list....: %s\n",global_data->filenames[EXCL_FILE_INDEX+4]);
		res=readin_exclfile(	Nr_of_Atoms,
					1, /* flag, because header of 14-exclusion is different */
					global_data->filenames[EXCL_FILE_INDEX+4],
					&excl14data,
					global_data->old2new_map,
					last_maxexcl14);
		last_maxexcl14=res;
	}

	/********************************************************************/
	/********** we need the angle-list in memory ************************/
	/********** atomnumbers are mapped, but no resorting is needed ******/
	/********************************************************************/

	/*** allocated space is eventuell increased in this function *****/
	txtout("Reading anglelist............: %s\n",global_data->filenames[ANGLELIST_FILE_INDEX]);
	res=readin_anglefile(	&size_anglelist,
				global_data->filenames[ANGLELIST_FILE_INDEX],
				&angledata,
				global_data->old2new_map);
	cur_size_anglelist = res;

	/*******************************************************************/
	/********** we need the dihedral-list in memory ********************/
	/********** atomnumbers are mapped, but no resorting is needed *****/
	/*******************************************************************/

	/*** allocated space is eventuell increased in this function *****/
	txtout("Reading dihedrallist.........: %s\n",global_data->filenames[DIHELIST_FILE_INDEX]);
	res=readin_dihefile(	&size_dihelist,
				global_data->filenames[DIHELIST_FILE_INDEX],
				&dihedata,
				global_data->old2new_map);

	cur_size_dihelist = res;

	/*******************************************************************/
	/********** we need the improper-list in memory ********************/
	/********** atomnumbers are mapped, but no resorting is needed *****/
	/*******************************************************************/

	/*** allocated space is eventuell increased in this function *****/
	txtout("Reading improperlist.........: %s\n",global_data->filenames[IMPRLIST_FILE_INDEX]);
	res=readin_dihefile(	&size_imprlist,
				global_data->filenames[IMPRLIST_FILE_INDEX],
				&imprdata,
				global_data->old2new_map);

	cur_size_imprlist = res;

	/******************************************************************/
	/********** ev. we need the shake-list in memory    		***/
	/********** atomnumbers are mapped, but no resorting is needed  ***/
	/******************************************************************/

	if (global_data->shake) {
		txtout("Reading shakelist............: %s\n",global_data->filenames[SHAKE_FILE_INDEX]);
		read_shake_file(global_data,
				&(global_data->num_shake),
				&(global_data->num_shake_h2o),
				&(global_data->num_shake_protein),
				global_data->filenames[SHAKE_FILE_INDEX],
				&shakedata,
				global_data->old2new_map);
	}

	/**** see how many atoms each node will get ***********************/
	global_data->first_atom[0] = 0;
	last_atom[0] = atoms_per_node[0] - 1;

	for (i=1;i<global_data->no_nodes;i++) {
		global_data->first_atom[i] = last_atom[i-1] + 1;
		last_atom[i]  = global_data->first_atom[i] + atoms_per_node[i] - 1;
	}

	/******************************************************************/
	/****** that is the big loop over all nodes ***********************/
	/******************************************************************/
	txtout("\nStart now distributing data to nodes ...\n");
	for (nodenum=0; nodenum<global_data->no_nodes; nodenum++) {
		int	first = global_data->first_atom[nodenum];
		int	last  = last_atom[nodenum];
		int	Atomanzahl=last-first+1;
		
		if (Atomanzahl>=start_msg_numalloc) {
			start_msg.no_atoms = Atomanzahl; /* number of atoms for this node */
			realloc_LOAD_MSG_ONE_NODE(&start_msg);
			start_msg_numalloc = start_msg.no_more_atoms; /* copy back how many atoms we actually allocated */
		}

		/*******************************************************/
		res=distri_atomdata(	nodenum,
					first,last,
					global_data->atomdata,
					&start_msg,
					global_data->new2old_map,
					global_data);

		/*******************************************************/
		distri_harmonic(first,last,
				global_data->atomdata,
				&start_msg,
				global_data->new2old_map);
		if (global_data->harmonic) {
			txtout("Reference position list    : % 4d\n",Atomanzahl);
		} else { /*** fill with dummy-data==0.0 --- important ! ***/
			int atom_number;
			for (atom_number=0;atom_number<Atomanzahl;atom_number++) {	
				start_msg.harmon_k[atom_number] = 0.0;
			}	
		}	
		if (global_data->stochastic) {
			txtout("Stochastic list length     : % 4d\n",Atomanzahl);
		} else { /*** fill with dummy-data==0.0 --- important ! ***/
			int atom_number;
			for (atom_number=0;atom_number<Atomanzahl;atom_number++) {	
				start_msg.beta[atom_number]	= 0.0;
			}	
		}	

		/********************************************************/
		distri_excl(	first,last,
				excldata,
				&start_msg,
				&mxno_bond_excl,
				last_maxexcl,
				global_data);
		txtout("EXCLUSION list             : % 4d\n",
			start_msg.ex_ptr-Atomanzahl);
		
		if (global_data->special_14_excl) {
			distri_excl14(	first,last,
					excl14data,
					&start_msg,
					&mxno_bond_ex14,
					last_maxexcl14,
					global_data);
			txtout("1-4 EXCLUSION list         : % 4d\n",
				start_msg.ex14_ptr-Atomanzahl);
		}

		/*** signal that list is complete ***/
		set_qmexclusionatom(-1,-1,global_data,0);

		/**********************************************************/
		
		if (global_data->shake) {
			res = distri_shake(first,last,
					shakedata,
					&start_msg,
					&mxlen_local_shake_list,
					global_data);
			txtout("SHAKE list (intern/deleted): % 4d   % 4d\n",
				start_msg.shake_number/2,res);
		} else {
			start_msg.shake_number = 0; /*** important ! ***/
		}

		/***********************************************************/
		/*** make shure,that we have done the SHAKE-stuff before,***/
		/*** as we need it here !*/
		res=distri_bonds(	first,last,
					bonddata,
					&start_msg,
					&mxno_extern_bonds,
					&mxno_intern_bonds);

		txtout("BOND list   (intern/extern): % 4d   % 4d\n",
					start_msg.int_bond_ptr-Atomanzahl,res);
		/***********************************************************/
		res=distri_angles(	cur_size_anglelist,
					first,last,
					angledata,
					&start_msg,
					&mxno_angles,
					&mxno_extern_angles );

		txtout("ANGLE list  (intern/extern): % 4d   % 4d\n",
				start_msg.int_angle_ptr,res);

		/***********************************************************/
		res=distri_dihe(	cur_size_dihelist,
					first,last,
					dihedata,
					&start_msg,
					&mxno_dihed,
					&mxno_extern_dihed);

		txtout("DIHEDRAL l. (intern/extern): % 4d   % 4d\n",
				start_msg.int_dihe_ptr,res);

		/***********************************************************/
		res=distri_impr(	cur_size_imprlist,
					first,last,
					imprdata,
					&start_msg,
					&mxno_improp,
					&mxno_extern_improp);

		txtout("IMPROPER l. (intern/extern): % 4d   % 4d\n",
				start_msg.int_impr_ptr,res);

		/***********************************************************/
		res=distri_tree(	nodenum,global_data->no_nodes,
					first,
					&(global_data->treeinfo),
					global_data->levelinfo,
					cluster,
					&start_msg);

		txtout("Branch for node %2d loaded.\n",nodenum+1);


		/***********************************************************/
		res=distri_pickselect(	nodenum,
					first,last,
					my_force,
					&start_msg,
					global_data->old2new_map);

		txtout("Pick-force selection %2d loaded with %d elements.\n",
			nodenum+1,res);

#if defined(SEQ)

#else
		/************************************************************/
		/***** here at least we send the data to the node ***********/
		/************************************************************/
		PCI_send_msg_one_node(&start_msg,nodenum);
		txtout("Data sent to node %d.\n",nodenum+1);
#endif
	}

	/*******************************************************************/
	/******* that is the end of the big loop over all nodes ************/
	/*******************************************************************/
	set_allocinfo("XXX");
	return 0;
}



static
int distri_pickselect(	int nodenum,
			int first,int last,
			STRUCT_FORCES *my_force,
			LOAD_MSG_ONE_NODE *start_msg,
			int *old2new)	
{
	int 	i,j=nodenum,warning=0; /*** avoid warning ***/
	int	shake_number = start_msg->shake_number;
	int	*shake_list;
	shake_list = start_msg->shake_list;

	/******************************************************************/
	/*** PICK_BOND 							***/
	/******************************************************************/	
	start_msg->pickbond.cur_len = 0;
	for (i=0;i<2*my_force->pickbond.len;i+=2) {
		int bp1 = old2new[my_force->pickbond.atom_ids[i  ]];
		int bp2 = old2new[my_force->pickbond.atom_ids[i+1]];

		if (is_local_shake_hyd(bp1,shake_number,shake_list)) 
			warning = my_force->pickbond.atom_ids[i  ]+1;
		else if (is_local_shake_hyd(bp2,shake_number,shake_list))
			warning = my_force->pickbond.atom_ids[i+1]+1;;

		if ( (!inrange(bp1,first,last)) &&
		     (!inrange(bp2,first,last)) ) continue; /*** nothing for current node ***/

		j = 2*start_msg->pickbond.cur_len;

		start_msg->pickbond.atom_ids[j  ] = my_force->pickbond.atom_ids[i  ];
		start_msg->pickbond.atom_ids[j+1] = my_force->pickbond.atom_ids[i+1];
		start_msg->pickbond.cur_len++;
	}
	start_msg->pickbond.len = start_msg->pickbond.cur_len;

	/******************************************************************/
	/*** PICK_ELEC 							***/
	/******************************************************************/	
	start_msg->pickelec.cur_len = 0;
	for (i=0;i<2*my_force->pickelec.len;i+=2) {
		int bp1 = old2new[my_force->pickelec.atom_ids[i  ]];
		int bp2 = old2new[my_force->pickelec.atom_ids[i+1]];

		if ( (!inrange(bp1,first,last)) &&
		     (!inrange(bp2,first,last)) ) continue; /*** nothing for current node ***/

		j = 2*start_msg->pickelec.cur_len;

		start_msg->pickelec.atom_ids[j  ] = my_force->pickelec.atom_ids[i  ];
		start_msg->pickelec.atom_ids[j+1] = my_force->pickelec.atom_ids[i+1];
		start_msg->pickelec.cur_len++;
	}
	start_msg->pickelec.len = start_msg->pickelec.cur_len;

	/******************************************************************/
	/*** PICK_VDW 							***/
	/******************************************************************/	
	start_msg->pickvdw.cur_len = 0;
	for (i=0;i<2*my_force->pickvdw.len;i+=2) {
		int bp1 = old2new[my_force->pickvdw.atom_ids[i  ]];
		int bp2 = old2new[my_force->pickvdw.atom_ids[i+1]];

		if ( (!inrange(bp1,first,last)) &&
		     (!inrange(bp2,first,last)) ) continue; /*** nothing for current node ***/

		j = 2*start_msg->pickvdw.cur_len;

		start_msg->pickvdw.atom_ids[j  ] = my_force->pickvdw.atom_ids[i  ];
		start_msg->pickvdw.atom_ids[j+1] = my_force->pickvdw.atom_ids[i+1];
		start_msg->pickvdw.cur_len++;
	}
	start_msg->pickvdw.len = start_msg->pickvdw.cur_len;

	/******************************************************************/
	/*** PICK_ANGLE 						***/
	/******************************************************************/	
	start_msg->pickangle.cur_len = 0;
	for (i=0;i<3*my_force->pickangle.len;i+=3) {
		int bp1 = old2new[my_force->pickangle.atom_ids[i  ]];
		int bp2 = old2new[my_force->pickangle.atom_ids[i+1]];
		int bp3 = old2new[my_force->pickangle.atom_ids[i+2]];

		if ( (!inrange(bp1,first,last)) &&
		     (!inrange(bp2,first,last)) &&
		     (!inrange(bp3,first,last)) ) continue; /*** nothing for current node ***/

		j = 3*start_msg->pickangle.cur_len;

		start_msg->pickangle.atom_ids[j  ] = my_force->pickangle.atom_ids[i  ];
		start_msg->pickangle.atom_ids[j+1] = my_force->pickangle.atom_ids[i+1];
		start_msg->pickangle.atom_ids[j+2] = my_force->pickangle.atom_ids[i+2];
		start_msg->pickangle.cur_len++;
	}
	start_msg->pickangle.len = start_msg->pickangle.cur_len;

	/******************************************************************/
	/*** PICK_DIHE	 						***/
	/******************************************************************/	
	start_msg->pickdihe.cur_len = 0;
	for (i=0;i<4*my_force->pickdihe.len;i+=4) {
		int bp1 = old2new[my_force->pickdihe.atom_ids[i  ]];
		int bp2 = old2new[my_force->pickdihe.atom_ids[i+1]];
		int bp3 = old2new[my_force->pickdihe.atom_ids[i+2]];
		int bp4 = old2new[my_force->pickdihe.atom_ids[i+3]];

		if ( (!inrange(bp1,first,last)) &&
		     (!inrange(bp2,first,last)) &&
		     (!inrange(bp3,first,last)) &&
		     (!inrange(bp4,first,last)) ) continue; /*** nothing for current node ***/

		j = 4*start_msg->pickdihe.cur_len;

		start_msg->pickdihe.atom_ids[j  ] = my_force->pickdihe.atom_ids[i  ];
		start_msg->pickdihe.atom_ids[j+1] = my_force->pickdihe.atom_ids[i+1];
		start_msg->pickdihe.atom_ids[j+2] = my_force->pickdihe.atom_ids[i+2];
		start_msg->pickdihe.atom_ids[j+3] = my_force->pickdihe.atom_ids[i+3];
		start_msg->pickdihe.cur_len++;
	}
	start_msg->pickdihe.len = start_msg->pickdihe.cur_len;

	/******************************************************************/
	/*** PICK_IMPR	 						***/
	/******************************************************************/	
	start_msg->pickimpr.cur_len = 0;
	for (i=0;i<4*my_force->pickimpr.len;i+=4) {
		int bp1 = old2new[my_force->pickimpr.atom_ids[i  ]];
		int bp2 = old2new[my_force->pickimpr.atom_ids[i+1]];
		int bp3 = old2new[my_force->pickimpr.atom_ids[i+2]];
		int bp4 = old2new[my_force->pickimpr.atom_ids[i+3]];

		if ( (!inrange(bp1,first,last)) &&
		     (!inrange(bp2,first,last)) &&
		     (!inrange(bp3,first,last)) &&
		     (!inrange(bp4,first,last)) ) continue; /*** nothing for current node ***/

		j = 4*start_msg->pickimpr.cur_len;

		start_msg->pickimpr.atom_ids[j  ] = my_force->pickimpr.atom_ids[i  ];
		start_msg->pickimpr.atom_ids[j+1] = my_force->pickimpr.atom_ids[i+1];
		start_msg->pickimpr.atom_ids[j+2] = my_force->pickimpr.atom_ids[i+2];
		start_msg->pickimpr.atom_ids[j+3] = my_force->pickimpr.atom_ids[i+3];
		start_msg->pickimpr.cur_len++;
	}
	start_msg->pickimpr.len = start_msg->pickimpr.cur_len;

	return (start_msg->pickbond.len + start_msg->pickelec.len  +
		start_msg->pickvdw.len  + start_msg->pickangle.len +
		start_msg->pickdihe.len + start_msg->pickimpr.len);	      
}	

/**********************************************************************/
/** make old2new_map, new2old_map and atoms_per_node                 **/
/** any atoms my belong to any node. we want to change the numbering,**/
/** so that the numbering is ascending from node to node. We need    **/
/** therefore two lists for changing the numbering vice-versa        **/
/**********************************************************************/
static
int	make_atommap(	int nr_of_atoms,
			int nr_of_nodes,
			struct_treeinfo *treeinfo,
			struct_levelinfo *levelinfo,
			struct_cluster_reduced *cluster[],
			int *child_nr[],
			int *old2new_map,
			int *new2old_map,
			int *atoms_per_node)
{
	int	branch,cur_node=0,cur_branch_per_node=0;
	int	level=treeinfo->levels-1;
	int	nr_branches=levelinfo[level].nr_of_clusters;
	int 	*atom_nr,maxatoms=100;
	int	branch_atoms=0,sum_atoms=0;
	int	i;

	atom_nr=(int*)alloc_ints(maxatoms);

	for (branch=0; branch<nr_branches; branch++) {
		branch_atoms=CLU_get_all_atoms(level,branch,&maxatoms,&atom_nr,
						cluster,child_nr[0]);
/***
txtout("branch %d   atoms %d\n",branch,branch_atoms);
***/
		if (sum_atoms+branch_atoms>nr_of_atoms) {
			errout("Fatal error 0: Illegal cluster-structure in make_atommap.\n");
			myfree(atom_nr);
			pex(-1);
		}

		memcpy(new2old_map+sum_atoms,atom_nr,branch_atoms*sizeof(int));
		sum_atoms+=branch_atoms;
	
		if (cur_node>=nr_of_nodes) {
			errout("Fatal error 1: Illegal cluster-structure in make_atommap.\n");
			myfree(atom_nr);
			pex(-1);;
		}

		if (cur_branch_per_node==0) {
			atoms_per_node[cur_node]=branch_atoms;
		} else {
			atoms_per_node[cur_node]+=branch_atoms;
		}
	
		if (++cur_branch_per_node==treeinfo->branches/nr_of_nodes) {
			cur_node++;
			cur_branch_per_node=0;
			txtout("Number of atoms on Node % 3d: % 5d\n",
				cur_node,atoms_per_node[cur_node-1]);
		}
	}
	if (cur_node!=nr_of_nodes) {
		errout("Fatal error 2: Illegal cluster-structure in make_atommap.\n");
		myfree(atom_nr);
		pex(-1);
	}

	for (i=0;i<nr_of_atoms;i++) {
		old2new_map[new2old_map[i]]=i;
	}

	myfree(atom_nr);
	return 0;
}

