/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/**************************************************************************/
/*** egmach.c contains all parallel comunication stuff.			***/
/*** all functions begin with "PCI_" which stands for Parallel Computer ***/
/*** Interface.								***/
/**************************************************************************/

#if defined(SEQ)

/*** egmach.c is empty for the sequential case ***/

#else

/* DEBUG = 0:	no debugging */
/* DEBUG = 1:	standard debugging */
/* DEBUG = 2:	PCI_exit() debugging */
#define DEBUG		2
#define WITH_FORCESTAT	1
#define SLOW_MPI	1	/* 0 or 1 */
#define	SAVE_COM	1

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>

#include "ego.h"
#include "util_lib.h"
#include "egcontro.h"
#include "egalloc.h"
#include "egmach.h"
#include "timing.h"
#include "egnolib.h"
#include "egbolib.h"
#include "egtree.h"
#include "egqminter.h"

#if defined(PARIX) || defined(PARIXHAL)
/**************************************************/
/***************** ONLY for PARIX-case ************/
/**************************************************/
#include "egmepvm.h"

static int do_send_data(ARGUMENTS_THREAD *arg_send);
static int do_recv_data(ARGUMENTS_THREAD *arg_recv);

int    	LocalLink(LinkCB_t *Link[2]);
static LinkCB_t *link_to_thread_send[2];
static LinkCB_t *link_to_thread_recv[2];
static int	*StarNodeId;
static int 	len;
static void 	*ptr;


#if defined(RS6K)
#include <epx/select.h>
#include <epx/sem.h>

#define MAXNODES	100
static	int	masterstar_top,nodestar_top;
static	int	masternode_link[MAXNODES];
static	int	nodenode_link[MAXNODES];
static	LinkCB_t *nodenode_CB[MAXNODES];
static	Option_t nodenode_opt[MAXNODES];

#if SAVECOMD

#else
/**** here comes the stuff from Regina Wanger ****/
static LinkCB_t *links[MAXNODES];
static LinkCB_t *e_links[MAXNODES];
static Option_t   opts[MAXNODES];
static int nlinks, nopts;
static int dolink = 0;

static
int SendNode(int i, int req, void *buffer,int size)
{
   int err, j=req;
   if(dolink == 0){
      dolink = 1;
      nlinks=GET_ROOT()->ProcRoot->nProcs;
      nopts = nlinks-1;
      for(j=0;j<GET_ROOT()->ProcRoot->nProcs;j++)
        if(j!=GET_ROOT()->ProcRoot->MyProcID){
          links[j]=ConnectLink(j,0,&err);
          e_links[j]=ConnectLink(j,0,&err);
       	  opts[j] = ReceiveOption(e_links[j]);
	}
   }
   if(i>=GET_ROOT()->ProcRoot->nProcs){
      txtout("Wrong proc to send to ... %d (%d) \n",i, size);
      while(1);
   }
   return SendLink(links[i],buffer,size);   
}

static
int RecvNode(int i, int req, void *buffer,int size)
{
   int err, j, flag;
   if(dolink == 0){
/**      int me = GET_ROOT()->ProcRoot->MyProcID; **/
      dolink = 1;
      nlinks=GET_ROOT()->ProcRoot->nProcs;
      nopts = nlinks-1;
      for(j=0;j<nlinks;j++){
	 if(GET_ROOT()->ProcRoot->MyProcID!=j){
          links[j]=ConnectLink(j,0,&err);
          e_links[j]=ConnectLink(j,0,&err);
       	  opts[j] = ReceiveOption(e_links[j]);
	 }
	}
   }
   return RecvLink(links[i],buffer,size);
}

static
int e_SendNode(int i, int req, void *buffer,int size)
{
   int j=req;j++;
   return SendLink(e_links[i],buffer,size);   
}

static
int e_RecvNode(int i, int req, void *buffer,int size)
{
   int err, j, flag=req;
   if(i!=-1) return RecvLink(e_links[i],buffer,size);
   if(GET_ROOT()->ProcRoot->MyProcID != 0) return (-1);
   flag = SelectList( nopts, & opts[1]);
   if(flag < 0) return (-1);
   if(flag > nlinks) return -1;
   return RecvLink(e_links[flag+1],buffer,size);      
}
#endif	/*** end of SAVE_COMD	***/
#endif	/*** end of RS6K special case ***/

/******************* end of PARIX-case ************/


#elif defined(MPI)
/**************************************************/
/***************** ONLY for MPI-case **************/
/**************************************************/
#include <mpi.h>
#include "egmepvm.h"

static MPI_Group 	MPI_GROUP_WORLD,only_workers_group;
static MPI_Comm		only_workers_comm;

static int 	len;
static void 	*ptr;

static	char	*mpi_send_buffer;
static	int	mpi_buffer_size;		


#else 	/******************* end of MPI-case **************/


/***************************************************/
/****************** ONLY for PVM-case **************/
/***************************************************/
#include "pvm3.h"
static int	*node_tids;
#define	NODEFILENAME	"node"

#if (defined(T3D) && !defined(VPP)) 	/***** PVM-Subcase T3D *****/
extern long _MPP_N_PES;
#elif defined(PARIXPPC) /***** PVM-Subcase PPC ****/
#undef NODEFILENAME
#define	NODEFILENAME	"node.px"
#endif

#endif 	/*************** end of PVM-case ******************/


static int really_my_number=-1;
static int my_pvm_freebuf(int ido);
static int encoding=0;
static int num_compnodes=0;
static int ProcNum=0;
static int spawn_pvm_task(int no_nodes);

/* Tokens used in communication */
#define LOAD_SINGLE 	0
#define COORDINATES 	1
#define TNPROCS		2
#define TMYDATA		3
#define TGLOBALDATA	4
#define TTCOORDS	5
#define TTPACKS		6
#define TENERGY		7
#define TSCALING	8
#define TFITEM		9
#define PACKETS     	10
#define ENERGY      	11
#define SHUFFLE     	12
#define TTBRANCHES	13
#define TFORCES		14
#define COANDBR		15
#define TNONODES	16
#define TNODEIDS	17
#define TMYNUMBER	18
#define TQMFORCE	19
#define TETCDATA	20
#define TPOSITIONS      21
#define	TSYNCHRONISE	12345

/******************************************************************/
/*** Called once at the start of EGO on every node (except PVM) ***/
/*** returns ProcId, the MasterID and number of compute nodes	***/
/*** For number of nodes we have to subtract here 1		***/
/*** since the master uses up one node. 			***/
/******************************************************************/
int   PCI_start(int *MyProcID,int *nProcs,int argc,char **argv)
{
	int MasterID=0;
	
#if defined(PARIX) || defined(PARIXHAL)
	int a=argc;
	char b=argv[0][0];
	a++;b++;

        *MyProcID = GET_ROOT()->ProcRoot->MyProcID;
	*nProcs   = GET_ROOT()->ProcRoot->nProcs - 1;
	StarNodeId= alloc_ints(*nProcs); /*** only used on master ***/
	MasterID  = 0;

#elif defined(MPI) /**********************************************/
	MPI_Init(&argc,&argv);
	MPI_Comm_rank(MPI_COMM_WORLD,MyProcID);
	MPI_Comm_size(MPI_COMM_WORLD,nProcs);
	(*nProcs) = (*nProcs) - 1;
	MasterID  = 0;

#elif (defined(T3D) && !defined(VPP)) /*********************************************/
	*MyProcID = pvm_get_PE(pvm_mytid());
	*nProcs   = pvm_gsize(0) - 1;
	encoding  = PvmDataRaw;	
	MasterID  = 0;
	
#elif defined(VPP) /*********************************************/
	*MyProcID = pvm_mytid();
	*nProcs   = 999; /* No way to find out how many we have on a VPP */
	encoding  = PvmDataRaw;	
	MasterID  = 0;
	
#elif defined(SP2) /*********************************************/
	int narch;
	*MyProcID = pvm_mytid();
	pvm_config(nProcs,&narch,NULL);
	(*nProcs) = (*nProcs)-1;
	encoding  = PvmDataRaw;
	MasterID  = *MyProcID;
	
#elif defined(PARIXPPC) /****************************************/
	*MyProcID = pvm_mytid();
	*nProcs   = 16;
	encoding  = PvmDataRaw; 
	MasterID  = *MyProcID;
#else                   /****************************************/
	int    start_pvm_count=0,narch=0;
	*MyProcID = -1;

	/******* enroll in pvm *******/
	while (	(start_pvm_count<MXNOTRYPVM) && (*MyProcID<0)) {

		*MyProcID = pvm_mytid();

		if (*MyProcID == PvmSysErr) {
			errout("\n\nPVM is not running. I start it for you, wait!\n");
			system("echo quit | pvm >/dev/null\n");
			start_pvm_count++;
		}
	}

	if (*MyProcID<0) {
		txtout("Can not start PVM (%d). Program terminating.\n",*MyProcID);
		exit(-2);
	}

	pvm_config(nProcs,&narch,NULL);
	/* not necessary to subtract 1 since we don't use the host node as a
	 * computational node because PVM does not include the host node here */

	if (narch==1) {
		encoding = PvmDataRaw;	/* we are sure to have only the same
					 * architecture everywhere */
	} else {
		encoding = PvmDataDefault;/* let PVM handle the situation */
	}
	MasterID  = *MyProcID;
#endif

	num_compnodes = *nProcs; /* Global variable. Used in PCI_exit() to kill nodes */
	ProcNum       = *MyProcID;
	return MasterID;
}


/************************************************************************/
int    PCI_set_no_nodes(int req_no_nodes, int nProcs)
{
#if (defined(PARIX) || defined(PARIXHAL) || defined(T3D) || defined(SP2)) && !defined(VPP)
	if (req_no_nodes < nProcs)	return req_no_nodes;
	else				return nProcs;
#elif defined(MPI)
	return nProcs;		/* for MPI always like this! */
#else
	return req_no_nodes;	/* for PVM always like this! */
#endif
}

/**************************************************************************/
/*** PCI_master_setup is called only once on the master during setup 	***/
/*** and corresponds to PCI_node_setup which is called on the nodes.    ***/
/**************************************************************************/
int   PCI_master_setup(	int no_nodes, 	/*** compute nodes (without master) ***/
			int nProcs, 	/*** real physical nodes (without master) ***/
			int host_tid)
{
#if defined(PARIX) || defined(PARIXHAL) /*************************************/
  {
	int	i;
	int	b = host_tid;b += 0;

	/*** each node gets to know how many (without master node) will participate ****/
	for (i=0;i<nProcs;i++) {
		SendNode(i+1,TNONODES,(void *)&no_nodes,sizeof(int));
	}

	for (i=0;i<nProcs;i++) {
		RecvNode(i+1,TMYNUMBER,(void *)&StarNodeId[i],sizeof(int));
	}

	for (i=0;i<no_nodes;i++) {
		SendNode(StarNodeId[i],TNODEIDS,StarNodeId,sizeof(int)*no_nodes);
	}

#if defined(RS6K)
   	/*** star-topologie: links from the master to all nodes ***/
	masterstar_top = NewTop(no_nodes);  	
  	for (i=1;i<=no_nodes;i++ ) {
		masternode_link[i-1] = AddTop(masterstar_top,ConnectLink(i,0,&b));
	}
#endif
	return 0;
   }
#elif defined(MPI) /*********************************************************/
   {
	int	star_node;
	int	excl_ranks[2];

	for (star_node=0;star_node<nProcs;star_node++) {
		MPI_Ssend(&no_nodes,1,MPI_INT,star_node+1,TNONODES,MPI_COMM_WORLD);
	}
/****
	excl_ranks[0]=0;

	MPI_Comm_group(MPI_COMM_WORLD,&MPI_GROUP_WORLD);
	MPI_Group_excl(MPI_GROUP_WORLD,1,excl_ranks,&only_workers_group);
	MPI_Comm_create(MPI_COMM_WORLD,only_workers_group,&only_workers_comm);
	MPI_Group_free(&only_workers_group);
	MPI_Group_free(&MPI_GROUP_WORLD);
****/
	return 0;
   }
#else		/************************************************************/
   {
	int    info,bid,send_no_nodes=no_nodes;

	spawn_pvm_task(no_nodes);

	/* for easy handling we put the host tid as n+1tid into node_tids */
	node_tids[no_nodes] = host_tid;

	/* each node gets to know how many (without host node) will participate */
	bid  = pvm_initsend(encoding);
	info = pvm_pkint(&send_no_nodes,1,1);
	info |= pvm_pkint(node_tids, no_nodes + 1, 1);
	info |= pvm_pkint(&encoding, 1, 1);
	info |= pvm_mcast(node_tids, no_nodes, TNPROCS);

	my_pvm_freebuf(bid);
	if (info<0) {
		errout("Fatal error: pvm_mcast failed in PCI_master_setup ! info=%d\n",info);
		pex(-1);
 	}
	return 0;
  }
#endif
}


/************************************************************************/
void   PCI_exit(int action) /*** called on the master (or, if MPI, also on the nodes) ***/
{
#if defined(PARIX) || defined(PARIXHAL) /********************************/
	pvm_free();
	myfree(StarNodeId);
	if (ProcNum==0) {
		TimeWait(TimeNow()+CLOCK_TICK* (unsigned int) 3);
		AbortServer(0);
	}
	
#elif defined(MPI) /****************************************/
#if (SLOW_MPI==0) && ((DEBUG & 2) == 2)
	errout("PCI_exit: MPI\n");
	MPI_Buffer_detach(&dptr,&di);
	errout("PCI_exit: MPI, after MPI_Buffer_detach\n");
	myfree(mpi_send_buffer);
	errout("PCI_exit: MPI, after myfree\n");
#endif
	SLEEP((unsigned int) 10);	/* feeble attempt to silence all communication */
	pvm_free();	/* free send and receive buffers of EGO packing functions */
#if ((DEBUG & 2) == 2)
	errout("Node %d in PCI_exit: MPI after pvm_free\n", really_my_number);
#endif
	if (action == SIGTERM)
	{
	  MPI_Abort(MPI_COMM_WORLD,action); /* MPI_Abort() does send SIGTERM to all other nodes,
	  				        but then it does not return! This may be because 
					        we handle SIGTERM ourselves already. But if we don't 
					        call MPI_Abort() then the other nodes don't know about
					        the signal and don't die. And worse: even the node that
					        received SIGTERM doesn't die! Sigh.
						Ah! I just found out that our mendling with the
						signals forced MPI_Abort() to no longer work,
						as it uses SIGTERM itself! So I reduced our signal
						handling to the two user signals 
						SIGUSR1 and SIGUSR2.*/
	}
	else
	  MPI_Finalize();	/* after this point no more MPI calls can be made. 
			   	   BUT: it must be assured that all communication 
				   has ceased BEFORE calling this routine. 
				   How shall we do this in case of a fatal error? 
				   On the SP2 MPI_Finalize() does not return.*/
#elif defined(T3D) /****************************************/
	myfree(node_tids);
	pvm_exit();
#else		   /****************************************/
	int    i;
	/* here we kill all the node processes since they would run for ever! */
	for (i=0;i<num_compnodes;i++) {
		pvm_kill(node_tids[i]);
	}
	myfree(node_tids);
	pvm_exit();
#endif
#if ((DEBUG & 2) == 2)
	errout("Node %d in PCI_exit: MPI after MPI_Finalize/Abort\n", really_my_number);
#endif
	return;
}

/******************************************************************/
void   PCI_node_cleanup(void) /*** called on the nodes ***/
{
#if defined(PVM)
	myfree(node_tids);
	pvm_exit();
#endif
}

/******************************************************************/
int    PCI_enroll(void)
{
	int myid=0;
#if defined(PARIX) || defined(PARIXHAL)
	myid = (GET_ROOT()->ProcRoot->MyProcID);
#elif defined(MPI)
	MPI_Comm_rank(MPI_COMM_WORLD,&myid);
#elif (defined(T3D) && !defined(VPP))
	myid = pvm_get_PE(pvm_mytid());
#else
	myid = pvm_mytid();
#endif

#if ((DEBUG & 1) == 1)
txtout("In PCI_enroll myid %d\n",myid);
#endif
	return myid;
}

/**************************************************************************/
/*** PCI_node_setup is called only once on the nodes during setup 	***/
/*** and corresponds to PCI_master_setup which is called on the master. ***/
/**************************************************************************/
void   PCI_node_setup(int MyProcID, int *no_nodes, int *my_number)
{
#if defined(PARIX) || defined(PARIXHAL) /*****************************************/
   {
	int    i,b,gap,my_processornum=MyProcID;

  	/* we need to know how many procs we will use */
	if (RecvNode(0,TNONODES,(void *)no_nodes,sizeof(int))<0) {
		errout("Node %d: PCI_node_setup failed #1\n",MyProcID);
		pex(-1);
	}

	*my_number = my_processornum-1;
	SendNode(0,TMYNUMBER,(void*)&my_processornum,sizeof(int));

	StarNodeId = alloc_ints(*no_nodes+1);
	RecvNode(0,TNODEIDS,(void*)StarNodeId,sizeof(int)*(*no_nodes));
	StarNodeId[*no_nodes] = 0;
	really_my_number=*my_number;
	
#if defined(RS6K)
   	/*** link to the master node ***/
	masterstar_top     = NewTop(1);  	

  	masternode_link[0] = AddTop(masterstar_top,ConnectLink(0,0,&b));

  	/*** now make links only between nodes 				***/
  	/*** star-topologie: links from one node to all other nodes	***/
	nodestar_top = NewTop(*no_nodes-1);  

  	for (i=1,gap=0;i<=*no_nodes;i++) {
  		if (i-1==*my_number) { gap=1; continue; }
  		nodenode_CB[i-gap-1]   = ConnectLink(i,1,&b);
  		if (nodenode_CB[i-gap-1]==NULL) {
  			errout("Node %d: ConnectLink to node %d failed.\n",*my_number,i-1);
  			pex(-1);
  		}
  		
  		nodenode_link[i-gap-1] = AddTop(nodestar_top,nodenode_CB[i-gap-1]);
		nodenode_opt[i-gap-1]  = ReceiveOption(nodenode_CB[i-gap-1]);
	}

#endif
	return;
   }
   
#elif defined(MPI) /*****************************************/
   
   {
	int excl_ranks[2];
	MPI_Status status;

	MPI_Recv(no_nodes,1,MPI_INT,0,TNONODES,MPI_COMM_WORLD,&status);
/***
	excl_ranks[0]=0;
	MPI_Comm_group(MPI_COMM_WORLD,&MPI_GROUP_WORLD);
	MPI_Group_excl(MPI_GROUP_WORLD,1,excl_ranks,&only_workers_group);
	MPI_Comm_create(MPI_COMM_WORLD,only_workers_group,&only_workers_comm);
	MPI_Group_free(&only_workers_group);
	MPI_Group_free(&MPI_GROUP_WORLD);
***/
	*my_number     = MyProcID-1;
	really_my_number= *my_number;
   }	
#else 	/******************* PVM ***********************/
   {	int    bid,info,i;

	/* we need to know how many procs we will use */
	bid = pvm_recv(-1,TNPROCS);
	if (bid<0) {
		txtout("PCI_node_setup: Receiving failed! info=%d\n",bid);
		pex(-1);
	}
	info = pvm_upkint(no_nodes, 1, 1);
	if (info<0) {
		txtout("PCI_node_setup: Unpacking failed! info=%d\n",info);
		pex(-1);
	}
	node_tids = alloc_ints(*no_nodes + 1);

#if defined(T3D)
	/*** In this special case, the nodes loaded but not participating ***/
	/*** have to exit NOW. ***/
	if (MyProcID > *no_nodes)	pex(0);
#endif

	info = pvm_upkint(node_tids,*no_nodes+1, 1);
	info = pvm_upkint(&encoding,1, 1);
	my_pvm_freebuf(bid);

	*my_number = -1;
	for (i = 0; i < *no_nodes; i++) {
		if (MyProcID == node_tids[i]) *my_number = i;
	}
	if (*my_number == -1) {
		errout("PCI_node_setup could not identify my own node ID!\n");
		pex(-1);
	}
	really_my_number=*my_number;
	return;
   }
#endif
}

static
int my_pvm_freebuf(int ido)
{
#if defined(PARIX) || defined(PARIXHAL) || defined(MPI)
	int    a=ido;a=0;
	check_checksum();
	return (0);
#else
	return (pvm_freebuf(ido));
#endif
}

/************************************************************************/
void   PCI_send_msg_one_node(LOAD_MSG_ONE_NODE *start_msg, int ring_node)
{
	int    info,bid;

	bid  = pvm_initsend(encoding);
	info = pk_LOAD_MSG_ONE_NODE(start_msg);
	if (info!=0) {
		errout("Packing failed in PCI_send_msg_one_node! info=%d\n",info);
		pex(-1);
	}

#if defined(PARIX) || defined(PARIXHAL)
	len = pvm_buffer(&ptr);

	if (SendNode(StarNodeId[ring_node],TMYDATA,(byte*)&len,sizeof(int))<0) {
		errout("Master: PCI_send_msg_one_node failed to Node %d\n",ring_node);
		pex(-1);
	}
	if (SendNode(StarNodeId[ring_node],TMYDATA,ptr,len)<0) {
		errout("Master: PCI_send_msg_one_node failed to Node %d\n",ring_node);
		pex(-1);
	}
#elif defined(MPI)
	len = pvm_buffer(&ptr);
	MPI_Send(ptr,len,MPI_BYTE,ring_node+1,TMYDATA,MPI_COMM_WORLD);
#else
	info = pvm_send(node_tids[ring_node],TMYDATA);
#endif

	my_pvm_freebuf(bid);
	if (info!=0) {
		errout("Sending failed in PCI_send_msg_one_node ! info=%d\n",info);
		pex(-1);
	}
	
	return;
}


/******************************************************************/
void   PCI_recv_load_msg_one_node(int nProcs,LOAD_MSG_ONE_NODE *my_data)
{
	int    bid=0,info=0;
	
#if defined(PARIX) || defined(PARIXHAL)
	{int len;
	void *ptr;
	int    a = nProcs; a = 0;

	if (RecvNode(0,TMYDATA,(byte*)&len,sizeof(int))<0) {
		errout("Node ?: PCI_recv_load_msg_one_node failed #1\n");
		pex(-1);
	}

	ptr = pvm_set_readbuf(len);
	if (RecvNode(0,TMYDATA,ptr,len)<0) {
		errout("Node ?: PCI_recv_load_msg_one_node failed #2\n");
		pex(-1);
	}
	}
#elif defined(MPI)
   {	MPI_Status status;

	MPI_Probe(0,TMYDATA,MPI_COMM_WORLD,&status);
	MPI_Get_count(&status,MPI_BYTE,&len);

	ptr = pvm_set_readbuf(len);
	MPI_Recv(ptr,len,MPI_BYTE,0,TMYDATA,MPI_COMM_WORLD,&status);
   }
#else
	bid = pvm_recv(node_tids[nProcs],TMYDATA);
#endif
	info |= upk_LOAD_MSG_ONE_NODE(my_data);
	if (info!=0) {
		errout("PCI_recv_load_msg_one_node() failed!\n");
		pex(-1);
	}
	my_pvm_freebuf(bid);
	return;
}

/************************************************************************/
void   PCI_send_update(LOAD_MSG_ALL_NODE *global_data)
{
	int    info,bid;

	bid  = pvm_initsend(encoding);
	info = pk_ALL_NODE_UPDATE(global_data);
	if (info!=0) {
		errout ("PCI_send_update info=%d\n", info);
		pex (-1);
	}

#if defined(PARIX) || defined(PARIXHAL)
    {
	int    ring_node;
	len = pvm_buffer(&ptr);
	for (ring_node=0;ring_node<global_data->no_nodes;ring_node++) {
		if (SendNode(StarNodeId[ring_node],TGLOBALDATA,(byte*)&len,sizeof(int))<0) {
			errout("Master: PCI_send_update failed to node %d.\n",ring_node);
			pex(-1);
		}
		if (SendNode(StarNodeId[ring_node],TGLOBALDATA,ptr,len)<0) {
			errout("Master: PCI_send_update failed to node %d.\n",ring_node);
			pex(-1);
		}
	}
    }
#elif defined(MPI)
    {
	int    ring_node;
	len = pvm_buffer(&ptr);
	for (ring_node=0;ring_node<global_data->no_nodes;ring_node++) {
		MPI_Ssend(ptr,len,MPI_BYTE,ring_node+1,TGLOBALDATA,MPI_COMM_WORLD);
	}
    }
#else
	info = pvm_mcast(node_tids,global_data->no_nodes,TGLOBALDATA);
#endif

	my_pvm_freebuf(bid);
	if (info!=0) {
		errout("PCI_send_update failed! info=%d\n",info);
		pex(-1);
	}
	return;
}

/******************************************************************/
void   PCI_recv_update(int nProcs,LOAD_MSG_ALL_NODE *global_data)
{
	int    bid = 0,info=0;

#if defined(PARIX) || defined(PARIXHAL)
	int    a = nProcs;a = 0;
	if (RecvNode(0,TGLOBALDATA,(byte*)&len,sizeof(int))<0) {
		errout("Node ?: PCI_recv_update failed #1\n");
		pex(-1);
	}
	ptr = pvm_set_readbuf(len);
	if (RecvNode(0,TGLOBALDATA,ptr,len)<0) {
		errout("Node ?: PCI_recv_update failed #2\n");
		pex(-1);
   	}
#elif defined(MPI)
   {	MPI_Status status;

	MPI_Probe(0,TGLOBALDATA,MPI_COMM_WORLD,&status);
	MPI_Get_count(&status,MPI_BYTE,&len);

	ptr = pvm_set_readbuf(len);
	MPI_Recv(ptr,len,MPI_BYTE,0,TGLOBALDATA,MPI_COMM_WORLD,&status);
   }
#else
	bid = pvm_recv(node_tids[nProcs],TGLOBALDATA);	/* this is the HOST tid! */
#endif

	info = upk_ALL_NODE_UPDATE(global_data);
	my_pvm_freebuf(bid);

	if (info!=0) {
		errout ("PCI_recv_update() failed!\n");
		pex (-1);
	}
	return;
}

/************************************************************************/
void   PCI_send_msg_all_node (LOAD_MSG_ALL_NODE *global_data)
{
	int    info,bid;

	bid  = pvm_initsend(encoding);
	info = pk_LOAD_MSG_ALL_NODE(global_data);
	if (info!=0) {
		errout("Packing of global_data failed! info=%d\n", info);
		pex(-1);
	}
#if defined(PARIX) || defined(PARIXHAL)
   {
	int    ring_node;
	len = pvm_buffer(&ptr);
	for (ring_node = 0; ring_node<global_data->no_nodes; ring_node++) {
		if (SendNode(StarNodeId[ring_node],TGLOBALDATA,(byte*)&len,sizeof(int))<0) {
			errout("Master: PCI_send_update failed to Node %d\n",ring_node);
			pex(-1);
		}
		if (SendNode(StarNodeId[ring_node],TGLOBALDATA,ptr,len)<0) {
			errout("Master: PCI_send_update failed to Node %d\n",ring_node);
			pex(-1);
		}
	}
   }
#elif defined(MPI)
    {
	int    ring_node;
	len = pvm_buffer(&ptr);
/****
	MPI_Bcast(&len,1,MPI_INT,0,MPI_COMM_WORLD);
	MPI_Bcast(ptr,len,MPI_BYTE,0,MPI_COMM_WORLD);
****/
	for (ring_node=0;ring_node<global_data->no_nodes;ring_node++) {
		MPI_Ssend(ptr,len,MPI_BYTE,ring_node+1,TGLOBALDATA,MPI_COMM_WORLD);
	}
   }
#else
	info = pvm_mcast(node_tids, global_data->no_nodes, TGLOBALDATA);
#endif

	my_pvm_freebuf(bid);
	if (info!=0) {
		errout("Sending of global_data failed! info=%d\n",info);
		pex (-1);
	}
	return;
}


/******************************************************************/
void   PCI_recv_load_msg_all_node(int nProcs,LOAD_MSG_ALL_NODE *global_data)
{
	int    bid=0,info=0;

#if defined(PARIX) || defined(PARIXHAL)
	int    a=nProcs;a=0;
	if (RecvNode(0,TGLOBALDATA,(byte*)&len,sizeof(int))<0) {
		errout("Node ?: PCI_recv_update failed #1\n");
		pex(-1);
	}
	ptr = pvm_set_readbuf(len);
	if (RecvNode(0,TGLOBALDATA,ptr,len)<0) {
		errout("Node ?: PCI_recv_update failed #2\n");
		pex(-1);
	}
#elif defined(MPI)
   {	MPI_Status status;

	MPI_Probe(0,TGLOBALDATA,MPI_COMM_WORLD,&status);
	MPI_Get_count(&status,MPI_BYTE,&len);

	ptr = pvm_set_readbuf(len);
	MPI_Recv(ptr,len,MPI_BYTE,0,TGLOBALDATA,MPI_COMM_WORLD,&status);
   }
#else
	bid = pvm_recv(node_tids[nProcs],TGLOBALDATA);	/* this is the HOST tid! */
	if (bid<0) {
		errout("PCI_recv_load_msg_all_node() failed in pvm_recv %d!\n",bid);
		pex(-1);
	}
#endif

	info = upk_LOAD_MSG_ALL_NODE(global_data);
	my_pvm_freebuf(bid);

	if (info!=0) {
		errout("PCI_recv_load_msg_all_node() failed!\n");
		pex(-1);
	}
	return;
}


/******************************************************************/
static
int    spawn_pvm_task(int no_nodes)
{
#if defined(PARIX) || defined(PARIXHAL) || defined(MPI)
	int b = no_nodes;
	b = 0;			/* suppress warnings */
	return 0;
/**********************************/
#elif (defined(T3D) && !defined(VPP))
	int    i,info,bid,dummy_no_nodes=no_nodes;

	/*** no_nodes is the number of nodes which we will really use. ***/
	/*** It is guaranteed to be <= the available number of nodes.  ***/

	node_tids = alloc_ints (no_nodes + 1);
	for (i = 0; i < no_nodes; i++) node_tids[i] = i + 1;

	/*** Now we have to give this info to the non-participating nodes, too. ***/
 	/*** Very awkward, uggh. ***/

	bid = pvm_initsend(encoding);
	info = pvm_pkint(&dummy_no_nodes,1,1);
	if (info < 0) 
	{
		errout("Packing failed! info=%d\n", info);
		pex (-1);
	}
	info = info | pvm_pkint(node_tids, no_nodes + 1, 1);
	info = info | pvm_pkint(&encoding, 1, 1);
	for (i = no_nodes + 1; i < _MPP_N_PES; i++) 
	{
		pvm_send (i, TNPROCS);
	}
/**********************************/
#elif defined(VPP)
	int    i,info,bid;

	/*** no_nodes is the number of nodes which we will really use (excluding the master!). ***/
	/*** It is guaranteed to be < the available number of nodes.  ***/

	node_tids = alloc_ints(no_nodes + 1);
	
	/* This is now special for the VPP: */
        pvm_spawn(NODEFILENAME,NULL,0, "",no_nodes,node_tids);
/***********************************/
#else
	int    numt,nhost,narch;

	node_tids = alloc_ints(no_nodes + 1);

	/****** first we find out about the dimensions we might need ******/
	pvm_config(&nhost,&narch,NULL);

	numt = pvm_spawn(NODEFILENAME,NULL,0,NULL,no_nodes,node_tids);
	if (numt != no_nodes) {
		errout ("Fatal error: Can only start %d out of %d processes. Aborting...\n",
			numt, no_nodes);
		pex (-1);
	}
#if ((DEBUG & 1) == 1)
	txtout("In pvm_spawn_task: NumTaks=%d NumHost=%d NumArch=%d\n",numt,nhost,narch);
#endif
	
#if OLDSTUFF
	int    i,j,names_match = FALSE,found,pvmd_tid,ntask,k,l;
	struct pvmhostinfo **hostp;
	struct pvmtaskinfo **taskp;
	
	/****** now we reserve the right amount of memory *******/
	hostp = alloc_pvmhostinfo(nhost);

	l = 3;				/* look at most 3 times */
	SLEEP(2);			/* wait till everything is settled */
	while (l>0) {
		/* first we find out about the dimensions we might need */
		pvm_tasks(0,&ntask,NULL);
		taskp = alloc_pvmtaskinfo(ntask);
		pvm_tasks(0,NULL,taskp);
		for (i = 0; i < no_nodes; i++)  {/* over all MY tasks */
			j = 0;
			while (j<ntask) {	/* over ALL tasks */
	  			if ((*taskp)[j].ti_tid == node_tids[i]) {
	    				/* now we have the id of the deamon, next we search the host */
	    				pvmd_tid = (*taskp)[j].ti_host;
	    				k = 0;
	    				while (k<nhost) {	/* over all hosts */
	      					if ((*hostp)[k].hi_tid == pvmd_tid) {
									/* now we found the right host! */
							k = nhost + 2;	/* stop the k loop and flag "found host" */
	      					}
	      					k++;
	    				}
	    				if (k==nhost){
	      					errout("Fatal error: Host ID not found in spwan_pvm_task!\n");
	      					pex (-1);
	    				}
	    				j = ntask + 2;	/* stop the j loop and flag "found task" */
				} /* end of if task found */
	  			j++;
			} /* end of while j loop */
	
			if (j == ntask) {
	 	 		/* we didn't find the task ID, probably because the network was
  		  		* too slow. So we try again. */
	  			SLEEP (1);		/* give the network time */
	  			i = no_nodes + 1;	/* don't have to look for the other IDs in
					 		* this run any more, but set flag to look
					 		* again */
	  			l--;			/* don't look for ever!! */
			}
      		}	/* for i loop */
      		if (i==no_nodes){
				/* we found all task IDs. Done (don't look again). */
			l = -1;	/* don't look again and flag "found all" */
      		} else {
			free_pvmtaskinfo(taskp,ntask);/* we will realloc, so free here */
      		}
	} /* while l loop */
	
    	if (l==0) {
      		errout("Fatal error: Task IDs not match in spwan_pvm_task!\n");
      		pex (-1);
    	}
    	free_pvmtaskinfo(taskp,ntask);
	free_pvmhostinfo(hostp,nhost);	
#endif /*** end of OLDSTUFF ***/
#endif
	return 0;
}

/*** recv data from master at every node ***/
void PCI_etcdata_recv(	int no_nodes,
			int my_number,
			ENERGY_INFO *local_energy,
			LOAD_MSG_ONE_NODE *my_data,
			LOAD_MSG_ALL_NODE *global_data)
{
	int bid=0;

#if ((DEBUG & 1) == 1)
txtout("Master: In start of PCI_etcdata_recv\n");
#endif
		
#if defined(PARIX) || defined(PARIXHAL)
   {	int 	infobuf[5],info=0;
	void	*ptr;
#if defined(RS6K)
	info=RecvNode(StarNodeId[no_nodes],TETCDATA,(byte*)infobuf,sizeof(int));
	ptr = pvm_set_readbuf(infobuf[0]);
	info=RecvNode(StarNodeId[no_nodes],TETCDATA,(byte*)ptr,infobuf[0]);
#else
	info=RecvNode(StarNodeId[no_nodes],TETCDATA,(byte*)infobuf,sizeof(int));
	ptr = pvm_set_readbuf(infobuf[0]);
	info=RecvNode(StarNodeId[no_nodes],TETCDATA,(byte*)ptr,infobuf[0]);
#endif
	if (info<0) {
		errout("Node %d:PCI_etcdata_recv failed\n",my_number);
		pex(-1);
	}
   }

#elif defined(MPI)
   {	MPI_Status status;

	MPI_Probe(MPI_ANY_SOURCE,TETCDATA,MPI_COMM_WORLD,&status);
	MPI_Get_count(&status,MPI_BYTE,&len);

	ptr = pvm_set_readbuf(len);
	MPI_Recv(ptr,len,MPI_BYTE,status.MPI_SOURCE,TETCDATA,MPI_COMM_WORLD,&status);
   }
#else
   {
	bid = pvm_recv(node_tids[no_nodes],TETCDATA);
   }
#endif
	upk_ENERGY_INFO(local_energy,no_nodes);
	my_pvm_freebuf(bid);
#if ((DEBUG & 1) == 1)
txtout("Master: PCI_etcdata_recv passed\n");
#endif	
	return;	
}

/***** send some special data from the master to the nodes ***/
void   PCI_etcdata_send(ENERGY_INFO *local_energy,
			LOAD_MSG_ALL_NODE *global_data)
{
	int bid=0,len=0;
	bid = pvm_initsend(encoding);

#if ((DEBUG & 1) == 1)
txtout("Master: In start of PCI_etcdata_send\n");
#endif

	pk_ENERGY_INFO(local_energy,global_data->no_nodes);

#if defined(PARIX) || defined(PARIXHAL)
    {
	int    ring_node;
	len = pvm_buffer(&ptr);
	for (ring_node=0;ring_node<global_data->no_nodes;ring_node++) {
		if (SendNode(StarNodeId[ring_node],TETCDATA,(byte*)&len,sizeof(int))<0) {
			errout("Master: PCI_etcdata_send failed to node %d.\n",ring_node);
			pex(-1);
		}
		if (SendNode(StarNodeId[ring_node],TETCDATA,ptr,len)<0) {
			errout("Master: PCI_etcdata_send failed to node %d.\n",ring_node);
			pex(-1);
		}
	}
    }
#elif defined(MPI)
    {
	int    ring_node;
	len = pvm_buffer(&ptr);
	for (ring_node=0;ring_node<global_data->no_nodes;ring_node++) {
		MPI_Ssend(ptr,len,MPI_BYTE,ring_node+1,TETCDATA,MPI_COMM_WORLD);
	}
    }
#else
	pvm_mcast(node_tids,global_data->no_nodes,TETCDATA);
#endif

#if ((DEBUG & 1) == 1)
txtout("Master: PCI_etcdata_send passed\n");
#endif
	my_pvm_freebuf(bid);
	return;
}


/****************************************************************/

void   PCI_positions_send(LOAD_MSG_ALL_NODE *global_data)
{
	int bid=0,len=0;
	bid = pvm_initsend(encoding);

#if ((DEBUG & 1) == 1)
txtout("Master: In start of PCI_positions_send\n");
#endif

	pk_POSITIONS_INFO(global_data);

#if defined(PARIX) || defined(PARIXHAL)
    {
	int    ring_node;
	len = pvm_buffer(&ptr);
	for (ring_node=0;ring_node<global_data->no_nodes;ring_node++) {
		if (SendNode(StarNodeId[ring_node],TPOSITIONS,(byte*)&len,sizeof(int))<0) {
			errout("Master: PCI_positions_send failed to node %d.\n",ring_node);
			pex(-1);
		}
		if (SendNode(StarNodeId[ring_node],TPOSITIONS,ptr,len)<0) {
			errout("Master: PCI_positions_send failed to node %d.\n",ring_node);
			pex(-1);
		}
	}
    }
#elif defined(MPI)
    {
	int    ring_node;
	len = pvm_buffer(&ptr);
	for (ring_node=0;ring_node<global_data->no_nodes;ring_node++) {
		MPI_Ssend(ptr,len,MPI_BYTE,ring_node+1,TPOSITIONS,MPI_COMM_WORLD);
	}
    }
#else
	pvm_mcast(node_tids,global_data->no_nodes,TPOSITIONS);
#endif

#if ((DEBUG & 1) == 1)
txtout("Master: PCI_positions_send passed\n");
#endif
	my_pvm_freebuf(bid);
	return;
}

/****************************************************************/
/*** recv data from master at every node ***/
void PCI_positions_recv(int no_nodes,
			int my_number,
			LOAD_MSG_ALL_NODE *global_data)
{
	int bid=0;

#if ((DEBUG & 1) == 1)
txtout("Master: In start of PCI_positions_recv\n");
#endif
		
#if defined(PARIX) || defined(PARIXHAL)
   {	int 	infobuf[5],info=0;
	void	*ptr;
#if defined(RS6K)
	info=RecvNode(StarNodeId[no_nodes],TPOSITIONS,(byte*)infobuf,sizeof(int));
	ptr = pvm_set_readbuf(infobuf[0]);
	info=RecvNode(StarNodeId[no_nodes],TPOSITIONS,(byte*)ptr,infobuf[0]);
#else
	info=RecvNode(StarNodeId[no_nodes],TPOSITIONS,(byte*)infobuf,sizeof(int));
	ptr = pvm_set_readbuf(infobuf[0]);
	info=RecvNode(StarNodeId[no_nodes],TPOSITIONS,(byte*)ptr,infobuf[0]);
#endif
	if (info<0) {
		errout("Node %d:PCI_etcdata_recv failed\n",my_number);
		pex(-1);
	}
   }

#elif defined(MPI)
   {	MPI_Status status;

	MPI_Probe(MPI_ANY_SOURCE,TPOSITIONS,MPI_COMM_WORLD,&status);
	MPI_Get_count(&status,MPI_BYTE,&len);

	ptr = pvm_set_readbuf(len);
	MPI_Recv(ptr,len,MPI_BYTE,status.MPI_SOURCE,TPOSITIONS,MPI_COMM_WORLD,&status);
   }
#else
   {
	bid = pvm_recv(node_tids[no_nodes],TPOSITIONS);
   }
#endif
	upk_POSITIONS_INFO(global_data);
	my_pvm_freebuf(bid);
#if ((DEBUG & 1) == 1)
txtout("Master: PCI_positions_recv passed\n");
#endif	
	return;	
}


/******************************************************************/
/*** send energy-info and forces from every node to the master	***/
/******************************************************************/
void   PCI_energy_send(		int no_nodes,
				int my_number,
				ENERGY_INFO *local_energy,
				LOAD_MSG_ONE_NODE *my_data,
				LOAD_MSG_ALL_NODE *global_data)
{
	int bid=0,len=0;
	bid = pvm_initsend(encoding);
	len = my_number;

	pvm_pkint(&len,1,1);
	pk_ENERGY_INFO(local_energy,no_nodes);
	pk_FORCE_INFO(my_number,my_data,global_data);

#if ((DEBUG & 1) == 1)
txtout("Node %d: In start of PCI_energy_send\n",my_number);
#endif

#if defined(PARIX) || defined(PARIXHAL)
  	{int	infobuf[5],info=0;
	void 	*ptr;
	
	len        = pvm_buffer(&ptr);
	infobuf[0] = my_number;
	infobuf[1] = len;

	if(StarNodeId[no_nodes]!=0){	/*** StarNodeId[no_nodes] is the master's id ***/
	     txtout("PCI_ENERGY SEND: wrong id..\n");
	     pex(-1);
	}
#if SAVE_COM
	info=SendNode(StarNodeId[no_nodes],TENERGY,infobuf,2*sizeof(int));
	info=SendNode(StarNodeId[no_nodes],TENERGY,(byte*)ptr,len);
#else	   
	info=e_SendNode(StarNodeId[no_nodes],TENERGY,infobuf,2*sizeof(int));
	info=e_SendNode(StarNodeId[no_nodes],TENERGY,(byte*)ptr,len);	
#endif
	if (info<0) {
		errout("PCI_energy_send failed\n");
		pex(-1);
	}
   	}
#elif defined(MPI)
   	{void 	*ptr;

	len = pvm_buffer(&ptr);
	MPI_Ssend(ptr,len,MPI_BYTE,0,TENERGY,MPI_COMM_WORLD);
	}
#else
	pvm_send(node_tids[no_nodes],TENERGY); /***send to master ***/
#endif

#if ((DEBUG & 1) == 1)
txtout("Node %d: PCI_energy_send passed\n",my_number);
#endif
	my_pvm_freebuf(bid);
	return;
}

/******************************************************************/
/*** get local energy and forces from every node to the master	***/
/******************************************************************/
void   PCI_energy_recv(	int recv_node,
			ENERGY_INFO *energy,
			int no_nodes,
			int *other_number,
			STRUCT_FORCES *my_forces)
{
	int	bid=recv_node;
	bid = 0;
#if ((DEBUG & 1) == 1)
txtout("Master: PCI_energy_recv entered recv_node=%d\n",recv_node);
#endif
	
#if defined(PARIX) || defined(PARIXHAL) /***********************************************/
   {	int 	len,infobuf[5],info=0;
	void	*ptr;
#if defined(RS6K)
#if SAVE_COM
	info=RecvNode(StarNodeId[recv_node],TENERGY,(byte*)infobuf,2*sizeof(int));
	*other_number = infobuf[0];	/*** should be equal to recv_node ***/
	len           = infobuf[1];
	ptr = pvm_set_readbuf(len);
	info=RecvNode(StarNodeId[*other_number],TENERGY,(byte*)ptr,len);
#else
	info=e_RecvNode(-1,TENERGY,(byte*)infobuf,2*sizeof(int));
	*other_number = infobuf[0];
	len           = infobuf[1];
	ptr = pvm_set_readbuf(len);
	info=e_RecvNode(StarNodeId[*other_number],TENERGY,(byte*)ptr,len);
#endif
#else
	info=RecvNode(-1,TENERGY,(byte*)infobuf,2*sizeof(int));
	*other_number = infobuf[0];
	len           = infobuf[1];
	ptr = pvm_set_readbuf(len);
	info=RecvNode(StarNodeId[*other_number],TENERGY,(byte*)ptr,len);
#endif

	if (info<0) {
		errout("PCI_energy_recv #1 failed %d",info);
		pex(-1);
	}

   }
#elif defined(MPI) /***********************************************/
   {	MPI_Status status;

	MPI_Probe(MPI_ANY_SOURCE,TENERGY,MPI_COMM_WORLD,&status);
	MPI_Get_count(&status,MPI_BYTE,&len);
	*other_number = status.MPI_SOURCE-1;

	ptr = pvm_set_readbuf(len);
	MPI_Recv(ptr,len,MPI_BYTE,*other_number+1,TENERGY,MPI_COMM_WORLD,&status);
   }
#else		   /***********************************************/
	bid = pvm_recv(-1,TENERGY);
#endif

	pvm_upkint(other_number,1,1);
#if ((DEBUG & 1) == 1)
txtout("Master: PCI_energy_recv passed from %d\n",*other_number);
#endif
	upk_ENERGY_INFO(energy,no_nodes);
	upk_FORCE_INFO (my_forces);

	my_pvm_freebuf(bid);
	return;
}

/******************************************************************/
void   PCI_send_qmforces(LOAD_MSG_ALL_NODE *global_data,
			 STRUCT_QM_INFO *finfo)
{
	int tonode,i,check,info=0,count=0;

	for (tonode=0,check=0;tonode<global_data->no_nodes;tonode++) {
		int bid = pvm_initsend(encoding);

		for (i=0,count=0;i<finfo->len;i++) {
			if (finfo->node[i]!=tonode) continue;
			pvm_pkint(finfo->atom_id+i,1,1);	
			pvm_pkint(finfo->nodeatomi+i,1,1);
			pvm_pkdouble(finfo->qm_q+i,1,1);
			pvm_pkdouble(finfo->qm_force+3*i,3,1);
			check++;
			count++;
		}
		if (count==0) {
			count=-1;
			pvm_pkint(&count,1,1);
		}
#if defined(PARIX) || defined(PARIXHAL)
		{int len;
		void *ptr;
		len = pvm_buffer(&ptr);

		SendNode(StarNodeId[tonode],TQMFORCE,(byte*)&len,sizeof(int));
		SendNode(StarNodeId[tonode],TQMFORCE,ptr,len);
		}
#elif defined(MPI)
		{int len;
		void *ptr;
		len = pvm_buffer(&ptr);

		MPI_Send(ptr,len,MPI_BYTE,tonode+1,TQMFORCE,MPI_COMM_WORLD);
		}
#else
		info = pvm_send(node_tids[tonode],TQMFORCE);
#endif
		if (info) {
			errout("Sending failed in PCI_send_qmforces.\n");
			pex(-1);
		}
		my_pvm_freebuf(bid);
	}

	if (check!=finfo->len) {
		errout("Fatal error: Something wrong in PCI_send_qmforces %d %d.\n",
			check,finfo->len);
		pex(-1);
	}
	return;
}

void   PCI_recv_qmforces(int nProcs,
			 LOAD_MSG_ALL_NODE *global_data,
			 LOAD_MSG_ONE_NODE *my_data,
			 int my_number)
{
	int    bid=0,i=global_data->nr_qmatoms,limit=nProcs;
#if defined(PARIX) || defined(PARIXHAL)
	{int len;
	void *ptr;

	RecvNode(0,TQMFORCE,(byte*)&len,sizeof(int));
	ptr = pvm_set_readbuf(len);
	RecvNode(0,TQMFORCE,ptr,len);
	}
#elif defined(MPI)
   	{MPI_Status status;

	MPI_Probe(0,TQMFORCE,MPI_COMM_WORLD,&status);
	MPI_Get_count(&status,MPI_BYTE,&len);

	ptr = pvm_set_readbuf(len);
	MPI_Recv(ptr,len,MPI_BYTE,0,TQMFORCE,MPI_COMM_WORLD,&status);
   	}
#else
	bid = pvm_recv(node_tids[nProcs],TQMFORCE);
#endif

	/**************************************************/
	/*** now we unpack what we got 			***/
	/**************************************************/	
	limit = max(my_data->qmlocal.cur_len,1);
	for (i=0;i<limit;i++) {
		int atom_id,j;

		pvm_upkint(&atom_id,1,1);
		if (atom_id==-1) break;

		j = get_qmnode_index(atom_id,my_data);
		if (j>=limit) {
			errout("Node %d: Fatal error (%d) in PCI_recv_qmforces.\n",
				my_number,j);
			pex(-1);
		}
		pvm_upkint   (my_data->qmlocal.atomi+j,1,1);
		pvm_upkdouble(my_data->qmlocal.qm_q+j,1,1);
		pvm_upkdouble(my_data->qmlocal.qm_force+3*j,3,1);	
	}

	my_pvm_freebuf(bid);
	return;
}

/******************************************************************/
#define DBUF_SIZE (12+MXNO_SBOUND_PARAM)
int   PCI_send_scaling(double scaling_factor,LOAD_MSG_ALL_NODE *global_data)
{
	double	dbuf[DBUF_SIZE];
	int i,res=0;

	dbuf[0] = scaling_factor;
	dbuf[1] = global_data->deltaimob[0];
	dbuf[2] = global_data->deltaimob[1];
	dbuf[3] = global_data->deltaimob[2];
	dbuf[4] = global_data->deltaimob[3];
	dbuf[5] = global_data->deltaimob[4];
	dbuf[6] = global_data->deltaimob[5];
	dbuf[7] = global_data->deltaimob[6];
	dbuf[8] = global_data->deltaimob[7];
	dbuf[9] = global_data->deltaimob[8];
	dbuf[10] = (double)global_data->final_cycle_no;
	
	if (get_terminate_flag()==AFTER_STEP) {
		dbuf[11] = -1.0; /*** stop signal ***/
		res=-1;
	} else {
		dbuf[11] = 1.0;	/*** continue signal ***/
		res=1;
	}
	
	/* for(i=0;i<6;i++)
        	dbuf[i+6] = global_data->pullspring[i]; */
        
	for(i=0;i<MXNO_SBOUND_PARAM;i++)
	  	dbuf[i+12] = global_data->sbound_droff[i];

#if defined(PARIX) || defined(PARIXHAL)
	for (i=0;i<global_data->no_nodes;i++) {
		if (SendNode(StarNodeId[i],TSCALING,(byte*)dbuf,
			DBUF_SIZE*sizeof(double))<0) {
			errout("Master: PCI_send_scaling failed to Node %d\n",i);
			pex(-1);
		}
	}
#elif defined(MPI)
	for (i=0;i<global_data->no_nodes;i++) {
		MPI_Ssend(dbuf, DBUF_SIZE,MPI_DOUBLE,i+1,TSCALING,MPI_COMM_WORLD);
	}
#else
   	i = pvm_initsend(encoding);
	pvm_pkdouble(dbuf,DBUF_SIZE,1);
	pvm_mcast(node_tids,global_data->no_nodes,TSCALING);
	my_pvm_freebuf(i);
#endif
	return res;
}

/******************************************************************/
int   PCI_recv_scaling(int my_number,
			double   *scaling_factor,
			LOAD_MSG_ALL_NODE *global_data)
{
	double	dbuf[DBUF_SIZE];
	int i=global_data->no_nodes;i++; /*** avoid warnings ***/
	
#if defined(PARIX) || defined(PARIXHAL)
	if (RecvNode(0,TSCALING,(byte*)dbuf, DBUF_SIZE*sizeof(double))<0) {
		errout("Node %d: PCI_recv_scaling failed\n",my_number);
		pex(-1);
	}
#elif defined(MPI)
   {	MPI_Status status;
	MPI_Recv(dbuf, DBUF_SIZE,MPI_DOUBLE,0,TSCALING,MPI_COMM_WORLD,&status);
   }
#else
   {	int    bid = pvm_recv(-1,TSCALING);
	pvm_upkdouble(dbuf, DBUF_SIZE,1);
	my_pvm_freebuf(bid);
   }
#endif
	*scaling_factor = dbuf[0];
	global_data->deltaimob[0]   = dbuf[1];
	global_data->deltaimob[1]   = dbuf[2];
	global_data->deltaimob[2]   = dbuf[3];
	global_data->deltaimob[3]   = dbuf[4];
	global_data->deltaimob[4]   = dbuf[5];
	global_data->deltaimob[5]   = dbuf[6];
	global_data->deltaimob[6]   = dbuf[7];
	global_data->deltaimob[7]   = dbuf[8];
	global_data->deltaimob[8]   = dbuf[9];
	global_data->final_cycle_no = (int)dbuf[10];
	
/*	for (i=0;i<6;i++)
        	global_data->pullspring[i] = dbuf[i+6]; */
	for(i=0;i<MXNO_SBOUND_PARAM;i++)
	  	global_data->sbound_droff[i] = dbuf[i+12];

	if (dbuf[11]<0.0) return -1; /*** stop signal ***/
	else		 return 1;
}

/************************************************************************/

int	send_start_node(int no_nodes)
{
#if defined(PARIX) || defined(PARIXHAL)
	int i,hgf=0;
	for (i=0;i<no_nodes;i++) {
		SendNode(StarNodeId[i],TSYNCHRONISE,(byte*)&hgf,sizeof(int));
	}
#elif defined(MPI)
	int i,hgf=0;
	for (i=0;i<no_nodes;i++) {
		MPI_Ssend(&hgf,1,MPI_INT,i+1,TSYNCHRONISE,MPI_COMM_WORLD);
	}
#else


#endif
	return 0;
}

int	recv_start_node()
{
#if defined(PARIX) || defined(PARIXHAL)
	int i;
	RecvNode(0,TSYNCHRONISE,(byte*)&i,sizeof(int));
#elif defined(MPI)
	int i;	
	MPI_Status status;
	MPI_Recv(&i,1,MPI_INT,0,TSYNCHRONISE,MPI_COMM_WORLD,&status);
#else

#endif
	return 0;
}

void
PCI_synchronise(int no_nodes)
{
#if defined(PARIX) || defined(PARIXHAL)
	int info=no_nodes;info++; /*** avoid warnings ***/

#elif defined(MPI)
	int info=no_nodes;info++; /*** avoid warnings ***/

#else
	int info=0,bid;
	bid = pvm_initsend(encoding);
	pvm_pkint(&info,1,1);
	pvm_mcast(node_tids,no_nodes,TSYNCHRONISE);
	my_pvm_freebuf(bid);
#endif
}

/********************************************************************/
/********************************************************************/

/**********************************************************************************/
/***										***/
/**********************************************************************************/

int    PCI_start_thread_send(	LOAD_MSG_ONE_NODE *my_data,
				LOAD_MSG_ALL_NODE *global_data,
				TRAVELLING_PACKETS *packet,
				int my_number,
				int *node_status,
				ARGUMENTS_THREAD *arg_send)
{
	arg_send->my_data     = my_data;
	arg_send->global_data = global_data;
	arg_send->packet      = packet;
	arg_send->my_number   = my_number;
	arg_send->node_status = node_status;

#if defined(PARIX) || defined(PARIXHAL)

#if defined(RS6K)
	arg_send->nodestar_top= nodestar_top;
	arg_send->nodenode_link= nodenode_link;	
#endif

#if ((DEBUG & 1) == 1)
txtout("Node %d: before create send-thread\n",my_number);
#endif
	{int error;
	Thread_t *send_thread;

	error = LocalLink(link_to_thread_send);
	if (error<0) {
		errout("Node %d: Could not generate link for thread_send: %d\n",my_number,error);
		pex(-1);
	}

	send_thread = CreateThread (NULL,100000,do_send_data,&error,arg_send);
	if (send_thread==NULL) {
		errout("Node %d: Error %d in spawning thread_send.\n",my_number,error);
		pex (-1);
	}
	}
#if ((DEBUG & 1) == 1)
txtout("Node %d: after create send-thread\n",my_number);
#endif

#elif defined(MPI)
#if SLOW_MPI

#else
   {	
	mpi_buffer_size = 	(global_data->no_nodes-1) * 
				global_data->maxatoms_per_node *
				(3*sizeof(float) + sizeof(struct_cluster)/3);
	mpi_send_buffer = alloc_chars(mpi_buffer_size);
	MPI_Buffer_attach((void*)mpi_send_buffer,mpi_buffer_size);
   }
#endif
#else	/**** PVM ***/

#endif
	return 0;
}


#if defined(PARIX) || defined(PARIXHAL)
static
int do_send_data(ARGUMENTS_THREAD *arg_send)
{
        int     flag=0,bid;
        int     no_nodes  = arg_send->global_data->no_nodes;
        int     my_number = arg_send->my_number;
        int     i,recv_node,infobuf[5],len,info,trueval=TRUE;
        void    *ptr;

        while (trueval) {
#if ((DEBUG & 1) == 1)
txtout("Node %d: in thread waiting for start signal.\n",my_number);
#endif
                /*** wait for the signal to start comming from the main-thread ***/
                RecvLink(link_to_thread_send[1],(void*)&flag,sizeof(flag));

#if ((DEBUG & 1) == 1)
txtout("Node %d: in thread after start signal.\n",my_number);
#endif
		bid   = pvm_initsend(encoding);
                pk_TRAVELLING_COORDINATES(&(arg_send->packet[my_number].coordinates));
                pk_TRAVELLING_BRANCHES   (&(arg_send->packet[my_number].branch));
                len   = pvm_buffer(&ptr);
#if defined(RS6K)
              for (i=my_number;i<my_number+no_nodes-1;i++) {
                        if (i>=no_nodes-1)	recv_node = i - no_nodes+1;
                        else 			recv_node = i;
		   
	   		infobuf[0] = my_number;
                        infobuf[1] = len;
                        
                        info = Send(arg_send->nodestar_top,
				    arg_send->nodenode_link[recv_node],
				    (byte*)infobuf,2*sizeof(int));
                        if (info<0) {
                                errout("Node %d: do_send_data failed #1 %d\n",my_number,info);
                                pex(-1);
                        }
                                              
                        info = Send(arg_send->nodestar_top,
				    arg_send->nodenode_link[recv_node],
				    (byte*)ptr,len);
                        if (info<0) {
                                errout("Node %d: do_send_data failed #2 %d\n",my_number,info);
                                pex(-1);
                        } 
		}
#else
                for (i=my_number+1;i<my_number+no_nodes;i++) {
                        if (i>=no_nodes)	recv_node = i - no_nodes;
                        else 			recv_node = i;

                        infobuf[0] = my_number;
                        infobuf[1] = len;
                        info = SendNode(StarNodeId[recv_node],COANDBR,(byte*)infobuf,2*sizeof(int));
                        if (info<0) {
                                errout("Node %d: do_send_data failed #1 %d\n",my_number,info);
                                pex(-1);
                        }
                        info = SendNode(StarNodeId[recv_node],COANDBR,(byte*)ptr,len);
                        if (info<0) {
                                errout("Node %d: do_send_data failed #2 %d\n",my_number,info);
                                pex(-1);
                        }
		}
#endif
/**
                my_pvm_freebuf(bid);
**/ 
                /*** tell the main-thread that work is done ***/
                SendLink(link_to_thread_send[1],(void*)&flag,sizeof(flag));
#if ((DEBUG & 1) == 1)
txtout("Node %d: in thread_send after sending stop signal.\n",my_number);
#endif
        }
        return 0;
}
#endif /**** end of PARIX ******/

#if defined(MPI)
static	MPI_Request request[500];
#endif

/**************************************************************************/
/*** here the main-thread tells his send-thread to start 		***/
/*** or in case of PVM the Broad-cast is done 				***/
/**************************************************************************/
void   PCI_start_send(	LOAD_MSG_ALL_NODE *global_data,
			TRAVELLING_PACKETS *packet,
			int my_number,
			int *node_status)
{
#if defined(PARIX) || defined(PARIXHAL)
   {	/******************** PARIX ******************************/
	int flag = 1;
	TRAVELLING_PACKETS *a; /*** no warnings ***/
	int b=global_data->no_nodes;
	int *c;
	a=packet;
	c=node_status;
	a++;b++;c++;

	if (SendLink(link_to_thread_send[0],&flag,sizeof(flag))==-1) {
		errout("Node %d: PCI_start_send failed\n",my_number);
		pex(-1);
	}
   }
#elif defined(MPI)
#if SLOW_MPI
	/*********************************************************************/
	/*************** slow but secure MPI communication *******************/
	/*********************************************************************/
   {	int i,j,bid,info,sinfobuf[5],rinfobuf[5],dest_node,source_node,res,
		source_len;
	char *source_ptr;
	MPI_Status status;

	bid  = pvm_initsend(encoding);
	info = pvm_pkint(&my_number,1,1);
	pk_TRAVELLING_COORDINATES(&(packet[my_number].coordinates));
	pk_TRAVELLING_BRANCHES   (&(packet[my_number].branch));
	len   = pvm_buffer(&ptr);
	sinfobuf[0] = my_number;
	sinfobuf[1] = len;
#if ((DEBUG & 1) == 1)
txtout("Node %d: PCI_start send packing done %d %d\n",
my_number,len,global_data->no_nodes);
#endif
	for (j=0;j<global_data->no_nodes;j++) {
		if (j==my_number)
		for (i=my_number+1;i<my_number+global_data->no_nodes;i++) {
			if (i>=global_data->no_nodes)	dest_node = i - global_data->no_nodes;
			else 				dest_node = i;

			MPI_Ssend(ptr,len,MPI_BYTE,dest_node+1,COANDBR,
				MPI_COMM_WORLD);

		} else {
			source_node = j;	

			MPI_Probe(source_node+1,COANDBR,MPI_COMM_WORLD,&status);
			MPI_Get_count(&status,MPI_BYTE,&source_len);

			source_ptr  = pvm_set_readbuf(source_len);
			MPI_Recv(source_ptr,source_len,MPI_BYTE,source_node+1,
				COANDBR,MPI_COMM_WORLD,&status);

			pvm_upkint(&source_node,1,1);
			if (source_node==my_number) {
				errout("Node %d: pvm_recv could not be from same node\n",
				my_number);
				pex(-1);			
			}
			upk_TRAVELLING_COORDINATES(&(packet[source_node].coordinates));
			upk_TRAVELLING_BRANCHES   (&(packet[source_node].branch));

			my_pvm_freebuf(0);

			node_status[source_node] = PACKET_FOR_CALC;
		}
	}
#if ((DEBUG & 1) == 1)
txtout("Node %d: finished now distributing.\n",my_number,dest_node);
#endif
  }
#else
	/*********************************************************************/
	/*************** fast but memory consuming MPI communication *********/
	/*********************************************************************/
	int j,info,dest_node,len,bid;
	MPI_Status status;

	bid  = pvm_initsend(encoding);
	info = pvm_pkint(&my_number,1,1);
	pk_TRAVELLING_COORDINATES(&(packet[my_number].coordinates));
	pk_TRAVELLING_BRANCHES   (&(packet[my_number].branch));
	len   = pvm_buffer(&ptr);
#if ((DEBUG & 1) == 1)
txtout("Node %d: PCI_start send packing done %d %d\n",
my_number,len,global_data->no_nodes);
#endif
	for (j=0;j<global_data->no_nodes;j++) {
		if (j==my_number) continue;
		dest_node = j;
		MPI_Ibsend(ptr,len,MPI_BYTE,dest_node+1,COANDBR,
			MPI_COMM_WORLD,&(request[dest_node]));
#if ((DEBUG & 1) == 1)
txtout("Node %d: PCI_start send done to node %d\n",
my_number,dest_node);
#endif
	}
#endif
#else 	/**********************************************************/
	/**** do now the PVM broad-cast sending			***/
	/**********************************************************/
  {	int bid,info,infobuf[2];

   	bid = pvm_recv(-1,TSYNCHRONISE);
	pvm_upkint(&info,1,1);
	my_pvm_freebuf(bid);
  
	bid = pvm_initsend(encoding);
	if (bid<0) {
		errout("Node %d: pvm_initsend in PCI_start_send failed %d\n",
			my_number,bid);
		pex(-1);
	}
	infobuf[0] = my_number;
	info = pvm_pkint(infobuf,1,1);
	if (info<0) {
		errout("Node %d: pvm_pkint in PCI_start_send failed %d\n",
			my_number,info);
		pex(-1);
	}

	pk_TRAVELLING_COORDINATES(&(packet[my_number].coordinates));
	pk_TRAVELLING_BRANCHES   (&(packet[my_number].branch));

	info = pvm_mcast(node_tids,global_data->no_nodes,COANDBR);
	if (info<0) {
		errout("Node %d: pvm_mcast in PCI_start_send broadcast failed %d\n",
			my_number,info);
		pex(-1);
	}
	my_pvm_freebuf(bid);
  }

	/***************************************************/
	/**** now do the pvm-receiving 			 ***/
	/***************************************************/
  {	int bid,i,infobuf[5],other_number,info;

	for (i=0;i<global_data->no_nodes-1;i++) {
		bid = pvm_recv(-1,COANDBR);
		if (bid<0) {
			errout("Node %d: pvm_recv in PCI_start_send failed %d\n",
				my_number,bid);
			pex(-1);
		}

		info = pvm_upkint(infobuf,1,1);
		if (info<0) {
			errout("Node %d: pvm_upkint in PCI_start_send failed %d\n",
				my_number,info);
			pex(-1);
		}
		other_number = infobuf[0];
		if (other_number==my_number) {
			errout("Node %d: pvm_recv could not be from same node\n",
				my_number);
			pex(-1);			
		}
		if (node_status[other_number]==PACKET_FOR_CALC) {
			/*** This error occurs, if one node overtooks others ***/
			errout("Node %d: Synchronisation error in PCI_start_send!\n",
				my_number);	
			pex(-1);
		}	

		upk_TRAVELLING_COORDINATES(&(packet[other_number].coordinates));
		upk_TRAVELLING_BRANCHES   (&(packet[other_number].branch));

		my_pvm_freebuf(bid);
		
		node_status[other_number] = PACKET_FOR_CALC;
	}
   }
#endif
	return;
}

void	PCI_get_packet(	LOAD_MSG_ALL_NODE *global_data,
			TRAVELLING_PACKETS *packet,
			int my_number,
			int *node_status)
{
#if defined(PARIX) || defined(PARIXHAL)
	{
	int a=global_data->no_nodes;
	int b=my_number;
	int c=node_status[0];
	int d=packet[0].coordinates.no_atoms;
	a++;b++;c++;d++;		
	}
#elif defined(MPI)
#if SLOW_MPI

#else
   {	int something_there=0,source_len,source_node;
	char *source_ptr;
	MPI_Status status;
	MPI_Request request_recv;
/***
	MPI_Iprobe(MPI_ANY_SOURCE,COANDBR,MPI_COMM_WORLD,
		&something_there,&status);
	if (something_there) {
***/
		MPI_Probe(MPI_ANY_SOURCE,COANDBR,MPI_COMM_WORLD,&status);

		MPI_Get_count(&status,MPI_BYTE,&source_len);
		source_node = status.MPI_SOURCE-1;
#if ((DEBUG & 1) == 1)	
txtout("Node %d: after probe from %d len %d\n",
my_number,source_node,source_len);
#endif
		source_ptr  = pvm_set_readbuf(source_len);
/**
		MPI_Recv(source_ptr,source_len,MPI_BYTE,source_node+1,
			COANDBR,MPI_COMM_WORLD,&status);
**/
		MPI_Irecv(source_ptr,source_len,MPI_BYTE,source_node+1,
			COANDBR,MPI_COMM_WORLD,&request_recv);
		MPI_Wait(&request_recv,&status);

#if ((DEBUG & 1) == 1)
txtout("Node %d: after recv from %d len %d.\n",my_number,source_node,source_len);
#endif
		pvm_upkint(&source_node,1,1);
		if (source_node==my_number) {
			errout("Node %d: mpi_recv can not be from same node!\n",
				my_number);
			pex(-1);			
		}
		if (node_status[source_node]==PACKET_FOR_CALC) {
			/*** This error occurs, if one node overtooks others ***/
			errout("Node %d: Synchronisation error in PCI_get_packet!\n",
				my_number);	
			pex(-1);

		}
		upk_TRAVELLING_COORDINATES(&(packet[source_node].coordinates));
		upk_TRAVELLING_BRANCHES   (&(packet[source_node].branch));

		my_pvm_freebuf(0);

		node_status[source_node] = PACKET_FOR_CALC;
/***
	} else {

		return;
	}
***/
   }
#endif
#else

#endif
}


/******************************************************************/
void   PCI_stop_send(int no_nodes)
{
#if defined(PARIX) || defined(PARIXHAL)
	/******************** PARIX ******************************/
	int    flag = 1,a=no_nodes;
	a++;
	if (RecvLink(link_to_thread_send[0],&flag,sizeof(flag))==-1) {
		errout("Node ?: PCI_stop_send failed\n");
		pex(-1);
	}
#elif defined(MPI)
#if SLOW_MPI

#else
   {	void *dptr;
	int di,dest_node;
	MPI_Status status;

	for (dest_node=0;dest_node<no_nodes;dest_node++) {
		MPI_Wait(&(request[dest_node]),&status);
	}

	/*** that is needed: I (ME) dont know exactly why ? ***/
	MPI_Buffer_detach(&dptr,&di);
	MPI_Buffer_attach((void*)mpi_send_buffer,mpi_buffer_size);
   
	}
#endif
#else	/**************************** PVM, MPI *******************/

	/*** nothing to do here ***/
#endif
	return;
}



/**********************************************************************************/
/***										***/
/**********************************************************************************/
int    PCI_start_thread_recv(	LOAD_MSG_ONE_NODE *my_data,
				LOAD_MSG_ALL_NODE *global_data,
				TRAVELLING_PACKETS *packet,				
				int my_number,
				int *node_status,
				ARGUMENTS_THREAD *arg_recv)
{
	arg_recv->my_data     = my_data;
	arg_recv->global_data = global_data;
	arg_recv->packet      = packet;
	arg_recv->my_number   = my_number;
	arg_recv->node_status = node_status;

#if defined(PARIX) || defined(PARIXHAL)

#if defined(RS6K)
	arg_recv->nodestar_top= nodestar_top;
	arg_recv->nodenode_link= nodenode_link;
	arg_recv->nodenode_CB = nodenode_CB;	
#endif

#if ((DEBUG & 1) == 1)
txtout("Node %d: before create recv-thread\n",my_number);
#endif
	{int error;
	Thread_t *recv_thread;

	error = LocalLink(link_to_thread_recv);
	if (error<0) {
		errout("Node %d: Could not generate link for thread_recv: %d\n",my_number,error);
		pex(-1);
	}

	recv_thread = CreateThread(NULL,100000,do_recv_data,&error,arg_recv);
	if (recv_thread==NULL) {
		errout("Node %d: Error %d in spawning thread_recv.\n",my_number,error);
		pex (-1);
	}
	}
#if ((DEBUG & 1) == 1)
txtout("Node %d: after create recv-thread\n",my_number);
#endif

#else	/**** PVM, MPI ***/

	/***** nothing to do here ******/

#endif
	return 0;
}


#if defined(PARIX) || defined(PARIXHAL)
static
int do_recv_data(ARGUMENTS_THREAD *arg_recv)
{
        int     flag =0,info,trueval=TRUE;
        int     no_nodes  = arg_recv->global_data->no_nodes;
        int     my_number = arg_recv->my_number,other_number,len;
        int     infobuf[5];
        int     i;
        void    *ptr;
 
 
        while (trueval) {
#if ((DEBUG & 1) == 1)
txtout("Node %d: in recv_thread waiting for start signal.\n",my_number);
#endif
	        RecvLink(link_to_thread_recv[1],(void*)&flag,sizeof(flag));

                /*** wait for the signal to start comming from the main-thread ***/
#if ((DEBUG & 1) == 1)
txtout("Node %d: in thread after start signal.\n",my_number);
#endif

#if defined(RS6K)
	   	flag = my_number-1;
		for (i=0;i<no_nodes-1;i++,flag--) {
			if(flag < 0) flag = flag + no_nodes - 1;
			info = RecvLink(arg_recv->nodenode_CB[flag],
					(byte*)infobuf,2*sizeof(int));
                        if (info<0) {
                                errout("Node %d: do_recv_data failed #1 %d\n",my_number,info);
                                pex(-1);
                        }
		   	
                        other_number = infobuf[0];
                        len          = infobuf[1];
 
                        ptr = pvm_set_readbuf(len);	

                        while (arg_recv->node_status[other_number]==PACKET_FOR_CALC) {
                           printf("BUSY!!!!!!!!!!!!!!!!!!!\n");
			   TimeWait(TimeNow()+1000);	/*** wait until status changes! ***/
                        }
                        info = RecvLink(arg_recv->nodenode_CB[flag],
                        	(byte*)ptr,len);
                        if (info<0) {
                                errout("Node %d: do_recv_data failed #2 %d\n",my_number,info);
                                pex(-1);
                        }
                        upk_TRAVELLING_COORDINATES(&(arg_recv->packet[other_number].coordinates));
                        upk_TRAVELLING_BRANCHES   (&(arg_recv->packet[other_number].branch));
 
 			my_pvm_freebuf(0);
 			                       
                        arg_recv->node_status[other_number] = PACKET_FOR_CALC;
		}
#else
                for (i=1;i<no_nodes;i++) {
                        info = RecvNode(-1,COANDBR,(byte*)infobuf,2*sizeof(int));
                        if (info<0) {
                                errout("Node %d: do_recv_data failed #1 %d\n",my_number,info);
                                pex(-1);
                        }
                        other_number = infobuf[0];
                        len          = infobuf[1];
 
                        ptr = pvm_set_readbuf(len);
 
                        while (arg_recv->node_status[other_number]==PACKET_FOR_CALC) {
                        	/*** wait until status changes! ***/
                        }
 
                        info = RecvNode(StarNodeId[other_number],COANDBR,(byte*)ptr,len);
                        if (info<0) {
                                errout("Node %d: do_recv_data failed #2 %d\n",my_number,info);
                                pex(-1);
                        }
                        upk_TRAVELLING_COORDINATES (&(arg_recv->packet[other_number].coordinates));
                        upk_TRAVELLING_BRANCHES    (&(arg_recv->packet[other_number].branch));
		
                       	my_pvm_freebuf(0);
			
                        arg_recv->node_status[other_number] = PACKET_FOR_CALC;
                }
#endif
                /*** tell the main-thread that work is done ***/

                SendLink(link_to_thread_recv[1],(void*)&flag,sizeof(flag));

#if ((DEBUG & 1) == 1)
txtout("Node %d: in recv_thread after sending stop signal.\n",my_number);
#endif
        }
        return 0;
}
#endif /**** end of PARIX ****/


/**************************************************************************/
/*** here the main-thread tells his send-thread to start 		***/
/**************************************************************************/
void   PCI_start_recv()
{
#if defined(PARIX) || defined(PARIXHAL)
	{/******************** PARIX ******************************/
	int    flag = 1;
	if (SendLink(link_to_thread_recv[0],&flag,sizeof(flag))==-1) {
		errout("Node ?: PCI_start_recv failed\n");
		pex(-1);
	}
	}
#else 	/*********************** PVM, MPI *************************/

#endif
	return;
}

/******************************************************************/
void   PCI_stop_recv()
{
#if defined(PARIX) || defined(PARIXHAL)
	int    flag = 1;
	if (RecvLink(link_to_thread_recv[0],&flag,sizeof(flag))==-1) {
		errout("Node ?: PCI_stop_recv failed\n");
		pex(-1);
	}
#else	/**** PVM, MPI case ***/

	
#endif
	return;
}

/*** done on the nodes, send coordinates to the master ***/
void PCI_coords_send(	int no_nodes,
			int my_number,
			LOAD_MSG_ONE_NODE *my_data,
			LOAD_MSG_ALL_NODE *global_data)
{
	int bid,info;	

	bid = pvm_initsend(encoding);

	info  = pvm_pkint   (&(my_data->no_atoms),1,1);
	info |= pvm_pkint   (&my_number,1,1);
	info |= pvm_pkdouble(my_data->x_old,my_data->no_atoms,1);
	info |= pvm_pkdouble(my_data->y_old,my_data->no_atoms,1);
	info |= pvm_pkdouble(my_data->z_old,my_data->no_atoms,1);

	if (global_data->rotation_mode) {
	  info |= pvm_pkdouble(global_data->rot1_x,global_data->nr_of_rot_atoms,1);
	  info |= pvm_pkdouble(global_data->rot1_y,global_data->nr_of_rot_atoms,1);
	  info |= pvm_pkdouble(global_data->rot1_z,global_data->nr_of_rot_atoms,1);

	  info |= pvm_pkdouble(global_data->rot2_x,global_data->nr_of_rot_atoms,1);
	  info |= pvm_pkdouble(global_data->rot2_y,global_data->nr_of_rot_atoms,1);
	  info |= pvm_pkdouble(global_data->rot2_z,global_data->nr_of_rot_atoms,1);

	  info |= pvm_pkdouble(global_data->rotation_torque,global_data->nr_of_rot_atoms,1);
	}

	if (info!=0) {
		errout("Node %d: Packing failed in PCI_coords_send !\n",my_number);
		pex(-1);
	}

#if defined(PARIX) || defined(PARIXHAL)
	{int len=0,infobuf[5];
	void *ptr;

	len = pvm_buffer(&ptr);
	infobuf[0] = my_number;
	infobuf[1] = len;

	info=SendNode(StarNodeId[no_nodes],TTCOORDS,(byte*)infobuf,2*sizeof(int));
	if (info<0) {
		errout("Node %d: PCI_coords_send #1 failed %d\n",my_number,info);
		pex(-1);
	}	

	info=SendNode(StarNodeId[no_nodes],TTCOORDS,(byte*)ptr,len);	
	if (info<0) {
		errout("Node %d: PCI_coords_send #2 failed %d\n",my_number,info);
		pex(-1);
	}

	}
#elif defined(MPI)
   	{int len=0;
	void *ptr;

	len = pvm_buffer(&ptr);
	MPI_Ssend(ptr,len,MPI_BYTE,0,TTCOORDS,MPI_COMM_WORLD);
   	}
#else
	pvm_send(node_tids[no_nodes],TTCOORDS);
#endif
	my_pvm_freebuf(bid);

	return;
}

/*** done on the master, collects the coordinates from nodes ***/
void PCI_coords_recv(int recv_node,TRAVELLING_COORDINATES *p,LOAD_MSG_ALL_NODE *q)
{
	int info,bid=0,no_atoms,len=0,other_number=0;

#if defined(PARIX) || defined(PARIXHAL)
   {
	int infobuf[5],a=recv_node;
	void *ptr;
	a++;	
#if defined(RS6K)
	info=RecvNode(StarNodeId[recv_node],TTCOORDS,(byte*)infobuf,2*sizeof(int));
#else
	info=RecvNode(-1,TTCOORDS,(byte*)infobuf,2*sizeof(int));
#endif
	if (info<0) {
		errout("PCI_coords_recv #1 failed %d",info);
		pex(-1);
	}
	other_number = infobuf[0];
	len          = infobuf[1];

	ptr  = pvm_set_readbuf(len);
	info = RecvNode(StarNodeId[other_number],TTCOORDS,(byte*)ptr,len);
	if (info<0) {
		errout("PCI_coords_recv #2 failed %d  (%d %d)",info,other_number,len);
		pex(-1);
	}
   }
#elif defined(MPI)
   {
	void *ptr;
	MPI_Status status;
	MPI_Probe(MPI_ANY_SOURCE,TTCOORDS,MPI_COMM_WORLD,&status);
	MPI_Get_count(&status,MPI_BYTE,&len);
	other_number = status.MPI_SOURCE-1;
	ptr = pvm_set_readbuf(len);
	MPI_Recv(ptr,len,MPI_BYTE,other_number+1,TTCOORDS,MPI_COMM_WORLD,&status);
   }
#else
	bid = pvm_recv(-1,TTCOORDS);
#endif

	info =        pvm_upkint(&no_atoms,1,1);
	realloc_TRAVELLING_COORDINATES(p,no_atoms);
	p->no_atoms = no_atoms;
	info = info | pvm_upkint(&(p->node),1,1);
	info = info | pvm_upkdouble(p->x,p->no_atoms,1);
	info = info | pvm_upkdouble(p->y,p->no_atoms,1);
	info = info | pvm_upkdouble(p->z,p->no_atoms,1);

	if (q->rotation_mode) {
	  info = info | pvm_upkdouble(q->rot_t1_x,q->nr_of_rot_atoms,1);
	  info = info | pvm_upkdouble(q->rot_t1_y,q->nr_of_rot_atoms,1);
	  info = info | pvm_upkdouble(q->rot_t1_z,q->nr_of_rot_atoms,1);

	  info = info | pvm_upkdouble(q->rot_t2_x,q->nr_of_rot_atoms,1);
	  info = info | pvm_upkdouble(q->rot_t2_y,q->nr_of_rot_atoms,1);
	  info = info | pvm_upkdouble(q->rot_t2_z,q->nr_of_rot_atoms,1);

	  info = info | pvm_upkdouble(q->rotation_t_torque,q->nr_of_rot_atoms,1);
	}

	if (info<0) {
		errout("PCI_coords_recv #3 failed %d  (%d %d)",info,other_number,len);
		pex(-1);
	}
	my_pvm_freebuf(bid);
	return;
}
#endif
