/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/* other local constants */
#define MXNOTRYPVM 	3

int    PCI_start(int *MyProcID, int *nProcs, int argc, char **argv);
void   PCI_exit(int i);
void   PCI_kill(int no_nodes);
int    PCI_set_no_nodes(int req_no_nodes, int nProc);
int    PCI_master_setup(int no_nodes, int nProcs,int host_tid);
void   PCI_node_setup(int MyProcID, int *nProcs, int *my_number);

void   PCI_send_update(LOAD_MSG_ALL_NODE *global_data);
void   PCI_recv_update(int nProcs,LOAD_MSG_ALL_NODE *global_data);
void   PCI_send_msg_one_node(LOAD_MSG_ONE_NODE *start_msg,int ring_node);
void   PCI_send_msg_all_node(LOAD_MSG_ALL_NODE *global_data);

int    PCI_enroll(void);

void   PCI_recv_load_msg_one_node (int nProcs, LOAD_MSG_ONE_NODE *my_data);
void   PCI_recv_load_msg_all_node (int nProcs, LOAD_MSG_ALL_NODE *global_data);

int   PCI_send_scaling(double scaling_factor,
			LOAD_MSG_ALL_NODE *global_data);
int   PCI_recv_scaling(int my_number,
			double   *scaling_factor,
			LOAD_MSG_ALL_NODE *global_data);
void   PCI_etcdata_send(ENERGY_INFO *local_energy,
			LOAD_MSG_ALL_NODE *global_data);
void PCI_etcdata_recv(	int no_nodes,
			int my_number,
			ENERGY_INFO *local_energy,
			LOAD_MSG_ONE_NODE *my_data,
			LOAD_MSG_ALL_NODE *global_data);
void   PCI_positions_send(LOAD_MSG_ALL_NODE *global_data);
void PCI_positions_recv(int no_nodes,
			int my_number,
			LOAD_MSG_ALL_NODE *global_data);

void   PCI_node_cleanup (void);

#if defined(PARIX)
#include <sys/link.h>
#include <sys/root.h>
#include <sys/select.h>
#include <sys/thread.h>
#include <sys/time.h>
#endif

#if defined(PARIXHAL)
#include </export/epx/epx/include/epx/link.h>
#include </export/epx/epx/include/epx/root.h>
#include </export/epx/epx/include/epx/select.h>
#include </export/epx/epx/include/epx/thread.h>
#include </export/epx/epx/include/epx/time.h>
#endif


typedef struct
{
	LOAD_MSG_ONE_NODE *my_data;
	LOAD_MSG_ALL_NODE *global_data;
	TRAVELLING_PACKETS *packet;
	int	my_number;
	int	*node_status;
	int	nodestar_top;
	int	*nodenode_link;
#if defined(PARIX) || defined(PARIXHAL)	
	LinkCB_t **nodenode_CB;
#endif
}      ARGUMENTS_THREAD;


int	recv_start_node(void);
int	send_start_node(int no_nodes);
void	PCI_synchronise(int no_nodes);
int	PCI_start_thread_send(	LOAD_MSG_ONE_NODE *my_data,
				LOAD_MSG_ALL_NODE *global_data,
				TRAVELLING_PACKETS *packet,
				int my_number,
				int *node_status,
				ARGUMENTS_THREAD *arg_send);

void	PCI_start_send(	LOAD_MSG_ALL_NODE *global_data,
			TRAVELLING_PACKETS *packet,
			int my_number,
			int *node_status);
void	PCI_stop_send(int no_nodes);

void	PCI_get_packet(	LOAD_MSG_ALL_NODE *global_data,
			TRAVELLING_PACKETS *packet,
			int my_number,
			int *node_status);

int	PCI_start_thread_recv (	LOAD_MSG_ONE_NODE *my_data,
				LOAD_MSG_ALL_NODE *global_data,
				TRAVELLING_PACKETS *packet,				
				int my_number,
				int *node_status,
				ARGUMENTS_THREAD *arg_recv);
void	PCI_start_recv(void);
void	PCI_stop_recv(void);


void	PCI_coords_send(int no_nodes,
			int my_number,
			LOAD_MSG_ONE_NODE *my_data,
			LOAD_MSG_ALL_NODE *global_data);

void	PCI_coords_recv(int recv_node,
			TRAVELLING_COORDINATES *p,
			LOAD_MSG_ALL_NODE *q);


void   PCI_energy_recv(	int recv_node,
			ENERGY_INFO *energy,
			int no_nodes,
			int *other_number,
			STRUCT_FORCES *my_forces);

void   PCI_energy_send( int no_nodes,
			int my_number,
			ENERGY_INFO *local_energy,
			LOAD_MSG_ONE_NODE *my_data,
			LOAD_MSG_ALL_NODE *global_data);

void   PCI_send_qmforces(LOAD_MSG_ALL_NODE *global_data,
			 STRUCT_QM_INFO *finfo);

void   PCI_recv_qmforces(int nProcs,
			 LOAD_MSG_ALL_NODE *global_data,
			 LOAD_MSG_ONE_NODE *my_data,
			 int my_number);
