/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>	/* on SP2 do not link with -lbsd !! */

#include "ego.h"
#include "egalloc.h"
#include "util_lib.h"

/* DEBUG == 1 : standard debugging */
/* DEBUG == 2 : debug loop */
/* DEBUG == 4 : */
#define DEBUG			0

int do_neuroclust_ms(	int inpn, double *xinp, double *yinp, double *zinp,
			int outn, double *xout, double *yout, double *zout,
			int ini_outvecs);

int free_neuroclust_ms(void);

static int shake_index_array(int numshake, int inpn, int *rndarr);

static	int	first_time=1;

static	int	*rndarr;
static	int	alloc_rndarr=0;

static	double	*xdif,*ydif,*zdif,*dist;
static	double	*move;
static	int	alloc_dif=0;

int
do_neuroclust_ms(	int inpn, double *xinp, double *yinp, double *zinp,
			int outn, double *xout, double *yout, double *zout,
			int ini_outvecs)
{
	int	i;
	int	lerni,makri,veci,outi;
	int	makr_limit,lern_limit;
	double	ei,ef,el,efac;
	double	ti,tf,tl,tfac;

#if (DEBUG & 1) == 1
txtout("Master: Enter do_neuroclust_ms, outn=%d\n",outn);
#endif
	if ((inpn<1)||(outn>inpn)) {
		errout("Illegal input for do_neuroclust_ms (inpn=%d, outn=%d)\n",
			inpn,outn);
		pex(-1);
	}
	if (inpn==outn) { 
		memcpy(xout,xinp,inpn*sizeof(double));
		memcpy(yout,yinp,inpn*sizeof(double));
		memcpy(zout,zinp,inpn*sizeof(double));
#if (DEBUG & 1) == 1
txtout("Master: Exit 1 do_neuroclust_ms\n");
#endif
		return 0;
	}

#if (DEBUG & 1) == 1
txtout("Master, do_neuroclust_ms: after memcpy\n");
#endif
	if (first_time) {
		rndarr = alloc_ints(inpn);
		alloc_rndarr = inpn;	
		set_allocinfo("egmarti start-up");
		xdif   = alloc_doubles(outn+10);
		ydif   = alloc_doubles(outn+10);
		zdif   = alloc_doubles(outn+10);
		dist   = alloc_doubles(outn+10);
		move   = alloc_doubles(outn+10);
		set_allocinfo("XXX");
		alloc_dif = outn;
	} else {
		if (alloc_rndarr<inpn) {
			set_allocinfo("egmarti realloc inp");
			realloc_ints(&rndarr,inpn,alloc_rndarr);
			alloc_rndarr = inpn;
		}
		if (alloc_dif<outn) {
			set_allocinfo("egmarti realloc out");
			realloc_doubles(&xdif,outn,alloc_dif);
			realloc_doubles(&ydif,outn,alloc_dif);
			realloc_doubles(&zdif,outn,alloc_dif);
			realloc_doubles(&dist,outn,alloc_dif);
			realloc_doubles(&move,outn,alloc_dif);
			alloc_dif=outn;
		}
		set_allocinfo("XXX");
	}
#if (DEBUG & 1) == 1
txtout("Master, do_neuroclust_ms: after alloc\n");
#endif
	/******* Make a random index array *******/
	for (i=0;i<inpn;i++) rndarr[i] = i;
	shake_index_array(4,inpn,rndarr);

	/******* ini out vecs ********/
	if (ini_outvecs) {
		for (i=0;i<outn;i++) {
			xout[i] = xinp[i];
			yout[i] = yinp[i];
			zout[i] = zinp[i];
		}

		if (outn<=6) {
			double dx,dy,dz;
			double cenx=0.0,ceny=0.0,cenz=0.0;

			for (i=0;i<inpn;i++) {
				if (xout[0]>xinp[i]) xout[0]=xinp[i];
				if (xout[1]<xinp[i]) xout[1]=xinp[i];

				if (yout[0]>yinp[i]) yout[0]=yinp[i];
				if (yout[1]<yinp[i]) yout[1]=yinp[i];

				if (zout[0]>zinp[i]) zout[0]=zinp[i];
				if (zout[1]<zinp[i]) zout[1]=zinp[i];

				cenx += xinp[i];
				ceny += yinp[i];
				cenz += zinp[i];			
			}
			cenx /= inpn;	    ceny /= inpn;	cenz /= inpn;
			dx=xout[1]-xout[0]; dy=yout[1]-yout[0];	dz=zout[1]-zout[0];

			xout[0]=cenx-dx/3; yout[0]=ceny; zout[0]=cenz;
			xout[1]=cenx+dx/3; yout[1]=ceny; zout[1]=cenz;
			if (outn>=3) {
				xout[2]=cenx; yout[2]=ceny+dy/3; zout[2]=cenz;
			}
			if (outn>=4) {
				xout[3]=cenx; yout[3]=ceny-dy/3; zout[3]=cenz;
			}
			if (outn>=5) {
				xout[4]=cenx; yout[4]=ceny; zout[4]=cenz+dz/3;
			}
			if (outn>=6) {
				xout[5]=cenx; yout[5]=ceny; zout[5]=cenz-dz/3;
			}
		}
	}	

#if (DEBUG & 1) == 1
txtout("Master, do_neuroclust_ms: after ini out vecs\n");
#endif
	shake_index_array(1,inpn,rndarr);
#if (DEBUG & 1) == 1
txtout("Master, do_neuroclust_ms: after shake_index_array\n");
#endif

	ei         = 0.95;
	ef	   = 0.03;	
	tf	   = 0.25;
	lern_limit = 10;
	ti	   = max(10,outn/2);
	if (outn<=6) { 
		ei = 0.90;
		ef = 0.03;
		tf = 0.25;
		lern_limit=10;
		ti = outn/2;
	}

#if (DEBUG & 1) == 1
txtout("Master, do_neuroclust_ms: after set params: inpn=%d, ef=%f\n",inpn,ef);
#endif
	makr_limit = (int) ((3.0*outn)/(inpn*ef));
#if (DEBUG & 1) == 1
txtout("Master, do_neuroclust_ms: makr_limit=%d, lern_limit=%d,ef=%f, ei=%f\n", 
	makr_limit, lern_limit,ef, ei);
#endif
	efac = pow(ef/ei,(double)(1.0/(lern_limit-1.0)));
#if (DEBUG & 1) == 1
txtout("Master, do_neuroclust_ms: efac=%f\n", efac);fflush(stdout);
#endif
	tfac = pow(tf/ti,(double)(1.0/(lern_limit-1.0)));
#if (DEBUG & 1) == 1
txtout("Master, do_neuroclust_ms: tfac=%f\n", tfac);
#endif
	el = ei;
	tl = ti;

#if (DEBUG & 1) == 1
txtout("Master, do_neuroclust_ms: before loop\n");
#endif
	for (lerni=0;lerni<lern_limit;lerni++) {		/*** One Learn Step 	***/
		int cuti=outn-1;
		int cur_makr=1+(int)(makr_limit*lerni/(lern_limit-1)+0.5);			
#if (DEBUG & 2) == 2
txtout("Master, do_neuroclust_ms: before learn step %d\n", lerni);
#endif
		for (outi=0;outi<=outn;outi++) {
			if (outn>=6)		move[outi] = el*exp(-(  outi+1.0)/tl);
			else if (outn>3)	move[outi] = el*exp(-(2*outi+1.0)/tl);
			else 			move[outi] = el*exp(-(2*outi+1.0)/tl);

			if ((move[outi]<0.0005)&&(cuti>outi)) cuti=outi;		
		}

		for (makri=0;makri<cur_makr;makri++) {	  /*** Makro Integration Step ***/
			shake_index_array(1,inpn,rndarr); /*** new random index array ***/			
			for (veci=0;veci<inpn;veci++) {	  /*** lets see every inpvec ***/
				int vecsel = rndarr[veci];/*** we select this vector ***/
				double x = xinp[vecsel];
				double y = yinp[vecsel];
				double z = zinp[vecsel];

				for (outi=0;outi<outn;outi++) {
					xdif[outi] = x - xout[outi];
					ydif[outi] = y - yout[outi];
					zdif[outi] = z - zout[outi];
					dist[outi] = 	((xdif[outi]<0.0)?-xdif[outi]:xdif[outi]) +
							((ydif[outi]<0.0)?-ydif[outi]:ydif[outi]) +
							((zdif[outi]<0.0)?-zdif[outi]:zdif[outi]);
				}

				for (outi=0;outi<outn;outi++) {
					double t=dist[outi];	
					int ka=0,i;
					for (i=0;i<outn;i++) {
						if (t>=dist[i]) { 
							ka++;
							if (ka>cuti) break;	
						}									
					}	
					if (ka>cuti) continue;						
					xout[outi] += move[ka]*xdif[outi];
					yout[outi] += move[ka]*ydif[outi];
					zout[outi] += move[ka]*zdif[outi];
				}
			} /*** end: lets see every inpvec ***/
		} /*** Makro Integration Step ***/
		el = efac*el;
		tl = tfac*tl;
	} /*** One Learn Step 	***/	

	first_time = 0;
#if (DEBUG & 1) == 1
txtout("Master: Exit 2 do_neuroclust_ms\n");
#endif
	return 0;
}

int
free_neuroclust_ms()
{
	if (!first_time) {
		myfree(rndarr);
		alloc_rndarr = 0;
		myfree(xdif);
		myfree(ydif);
		myfree(zdif);
		myfree(dist);
		alloc_dif=0;
		first_time = 1;
	}	
	return 0;
}

static int
shake_index_array(int numshake, int inpn, int *rndarr)
{
	int i,n;
	double fac=inpn-1;

	for (i=0;i<numshake;i++) {
		for (n=0;n<inpn;n++) {
			int ex1 = (int) (ego_rand()*fac);
			int ex2 = (int) (ego_rand()*fac);
			int sav = rndarr[ex1];
			rndarr[ex1] = rndarr[ex2];
			rndarr[ex2] = sav;
		}
	}
	return 0;
}
