/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: egmepvm.c	by M.Eichinger, Munich			**
**	Belongs2: EGO		The parallel MD-Simulation programm	**
**	Contents: "PVM" packing functions for PARIX and MPI.		**			
**		  This file is only included if PARIX or MPI is defined **
**  	Modif's:							**
**									**		
**									**
*************************************************************************/



#if defined(PARIX) || defined(PARIXHAL) || defined(MPI)

#include <stdio.h>
#include <string.h>

#include "ego.h"
#include "egmepvm.h"
#include "egalloc.h"
#include "util_lib.h"

#define DEBUG		0
#define	SEND_FLOAT	0	/* set this to 1 ONLY FOR TESTING!! */
#define CHECKSUM	0	/* check communication with checksumms */


/************************************************************************/

static	int	increase_dynbuf_send(int incr);
static	int	increase_dynbuf_recv(int incr);

/***********************************************************************/

static	char	*dynbuf_send;
static	char	*dynbuf_recv;

static	int	dynbuf_here_send=0; 
static	int	dynbuf_here_recv=0; 

static	int	cur_bufsize_send=0,incr_size=0;
static	int	cur_used_send=0;

static	int	cur_bufsize_recv=0;
static	int	cur_used_recv=0;
static	int	cur_expected_recv=0,testnow=0;

/****************************************************************************/

int	pvm_initsend(int encoding)
{
	int a=encoding; a=0;
	
	if (dynbuf_here_send) {
		cur_used_send=0;
		return 0;
	}

	dynbuf_send=(char*) alloc_chars(CHUNK_BOND_EXCL);
	if (dynbuf_send==NULL) {
		errout("pvm_initsend in mepvm failed !\n");
		pex(-1);
	}
	cur_bufsize_send=CHUNK_BOND_EXCL;
	cur_used_send=0;

	dynbuf_here_send=1;
	incr_size=CHUNK_BOND_EXCL;
	
	return 0;
}

/****************************************************************************/
void *pvm_set_readbuf(int size)
{
	int inc;


	cur_expected_recv = size;
	incr_size     = CHUNK_BOND_EXCL;
	cur_used_recv = 0;

	if (!dynbuf_here_recv) {
		dynbuf_recv=(char*) alloc_chars(max(CHUNK_BOND_EXCL,size));
		if (dynbuf_recv==NULL) {
			errout("pvm_initrecv in mepvm failed !\n");
			pex(-1);
		}
		cur_bufsize_recv = max(CHUNK_BOND_EXCL,size);
		dynbuf_here_recv = 1;
	}

	if (size>cur_bufsize_recv) {
		inc=size-cur_bufsize_recv;
		if (increase_dynbuf_recv(inc)==-1)	{
			errout("Could not allocate memory in mepvm for recv!\n");
			pex(-1);
		}
	}
	testnow=1;
#if DEBUG
	if ((dynbuf_recv==NULL)||(!dynbuf_here_recv)) {
		errout("No dynbuf_recv in mepvm #2\n");
		pex(-1);
	}
#endif
	return (void *)dynbuf_recv;
}

/*****************************************************************************/

int	pvm_pkdouble(double *val,int count,int dummy)
{
	int	new_bytes,inc;

	if (count<=0) return 0;

	if (dummy != 1) { errout("Stride %d not implemented!\n",dummy);pex(-1);}
	
	if (!dynbuf_here_send) {
		errout("Fatal error: no mepvm buffer for pkdouble.\n");
		pex(-1);
	}
#if SEND_FLOAT
	new_bytes=sizeof(float)*count;
#else
	new_bytes=sizeof(double)*count;
#endif


#if DEBUG
txtout("dbl: cur_used=%d, new_bytes=%d, cur_bufsize=%d\n",
cur_used, new_bytes, cur_bufsize);
#endif

	if (cur_used_send+new_bytes>=cur_bufsize_send) {
		inc=max(new_bytes-(cur_bufsize_send-cur_used_send),incr_size);
		if (increase_dynbuf_send(inc)==-1)	return -1;
	}

#if SEND_FLOAT
	{float b;
	int i;
	char 	*flptr;
	flptr = (char*) dynbuf_send+cur_used_send;

	for (i=0;i<count;i++) {
		b = (float) val[i]; 
		memcpy(flptr,&b,sizeof(float));
		flptr+=sizeof(float);
	}
	}
#else
	memcpy(dynbuf_send+cur_used_send,val,new_bytes);
#endif
	cur_used_send += new_bytes;

	return 0;
}

int	pvm_upkdouble(double *val,int count,int dummy)
{
	int	new_bytes;

	if (count<=0) return 0;

	if (!dynbuf_here_recv) {
		errout("Fatal error: no mepvm buffer for pvm_upkdouble.\n");
		pex(-1);
	}

	if (dummy != 1) {errout("Stride %d not implemented!\n",dummy);pex(-1);}

#if SEND_FLOAT
	new_bytes=sizeof(float)*count;
#else
	new_bytes=sizeof(double)*count;
#endif

	if (cur_used_recv+new_bytes>cur_bufsize_recv) {
		errout("Fatal error upkdouble: buffer should be bigger. %d %d\n",
			cur_used_recv+new_bytes,cur_bufsize_recv);
		pex(-1);
	}

#if SEND_FLOAT
	{float b;
	int i;
	char 	*flptr;

	flptr = (char*) dynbuf_recv+cur_used_recv;
	for (i=0;i<count;i++) {
		memcpy(&b,flptr,sizeof(float));
		val[i] = (double) b;
		flptr+=sizeof(float);
	}
	}
#else
	memcpy(val,dynbuf_recv+cur_used_recv,new_bytes);
#endif

	cur_used_recv += new_bytes;

	return 0;
}

/*****************************************************************************/

int	pvm_pkint(int *val,int count,int dummy)
{
	int	new_bytes,inc;

	if (count<=0) return 0;

	if (dummy != 1) {errout("Stride %d not implemented!\n",dummy);pex(-1);}

	if (!dynbuf_here_send) {
		errout("Fatal error: no mepvm buffer for pkint.\n");
		pex(-1);
	}

	new_bytes=sizeof(int)*count;

	if (cur_used_send+new_bytes>=cur_bufsize_send) {
		inc=max(new_bytes-(cur_bufsize_send-cur_used_send),incr_size);
		if (increase_dynbuf_send(inc)==-1)	return -1;
	}

	memcpy(dynbuf_send+cur_used_send,val,new_bytes);
	cur_used_send+=new_bytes;

	return 0;
}

int	pvm_upkint(int *val,int count,int dummy)
{
	int	new_bytes;

	if (count<=0) return 0;

	if (!dynbuf_here_recv) {
		errout("Fatal error: no mepvm buffer for pvm_upkint.\n");
		pex(-1);
	}

	if (dummy != 1) {errout("Stride %d not implemented!\n",dummy);pex(-1);}
	new_bytes=sizeof(int)*count;

	if (cur_used_recv+new_bytes>cur_bufsize_recv) {
		errout("Fatal error upkint: buffer should be bigger. %d %d\n",
		cur_used_recv+new_bytes,cur_bufsize_recv);
		pex(-1);
	}

	memcpy(val,dynbuf_recv+cur_used_recv,new_bytes);
	cur_used_recv+=new_bytes;

	return 0;
}

/*********************************************************************************/

int	pvm_pkbyte(char *val,int count,int dummy)
{
	int	new_bytes,inc;

	if (count<=0) return 0;

	if (dummy != 1) {errout("Stride %d not implemented!\n",dummy);pex(-1);}

	if (!dynbuf_here_send) {
		errout("Fatal error: no mepvm buffer for pkbyte.\n");
		pex(-1);
	}

	new_bytes=sizeof(char)*count;

	if (cur_used_send+new_bytes>=cur_bufsize_send) {
		inc=max(new_bytes-(cur_bufsize_send-cur_used_send),incr_size);
		if (increase_dynbuf_send(inc)==-1)	return -1;
	}

	memcpy(dynbuf_send+cur_used_send,val,new_bytes);
	cur_used_send+=new_bytes;

	return 0;
}

int	pvm_upkbyte(char *val,int count,int dummy)
{
	int	new_bytes;

	if (count<=0) return 0;

	if (!dynbuf_here_recv) {
		errout("Fatal error: no mepvm buffer for pkstr.\n");
		pex(-1);
	}

	if (dummy != 1) {errout("Stride %d not implemented!\n",dummy);pex(-1);}

	new_bytes=sizeof(char)*count;

	if (cur_used_recv+new_bytes>cur_bufsize_recv) {
		errout("Fatal error upkbyte: buffer should be bigger.\n");
		pex(-1);
	}

	memcpy(val,dynbuf_recv+cur_used_recv,new_bytes);
	cur_used_recv+=new_bytes;

	return 0;
}

/********************************************************************************/

int	pvm_pkstr(char *str)
{
	int	new_bytes,inc;

	if (!dynbuf_here_send) {
		errout("Fatal error: no mepvm buffer for pkstr.\n");
		pex(-1);
	}

	new_bytes=strlen(str)+1;

	if (cur_used_send+new_bytes>=cur_bufsize_send) {
		inc=max(new_bytes-(cur_bufsize_send-cur_used_send),incr_size);
		if (increase_dynbuf_send(inc)==-1)	return -1;
	}

	memcpy(dynbuf_send+cur_used_send,str,new_bytes);
	cur_used_send+=new_bytes;

	return 0;
}

int	pvm_upkstr(char *str)
{
	int	new_bytes;
	
	if (!dynbuf_here_recv) {
		errout("Fatal error: no mepvm buffer for pkstr.\n");
		return -1;
	}

	new_bytes=strlen(dynbuf_recv+cur_used_recv)+1;

	if (cur_used_recv+new_bytes>cur_bufsize_recv) {
		errout("Fatal error upkstr: buffer should be bigger.\n");
		pex(-1);
	}

	memcpy(str,dynbuf_recv+cur_used_recv,new_bytes);
	cur_used_recv+=new_bytes;

	return 0;
}


/*****************************************************************************/
int	pvm_buffer(void ** ptr)
{
#if DEBUG
	if ((dynbuf_send==NULL)||(!dynbuf_here_send)) {
		errout("No dynbuf in mepvm #1\n");
		pex(-1);
	}
#endif

#if CHECKSUM
	{int i,sendsum;
	unsigned char *charp;
	
	charp = (unsigned char*) dynbuf_send;
	for (i=0,sendsum=0;i<cur_used_send;i++) {
		sendsum += (int) charp[i];
	}	
	pvm_pkint(&sendsum,1,1);	
	}	
#endif
	*ptr = (void *)dynbuf_send;
	return (cur_used_send);	/*** size of data in bytes ***/
}


int	check_checksum()
{
#if CHECKSUM
	int oldchecksum=0,i,newsum=0;
	unsigned char *charp;

	if (testnow==0) return 0;
	
	testnow=0;
		
	pvm_upkint(&oldchecksum,1,1);
	
	charp = (unsigned char*) dynbuf_recv;
	for (i=0,newsum=0;i<cur_expected_recv-sizeof(int);i++) {
		newsum += (int) charp[i];
	}
	if (newsum!=oldchecksum) {
		errout("Fatal error: Checksum error send %d   recv %d\n",
			oldchecksum,newsum);
		pex(-1);
	}
	return 0;
#else
	
	return 0;	
#endif	
}

/**************************************/
int	pvm_free()
{
	if (dynbuf_here_send) {
		myfree(dynbuf_send);
		dynbuf_here_send=0;
		cur_bufsize_send=0;
		cur_used_send=0;
	} else {
		errout("Info: No mepvm-buffer allocated yet!\n");
	}

	if (dynbuf_here_recv) {
		myfree(dynbuf_recv);
		dynbuf_here_recv=0;
		cur_bufsize_recv=0;
		cur_used_recv=0;
	} else {
		errout("Info: No mepvm-buffer allocated yet!\n");
	}
	return 0;
}

/**************************************/
static
int	increase_dynbuf_send(int incr)
{
	realloc_chars(&dynbuf_send, incr, cur_bufsize_send);
	cur_bufsize_send+=incr;
	return 0;
}

static
int	increase_dynbuf_recv(int incr)
{
	realloc_chars(&dynbuf_recv, incr, cur_bufsize_recv);
	cur_bufsize_recv+=incr;
	return 0;
}
#else

static int
do_nothing_atall()
{
	return 0;
}

#endif /*** endif PARIX or MPI ***/
