/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: egnolib.c	by M.Eichinger, H.Grubmueller, Munich	**
**	Belongs2: EGO/node	The parallel MD-Simulation programm	**
**                              This library file contains routines used**
**                              on the NODES (slaves).                  **
**	Contents: 	          					**
**									**
**  	Modif's:		See RCS log (wishful thinking of HH)!!	**
**									**		
*************************************************************************/
#define DEBUG 0

#include <math.h>
#include <stdio.h>
#include <string.h>

#include "ego.h"
#include "egalloc.h"
#include "util_lib.h"
#include "egnolib.h"
#include "egbolib.h"
#include "egrotflood.h"
#include "egbound.h"
#include "egpull.h"

static	void   map_forceout_index(int my_number,
				LOAD_MSG_ALL_NODE *global_data,
			  	LOAD_MSG_ONE_NODE *my_data);

/************************************************************************/
void   zero_forces(TRAVELLING_FORCES *forces)
{
	int	j;

	for (j = 0; j < forces->no_atoms; j++) {
		forces->Fx[j] = 0.0;
		forces->Fy[j] = 0.0;
		forces->Fz[j] = 0.0;
	}
}
 

/************************************************************************/
void   rescale(	double scale,
		LOAD_MSG_ALL_NODE *global_data,
		LOAD_MSG_ONE_NODE *my_data)
{
	/* rescaling of velocities; equilibration	 */
	/* the VELOCITIES are rescaled to yield the target temperature; */

	if (global_data->equilibration) {
		int    i;
		double s = 1.0 - scale;
		if ((global_data->rot_correction)||(global_data->flooding)){
			/*** in this case we have to rescale the new-positions ***/
			/*** because otherwise we would have to send also the  ***/
			/*** old coordinates to the other nodes.Grubis rotation***/
			/*** correction would need this ***/		
			for (i=0;i<my_data->no_atoms;i++) {
		 	   my_data->x_new[i] = scale * my_data->x_new[i] + s * my_data->x_old[i];
			   my_data->y_new[i] = scale * my_data->y_new[i] + s * my_data->y_old[i];
			   my_data->z_new[i] = scale * my_data->z_new[i] + s * my_data->z_old[i];
			}
		} else {
			/*** otherwise we use the original algorithm, which ***/
			/*** rescales the old-coordinates ***/			
			for (i=0;i<my_data->no_atoms;i++) {
		 	   my_data->x_old[i] = scale * my_data->x_old[i] + s * my_data->x_new[i];
			   my_data->y_old[i] = scale * my_data->y_old[i] + s * my_data->y_new[i];
			   my_data->z_old[i] = scale * my_data->z_old[i] + s * my_data->z_new[i];
			}
		}	
	}
}

/************************************************************************/
void   init_for_each_step(	ENERGY_INFO *energy,
				ENERGY_INFO *estm_local_energy,
				ENERGY_INFO *dummy_energy,
				LOAD_MSG_ALL_NODE *global_data,
				LOAD_MSG_ONE_NODE *my_data,
				TRAVELLING_FORCES *my_new_forces,
				TRAVELLING_FORCES *my_exact_LRforces,
				int my_number,
				TRAVELLING_PACKETS *packet,
				int *node_status)
{
	int i,level,toplevel;
	

	/************* here we set all energies to zero ******************/
 	dummy_energy->elec     = estm_local_energy->elec     = energy->elec 	= 0.0;
	dummy_energy->vdw      = estm_local_energy->vdw      = energy->vdw 	= 0.0;
	dummy_energy->bond     = estm_local_energy->bond     = energy->bond	= 0.0;
	dummy_energy->angle    = estm_local_energy->angle    = energy->angle 	= 0.0;
	dummy_energy->dihe     = estm_local_energy->dihe     = energy->dihe 	= 0.0;
	dummy_energy->impr     = estm_local_energy->impr     = energy->impr 	= 0.0;
	dummy_energy->kinetic  = estm_local_energy->kinetic  = energy->kinetic 	= 0.0;
	dummy_energy->kinetic_h2o     = estm_local_energy->kinetic_h2o     = energy->kinetic_h2o 	= 0.0;
	dummy_energy->kinetic_protein = estm_local_energy->kinetic_protein = energy->kinetic_protein 	= 0.0;
	dummy_energy->restraint= estm_local_energy->restraint= energy->restraint= 0.0;
	dummy_energy->hbond    = estm_local_energy->hbond    = energy->hbond 	= 0.0;
	dummy_energy->flooding = estm_local_energy->flooding = energy->flooding	= 0.0;
	dummy_energy->sbound   = estm_local_energy->sbound   = energy->sbound	= 0.0;
	dummy_energy->imobil   = estm_local_energy->imobil   = energy->imobil	= 0.0;
	energy->rotation_energy = 0.0;

	dummy_energy->gradinfo[0]= estm_local_energy->gradinfo[0]= energy->gradinfo[0]=0.0;
	dummy_energy->gradinfo[1]= estm_local_energy->gradinfo[1]= energy->gradinfo[1]=0.0;
	
	for (i=0;i<QMENERGIES;i++) {
		estm_local_energy->qm_energy[i] = 0.0;
		dummy_energy->qm_energy[i]      = 0.0;
		energy->qm_energy[i]            = 0.0;		
	}

	/*************** and zero all forces ****************************/
	zero_forces(my_new_forces);
	zero_forces(my_exact_LRforces);
	zero_pressure(&(energy->pressure));
	zero_pressure(&(energy->density));
	zero_pressure(&(estm_local_energy->pressure));
	zero_pressure(&(estm_local_energy->density));
	zero_pressure(&(dummy_energy->pressure));
	zero_pressure(&(dummy_energy->density));
#if DEBUG
txtout("Node %d: in init_for_each_step phase1\n",my_number);
#endif
	/*********************************************************************/
 	/******* copy local coords into right structure format (packet !) ****/
	/*********************************************************************/
	memcpy(	packet[my_number].coordinates.x,my_data->x_new,
		my_data->no_atoms*sizeof(double));

	memcpy(	packet[my_number].coordinates.y,my_data->y_new,
		my_data->no_atoms*sizeof(double));

	memcpy(	packet[my_number].coordinates.z,my_data->z_new,
		my_data->no_atoms*sizeof(double));

	memcpy(	packet[my_number].coordinates.q,my_data->q,
		my_data->no_atoms*sizeof(double));

	memcpy(	packet[my_number].coordinates.typ,my_data->typ,
		my_data->no_atoms*sizeof(int));

	memcpy(	packet[my_number].coordinates.atom_id,my_data->atom_id,
		my_data->no_atoms*sizeof(int));

	memcpy(	packet[my_number].coordinates.atom_sele_flag1,
		my_data->atom_sele_flag1,
		my_data->no_atoms*sizeof(int));

	memcpy(	packet[my_number].coordinates.atom_sele_flag2,
		my_data->atom_sele_flag2,
		my_data->no_atoms*sizeof(int));

	packet[my_number].coordinates.no_atoms = my_data->no_atoms;
	packet[my_number].coordinates.node     = my_number;
#if DEBUG
txtout("Node %d: in init_for_each_step phase2\n",my_number);
#endif
	/*********************************************************************/
	/* copy local branch into right structure format (travelling_branch) */
	/*********************************************************************/
	
	toplevel = my_data->levels-1;
	packet[my_number].branch.levels = toplevel+1;
	packet[my_number].branch.node   = my_number;

	for (level=0;level<=toplevel;level++) {
		int nr_cl=my_data->cl_per_level[level];
		packet[my_number].branch.cl_per_level[level] = nr_cl;

		if (nr_cl<1) {
			errout("Fatal error: Illegal Tree-Structure in shift_and_gather.");
			pex(-1);
		}

		memcpy(	packet[my_number].branch.cluster[level],
			my_data->cluster[level],
			sizeof(struct_cluster)*nr_cl);
	}
#if DEBUG
txtout("Node %d: in init_for_each_step phase3\n",my_number);
#endif
	/*********************************************************************/
	/*** reset status of data-buffers (shared memory of all threads !) ***/
	/*********************************************************************/
	for (i=0;i<global_data->no_nodes;i++) node_status[i]=PACKET_NOT_READY;
	node_status[my_number]=PACKET_FOR_CALC;

	/*** zero forces ***/
	for (i=0;i<4*my_data->no_atoms;i++) my_data->force_total[i] = 0.0;
	memcpy(my_data->force_elecvdw,my_data->force_total,4*my_data->no_atoms*sizeof(double));
	memcpy(my_data->force_bond   ,my_data->force_total,4*my_data->no_atoms*sizeof(double));
	memcpy(my_data->force_elec   ,my_data->force_total,4*my_data->no_atoms*sizeof(double));
	memcpy(my_data->force_vdw    ,my_data->force_total,4*my_data->no_atoms*sizeof(double));
	memcpy(my_data->force_angle  ,my_data->force_total,4*my_data->no_atoms*sizeof(double));
	memcpy(my_data->force_dihe   ,my_data->force_total,4*my_data->no_atoms*sizeof(double));
	memcpy(my_data->force_impr   ,my_data->force_total,4*my_data->no_atoms*sizeof(double));
	memcpy(my_data->force_restr  ,my_data->force_total,4*my_data->no_atoms*sizeof(double));
	memcpy(my_data->force_sbound ,my_data->force_total,4*my_data->no_atoms*sizeof(double));
	memcpy(my_data->force_flood  ,my_data->force_total,4*my_data->no_atoms*sizeof(double));
	memcpy(my_data->force_intern ,my_data->force_total,4*my_data->no_atoms*sizeof(double));
	memcpy(my_data->force_kin    ,my_data->force_total,4*my_data->no_atoms*sizeof(double));

	my_data->pickbond.cur_len = 0;
	my_data->pickelec.cur_len = 0;
	my_data->pickvdw.cur_len  = 0;
	my_data->pickangle.cur_len= 0;
	my_data->pickdihe.cur_len = 0;
	my_data->pickimpr.cur_len = 0;
#if DEBUG
txtout("Node %d: in init_for_each_step phase4\n",my_number);
#endif
	/*** Here we set special pick-force-buffers to 0.0 ***/
	/*** The PICK_BOND buffer must be set to zero as SHAKED atoms will  ***/
	/*** not be in any bond list and will therefore not contribute to   ***/
	/*** any PICK_BOND statement --> avoid uninitialized buffers by	    ***/
	/*** initializing it with zero.					    ***/
	/*** For PICK_ELEC and PICK_VDW the same situation can occur	    ***/
	/*** because of the multiple timestep method and the FMM: Only near ***/
	/*** atoms can be picked. If atoms are seperated too far, there will***/
	/*** be no contribution, so we initialize (or mark) it with 0.0	    ***/	
	for (i=0;i<5*my_data->pickbond.len;i++) {
		my_data->pickbond.force[i] = 0.0;
	}
	for (i=0;i<5*my_data->pickelec.len;i++) {
		my_data->pickelec.force[i] = 0.0;
	}
	for (i=0;i<5*my_data->pickvdw.len;i++) {
		my_data->pickvdw.force[i] = 0.0;
	}

	/**************************************************/
	/*** All things for the QM-interface 		***/
	/**************************************************/
	my_data->qmlocal.cur_len = 0;
	
	for (i=0;i<my_data->no_atoms;i++) {
		my_data->atom_sele_flag1[i] &= (~FLAG_QMDUMMY);
	}
	
	for (i=0;i<my_data->qmlocal.len;i++) {
		my_data->qmlocal.atomi[i]        = 0;
		my_data->qmlocal.incllen[i]      = 0;
		my_data->qmlocal.qm_q[i]         = 0.0;

		my_data->qmlocal.qm_force[3*i+0] = 0.0;
		my_data->qmlocal.qm_force[3*i+1] = 0.0;
		my_data->qmlocal.qm_force[3*i+2] = 0.0;

		my_data->qmlocal.bondf[4*i+0] = 0.0; /*** energy ***/
		my_data->qmlocal.bondf[4*i+1] = 0.0; /*** fx ***/
		my_data->qmlocal.bondf[4*i+2] = 0.0; /*** fy ***/
		my_data->qmlocal.bondf[4*i+3] = 0.0; /*** fz ***/
		
		my_data->qmlocal.anglef[4*i+0] = 0.0; /*** energy ***/
		my_data->qmlocal.anglef[4*i+1] = 0.0; /*** fx ***/
		my_data->qmlocal.anglef[4*i+2] = 0.0; /*** fy ***/
		my_data->qmlocal.anglef[4*i+3] = 0.0; /*** fz ***/

		my_data->qmlocal.dihef[4*i+0] = 0.0; /*** energy ***/
		my_data->qmlocal.dihef[4*i+1] = 0.0; /*** fx ***/
		my_data->qmlocal.dihef[4*i+2] = 0.0; /*** fy ***/
		my_data->qmlocal.dihef[4*i+3] = 0.0; /*** fz ***/
		
		my_data->qmlocal.imprf[4*i+0] = 0.0; /*** energy ***/
		my_data->qmlocal.imprf[4*i+1] = 0.0; /*** fx ***/
		my_data->qmlocal.imprf[4*i+2] = 0.0; /*** fy ***/
		my_data->qmlocal.imprf[4*i+3] = 0.0; /*** fz ***/
	}
	
	for (i=0;i<10*my_data->qmlocal.len;i++) {
		my_data->qmlocal.local[i] = 0.0;
	}
	
	for (i=0;i<MAXDUMMYSPERQMATOM*my_data->qmlocal.len;i++) {
		my_data->qmlocal.incllist[i] = 0;
	}
	
	ini_special(global_data->keep_press_const); /* set up the "special" for SBOUND */
	
	return;
}

/****************************************************************************/

static	int	no_more_atoms = 0;
static	int	no_more_extptr= 0;

void   reinit_node(	LOAD_MSG_ALL_NODE *global_data,
			LOAD_MSG_ONE_NODE *my_data,
			TRAVELLING_FORCES *my_new_forces,
			TRAVELLING_FORCES *my_exact_LRforces,
			int my_number,
			double *rnd_force_std_deviation,
			double *Gamma,
			TRAVELLING_PACKETS *packet)
{	int i,j;

	if  (	(my_data->no_atoms>no_more_atoms) ||
		(max(my_data->ext_angle_ptr,max(my_data->ext_dihe_ptr,my_data->ext_impr_ptr)) >
		 no_more_extptr)) {
		errout("Node %d: in reinit_node after new clustering: now you must implement realloc !\n",my_number);
		errout("Node %d: no_atoms=%d, no_more_atoms=%d\n", my_number,my_data->no_atoms,no_more_atoms);
		errout("Node %d: ext_angle_ptr=%d, ext_dihe_ptr=%d, ext_impr_ptr=%d, no_more_extptr=%d\n", 
			my_number,my_data->ext_angle_ptr,my_data->ext_dihe_ptr, my_data->ext_impr_ptr, no_more_extptr);
		pex(-1);
	}

	compute_Gamma(global_data,my_data,Gamma,rnd_force_std_deviation, my_number);

	i = my_number;
	free_TRAVELLING_PACKETS(&(packet[i]));
	packet[i].coordinates.no_atoms	= my_data->no_atoms;
	packet[i].coordinates.node 	= my_number;

	packet[i].branch.levels		= my_data->levels;
	packet[i].branch.node 		= my_number;
	for (j=0;j<my_data->levels;j++) {
		packet[i].branch.cl_per_level[j]   =my_data->cl_per_level[j];
		packet[i].branch.alloc_per_level[j]=my_data->cl_per_level[j];
	}	
	alloc_TRAVELLING_PACKETS(&(packet[i]));

	my_new_forces->no_atoms 	= my_data->no_atoms;
	zero_forces(my_new_forces);

	my_exact_LRforces->no_atoms	= my_data->no_atoms;
	zero_forces(my_exact_LRforces);

	map_forceout_index(my_number,global_data,my_data);
	
	make_inv_rot_map(my_number,my_data,global_data);

	for (i=0;i<my_data->int_bond_ptr;i++)
		if (my_data->int_bond_list[i]!=END_OF_LIST)
			my_data->int_bond_list[i] -= global_data->first_atom[my_number];

}

/************************************************************************/
void   init_node(	LOAD_MSG_ALL_NODE *global_data,
			LOAD_MSG_ONE_NODE *my_data,
			TRAVELLING_FORCES *my_new_forces,
			TRAVELLING_FORCES *my_exact_LRforces,
			int my_number,
			double **rnd_force_std_deviation,
			double **Gamma,
			TRAVELLING_PACKETS *packet,
			EXT_3_ATOM **ext_angle,
			EXT_4_ATOM **ext_dihe,
			EXT_4_ATOM **ext_impr)
{	int	i,j;

#if DEBUG
errout("Node %d: in init_node\n",my_number);
#endif
	if (global_data->no_nodes>1) 
			no_more_atoms = (int)(2*global_data->maxatoms_per_node);
		else 	no_more_atoms = my_data->no_atoms+10;

	no_more_extptr= 100*max(my_data->ext_angle_ptr,
		   	 max( my_data->ext_dihe_ptr,
			 my_data->ext_impr_ptr));
	if (no_more_extptr<100) no_more_extptr=100;

	set_allocinfo("init_node");	
	ini_rot_flood(global_data,no_more_atoms);
 	ini_pull(my_data,global_data,global_data->atom_sele_flag1,no_more_atoms);
	ini_range(global_data->sbound_range); /* set up the range for SBOUND */

	/*******************************************************************/
	/******** for stochastic boundry ***********************************/	
	*Gamma                   = alloc_doubles(no_more_atoms);
	*rnd_force_std_deviation = alloc_doubles(no_more_atoms);
	compute_Gamma(global_data,my_data,*Gamma,*rnd_force_std_deviation, my_number);
#if DEBUG
	errout("Node %d: in init_node after compute_Gamma\n", my_number);
#endif
	/****************************************************************/

	for (i=0;i<global_data->no_nodes;i++) {
		packet[i].coordinates.no_atoms	= my_data->no_atoms;
		packet[i].coordinates.node 	= my_number;

		packet[i].branch.levels		= my_data->levels;
		packet[i].branch.node 		= my_number;
		for (j=0;j<my_data->levels;j++) {
			packet[i].branch.cl_per_level[j]   =my_data->cl_per_level[j];
			packet[i].branch.alloc_per_level[j]=my_data->cl_per_level[j];
		}	

		alloc_TRAVELLING_PACKETS(&(packet[i]));
	}

	*ext_angle= alloc_EXT_3_ATOMs(no_more_extptr);
	*ext_dihe = alloc_EXT_4_ATOMs(no_more_extptr);
	*ext_impr = alloc_EXT_4_ATOMs(no_more_extptr);
#if DEBUG
	errout("Node %d: in init_node after alloc_TRAVELLING_PACKETS\n", my_number);
#endif
	
	/********************************************************************/
	my_new_forces->no_atoms		= no_more_atoms;
	my_new_forces->node		= my_number;
	alloc_TRAVELLING_FORCES(my_new_forces);
	my_new_forces->no_atoms 	= my_data->no_atoms;
	zero_forces(my_new_forces);

#if DEBUG
	errout("Node %d: in init_node beg:MEMXXX\n", my_number);
#endif
	my_exact_LRforces->no_atoms	= no_more_atoms;
	my_exact_LRforces->node		= my_number;
	alloc_TRAVELLING_FORCES(my_exact_LRforces);
	my_exact_LRforces->no_atoms	= my_data->no_atoms;
	zero_forces(my_exact_LRforces);

	/**** All things for force-selection output ****/
	my_data->atom_sele_flag1 	= alloc_ints(no_more_atoms);
	my_data->atom_sele_flag2 	= alloc_ints(no_more_atoms);

	map_forceout_index(my_number,global_data,my_data);

	alloc_temporary(global_data);
	if (global_data->rotation_mode) {
	  my_data->inv_rotation_map = alloc_ints(global_data->nr_of_rot_atoms);
	}
	make_inv_rot_map(my_number,my_data,global_data);

	/*** factor 4 because of E,fx,fy,fz ***/
	my_data->force_total		= alloc_doubles(no_more_atoms*4);
	my_data->force_elecvdw		= alloc_doubles(no_more_atoms*4);
	my_data->force_bond		= alloc_doubles(no_more_atoms*4);
	my_data->force_elec		= alloc_doubles(no_more_atoms*4);
	my_data->force_vdw		= alloc_doubles(no_more_atoms*4);
	my_data->force_angle		= alloc_doubles(no_more_atoms*4);
	my_data->force_dihe		= alloc_doubles(no_more_atoms*4);
	my_data->force_impr		= alloc_doubles(no_more_atoms*4);
	my_data->force_restr		= alloc_doubles(no_more_atoms*4);
	my_data->force_sbound		= alloc_doubles(no_more_atoms*4);
	my_data->force_flood		= alloc_doubles(no_more_atoms*4);
	my_data->force_intern		= alloc_doubles(no_more_atoms*4);
	my_data->force_kin		= alloc_doubles(no_more_atoms*4);

	for (i=0;i<my_data->int_bond_ptr;i++)
		if (my_data->int_bond_list[i]!=END_OF_LIST)
			my_data->int_bond_list[i] -= global_data->first_atom[my_number];

	/*****************************************/
	/**** All things for the QM-Interface ****/
	/*****************************************/
	if (global_data->nr_qmatoms>0) {
#if DEBUG
errout("Node %d: ini qm alloc %d\n",my_number,global_data->nr_qmatoms);
#endif
		/*** I allocate here on every node enough space to ***/
		/*** hold all data of the QM-atoms --> no realloc  ***/
		/*** needed after reclustering.			   ***/
		my_data->qmlocal.len     = global_data->nr_qmatoms;
		my_data->qmlocal.cur_len = 0;
		my_data->qmlocal.atom_id = alloc_ints(my_data->qmlocal.len);
		my_data->qmlocal.atomi   = alloc_ints(my_data->qmlocal.len);
		my_data->qmlocal.incllen = alloc_ints(my_data->qmlocal.len);
		my_data->qmlocal.incllist= alloc_ints(my_data->qmlocal.len*MAXDUMMYSPERQMATOM);		
		my_data->qmlocal.local   = alloc_doubles(my_data->qmlocal.len*10);
		my_data->qmlocal.qm_q	 = alloc_doubles(my_data->qmlocal.len);
		my_data->qmlocal.qm_force= alloc_doubles(my_data->qmlocal.len*3);
		my_data->qmlocal.bondf   = alloc_doubles(my_data->qmlocal.len*4);
		my_data->qmlocal.anglef  = alloc_doubles(my_data->qmlocal.len*4);
		my_data->qmlocal.dihef   = alloc_doubles(my_data->qmlocal.len*4);
		my_data->qmlocal.imprf   = alloc_doubles(my_data->qmlocal.len*4);		
	}

#if DEBUG
errout("Node %d: done with init_node\n",my_number);
#endif
}

/**************************************************************************/
/*** Here we have to map the atom-node-indices to the atom-master-indices */
/*** in order to know which atoms have to be sent to the master 	  */
/**************************************************************************/
static
void   map_forceout_index(	int my_number,
				LOAD_MSG_ALL_NODE *global_data,
			  	LOAD_MSG_ONE_NODE *my_data)
{
	int i;

	my_data->forceout_nr_atoms = 0;
	for (i=0;i<my_data->no_atoms;i++) {
		int j    = global_data->first_atom[my_number] + i;
		int mapi = global_data->new2old_map[j];

		if (mapi!=my_data->atom_id[i]) {
			errout("Node %d: Fatal error: mapi %d atom_id %d are different\n",
				my_number,mapi,my_data->atom_id[i]);
			errout("Node %d: This can only be caused by severe\n",my_number);
			errout("Node %d: programming or hardware errors !\n",my_number);
			pex(-1);
		}

		my_data->atom_sele_flag1[i] = global_data->atom_sele_flag1[mapi];
		my_data->atom_sele_flag2[i] = global_data->atom_sele_flag2[mapi];

		if (my_data->atom_sele_flag1[i]!=0) {
			my_data->forceout_nr_atoms++;
#if DEBUG
			txtout("Atom-Id %5d Res-Id %5d Atom-Str <%s> Res-Str <%s>\n",
				my_data->atom_id[i],my_data->res_id[i],
				my_data->atom_string[i],my_data->res_string[i]);
#endif
		}
	}
}

/************************************************************************/

