/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

#define DEBUG		0
#define DEBUG_PICK	0

#include <math.h>
#include <stdio.h>

#include "ego.h"
#include "egnomath.h"
#include "egqminter.h"
#include "util_lib.h"

static int
sort_to_pickangle(	int *atom_id,
			double E, 
			double *f1,double *f3,
			double angle,STRUCT_PICK_BOND *pickangle);

static int
sort_to_pickdihe(	int *atom_id,
			double E, 
			double *f1,double *f2,double *f3,double *f4,
			double angle,
			STRUCT_PICK_BOND *pickdihe);

/**********************************************************/
/*** calculate internal ANGLE interaction:		***/
/*** energy:  Kr*(Theta-Theta0)^2			***/
/*** force: 2*Kr*(Theta-Theta0)				***/
/**********************************************************/

double angle_energy(	int angle_ptr,
			int *angle_Nr1,
			int *angle_Nr2,
			int *angle_Nr3,
			int *angle_index,
			int *forceout_atom,
			int *atom_id,
			double *x_new,
			double *y_new,
			double *z_new,
			double *theta0,
			double *k_angle,
			double *ubk,
			double *ubd,
			char is_internal,
			double *force_angle,
			int my_first_atom,
			int my_number,
			EXT_3_ATOM *ext_angle,
			STRUCT_PICK_BOND *pickangle,
			LOAD_MSG_ONE_NODE *my_data)
{
	double	cosinus_theta,sinus_theta,delta_theta;
	double	r,kt,a11,a12,a22;
	double	b1_x,b1_y,b1_z,b1_r2;
	double	b2_x,b2_y,b2_z,b2_r2;
	double	rx,ry,rz;
	double	cur_energy,cur_angle;
	double	f1[3],f3[3];	/*** x,y,z-force on atom 1 and atom 3 ***/
	int	forceout_flag[3],myatom_id[3];
	int	onthisnode[3];
	int	i,ai[3];

	onthisnode[0]	 = 1;	/*** default for is_internal==TRUE ***/
	onthisnode[1]	 = 1;	/*** is TRUE for onthisnode	   ***/
	onthisnode[2]	 = 1;

	for (i=0;i<angle_ptr;i++) {
		int	index = angle_index[i];
		/*** ai[0],... are the indices of the atom ***/
		/*** data in the my_data buffers	  ***/
		ai[0] = angle_Nr1[i]-my_first_atom;
		ai[1] = angle_Nr2[i]-my_first_atom;
		ai[2] = angle_Nr3[i]-my_first_atom;

		if (is_internal) {
			b1_x = x_new[ai[0]] - x_new[ai[1]];
			b1_y = y_new[ai[0]] - y_new[ai[1]];
			b1_z = z_new[ai[0]] - z_new[ai[1]];
			
 			b2_x = x_new[ai[2]] - x_new[ai[1]];
			b2_y = y_new[ai[2]] - y_new[ai[1]];
			b2_z = z_new[ai[2]] - z_new[ai[1]];

 			rx   = x_new[ai[2]] - x_new[ai[0]];
			ry   = y_new[ai[2]] - y_new[ai[0]];
			rz   = z_new[ai[2]] - z_new[ai[0]];
			
			forceout_flag[0] = forceout_atom[ai[0]];
			forceout_flag[1] = forceout_atom[ai[1]];
			forceout_flag[2] = forceout_atom[ai[2]];
			myatom_id[0]     = atom_id[ai[0]];
			myatom_id[1]     = atom_id[ai[1]];
			myatom_id[2]     = atom_id[ai[2]];
		} else {
			b1_x = ext_angle[i].a1.x - ext_angle[i].a2.x;
			b1_y = ext_angle[i].a1.y - ext_angle[i].a2.y;
			b1_z = ext_angle[i].a1.z - ext_angle[i].a2.z;
			
			b2_x = ext_angle[i].a3.x - ext_angle[i].a2.x;
			b2_y = ext_angle[i].a3.y - ext_angle[i].a2.y;
			b2_z = ext_angle[i].a3.z - ext_angle[i].a2.z;

			rx   = ext_angle[i].a3.x - ext_angle[i].a1.x;
			ry   = ext_angle[i].a3.y - ext_angle[i].a1.y;
			rz   = ext_angle[i].a3.z - ext_angle[i].a1.z;
			
			forceout_flag[0] = ext_angle[i].a1.forceout_atom;
			forceout_flag[1] = ext_angle[i].a2.forceout_atom;
			forceout_flag[2] = ext_angle[i].a3.forceout_atom;
			myatom_id[0]     = ext_angle[i].a1.atom_id;
			myatom_id[1]     = ext_angle[i].a2.atom_id;
			myatom_id[2]     = ext_angle[i].a3.atom_id;
			onthisnode[0]    = (ext_angle[i].a1.node_id == my_number);
			onthisnode[1]    = (ext_angle[i].a2.node_id == my_number);
			onthisnode[2]    = (ext_angle[i].a3.node_id == my_number);
		}

		b1_r2 = b1_x*b1_x + b1_y*b1_y + b1_z*b1_z;
		b2_r2 = b2_x*b2_x + b2_y*b2_y + b2_z*b2_z;

		r = 1.0 / sqrt(b1_r2*b2_r2);
		cosinus_theta = (b1_x*b2_x + b1_y*b2_y + b1_z*b2_z) * r;

		if      (cosinus_theta> 1.0)	cosinus_theta =  1.0;
		else if (cosinus_theta<-1.0)	cosinus_theta = -1.0;
	
		cur_angle   = acos(cosinus_theta);
		delta_theta = cur_angle - theta0[index];
		kt          = delta_theta * k_angle[index];
		cur_energy  = kt*delta_theta;
		cur_angle  *= RADDEGREEFAC; 

		sinus_theta = sqrt(1.0 - cosinus_theta*cosinus_theta);
		if (sinus_theta<SMALL) sinus_theta = SMALL; /* important for inverse! */

		kt  = 2*kt/sinus_theta;
		a12 = -kt*r;
		kt  =  kt*cosinus_theta;
		a11 =  kt/b1_r2;
		a22 =  kt/b2_r2;

		f1[0] = a11*b1_x + a12*b2_x;
		f1[1] = a11*b1_y + a12*b2_y;
		f1[2] = a11*b1_z + a12*b2_z;

		f3[0] = a22*b2_x + a12*b1_x;
		f3[1] = a22*b2_y + a12*b1_y;
		f3[2] = a22*b2_z + a12*b1_z;

		/*** now we add the Urey-Bradley term ***/
		if (ubk[index]!=0.0) {
			double 	R  = sqrt(rx*rx + ry*ry + rz*rz);
			double	dr = R - ubd[index];
			double	kr = ubk[index] * dr / 2;
			double	bforce[4];
			
			bforce[0]   = kr*dr;	/*** energy ***/					
			cur_energy += 2.0*bforce[0];
			
			kr = 4 * kr / R;
			bforce[1] = kr * rx;		
			bforce[2] = kr * ry;
			bforce[3] = kr * rz;
			
			f1[0] -= bforce[1];
			f1[1] -= bforce[2];
			f1[2] -= bforce[3];		

			f3[0] += bforce[1];
			f3[1] += bforce[2];
			f3[2] += bforce[3];			
		}

		/*** store the forces now at the right place ***/
		if (onthisnode[0]) {
			int	ii=4*ai[0];
			force_angle[ii  ] += cur_energy/3.0; 
			force_angle[ii+1] += f1[0];
			force_angle[ii+2] += f1[1];
			force_angle[ii+3] += f1[2];
		}

		if (onthisnode[1]) {
			int ii=4*ai[1];
			force_angle[ii  ] += cur_energy/3.0; 
			force_angle[ii+1] -= (f1[0] + f3[0]);
			force_angle[ii+2] -= (f1[1] + f3[1]);
			force_angle[ii+3] -= (f1[2] + f3[2]);
		}

		if (onthisnode[2]) {
			int ii=4*ai[2];
			force_angle[ii  ] += cur_energy/3.0; 
			force_angle[ii+1] += f3[0];
			force_angle[ii+2] += f3[1];
			force_angle[ii+3] += f3[2];
		} 
	
		if (	forceout_flag[0] || 
			forceout_flag[1] || 
			forceout_flag[2]  ) {
				
			/*** do all stuff for the QM-part if any ***/
			handle_qmangles(onthisnode,forceout_flag,myatom_id,
					my_data,
					cur_energy,
					f1,f3,ai);
		
			/*** Here we eventually sort the calculated forces ***/
			/*** to the PICK_ANGLE buffer which is sent to the ***/
			/*** master and can be saved to disk.		   ***/		
			if (	(forceout_flag[0]&PICK_ANGLE)&&
		    		(forceout_flag[1]&PICK_ANGLE)&&
		   	 	(forceout_flag[2]&PICK_ANGLE)){
				sort_to_pickangle(myatom_id,cur_energy,f1,f3,
						cur_angle,pickangle);				
			}
		
		}
	}
	return 0.0;
}

/**************************************************************************
*** IMPROPER interaction:    						***
*** energy: Ki*(omega-omeag0)^2						***
*** Atoms (a1, a2, a3) define plane1 and atoms (a2, a3, a4) define	***
*** plane2. The improper angle is the angle defined by the intersection ***
*** of those two planes.						***	
*** This code has been take partially from the X-PLOR distribution 	***
**************************************************************************/

double improper_energy( int impr_ptr,
			int *impr_Nr1,
			int *impr_Nr2,
			int *impr_Nr3,
			int *impr_Nr4,
			int *impr_index,
			int *forceout_atom,
			int *atom_id,
			double *x_new,
			double *y_new,
			double *z_new,
			double *omega0,
			double *k_impr,
			char is_internal,
			double *force_impr,
			int my_first_atom,
			int my_number,
			EXT_4_ATOM *ext_impr,
			STRUCT_PICK_BOND *pickimpr,
			LOAD_MSG_ONE_NODE *my_data)
{
	double cur_energy,cur_angle;
	double CP,SP,DI,DF,PHI;
	double RIJX,RIJY,RIJZ,RJKX,RJKY,RJKZ,RKLX,RKLY,RKLZ;
	double AX,AY,AZ,BX,BY,BZ,CX,CY,CZ;
	double RAR,RBR,RCR;
	double SWITCH,RECSP,RECCP;
	double	f1[3],f2[3],f3[3],f4[3];	
	int	myatom_id[4],forceout_flag[4];
	int	onthisnode[4];
	int	i,ai[4];

	onthisnode[0]	 = 1;	/*** default for is_internal==TRUE ***/
	onthisnode[1]	 = 1;	/*** is TRUE for onthisnode	   ***/
	onthisnode[2]	 = 1;
	onthisnode[3]	 = 1;
			
	for (i=0;i<impr_ptr;i++) {
		int index = impr_index[i]; /*** the index constants list ***/
		ai[0] = impr_Nr1[i]-my_first_atom;
		ai[1] = impr_Nr2[i]-my_first_atom;
		ai[2] = impr_Nr3[i]-my_first_atom;
		ai[3] = impr_Nr4[i]-my_first_atom;

    		/*** first compute differences ***/
		if (is_internal) {
			RIJX = x_new[ai[0]] - x_new[ai[1]];
			RIJY = y_new[ai[0]] - y_new[ai[1]];
			RIJZ = z_new[ai[0]] - z_new[ai[1]];

			RJKX = x_new[ai[1]] - x_new[ai[2]];
			RJKY = y_new[ai[1]] - y_new[ai[2]];
			RJKZ = z_new[ai[1]] - z_new[ai[2]];

			RKLX = x_new[ai[2]] - x_new[ai[3]];
			RKLY = y_new[ai[2]] - y_new[ai[3]];
			RKLZ = z_new[ai[2]] - z_new[ai[3]];

			forceout_flag[0] = forceout_atom[ai[0]];
			forceout_flag[1] = forceout_atom[ai[1]];
			forceout_flag[2] = forceout_atom[ai[2]];
			forceout_flag[3] = forceout_atom[ai[3]];

			myatom_id[0]     = atom_id[ai[0]];
			myatom_id[1]     = atom_id[ai[1]];
			myatom_id[2]     = atom_id[ai[2]];
			myatom_id[3]     = atom_id[ai[3]];
		} else {
			RIJX = ext_impr[i].a1.x - ext_impr[i].a2.x;
			RIJY = ext_impr[i].a1.y - ext_impr[i].a2.y;
			RIJZ = ext_impr[i].a1.z - ext_impr[i].a2.z;

			RJKX = ext_impr[i].a2.x - ext_impr[i].a3.x;
			RJKY = ext_impr[i].a2.y - ext_impr[i].a3.y;
			RJKZ = ext_impr[i].a2.z - ext_impr[i].a3.z;

			RKLX = ext_impr[i].a3.x - ext_impr[i].a4.x;
			RKLY = ext_impr[i].a3.y - ext_impr[i].a4.y;
			RKLZ = ext_impr[i].a3.z - ext_impr[i].a4.z;

			forceout_flag[0] = ext_impr[i].a1.forceout_atom;
			forceout_flag[1] = ext_impr[i].a2.forceout_atom;
			forceout_flag[2] = ext_impr[i].a3.forceout_atom;
			forceout_flag[3] = ext_impr[i].a4.forceout_atom;

			myatom_id[0]       = ext_impr[i].a1.atom_id;
			myatom_id[1]       = ext_impr[i].a2.atom_id;
			myatom_id[2]       = ext_impr[i].a3.atom_id;
			myatom_id[3]       = ext_impr[i].a4.atom_id;
			
			onthisnode[0]    = (ext_impr[i].a1.node_id==my_number);
			onthisnode[1]    = (ext_impr[i].a2.node_id==my_number);
			onthisnode[2]    = (ext_impr[i].a3.node_id==my_number);
			onthisnode[3]    = (ext_impr[i].a4.node_id==my_number);			
		}

		/*** now compute A= RIJ X RJK, B= RJK X RKL,	***/
		/*** C= RJK X ( RIJ X RJK ) 			***/
		AX = RIJY * RJKZ - RIJZ * RJKY;
		AY = RIJZ * RJKX - RIJX * RJKZ;
		AZ = RIJX * RJKY - RIJY * RJKX;

		BX = RJKY * RKLZ - RKLY * RJKZ;
		BY = RJKZ * RKLX - RKLZ * RJKX;
		BZ = RJKX * RKLY - RKLX * RJKY;

		CX = RJKY * AZ - RJKZ * AY;
		CY = RJKZ * AX - RJKX * AZ;
		CZ = RJKX * AY - RJKY * AX;

		/*** compute the norm of A, B, C and set to MCONST if too***/
		/*** small than SMALL					 ***/
 		RAR = 1.0 / sqrt(max(SMALL,AX*AX + AY*AY + AZ*AZ));
 		RBR = 1.0 / sqrt(max(SMALL,BX*BX + BY*BY + BZ*BZ));
		RCR = 1.0 / sqrt(max(SMALL,CX*CX + CY*CY + CZ*CZ));

		/* normalize A, B and C */
		AX *= RAR;	AY *= RAR;	AZ *= RAR;
		BX *= RBR;	BY *= RBR;	BZ *= RBR;
		CX *= RCR;	CY *= RCR;	CZ *= RCR;

    		/* compute CP=cos(phi)  and SP=sin(phi) */
		CP = AX*BX + AY*BY + AZ*BZ;
		SP = CX*BX + CY*BY + CZ*BZ;

		/*** compute PHI ( make sure CP within boundaries and get ***/
		/*** sign from SP ) */
		PHI = -copysign(acos(min(1.0,max(-1.0,CP))),SP);
		cur_angle = PHI*RADDEGREEFAC;

		/* take the difference between phi and phi0 */
		DI = PHI - omega0[index];

		/* the difference phi-phi0 is always mapped between -pi and +pi */
    		DI += 4*PI;
    		while (DI > PI) DI -= 2*PI;

    		DF = k_impr[index] * DI;
		cur_energy = DI * DF;

		/* now compute the force */
		DF = 2*DF;

		/**********************************************************
		*** compute heavyside function:				***	
		***		  /  -1   |SP| > SMALL			***		
		*** SWITCH(SP) = |					***	
		*** 	          \   0   |SP| < SMALL			***
		*** this function allows us to switch between two 	***
		*** alternatives to compute d phi/ d r			***
		*** compute switched and truncated reciprocals of CP and***
		*** SP multiplied by the derivative of the function DF	***	
		**********************************************************/
		SWITCH = (fabs(SP)>SMALL) ? -1.0 : 0.0;
		RECSP = DF * SWITCH     * copysign(1.0,SP)/max(fabs(SP),SMALL);
		RECCP = DF *(SWITCH+1.0)* copysign(1.0,CP)/max(fabs(CP),SMALL);

           {/* begin skope */
		/**********************************
		*** d cos( PHI )  d cos ( PHI )	***
		*** ------------, -------------	***
		*** d A             d B		***
		**********************************/
		double DCPAX = RAR * (CP * AX - BX);
		double DCPAY = RAR * (CP * AY - BY);
		double DCPAZ = RAR * (CP * AZ - BZ);
		double DCPBX = RBR * (CP * BX - AX);
		double DCPBY = RBR * (CP * BY - AY);
		double DCPBZ = RBR * (CP * BZ - AZ);

		/**********************************
		*** d sin( PHI )  d sin ( PHI )	***
		*** ------------, -------------	***
		*** d C             d B		***
		**********************************/
		double DSPCX = RCR * (SP * CX - BX);
		double DSPCY = RCR * (SP * CY - BY);
		double DSPCZ = RCR * (SP * CZ - BZ);
		double DSPBX = RBR * (SP * BX - CX);
		double DSPBY = RBR * (SP * BY - CY);
		double DSPBZ = RBR * (SP * BZ - CZ);

		/**********************************************************		
		***    d (phi)						***
		*** DF -------,  etc.					***
		***    d rij						***	
		*** get rid of singularity by using two alternatives	***
		**********************************************************/

		double DPRJKX = RECSP * (DCPAY * RIJZ - DCPAZ * RIJY + DCPBZ * RKLY - DCPBY * RKLZ) +
      				RECCP * (-(RJKY * RIJY + RJKZ * RIJZ) * DSPCX
	       			+ (2 * RJKX * RIJY - RIJX * RJKY) * DSPCY
	       			+ (2 * RJKX * RIJZ - RIJX * RJKZ) * DSPCZ
	       			+ DSPBZ * RKLY - DSPBY * RKLZ);
		double DPRJKY = RECSP * (DCPAZ * RIJX - DCPAX * RIJZ + DCPBX * RKLZ - DCPBZ * RKLX) +
				RECCP * (-(RJKZ * RIJZ + RJKX * RIJX) * DSPCY
				+ (2 * RJKY * RIJZ - RIJY * RJKZ) * DSPCZ
				+ (2 * RJKY * RIJX - RIJY * RJKX) * DSPCX
				+ DSPBX * RKLZ - DSPBZ * RKLX);
		double DPRJKZ = RECSP * (DCPAX * RIJY - DCPAY * RIJX + DCPBY * RKLX - DCPBX * RKLY) +
				RECCP * (-(RJKX * RIJX + RJKY * RIJY) * DSPCZ
				+ (2 * RJKZ * RIJX - RIJZ * RJKX) * DSPCX
				+ (2 * RJKZ * RIJY - RIJZ * RJKY) * DSPCY
				+ DSPBY * RKLX - DSPBX * RKLY);

		f1[0] = RECSP * (RJKY * DCPAZ - DCPAY * RJKZ) +
      				RECCP * ((RJKY * RJKY + RJKZ * RJKZ)*DSPCX - 
					RJKX*RJKY*DSPCY - RJKX*RJKZ*DSPCZ);
		f1[1] = RECSP * (RJKZ * DCPAX - DCPAZ * RJKX) +
      				RECCP * ((RJKZ * RJKZ + RJKX * RJKX)*DSPCY - 
					RJKY*RJKZ*DSPCZ - RJKY*RJKX*DSPCX);
		f1[2] = RECSP * (RJKX * DCPAY - DCPAX * RJKY) +
				RECCP * ((RJKX * RJKX + RJKY * RJKY)*DSPCZ - 
					RJKZ*RJKX*DSPCX - RJKZ*RJKY*DSPCY);

		f2[0] = DPRJKX - f1[0];
		f2[1] = DPRJKY - f1[1];
		f2[2] = DPRJKZ - f1[2];

		f4[0] = RECSP * (RJKY * DCPBZ - DCPBY * RJKZ) +
				RECCP * (RJKY * DSPBZ - DSPBY * RJKZ);
		f4[1] = RECSP * (RJKZ * DCPBX - DCPBZ * RJKX) +
				RECCP * (RJKZ * DSPBX - DSPBZ * RJKX);
		f4[2] = RECSP * (RJKX * DCPBY - DCPBX * RJKY) +
				RECCP * (RJKX * DSPBY - DSPBX * RJKY);

		f3[0] = -f4[0] - DPRJKX;
		f3[1] = -f4[1] - DPRJKY;
		f3[2] = -f4[2] - DPRJKZ;

		if (onthisnode[0]) {
			int ii=4*ai[0];
			force_impr[ii  ] += cur_energy/4.0;
	  		force_impr[ii+1] += f1[0];
	 		force_impr[ii+2] += f1[1];
	  		force_impr[ii+3] += f1[2];
		}
		
		if (onthisnode[1]) {
			int ii=4*ai[1];
			force_impr[ii  ] += cur_energy/4.0;
	 		force_impr[ii+1] += f2[0];
	 		force_impr[ii+2] += f2[1];
	  		force_impr[ii+3] += f2[2];
		}
		
		if (onthisnode[2]) {		
			int ii=4*ai[2];
			force_impr[ii  ] += cur_energy/4.0;
	 		force_impr[ii+1] += f3[0];
	  		force_impr[ii+2] += f3[1];
	  		force_impr[ii+3] += f3[2];
		}
		
		if (onthisnode[3]) {		
			int ii=4*ai[3];
			force_impr[ii  ] += cur_energy/4.0;
	 		force_impr[ii+1] += f4[0];
	  		force_impr[ii+2] += f4[1];
	  		force_impr[ii+3] += f4[2];
		}

		if (	forceout_flag[0] || 
			forceout_flag[1] || 
			forceout_flag[2] || 
			forceout_flag[3]  ) { 
			
			/*** do all stuff for the QM-part if any ***/
			handle_qmimpr(	onthisnode,forceout_flag,myatom_id,
					my_data,
					cur_energy,
					f1,f2,f3,f4,ai);

			/*** Here we eventually sort the calculated forces ***/
			/*** to the PICK_IMPR buffer which is sent to the  ***/
			/*** master and can be saved to disk.		   ***/
			if ((forceout_flag[0]&PICK_IMPR)&&
		   	    (forceout_flag[1]&PICK_IMPR)&&
		   	    (forceout_flag[2]&PICK_IMPR)&&
		  	    (forceout_flag[3]&PICK_IMPR)){
		  	    	/*** here we can use the same function ***/
		  	    	/*** as for dihedrals 		       ***/
				sort_to_pickdihe(myatom_id,	
						cur_energy,
						f1,f2,f3,f4,
						cur_angle,
						pickimpr);				
			}
		}

            }/* end skope */
	} /*** end of loop of all impropers ***/
	return 0.0;
}


/**********************************************************/
/*** calculate internal DIHEDRAL interaction:		***/
/*** energy: Kd*(1+cos(n*Phi+d))			***/
/*** n is the symmetry (repeat)				***/
/*** d is a phase constant				***/
/*** for saving storage space:				***/
/*** n<128: (minimum) atom group at 0 deg   (d=0) 	***/
/*** n>128: (minimum) atom group at 180 Grad  (d=Pi)	***/
/*** 50<=n<=60: general case, where 			***/
/*** n-50 yields the correct n-value.			***/
/*** the dihedral angle is defined as the angle between	***/
/*** two planes:					***/
/*** plane 1 is defined by atoms A0, A1, A2; 		***/	
/*** plane 2 is given by the atoms A1,A2, A3; 		***/
/*** a dihedral angle of 180 deg corresponds to a trans	***/
/*** configuration, 0 deg to a cis configuration; 	***/	
/*** d is NOT simply the angle of the (one) minimum 	***/
/*** position, but a phase which does NOT distinguish 	***/
/*** cis and trans; however, by slowly varying d one can***/
/*** transform CIS into TRANS. 				***/
/**********************************************************/

double dihedral_energy( int dihe_ptr,
			int *dihe_Nr1,
			int *dihe_Nr2,
			int *dihe_Nr3,
			int *dihe_Nr4,
			int *dihe_index,
			int *forceout_atom,
			int *atom_id,
			double *x_new,
			double *y_new,
			double *z_new,
			double *phi0,
			double *k_dihe,
			int *n_dihe,
			char is_internal,
			double *force_dihe,
			int my_first_atom,
			int my_number,
			EXT_4_ATOM *ext_dihe,
			STRUCT_PICK_BOND *pickdihe,
			LOAD_MSG_ONE_NODE *my_data)
{
	double	cur_energy;	
	double	b1_x,b1_y,b1_z,b2_x,b2_y,b2_z,b3_x,b3_y,b3_z;
	double	b1_r2,b1_r,b2_r2,b2_r,b3_r2,b3_r;
	double	cosinus1,cosinus2,cosinus,cosinus_phi;
	double	P,Pd;
	double	f1[3],f2[3],f3[3],f4[3];
	double	myhx;
	double	myhy;
	double	myhz; /* using the variable name hz gives an error on AIX4.2 */
	double	sinus1, sinus2, sinus12;
	double	A,a11,a22,a33,a12,a13,a23;
	double	cosinus_quadrat;
	double	sinus,phi,N;
	int	onthisnode[4],forceout_flag[4],myatom_id[4];
	int	i,ai[4];

	onthisnode[0]	 = 1;	/*** default for is_internal==TRUE ***/
	onthisnode[1]	 = 1;	/*** is TRUE for onthisnode	   ***/
	onthisnode[2]	 = 1;
	onthisnode[3]	 = 1;

	for (i=0;i<dihe_ptr;i++) {
		int	index = dihe_index[i];
		ai[0] = dihe_Nr1[i] - my_first_atom;
		ai[1] = dihe_Nr2[i] - my_first_atom;
		ai[2] = dihe_Nr3[i] - my_first_atom;
		ai[3] = dihe_Nr4[i] - my_first_atom;

		if (is_internal) {
			b1_x = x_new[ai[0]] - x_new[ai[1]];
			b1_y = y_new[ai[0]] - y_new[ai[1]];
			b1_z = z_new[ai[0]] - z_new[ai[1]];

			b2_x = x_new[ai[2]] - x_new[ai[1]];
			b2_y = y_new[ai[2]] - y_new[ai[1]];
			b2_z = z_new[ai[2]] - z_new[ai[1]];

			b3_x = x_new[ai[3]] - x_new[ai[2]];
			b3_y = y_new[ai[3]] - y_new[ai[2]];
			b3_z = z_new[ai[3]] - z_new[ai[2]];

			forceout_flag[0] = forceout_atom[ai[0]];
			forceout_flag[1] = forceout_atom[ai[1]];
			forceout_flag[2] = forceout_atom[ai[2]];
			forceout_flag[3] = forceout_atom[ai[3]];

			myatom_id[0]       = atom_id[ai[0]];
			myatom_id[1]       = atom_id[ai[1]];
			myatom_id[2]       = atom_id[ai[2]];
			myatom_id[3]       = atom_id[ai[3]];
		} else {
 			b1_x = ext_dihe[i].a1.x - ext_dihe[i].a2.x;
			b1_y = ext_dihe[i].a1.y - ext_dihe[i].a2.y;
			b1_z = ext_dihe[i].a1.z - ext_dihe[i].a2.z;

			b2_x = ext_dihe[i].a3.x - ext_dihe[i].a2.x;
			b2_y = ext_dihe[i].a3.y - ext_dihe[i].a2.y;
			b2_z = ext_dihe[i].a3.z - ext_dihe[i].a2.z;

			b3_x = ext_dihe[i].a4.x - ext_dihe[i].a3.x;
			b3_y = ext_dihe[i].a4.y - ext_dihe[i].a3.y;
			b3_z = ext_dihe[i].a4.z - ext_dihe[i].a3.z;

			forceout_flag[0] = ext_dihe[i].a1.forceout_atom;
			forceout_flag[1] = ext_dihe[i].a2.forceout_atom;
			forceout_flag[2] = ext_dihe[i].a3.forceout_atom;
			forceout_flag[3] = ext_dihe[i].a4.forceout_atom;

			myatom_id[0]     = ext_dihe[i].a1.atom_id;
			myatom_id[1]     = ext_dihe[i].a2.atom_id;
			myatom_id[2]     = ext_dihe[i].a3.atom_id;
			myatom_id[3]     = ext_dihe[i].a4.atom_id;
			
			onthisnode[0]    = (ext_dihe[i].a1.node_id==my_number);
			onthisnode[1]    = (ext_dihe[i].a2.node_id==my_number);
			onthisnode[2]    = (ext_dihe[i].a3.node_id==my_number);
			onthisnode[3]    = (ext_dihe[i].a4.node_id==my_number);	
		}	
			
		b1_r2 = 1.0 / (b1_x*b1_x + b1_y*b1_y + b1_z*b1_z);
		b1_r = sqrt(b1_r2);

		b2_r2 = 1.0 / (b2_x*b2_x + b2_y*b2_y + b2_z*b2_z);
		b2_r = sqrt(b2_r2);
		
		b3_r2 = 1.0 / (b3_x*b3_x + b3_y*b3_y + b3_z*b3_z);
		b3_r = sqrt(b3_r2);
		
		/* set sine and cosine */
		cosinus1 = (b1_x*b2_x + b1_y*b2_y + b1_z*b2_z)*b1_r*b2_r;
		cosinus2 = (b2_x*b3_x + b2_y*b3_y + b2_z*b3_z)*b2_r*b3_r;
		cosinus  = (b1_x*b3_x + b1_y*b3_y + b1_z*b3_z)*b1_r*b3_r;
		sinus1   = 1.0 - cosinus1*cosinus1;

		if (sinus1<SMALL) sinus1 = SMALL; /* important for inverse! */
		sinus1 = 1.0 / sinus1;

		sinus2 = 1.0 - cosinus2*cosinus2;
		if (sinus2<SMALL) sinus2 = SMALL; /* important for inverse! */

		sinus2  = 1.0 / sinus2;
		sinus12 = sqrt(sinus1 * sinus2);

		cosinus_phi = (cosinus - cosinus1*cosinus2) * sinus12;
		if (cosinus_phi<-1.0) 		cosinus_phi = -1.0;
		else if (cosinus_phi>1.0) 	cosinus_phi =  1.0;

		sinus =   (b1_y * b3_z - b1_z * b3_y) * b2_x +
			  (b1_z * b3_x - b1_x * b3_z) * b2_y +
			  (b1_x * b3_y - b1_y * b3_x) * b2_z;

		phi = acos(cosinus_phi);
		if (sinus<0.0) phi = -phi; /* make the sign of phi right */

		switch (n_dihe[index]) {
		case 2:
			Pd = 2.0 * cosinus_phi;
			P  = Pd * cosinus_phi;
			break;
		case 130:
			Pd = -2.0 * cosinus_phi;
			P  = 2.0 + Pd * cosinus_phi;
			break;
		case 3:
			cosinus_quadrat = cosinus_phi * cosinus_phi;
			P  = (4.0 * cosinus_quadrat - 3.0) * cosinus_phi + 1.0;
			Pd = 6.0 * cosinus_quadrat - 1.5;
			break;
		case 131:
			cosinus_quadrat = cosinus_phi * cosinus_phi;
			P  = (3.0 - 4.0 * cosinus_quadrat) * cosinus_phi + 1.0;
			Pd = 1.5 - 6.0 * cosinus_quadrat;
			break;
		case 4:
			cosinus_quadrat = cosinus_phi * cosinus_phi;
			P  = 8.0 * (cosinus_quadrat - 1.0) * cosinus_quadrat + 2.0;
			Pd = (16.0 * cosinus_quadrat - 8.0) * cosinus_phi;
			break;
		case 132:
			cosinus_quadrat = cosinus_phi * cosinus_phi;
			P  = 8.0 * (1.0 - cosinus_quadrat) * cosinus_quadrat;
			Pd = (8.0 - 16.0 * cosinus_quadrat) * cosinus_phi;
			break;
		case 6:
			cosinus_quadrat = cosinus_phi * cosinus_phi;
			P  = ((32.0*cosinus_quadrat - 48.0)*cosinus_quadrat + 18.0)*
				cosinus_quadrat;
			Pd = (96.0*(cosinus_quadrat - 1.0)*cosinus_quadrat + 18.0)*
				cosinus_phi;
			break;
		case 134:
			cosinus_quadrat = cosinus_phi * cosinus_phi;
			P  = ((48.0 - 32.0*cosinus_quadrat)*cosinus_quadrat-18.0)*
				cosinus_quadrat + 2.0;
			Pd =  (96.0*(1.0 - cosinus_quadrat)*cosinus_quadrat-18.0)*
				cosinus_phi;
			break;
		default:
			N = (double)(n_dihe[index] - 50);
							/* check for range */
			if (N>=128) N = N - 128;		
			if ((N>=ZERO)&&(N<=10.0)) {
				/* only symmetries between 0 and 10 are valid */
			} else {
				errout("egintern.c (dihedrals): invalid symmetry detected %d\n",N);
				pex(-1);
 			}

			cosinus_quadrat = sqrt ((1.0 - cosinus1 * cosinus1) *
						(1.0 - cosinus2 * cosinus2));

			/* linear dihedral, important for inverse! */
			if (cosinus_quadrat<SMALL) cosinus_quadrat = SMALL;

			sinus = sinus * ((b1_r * b2_r * b3_r) / cosinus_quadrat);
			/* test of value of sinus */
			if ((sinus >= SMALL) || (sinus <= (-SMALL))) {

			} else 
			if ((sinus < SMALL) && (sinus >= ZERO)) {
				sinus = SMALL;	/* important for inverse! */
			} else 
			if ((sinus > (-SMALL)) && (sinus < ZERO)) {
				sinus = -SMALL;	/* important for inverse! */
			}
	
			Pd = cos(N*phi + phi0[index]);
			P = 1.0 + Pd;
			Pd = N * sin(N*phi + phi0[index])/(sinus*2.0);
			break;  /* case default */
	    	} /******* end of switch ******/

		A = k_dihe[index];
		cur_energy = A*P;

		A           *= 2.0 * Pd;
		cosinus_phi *= A;	/* this scales all aij by A */
		sinus12     *= A;

		/* set aij */
		a11 = -cosinus_phi * b1_r2 * sinus1;
		a22 = b2_r2 * (2.0 * cosinus * sinus12 - cosinus_phi * (sinus1 + sinus2));
		a33 = -cosinus_phi * b3_r2 * sinus2;
		a12 = b1_r * b2_r * (cosinus1 * cosinus_phi * sinus1 - cosinus2 * sinus12);
		a13 = b1_r * b3_r * sinus12;
		a23 = b2_r * b3_r * (cosinus2 * cosinus_phi * sinus2 - cosinus1 * sinus12);

		/* force update */
		f1[0] = a11 * b1_x + a12 * b2_x + a13 * b3_x;
		myhx    = a12 * b1_x + a22 * b2_x + a23 * b3_x;
		f4[0] = a13 * b1_x + a23 * b2_x + a33 * b3_x;

		f1[1] = a11 * b1_y + a12 * b2_y + a13 * b3_y;
		myhy    = a12 * b1_y + a22 * b2_y + a23 * b3_y;
		f4[1] = a13 * b1_y + a23 * b2_y + a33 * b3_y;

		f1[2] = a11 * b1_z + a12 * b2_z + a13 * b3_z;
		myhz    = a12 * b1_z + a22 * b2_z + a23 * b3_z;
		f4[2] = a13 * b1_z + a23 * b2_z + a33 * b3_z;

		f2[0] = -f1[0] - myhx;
		f2[1] = -f1[1] - myhy;
		f2[2] = -f1[2] - myhz;

		f3[0] = myhx - f4[0];
		f3[1] = myhy - f4[1];
		f3[2] = myhz - f4[2];

		if (onthisnode[0]) {		
			int ii=4*ai[0];
			force_dihe[ii  ] += cur_energy/4.0;	
			force_dihe[ii+1] += f1[0];
			force_dihe[ii+2] += f1[1];
			force_dihe[ii+3] += f1[2];
		}
		
		if (onthisnode[1]) {
			int ii=4*ai[1];
			force_dihe[ii  ] += cur_energy/4.0;
			force_dihe[ii+1] += f2[0];
			force_dihe[ii+2] += f2[1];
			force_dihe[ii+3] += f2[2];
		}
		
		if (onthisnode[2]) {	
			int ii=4*ai[2];
			force_dihe[ii  ] += cur_energy/4.0;
			force_dihe[ii+1] += f3[0];
			force_dihe[ii+2] += f3[1];
			force_dihe[ii+3] += f3[2];
		}
		
		if (onthisnode[3]) {		
			int ii=4*ai[3];
			force_dihe[ii  ] += cur_energy/4.0;
			force_dihe[ii+1] += f4[0];
			force_dihe[ii+2] += f4[1];
			force_dihe[ii+3] += f4[2];
		}

		if (	forceout_flag[0] || 
			forceout_flag[1] || 
			forceout_flag[2] || 
			forceout_flag[3]  ) {
				
			/*** do all stuff for the QM-part if any ***/	 
			handle_qmdihe(	onthisnode,forceout_flag,myatom_id,
					my_data,
					cur_energy,
					f1,f2,f3,f4,ai);
				
			/*** Here we eventually sort the calculated forces ***/
			/*** to the PICK_IMPR buffer which is sent to the  ***/
			/*** master and can be saved to disk.		   ***/
			if ((forceout_flag[0]&PICK_DIHE)&&
		 	    (forceout_flag[1]&PICK_DIHE)&&
		  	    (forceout_flag[2]&PICK_DIHE)&&
		  	    (forceout_flag[3]&PICK_DIHE)){
				sort_to_pickdihe(myatom_id,	
						cur_energy,
						f1,f2,f3,f4,
						phi*RADDEGREEFAC,
						pickdihe);				
			}
		}
	}

	return 0.0;
}


/******************************************************************************/
/*** We store only the force on the first noted atom pickangle->atom_ids[i  ]**/
/*** and on the third noted atom pickangle->atom_ids[i+2]		     **/	
/*** because the force on the second noted atom pickbond->atom_ids[i+1] is   **/
/*** just the negative sum of the other two forces.			     **/
/******************************************************************************/	
static int
sort_to_pickangle(	int *atom_id,
			double E, 
			double *f1,double *f3,
			double angle,STRUCT_PICK_BOND *pickangle)
{
	int	i,j;
	double aforce1[3],aforce3[3];

	/**** test every picked angle against current angle ****/
	for (i=0,j=0;i<3*pickangle->len;i+=3,j+=8) {

		/*** test the atom in the middle ***/
		if (pickangle->atom_ids[i+1]!=atom_id[1]) continue;

		if (	(pickangle->atom_ids[i  ]==atom_id[0]) &&
			(pickangle->atom_ids[i+2]==atom_id[2]) ) {
			
			/*** thats the right angle, keep forces ***/
			/*** copy them normal order             ***/
			memcpy(aforce1,f1,3*sizeof(double));
			memcpy(aforce3,f3,3*sizeof(double));
		} else
		if (	(pickangle->atom_ids[i  ]==atom_id[2]) &&
			(pickangle->atom_ids[i+2]==atom_id[0]) ) {
			
			/*** thats the right angle, but in reversed order ***/
			/*** copy forces in swaped order                  ***/	
			memcpy(aforce3,f1,3*sizeof(double));
			memcpy(aforce1,f3,3*sizeof(double));
		} else
			continue; /*** not the right angle, try next ***/

		if (pickangle->cur_len>=pickangle->len)
			goto inconsistent_pickangle;

		/*** store forces in the structure which is passed ***/
		/*** to the master for printing later to file	   ***/
		pickangle->force[j  ] = E;
		pickangle->force[j+1] = aforce1[0];
		pickangle->force[j+2] = aforce1[1];
		pickangle->force[j+3] = aforce1[2];
		pickangle->force[j+4] = aforce3[0];
		pickangle->force[j+5] = aforce3[1];
		pickangle->force[j+6] = aforce3[2];
		pickangle->force[j+7] = angle;
		pickangle->cur_len++;
#if DEBUG_PICK
txtout("PICKANGLE %d %d %d + %d: %f   %f %f %f    %f %f %f   %f\n",
pickangle->atom_ids[i  ],
pickangle->atom_ids[i+1],
pickangle->atom_ids[i+2],
pickangle->cur_len,E,
aforce1[0],aforce1[1],aforce1[2],
aforce3[0],aforce3[1],aforce3[2],
angle);
#endif
		return 0;	/*** data of this angle stored	***/
				/*** -> go home 		***/
	}
	return 0;

inconsistent_pickangle:
	errout("Fatal error: Inconsistent data in sort_to_pickangle %d %d %d!\n",
		atom_id[0],atom_id[1],atom_id[2]);
	errout("This is surely a programming (or hardware) bug!\n");
	pex(-1);
	return -1;
}

/******************************************************************************/
/******************************************************************************/

/*** this function is used for dihe and impropper ***/
static int
sort_to_pickdihe(	int *atom_id,
			double E, 
			double *f1,double *f2,double *f3,double *f4,
			double angle,
			STRUCT_PICK_BOND *pickdihe)
{
	int	i,j;
	double aforce1[3],aforce2[3],aforce3[3],aforce4[3];
	
	for (i=0,j=0;i<4*pickdihe->len;i+=4,j+=14) {
		if (	(pickdihe->atom_ids[i  ]==atom_id[0]) &&
			(pickdihe->atom_ids[i+1]==atom_id[1]) &&
			(pickdihe->atom_ids[i+2]==atom_id[2]) &&
			(pickdihe->atom_ids[i+3]==atom_id[3])) {
			
			/*** thats the right dihe, keep forces ***/
			/*** copy them normal order            ***/			
			memcpy(aforce1,f1,3*sizeof(double));
			memcpy(aforce2,f2,3*sizeof(double));
			memcpy(aforce3,f3,3*sizeof(double));
			memcpy(aforce4,f4,3*sizeof(double));
			
		} else
		if (	(pickdihe->atom_ids[i  ]==atom_id[3]) &&
			(pickdihe->atom_ids[i+1]==atom_id[2]) &&
			(pickdihe->atom_ids[i+2]==atom_id[1]) &&
			(pickdihe->atom_ids[i+3]==atom_id[0])) {

			/*** thats the right dihe, but in reversed ***/
			/*** order copy forces in swaped order     ***/				
			memcpy(aforce1,f4,3*sizeof(double));
			memcpy(aforce2,f3,3*sizeof(double));
			memcpy(aforce3,f2,3*sizeof(double));
			memcpy(aforce4,f1,3*sizeof(double));
		} else
			continue; /*** not the right quadruple ***/

		if (pickdihe->cur_len>=pickdihe->len)
			goto inconsistent_pickdihe;

		pickdihe->force[j  ] = E;
		pickdihe->force[j+1] = aforce1[0];
		pickdihe->force[j+2] = aforce1[1];
		pickdihe->force[j+3] = aforce1[2];
		pickdihe->force[j+4] = aforce2[0];
		pickdihe->force[j+5] = aforce2[1];
		pickdihe->force[j+6] = aforce2[2];
		pickdihe->force[j+7] = aforce3[0];
		pickdihe->force[j+8] = aforce3[1];
		pickdihe->force[j+9] = aforce3[2];
		pickdihe->force[j+10]= aforce4[0];
		pickdihe->force[j+11]= aforce4[1];
		pickdihe->force[j+12]= aforce4[2];
		pickdihe->force[j+13]= angle;
		pickdihe->cur_len++;
#if DEBUG_PICK
txtout("PICKANGLE %d %d %d + %d: %f   %f %f %f    %f %f %f   %f %f %f    %f %f %f   %f\n",
pickangle->atom_ids[i  ],
pickangle->atom_ids[i+1],
pickangle->atom_ids[i+2],
pickangle->cur_len,E,
aforce1[0],aforce1[1],aforce1[2],
aforce2[0],aforce2[1],aforce2[2],
aforce3[0],aforce3[1],aforce3[2],
aforce4[0],aforce4[1],aforce4[2],
angle);
#endif
		return 0;   /*** data sorted in -> go home ***/
	}
	return 0;

inconsistent_pickdihe:
	errout("Fatal error: Inconsistent data in sort_to_pickdihe/impr %d %d %d %d!\n",
		atom_id[0],atom_id[1],atom_id[2],atom_id[3]);
	errout("This is surely a programming bug!\n");
	pex(-1);
	return -1;
}
