/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/**-----------------------------------------------------**/
/**--- Conventions:                                  	**/
/**--- -- #define-names in capital letters only;     	**/
/**--- -- comment lines precede the commented code;  	**/
/**--- -- array-bounds start with MXNO_..., to be    	**/
/**--- read as "maximum No of ...";                  	**/
/**--- conversion factors begin with F_...              **/
/** in every structure the doubles should be noted first**/
/** because of alignment exceptions in parix ! 		**/
/** we also include padding-bytes in this case !	**/	
/**-----------------------------------------------------**/


/* In case of memory shortage, EGO can be made to use single precision (floats)
   instead of double precision (doubles). Just change the following #define: */

#define SINGLE 1	/* all in single precision: save memory */
#define DOUBLE 2	/* This is the default */
#define BIGMEM 3	/* Running really large systems might require this.
			   You then have to have a compiler that has long long (64bit)
			   integers. On the SP2, e.g., do: 
			   aimk CC=xlc CFLAGS="-qlonglong" ego_seq
			   You have to manually change all "long int" to "int"
			   before you compile (egbound.c, egcolib.c.
			   However, this has never worked so far. */

#define PRECISION DOUBLE	/* allowed values are DOUBLE, SINGLE and BIGMEM*/

#if PRECISION == SINGLE
#define double float
#endif

#if PRECISION == BIGMEM
#define int long long
#endif


#define max(a,b) ((a) > (b) ? (a) : (b))
#define min(a,b) ((a) < (b) ? (a) : (b))
#define	inrange(a,b,c) (((a)>=(b))&&((a)<=(c)))

#ifdef LINUX
#define EGO_MACHINE "C program on a LINUX computer"
#elif defined(SUN4SOL2)
#define EGO_MACHINE "C program on a SUN computer"
#elif defined(SR8000)
#define EGO_MACHINE "C program on a Hitachi SR8000"
#elif defined(RS6KMP)
#define EGO_MACHINE "C program on an RS6000"
#else
#define EGO_MACHINE "C program on a " MACHINE_TYPE
#endif

void   	pex(int i);	/*** makes the program exit ***/
void 	set_terminate_flag(int flag);
int 	get_terminate_flag();

/********* global constants: *********/

#ifndef TRUE
#define TRUE (1)
#endif

#ifndef FALSE
#define FALSE (0)
#endif

#ifndef PI
#define PI    3.141592653589793
#endif

/*** flags after signals ***/
#define	DONT_STOP	0
#define AFTER_STEP	1	/*** finish after next integration step ***/
#define AFTER_FILE	2	/*** finish after next restart file     ***/

#define	NEUTRAL		0
#define POLAR		1
#define CHARGED		2

#define	WORK_ON_PACKET		1
#define PACKET_FOR_CALC		2
#define PACKET_DONE		3
#define	PACKET_NOT_READY	4
#define WORK_DONE		5

#define	COMEFFORT_ALL		0
#define COMEFFORT_REDUCED	1
#define COMEFFORT_SKIP		2

#define MAX_NO_LEVELS		7  /* Max No of hierarchy levels for FAMUSAMM */

#define MXNO_USERENERGIES	20 /* Max No of user defined energies	  */
#define MXNO_USERINTS		20 /* Max No of user defined ints         */
#define MXNO_USERDOUBLES	20 /* Max No of user defined doubles      */
#define LEN_CTLDESCR		60 /* String len for ctl file description */

#define LEN_ATOM_TYPE		4 /* No of letters to identify atom types    */
#define LEN_RES_TYPE		4 /* No of letters to identify residue types */
#define LEN_SEGID_TYPE		4 /* No of letters to identify residue types */

/* explicit hbond calculation: not yet implemented */
#define MXNO_HYDROGEN_TYPES	10
#define MXNO_DONORS  		some_number
#define MXNO_ACCEPTORS  	MXNO_DONORS
#define MXSIZ_HBOND_AB_MATRIX 	5 	/* maximum size of van-der-Waals? */
				   	/* A & B matrices */
#define MXNO_HBOND_PARAM 	10	/* maximum No of hbond parameters, 
					   must be > 0!!! */
#define F_HBOND  		0.2

#define MXNO_HBOND_NODES 0	/* nodes entirely devoted to Hbond 
				 * calculation: not yet implemented! */

#define MAX_STRING_SIZE		1000	/*** mod 4 should be 0 ***/
#define MINNO_NODES		2
#define MINNO_ATOMS_PER_NODE	5
#define MXNO_ATOM_TYPES		80
#define MXNO_RES_TYPES		200

#define LIMIT_NO_ATOMS		2000	/* Threshhold to suppress printout of
					 * energ. every step for small molec */
						
#define END_OF_LIST		-1	/* mark for end of the exclusion list*/

#define MXNO_ANGLE_PARAM   (8*MXNO_ATOM_TYPES)	 /*max# of angle params       */
#define MXNO_DIHED_PARAM   (8*MXNO_ATOM_TYPES)   /*max# of dihedral param     */
#define MXNO_IMPROP_PARAM  (MXNO_DIHED_PARAM+10) /*max# of improper param     */
#define	MXNO_HBOND_PARAMS	100		 /*max# of HBond-params, only */					
						 /*used in interface program  */

#define CHUNK_BOND_EXCL		1000	/* used for alloc integer lists */

#define MXNO_SHAKE_ITERATIONS	500	/* SHAKE: */
#if PRECISION == DOUBLE
#define SHAKE_TOLERANCE		1.0E-6
#else
#define SHAKE_TOLERANCE		1.0E-4
#endif

#define LOAD_DISP_NUM		10	/* number of LOAD # in one line */

				/* implicit hbond calculation:              */
#define AEXP 4			/* only these exponents for the hbond       */
#define REXP 6			/* interaction are supported by the program */
#define HAEX 4
#define AAEX 2

/*** different timers for profiling ***/
#define CALC_TREE1		0
#define CALC_TREE2		1
#define CALC_OTHER		2	
#define CALC_EXACT		3
#define	ORGA_STUFF		4
#define COMMUN1			5
#define COMMUN2			6
#define COMMUN_ENE		7
#define COMMUN_KO		8
#define NUM_TIMERS		9			

/***** for parameter-files **********/
#define MXNO_FIELDS_VDW		5	/* max# of data fields per line of
				 	 * the vdW matrix */
#define MXNO_FIELDS_BOND	10	/* max# of bonds per atom */
#define MXNO_FIELDS_EXCL	20	/* max# of exclusions per atoms */
					/* used only on woexcl.c ! */
#define MXNO_FIELDS_BYTE	10

/* constants: */
/* since double constants are used most, they are denoted by ONE, TWO, ... */
#define ZERO  0.0		/* double ZERO */
#define ONE   1.0		/* double ONE  */
#define TWO   2.0		/* double TWO  */

#define RADDEGREEFAC	57.29577951 /* Conversionfactor from Rad to Degree */

#define SMALL 0.0000001		/* double SMALL   */


/* SBOUND2 parameters for WATER only: */
#define SBOUND_FORCE_H2O 0.0025
#define SBOUND_P1_H2O    49.0

/* SBOUND parameters for everything but water: */
/* #define SBOUND_FORCE		4.8077E-5 */ /* is 1/52 of the water constant 
					     because 52 atoms are in one lipid */
#define SBOUND_FORCE		2.30784E-5 /* is 1/52/2 of the water constant 
					     because 52 atoms are in one lipid
					     but 1/52 leads to clustering in the
					     boundary. This is a trial&error
					     approach! */
#define SBOUND_P1		49.0 /* MUST have the same length as for water! */

#define SBOUND_EXPONENT 	4
#define MXNO_SBOUND_PARAM	14
#define MXNO_ARBOUND_GAUSSIANS	20

/*********** time and energy conversion factors: *************/
#define KBOLTZ 1.987191E-3	/* Boltzman constant in AKMA units: */

				/* electrostatics constants: */
#define EPSILON  1.0		/* EPSILON = default dielectric constant */

#define F_ELSTAT  (332.0636/EPSILON)	/* electrostatics conversion
					 * factor, F_ELSTAT =
					 * 1/(4*PI*EPSILON) in AKMA units */

#define ENERGY_CONVERSION 0.00239/* energy conversion factor
				  * from amu*AA^2/ps^2 to AKMA (kcal/mole): 
				  * ENERGY_CONVERSION = 
				  * kcal / NAvogadro / (amu * Angstrom^2 / ps^2 ) 
				  * multiply an energy in amu*AA^2/ps^2 with 
				  * ENERGY_CONVERSION to get kcal/mole */

#define TIMEFAC  0.04888821	/* time conversion factor from AKMA to
				 * seconds: TIMEFAC = DSQRT(Angstrom^2 *
				 * amu * NAvogadro / kcal) 
				 * = SQRT(ENERGY_CONVERSION) */

#define TIMEDIV  (TIMEFAC*1.0E-12)	/* and the corresponding factor
					 * with 1 ps as the unit of time: */

#define KCALPN	69.57769	/*** Multiply factor for conversion of	***/
				/*** kcal/mol/A to pN			***/

#define MXNO_PULLING_ATOMS  5000

/*** multiply-factor to convert from Angstroem to atomic units (bohr)	***/
#define ANGS_TO_BOHR	1.88972688 

/*** multiply-factor to convert from kcal/mol to Hartree = 2Ry		***/
#define KCAL_TO_HART	0.001594153

/*** multiply-factor to convert forces in atomic units (Hartree/bohr)	***/
/*** to kcal/mol A and to kcal/mol A^2 					***/		
#define FORCE_AU_TO_KCAL 1185.41
#define HESS_AU_TO_KCAL	 2239.5566

#define MIN_TIME_STEP  1.0E-20

					/* for controller_load */
#define MAX_FILES 		40
#define CUR_LISFILES		35

#define SHAKE_FILE_INDEX 	0	
#define COORD_FILE_INDEX 	1
#define EXCL_FILE_INDEX  	2
#define BONDLIST_FILE_INDEX 	11
#define ANGLELIST_FILE_INDEX 	12
#define DIHELIST_FILE_INDEX 	13
#define IMPRLIST_FILE_INDEX 	14
#define VDWPARA_FILE_INDEX      15
#define VDW14PARA_FILE_INDEX    16
#define BONDPARA_FILE_INDEX     17
#define ANGLEPARA_FILE_INDEX    18
#define DIHEPARA_FILE_INDEX     19
#define IMPRPARA_FILE_INDEX     20
#define HBTRAN_FILE_INDEX       21
#define HBACC_FILE_INDEX        22
#define HBDON_FILE_INDEX        23
#define HBAH_FILE_INDEX         24
#define HBBH_FILE_INDEX         25
#define HBHT_FILE_INDEX         26
#define PGA_FILE_INDEX          27
#define TYPE_FILE_INDEX         28
#define UNIT_FILE_INDEX		29
#define FLOOD_FILE_INDEX	30
#define RESTART_FILE_INDEX	31
#define FORCE_FILE_INDEX	32
#define BOUNDARY_FILE_INDEX	33
#define SBOUND_FILE_INDEX	34

#define TMP_RESTART_FILE 	"restart.tmp"

/**********************************************************/
/*** Here come Bit-Flags of atoms, sent to the nodes	***/
/*** in the array *atom_sele_flag1, which control       ***/
/*** the force-output-stuff	and other things        ***/
/*** (Pulling/Rotation, QM-selection, etc)		***/
/**********************************************************/	
#define	FORCE_TOTAL		0x00000001
#define	FORCE_ELECVDW		0x00000002
#define	FORCE_BOND		0x00000004
#define	FORCE_ELEC		0x00000008
#define	FORCE_VDW		0x00000010
#define FORCE_ANGLE		0x00000020
#define FORCE_DIHE		0x00000040
#define FORCE_IMPR		0x00000080
#define FORCE_RESTR		0x00000100
#define FORCE_SBOUND		0x00000200
#define FORCE_INTERN		0x00000400
#define FORCE_FLOOD		0x00000800
#define FORCE_KIN		0x00001000

#define PICK_BOND		0x00002000
#define PICK_ELEC		0x00004000
#define PICK_VDW		0x00008000
#define PICK_ANGLE		0x00010000
#define PICK_DIHE		0x00020000
#define PICK_IMPR		0x00040000
#define FLAG_FORCEOFF		0x00080000
#define FLAG_QMATOM		0x00100000
#define FLAG_QMDUMMY		0x00200000
#define FLAG_SHAKEOFF		0x00400000
#define FLAG_SPEEDOFF		0x00800000
#define FLAG_QMLINK    		0x01000000
#define FLAG_SCHWING    	0x02000000

#define FLAG_PART1		0x08000000
#define FLAG_PART2		0x10000000
#define FLAG_PULL		0x20000000
#define FLAG_PULL_AGAINST	0x40000000

/**-----------------------------------------------------**/
/** here we define flags for rotation and --------------**/
/** immobilization staff sent to the nodes in the array **/
/** *atom_sele_flag2, which take control of rotation, --**/
/** pulling and immobilization -------------------------**/
#define FLAG_ROT                0x00000001
#define FLAG_IMOBIL1            0x00000002
#define FLAG_IMOBIL2            0x00000004
#define FLAG_IMOBIL3            0x00000008
/**-----------------------------------------------------**/

#define FFN_TOTAL	"total_force.out"
#define FFN_ELECVDW	"elecvdw_force.out"
#define FFN_BOND	"bond_force.out"
#define FFN_ELEC	"elec_force.out"
#define FFN_VDW		"vdw_force.out"
#define FFN_ANGLE	"angle_force.out"
#define FFN_DIHE	"dihe_force.out"
#define FFN_IMPR	"impr_force.out"
#define FFN_RESTR	"restr_force.out"
#define FFN_SBOUND	"sbound_force.out"
#define FFN_INTERN	"intern_force.out"
#define FFN_FLOOD	"flood_force.out"
#define FFN_KIN		"kin_force.out"

#define FFN_PICKBOND	"bond_pick.out"
#define FFN_PICKANGLE	"angle_pick.out"
#define FFN_PICKDIHE	"dihe_pick.out"
#define FFN_PICKIMPR	"impr_pick.out"
#define FFN_PICKELEC	"elec_pick.out"
#define FFN_PICKVDW	"vdw_pick.out"

#define	MAX_QMLINK_ATOMS	20
#define QMENERGIES		2
#define QMATOMFORCE_OFF		0
#define QMATOMFORCE_QM		1
#define QMATOMFORCE_CLASSIC	2
#define MAXDUMMYSPERQMATOM	1000
#define FORCEINFOLEN	7	/*** number of floats in *force		***/

#define	QMLINKMODE_ACTIVEPOS	0
#define	QMLINKMODE_FIXEDPOS	1
#define	QMLINKMODE_MINIPOS	2
#define	QMLINKMODE_BAKOWIES	3

#define	QMCHARGEMODE_FIXED	0
#define	QMCHARGEMODE_VARIABLE	1

typedef struct {
	int	len;		/*** number of allocated elements	***/
	int	cur_len;	/*** current number of QM-atoms		***/

	int	len_dummys;	/*** number of allocated elements 	***/
	int	cur_nr_dummys;  /*** actual number of dummy atoms	***/

	int	qmlink_len;	/*** current number of link-atoms	***/
	int	qmlink_p1[MAX_QMLINK_ATOMS];
	int	qmlink_p2[MAX_QMLINK_ATOMS];

	double	qmlink_dist1[MAX_QMLINK_ATOMS];
	double	qmlink_fconst1[MAX_QMLINK_ATOMS];
	double	qmlink_dist2[MAX_QMLINK_ATOMS];
	double	qmlink_fconst2[MAX_QMLINK_ATOMS];
		
	double	offs[3];
	double	mass;
	double	qm_energy[2];	/*** Energy of the QM system		***/
	double	dipol[4];	/*** Px,Py,Pz,P Dipol-Moment		***/
	double	sumlinkforce[3];/*** needed for the translation correct.***/
			
	int	*qmmap;
	int	*flag;		/*** for some flags			***/
	int	*atom_id;	/*** EGO-atom-ids			***/
	int	*nodeatomi;	/*** index of this atom on the node	***/
	int	*node;		/*** the target node 			***/
	double	*ko;		/*** x, y, z				***/
	double	*kolink;	/*** position of link-atom		***/
	double	*elecf;		/*** Elec-forces from EGO		***/
	double  *local;		/*** local expansion			***/
	double  *bondf;		/*** bondforce from EGO			***/
	double  *anglef;	/*** angleforce from EGO		***/
	double  *dihef;		/*** diheforce from EGO			***/	
	double  *imprf;		/*** imprforce from EGO			***/
	
	int	*incllen;	/*** number of dummy atoms for this atom***/
	int 	*incllist;	/*** list of dummy atoms for this atom	***/

	int	*excltype;	/*** 0, 1=total_excl, 2=14_excl		***/	
	int	*dummy_id;
	double	*dummy_q;	/*** charge of the dummy atoms		***/
	double	*dummy_ko;	/*** position of dummy atoms		***/

	double	*qm_force;	/*** forces from the QM-program		***/
	double	*qm_q;		/*** charges from the QM-program	***/
	double	*asize;		/*** Size of an atom			***/
	double	*qcore;		/*** Charge of the atom-core		***/
} STRUCT_QM_INFO;

typedef struct {
	int	len;		/*** number of allocated elements	***/
	int	cur_len;	/*** current number of elements		***/
	double	sumlinkforce[3];
	
	int	*atom_id;	/*** corresponding atom			***/
	int	*atomi;		/*** atom index on the node 		***/
	int	*incllen;	/*** atom numbers of dummy atoms	***/
	int	*incllist;	/*** atom numbers of dummy atoms	***/ 
	double	*local;		/*** local expansion			***/
	double	*qm_force;	/*** forces from the QM-program		***/
	double	*qm_q;		/*** charges from the QM-program	***/
	double  *bondf;		/*** bondforce from EGO			***/
	double  *anglef;	/*** angleforce from EGO		***/
	double  *dihef;		/*** diheforce from EGO			***/	
	double  *imprf;		/*** imprforce from EGO			***/
} STRUCT_QM_LOCALINFO;

typedef struct {
	int	len;		/*** number of allocated elements	***/
	int	cur_len;	/*** current number of elements		***/
	int	*atom_id;	/*** EGO-atom-ids			***/
	float	*force;		/*** E, Fx, Fy, Fz, Kx, Ky, Kz		***/
} STRUCT_FORCE_INFO;

typedef struct {
	int	len;		/*** number of picked bonds	***/
	int	cur_len;	/*** actually scanned in bonds	***/
	int	*atom_ids;	/*** first and second partner	***/
	double	*force;		/*** E, Fx, Fy, Fz, BondLength	***/
} STRUCT_PICK_BOND;		/*** This structure is also used***/
				/*** for VDW and ELEC forces	***/

typedef struct {
	STRUCT_FORCE_INFO	total;
	STRUCT_FORCE_INFO	elecvdw;
	STRUCT_FORCE_INFO	bond;
	STRUCT_FORCE_INFO	elec;
	STRUCT_FORCE_INFO	vdw;
	STRUCT_FORCE_INFO	angle;
	STRUCT_FORCE_INFO	dihe;
	STRUCT_FORCE_INFO	impr;
	STRUCT_FORCE_INFO	restr;
	STRUCT_FORCE_INFO	sbound;
	STRUCT_FORCE_INFO	intern;
	STRUCT_FORCE_INFO	flood;
	STRUCT_FORCE_INFO	kin;

	STRUCT_QM_INFO		qm;

	STRUCT_PICK_BOND	pickbond;
	STRUCT_PICK_BOND	pickelec;
	STRUCT_PICK_BOND	pickvdw;
	STRUCT_PICK_BOND	pickangle;
	STRUCT_PICK_BOND	pickdihe;
	STRUCT_PICK_BOND	pickimpr;
} STRUCT_FORCES;		/*** This structure is only used on the	***/
				/*** master node 			***/

typedef struct {
	double	ko[3];
	double	vel[3];
	double	refp[3];
	double	charge;
	double	beta;
	double	harmon;
	int	atomnr;
	int	atomtype;
	int	residnr;
	int	restype;
#if defined(PARIX) || defined(PARIXHAL)
	int	padding;
#endif
	char	atomtypestr[LEN_ATOM_TYPE+1];
	char	atomstr[LEN_ATOM_TYPE+1];
	char	segidstr[LEN_SEGID_TYPE+1];
	char	residstr[LEN_RES_TYPE+1];
} STRUCT_ATOMDATA;

/**********************************************/
/** structure definition for pressure values **/
/**********************************************/
typedef struct {
	/* force on the sphere */
	double sphere;
	
	/* force on the cylinder */
	double cyl_rad; /* on the round side */
	/* the flat surfaces use the normal planar SBOUND */
		
	/* forces on the six planar faces: 
	 * plane_planar[0] = -X
	 * plane_planar[1] = +X
	 * plane_planar[2] = -Y
	 * plane_planar[3] = +Y
	 * plane_planar[4] = -Z
	 * plane_planar[5] = +Z
	 */
	double plane_planar[6]; 
	
	/* radial forces on the 12 edges of the cube: 
	 * plane_edge[0] = -X, ..., +X; -Y; -Z;
	 * plane_edge[1] = -X, ..., +X; +Y; -Z;
	 * plane_edge[2] = -X, ..., +X; -Y; +Z;
	 * plane_edge[3] = -X, ..., +X; +Y; +Z;
	 * plane_edge[4] = -X; -Y, ..., +Y; -Z;
	 * plane_edge[5] = +X; -Y, ..., +Y; -Z;
	 * plane_edge[6] = -X; -Y, ..., +Y; +Z;
	 * plane_edge[7] = +X; -Y, ..., +Y; +Z;
	 * plane_edge[8] = -X; -Y; -Z, ..., +Z;
	 * plane_edge[9] = +X; -Y; -Z, ..., +Z;
	 * plane_edge[10]= -X; +Y; -Z, ..., +Z;
	 * plane_edge[11]= +X; +Y; -Z, ..., +Z;
	 */
	double plane_edge[12];
	
	/* radial forces on the 8 corners of the cube: 
	 * plane_corner[0] = -X; -Y; -Z;
	 * plane_corner[1] = +X; -Y; -Z;
	 * plane_corner[2] = -X; +Y; -Z;
	 * plane_corner[3] = +X; +Y; -Z;
	 * plane_corner[4] = -X; -Y; +Z;
	 * plane_corner[5] = +X; -Y; +Z;
	 * plane_corner[6] = -X; +Y; +Z;
	 * plane_corner[7] = +X; +Y; +Z;
	 */
	double plane_corner[8];
	
        /* rok140696 Virial Tensor */
        double virial[3][3]; 

        /* rok091096 Total internal Virial */
        double virial_tot;

        /*rok051196 External Virial (interactions with the boundary) */
        double virial_ext[3][3];
        double virial_ext_tot;

        /* rok251096 Total Virial (with all Forces) I'll call it "virial kinetic energy" */
        double virial_ekin[3][3];
        double virial_ekin_tot;
      
        /* rok140696 pressure tensor */
        double pressure_tensor[3][3];
        double pressure_tot;

        /* rok121196 external virial pressure */
        double pressure_ext_tensor[3][3];
        double pressure_ext_tot;
} PRESSURE;


typedef struct {
	int	nr_of_clusters;
	int	nr_of_childs;
	int	*first_cluster;
	int	*last_cluster;
} struct_levelinfo;


typedef struct {
	int	algo;
	int	levels;

	int	branches;
	int	steps;

	int	clusterseq[MAX_NO_LEVELS];

#if defined(PARIX) || defined(PARIXHAL)
	int	padding;
#endif	
} struct_treeinfo;


typedef struct {
	double	bbox[6];
	double	moment[9]; 	/* [0]       = charge		*/
				/* [1][2][3] = Dipol-moment	*/
				/* [4][5]    = Qxx,Qyy		*/
				/* [6][7][8] = Qxy,Qyz,Qzx	*/
	double	local[10];	/* local expansion		*/
				/* [0]       = Potential	*/
				/* [1][2][3] = Gradient		*/	
				/* [4]-[9]   = Hesse-Matrix	*/
				/* [4]=xx, [5]=yy, [6]=zz	*/
				/* [7]=xy, [8]=xz, [9]=yz	*/	
	double	cen_geom[4];
	double	cen_refp[3];
	int	child_pos;
	int	status;
	int	nr_childs;
	int	typ;
} struct_cluster;

typedef struct {
	double	bbox[6];
	double	moment[1]; 	/* [0]       = charge		*/
				/* [1][2][3] = Dipol-moment	*/
				/* [4][5]    = Qxx,Qyy		*/
				/* [6][7][8] = Qxy,Qyz,Qzx	*/
	double	cen_geom[4];
	int	child_pos;
	int	status;
	int	nr_childs;
	int	typ;
} struct_cluster_reduced;

typedef struct {
	double sbound;
/* 	double polar_fac; perhaps later needed for depolarisation*/
} STRUCT_POTGRID_ELEMENT;

typedef struct {
	double stepsize;
	double xmin, xmax;
	double ymin, ymax;
	double zmin, zmax;
	int    nx,ny,nz;
	STRUCT_POTGRID_ELEMENT ****pt;
} STRUCT_ADD_POTENTIAL;


/* structure definition for the info packet distributed to each node
 * individually in the loading phase */
typedef struct
{
	int	no_atoms;	/*** number of atoms loaded to this node ***/
	int	no_more_atoms;  /*** space allocated for this node ***/
	int	ex_ptr,ex14_ptr;

	int	shake_number;
	int	forceout_nr_atoms;

	int	int_bond_ptr,ext_bond_ptr;
	int	int_angle_ptr,ext_angle_ptr;
	int	int_dihe_ptr,ext_dihe_ptr;
	int	int_impr_ptr,ext_impr_ptr;

	int	levels;

	int	cl_per_level[MAX_NO_LEVELS];
	int	cl_more_per_level[MAX_NO_LEVELS];
#if defined(PARIX) || defined(PARIXHAL)
	int	padding;
#endif

	STRUCT_PICK_BOND	pickbond;
	STRUCT_PICK_BOND	pickelec;
	STRUCT_PICK_BOND	pickvdw;
	STRUCT_PICK_BOND	pickangle;
	STRUCT_PICK_BOND	pickdihe;
	STRUCT_PICK_BOND	pickimpr;

	STRUCT_QM_LOCALINFO	qmlocal;

	double *x_new;	/*** the most recent coordinates ***/ 
	double *y_new;
	double *z_new;

	double *x_old;	/*** the last coordinates ***/ 
	double *y_old;
	double *z_old;

			/*** all subsequent force arrays have the format ***/
			/*** E,fx,fy,fz. Forces are in opposite direction***/
	double 	*force_total;
	double 	*force_elecvdw;
	double 	*force_bond;
	double	*force_elec;
	double	*force_vdw;
	double	*force_angle;
	double	*force_dihe;
	double	*force_impr;
	double	*force_restr;
	double	*force_sbound;
	double	*force_intern;
	double	*force_flood;
	double	*force_kin;

	int    *atom_id;/*** atom number in coord.lis file, which starts ***/
			/*** with 0 for the first atom. Note that in the ***/
			/*** PDB file atom numbers start with 1		 ***/
 	int    *typ;	/*** atom type index				 ***/
			/*** array of strings of length LEN_ATOM_TYPE 	 ***/
			/*** e.g. HT1, CA, CB, ... 			 ***/
 	char   (*atom_string)[LEN_ATOM_TYPE+1];

 	int    *res_id;	/*** residue number, same as in PDB/PSF file	***/

			/*** array of strings of length LEN_RES_TYPE	***/
			/*** residue string, e.g ARG, LYS,...		***/
	char   (*res_string)[LEN_RES_TYPE+1];	

	double *q;	/*** the charge of the atom			***/
	double *beta;	/*** the friction constant			***/

	double *harmon_x;	/*** harmonic restraint positions	***/
	double *harmon_y;	/*** loaded from the start pos in the   ***/
	double *harmon_z;	/*** coord.lis file.			***/	
	double *harmon_k;	/*** the harmonic constant		***/

	int	*ex_list,*ex14_list;
	int	*int_bond_list,*ext_bond_list;
	
	int	*shake_list;

	int	*int_angle_Nr1,
		*int_angle_Nr2,
		*int_angle_Nr3,
		*int_angle_index;

	int	*ext_angle_Nr1,
		*ext_angle_Nr2,
		*ext_angle_Nr3,
		*ext_angle_index;

	int	*int_dihe_Nr1,
		*int_dihe_Nr2,
		*int_dihe_Nr3,
		*int_dihe_Nr4,
		*int_dihe_index;

	int	*ext_dihe_Nr1,
		*ext_dihe_Nr2,
		*ext_dihe_Nr3,
		*ext_dihe_Nr4,
		*ext_dihe_index;

	int	*int_impr_Nr1,
		*int_impr_Nr2,
		*int_impr_Nr3,
		*int_impr_Nr4,
		*int_impr_index;

	int	*ext_impr_Nr1,
		*ext_impr_Nr2,
		*ext_impr_Nr3,
		*ext_impr_Nr4,
		*ext_impr_index;

	int	*atom_sele_flag1;
        int     *atom_sele_flag2;
        int     *inv_rotation_map;

	struct_cluster *cluster[MAX_NO_LEVELS];
}      LOAD_MSG_ONE_NODE;


/* structure definition for the info packet distributed to all 
   nodes globally in the loading phase */
typedef struct
{
	double	elec14;
	double	fric_coeff;
	double	max_move;
	double	target_temp;
	double	time_step;
	double	tau_T;
	double	total_mass;
        double	f_elstat;
	double	AB[MXNO_ATOM_TYPES][MXNO_ATOM_TYPES];
	double	AB14[MXNO_ATOM_TYPES][MXNO_ATOM_TYPES];
	double	k_bond[MXNO_ATOM_TYPES][MXNO_ATOM_TYPES];
	double	k_angle[MXNO_ANGLE_PARAM];
	double	theta0[MXNO_ANGLE_PARAM];
	double	ubk[MXNO_ANGLE_PARAM]; /*** Urey-Bradly ***/
	double	ubd[MXNO_ANGLE_PARAM];
	double	k_dihe[MXNO_DIHED_PARAM];
	double	phi0[MXNO_DIHED_PARAM];
	double	k_impr[MXNO_IMPROP_PARAM];
	double	omega0[MXNO_IMPROP_PARAM];
	double	outer_classrad[MAX_NO_LEVELS];

	/*** all doubles for flooding ***/
	double	flood_energy_ini;
	double	flood_energy_final;
	double	flood_energy_inc;
	double	flood_time_const;
	double	flood_energy_average;
	double	flood_energy_strength;
	double	flood_scale,flood_distance,flood_time_distance;

	/*** all doubles for ARBITRARY SHAPED SBOUND ***/
	double arbs_water_shell_thickness;
	double arbs_surface_level;
	double arbs_gauss_height[MXNO_ARBOUND_GAUSSIANS];
	double arbs_gauss_y[3][MXNO_ARBOUND_GAUSSIANS];
	double arbs_gauss_w[3][3][MXNO_ARBOUND_GAUSSIANS];
	double arbs_gauss_rho2[3][MXNO_ARBOUND_GAUSSIANS];

	/*** all doubles for SBOUND ***/
	double	sbound_droff[MXNO_SBOUND_PARAM]; /* -x, +x, -y, +y, -z, +z */
						 /* ori_x, ori_y, ori_z, R */
						 /* curvature		*/
						 /* cyl_x, cyl_y, cyl_R */
	double	sbound_fbetagradient,sbound_layerthick;
  /********* user defined things ****/
	double	userdouble[MXNO_USERDOUBLES];
  /********* immobilization stuff *****/
	double	selimob_mass[3];
	double	selimob_const;
	double	schwing_offs;
  /******all for pulling ******/
	double	pull_vel;
	double	pull_dir[3];
	double	pull_const[2];
	double	refC[9]; 	/*** Reference center of mass of immob. ***/
	double	deltaimob[9];	/*** deviation from that center 	***/
	double	pullspringpos[7];  /*** positions of pulling springs ***/
        double  mass_pull_tot[2];  /*** total mass of atoms to be pulled ***/
        double  xcom_pull[2];      /*** CoM of atoms to be pulled  ***/
        double  ycom_pull[2];
        double  zcom_pull[2];
  /*****rotation things ****/
        double  angular_vel;
        double  rotation_dir[3];
        double  rotation_const[3];
        double  rotation_point[3];
        int     rotation_mode;
        double  rotation_sigma;

        int     nr_of_rot_atoms;
        double  *rot_x,*rot_y,*rot_z; /** actual coordinates of rotated atoms **/
        double  *rot1_x,*rot1_y,*rot1_z; /** (actual-1) coordinates of rotated atoms **/
        double  *rot2_x,*rot2_y,*rot2_z; /** (actual-2) coordinates of rotated atoms **/
        double  *rot_t1_x,*rot_t1_y,*rot_t1_z; /** temporary actual coordinates of rotated atoms **/
        double  *rot_t2_x,*rot_t2_y,*rot_t2_z; /** temporary (actual-1) coordinates of rotated atoms **/

        double  *rotation_torque,*rotation_t_torque;
	double  rotation_energy;

  /****QM gedoens****/
	double	qmbox[6];
	double	qmboxext[6];	/*** this is the bounding box of the QM ***/
				/*** area. Must be kept fixed during an ***/
				/*** simulation and loader from restartf***/
	double	qmenergyoffset;
	double	qmexcl[2];
	double	minikrit[2];	/*** stop criterion for minimsation	***/
	double	qmcharge;
        double  dipol[4];
        double  mass_pull[2][MXNO_PULLING_ATOMS];  /*** mass of each atom to be pulled ***/
        int     indpull[2][MXNO_PULLING_ATOMS];    /*** index for each atom to be pulled ***/

        int     nr_of_pull_atoms[2];

	int	newtransrotcor;
	int 	arbs_ncodebook;

			/*** all elements for an ADDITIONAL POTENTIAL GRID ***/
			/*** the grid elements contain for example values  ***/
	     		/*** for the SBound                                ***/
	STRUCT_ADD_POTENTIAL potgrid;
			/*** array of strings of length LEN_RES_TYPE	***/
			/*** residue string, e.g ARG, LYS,...		***/
	int    no_sbound_res_string;
	char   (*sbound_res_string)[LEN_RES_TYPE+1];	
	
	int	keep_press_const; /* flag for constant pressure simulation.
				   * used on master and nodes to decide if
				   * scaling data (new sbounds) should be
				   * sent/received. 
				   * 0 = constant volume
				   * 1 = constant pressure
				   * 2 = constant density ***************/
	int    	restarting;
	int    	special_14_excl;
	int    	num_shake,num_shake_h2o,num_shake_protein;
	int    	no_nodes;
	int    	total_no_atoms,maxatoms_per_node;
        int     total_h2o_atoms,total_protein_atoms;
	int	nr_of_atom_typs;
	int	Analyse_nth;
	int	Restart_nth;
	int	step_of_energy;
	int	cycles_passed;
	int	final_cycle_no;
	int	ex_number;
	int	nr_distclasses;
	int	nr_of_flooding_gaussians;
	int	flood_nr_atoms;

	int	writeout_nr_atoms;
	int	forceout_nr_atoms;
	int	forceout_start_step;
	int	forceout_stride_step;

	int	nr_of_dofls;
	int	n_dihe[MXNO_DIHED_PARAM];
	int	syscallstride;
	int	num_userints;
	int	num_userdoubles;
	int	flag_selimob,flag_pull;
	int	pull_atoms[2];
	int	selimob_start;
	int	start_refC;
	int	nr_qmatoms,qmstart_step,qmcontin_step;
	int	qmnoextelec,qmfixbox;
	int	qmatomforce;
	int	qmcutoff,qmdipol,qmlinkmode,qmchargemode;
	int	extdyn_step[3];
	
	/*** Switches ***/
	int	flooding,rot_correction;
	int	flag_adaptive_flood;
	int    	equilibration;
	int    	minimisation;
	int	shake;
	int    	stochastic;
	int    	harmonic;
	int     sbound_range;
	int	schwing_mode;
	
	int	userint[MXNO_USERINTS];
	int	debug_flags[20];

	struct_treeinfo	treeinfo;
	struct_levelinfo levelinfo[MAX_NO_LEVELS];
	PRESSURE outside_pressure;/* only used on master, not packaged/sent */
	
#ifdef SP2
	/* information for signal forwarding */
	
#endif

	/*** Strings ***/
	char	syscallstring[MAX_STRING_SIZE];
	char	outpath[MAX_STRING_SIZE];
	char	workpath[MAX_STRING_SIZE];
	char	lispath[MAX_STRING_SIZE];
	char	restartfile[MAX_STRING_SIZE];
	char	ctlfile[MAX_STRING_SIZE];	
	char	writeout_atom_selection[MAX_STRING_SIZE];
	char	str_userint[MXNO_USERINTS][LEN_CTLDESCR];
	char	str_userdouble[MXNO_USERDOUBLES][LEN_CTLDESCR];
	char	selimob_str[3][MAX_STRING_SIZE];
	char	pull_str[2][MAX_STRING_SIZE];
        char    rotation_str[MAX_STRING_SIZE];
	char	qmstartstr[MAX_STRING_SIZE];
	char	qmcontinstr[MAX_STRING_SIZE];
	char	idstr[MAX_STRING_SIZE];
	char	extdynstr[MAX_STRING_SIZE];
	char	filenames[MAX_FILES][MAX_STRING_SIZE];
	
	int	*first_atom;
	double	*flood_cm[10];
	double	*flood_matrix[10];
	int	*flood_used_atoms;
	int	*writeout_atoms;
	int	*atom_sele_flag1;
	int	*atom_sele_flag2;
	STRUCT_ATOMDATA	*atomdata;
	int	*new2old_map;
	int	*old2new_map;
        int     *rotation_map;
}      LOAD_MSG_ALL_NODE;

/****************************************************/
/* structure definition for travelling force packet */
/****************************************************/
typedef struct
{
  	int	no_atoms,node;
  	int	len;
#if defined(PARIX) || defined(PARIXHAL)
	int	padding;
#endif
  	double	*Fx,*Fy,*Fz;
}	TRAVELLING_FORCES;

/*********************************************************/
/* structure definition for travelling coordinate packet */
/*********************************************************/
typedef struct
{
  	int	no_atoms,node;
  	int	len;
#if defined(PARIX) || defined(PARIXHAL)
	int	padding;
#endif
  	double	*x,*y,*z;
  	double	*q;
  	int	*typ;
	int	*atom_id;
	int	*atom_sele_flag1;
	int	*atom_sele_flag2;
}      TRAVELLING_COORDINATES;

/********************************************************/
/* structure definition for travelling branches packet **/
/********************************************************/
typedef struct
{
	int	node,len;
	int	levels;
	int	cl_per_level[MAX_NO_LEVELS];
	int	alloc_per_level[MAX_NO_LEVELS];
#if defined(PARIX) || defined(PARIXHAL)
	int	padding;
#endif 
	struct_cluster *cluster[MAX_NO_LEVELS];
}	TRAVELLING_BRANCHES;

/*****************************************************************/
/* definition for travelling coordinates+forces+Branches packets */
/*****************************************************************/
typedef struct
{
	TRAVELLING_COORDINATES	coordinates;
	TRAVELLING_FORCES 	forces;
	TRAVELLING_BRANCHES	branch;

}	TRAVELLING_PACKETS;

/* structure definition for energy information */
typedef struct
{
	double elec;
	double vdw;
	double bond;
	double angle;
	double dihe;
	double impr;
	double kinetic;
        double kinetic_h2o;
	double kinetic_protein;
	double restraint;
	double hbond;
	double flooding;
	double flood_energy_average;
	double flood_energy_strength;
	double flooding_distance;
	double sbound;
	double imobil;
	double qm_energy[QMENERGIES];/*** energies from the QM program. They***/
				     /*** are only known at the Master-node ***/
	double userene[MXNO_USERENERGIES];

	double C[3];	/*** nodepart of center of mass   		***/
	double P[3];	/*** nodepart of momentum	  		***/
	double L[3];	/*** nodepart of angular momentum 		***/
	double imobC[9];/*** nodepart COM of immobilised atoms 		***/
	double qmftot[3],mmftot[3];
	double qmang[3],mmang[3];	
  	double pullspringpos[7]; /*** initial position of the pulled atom  
                                      and average spring elongation difference
                                      (for pulling mode 3) ***/
	double pull_indir[2];
	double pull_outdir[2];

        double rot_force[3];

        double rotation_energy;

	double gradinfo[3];	/*** maximum and average gradient	***/
	double time[NUM_TIMERS];

                    /*** rok140696                                          ***/	
        double dFx; /*** Error in the total force, that results from the    ***/
        double dFy; /*** approximative character of the total force         ***/
        double dFz; /*** Used to correct the Virial of an translated System ***/

        double ekin_tensor[3][3]; /*** rok170696                            ***/
                                  /*** Tensor for kinetic energy            ***/
        double ekin_trace;        /*** rok191196 Trace of ekin tensor       ***/ 

	double special_YP_pos, special_YM_pos; /* position of P43 for special mode */
	int    special_YP_count, special_YM_count; /* count of P43 for special mode */
	
	PRESSURE pressure;/*** elements of pressure on surfaces of SBOUND ***/
	PRESSURE density; /*** number of heavy atoms in SBOUND ***/
	
} ENERGY_INFO;

typedef struct
{
	int    no_nodes;
	int    *no_atoms;
	char   **name;
}      LOAD_DISTRIBUTION;

/**********************************************************************/

typedef struct
{
	double	x;
	double	y;
	double	z;

	int	node_id;
	int	forceout_atom;
	int	atom_id;
#if defined(PARIX) || defined(PARIXHAL)
	int	padding;
#endif	
} EXT_ATOM;

typedef struct
{
	EXT_ATOM a1,a2,a3,a4;
} EXT_4_ATOM;

typedef struct
{
	EXT_ATOM a1,a2,a3;
} EXT_3_ATOM;





