/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/***************************************************************************/
/* egpressure.c                                                            */
/*                                                                         */
/* Berechnung des Drucks mittels einer Einfachsumme                        */
/* Lit.:                                                                   */
/*                                                                         */
/* rok160496      R.Kossmann                                               */
/***************************************************************************/

#include <stdio.h>
#include <math.h>

#include "ego.h"
#include "util_lib.h"
#include "egpressure.h"
#include "egbound.h"

#define DEBUG	0

  /* If you use SHAKE for your calculations, you have to calculate the virial and ekin */
  /* terms by using the COM's of the molecules. If you don't want to do so, you can    */
  /* switch this of by setting the define switch NO_COMS to 1, then everything will    */
  /* be calculated using only the atomar view of things.                               */ 
#define NO_COMS 0

static	double	calc_volume(double *sbound_droff);

void calc_virial(ENERGY_INFO *local_energy,
                 LOAD_MSG_ONE_NODE *my_data,
                 LOAD_MSG_ALL_NODE *global_data,
                 TRAVELLING_FORCES *my_new_forces)

{
  int i,j;
  
  double factor;
  double vx,vy,vz;
  int no_coms = NO_COMS;

  int res_id,res_id_old = -1,ty;
  double mass,mass_tot;
  double mol_x,mol_y,mol_z;
  double mol_force_x,mol_force_y,mol_force_z;
  double mol_force_int_x,mol_force_int_y,mol_force_int_z;
  double mol_force_ext_x,mol_force_ext_y,mol_force_ext_z;
  double mol_vx,mol_vy,mol_vz;

  
 
  /* Initialisierung der Akkus fuer den Virialtensor */
  for(i=0;i<3;i++)
    for(j=0;j<3;j++)
      {
      local_energy->pressure.virial[i][j] = 0.0;
      local_energy->pressure.virial_ekin[i][j] = 0.0;
      local_energy->ekin_tensor[i][j]     = 0.0;
      }

  local_energy->pressure.virial_tot = 0.0;
  local_energy->pressure.virial_ekin_tot = 0.0;
  local_energy->pressure.virial_ext_tot  = 0.0;
  local_energy->ekin_trace = 0.0;

  local_energy->dFx = 0.0;
  local_energy->dFy = 0.0;
  local_energy->dFz = 0.0;


  /* Variables for COM calculations */
      mol_force_x = 0.0;
      mol_force_y = 0.0;
      mol_force_z = 0.0;
      mol_force_int_x = 0.0;
      mol_force_int_y = 0.0;
      mol_force_int_z = 0.0;
      mol_force_ext_x = 0.0;
      mol_force_ext_y = 0.0;
      mol_force_ext_z = 0.0;
      mol_x=0.0;
      mol_y=0.0;
      mol_z=0.0;
      mass_tot = 0.0;
      res_id_old=-1;
      res_id=0;
      mol_vx=mol_vy=mol_vz=0.0;

     
for(i = 0;i < my_data->no_atoms; i++)
  {

  if(strcmp(my_data->res_string[i],"TIP3")==0 && global_data->shake && no_coms == 0)
    {
      /* During the first loop *_old holds the velocities allready                   */   
      if (global_data->restarting) 
        {
        vx  = (my_data->x_new[i] - my_data->x_old[i])/global_data->time_step;
        vy  = (my_data->y_new[i] - my_data->y_old[i])/global_data->time_step;
        vz  = (my_data->z_new[i] - my_data->z_old[i])/global_data->time_step;
        } 
      else 
        {
        vx  = my_data->x_old[i];
        vy  = my_data->y_old[i];
        vz  = my_data->z_old[i];
        }

      ty=my_data->typ[i]; 
      mass=global_data->AB[ty][ty];
      res_id=my_data->res_id[i];
      if(res_id_old == -1) res_id_old = res_id;


      mol_force_x += my_new_forces->Fx[i];
      mol_force_y += my_new_forces->Fy[i];
      mol_force_z += my_new_forces->Fz[i];

      mol_force_int_x += my_new_forces->Fx[i] - my_data->force_sbound[4*i + 1];
      mol_force_int_y += my_new_forces->Fy[i] - my_data->force_sbound[4*i + 2];
      mol_force_int_z += my_new_forces->Fz[i] - my_data->force_sbound[4*i + 3];

      mol_force_ext_x += my_data->force_sbound[4*i + 1];
      mol_force_ext_y += my_data->force_sbound[4*i + 2];
      mol_force_ext_z += my_data->force_sbound[4*i + 3];

      mol_x += mass*my_data->x_new[i];
      mol_y += mass*my_data->y_new[i];
      mol_z += mass*my_data->z_new[i];
      mol_vx += mass*vx;
      mol_vy += mass*vy;
      mol_vz += mass*vz;
      mass_tot += mass;
 
      /* if the next atom in list has an other res_id this was the last */
      /*  atom of this molecule and we can calculate the Virialterms    */
      /* for this molecule                                              */
      if(my_data->res_id[i+1]!= res_id)
        {
        mol_x = mol_x/mass_tot;
        mol_y = mol_y/mass_tot;
        mol_z = mol_z/mass_tot;

        mol_vx = mol_vx/mass_tot;
        mol_vy = mol_vy/mass_tot;
        mol_vz = mol_vz/mass_tot;

        local_energy->pressure.virial[0][0]  += -mol_x * mol_force_int_x;
        local_energy->pressure.virial[0][1]  += -mol_x * mol_force_int_y;
        local_energy->pressure.virial[0][2]  += -mol_x * mol_force_int_z;

        local_energy->pressure.virial[1][0]  += -mol_y * mol_force_int_x;
        local_energy->pressure.virial[1][1]  += -mol_y * mol_force_int_y;
        local_energy->pressure.virial[1][2]  += -mol_y * mol_force_int_z;

        local_energy->pressure.virial[2][0]  += -mol_z * mol_force_int_x;
        local_energy->pressure.virial[2][1]  += -mol_z * mol_force_int_y;
        local_energy->pressure.virial[2][2]  += -mol_z * mol_force_int_z;

        local_energy->pressure.virial_ext[0][0] += -mol_x * mol_force_ext_x;
        local_energy->pressure.virial_ext[0][1] += -mol_x * mol_force_ext_y;
        local_energy->pressure.virial_ext[0][2] += -mol_x * mol_force_ext_z;
      
        local_energy->pressure.virial_ext[1][0] += -mol_y * mol_force_ext_x;
        local_energy->pressure.virial_ext[1][1] += -mol_y * mol_force_ext_y;
        local_energy->pressure.virial_ext[1][2] += -mol_y * mol_force_ext_z;
      
        local_energy->pressure.virial_ext[2][0] += -mol_z * mol_force_ext_x;
        local_energy->pressure.virial_ext[2][1] += -mol_z * mol_force_ext_y;
        local_energy->pressure.virial_ext[2][2] += -mol_z * mol_force_ext_z;

        local_energy->pressure.virial_ekin[0][0] += -1.*(-0.5)*mol_x * mol_force_x; 
        local_energy->pressure.virial_ekin[0][1] += -1.*(-0.5)*mol_x * mol_force_y; 
        local_energy->pressure.virial_ekin[0][2] += -1.*(-0.5)*mol_x * mol_force_z;
 
        local_energy->pressure.virial_ekin[1][0] += -1.*(-0.5)*mol_y * mol_force_x; 
        local_energy->pressure.virial_ekin[1][1] += -1.*(-0.5)*mol_y * mol_force_y; 
        local_energy->pressure.virial_ekin[1][2] += -1.*(-0.5)*mol_y * mol_force_z;

        local_energy->pressure.virial_ekin[2][0] += -1.*(-0.5)*mol_z * mol_force_x; 
        local_energy->pressure.virial_ekin[2][1] += -1.*(-0.5)*mol_z * mol_force_y; 
        local_energy->pressure.virial_ekin[2][2] += -1.*(-0.5)*mol_z * mol_force_z;

        factor = 0.5*mass_tot;

        local_energy->ekin_tensor[0][0] += factor * mol_vx * mol_vx;
        local_energy->ekin_tensor[0][1] += factor * mol_vx * mol_vy;
        local_energy->ekin_tensor[0][2] += factor * mol_vx * mol_vz;
  
        local_energy->ekin_tensor[1][0] += factor * mol_vy * mol_vx;
        local_energy->ekin_tensor[1][1] += factor * mol_vy * mol_vy;
        local_energy->ekin_tensor[1][2] += factor * mol_vy * mol_vz;
  
        local_energy->ekin_tensor[2][0] += factor * mol_vz * mol_vx;
        local_energy->ekin_tensor[2][1] += factor * mol_vz * mol_vy;
        local_energy->ekin_tensor[2][2] += factor * mol_vz * mol_vz;

        mol_force_x = 0.0;
        mol_force_y = 0.0;
        mol_force_z = 0.0;
        mol_force_int_x = 0.0;
        mol_force_int_y = 0.0;
        mol_force_int_z = 0.0;
        mol_force_ext_x = 0.0;
        mol_force_ext_y = 0.0;
        mol_force_ext_z = 0.0;
        mol_x=0.0;
        mol_y=0.0;
        mol_z=0.0;
        mol_vx=0.0;
        mol_vy=0.0;
        mol_vz=0.0;
        mass_tot = 0.0;
        }/* endif res_id */
  
      /* Aufsummieren der Gesamtkraft. Aufgrund der verwendeten Naeherungen ist diese */
      /* bei einer Translation der Kugel i.A. von Null verschieden.                   */
      local_energy->dFx += my_new_forces->Fx[i];
      local_energy->dFy += my_new_forces->Fy[i];
      local_energy->dFz += my_new_forces->Fz[i];
    

    }/*endif shake*/
  else /* no shake or not TIP3*/
    {
        /* Calculating the internal Virial only electrostatic an VDW interactions and bond-interactions */
      local_energy->pressure.virial[0][0] += -my_data->x_new[i] * (my_new_forces->Fx[i] - 
                                                                   my_data->force_sbound[4*i + 1]);                              local_energy->pressure.virial[0][1] += -my_data->x_new[i] * (my_new_forces->Fy[i] -
                                                                   my_data->force_sbound[4*i + 2]);
      local_energy->pressure.virial[0][2] += -my_data->x_new[i] * (my_new_forces->Fz[i] -
                                                                   my_data->force_sbound[4*i + 3]);

      local_energy->pressure.virial[1][0] += -my_data->y_new[i] * (my_new_forces->Fx[i] - 
                                                                   my_data->force_sbound[4*i + 1]);
      local_energy->pressure.virial[1][1] += -my_data->y_new[i] * (my_new_forces->Fy[i] -
                                                                   my_data->force_sbound[4*i + 2]);
      local_energy->pressure.virial[1][2] += -my_data->y_new[i] * (my_new_forces->Fz[i] -
                                                                   my_data->force_sbound[4*i + 3]);
 
      local_energy->pressure.virial[2][0] += -my_data->z_new[i] * (my_new_forces->Fx[i] - 
                                                                   my_data->force_sbound[4*i + 1]);
      local_energy->pressure.virial[2][1] += -my_data->z_new[i] * (my_new_forces->Fy[i] -
                                                                   my_data->force_sbound[4*i + 2]);
      local_energy->pressure.virial[2][2] += -my_data->z_new[i] * (my_new_forces->Fz[i] -
                                                                   my_data->force_sbound[4*i + 3]);

        /* Calculating the external virial (Interaction with sbound) */
      local_energy->pressure.virial_ext[0][0] += -my_data->x_new[i] * my_data->force_sbound[4*i + 1];
      local_energy->pressure.virial_ext[0][1] += -my_data->x_new[i] * my_data->force_sbound[4*i + 2];
      local_energy->pressure.virial_ext[0][2] += -my_data->x_new[i] * my_data->force_sbound[4*i + 3];

      local_energy->pressure.virial_ext[1][0] += -my_data->y_new[i] * my_data->force_sbound[4*i + 1];
      local_energy->pressure.virial_ext[1][1] += -my_data->y_new[i] * my_data->force_sbound[4*i + 2];
      local_energy->pressure.virial_ext[1][2] += -my_data->y_new[i] * my_data->force_sbound[4*i + 3];
  
      local_energy->pressure.virial_ext[2][0] += -my_data->z_new[i] * my_data->force_sbound[4*i + 1];
      local_energy->pressure.virial_ext[2][1] += -my_data->z_new[i] * my_data->force_sbound[4*i + 2];
      local_energy->pressure.virial_ext[2][2] += -my_data->z_new[i] * my_data->force_sbound[4*i + 3];
  



       /* Caculating the virial kinetic energy (corresponds to the total virial with all interactions) */
      local_energy->pressure.virial_ekin[0][0] += -1.*(-0.5)*my_data->x_new[i] * my_new_forces->Fx[i];
      local_energy->pressure.virial_ekin[0][1] += -1.*(-0.5)*my_data->x_new[i] * my_new_forces->Fy[i];
      local_energy->pressure.virial_ekin[0][2] += -1.*(-0.5)*my_data->x_new[i] * my_new_forces->Fz[i];

      local_energy->pressure.virial_ekin[1][0] += -1.*(-0.5)*my_data->y_new[i] * my_new_forces->Fx[i];
      local_energy->pressure.virial_ekin[1][1] += -1.*(-0.5)*my_data->y_new[i] * my_new_forces->Fy[i];
      local_energy->pressure.virial_ekin[1][2] += -1.*(-0.5)*my_data->y_new[i] * my_new_forces->Fz[i];
  
      local_energy->pressure.virial_ekin[2][0] += -1.*(-0.5)*my_data->z_new[i] * my_new_forces->Fx[i];
      local_energy->pressure.virial_ekin[2][1] += -1.*(-0.5)*my_data->z_new[i] * my_new_forces->Fy[i];
      local_energy->pressure.virial_ekin[2][2] += -1.*(-0.5)*my_data->z_new[i] * my_new_forces->Fz[i];
  
    
      /* Calculating kinetic energy tensor */
      factor = 0.5 * global_data->AB[my_data->typ[i]][my_data->typ[i]]; 

      /* During the first loop *_old holds the velocities allready                   */   
      if (global_data->restarting) 
        {
        vx  = (my_data->x_new[i] - my_data->x_old[i])/global_data->time_step;
        vy  = (my_data->y_new[i] - my_data->y_old[i])/global_data->time_step;
        vz  = (my_data->z_new[i] - my_data->z_old[i])/global_data->time_step;
        } 
      else 
        {
        vx  = my_data->x_old[i];
        vy  = my_data->y_old[i];
        vz  = my_data->z_old[i];
        }
  
      local_energy->ekin_tensor[0][0] += factor * vx * vx;
      local_energy->ekin_tensor[0][1] += factor * vx * vy;
      local_energy->ekin_tensor[0][2] += factor * vx * vz;
  
      local_energy->ekin_tensor[1][0] += factor * vy * vx;
      local_energy->ekin_tensor[1][1] += factor * vy * vy;
      local_energy->ekin_tensor[1][2] += factor * vy * vz;
  
      local_energy->ekin_tensor[2][0] += factor * vz * vx;
      local_energy->ekin_tensor[2][1] += factor * vz * vy;
      local_energy->ekin_tensor[2][2] += factor * vz * vz;
  
      /* Aufsummieren der Gesamtkraft. Aufgrund der verwendeten Naeherungen ist diese */
      /* bei einer Translation der Kugel i.A. von Null verschieden.                   */
      local_energy->dFx += my_new_forces->Fx[i];
      local_energy->dFy += my_new_forces->Fy[i];
      local_energy->dFz += my_new_forces->Fz[i];

    }/* endelse no  shake */

  }/* endfor i */

  /* Calculating the Trace of the Virial Tensors. */
    for(i=0;i<3;i++)
      {
      local_energy->pressure.virial_tot      += local_energy->pressure.virial[i][i];
      local_energy->pressure.virial_ekin_tot += local_energy->pressure.virial_ekin[i][i];
      local_energy->pressure.virial_ext_tot  += local_energy->pressure.virial_ext[i][i];
      local_energy->ekin_trace += local_energy->ekin_tensor[i][i];
      }

}

void calc_vir_pressure(PRESSURE *press,double Ekin,
			double *sbound_droff,
			double ekin_tensor[3][3])
{
  int i,j;

  double virial_tot = 0.0;
  double volume=Ekin; /*** avoid warnings ***/
  double pspur = 0.0;
  double espur = 0.0;
  double pexspur = 0.0;

  double conv = KCALPN*1.0e8/1.013e5;

  /*First its necessary to calculate the volume of the system */
  volume = calc_volume(sbound_droff);
  if (volume<=0.0) /*** no bounding specified ***/
    {
    for(i=0;i<3;i++)
      for(j=0;j<3;j++)
        {
     	press->pressure_tensor[i][j] = 0.0;
        press->pressure_ext_tensor[i][j] = 0.0;
        }

	press->pressure_tot = 0.0;
        press->pressure_ext_tot = 0.0;
     return; 
  }
  
  
  for(i=0;i<3;i++)
    for(j=0;j<3;j++)
      {
      press->pressure_tensor[i][j] = conv * (2. / volume) * (ekin_tensor[i][j] + 
                                               (0.5 * press->virial[i][j]));

      press->pressure_ext_tensor[i][j] = conv * (-1. / volume) * press->virial_ext[i][j];     
      }


  for(i=0;i<3;i++)
    {
    pspur      += press->pressure_tensor[i][i]/3.;
    pexspur    += press->pressure_ext_tensor[i][i]/3.;
    espur      += ekin_tensor[i][i]/3.;
    }

  virial_tot = press->virial_tot/3.;
 
  press->pressure_tot = conv * (2. / volume)*(espur + (0.5 * virial_tot));
  press->pressure_ext_tot = pexspur;

#if (DEBUG == 3)
for(i=0;i<3;i++)
  for(j=0;j<3;j++)
    txtout("## V[%d][%d] = %g    P[%d][%d] = %g   EK[%d][%d] = %g\n",i,j,press->virial[i][j],
                                              i,j,press->pressure_tensor[i][j],
                                              i,j,ekin_tensor[i][j]);

txtout("## Vir_tot = %g     P_tot = %g  PS= %g EK= %g ES = %g  Volume= %g\n",press->virial_tot,
                                                                 press->pressure_tot,
                                                                 pspur,Ekin,espur, volume);
#endif

}

void print_vir_pressure(PRESSURE *pressure)
{
#if (DEBUG == 1)
   int i;
   txtout("Virial-Pressure-Tensor :\n");
   for(i=0;i<3;i++)
     {
     txtout("   %g  %g  %g\n",
            pressure->pressure_tensor[i][0],
            pressure->pressure_tensor[i][1],
            pressure->pressure_tensor[i][2]);
     }

   txtout("External Virial-Pressure-Tensor :\n");
   for(i=0;i<3;i++)
     {
     txtout("   %g  %g  %g\n",
            pressure->pressure_ext_tensor[i][0],
            pressure->pressure_ext_tensor[i][1],
            pressure->pressure_ext_tensor[i][2]);
     }

   txtout("Virial-Tensor :\n");
   for(i=0;i<3;i++)
     {
     txtout("   %g  %g  %g\n",
            pressure->virial[i][0],
            pressure->virial[i][1],
            pressure->virial[i][2]);
     }

   txtout("Virial-Extern-Tensor :\n");
   for(i=0;i<3;i++)
     {
     txtout("   %g  %g  %g\n",
            pressure->virial_ext[i][0],
            pressure->virial_ext[i][1],
            pressure->virial_ext[i][2]);
     }

   txtout("Virial-KineticEnergy-Tensor :\n");
   for(i=0;i<3;i++)
     {
     txtout("   %g  %g  %g\n",
            pressure->virial_ekin[i][0],
            pressure->virial_ekin[i][1],
            pressure->virial_ekin[i][2]);
     }
#endif


   txtout("Virial Pressure.....: %# 10.7g atm.\n",pressure->pressure_tot);
   txtout("Extern Vir Pressure.: %# 10.7g atm.\n",pressure->pressure_ext_tot);
   txtout("Virial Intern.......: %# 10.7g kcal/mol.\n",pressure->virial_tot);
   txtout("Virial Extern.......: %# 10.7g kcal/mol.\n",pressure->virial_ext_tot);
   txtout("Virial Ekin.........: %# 10.7g kcal/mol.\n",pressure->virial_ekin_tot);
   
}

int fprintf_vir_pressure(FILE *to_file, PRESSURE *pressure)
{
int err=0;

  err|=fprintf(to_file,"%.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g ",
             pressure->pressure_tensor[0][0],             
             pressure->pressure_tensor[0][1],
             pressure->pressure_tensor[0][2],             
             pressure->pressure_tensor[1][0],             
             pressure->pressure_tensor[1][1],
             pressure->pressure_tensor[1][2],             
             pressure->pressure_tensor[2][0],             
             pressure->pressure_tensor[2][1],
             pressure->pressure_tensor[2][2],
             pressure->pressure_tot);

  err|=fprintf(to_file,"%.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g ",
             pressure->pressure_ext_tensor[0][0],             
             pressure->pressure_ext_tensor[0][1],
             pressure->pressure_ext_tensor[0][2],             
             pressure->pressure_ext_tensor[1][0],             
             pressure->pressure_ext_tensor[1][1],
             pressure->pressure_ext_tensor[1][2],             
             pressure->pressure_ext_tensor[2][0],             
             pressure->pressure_ext_tensor[2][1],
             pressure->pressure_ext_tensor[2][2],
             pressure->pressure_ext_tot);

   err|=fprintf(to_file,"%.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g ",
             pressure->virial[0][0],
             pressure->virial[0][1],
             pressure->virial[0][2],
             pressure->virial[1][0],
             pressure->virial[1][1],
             pressure->virial[1][2],
             pressure->virial[2][0],
             pressure->virial[2][1],
             pressure->virial[2][2],
             pressure->virial_tot);

  err|=fprintf(to_file,"%.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g ",
             pressure->virial_ext[0][0],
             pressure->virial_ext[0][1],
             pressure->virial_ext[0][2],
             pressure->virial_ext[1][0],
             pressure->virial_ext[1][1],
             pressure->virial_ext[1][2],
             pressure->virial_ext[2][0],
             pressure->virial_ext[2][1],
             pressure->virial_ext[2][2],
             pressure->virial_ext_tot);

  err|=fprintf(to_file,"%.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g ",
             pressure->virial_ekin[0][0],
             pressure->virial_ekin[0][1],
             pressure->virial_ekin[0][2],
             pressure->virial_ekin[1][0],
             pressure->virial_ekin[1][1],
             pressure->virial_ekin[1][2],
             pressure->virial_ekin[2][0],
             pressure->virial_ekin[2][1],
             pressure->virial_ekin[2][2],
             pressure->virial_ekin_tot);

  return(err);
}

static
double calc_volume(double *sbound_droff)
{

  /* Volume of the simulated system */
  /* to be calculated               */
  double volume = 0.0;

  /*Shortcut for radiusdata         */		
  double r;

  /* Sides of the planar set        */
  double a,b,c;

  /*************************************************/
  /******** Spheric version ************************/
  /*************************************************/
  if (sbound_droff[R_SPHERE]>SMALL) 
    {
    volume = (double)((4./3.) * PI * (sbound_droff[R_SPHERE] *
                             sbound_droff[R_SPHERE] * 
                             sbound_droff[R_SPHERE]));
#if DEBUG
    txtout("r=%.7g V=%.7g\n",sbound_droff[R_SPHERE],volume);
#endif
    return(volume);
    }

  /*************************************************/
  /******** Cylindric version *********************/
  /*************************************************/
  if (sbound_droff[R_CYL]>SMALL) 
  {
      volume = PI * sbound_droff[R_SPHERE] * 
               sbound_droff[R_SPHERE] * fabs(sbound_droff[5] - sbound_droff[4]) ;
  }

  /*************************************************/
  /********* Version with planes *******************/
  /*************************************************/
  if (any_planar_sbound_set(sbound_droff)) 
    {
    if (sbound_droff[R_CURVE] < SMALL)
      {
      /* we have sharp corners, easy to compute */
      volume = fabs(sbound_droff[1] - sbound_droff[0]) *
               fabs(sbound_droff[3] - sbound_droff[2]) *
               fabs(sbound_droff[5] - sbound_droff[4]);

      } /* if sharp corners */
    else
      {
      /* we have round corners */
      /* In this case we have all 8 sides active, even if some of them are */
      /* placed through the origin. */

      /* Curvature radius */
      r = sbound_droff[10];

      /* Sidelength of the inner cube: */
      a = fabs(sbound_droff[1] - sbound_droff[0]) - 2*r;
      b = fabs(sbound_droff[3] - sbound_droff[2]) - 2*r;
      c = fabs(sbound_droff[5] - sbound_droff[4]) - 2*r;
#if DEBUG
    txtout("calc_volume(): Seiten des inneren Wuerfels: %.7g %.7g %.7g\n",a,b,c);
    txtout("calc_volume(): Radius: %.7g\n",r);
#endif
      /* volume of inner cube */    
      volume = a*b*c;

      /* volume of the front 'planes' */
      volume += 2 * r * (a*b + a*c + b*c);

      /* volume of the cylinders */
      volume += 2 * PI * r *(a + b + c);

      /*volume of the sphere */
      volume += (4./3.) * PI * r*r*r;
      
      } /* round corners */
#if DEBUG
      txtout("calc_volume() planes : Volume = %.7g\n",volume);
#endif
      return(volume);
    } 

    /* no bounds specified */    
    return(-1.0);
}

