/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									  **
**	Module	: egpull.c	by M.Eichinger and B. Heymann, Munich	  **
**	Belongs2: EGO							  **
**	Contents: All about pulling and immobilisation and rotation	  **
**		global_data->flag_pull = 0	No pulling		  **
**					 1 	Normal Pulling		  **
**					 2	Stepping Mode		  **
**					 3	Pulling in opposite dirs  **
**		global_data->rotation_mode = 1				  **
****************************************************************************/

#include <math.h>
#include <stdio.h>

#include "ego.h"
#include "util_lib.h"
#include "egalloc.h"
#include "egpull.h"

#define DEBUG1 0 
#define DEBUG2 1 
#define DEBUG3 1 

static int
set_immobilized_atoms(  int nr_of_atoms,
                        STRUCT_ATOMDATA *atom_data,
                        int *atom_sele_flag2,
                        LOAD_MSG_ALL_NODE *global_data);

static int
set_rotation_atoms(  int nr_of_atoms,
		     STRUCT_ATOMDATA *atom_data,
		     int *atom_sele_flag2,
		     LOAD_MSG_ALL_NODE *global_data);

static int
set_pull_atoms( int nr_of_atoms,
                STRUCT_ATOMDATA *atom_data,
                int *atom_sele_flag1,
                LOAD_MSG_ALL_NODE *global_data);

static  int     *new2old_map;
static  int     my_num_atoms=0;
static  int     my_start=0;
static  int     immobilization=0;
static  int     pulling=0;
static  int     flag_rot=0;

static  int     countptr=0;
static  int     my_number;

static  double  *x2,*y2,*z2;
static  double  *x2ord,*y2ord,*z2ord;

static  int           *temp_rot; 
static  double        **kmatrix,force[3];
static  double        **dmat,**dmat_inv;
static  double        **umat,**umat_inv;
static  double        **tmat,**tmat_inv;
static  double        **fmat;
static  double        **temp_mat,**udu_mat;
static  double        er[3],ep[3],ez[3]; /* orthogonal & normalized vectors with (ez || rotation_dir) */



int 
ini_pull(LOAD_MSG_ONE_NODE *my_data,LOAD_MSG_ALL_NODE *global_data,int *atom_sele_flag1,int no_max_atoms)
{
        int i=no_max_atoms;i++; /*** no warning ***/

        new2old_map = global_data->new2old_map;

        immobilization = global_data->flag_selimob;
        pulling       = global_data->flag_pull;

        if ((!immobilization)&&(!pulling)) return 0;

        x2 = alloc_doubles(global_data->total_no_atoms+10);
        y2 = alloc_doubles(global_data->total_no_atoms+10);
        z2 = alloc_doubles(global_data->total_no_atoms+10);

        x2ord = alloc_doubles(global_data->total_no_atoms+10);
        y2ord = alloc_doubles(global_data->total_no_atoms+10);
        z2ord = alloc_doubles(global_data->total_no_atoms+10);

        return 0;
}

int  free_pull()
{
        if ((!immobilization)&&(!pulling)) return 0;

        myfree(x2);
        myfree(y2);
        myfree(z2);

        myfree(x2ord);
        myfree(y2ord);
        myfree(z2ord);

        return 0;
}

/***********************************************************************/
/***         here we collect the coordinates from all nodes so that  ***/
/***         each node knows about all coordinates. This is needed   ***/
/***         for the calculation of the CoM of all pulling atoms     ***/
/***********************************************************************/

int     
merge_packet_to_buffer_pull( int myn,
                        int nr_of_atoms,
                        TRAVELLING_COORDINATES *coord,
                        STRUCT_ATOMDATA *atom_data)
{


        static  int     i; /*** ME: static needed for PARIX ?? ***/

        my_number=myn;
        if ((!immobilization)&&(!pulling)) return 0;
#if DEBUG
txtout("Node %d: merging packet %d to buffer all %d  act %d\n",
my_number,coord->node,nr_of_atoms,coord->no_atoms);
#endif
         if (coord->node==0) countptr=0;


        if (my_number==coord->node) {
                my_start     = countptr;
                my_num_atoms = coord->no_atoms;
        }

        /*** merge current coords to x2,y2,z2 ***/
        memcpy(x2+countptr,coord->x,sizeof(double)*coord->no_atoms);
        memcpy(y2+countptr,coord->y,sizeof(double)*coord->no_atoms);
        memcpy(z2+countptr,coord->z,sizeof(double)*coord->no_atoms);
        countptr += coord->no_atoms;
    
        /**** now we reorder the new atoms x2,y2,z2 *****/

        if (countptr==nr_of_atoms) {

          for (i=0;i<nr_of_atoms;i++) {
                  int mapi = new2old_map[i];

                  x2ord[mapi] = x2[i];
                  y2ord[mapi] = y2[i];
                  z2ord[mapi] = z2[i];
           }

         }

#if DEBUG
txtout("Node %d: exit merge now %d with %d\n",
my_number,coord->node,all_filled);
#endif
        return 0;
}

/** initialize rotation called by the loader in egloader.c**/

int
ini_rotation(LOAD_MSG_ALL_NODE *global_data,STRUCT_ATOMDATA *atom_data)
{
  double norm;
  int i;
    flag_rot = global_data->rotation_mode;
    if (flag_rot == 1){
      norm = global_data->rotation_dir[0]*global_data->rotation_dir[0] +
	     global_data->rotation_dir[1]*global_data->rotation_dir[1] +
	     global_data->rotation_dir[2]*global_data->rotation_dir[2];
      if (norm<=0.0001) {
	errout("Fatal error: You have to define a rotationm axis!\n");
	pex(-1);
      }
      global_data->rotation_dir[0] /= sqrt(norm);
      global_data->rotation_dir[1] /= sqrt(norm);
      global_data->rotation_dir[2] /= sqrt(norm);

      txtout("Calculated normalized rotation axis for fixed rotation axis mode: %f %f %f\n",
	     global_data->rotation_dir[0],
	     global_data->rotation_dir[1],
	     global_data->rotation_dir[2]);		

      set_rotation_atoms(global_data->total_no_atoms,
			 atom_data,
			 global_data->atom_sele_flag2,
			 global_data);

      txtout("Number of Atoms selected for Rotation Forces: %d \n",
	     global_data->nr_of_rot_atoms);

     temp_rot = alloc_ints(global_data->nr_of_rot_atoms);
     for (i=0;i<global_data->nr_of_rot_atoms;i++) {
       temp_rot[i] = 0;
     }
    }
    return 0;
}

/** end of ini_rotation **/

int
ini_imob_and_pull(LOAD_MSG_ALL_NODE *global_data,STRUCT_ATOMDATA *atom_data)
{	
	double norm;
	int n,m;
	/*** calculate pulling direction from atom positions ***/
	if (global_data->flag_pull>=4) {
		double ko1[3],ko2[3];

		ko1[0] = atom_data[global_data->pull_atoms[0]-1].ko[0];
		ko1[1] = atom_data[global_data->pull_atoms[0]-1].ko[1];
		ko1[2] = atom_data[global_data->pull_atoms[0]-1].ko[2];		

		ko2[0] = atom_data[global_data->pull_atoms[1]-1].ko[0];
		ko2[1] = atom_data[global_data->pull_atoms[1]-1].ko[1];
		ko2[2] = atom_data[global_data->pull_atoms[1]-1].ko[2];
		
		global_data->pull_dir[0] = ko2[0] - ko1[0];
		global_data->pull_dir[1] = ko2[1] - ko1[1];
		global_data->pull_dir[2] = ko2[2] - ko1[2];		
	}

	/*** normalize the pulling vector ***/
	norm =  global_data->pull_dir[0]*global_data->pull_dir[0] +
			global_data->pull_dir[1]*global_data->pull_dir[1] +
			global_data->pull_dir[2]*global_data->pull_dir[2];
	if (norm<=0.0001) {
		errout("Fatal error: You have to define a pulling direction!\n");
		pex(-1);
	}
	global_data->pull_dir[0] /= sqrt(norm);
	global_data->pull_dir[1] /= sqrt(norm);
	global_data->pull_dir[2] /= sqrt(norm);

	/*** now we have the pulling direction as a vector ***/
	/*** and switch back to the usual mode	           ***/
	if (global_data->flag_pull>=4) {
		global_data->flag_pull-=3;	/*** map it to usual mode ***/
		txtout("Calculated pulling (atom %d, atom%d) direction: %f %f %f\n",
			global_data->pull_atoms[0],
			global_data->pull_atoms[1],
			global_data->pull_dir[0],
			global_data->pull_dir[1],
			global_data->pull_dir[2]);		
		txtout("Switching back to pulling mode %d\n",global_data->flag_pull);
	}

	/*** some printouts and error checks whether we have all data ***/
	if (	((global_data->flag_selimob)||(global_data->flag_pull)) &&
		(global_data->selimob_start<global_data->cycles_passed) ) {
		
		if (global_data->selimob_start != global_data->start_refC) {
			if (global_data->start_refC<0) {
				errout("Fatal error: Found no COM/pull-info for harmonic immobilisation\n");
				errout("and pulling in restart-file.\n");
				pex(-1);
			} else {
				errout("Fatal error: Found COM/pull-info from integration step %d, but you\n",
						global_data->start_refC);
				errout("want to work with data from step %d restarting with step %d.\n",
						global_data->selimob_start,
						global_data->cycles_passed);
				errout("That is impossible !\n");
				pex(-1);
			}		
		} else {
		  if (global_data->flag_selimob) {
		    for (n=0;n<global_data->flag_selimob;n++) {
		      m=3*n;
		      txtout("\nUsing (%f, %f, %f) at step %d as center %d for immobilisation.\n",
			       global_data->refC[m],
			       global_data->refC[m+1],
			       global_data->refC[m+2],
			       global_data->start_refC,
			       n+1);
		    }
		  }
                }
	}
	
	if (global_data->flag_selimob) {
		set_immobilized_atoms(	global_data->total_no_atoms,
					atom_data,
					global_data->atom_sele_flag2,
					global_data);
	}

	if (global_data->flag_pull) {
		set_pull_atoms(	global_data->total_no_atoms,
				atom_data,
				global_data->atom_sele_flag1,
				global_data);
	}	
	
	return 0;
}

/**************************************************************************/
/*** here I mark all atoms selected for immobilisation with FLAG_IMOBIL	***/
/**************************************************************************/
static int
set_immobilized_atoms(	int nr_of_atoms,
			STRUCT_ATOMDATA *atom_data,
			int *atom_sele_flag2,
			LOAD_MSG_ALL_NODE *global_data)
{
	int	h,i,j,j1,j2,j3;
	int	*used_atoms;
	j1=0;
	j2=0;
	j3=0;
	used_atoms = alloc_ints(nr_of_atoms);

	for (h=0;h<global_data->flag_selimob;h++) {
	  for (i=0;i<nr_of_atoms;i++) used_atoms[i] = 0;
	  
	  set_used_atoms(	global_data->selimob_str[h],
				nr_of_atoms,
				atom_data,
				used_atoms);

	  global_data->selimob_mass[h] = 0.0;
	  if (h==0) {
	    for (i=0,j1=0;i<nr_of_atoms;i++) {
	      if (used_atoms[i]) {
		int	ty = (int)atom_data[i].atomtype;
		double	m  = global_data->AB[ty][ty];
		j1++;
		atom_sele_flag2[i]         |= FLAG_IMOBIL1;
		global_data->selimob_mass[h] += m;
	      }	
	    }
	  } else if (h==1) {
	    for (i=0,j2=0;i<nr_of_atoms;i++) {
	      if (used_atoms[i]) {
		int	ty = (int)atom_data[i].atomtype;
		double	m  = global_data->AB[ty][ty];
		j2++;
		atom_sele_flag2[i]         |= FLAG_IMOBIL2;
		global_data->selimob_mass[h] += m;
	      }	
	    }
	  } else if (h==2) {
	    for (i=0,j3=0;i<nr_of_atoms;i++) {
	      if (used_atoms[i]) {
		int	ty = (int)atom_data[i].atomtype;
		double	m  = global_data->AB[ty][ty];
		j3++;
		atom_sele_flag2[i]         |= FLAG_IMOBIL3;
		global_data->selimob_mass[h] += m;
	      }	
	    }
	  }
	}
	myfree(used_atoms);

	txtout("%d -- %d -- %d atoms are selected for immobilisation.\n",j1,j2,j3);
	j=j1+j2+j3;
	if (j<=0) {
		errout("If you choose immobilisation you must select at least\n");
		errout("one atom! Please check your selection-string in ctl.lis.\n");
		pex(-1);	
	}
	
	return j;
}

/**************************************************************************/
/*** here I mark all atoms selected for rotation with FLAG_ROT		***/
/*** these atoms are indexed in the array rotation map                  ***/
/*** called by ini_rotation                                             ***/
/**************************************************************************/
static int
set_rotation_atoms(	int nr_of_atoms,
			STRUCT_ATOMDATA *atom_data,
			int *atom_sele_flag2,
			LOAD_MSG_ALL_NODE *global_data)
{
  int	i,j;
  int	*used_atoms;

  used_atoms = alloc_ints(nr_of_atoms);

  for (i=0;i<nr_of_atoms;i++) used_atoms[i] = 0;
  set_used_atoms(	global_data->rotation_str,
			nr_of_atoms,
			atom_data,
			used_atoms);


  /***inv_rot_map = alloc_ints(global_data->nr_of_rot_atoms);***/

  /*** The following IF-Statement is necessary since allocation and initialization 
       of global_data->rot_atom_x,y,z might already be done in "restart"! 
       (egcolib.c)***/

  if (global_data->nr_of_rot_atoms==0) {
    for (i=0,j=0;i<nr_of_atoms;i++) {
      if (used_atoms[i]) {
	atom_sele_flag2[i] |= FLAG_ROT;
	global_data->nr_of_rot_atoms++;
	j++;
      }
    }

    /*** some allocation needed for rotation purposes ***/
    global_data->rot_x  = alloc_doubles(global_data->nr_of_rot_atoms);
    global_data->rot_y  = alloc_doubles(global_data->nr_of_rot_atoms);
    global_data->rot_z  = alloc_doubles(global_data->nr_of_rot_atoms);
    global_data->rot1_x = alloc_doubles(global_data->nr_of_rot_atoms);
    global_data->rot1_y = alloc_doubles(global_data->nr_of_rot_atoms);
    global_data->rot1_z = alloc_doubles(global_data->nr_of_rot_atoms);
    global_data->rot2_x = alloc_doubles(global_data->nr_of_rot_atoms);
    global_data->rot2_y = alloc_doubles(global_data->nr_of_rot_atoms);
    global_data->rot2_z = alloc_doubles(global_data->nr_of_rot_atoms);

    global_data->rotation_torque  = alloc_doubles(global_data->nr_of_rot_atoms);
    global_data->rot_t1_x = alloc_doubles(global_data->nr_of_rot_atoms);
    global_data->rot_t1_y = alloc_doubles(global_data->nr_of_rot_atoms);
    global_data->rot_t1_z = alloc_doubles(global_data->nr_of_rot_atoms);
    global_data->rot_t2_x = alloc_doubles(global_data->nr_of_rot_atoms);
    global_data->rot_t2_y = alloc_doubles(global_data->nr_of_rot_atoms);
    global_data->rot_t2_z = alloc_doubles(global_data->nr_of_rot_atoms);
    global_data->rotation_t_torque  = alloc_doubles(global_data->nr_of_rot_atoms);

    global_data->rotation_map     = alloc_ints(global_data->nr_of_rot_atoms);
    /*** end of rot-allocation ***/

    /*** set coordinates of rot-atoms and build map from global_data coordinates 
	 to rot-atom coordinates ***/
    for (i=0,j=0;i<global_data->total_no_atoms;i++) {
      if (used_atoms[i]) {
	global_data->rotation_map[j]     = i;
	global_data->rot_x[j]  = global_data->atomdata[i].ko[0];
	global_data->rot_y[j]  = global_data->atomdata[i].ko[1]; 
	global_data->rot_z[j]  = global_data->atomdata[i].ko[2]; 
	global_data->rot1_x[j] = global_data->atomdata[i].ko[0];
	global_data->rot1_y[j] = global_data->atomdata[i].ko[1]; 
	global_data->rot1_z[j] = global_data->atomdata[i].ko[2]; 
	global_data->rot2_x[j] = global_data->atomdata[i].ko[0];
	global_data->rot2_y[j] = global_data->atomdata[i].ko[1]; 
	global_data->rot2_z[j] = global_data->atomdata[i].ko[2]; 

	global_data->rotation_torque[j]  = 0.0; 
	global_data->rot_t1_x[j] = 0.0; 
	global_data->rot_t1_y[j] = 0.0; 
	global_data->rot_t1_z[j] = 0.0; 
	global_data->rot_t2_x[j] = 0.0; 
	global_data->rot_t2_y[j] = 0.0; 
	global_data->rot_t2_z[j] = 0.0; 

	global_data->rotation_t_torque[j] = 0.0; 

	txtout("Atom %d (%s) in Residue %d (%s) selected for pulling in pos. direction.\n %f %f %f \n",
	       global_data->atomdata[i].atomnr+1,global_data->atomdata[i].atomstr,
	       global_data->atomdata[i].residnr,global_data->atomdata[i].residstr,
	       *((global_data->atomdata[i]).ko+0),*((global_data->atomdata[i]).ko+1),
	       *((global_data->atomdata[i]).ko+2));
	txtout("rot_atom_coord %f %f %\n",global_data->rot_x[j], global_data->rot_y[j],global_data->rot_z[j]);
	txtout("rot_atom_coord2 %f %f %\n",global_data->rot1_x[j], global_data->rot1_y[j],global_data->rot1_z[j]);

	j++;
      }
    }
  }
  else {
    for (i=0;i<global_data->nr_of_rot_atoms;i++) {
      j=global_data->rotation_map[i];
      atom_sele_flag2[j] |= FLAG_ROT;
    }
  }

  for (i=0;i<global_data->nr_of_rot_atoms;i++) {
    txtout("Atom %d with coordinates %f %f %f selected.\n",
	   global_data->atomdata[i].atomnr+1,global_data->rot_x[i],global_data->rot_y[i],global_data->rot_z[i]);
  }
  
  myfree(used_atoms);


  txtout("\n%d atoms selected for rotation mode.\n",j);

  return j;
}


/**************************************************************************/
/*** here I mark all atoms selected for pulling with FLAG_PULL		***/
/*** these atoms are indexed in the array indpull                       ***/
/**************************************************************************/
static int
set_pull_atoms(	int nr_of_atoms,
		STRUCT_ATOMDATA *atom_data,
		int *atom_sele_flag1,
		LOAD_MSG_ALL_NODE *global_data)
{

	int	i,j,k;
	int	*used_atoms;

	used_atoms = alloc_ints(nr_of_atoms);

        for (k=0;k<2;k++) {
	     for (i=0;i<nr_of_atoms;i++) used_atoms[i] = 0;
	     set_used_atoms(	global_data->pull_str[k],
	               		nr_of_atoms,
          			atom_data,
	          		used_atoms);

             global_data->nr_of_pull_atoms[k] = 0;
             global_data->mass_pull_tot[k] = 0.0;

             for (i=0,j=0;i<nr_of_atoms;i++) {
      		if (used_atoms[i]) {
                              int     ty = (int)atom_data[i].atomtype;
                              double  m  = global_data->AB[ty][ty];

                              if (k==0) {
			            atom_sele_flag1[i] |= FLAG_PULL;
                              }
                              else  {
			            atom_sele_flag1[i] |= FLAG_PULL;
                                    atom_sele_flag1[i] |= FLAG_PULL_AGAINST;
                              }


                              global_data->indpull[k][j] = i; 
                              global_data->mass_pull[k][j] = m ; 
                              global_data->mass_pull_tot[k] += m;
                              global_data->nr_of_pull_atoms[k]++;
			      j++;
		      }	
	      }
        }
	myfree(used_atoms);

	if (global_data->flag_pull==2) {
		txtout("\n%d atoms selected for stepping mode.\n",j);
	} else	
	   if ((global_data->flag_pull==1)&&(global_data->nr_of_pull_atoms[0]<1)) {
		errout("Fatal error: You must select at least one atom for force pulling \n");
		errout("mode 1! But you have selected %d atoms. Change your ctl.lis file!\n",j);
		myfree(used_atoms);
		pex(-1);
        } else   
        if ((global_data->flag_pull==3)&&(global_data->nr_of_pull_atoms[0]<1 || global_data->nr_of_pull_atoms[1]<1)) {
                errout("Fatal error: You must select at least one atom for each direction in force pulling \n");
                errout("mode 3! But you have selected %d atom for the positive\n",global_data->nr_of_pull_atoms[0]);
                errout("and %d atoms for the negative direction. Change your ctl.lis file!\n",global_data->nr_of_pull_atoms[1]);
                myfree(used_atoms);
                pex(-1);
	} else {
                       for (i=0;i<global_data->nr_of_pull_atoms[0];i++) {
			       txtout("Atom %d (%s) in Residue %d (%s) selected for pulling in pos. direction.\n",
				atom_data[global_data->indpull[0][i]].atomnr+1,atom_data[global_data->indpull[0][i]].atomstr,
				atom_data[global_data->indpull[0][i]].residnr,atom_data[global_data->indpull[0][i]].residstr);
                       }
                       for (i=0;i<global_data->nr_of_pull_atoms[1];i++) {
			txtout("Atom %d (%s) in Residue %d (%s) selected for pulling in neg. direction.\n",
				atom_data[global_data->indpull[1][i]].atomnr+1,atom_data[global_data->indpull[1][i]].atomstr,
				atom_data[global_data->indpull[1][i]].residnr,atom_data[global_data->indpull[1][i]].residstr);
                       }
		}

       	return j;
}

void calc_imob_energy(ENERGY_INFO *energy,LOAD_MSG_ALL_NODE *global_data)
{
  int n,m;
	global_data->deltaimob[0] = 0.0;
	global_data->deltaimob[1] = 0.0;
	global_data->deltaimob[2] = 0.0;
	global_data->deltaimob[3] = 0.0;
	global_data->deltaimob[4] = 0.0;
	global_data->deltaimob[5] = 0.0;
	global_data->deltaimob[6] = 0.0;
	global_data->deltaimob[7] = 0.0;
	global_data->deltaimob[8] = 0.0;

	energy->imobil            = 0.0;

	if (global_data->flag_selimob) {
	  for (n=0;n<global_data->flag_selimob;n++) {
	    m=3*n;
	    energy->imobC[m]   /= global_data->selimob_mass[n];
	    energy->imobC[m+1] /= global_data->selimob_mass[n];
	    energy->imobC[m+2] /= global_data->selimob_mass[n];

	    if (global_data->cycles_passed==global_data->selimob_start) {
	      global_data->start_refC   = global_data->cycles_passed;
	      global_data->refC[m]      = energy->imobC[m];
	      global_data->refC[m+1]    = energy->imobC[m+1];
	      global_data->refC[m+2]    = energy->imobC[m+2];
	      txtout("\nUsing (%f, %f, %f) at step %d as center no. %d for immobilisation.\n",
		     global_data->refC[m],
		     global_data->refC[m+1],
		     global_data->refC[m+2],
		     global_data->start_refC,
		     n+1);
	    }
	    if (global_data->cycles_passed>=global_data->selimob_start) {
	      global_data->deltaimob[m]   = energy->imobC[m]  -global_data->refC[m];
	      global_data->deltaimob[m+1] = energy->imobC[m+1]-global_data->refC[m+1];
	      global_data->deltaimob[m+2] = energy->imobC[m+2]-global_data->refC[m+2];
	      energy->imobil += global_data->selimob_const * (
			       global_data->deltaimob[m]  *global_data->deltaimob[m] +
			       global_data->deltaimob[m+1]*global_data->deltaimob[m+1] +
			       global_data->deltaimob[m+2]*global_data->deltaimob[m+2]);			
	    }
	  }
	}

}

void add_imob_forces(	LOAD_MSG_ONE_NODE *my_data,
			LOAD_MSG_ALL_NODE *global_data,
			double *Fx,
			double *Fy,
			double *Fz)
{
        int i;
	double fac,facx,facy,facz;


	if (!global_data->flag_selimob) return;

	/*** in the next condition '<=' is neccessary, not only '<'	***/
	/*** because we have 'deltaimob' not for all cases !		***/
	if (global_data->cycles_passed<=global_data->selimob_start) return;

	if (	(global_data->deltaimob[0]==0.0) &&
		(global_data->deltaimob[1]==0.0) &&
		(global_data->deltaimob[2]==0.0) &&
		(global_data->deltaimob[3]==0.0) &&
		(global_data->deltaimob[4]==0.0) &&
		(global_data->deltaimob[5]==0.0) &&
		(global_data->deltaimob[6]==0.0) &&
		(global_data->deltaimob[7]==0.0) &&
		(global_data->deltaimob[8]==0.0) ) return;


	for (i=0;i<my_data->no_atoms; i++) {
	  if (my_data->atom_sele_flag2[i]&FLAG_IMOBIL1) {
	    int	ty  = my_data->typ[i];
	    double	m   = global_data->AB[ty][ty];
	    fac  = -2.0*global_data->selimob_const/global_data->selimob_mass[0];
	    facx = fac * global_data->deltaimob[0];
	    facy = fac * global_data->deltaimob[1];
	    facz = fac * global_data->deltaimob[2];

	    Fx[i] -=  m * facx; 
	    Fy[i] -=  m * facy; 
	    Fz[i] -=  m * facz; 
	  } else if (my_data->atom_sele_flag2[i]&FLAG_IMOBIL2) {
	    int	ty  = my_data->typ[i];
	    double	m   = global_data->AB[ty][ty];
	    fac  = -2.0*global_data->selimob_const/global_data->selimob_mass[1];
	    facx = fac * global_data->deltaimob[3];
	    facy = fac * global_data->deltaimob[4];
	    facz = fac * global_data->deltaimob[5];

	    Fx[i] -=  m * facx; 
	    Fy[i] -=  m * facy; 
	    Fz[i] -=  m * facz; 
	  } else if (my_data->atom_sele_flag2[i]&FLAG_IMOBIL3) {
	    int	ty  = my_data->typ[i];
	    double	m   = global_data->AB[ty][ty];
	    fac  = -2.0*global_data->selimob_const/global_data->selimob_mass[2];
	    facx = fac * global_data->deltaimob[6];
	    facy = fac * global_data->deltaimob[7];
	    facz = fac * global_data->deltaimob[8];
	    
	    Fx[i] -=  m * facx; 
	    Fy[i] -=  m * facy; 
	    Fz[i] -=  m * facz; 
	  }

	    /*** note the hidden '-' in the integrator -> '-=' instead of '+=' ***/
	}
}


void add_pull_forces(	int myn,
                        ENERGY_INFO *energy,	
			LOAD_MSG_ONE_NODE *my_data,
			LOAD_MSG_ALL_NODE *global_data,
			double *Fx,
			double *Fy,
			double *Fz)
{
	int i,k;
        int loop_end = 1;
	double	timediff;	/*** time-difference in ps ***/
	double	equipos[6];
	double	offset;		/*** in A ***/
        my_number=myn;

	energy->pull_indir[0]  = 0.0;
	energy->pull_outdir[0] = 0.0;
	energy->pull_indir[1]  = 0.0;
	energy->pull_outdir[1] = 0.0;

	if (!global_data->flag_pull) return;
	if (global_data->flag_pull==2) return;

        if (global_data->flag_pull==3) loop_end = 2;
		
	/*** in the next condition '<=' is neccessary, not only '<'	***/
	/*** because we have 'pullspringpos' not in all cases !		***/
	if (global_data->cycles_passed<=global_data->selimob_start) return;

	/*** calculate the new equilibration position of the pulled atom***/
	/*** note that the length of the pull_dir vector is 1 		***/

        for (k=0;k<loop_end; k++){
               global_data->xcom_pull[k] = 0.0;
               global_data->ycom_pull[k] = 0.0;
               global_data->zcom_pull[k] = 0.0;

         	for (i=0;i<global_data->nr_of_pull_atoms[k]; i++){
                       global_data->xcom_pull[k] +=  global_data->mass_pull[k][i] * x2ord[global_data->indpull[k][i]];
                       global_data->ycom_pull[k] +=  global_data->mass_pull[k][i] * y2ord[global_data->indpull[k][i]];
                       global_data->zcom_pull[k] +=  global_data->mass_pull[k][i] * z2ord[global_data->indpull[k][i]];

                }


                if (global_data->mass_pull_tot[k]>0.0) {

                       global_data->xcom_pull[k] /= global_data->mass_pull_tot[k];   
                       global_data->ycom_pull[k] /= global_data->mass_pull_tot[k];
                       global_data->zcom_pull[k] /= global_data->mass_pull_tot[k];   
                }
           }

            if (global_data->cycles_passed==global_data->selimob_start+1) {

	      global_data->start_refC    = global_data->cycles_passed - 1; 
	      /*-1 to get right number in restart-file */
	      energy->pullspringpos[0] = global_data->xcom_pull[0];
	      energy->pullspringpos[1] = global_data->ycom_pull[0];
	      energy->pullspringpos[2] = global_data->zcom_pull[0];
	      energy->pullspringpos[3] = global_data->xcom_pull[1];
	      energy->pullspringpos[4] = global_data->ycom_pull[1];
	      energy->pullspringpos[5] = global_data->zcom_pull[1];
	      energy->pullspringpos[6] = 0.0;

	      if (my_number==0)  {
		txtout("Using (%f, %f, %f) for pulling start position.\n",
		       energy->pullspringpos[0],
		       energy->pullspringpos[1],
		       energy->pullspringpos[2]);
		if ( global_data->flag_pull==3 ) {
		  txtout("Using (%f, %f, %f) for pulling start position in opposite dir.\n",
			 energy->pullspringpos[3],
			 energy->pullspringpos[4],
			 energy->pullspringpos[5]);
		}
		
	      }
	    }

                timediff   =  1.0 *                                                      /*** number of steps  ***/
                ((global_data->time_step * TIMEDIV) /                        /*** time per step in sec  ***/
                1.000E-12);                                                  /*** factor to ps        ***/

                if (global_data->flag_pull!=3) {

                       offset     = timediff * global_data->pull_vel;

                       energy->pullspringpos[0] += global_data->pull_dir[0] * offset;
                       energy->pullspringpos[1] += global_data->pull_dir[1] * offset;
                       energy->pullspringpos[2] += global_data->pull_dir[2] * offset;
                       equipos[0] = energy->pullspringpos[0];
                       equipos[1] = energy->pullspringpos[1];
                       equipos[2] = energy->pullspringpos[2];
                       global_data->pullspringpos[0]=energy->pullspringpos[0];
                       global_data->pullspringpos[1]=energy->pullspringpos[1];
                       global_data->pullspringpos[2]=energy->pullspringpos[2];

              } else {

                       /*** divide by 2.0 because we pull two atoms in opposite direction ***/
                       offset     = timediff * global_data->pull_vel / 2.0;


			/*** we have to calc the place of the springs and put this ***/
			/*** information in energy->pullspringpos for the master       ***/

				energy->pullspringpos[0] += global_data->pull_dir[0] * offset;
				energy->pullspringpos[1] += global_data->pull_dir[1] * offset;
				energy->pullspringpos[2] += global_data->pull_dir[2] * offset; 
				energy->pullspringpos[3] -= global_data->pull_dir[0] * offset;
				energy->pullspringpos[4] -= global_data->pull_dir[1] * offset;
				energy->pullspringpos[5] -= global_data->pull_dir[2] * offset; 
				equipos[0] = energy->pullspringpos[0];
				equipos[1] = energy->pullspringpos[1];
				equipos[2] = energy->pullspringpos[2]; 
				equipos[3] = energy->pullspringpos[3];
				equipos[4] = energy->pullspringpos[4];
				equipos[5] = energy->pullspringpos[5]; 
                                global_data->pullspringpos[0]=energy->pullspringpos[0];
                                global_data->pullspringpos[1]=energy->pullspringpos[1];
                                global_data->pullspringpos[2]=energy->pullspringpos[2];
                                global_data->pullspringpos[3]=energy->pullspringpos[3];
                                global_data->pullspringpos[4]=energy->pullspringpos[4];
                                global_data->pullspringpos[5]=energy->pullspringpos[5];

		}

	for (i=0;i<my_data->no_atoms; i++){
                int     ty  = my_data->typ[i];
                double  m   = global_data->AB[ty][ty];
		double vec[3];	/*** difference vector between atom and equipos ***/
		double outvec[3],norm;
		double offset_in_dir;	
		double offset_out_dir;
		double force_in_dir;
		double force_out_dir;

		if (!(my_data->atom_sele_flag1[i]&FLAG_PULL)) continue;

		
		if (!(my_data->atom_sele_flag1[i]&FLAG_PULL_AGAINST)) {
			vec[0] = global_data->xcom_pull[0] - equipos[0];
			vec[1] = global_data->ycom_pull[0] - equipos[1];
			vec[2] = global_data->zcom_pull[0] - equipos[2];
		} else {
                        vec[0] = global_data->xcom_pull[1] - equipos[3];
                        vec[1] = global_data->ycom_pull[1] - equipos[4];
                        vec[2] = global_data->zcom_pull[1] - equipos[5];
		}

		offset_in_dir = vec[0] * global_data->pull_dir[0] +
				vec[1] * global_data->pull_dir[1] +
				vec[2] * global_data->pull_dir[2];

		offset_out_dir = sqrt(fabs(vec[0]*vec[0] + 
				   	vec[1]*vec[1] + 
				   	vec[2]*vec[2] -
				   offset_in_dir * offset_in_dir));

                if (!(my_data->atom_sele_flag1[i]&FLAG_PULL_AGAINST)) {
		      force_in_dir  = -2.0*global_data->pull_const[0]*offset_in_dir * m/global_data->mass_pull_tot[0];
		      force_out_dir = -2.0*global_data->pull_const[1]*offset_out_dir * m/global_data->mass_pull_tot[0];
                 } else {
		      force_in_dir  = -2.0*global_data->pull_const[0]*offset_in_dir * m/global_data->mass_pull_tot[1];
		      force_out_dir = -2.0*global_data->pull_const[1]*offset_out_dir * m/global_data->mass_pull_tot[1];
                 }


		/*** note the hidden '-' in the integrator -> '-=' instead of '+=' ***/
		Fx[i] -= force_in_dir*global_data->pull_dir[0];
		Fy[i] -= force_in_dir*global_data->pull_dir[1];
		Fz[i] -= force_in_dir*global_data->pull_dir[2];

		outvec[0] = vec[0] - offset_in_dir*global_data->pull_dir[0];
		outvec[1] = vec[1] - offset_in_dir*global_data->pull_dir[1];
		outvec[2] = vec[2] - offset_in_dir*global_data->pull_dir[2];
		norm      = sqrt(outvec[0]*outvec[0] + outvec[1]*outvec[1] + outvec[2]*outvec[2]);
		if (norm<=0.0)	norm = 1.0;

		/*** note the hidden '-' in the integrator -> '-=' instead of '+=' ***/
		Fx[i] -= force_out_dir * outvec[0]/norm;
		Fy[i] -= force_out_dir * outvec[1]/norm;
		Fz[i] -= force_out_dir * outvec[2]/norm;

		if (!(my_data->atom_sele_flag1[i]&FLAG_PULL_AGAINST)) {
			energy->pull_indir[0]  += force_in_dir  * KCALPN;
			energy->pull_outdir[0] += force_out_dir * KCALPN;	
		} else {
			energy->pull_indir[1]  -= force_in_dir  * KCALPN; /* negative sign
                                                                             because we are pulling
                                                                             in the opposite direction */
			energy->pull_outdir[1] += force_out_dir * KCALPN;
		}
	}
	return;	
}



/** if the pulling forces in pulling mode 3 for the two	**/
/** springs are very different, we adopt the positions	**/
/** of the springs to counteract the strong movement	**/
/** of the protein. The force correction bases on the   **/
/** average force difference (with the chosen averaging **/
/** constant the force difference 500 integration steps **/
/** before has the weight 1/e)                          **/
/** 							**/
/** the force difference should in 1000 integration step**/
/** fall down to 1 percent				**/
#define WGHT_FOR_ADOPT_PULL_FRAME (1.0 - pow(0.01,0.001))
#define AVERAGE_WEIGHT 0.002
void adopt_pull_frame(	ENERGY_INFO	*energy,
			LOAD_MSG_ALL_NODE *global_data)
{
	double  delta_f, delta_x, av_diff_delta_x;

	if (global_data->cycles_passed<=global_data->selimob_start) return;
	if (global_data->flag_pull==3) {
	  
		/*** calc difference of forces in kcal/mol/A ***/
                av_diff_delta_x=energy->pullspringpos[6];
	  	delta_f = (energy->pull_indir[0] - energy->pull_indir[1]) / KCALPN; 
                delta_x = -delta_f / (2.0 * global_data->pull_const[0]) * AVERAGE_WEIGHT + (1-AVERAGE_WEIGHT) * av_diff_delta_x;

		global_data->pullspringpos[0] += WGHT_FOR_ADOPT_PULL_FRAME * delta_x * global_data->pull_dir[0];
		global_data->pullspringpos[1] += WGHT_FOR_ADOPT_PULL_FRAME * delta_x * global_data->pull_dir[1];
		global_data->pullspringpos[2] += WGHT_FOR_ADOPT_PULL_FRAME * delta_x * global_data->pull_dir[2];
		global_data->pullspringpos[3] += WGHT_FOR_ADOPT_PULL_FRAME * delta_x * global_data->pull_dir[0];
		global_data->pullspringpos[4] += WGHT_FOR_ADOPT_PULL_FRAME * delta_x * global_data->pull_dir[1];
		global_data->pullspringpos[5] += WGHT_FOR_ADOPT_PULL_FRAME * delta_x * global_data->pull_dir[2];
                global_data->pullspringpos[6] = delta_x;
                energy->pullspringpos[0]=global_data->pullspringpos[0];
                energy->pullspringpos[1]=global_data->pullspringpos[1];
                energy->pullspringpos[2]=global_data->pullspringpos[2];
                energy->pullspringpos[3]=global_data->pullspringpos[3];
                energy->pullspringpos[4]=global_data->pullspringpos[4];
                energy->pullspringpos[5]=global_data->pullspringpos[5];
                energy->pullspringpos[6]=global_data->pullspringpos[6];
	}

	return;
}

/**------------------------------------------------------------------**/
/*** ROTATION stuff follows....***/
void alloc_temporary( LOAD_MSG_ALL_NODE *global_data)
{
  static int i,i1,i2;
  if (!(global_data->rotation_mode)) return;
  temp_rot = alloc_ints(global_data->nr_of_rot_atoms);
  for (i=0;i<global_data->nr_of_rot_atoms;i++) {
    temp_rot[i] = 0;
  }

  dmat     = dmatrix(0,2,0,2);
  dmat_inv = dmatrix(0,2,0,2);
  
  umat     = dmatrix(0,2,0,2);
  umat_inv = dmatrix(0,2,0,2);
  tmat     = dmatrix(0,2,0,2);
  tmat_inv = dmatrix(0,2,0,2);

  fmat     = dmatrix(0,2,0,2);
  temp_mat = dmatrix(0,2,0,2);
  udu_mat  = dmatrix(0,2,0,2);
  kmatrix  = dmatrix(0,2,0,2);
  ez[0] = global_data->rotation_dir[0];
  ez[1] = global_data->rotation_dir[1];
  ez[2] = global_data->rotation_dir[2];

  for (i1=0;i1<=2;i1++){
    for (i2=0;i2<=2;i2++){
      fmat[i1][i2]     = 0.0;
      temp_mat[i1][i2] = 0.0;
      udu_mat[i1][i2]  = 0.0;
      umat[i1][i2]     = 0.0;
      umat_inv[i1][i2] = 0.0;
      tmat[i1][i2]     = 0.0;
      tmat_inv[i1][i2] = 0.0;
      dmat[i1][i2]     = 0.0;
      dmat_inv[i1][i2] = 0.0;
      kmatrix[i1][i2]  = 0.0;
    }
  }
  kmatrix[0][0] = global_data->rotation_const[1];
  kmatrix[1][1] = global_data->rotation_const[0];
  kmatrix[2][2] = global_data->rotation_const[2];

}
/*** The following routine "make_inv_rot_map" creates so to say the inverse
     field to rot_map - with the same dimension [nr_of_rot_atoms]! It is called
     from init_node AND reinit_node. This has the advantage of small fields
     and thus little communication effort. The inverse map has to be updated after 
     each reclustering!***/

void make_inv_rot_map(  int myn,
			LOAD_MSG_ONE_NODE *my_data,
			LOAD_MSG_ALL_NODE *global_data)
{
  static int i,m,k,index;
  my_number = myn;

  if (!(global_data->rotation_mode)) return;

  for (i=0,m=0;i<my_data->no_atoms;i++) {
    if (!(my_data->atom_sele_flag2[i] & FLAG_ROT)) continue;
    k = global_data->first_atom[my_number] + i;
    index = global_data->new2old_map[k];
    
    if (index!=my_data->atom_id[i]) {
      errout("Node %d: Fatal error: index %d atom_id %d are different \n",
	     my_number,index,my_data->atom_id[i]);
      pex(-1);
    }


    for (k=0;k<global_data->nr_of_rot_atoms;k++) {
      if (global_data->rotation_map[k] != index) continue;
      /*** now the inverse rot_map is filled ***/
      my_data->inv_rotation_map[m] = k;
      txtout("Node %d: Inv_Rotation_Map[%d] = %d \n", my_number,m,k);
    }
    m++;
  }
  for (i=0;i<global_data->nr_of_rot_atoms;i++) {
    temp_rot[i] = 0;
  }
}

void add_rotation_forces(	int myn,
				ENERGY_INFO *energy,	
				LOAD_MSG_ONE_NODE *my_data,
				LOAD_MSG_ALL_NODE *global_data,
				double *Fx,
				double *Fy,
				double *Fz)
{
  double	timediff;	/* time-difference in ps */
  double	offset;		/* in A */
  int           i,j,k,l,m,i1,i2;
  double        dvec[3];
  double        z[3],zo[3],znew[3],pos[3],pos2o[3],pos_rad[3],pos_perp[3],phi,temp[3],poso_rad[3];
  double check=0.0,a=0.0;
  double b,norm,radial_distance;
  double angle;

  if (!(global_data->rotation_mode)) return;

  for (i=0;i<=2;i++){
    energy->rot_force[i] = 0.0;
  }

  if (global_data->cycles_passed <= global_data->selimob_start) return;


  /* now the rotation routine for fixed rotation axis follows*/
  if (global_data->rotation_mode==1) {

    timediff = (global_data->cycles_passed - global_data->selimob_start) * 
      ((global_data->time_step * TIMEDIV) / 1.000E-12);					
    offset   = timediff * global_data->angular_vel;


    for (i=0,m=0;i<my_data->no_atoms;i++) {
      if (!(my_data->atom_sele_flag2[i] & FLAG_ROT)) continue;
      k = my_data->inv_rotation_map[m];
      for (i1=0;i1<=2;i1++){
	for (i2=0;i2<=2;i2++){
	  fmat[i1][i2]     = 0.0;
	  temp_mat[i1][i2] = 0.0;
	  udu_mat[i1][i2] = 0.0;
	}
      }
      for (i1=0;i1<=2;i1++){
	znew[i1]  = 0.0;
	force[i1] = 0.0;
      }

      /* set the origin of coordinate system on the rotation axis:*/
      /* this procedure should be implemented elsewhere since it is only necessary once*/
      z[0] = zo[0] = global_data->rot_x[k] - global_data->rotation_point[0];
      z[1] = zo[1] = global_data->rot_y[k] - global_data->rotation_point[1];
      z[2] = zo[2] = global_data->rot_z[k] - global_data->rotation_point[2];

      norm = dotprod(z,ez);
      for (l=0;l<=2;l++){
	z[l] = z[l] - norm*ez[l]; /* */
	/*er[l] =  zo[l]-z[l];*/
      }    

      for (l=0;l<=2;l++){
	er[l] = z[l]/veclen(z); /*er[l]/veclen(er);*/
      }
      /* calculate ep(hi)[3] */
      a=crossprod(ez,er,ep);
      /* U-Matrix for transformation to (fixed) rotation coordinate system*/
      for (l=0;l<=2;l++){
	umat[0][l]=er[l];
	umat[1][l]=ep[l];
	umat[2][l]=ez[l];
      }
      /* U^{-1} == U^{^T}*/
      for (l=0;l<=2;l++){
	umat_inv[l][0]=er[l];
	umat_inv[l][1]=ep[l];
	umat_inv[l][2]=ez[l];
      }

      /* rotation matrix D, D^{-1} */
      dmat     = rotation_matrix(dmat,offset);
      dmat_inv = rotation_matrix(dmat_inv,-offset);

      /* calculate the vector to the reference position, e.g. ((D))(omega,t)*((U))(x_i^R-x_A) */
      /* in cylinder coordinates */
      matrix_dot_matrix(dmat,umat,tmat);
      matrix_dot_matrix(umat_inv,tmat,udu_mat);
      mat_dot_vec(udu_mat,&zo[0],&znew[0]);

      /* force = ((U))^{-1} ((K)) ((U)) ( ((U))^{-1} ((D)) ((U)) x^R - x_i ) */
      matrix_dot_matrix(kmatrix,umat,fmat);

      matrix_dot_matrix(umat_inv,fmat,temp_mat);

      pos[0] = my_data->x_new[i] - global_data->rotation_point[0];
      pos[1] = my_data->y_new[i] - global_data->rotation_point[1];
      pos[2] = my_data->z_new[i] - global_data->rotation_point[2];


      dvec[0] = znew[0] - pos[0];
      dvec[1] = znew[1] - pos[1];
      dvec[2] = znew[2] - pos[2];

      mat_dot_vec(temp_mat,&dvec[0],&force[0]);

      Fx[i] -= force[0];
      Fy[i] -= force[1];
      Fz[i] -= force[2];

      b=dotprod(force,ez);
      energy->rot_force[1] += b * KCALPN;

      /* calculate delta(phi) for energy integration */
      pos2o[0] = global_data->rot2_x[k] - global_data->rotation_point[0];
      pos2o[1] = global_data->rot2_y[k] - global_data->rotation_point[1];
      pos2o[2] = global_data->rot2_z[k] - global_data->rotation_point[2];

      a = dotprod(pos2o,ez);
      for (l=0;l<=2;l++) {
	poso_rad[l] = pos2o[l] - a*ez[l];
      }
      a = dotprod(pos,ez);
      for (l=0;l<=2;l++) {
	pos_rad[l] = pos[l] - a*ez[l];
      }

      /* get the right sign for the rotation forces */
      a = crossprod(pos_rad,force,temp);
      a = dotprod(temp,ez);
      if (a >=0) {
	energy->rot_force[0] += sqrt(dotprod(force,force)-b*b) * KCALPN;
      } else {
	energy->rot_force[0] -= sqrt(dotprod(force,force)-b*b) * KCALPN;
      }

      /* get the change in angle for the energy integration */ 
      phi = dotprod(pos_rad,poso_rad) / (veclen(pos_rad)*veclen(poso_rad));
      a   = crossprod(poso_rad,pos_rad,temp);
      check   = temp[0]*ez[0] + temp[1]*ez[1] + temp[2]*ez[2];
      if (phi >= 1.0) {
	angle = 0.0;
      } else
	if (check >= 0.0) {
	  angle = acos(phi);
	} else {
	  angle = -acos(phi);
	}


      /* integrate the energy */
      energy->rotation_energy += angle/2.0 * global_data->rotation_torque[k];

      /*set coordinates for next step: n-1 and n-2*/
      global_data->rot1_x[k] = my_data->x_new[i];
      global_data->rot1_y[k] = my_data->y_new[i];
      global_data->rot1_z[k] = my_data->z_new[i];
      global_data->rot2_x[k] = my_data->x_old[i];
      global_data->rot2_y[k] = my_data->y_old[i];
      global_data->rot2_z[k] = my_data->z_old[i];


      a = crossprod(ez,pos_rad,pos_perp);
      b = veclen(pos_perp);
      for (l=0;l<=2;l++) {
	pos_perp[l] = pos_perp[l]/b;
      }
      radial_distance = veclen(pos_rad);

      global_data->rotation_torque[k] = dotprod(force,pos_perp) * radial_distance; /** KCALPN;*/
      /* output of total applied torque in pN A */
      energy->rot_force[2] += global_data->rotation_torque[k] * KCALPN;

      m++;
      temp_rot[k] = 1;

    }
    /* for communication purposes: to get the right values from different nodes (see eganaly.c)  */
    for (j=0;j<global_data->nr_of_rot_atoms;j++) {
      if (temp_rot[j] == 0) {
	global_data->rot1_x[j] = 0.0;
	global_data->rot1_y[j] = 0.0;
	global_data->rot1_z[j] = 0.0;
	global_data->rot2_x[j] = 0.0;
	global_data->rot2_y[j] = 0.0;
	global_data->rot2_z[j] = 0.0;
	global_data->rotation_torque[j]  = 0.0;
      }
    }
  }
}

double **rotation_matrix(double **mat,double omega)
{
 mat[0][0] = cos(omega);
 mat[0][1] = -sin(omega);
 mat[0][2] = 0.0;
 mat[1][0] = sin(omega);
 mat[1][1] = cos(omega);
 mat[1][2] = 0.0;
 mat[2][0] = 0.0;
 mat[2][1] = 0.0;
 mat[2][2] = 1.0;
 return mat;
}

void matrix_dot_matrix(double **mat1,double **mat2,double **mat3)
{
int j1,j2,j3;

 for (j1=0;j1<=2;j1++){
   for (j2=0;j2<=2;j2++){
     mat3[j1][j2]=0.0;
     for (j3=0;j3<=2;j3++){
       mat3[j1][j2] += mat1[j1][j3]*mat2[j3][j2];
     }
   }
 }
}

void mat_dot_vec(double **matti,double *veci,double *veco) 
{
	double v[3];
		
	v[0] = matti[0][0]*veci[0] + matti[0][1]*veci[1] + matti[0][2]*veci[2];
	v[1] = matti[1][0]*veci[0] + matti[1][1]*veci[1] + matti[1][2]*veci[2];
	v[2] = matti[2][0]*veci[0] + matti[2][1]*veci[1] + matti[2][2]*veci[2];
	veco[0]=v[0];
	veco[1]=v[1];
	veco[2]=v[2];	
}

int  free_rotation()
{
        if ((!immobilization)&&(flag_rot)) return 0;
	
	free_dmatrix(dmat,0,2,0,2);
	free_dmatrix(dmat_inv,0,2,0,2);
	free_dmatrix(umat,0,2,0,2);
	free_dmatrix(umat_inv,0,2,0,2);
	free_dmatrix(tmat,0,2,0,2);
	free_dmatrix(tmat_inv,0,2,0,2);
	free_dmatrix(fmat,0,2,0,2);
	free_dmatrix(temp_mat,0,2,0,2);
	free_dmatrix(udu_mat,0,2,0,2);
	free_dmatrix(kmatrix,0,2,0,2);


        return 0;
}

/**--------------------------------------------------------------**/
