/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**	Belongs to: EGO							**
**									**
**	Module	: tree.c	by M.Eichinger, Munich			**
**	Contents: Calculations on the tree of clusters			**
**		  calculate ww 	with local expansion (2. order)		**
**  	Modif's:							**
**									**
**									**
*************************************************************************/

/******* Usage of different Algol ********/
#define NO_EXTRAPOL		0
#define DO_CUTOFF		0	
#define REALLOC_SIZE		30000
#define FREEENERGY		0
				
/*****************************************/
#define DEBUG			0
#define DEBUG_INI		0
#define DEBUG_LR		0
#define DEBUG_MULTIPOL		0
#define IFCHECK_CALC_MULTIPOL	1
#define	IFCHECK_CALC_LONGRANGE	1
#define IFCHECKBUFSIZE		1
#define TRACE_REALLOC		0

#include <stdio.h>
#include <string.h>
#include <math.h>
#include "ego.h"
#include "egtree.h"
#include "timing.h"
#include "egalloc.h"
#include "egbolib.h"
#include "util_lib.h"
#include "egintern.h"
#include "egqminter.h"

/****************************************************************************/
/********************** defined functions ***********************************/
/****************************************************************************/

static	int	correct_refpoint(double *refp,double *cen_geom);

static int
calc_exact_interact(	int	my_number,int my_atom,
			LOAD_MSG_ALL_NODE *global_data,
			LOAD_MSG_ONE_NODE *my_data,
			TRAVELLING_PACKETS *packet,
			int mode);

static int
TREE_force_tree2tree(	int my_number,
			int my_cluster,
			int level,
			LOAD_MSG_ALL_NODE *global_data,
			LOAD_MSG_ONE_NODE *my_data,
			TRAVELLING_PACKETS *packet);

static int
TREE_local_cl2cl(	struct_cluster *my_cluster,
			struct_cluster *other_cluster);

static
int	inherit_local_expansion(	int my_cluster,
					int level,
					LOAD_MSG_ONE_NODE *my_data);
static	int
extrapolate_lshells(	int next_update_down,
			LOAD_MSG_ALL_NODE *global_data,
			LOAD_MSG_ONE_NODE *my_data);

static	int
update_lshells( int my_number,
		int level,
		LOAD_MSG_ALL_NODE *global_data,
		LOAD_MSG_ONE_NODE *my_data);


static int	put_to_wwbuf(int level,int other_number,int elem);

static int	put_to_unitwwlist(int other_number,int elem);

static
int	TREE_calc_innerww(	int	my_number,
			LOAD_MSG_ALL_NODE *global_data,
			LOAD_MSG_ONE_NODE *my_data,
			TRAVELLING_PACKETS *packet,
			int my_cluster,
			int *start_curunit,
			int mode);

static int
do_nearest_extrapol(	int my_number,
			int my_atom,
			LOAD_MSG_ALL_NODE *global_data,
			LOAD_MSG_ONE_NODE *my_data,
			TRAVELLING_PACKETS *packet);

static	int	reset_treestatus_flags(void);

static	int	ini_DC1d(void);
static	double	extrapolate_dc1d(int j,int i, double fnew,double fold);

/**************************************************************************/

static int	alloc_innerwwbuf(int max_no_atoms);
static int	free_innerwwbuf(void);
static int	reset_innerwwbuf(void);
static	int	posinnerwwbuf=0;
static	int	*innerwwbuf,*unitnumbuf;
static	int	max_size_innerwwbuf=1000;
static	int	cur_size_innerwwbuf=0;
static	int	my_unitnum=0;

/***************************************************/
/*** some varibales for the nasty exclusion list ***/
/***************************************************/

static	int	*ex_list,*ex14_list;
static	int	ex_ptr=0,ex14_ptr=0;
static	int	ext_first=0;
static	int	ex_elem=0,ex14_elem=0;
static	int	special_14_excl=0;
static	int	list_end=0;

/***************************************************/
/*** stuff for all kinds of interaction-buffers  ***/
/***************************************************/
static	int	**unitwwlist;
static	int	*max_size_unitwwlist;
static	int	cur_unitww_ptr=0;

static	int	**wwbuf[MAX_NO_LEVELS];
static	int	*antiwwbuf[MAX_NO_LEVELS];
static	int	*max_size_wwbuf[MAX_NO_LEVELS];
static	int	max_size_antiwwbuf[MAX_NO_LEVELS];

/***************************************************/
/*** buffers for extraploation			 ***/
/***************************************************/
static	double	*new_lshell[MAX_NO_LEVELS];
static	double	*old_lshell[MAX_NO_LEVELS];

static	double	*new_fx_elec,*new_fy_elec,*new_fz_elec,*new_E_elec;
static	double	*old_fx_elec,*old_fy_elec,*old_fz_elec,*old_E_elec;

static	double	*new_fx_elecvdw,*new_fy_elecvdw,*new_fz_elecvdw,*new_E_elecvdw;
static	double	*old_fx_elecvdw,*old_fy_elecvdw,*old_fz_elecvdw,*old_E_elecvdw;

static	int	*start_nearest_ptr;

static	int	*nearestwwbuf;
static	int	size_nearestwwbuf=0,nearest_ptr=0;

static	int	new_int_step[MAX_NO_LEVELS]; /*** we need to save the integrationstep***/
static	int	old_int_step[MAX_NO_LEVELS]; /*** at which update of lshell happened	***/
static	int	cur_int_step=0;

static	int	cl_per_shell[MAX_NO_LEVELS];	/*** just for checking ***/
static	int	alloc_per_shell[MAX_NO_LEVELS];

static	int	new_lshell_filled=0;
static	int	old_lshell_filled=0;

static	int	nearest_new_filled=0;
static	int	nearest_old_filled=0;
static	int	nearest_extrapol_possible=0;
static	int	nearest_old_int=0;
static	int	nearest_new_int=0;

static	int	packet_counter=0;
static	int	extrapol_possible=0;
static	int	next_update_down =-1;
static	int	last_update_down =-1;
static	int	cur_update_down  =-1,step_counter=0;

static	double	nearest_radius2 = 1/(5.0*5.0);
static	int	no_more_atoms=0;

#if FREEENERGY

#define LAMBDA get_freeenergy_lambda(global_data)

static double
get_freeenergy_lambda(LOAD_MSG_ALL_NODE *global_data)
{
	if (global_data->num_userdoubles>=1) {
		return global_data->userdouble[0];
	} else {
		errout("Fatal error: No LAMBDA defined as user-defined double\n");
		pex(-1);
	}
}
#endif

/**********************************************************************/

static	double	DC1d_A[MAX_NO_LEVELS+2][300];
static	double	DC1d_B[MAX_NO_LEVELS+2][300];

static	int
ini_DC1d()
{
	int j,i,limit;
	double A,B;

	limit = 1;
	for (j=0;j<MAX_NO_LEVELS;j++) {
		for (i=0;i<limit;i++) {
			A = (3.0*(double)limit*limit-2.0*limit+1.0) / ((double) limit*(limit+1.0));
			B = 3.0*(limit-1.0)/((double)limit*(limit+1.0));

			DC1d_A[j][i] = A-(double)i*B;
			DC1d_B[j][i] = 1.0 - DC1d_A[j][i];
		}
		limit *=2;
	}
	return 0;
}


static	double
extrapolate_dc1d(int j,int i,double fnew,double fold)
{
	double result;

	result = fnew * DC1d_A[j][i] + fold * DC1d_B[j][i];

	return result;
}

/************************************************************************/
/*** innerwwbuf holds all atoms which interact directly with an       ***/
/*** other atom closer then about 10 A. For the QM-interface we need  ***/
/*** also the unit-number of the interacting atom		      ***/
/************************************************************************/
static int
alloc_innerwwbuf(int max_no_atoms)
{
	max_size_innerwwbuf = max_no_atoms;

	innerwwbuf = (int*) alloc_ints(max_size_innerwwbuf);
	unitnumbuf = (int*) alloc_ints(max_size_innerwwbuf);
	cur_size_innerwwbuf=0; /*** number of atoms actually in the buffer ***/
	innerwwbuf[0] = 0;
	unitnumbuf[0] = 0;
	return 0;
}

static int
free_innerwwbuf()
{
	myfree(innerwwbuf);
	myfree(unitnumbuf);
	cur_size_innerwwbuf=0;
	max_size_innerwwbuf=0;
	return 0;
}

static	int
reset_innerwwbuf()
{
	cur_size_innerwwbuf=0;
	innerwwbuf[0] = 0;
	unitnumbuf[0] = 0;
	return 0;
}

/****************************************************************************/


/***************************************************************************/
/*** this function tests if ref_point and cen_geom are to far seperated. ***/
/*** If so, it calculates a new ref_point,which is shifted from cen_geom ***/
/*** to the direction of the ref_point. the "size of cluster" is enlarged***/
/*** by that shift.							 ***/
/***************************************************************************/

static
int	correct_refpoint(double *cen_refp,double *cen_geom)
{
	double dx = cen_refp[0] - cen_geom[0];
	double dy = cen_refp[1] - cen_geom[1];
	double dz = cen_refp[2] - cen_geom[2];
	double d  = sqrt(dx*dx+dy*dy+dz*dz);

	if (d>1.2) {
		double fac = 1.2/d;
		cen_refp[0] = dx*fac + cen_geom[0];
		cen_refp[1] = dy*fac + cen_geom[1];
		cen_refp[2] = dz*fac + cen_geom[2];
		cen_geom[3]+=1.2;
	}
	return 0;
}


/*********************************************************************************/
/*** this function calculates for every cluster the dipol and quadrupol-moment ***/
/*** and the ideal reference-point (cen_refp !). It starts with the unit-level ***/
/*** (which is level==0). The moments for this level are calculated by the     ***/
/*** atom-positions. For all higher levels the multipols are inherited.	       ***/
/*********************************************************************************/

int	TREE_start_loop()
{
	packet_counter=0;
	return 0;	
}

int TREE_calc_multipol(	int my_number,
			LOAD_MSG_ALL_NODE *global_data,
			TRAVELLING_PACKETS *packet)
{
	int	ind_first = global_data->first_atom[my_number];
	int	toplevel  = packet->branch.levels - 1;
	int	nr_units  = packet->branch.cl_per_level[0];
	int	level,atom,sta_atom,end_atom,typ,sta_child,end_child;
	int	unit,cluster,child;
	int	left_branch =  my_number * global_data->treeinfo.branches/global_data->no_nodes;
	struct_cluster *pa_cl,*ch_cl;
	double	*x_new,*y_new,*z_new,*q_new;
	double	sum_charge=0.0;

#if DEBUG_MULTIPOL
txtout("Node %d: other %d : In TREE_calc_multipol: nr_units %d ind_first %d left_branch %d\n",
	my_number,packet->branch.node,nr_units,ind_first,left_branch);
txtout("Node %d: branches %d no nodes %d lastlevel %d\n",
	my_number,global_data->treeinfo.branches,global_data->no_nodes,toplevel);
#endif

	x_new = packet->coordinates.x;
	y_new = packet->coordinates.y;
	z_new = packet->coordinates.z;
	q_new = packet->coordinates.q;

	/*****************************************************************/
	/***** multipols from atom-level to unit-level (cluster[0]) ******/
	/*****************************************************************/
	for (unit=0;unit<nr_units;unit++) {
		pa_cl    = &(packet->branch.cluster[0][unit]);
		sta_atom = pa_cl->child_pos - ind_first;
		end_atom = pa_cl->nr_childs + sta_atom;
		typ      = pa_cl->typ;

#if IFCHECK_CALC_MULTIPOL
/***checks***/	{
/***checks***/	double	sum_charge=0.0;
/***checks***/	if (sta_atom<0) {
/***checks***/		txtout("Node %d: OB_MULTIP: level 0 illegal atom-pos %3d %3d\n",
/***checks***/			my_number,
/***checks***/			sta_atom,end_atom);
/***checks***/	}
/***checks***/	for (atom=sta_atom;atom<end_atom;atom++) {
/***checks***/		sum_charge += q_new[atom]*global_data->f_elstat;
/***checks***/		if ((typ==NEUTRAL)&&(fabs(q_new[atom])>0.00001)) {
/***checks***/			txtout("Node %d: OB_MULTIP: charged atom in neutral unit %f\n",
/***checks***/				my_number,q_new[atom]);
/***checks***/			
/***checks***/		}
/***checks***/	}
/***checks***/	if (	(typ==CHARGED)&&
/***checks***/		fabs(pa_cl->moment[0]-sum_charge)/global_data->f_elstat>0.00001) {
/***checks***/		txtout("Node %d: OB_MULTIP: Q changed lev %2d cluster %3d % f % f\n",
/***checks***/			my_number,0,unit,
/***checks***/			sum_charge/global_data->f_elstat,
/***checks***/			pa_cl->moment[0]/global_data->f_elstat);
/***checks***/	}
/***checks***/	if (pa_cl->nr_childs<=0) {
/***checks***/		txtout("Node %d: OB_MULTIP: illegal number of childs %d (%d) on atomlevel\n",
/***checks***/			my_number,
/***checks***/			unit,
/***checks***/			pa_cl->nr_childs);
/***checks***/	}
/***checks***/	if (	(typ==POLAR)&&
/***checks***/		(fabs(sum_charge)/global_data->f_elstat>0.00001)) {
/***checks***/		txtout("Node %d: OB_MULTIP: Polar unit should have no net-charge ! %f\n",
/***checks***/			my_number,sum_charge/global_data->f_elstat);
/***checks***/	}
/***checks***/	}
#endif

		pa_cl->status = 0;
		for (atom=sta_atom,sum_charge=0.0;atom<end_atom;atom++) {
			sum_charge += q_new[atom]*global_data->f_elstat;
			if (packet->coordinates.atom_sele_flag1[atom]&FLAG_QMATOM)
				if (!(packet->coordinates.atom_sele_flag1[atom]&FLAG_QMLINK))
					pa_cl->status=1;
		}
		pa_cl->moment[0]=sum_charge;
		
		if (typ!=NEUTRAL) { /** most are POLAR or CHARGED, so we begin with it */
			double	qxx=0.0,qyy=0.0;
			double	qxy=0.0,qyz=0.0,qzx=0.0;
			double	px=0.0,py=0.0,pz=0.0;
#if FREEENERGY
			int	freeenergy_part=0;
#endif
	/*** calculate dipol-moment and quadrupol-tensor 		***/
	/*** We need in any case (CHARGED or POLAR) P and Q for		***/
	/*** center of koordinate-system. For CHARGED cluster P is	***/
	/*** 0 if choose the ideal ref-point, so we can use moment[1-3]	***/
	/*** for storing it. For POLAR units P is independant of the	***/
	/*** center of koordinate-system ! 				***/
	/*** We the Quadrupol-moments even for CHARGED clusters, if	***/
	/*** charges cancel out in a parentcluster, so the cluster	***/
	/*** becomes a POLAR-cluster ! 		  			***/

			start_bbox(pa_cl->bbox);
			for (atom=sta_atom;atom<end_atom;atom++) {
				double cur_q = q_new[atom];

				double x = x_new[atom];
				double y = y_new[atom];
				double z = z_new[atom];

				double x2 = x*x;
				double y2 = y*y;

				double r2q = cur_q * (x2 + y2 + z * z);
				double q3  = 3.0 * cur_q;
#if FREEENERGY
				freeenergy_part |= packet->coordinates.atom_sele_flag1[atom];
#endif			
				add_atom2bbox(x,y,z,pa_cl->bbox);
				
				px += cur_q * x;
				py += cur_q * y;
				pz += cur_q * z;
				
				qxx += q3 * x2 - r2q;
				qyy += q3 * y2 - r2q;
				qxy += q3 * x * y;
				qyz += q3 * y * z;
				qzx += q3 * z * x; 				
			}
			
			/*** now we have the center of geometry ***/
			bbox2cenrad(pa_cl->bbox,pa_cl->cen_geom);
			
			pa_cl->moment[1] = px*global_data->f_elstat;
			pa_cl->moment[2] = py*global_data->f_elstat;
			pa_cl->moment[3] = pz*global_data->f_elstat;

			pa_cl->moment[4] = qxx*global_data->f_elstat;
			pa_cl->moment[5] = qyy*global_data->f_elstat;
			pa_cl->moment[6] = qxy*global_data->f_elstat;
			pa_cl->moment[7] = qyz*global_data->f_elstat;
			pa_cl->moment[8] = qzx*global_data->f_elstat;
#if FREEENERGY
			/*** scale interaction between parts and the rest ***/
			if (freeenergy_part==FLAG_PART1) {
				int i;
				for (i=0;i<9;i++) pa_cl->moment[i] *= LAMBDA;
			} else
			if (freeenergy_part==FLAG_PART2) {
				int i;
				for (i=0;i<9;i++) pa_cl->moment[i] *= (1.0-LAMBDA);
			}
#endif		
			if (typ==POLAR) {
				double p2 = px * px + py * py + pz * pz;
				double p3 = 3.0 * p2;
				double p4 = 12.0 * p2 * p2;
				double qzz = -qxx - qyy;

				double qpx = qxx * px + qxy * py + qzx * pz;
				double qpy = qxy * px + qyy * py + qyz * pz;
				double qpz = qzx * px + qyz * py + qzz * pz;
				
				/*** ideal ref-point for POLAR ***/
				if (p2>1.0e-6) {
					double pqn = (qpx * px + qpy * py + qpz * pz)/p4;
					pa_cl->cen_refp[0] = qpx/p3 - pqn*px;
					pa_cl->cen_refp[1] = qpy/p3 - pqn*py;
					pa_cl->cen_refp[2] = qpz/p3 - pqn*pz;
				} else {
					pa_cl->cen_refp[0] = pa_cl->cen_geom[0];
					pa_cl->cen_refp[1] = pa_cl->cen_geom[1];
					pa_cl->cen_refp[2] = pa_cl->cen_geom[2];						
				}

			} else { /**** if CHARGED ****/
				double	sum_q=packet->branch.cluster[0][unit].moment[0]/global_data->f_elstat;
				
				/*** ideal ref-point for CHARGED ****/
				if (sum_q!=0.0) {
					pa_cl->cen_refp[0] = px/sum_q;
					pa_cl->cen_refp[1] = py/sum_q;
					pa_cl->cen_refp[2] = pz/sum_q;
				} else {
					pa_cl->cen_refp[0] = pa_cl->cen_geom[0];
					pa_cl->cen_refp[1] = pa_cl->cen_geom[1];
					pa_cl->cen_refp[2] = pa_cl->cen_geom[2];
				}
			}

			correct_refpoint(pa_cl->cen_refp,pa_cl->cen_geom);

		} else {
			/**** if NEUTRAL *****/
			/**** calculate the center of geometry ***********/

			start_bbox(pa_cl->bbox);
			for (atom=sta_atom;atom<end_atom;atom++) {
				add_atom2bbox(	x_new[atom],
						y_new[atom],
						z_new[atom],
						pa_cl->bbox);
			}
			bbox2cenrad(pa_cl->bbox,pa_cl->cen_geom);
		
			pa_cl->cen_refp[0] = pa_cl->cen_geom[0];
			pa_cl->cen_refp[1] = pa_cl->cen_geom[1];
			pa_cl->cen_refp[2] = pa_cl->cen_geom[2];
		}

		/*** reset the local expansion ***/
		pa_cl->local[0] = 0.0;
		pa_cl->local[1] = 0.0;
		pa_cl->local[2] = 0.0;
		pa_cl->local[3] = 0.0;
		pa_cl->local[4] = 0.0;
		pa_cl->local[5] = 0.0;
		pa_cl->local[6] = 0.0;
		pa_cl->local[7] = 0.0;
		pa_cl->local[8] = 0.0;
		pa_cl->local[9] = 0.0;
	} /*** end for every unit ***/

	/****************************************************************/
	/***** inherit multipols from unit-level to cluster-levels ******/
	/****************************************************************/

	for (level=1;level<=toplevel;level++) {	/*** for every level ****/
		int	nr_clusters= packet->branch.cl_per_level[level];
		int	ch_lev     = level-1;
		int	left_cl    = global_data->levelinfo[ch_lev].first_cluster[left_branch];
#if DEBUG_MULTIPOL
txtout("Node %d: In TREE_calc_multipol: toplevel %d nr_clusters %d leftcl %d\n",
my_number,toplevel,nr_clusters,left_cl);
#endif
						/*** for every cluster on this level ***/
		for (cluster=0;cluster<nr_clusters;cluster++) {
			pa_cl     = &(packet->branch.cluster[level][cluster]);
			sta_child = pa_cl->child_pos - left_cl;
			end_child = pa_cl->nr_childs + sta_child;
			typ       = pa_cl->typ;

#if IFCHECK_CALC_MULTIPOL
/***checks***/	{
/***checks***/	if (	(sta_child<0) ||
/***checks***/		(end_child>packet->branch.cl_per_level[ch_lev])) {
/***checks***/		txtout("Node %d: OB_MULTIP: illegal child-pos %d %d\n",
/***checks***/				my_number, 
/***checks***/				sta_child,end_child);
/***checks***/	}
/***checks***/	for (child=sta_child,sum_charge=0.0;child<end_child;child++) {
/***checks***/		sum_charge += packet->branch.cluster[ch_lev][child].moment[0];
/***checks***/		if (	(typ==NEUTRAL)&&
/***checks***/			(fabs(packet->branch.cluster[ch_lev][child].moment[0])/global_data->f_elstat>0.00001) ) {
/***checks***/			txtout("Node %d: OB_MULTIP: charged atom in neutral unit %f\n",
/***checks***/			my_number,
/***checks***/			packet->branch.cluster[ch_lev][child].moment[0]/global_data->f_elstat);
/***checks***/		}
/***checks***/	}
/***checks***/	if (	(typ==CHARGED) &&
/***checks***/		fabs(pa_cl->moment[0]-sum_charge)/global_data->f_elstat>0.00001) {
/***checks***/		txtout("Node %2d: OB_MULTIP: Error Q changed lev %2d cluster %3d % f % f\n",
/***checks***/			my_number,level,cluster,
/***checks***/			sum_charge/global_data->f_elstat,
/***checks***/			pa_cl->moment[0]/global_data->f_elstat);
/***checks***/	}
/***checks***/	if (pa_cl->nr_childs<=0) {
/***checks***/		txtout("Node %d: OB_MULTIP: illegal number of childs %d (%d) on clustlevel\n",
/***checks***/			my_number,
/***checks***/			cluster,
/***checks***/			pa_cl->nr_childs);
/***checks***/	}
/***checks***/	if (	(typ==POLAR)&&
/***checks***/		(fabs(sum_charge/global_data->f_elstat)>0.00001)) {
/***checks***/		txtout("Node %d: OB_MULTIP: Polar cluster should have no net-charge! %f\n",
/***checks***/				my_number,
/***checks***/				sum_charge/global_data->f_elstat);
/***checks***/	}
/***checks***/	}
#endif
			pa_cl->status = 0;
			for (child=sta_child,sum_charge=0.0;child<end_child;child++) {
				sum_charge += packet->branch.cluster[ch_lev][child].moment[0];			
				if (packet->branch.cluster[ch_lev][child].status==1) pa_cl->status = 1;
			}
			pa_cl->moment[0] = sum_charge;

			if (typ!=NEUTRAL) { /* most are POLAR or CHARGED so we begin with it */
				double	qxx=0.0,qyy=0.0;
				double	qxy=0.0,qyz=0.0,qzx=0.0;
				double	px=0.0,py=0.0,pz=0.0;
				double	maxbbox[6];
				double	maxcenrad[4];

				start_bbox(pa_cl->bbox);
				start_bbox(maxbbox);

				for (child=sta_child;child<end_child;child++) {
					ch_cl=&(packet->branch.cluster[ch_lev][child]);

					if (ch_cl->typ!=NEUTRAL) {
						px += ch_cl->moment[1];
						py += ch_cl->moment[2];
						pz += ch_cl->moment[3];

						qxx+= ch_cl->moment[4];
						qyy+= ch_cl->moment[5];
						qxy+= ch_cl->moment[6];
						qyz+= ch_cl->moment[7];
						qzx+= ch_cl->moment[8];
						merge_bbox(pa_cl->bbox,ch_cl->bbox,pa_cl->bbox);
					}
					merge_bbox(maxbbox,ch_cl->bbox,maxbbox);
				}
				
				bbox2cenrad(pa_cl->bbox,pa_cl->cen_geom);
				bbox2cenrad(maxbbox,maxcenrad);

				pa_cl->moment[1] = px;
				pa_cl->moment[2] = py;
				pa_cl->moment[3] = pz;

				pa_cl->moment[4] = qxx;
				pa_cl->moment[5] = qyy;
				pa_cl->moment[6] = qxy;
				pa_cl->moment[7] = qyz;
				pa_cl->moment[8] = qzx;

				if (typ==POLAR) {
					double p2 = px * px + py * py + pz * pz;
					double p3 = 3.0 * p2;
					double p4 = 12.0 * p2 * p2;
					double qzz = -qxx - qyy;

					double qpx = qxx * px + qxy * py + qzx * pz;
					double qpy = qxy * px + qyy * py + qyz * pz;
					double qpz = qzx * px + qyz * py + qzz * pz;

					if (p2>1.0e-6) {
						double pqn = (qpx * px + qpy * py + qpz * pz)/p4;
						pa_cl->cen_refp[0] = qpx/p3 - pqn*px;
						pa_cl->cen_refp[1] = qpy/p3 - pqn*py;
						pa_cl->cen_refp[2] = qpz/p3 - pqn*pz;
					} else {
						pa_cl->cen_refp[0] = pa_cl->cen_geom[0];
						pa_cl->cen_refp[1] = pa_cl->cen_geom[1];
						pa_cl->cen_refp[2] = pa_cl->cen_geom[2];						
					}
	
				} else {
				/*** if CHARGED *****/
					double	sum_q=pa_cl->moment[0];


					if (sum_q!=0.0) {
						pa_cl->cen_refp[0] = px/sum_q;
						pa_cl->cen_refp[1] = py/sum_q;
						pa_cl->cen_refp[2] = pz/sum_q;
					} else {
						pa_cl->cen_refp[0] = pa_cl->cen_geom[0];
						pa_cl->cen_refp[1] = pa_cl->cen_geom[1];
						pa_cl->cen_refp[2] = pa_cl->cen_geom[2];
					}
				}

				correct_refpoint(pa_cl->cen_refp,pa_cl->cen_geom);
				if (pa_cl->cen_geom[3]<maxcenrad[3]) pa_cl->cen_geom[3]=maxcenrad[3];

			} else { /*** if cluster is NEUTRAL *****/
				start_bbox(pa_cl->bbox);
				for (child=sta_child;child<end_child;child++) {
					merge_bbox(	pa_cl->bbox,
							packet->branch.cluster[ch_lev][child].bbox,
							pa_cl->bbox);
				}

				bbox2cenrad(pa_cl->bbox,pa_cl->cen_geom);
				pa_cl->cen_refp[0] = pa_cl->cen_geom[0];
				pa_cl->cen_refp[1] = pa_cl->cen_geom[1];
				pa_cl->cen_refp[2] = pa_cl->cen_geom[2];
			}

			/*** reset local expansion ***/
			pa_cl->local[0] = 0.0;
			pa_cl->local[1] = 0.0;
			pa_cl->local[2] = 0.0;
			pa_cl->local[3] = 0.0;
			pa_cl->local[4] = 0.0;
			pa_cl->local[5] = 0.0;
			pa_cl->local[6] = 0.0;
			pa_cl->local[7] = 0.0;
			pa_cl->local[8] = 0.0;
			pa_cl->local[9] = 0.0;

		} /*** end for every cluster on this level ****/
	} /*** end for every level ****/

#if DEBUG_MULTIPOL
txtout("Node %d: Exit TREE_calc_multipol:\n",my_number);
#endif
	return 0;
}

/****************************************************************************/

int	TREE_update(	int my_number,
			LOAD_MSG_ALL_NODE *global_data,
			LOAD_MSG_ONE_NODE *my_data)
{
	int	left_branch,right_branch;
	int	left_cl,right_cl,nr_cl,level;

	while (my_data->no_atoms > no_more_atoms) {
		realloc_extrapolation(&no_more_atoms, my_data->no_atoms - no_more_atoms +10,
			    &new_fx_elecvdw, &new_fy_elecvdw, &new_fz_elecvdw, &new_E_elecvdw,
			    &old_fx_elecvdw, &old_fy_elecvdw, &old_fz_elecvdw, &old_E_elecvdw,
			    &new_fx_elec, &new_fy_elec, &new_fz_elec, &new_E_elec,
			    &old_fx_elec, &old_fy_elec, &old_fz_elec, &old_E_elec);
	}

	for (level=0;level<global_data->treeinfo.levels;level++) {
		left_branch =  my_number   * global_data->treeinfo.branches/global_data->no_nodes;
		right_branch= (my_number+1)* global_data->treeinfo.branches/global_data->no_nodes - 1;

		left_cl = global_data->levelinfo[level].first_cluster[left_branch];
		right_cl= global_data->levelinfo[level].last_cluster [right_branch];		
		nr_cl   = right_cl - left_cl + 1;

		if (nr_cl>alloc_per_shell[level]) {
			errout("Node %d: Fatal error: Now you have to realloc in TREE_update #1 %d %d level %d\n",
				my_number,nr_cl,alloc_per_shell[level],level);
			errout("Node %d: left_cl %d  right_cl %d   left_br %d  right_br %d\n",
				my_number,left_cl,right_cl,left_branch,right_branch);
			pex(-1);
		}
		cl_per_shell[level]    = nr_cl;
	}

	reset_treestatus_flags();

	packet_counter    = 0;
	return 0;
}

static int
reset_treestatus_flags()
{
	new_lshell_filled = 0;
	old_lshell_filled = 0;
	extrapol_possible = 0;
	next_update_down  = -1;
	last_update_down  = -1;
	cur_int_step      = 0;

	nearest_new_filled=0;
	nearest_old_filled=0;
	nearest_extrapol_possible=0;
	nearest_old_int=0;
	nearest_new_int=0;
	return 0;
}


int	TREE_ini_all(	int my_number,
			LOAD_MSG_ALL_NODE *global_data,
			LOAD_MSG_ONE_NODE *my_data)
{
	int	level,node;
	int	cur_size;
	int	max_clplpn[MAX_NO_LEVELS],atoms_per_level[MAX_NO_LEVELS];
	int	left_branch,right_branch,left_cl,right_cl,nr_cl;
	double	clsize_on_level[MAX_NO_LEVELS],clvol_on_level[MAX_NO_LEVELS];
	int	sum_all_over=0;
#if DEBUG_INI
txtout("Node %d: Start TREE_ini_all\n",my_number);
#endif
	/********************************************************************/
	/*** we want the maximal number of clusters per node on any level ***/
	/********************************************************************/
	if (global_data->no_nodes>1)
			no_more_atoms = (int)(1.3*global_data->maxatoms_per_node);
		else	no_more_atoms = (int)(my_data->no_atoms+10);

	for (level=0;level<MAX_NO_LEVELS;level++) { /*** just for ini ***/
		max_clplpn[level]     = 10;
		atoms_per_level[0]    = 10;
		clvol_on_level[level] = 100.0;
		clsize_on_level[level]= 100.0;
	}

	for (level=0;level<global_data->treeinfo.levels;level++) {
		max_clplpn[level] = 10;
		if (level==0) {
			atoms_per_level[0] = 3;
		} else {
			atoms_per_level[level]=
				atoms_per_level[level-1]*global_data->treeinfo.clusterseq[level];
		}
		{
		double	atoms_per_vol = 0.1;	/*** typical value for atoms per Angstr^3 ***/
		double	vol           = atoms_per_level[level]/atoms_per_vol;
		clvol_on_level[level] = vol*1.2;	/*** factor 1.2 for sure-is-sure ****/
		clsize_on_level[level]= 0.5*sqrt(3.0)*pow(clvol_on_level[level],1.0/3.0);
		}
		for (node=0;node<global_data->no_nodes;node++) {
			left_branch =  node   * global_data->treeinfo.branches/global_data->no_nodes;
			right_branch= (node+1)* global_data->treeinfo.branches/global_data->no_nodes - 1;

			
			left_cl = global_data->levelinfo[level].first_cluster[left_branch];
			right_cl= global_data->levelinfo[level].last_cluster[right_branch];		
			nr_cl   = right_cl - left_cl + 1;
			if (nr_cl>max_clplpn[level]) max_clplpn[level] = nr_cl;
		}
	}

	if (global_data->no_nodes>1)
			for (level=0;level<global_data->treeinfo.levels;level++) 
				max_clplpn[level] = (int)(max_clplpn[level]*1.2);
		else
			for (level=0;level<global_data->treeinfo.levels;level++) 
				max_clplpn[level] = max_clplpn[level]+10;

	/*************************************************************************/
	/*** here we alloc the mem for local expansion for each shell on every ***/
	/*** level, which we need for extrapolation for multiple-time-step     ***/
	/*************************************************************************/
	set_allocinfo("TREE_ini_all lshell");
	for (level=0;level<global_data->treeinfo.levels;level++) {
		left_branch =  my_number   * global_data->treeinfo.branches/global_data->no_nodes;
		right_branch= (my_number+1)* global_data->treeinfo.branches/global_data->no_nodes - 1;

		left_cl = global_data->levelinfo[level].first_cluster[left_branch];
		right_cl= global_data->levelinfo[level].last_cluster [right_branch];		
		nr_cl   = right_cl - left_cl + 1;

		old_lshell[level] = alloc_doubles( 3 * 10 * nr_cl );
		new_lshell[level] = alloc_doubles( 3 * 10 * nr_cl );
		cl_per_shell[level]    = nr_cl;
		alloc_per_shell[level] = nr_cl*3;

		sum_all_over += 3* 2 * sizeof(double) * 10 * nr_cl;
	}
	set_allocinfo("XXX");
	new_lshell_filled = 0;
	old_lshell_filled = 0;
	extrapol_possible = 0;
	next_update_down  = -1;
	last_update_down  = -1;
	cur_int_step      = 0;

	nearest_new_filled=0;
	nearest_old_filled=0;
	nearest_extrapol_possible=0;
	nearest_old_int=0;
	nearest_new_int=0;

	/*********************************************************************/
	/*** here we calc the maximum number of atoms in the inner ww-list ***/
	/*********************************************************************/
	{
	double	atoms_per_vol = 0.1;		/*** typical value for atoms per Angstr^3 ***/
	double	mean_ext_unit = 4.0;		/*** in Angstroem			***/
	double	radius        = global_data->outer_classrad[0] + mean_ext_unit;
	double	vol           = 4.0/3.0 * PI * radius * radius * radius;
	double	no_atoms      = vol * atoms_per_vol;
	int	max_no_atoms  = (int)(no_atoms*4.0);	/*** factor 4.0 for sure-is-sure ***/	
	double	nearest_radius;

	set_allocinfo("TREE_ini_all innerww");
	alloc_innerwwbuf(max_no_atoms);
	sum_all_over += sizeof(int)*max_no_atoms;
		
	alloc_extrapolation(no_more_atoms, 
			    &new_fx_elecvdw, &new_fy_elecvdw, &new_fz_elecvdw, &new_E_elecvdw,
			    &old_fx_elecvdw, &old_fy_elecvdw, &old_fz_elecvdw, &old_E_elecvdw,
			    &new_fx_elec, &new_fy_elec, &new_fz_elec, &new_E_elec,
			    &old_fx_elec, &old_fy_elec, &old_fz_elec, &old_E_elec);
	sum_all_over += 6*sizeof(double)*no_more_atoms;

	nearest_radius = 5.0;
	vol            = 4.0/3.0 * PI * nearest_radius * nearest_radius * nearest_radius;
	max_no_atoms   = (int)(vol * atoms_per_vol * 2.0);
	size_nearestwwbuf = no_more_atoms * max_no_atoms;

	nearestwwbuf  = alloc_ints( size_nearestwwbuf );
	sum_all_over += sizeof(int)*size_nearestwwbuf;

	start_nearest_ptr = alloc_ints(global_data->no_nodes);
	}

	set_allocinfo("XXX");
	/**********************************************************************/
	/*** now we do the allocation stuff (there should be no reallocation **/
	/*** needed)							     **/ 		
	/**********************************************************************/

	{
	double 	ri      = global_data->outer_classrad[0]+clsize_on_level[0];
	double	ra	= global_data->outer_classrad[1]+clsize_on_level[0];
	double	shelvol = 4.0/3.0*PI*(ra*ra*ra - ri*ri*ri);

	cur_size = max_clplpn[0]*(int)(1.0+1.5*shelvol/(clvol_on_level[0]*global_data->no_nodes));
	cur_size = min(cur_size,max_clplpn[0]*(max_clplpn[0] + 1))+10;
	}
	set_allocinfo("TREE_ini_all unitww");
#if TRACE_REALLOC
if (my_number==1) {
	txtout("Default alloc for unitww-list: %d x %d x %d (max_clplpn[0] %d) Sum %d\n",
	cur_size,global_data->no_nodes,sizeof(int),max_clplpn[0],
	cur_size*global_data->no_nodes*sizeof(int));
	
}
#endif
	unitwwlist 	    = (int**)alloc_chars(global_data->no_nodes*sizeof(int*));
	max_size_unitwwlist = (int*) alloc_ints(global_data->no_nodes);
	
	for (node=0;node<global_data->no_nodes;node++) {
		unitwwlist[node] = (int*) alloc_ints(cur_size);
		sum_all_over    += sizeof(int)*cur_size;
		max_size_unitwwlist[node] = cur_size;
	}
	set_allocinfo("TREE_ini_all wwbuf");
	for (level=0;level<max(1,global_data->treeinfo.levels-1);level++) {
		wwbuf[level]          = (int**)alloc_chars(global_data->no_nodes*sizeof(int*));
		max_size_wwbuf[level] = (int*) alloc_ints(global_data->no_nodes);
	}

	for (level=0;level<max(1,global_data->treeinfo.levels-1);level++) {
		double 	ri      = global_data->outer_classrad[level]  +clsize_on_level[level];
		double	ra	= global_data->outer_classrad[level+1]+clsize_on_level[level];
		double	shelvol = 4.0/3.0*PI*(ra*ra*ra - ri*ri*ri);

		cur_size = max_clplpn[level]*
			(int)(1.0+2.0*shelvol/(clvol_on_level[level]*global_data->no_nodes));
if (level==0) cur_size*=2;
		cur_size = min(cur_size,max_clplpn[level]*(max_clplpn[level] + 1))+10;
#if TRACE_REALLOC
if (my_number==1) {
	txtout("Default alloc for wwbuf-list: %d x %d x %d (max_clplpn[%d] %d) Sum %d\n",
	cur_size,global_data->no_nodes,sizeof(int),level,max_clplpn[level],
	cur_size*global_data->no_nodes*sizeof(int));
	
}
#endif
		for (node=0;node<global_data->no_nodes;node++) {
			wwbuf[level][node]          = (int*) alloc_ints(cur_size);
			max_size_wwbuf[level][node] = cur_size;
			sum_all_over += sizeof(int)*cur_size;
		}
	}

	/********************************************************************/
	/*** antiwwbuf is used for every other branch. is only needed for ***/
	/*** building up the ww-list, which is coded in 		  ***/		
	/*** wwbuf[level][other_number] 				  ***/
	/********************************************************************/
	set_allocinfo("TREE_ini_all antiww");
	for (level=0;level<max(1,global_data->treeinfo.levels-1);level++) {
		cur_size = max_clplpn[level+1]*(max_clplpn[level] + 1)+10;
		antiwwbuf[level]          = (int*) alloc_ints(cur_size);
		max_size_antiwwbuf[level] = cur_size;
		sum_all_over += sizeof(int)*cur_size;
	}

	ini_DC1d();

#if DEBUG_INI
txtout("Node %d: Exit TREE_ini_all: %d\n",my_number,sum_all_over);
#endif
	return 0;
}

/***************************************************************************/

int	TREE_exit(LOAD_MSG_ALL_NODE *global_data)
{
	int	level,node;

	free_innerwwbuf();

	myfree(max_size_unitwwlist);

	for (level=0;level<global_data->treeinfo.levels;level++) {
		myfree(old_lshell[level]);
		myfree(new_lshell[level]);
	}

	free_extrapolation(	&no_more_atoms,
				new_fx_elecvdw, new_fy_elecvdw, new_fz_elecvdw, new_E_elecvdw,
				old_fx_elecvdw, old_fy_elecvdw, old_fz_elecvdw, old_E_elecvdw,
				new_fx_elec,    new_fy_elec,    new_fz_elec,    new_E_elec,
				old_fx_elec,    old_fy_elec,    old_fz_elec,    old_E_elec);
	myfree(nearestwwbuf);
	myfree(start_nearest_ptr);

	for (node=0;node<global_data->no_nodes;node++) {

		myfree(unitwwlist[node]);

		for (level=0;level<global_data->treeinfo.levels-1;level++) {
			myfree(wwbuf[level][node]);
			max_size_wwbuf[level][node] = 0;
		}
	}

	myfree(unitwwlist);
	for (level=0;level<global_data->treeinfo.levels-1;level++) {
		myfree(wwbuf[level]);
		myfree(max_size_wwbuf[level]);
	}

	for (level=0;level<global_data->treeinfo.levels-1;level++) {
		myfree(antiwwbuf[level]);
	}
	return 0;
}

/***************************************************************************/

static	int
update_lshells (int my_number,
		int level,
		LOAD_MSG_ALL_NODE *global_data,
		LOAD_MSG_ONE_NODE *my_data)
{
	int my_nr_clusters = my_data->cl_per_level[level];
	int cluster,bufptr,i=global_data->no_nodes;

#if IFCHECK_CALC_LONGRANGE
/***checks***/	if (my_nr_clusters!=cl_per_shell[level]) {
/***checks***/		errout("Node %d: got confusion in update_lshells %d %d\n",
/***checks***/		my_number,my_nr_clusters,cl_per_shell[level]);
/***checks***/		pex(-1);
/***checks***/	}
#endif
	if (new_lshell_filled) {
		memcpy(	&(old_lshell[level][0]),
			&(new_lshell[level][0]),
			10*sizeof(double)*cl_per_shell[level]);
		old_int_step[level] = new_int_step[level];
	}

	for (cluster=0,bufptr=0;cluster<my_nr_clusters;cluster++,bufptr+=10) {
		memcpy(	&(new_lshell[level][bufptr]),
			&(my_data->cluster[level][cluster].local[0]),
			10*sizeof(double) );

		if (!new_lshell_filled) continue;

		for (i=0;i<10;i++) {
			my_data->cluster[level][cluster].local[i] = 
					extrapolate_dc1d(
						level+2,
						0,
						new_lshell[level][bufptr+i],
						old_lshell[level][bufptr+i]);
		}
	}
	new_int_step[level] = cur_int_step;
	return 0;
}

/****************************************************************************/

static	int
extrapolate_lshells (	int next_update_down,
			LOAD_MSG_ALL_NODE *global_data,
			LOAD_MSG_ONE_NODE *my_data)
{
	int toplevel = max(0,my_data->levels-2);
	int level,cluster,bufptr,i=global_data->no_nodes;
	int di;	

	for (level=toplevel;level>=next_update_down+1;level--) {
		di = new_int_step[level]-old_int_step[level];

#if IFCHECK_CALC_LONGRANGE
/***checks***/	if ((di<=0)||(cur_int_step<=new_int_step[level])) {
/***checks***/		errout("Fatal error: That can not be in extrapolate_lshells %d  %d %d lev %d\n",
/***checks***/		di,new_int_step[level],old_int_step[level],level);
/***checks***/		pex(-1);
/***checks***/	}
#endif
		for (cluster=0,bufptr=0;cluster<cl_per_shell[level];cluster++,bufptr+=10) {
			for (i=0;i<10;i++) {
				my_data->cluster[level][cluster].local[i] = 
					extrapolate_dc1d(
						level+2,
						cur_int_step-new_int_step[level],
						new_lshell[level][bufptr+i],
						old_lshell[level][bufptr+i]);
			}
		}
	}
	return 0;
}

/*****************************************************************************/

static	int	next_wwbuf_ptr = 0;
static	int	cur_wwbuf_ptr  = 0;
static	int	last_wwbuf_ptr = 0;
static	int	interact_ptr   = 0;

/*****************************************************************************/
/*** mode: 1.Bit -> 1: calc new ww-list,	 0: use last ww-list	   ***/
/***	   2.Bit -> 1: calc new local expansion	 0:use old local expansion ***/
/*****************************************************************************/

int	TREE_calc_LongRange(	int	my_number,
				LOAD_MSG_ALL_NODE *global_data,
				LOAD_MSG_ONE_NODE *my_data,
				TRAVELLING_PACKETS *packet,
				int mode)

{
	int	my_cluster,level;
	int	toplevel    = max(0,packet->branch.levels - 2);
	int	other_number= packet->branch.node;
	int	my_nr_clusters;
	int	inner_wwptr;

#if DEBUG_LR
txtout("Node %d: In TREE_calc_LongRange: other %d  mode %d packet_counter %d\n",
my_number,packet->branch.node,mode,packet_counter);
#endif	
	packet_counter++;

	if (!(mode & (int)0x01)) goto skip_wwlist;

	if (packet_counter==1) reset_treestatus_flags();

	/******************************************************************/
	/*** here we update the new ww-list				***/
	/******************************************************************/
#if DEBUG_LR
if (packet_counter==1) txtout("Node %d: now calc ww-list other %d toplevel %d\n",
my_number,packet->branch.node,toplevel);
#endif
	{/******** here we mark the toplevel for work on it **********/
	int	my_nr_clusters    = my_data->cl_per_level[toplevel];
	int	other_nr_clusters = packet->branch.cl_per_level[toplevel];
	int	c=0,i;

	for (i=0;i<other_nr_clusters;i++) antiwwbuf[toplevel][c++] = i;
	antiwwbuf[toplevel][c++] = -my_nr_clusters;	/*** Minus signals the end and how      ***/
							/*** often to use this part (saves RAM) ***/

	next_wwbuf_ptr = c;

#if IFCHECKBUFSIZE
/***checks***/	if (my_nr_clusters<=0) {
/***checks***/		errout("Node %d: illegal value of my_nr_clusters %d\n",my_number,my_nr_clusters);
/***checks***/		pex(-1);
/***checks***/	}
/***checks***/	if (c>=max_size_antiwwbuf[toplevel]) {
/***checks***/		errout("Node %d: buffer to small #1 (%d %d)\n",my_number,c,max_size_antiwwbuf[toplevel]);
/***checks***/		pex(-1);
/***checks***/	}
#endif
	}/******** end ini toplevel ********/

	/**********************************************************************/
	/*** here we test interaction of cluster on every level and make the **/
	/*** final interaction list for the innermost ww-shell		     **/
	/**********************************************************************/

	cur_unitww_ptr=0;
	for (level=toplevel;level>=0;level--) {		/*** for every level **/
#if DEBUG_LR
txtout("Node %d: ww-list level %d\n",my_number,level);
#endif
		my_nr_clusters = my_data->cl_per_level[level];
		interact_ptr   = 0;
		last_wwbuf_ptr = next_wwbuf_ptr;
		cur_wwbuf_ptr  = 0;	/*** point to beginning of ww-list for this level ***/
		next_wwbuf_ptr = 0;	/*** point to beginning of ww-list for next level ***/
					/*** is going to be made in TREE_force_tree2tree  ***/

					/*** for every cluster on this level **/
		for (my_cluster=0;my_cluster<my_nr_clusters;my_cluster++) {						
			TREE_force_tree2tree(	my_number,
						my_cluster,
						level,
						global_data,
						my_data,
						packet);

			put_to_wwbuf(level,other_number,-1);

		} /*** end for every cluster on this level ***/

	} /*** for every level ***/


skip_wwlist:
#if DEBUG_LR
txtout("Node %d: now calc inner ww other %d\n",my_number,packet->branch.node);
#endif
	/*********************************************************************/
	/*** now we have the interaction-list on the unit-level		   ***/
	/*** and so we can calculate the interaction of the innermost shell***/
	/*********************************************************************/
	if (!extrapol_possible) next_update_down = toplevel;

	/**** ini the exclusion stuff ***/
	ex_list   = my_data->ex_list;
	ex14_list = my_data->ex14_list;
	ex_ptr    = 0;
	ex14_ptr  = 0;
	ext_first = global_data->first_atom[packet->coordinates.node];
	list_end  = END_OF_LIST - ext_first;
	special_14_excl = global_data->special_14_excl;

	my_nr_clusters = my_data->cl_per_level[0];
	inner_wwptr=0;

	/******************************************************************/
	/*** handle startup and filling of force-buffers, which		***/
	/*** we need for extrapolation.					***/
	/******************************************************************/
	if (	(!nearest_extrapol_possible) &&
		(packet_counter==1)) {
		int i;
		if (nearest_new_filled) {
			memcpy(old_fx_elecvdw,new_fx_elecvdw,sizeof(double)*my_data->no_atoms);
			memcpy(old_fy_elecvdw,new_fy_elecvdw,sizeof(double)*my_data->no_atoms);
			memcpy(old_fz_elecvdw,new_fz_elecvdw,sizeof(double)*my_data->no_atoms);
			memcpy(old_E_elecvdw ,new_E_elecvdw ,sizeof(double)*my_data->no_atoms);

			memcpy(old_fx_elec,new_fx_elec,sizeof(double)*my_data->no_atoms);
			memcpy(old_fy_elec,new_fy_elec,sizeof(double)*my_data->no_atoms);
			memcpy(old_fz_elec,new_fz_elec,sizeof(double)*my_data->no_atoms);
			memcpy(old_E_elec ,new_E_elec ,sizeof(double)*my_data->no_atoms);

			nearest_old_int    = nearest_new_int;

			if (nearest_old_filled==0)	nearest_old_filled = -1;
			else				nearest_old_filled = 1;
		}
		for (i=0;i<my_data->no_atoms;i++) {
			new_fx_elecvdw[i] = new_fy_elecvdw[i] = new_fz_elecvdw[i] = 0.0;
			new_fx_elec[i]    = new_fy_elec[i]    = new_fz_elec[i]    = 0.0;
			new_E_elec[i]     = new_E_elecvdw[i]  = 0.0;
		}
		nearest_new_int    = cur_int_step;
		nearest_new_filled = 1;
	}	
	/******************************************************************/
	/*** set the ptr-position to the nearest-wwbuffer needed for	***/
	/*** this packet.						***/
	/******************************************************************/
	if (mode & (int)0x01) {	/*** new wwlist ***/
		if (packet_counter==1)	nearest_ptr = 0;
		start_nearest_ptr[other_number] = nearest_ptr;
		
	} else {		/*** use old wwlist ***/
		nearest_ptr = start_nearest_ptr[other_number];
	}
#if DEBUG_LR
txtout("Node %d: before loop over inner clusters %d  nearest_ptr %d packetc %d\n",
my_number,my_nr_clusters,nearest_ptr,packet_counter);
#endif
	for (my_cluster=0;my_cluster<my_nr_clusters;my_cluster++) {	
		TREE_calc_innerww(	my_number,
					global_data,
					my_data,
					packet,
					my_cluster,
					&inner_wwptr,
					mode);
	}
#if DEBUG_LR
txtout("Node %d: after loop over inner clusters %d\n",my_nr_clusters);
#endif
	if (packet_counter==global_data->no_nodes) {
 		if (nearest_extrapol_possible) {
			nearest_extrapol_possible = 0;
		} else {
#if NO_EXTRAPOL
			nearest_extrapol_possible = 0;
#else
			if (nearest_old_filled==1)
				nearest_extrapol_possible = 1;
#endif
		}
	} 

	if ((mode & (int)0x02)==0) goto skip_localexp;

	/*********************************************************************/
	/*** here we calculate the contribution of the cluster in our      ***/
	/*** ww-list to the local expansion on any level		   ***/
	/*********************************************************************/
#if DEBUG_LR
if ((packet_counter==1)&&(my_number==0)) txtout("Node %d: now calc cluster-cluster other %d  next down %d\n",my_number,packet->branch.node,next_update_down);
#endif
	for (level=next_update_down;level>=0;level--) {		/*** for every level ****/
		my_nr_clusters = my_data->cl_per_level[level];
		interact_ptr   = 0;		
					/*** for every cluster on this level ***/
		for (my_cluster=0;my_cluster<my_nr_clusters;my_cluster++) {
			int cluster;

			while ((cluster=wwbuf[level][other_number][interact_ptr])!=-1) {
#if IFCHECK_CALC_LONGRANGE
/***checks***/	if ( (cluster<0) || (cluster>=packet->branch.cl_per_level[level]) ||
/***checks***/     (interact_ptr>=max_size_wwbuf[level][other_number])){
/***checks***/		errout("Node %d: going wrong cluster %d   limit %d my_cl %d lim %d lev %d othernum %d iptr %d lim %d\n",
/***checks***/		my_number,cluster,packet->branch.cl_per_level[level],my_cluster,my_nr_clusters,level,
/***checks***/		other_number,interact_ptr,max_size_wwbuf[level][other_number]);
/***checks***/		pex(-1);
/***checks***/	}
#endif
				TREE_local_cl2cl(&(my_data->cluster[level][my_cluster]),
						 &(packet->branch.cluster[level][cluster]));

				interact_ptr++;
			}
			interact_ptr++;		
		}
	}

	/**********************************************************************/
	/*** if the last branch has been done, we can now inherit the local ***/
	/*** expansion from toplevel to the unit-level or do our extra-     ***/
	/*** polation also the stuff for next multiple-timestep-mode is done***/
	/**********************************************************************/

	if (packet_counter==global_data->no_nodes) {
		int	effort_coord,effort_branch;

#if DEBUG_LR
txtout("Node %d: now inherit other %d\n",my_number,packet->branch.node);
#endif
		packet_counter = 0;	/*** reset for next cycle ***/

#if NO_EXTRAPOL
		next_update_down = toplevel;
		last_update_down = toplevel;
		extrapol_possible=0;
		new_lshell_filled=0;
		old_lshell_filled=0;
#endif
		if (extrapol_possible) {
			extrapolate_lshells(next_update_down,global_data,my_data);

			for (level=next_update_down;level>=0;level--) 
				update_lshells(my_number,level,global_data,my_data);

			if (step_counter==3) {
				if (cur_update_down<=0) {
					next_update_down = last_update_down + 1;
					last_update_down = next_update_down;

					if (last_update_down>=toplevel) {
						last_update_down = 0;
						next_update_down = toplevel;
					}
				} else {
					next_update_down = 0;
				}
				cur_update_down  = next_update_down;
			} else {
				next_update_down = -1;
			}
			step_counter++;
			if (step_counter>3) step_counter=0;		

		} else { /*** handle startup-situation ***/
			for (level=toplevel;level>=0;level--) 
				update_lshells(my_number,level,global_data,my_data);

			if (new_lshell_filled) {
				old_lshell_filled = 1;
				extrapol_possible = 1;
			} else {
				new_lshell_filled = 1;
			}

			if (extrapol_possible) {
				next_update_down = -1;
				last_update_down = -1;
				cur_update_down  = -1;
			} else {
				next_update_down = toplevel;
				last_update_down = toplevel;
			}
			step_counter=1;
		}

		for (level=toplevel;level>=1;level--) 
			for (my_cluster=0;my_cluster<my_data->cl_per_level[level];my_cluster++) 
				inherit_local_expansion(my_cluster,level,my_data);

		get_communication_effort(&effort_coord,&effort_branch);
		if (next_update_down<0) {
			effort_branch = COMEFFORT_SKIP;
		} else {
			effort_branch = COMEFFORT_REDUCED;
		}
		set_communication_effort(effort_coord,effort_branch);
		cur_int_step++;
	}

#if DEBUG_LR
txtout("Node %d: exit now TREE_calc_LongRange other %d\n",my_number,packet->branch.node);
#endif

skip_localexp:
	return 0;
}

/**************************************************************************/
/*** here the long-range interaction (VDW+Elec) of the inner ww-list	***/
/*** is calculated							***/
/**************************************************************************/

static int	first_atom=-1,last_atom=-1;

static int
TREE_calc_innerww(	int	my_number,
			LOAD_MSG_ALL_NODE *global_data,
			LOAD_MSG_ONE_NODE *my_data,
			TRAVELLING_PACKETS *packet,
			int my_cluster,
			int *inner_wwptr,
			int mode)
{
	int	ind_first = global_data->first_atom[packet->branch.node];
	int	unit,sta_child,end_child,sta_atom,other_atom,atom;
	int	special_14excl = global_data->special_14_excl;
	int	ex14_flag=0,c;
	int	other_number   = packet->branch.node;

	first_atom = my_data->cluster[0][my_cluster].child_pos - my_data->cluster[0][0].child_pos;
	last_atom = first_atom + my_data->cluster[0][my_cluster].nr_childs;

	if (	(nearest_extrapol_possible) &&
		(global_data->nr_qmatoms==0)&&
		(global_data->schwing_mode==0)
	   ) {
#if DEBUG
txtout("Node %d: before loop do_nearest_extrapol %d %d\n",
my_number,first_atom,last_atom);
#endif
		for (atom=first_atom;atom<last_atom;atom++) {
			do_nearest_extrapol(	my_number,
						atom,
						global_data,
						my_data,
						packet);
		}
#if DEBUG
txtout("Node %d: after loop do_nearest_extrapol %d %d\n",
my_number,first_atom,last_atom);
#endif
		
		return 0;	
	}
#if DEBUG
txtout("Node %d: before buildup loop %d %d\n",
my_number,first_atom,last_atom);
#endif
	/*********************************/
	/*** for every atom in my unit ***/
	/*********************************/
	c = *inner_wwptr;
	my_unitnum = my_cluster;
	for (atom=first_atom;atom<last_atom;atom++) {
		reset_innerwwbuf();

		if (packet->branch.node==my_number)	sta_atom = atom;
		else					sta_atom = -1;

		/*****************************************/
		/*** Manage start of exlusion list     ***/	
		/*****************************************/
		ex_elem = ex_list[ex_ptr] - ext_first;	/* this atom is skipped first */
							/* skip exclusions outside the range! */
		while (	(ex_elem <= sta_atom) && 
			(ex_elem != list_end)) {
			ex_elem = ex_list[++ex_ptr] - ext_first;
		}
						/* setup ex14_elem in appropriate range */
						/* skip exclusions outside the range! */
		ex14_elem = ex14_list[ex14_ptr] - ext_first;
		while (	(ex14_elem <= sta_atom) &&
			(ex14_elem != list_end) &&
			special_14_excl) {
			ex14_elem = ex14_list[++ex14_ptr] - ext_first;
		}

		c = *inner_wwptr;

		/*******************************************/
		/*** for every unit of the inner ww-list ***/
		/*******************************************/
		while ((unit=unitwwlist[other_number][c])!=-1) {
			sta_child = packet->branch.cluster[0][unit].child_pos - ind_first;
			end_child = packet->branch.cluster[0][unit].nr_childs + sta_child;
			c++;
			if (end_child-1<=sta_atom)	continue;
			if (sta_child  <=sta_atom)	sta_child=sta_atom+1;

			/*** check exclusion for every atom in unit ***/
			for (other_atom=sta_child;other_atom<end_child;other_atom++) {
#if IFCHECK_CALC_LONGRANGE
/***checks***/			if ((other_atom>ex_elem)&&(ex_elem!=list_end)) {
/***checks***/				errout("Node %d: Fatal error: Consistency check failed for exclusion list! other %d excl %d for atomnr %d\n",
/***checks***/					my_number,other_atom,ex_elem,atom);
/***checks***/				errout("Node %d: This can happen if you are not using the right clustering algorithm for\n",
/***checks***/					my_number);
/***checks***/				errout("Node %d: systems with extended areas containing neutral atoms, e.g., lipid membranes.\n",
/***checks***/					my_number);
/***checks***/				errout("Node %d: For membranes use as clustering algorithm \"2\" in ctl.lis, but not \"0\" or \"1\"!\n",
/***checks***/					my_number);
/***checks***/				errout("Node %d: This error also occurs if your structure has \"exploded\" and excluded atoms\n",
/***checks***/					my_number);
/***checks***/				errout("Node %d: are more than 5  apart.\n",
/***checks***/					my_number);
/***checks***/				pex(-1);
/***checks***/			}
#endif
				ex14_flag = special_14excl && (other_atom==ex14_elem);
				if (other_atom!=ex_elem) {
#if IFCHECKBUFSIZE
/***checks***/				if (cur_size_innerwwbuf+1>=max_size_innerwwbuf) {
/***checks***/					errout("Node %d: Fatal error: That should not be: we dont want reallocs or if-checks for innerwwbuf\n",
/***checks***/						my_number);
/***checks***/					errout("Node %d:we have %d other %d unit %d\n",
/***checks***/						my_number,max_size_innerwwbuf,other_atom,unit);
/***checks***/					pex(-1);
/***checks***/				}
#endif					
					unitnumbuf[cur_size_innerwwbuf  ] = unit;		
					innerwwbuf[cur_size_innerwwbuf++] = 
						(ex14_flag)?-other_atom-1:other_atom+1;
				} else {
					/*****************************************/
					/*** now we get the next excluded atom ***/
					/*****************************************/
					ex_elem = ex_list[++ex_ptr] - ext_first;
				}

				if (ex14_flag) {	/* set next ex14 element */
					ex14_elem = ex14_list[++ex14_ptr] - ext_first;
				}
			}
		} /*** end for every unit of the inner ww-list ***/
#if DEBUG
txtout("Node %d: before calc_extact_interact %d %d %d\n",
my_number,atom,first_atom,last_atom);
#endif
		calc_exact_interact(	my_number,
					atom,
					global_data,
					my_data,
					packet,
					mode);
#if DEBUG
txtout("Node %d: after calc_extact_interact %d %d %d\n",
my_number,atom,first_atom,last_atom);
#endif
		/***************************************/
		/**** find exclusions for next atom ****/
		/***************************************/
		while (ex_elem != list_end) {
			ex_elem = ex_list[++ex_ptr] - ext_first;
		}

		while (special_14_excl && (ex14_elem != list_end)) {
			ex14_elem = ex14_list[++ex14_ptr] - ext_first;
		}
			
		/*** skip end of exclusion lists ***/
		ex_ptr++;
		ex14_ptr++;

	} /*** end for every atom in my unit ***/
	*inner_wwptr = c+1;	/*** count on ***/

#if DEBUG
txtout("Node %d: after buildup loop %d %d\n",
my_number,first_atom,last_atom);
#endif
	return 0;
}

double get_min_unitdist(int flag,
		int my_atom,int other_unitnum,
		LOAD_MSG_ONE_NODE *my_data,
		TRAVELLING_PACKETS *packet)
{
	int j,other_atom;
 	double newdist=9999999.0,d,dxt,dyt,dzt;

	if (flag==0) {
		if (other_unitnum!=unitnumbuf[posinnerwwbuf]) {
			errout("Fatal error #1 in get_min_unitdist %d %d %d\n",
				other_unitnum,unitnumbuf[posinnerwwbuf],posinnerwwbuf);
			pex(-1);
		}

		for (j=posinnerwwbuf-1;j>0;j--) {
 			if (other_unitnum==unitnumbuf[j]) {
				if ((other_atom=innerwwbuf[j])<0) {
					other_atom = -other_atom - 1;
				} else {
					other_atom--;
				}
 				dxt = my_data->x_new[my_atom]-packet->coordinates.x[other_atom];
				dyt = my_data->y_new[my_atom]-packet->coordinates.y[other_atom];
				dzt = my_data->z_new[my_atom]-packet->coordinates.z[other_atom];
				d=dxt*dxt + dyt*dyt + dzt*dzt;
				if (d<newdist) newdist=d;					
 			} else break;	
 		}
		for (j=posinnerwwbuf;j<cur_size_innerwwbuf;j++) {
 			if (other_unitnum==unitnumbuf[j]) {
				if ((other_atom=innerwwbuf[j])<0) {
					other_atom = -other_atom - 1;
				} else {
					other_atom--;
				}
 				dxt = my_data->x_new[my_atom]-packet->coordinates.x[other_atom];
				dyt = my_data->y_new[my_atom]-packet->coordinates.y[other_atom];
				dzt = my_data->z_new[my_atom]-packet->coordinates.z[other_atom];
				d=dxt*dxt + dyt*dyt + dzt*dzt;
				if (d<newdist) newdist=d;					
 			} else break;	
 		}
 	} else 
 	if (flag==1) {
 		if (other_unitnum!=my_unitnum) {
 			errout("Fatal error #2 in get_min_unitdist %d %d\n",
				other_unitnum,my_unitnum);
			pex(-1);		
 			
 		}
 		for (other_atom=first_atom;other_atom<last_atom;other_atom++) {	
 			dxt = packet->coordinates.x[my_atom] - my_data->x_new[other_atom];
			dyt = packet->coordinates.y[my_atom] - my_data->y_new[other_atom];
			dzt = packet->coordinates.z[my_atom] - my_data->z_new[other_atom];	
			d=dxt*dxt + dyt*dyt + dzt*dzt;
			if (d<newdist) newdist=d;
 		}
 	}
 	
	return (double) sqrt(newdist);	
}


/**************************************************************************/
/*** here the forces comming from the local expansion are added to	***/
/*** the force-buffer							***/
/**************************************************************************/

int TREE_add_distforces(	int my_number,
				LOAD_MSG_ALL_NODE *global_data,
				LOAD_MSG_ONE_NODE *my_data)
{
	int	unit=my_number; /*** otherwise PARIX warnings ***/
	int	nr_units = my_data->cl_per_level[0];
	int	ind_first= my_data->cluster[0][0].child_pos;
	int	sta_atom,end_atom,my_atom=global_data->no_nodes;/*** otherwise PARIX warnings ***/
	double	dx,dy,dz,Q;
	double	e,fx,fy,fz;
	struct_cluster *my_unit;

#if DO_CUTOFF
	return 0;
#endif
	for (unit=0;unit<nr_units;unit++) {
		my_unit = &(my_data->cluster[0][unit]);

		if (my_unit->typ==NEUTRAL) continue;

		/************** for every unit which is not neutral *********/
		sta_atom = my_unit->child_pos - ind_first;
		end_atom = my_unit->nr_childs + sta_atom;

		for (my_atom=sta_atom;my_atom<end_atom;my_atom++) {
			int	ii=4*my_atom;
			int	ii1=ii+1,ii2=ii+2,ii3=ii+3;

			if ((Q=my_data->q[my_atom])==0.0) continue;
			
			/*** for every atom in that unit, which is charged ***/
			dx = my_data->x_new[my_atom] - my_unit->cen_refp[0];
			dy = my_data->y_new[my_atom] - my_unit->cen_refp[1];
			dz = my_data->z_new[my_atom] - my_unit->cen_refp[2];
		
			e  = 0.5*Q*(	my_unit->local[0] +
				     dx*my_unit->local[1] +
				     dy*my_unit->local[2] +
				     dz*my_unit->local[3] +				 
				 0.5*dx*my_unit->local[4]*dx +
				 0.5*dy*my_unit->local[5]*dy +
				 0.5*dz*my_unit->local[6]*dz +
				     dx*my_unit->local[7]*dy +
				     dx*my_unit->local[8]*dz +
				     dy*my_unit->local[9]*dz);

			fx = Q*(   my_unit->local[1] +
				dx*my_unit->local[4] + 
				dy*my_unit->local[7] +
				dz*my_unit->local[8]);

			fy = Q*(   my_unit->local[2]+
				dx*my_unit->local[7]+
				dy*my_unit->local[5]+
				dz*my_unit->local[9]);

			fz = Q*(   my_unit->local[3]+
				dx*my_unit->local[8]+
				dy*my_unit->local[9]+
				dz*my_unit->local[6]);

			my_data->force_elec[ii ] += e;
			my_data->force_elec[ii1] += fx;
			my_data->force_elec[ii2] += fy;
			my_data->force_elec[ii3] += fz;

			my_data->force_elecvdw[ii ] += e;
			my_data->force_elecvdw[ii1] += fx;
			my_data->force_elecvdw[ii2] += fy;
			my_data->force_elecvdw[ii3] += fz;
		}
	}

	return 0;
}

/*** here we inherit the local-expansion of structural units to ***/
/*** to the QM-atoms.						***/
int TREE_calc_qmlocal(	int my_number,
			LOAD_MSG_ALL_NODE *global_data,
			LOAD_MSG_ONE_NODE *my_data)
{
	int	unit=my_number; /*** otherwise PARIX warnings ***/
	int	nr_units = my_data->cl_per_level[0];
	int	ind_first= my_data->cluster[0][0].child_pos;
	int	sta_atom,end_atom,my_atom=global_data->no_nodes;/*** otherwise PARIX warnings ***/
	double	dx,dy,dz,rhx,rhy,rhz;
	struct_cluster *my_unit;
#if DEBUG
errout("Node %d: Now in TREE_calc_qmlocal %d %d\n",
my_number,my_data->qmlocal.len,my_data->qmlocal.cur_len);
#endif
	for (unit=0;unit<nr_units;unit++) {
		my_unit  = &(my_data->cluster[0][unit]);
		sta_atom = my_unit->child_pos - ind_first;
		end_atom = my_unit->nr_childs + sta_atom;

		for (my_atom=sta_atom;my_atom<end_atom;my_atom++) {
			int	local_index;
			double	*mylocal;

			if (!(my_data->atom_sele_flag1[my_atom]&FLAG_QMATOM)) continue;

			local_index = get_qmnode_index(my_data->atom_id[my_atom],
							   my_data);	
			mylocal = my_data->qmlocal.local + 10*local_index;

			dx = my_data->x_new[my_atom] - my_unit->cen_refp[0];
			dy = my_data->y_new[my_atom] - my_unit->cen_refp[1];
			dz = my_data->z_new[my_atom] - my_unit->cen_refp[2];

			rhx =	dx*my_unit->local[4] + 
				dy*my_unit->local[7] + 
				dz*my_unit->local[8];

			rhy =	dx*my_unit->local[7] + 
				dy*my_unit->local[5] + 
				dz*my_unit->local[9];

			rhz =	dx*my_unit->local[8] + 
				dy*my_unit->local[9] + 
				dz*my_unit->local[6];

			mylocal[0] += 	 my_unit->local[0] + 
					(my_unit->local[1] + 0.5*rhx)*dx+
					(my_unit->local[2] + 0.5*rhy)*dy+
					(my_unit->local[3] + 0.5*rhz)*dz;

			mylocal[1] += (my_unit->local[1] + rhx);
			mylocal[2] += (my_unit->local[2] + rhy);
			mylocal[3] += (my_unit->local[3] + rhz);

			mylocal[4] += my_unit->local[4];
			mylocal[5] += my_unit->local[5];
			mylocal[6] += my_unit->local[6];

			mylocal[7] += my_unit->local[7];
			mylocal[8] += my_unit->local[8];
			mylocal[9] += my_unit->local[9];
		}
	}
#if DEBUG
errout("Node %d: Now exit TREE_calc_qmlocal %d %d\n",
my_number,my_data->qmlocal.len,my_data->qmlocal.cur_len);
#endif
	return 0;
}
	
/***************************************************************************/

static int
TREE_force_tree2tree(	int my_number,
			int my_cluster,
			int level,
			LOAD_MSG_ALL_NODE *global_data,
			LOAD_MSG_ONE_NODE *my_data,
			TRAVELLING_PACKETS *packet)
{
	int	other_number= packet->branch.node;
	int	left_branch = other_number * global_data->treeinfo.branches/global_data->no_nodes;
	int	nr_clusters = packet->branch.cl_per_level[level];
	struct_cluster *my_cl,*other_cl;
	double	x,y,z,limit,dx,dy,dz,dist;
	int	left_cl=0,ch_level=level-1,child;
	int	cluster;
	int	ch_ptr    = next_wwbuf_ptr;
	int	sta_wwbuf = cur_wwbuf_ptr;

	/******************************************************************/
	/*** this parent (my_cluster) has no partners in his ww-list,	***/
	/*** so the childs inherit it 					***/
	/******************************************************************/
	my_cl = &(my_data->cluster[level][my_cluster]);

	if ((cluster=antiwwbuf[level][sta_wwbuf])<=-1)  {

		antiwwbuf[level][sta_wwbuf]++;/*** "decrement" how often used ! ***/

		if (cluster==-1) {
			cur_wwbuf_ptr++;	/*** point to next ww-element ***/
		}

		if (level==0) {	/*** mark end of ww-list for this unit ***/
			put_to_unitwwlist(other_number,-1);
			return 0;
		}
	
		goto inherit_now;
	}

	if (level>0) {
		left_cl = global_data->levelinfo[ch_level].first_cluster[left_branch];
	}

	x = my_cl->cen_refp[0];
	y = my_cl->cen_refp[1];
	z = my_cl->cen_refp[2];
	limit = global_data->outer_classrad[level]+my_cl->cen_geom[3];

	/****************************************************************************/
	/*** for every cluster, which is in in ww-list of the parent (my_cluster) ***/
	/*** we have to check the distance 					  ***/
	/****************************************************************************/

	while ((cluster=antiwwbuf[level][sta_wwbuf])>=0) {
#if IFCHECKBUFSIZE
/***checks***/	if (cluster>=nr_clusters) {
/***checks***/		errout("Node %d: Fatal error: Illegal cluster index %d  %d lev %d stww %d othnum %d\n",
/***checks***/			my_number,cluster,nr_clusters,level,sta_wwbuf,other_number);
/***checks***/		pex(-1);
/***checks***/	}
/***checks***/	if (last_wwbuf_ptr<sta_wwbuf) {
/***checks***/		errout("Node %d: Fatal error: bad things happen %d %d\n",
/***checks***/			my_number,sta_wwbuf,last_wwbuf_ptr);
/***checks***/	}
#endif

		other_cl=&(packet->branch.cluster[level][cluster]);
		dx = x - other_cl->cen_refp[0];
		dy = y - other_cl->cen_refp[1];
		dz = z - other_cl->cen_refp[2];

		dist = sqrt(dx*dx + dy*dy + dz*dz) - other_cl->cen_geom[3];

		if (dist>limit) {
			/**********************************************************/
			/*** this cluster is far enough away, so we can calculate**/
			/*** the contribution to the local expansion of our     ***/
			/**** current cluster (my_cluster) 			***/
			/**********************************************************/

			put_to_wwbuf(level,other_number,cluster);

		} else {
			/**********************************************************/
			/*** this cluster is NOT far enough away, so we have to	***/
			/*** add all child of this cluster to the possible 	***/
			/*** ww-list of the childs of my_cluster		***/
			/**********************************************************/

			if (level==0) {
				/*** if we are at the unit-level, we make a list***/
				/*** of the units which belong to the innermost ***/
				/*** interaction shell				***/
				
				put_to_unitwwlist(other_number,cluster);

			} else {
				/*** here we build up the possible ww-list for 	***/
				/*** one child of that parent later we mark the ***/
				/*** end with the negativ number of childs 	***/

				int sta_child = other_cl->child_pos - left_cl;
				int end_child = other_cl->nr_childs + sta_child;
				for (child=sta_child;child<end_child;child++) {

#if IFCHECKBUFSIZE
/***checks***/				if (ch_ptr+1>=max_size_antiwwbuf[ch_level]) {
/***checks***/						errout("Node %d: Fatal error: buf to small #33 at level %d (%d %d)\n",
/***checks***/						my_number,ch_level,ch_ptr,
/***checks***/						max_size_antiwwbuf[ch_level]);
/***checks***/					pex(-1);
/***checks***/				}
/***checks***/				if (child>=packet->branch.cl_per_level[ch_level]) {
/***checks***/					errout("Node %d: Fatal error: Inval entry %d %d lev %d chp %d nr %d\n",
/***checks***/						my_number,
/***checks***/						child,packet->branch.cl_per_level[ch_level],
/***checks***/						level,ch_ptr,other_cl->nr_childs);
/***checks***/					pex(-1);
/***checks***/				}
#endif

					antiwwbuf[ch_level][ch_ptr++] = child;
				}
			}
		}
		sta_wwbuf++;		/*** point to next ww-element ***/
	}

	antiwwbuf[level][sta_wwbuf]++;	/*** "decrement" how often used ! ****/

	if (cluster==-1) {
		cur_wwbuf_ptr=sta_wwbuf+1;		/*** point to next ww-element	***/
	}

	if (level==0) {
		put_to_unitwwlist(other_number,-1);	/*** mark end of ww-list for this unit ***/
		return 0;
	}

	/**************************************************************/
inherit_now:
#if IFCHECKBUFSIZE
/***checks***/	if (ch_ptr+1>=max_size_antiwwbuf[ch_level]) {
/***checks***/		errout("Node %d: Fatal error: buffer to small #4 at level %d (%d %d)\n",
/***checks***/			my_number,ch_level,ch_ptr,max_size_antiwwbuf[ch_level]);
/***checks***/		pex(-1);
/***checks***/	}
/***checks***/	if (my_cl->nr_childs<=0) {
/***checks***/		errout("Node %d: Fatal error: nr_childs<=0 %d!! lev %d mycl %d\n",
/***checks***/		my_number,my_cl->nr_childs,level,my_cluster);
/***checks***/		pex(-1);
/***checks***/	}
#endif

	/*** mark the end of the possible ww-list ***/
	/*** and code how often to use this part  ***/
	antiwwbuf[ch_level][ch_ptr++] = -my_cl->nr_childs;
	next_wwbuf_ptr = ch_ptr;
	
	return 0;
}


static int
put_to_unitwwlist(int other_number,int elem)
{
	if (cur_unitww_ptr>=max_size_unitwwlist[other_number]) {
#if TRACE_REALLOC
txtout("Node ?: realloc for unitwwlist[%d] new_size %d\n",
other_number,max_size_unitwwlist[other_number]);
#endif
		set_allocinfo("put_to_unitwwlist");
		unitwwlist[other_number] = (int*) merealloc(unitwwlist[other_number],
						sizeof(int)*(max_size_unitwwlist[other_number]+REALLOC_SIZE),
						sizeof(int)* max_size_unitwwlist[other_number]);
		max_size_unitwwlist[other_number]+=REALLOC_SIZE;
		set_allocinfo("XXX");
	}	

	unitwwlist[other_number][cur_unitww_ptr++] = elem;
	return 0;
}

static int
put_to_wwbuf(int level,int other_number,int elem)
{

	if (interact_ptr>=max_size_wwbuf[level][other_number]) {
#if TRACE_REALLOC
txtout("Node ?: realloc for wwbuf[%d][%d] new_size %d\n",
level,other_number,max_size_wwbuf[level][other_number]);
#endif
		set_allocinfo("put_to_wwbuf");
		wwbuf[level][other_number] = (int*) merealloc(wwbuf[level][other_number],
						sizeof(int)*(max_size_wwbuf[level][other_number]+REALLOC_SIZE),
						sizeof(int)* max_size_wwbuf[level][other_number]);
		max_size_wwbuf[level][other_number]+=REALLOC_SIZE;
		set_allocinfo("XXX");
	}	

	wwbuf[level][other_number][interact_ptr++] = elem;
	return 0;
}

/**************************************************************************/
/*** here the local expansion is inherited to the childs 		***/
/**************************************************************************/

static
int	inherit_local_expansion(	int my_cluster,
					int level,
					LOAD_MSG_ONE_NODE *my_data)
{
	int	sta_child,end_child;
	int	ch_level=level-1,child;
	struct_cluster *childcl;
	struct_cluster *parentcl;
	double	dx,dy,dz,rhx,rhy,rhz,xparent,yparent,zparent;

	parentcl  = &(my_data->cluster[level][my_cluster]);

	sta_child = parentcl->child_pos - my_data->cluster[level][0].child_pos;

	end_child = parentcl->nr_childs + sta_child;

	xparent = parentcl->cen_refp[0];
	yparent = parentcl->cen_refp[1];
	zparent = parentcl->cen_refp[2];

	for (child=sta_child;child<end_child;child++) {
		childcl = &(my_data->cluster[ch_level][child]);

		dx = childcl->cen_refp[0] - xparent;
		dy = childcl->cen_refp[1] - yparent;
		dz = childcl->cen_refp[2] - zparent;

		rhx = dx*parentcl->local[4] + dy*parentcl->local[7] + dz*parentcl->local[8];
		rhy = dx*parentcl->local[7] + dy*parentcl->local[5] + dz*parentcl->local[9];
		rhz = dx*parentcl->local[8] + dy*parentcl->local[9] + dz*parentcl->local[6];

		childcl->local[0] += parentcl->local[0] + 
					(parentcl->local[1]+0.5*rhx)*dx+
					(parentcl->local[2]+0.5*rhy)*dy+
					(parentcl->local[3]+0.5*rhz)*dz;

		childcl->local[1] += (parentcl->local[1] + rhx);
		childcl->local[2] += (parentcl->local[2] + rhy);
		childcl->local[3] += (parentcl->local[3] + rhz);

		childcl->local[4] += parentcl->local[4];
		childcl->local[5] += parentcl->local[5];
		childcl->local[6] += parentcl->local[6];

		childcl->local[7] += parentcl->local[7];
		childcl->local[8] += parentcl->local[8];
		childcl->local[9] += parentcl->local[9];
	}
	return 0;
}


/**************************************************************************/
/*** here we use the ww-list for the nearest atoms and do extrapolate 	***/
/*** the forces of atoms which did not belong to this list. 		***/
/**************************************************************************/

static int
do_nearest_extrapol(	int my_number,
			int my_atom,
			LOAD_MSG_ALL_NODE *global_data,
			LOAD_MSG_ONE_NODE *my_data,
			TRAVELLING_PACKETS *packet)
{
	double	*xother,x = my_data->x_new[my_atom];
	double	*yother,y = my_data->y_new[my_atom];
	double	*zother,z = my_data->z_new[my_atom];
	double	*qother,Q0= 0.5 * my_data->q[my_atom] * global_data->f_elstat;
	double	*force_elec,*force_elecvdw;
	double	Q0_14= Q0 * global_data->elec14;
	double	(*cur_AB)[MXNO_ATOM_TYPES];

	int	other_atom,num_types= global_data->nr_of_atom_typs;
	int	*typother,typ1      = my_data->typ[my_atom];
	int 	my_forceout_atom    = my_data->atom_sele_flag1[my_atom];
	int	copy_flag = (my_number==packet->branch.node);
	int	ii  = 4*my_atom;
	int	ii1 = ii+1,ii2=ii+2,ii3=ii+3;
	int	*other_forceout_atoms;
	
	xother  = packet->coordinates.x;
	yother  = packet->coordinates.y;
	zother  = packet->coordinates.z;
	qother  = packet->coordinates.q;
	typother= packet->coordinates.typ;
	force_elec    = my_data->force_elec;
	force_elecvdw = my_data->force_elecvdw;
	other_forceout_atoms = packet->coordinates.atom_sele_flag1;
	
	while ((other_atom=nearestwwbuf[nearest_ptr++])!=0) {
		double 	q2,Q,A,B;
		double 	rx,ry,rz,r_2,r_6,felecvdw=0.0,vdw_loc_energy=0.0;
		double	fx_h=0.0,fy_h=0.0,fz_h=0.0;
		double	fx_h_elec=0.0,fy_h_elec=0.0,fz_h_elec=0.0;
		double  elecstat_loc_energy=0.0,felec=0.0;
		int 	typ2;
	
		if (other_atom<0) {
			other_atom = -other_atom - 1;
			cur_AB = global_data->AB14;
			Q = Q0_14;
		} else {
			other_atom--;
			cur_AB = global_data->AB;
			Q = Q0;
		}
		
		typ2 = typother[other_atom];
		if (typ1 < typ2) {
			A = cur_AB[typ1][typ2];
			B = cur_AB[typ2][typ1];
		} else
		if (typ1 > typ2) {
			A = cur_AB[typ2][typ1];
			B = cur_AB[typ1][typ2];
		} else {
			A = cur_AB[typ1][num_types];
			B = cur_AB[num_types][typ1];
		}
		
		rx = x - xother[other_atom];
		ry = y - yother[other_atom];
		rz = z - zother[other_atom];
		r_2= 1/(rx*rx + ry*ry + rz*rz);
		
#if FREEENERGY
		/*** Dont calculate any forces between the two parts ***/
		if ((other_forceout_atoms[other_atom]&FLAG_PART1)&&
		    (   my_forceout_atom             &FLAG_PART2)) {
			goto energy_goon_here1;
		}
		if ((other_forceout_atoms[other_atom]&FLAG_PART2)&&
		    (   my_forceout_atom             &FLAG_PART1)) {
			goto energy_goon_here1;
		}

		/*** otherwise scale interaction between parts and the rest ***/
		if (my_forceout_atom&FLAG_PART1) {
			Q *= LAMBDA;
			A *= LAMBDA;
			B *= LAMBDA;
		} else
		if (my_forceout_atom&FLAG_PART2) {
			Q *= (1.0-LAMBDA);
			A *= (1.0-LAMBDA);
			B *= (1.0-LAMBDA);
		}
#endif

		/**********************************/
		/*** electrostatic energy 	***/
		/**********************************/
		q2 = qother[other_atom];
	    	if ((Q!=0.0)&&(q2!=0.0)) {
			elecstat_loc_energy = Q * q2 * sqrt(r_2);
			felec     = 2 * r_2 * elecstat_loc_energy;
	    		fx_h_elec = felec * rx;
	    		fy_h_elec = felec * ry;
	    		fz_h_elec = felec * rz;

			force_elec[ii ] += elecstat_loc_energy;
			force_elec[ii1] -= fx_h_elec;
			force_elec[ii2] -= fy_h_elec;
			force_elec[ii3] -= fz_h_elec;
		
			if (copy_flag) {
				int jj=4*other_atom;
				force_elec[jj++] += elecstat_loc_energy;
				force_elec[jj++] += fx_h_elec;
				force_elec[jj++] += fy_h_elec;
				force_elec[jj  ] += fz_h_elec;
			}
		}

		/**********************************/
		/*** add VDW energy 		***/
		/**********************************/
	    	r_6            = r_2*r_2*r_2;
		vdw_loc_energy = r_6 * (r_6 * A - B) / 2;

		felecvdw       = felec + r_2 * (24*vdw_loc_energy + 6*r_6*B);
	    	fx_h = felecvdw * rx;
	    	fy_h = felecvdw * ry;
	    	fz_h = felecvdw * rz;

		force_elecvdw[ii ] += elecstat_loc_energy+vdw_loc_energy;
		force_elecvdw[ii1] -= fx_h;
		force_elecvdw[ii2] -= fy_h;
		force_elecvdw[ii3] -= fz_h;
		
		if (copy_flag) {
			int jj=4*other_atom;
			force_elecvdw[jj++] += elecstat_loc_energy+vdw_loc_energy;
			force_elecvdw[jj++] += fx_h;
			force_elecvdw[jj++] += fy_h;
			force_elecvdw[jj  ] += fz_h;
		}
#if FREEENERGY
energy_goon_here1:
#endif
		if (!my_forceout_atom)	continue;

		/*****************************************************/
		/*** Here we eventually sort the calculated forces ***/
		/*** to the PICK_ELEC buffer which is sent to the  ***/
		/*** master and can be saved to disk		   ***/	
		/*****************************************************/
		if ((my_forceout_atom                &PICK_ELEC)&&
		    (other_forceout_atoms[other_atom]&PICK_ELEC)){
		    	double eforce[3]; int myatom_id[2];
		    	myatom_id[0] = packet->coordinates.atom_id[other_atom];
		    	myatom_id[1] = my_data->atom_id[my_atom];		    	
		    	eforce[0] = fx_h_elec;
		    	eforce[1] = fy_h_elec;
		    	eforce[2] = fz_h_elec;		    	
		    
			sort_to_pickbond(myatom_id,
					elecstat_loc_energy,
					eforce,
					sqrt(1/r_2),
					&(my_data->pickelec));				
		}

		/*****************************************************/
		/*** Here we eventually sort the calculated forces ***/
		/*** to the PICK_VDW buffer which is sent to the   ***/
		/*** master and can be saved to disk		   ***/	
		/*****************************************************/		
		if ((my_forceout_atom                &PICK_VDW)&&
		    (other_forceout_atoms[other_atom]&PICK_VDW)){
		   	double eforce[3]; int myatom_id[2];
		    	myatom_id[0] = packet->coordinates.atom_id[other_atom];
		    	myatom_id[1] = my_data->atom_id[my_atom];		    	
		    	eforce[0] = fx_h-fx_h_elec;
		    	eforce[1] = fy_h-fy_h_elec;
		    	eforce[2] = fz_h-fz_h_elec;
		    
			sort_to_pickbond(myatom_id,
					vdw_loc_energy,
					eforce,
					sqrt(1/r_2),
					&(my_data->pickvdw));				
		}
	}

	/**** extrapolate forces of the outer schell with DC-1d *****/
	if (packet_counter==global_data->no_nodes) {
		force_elecvdw[ii ] += extrapolate_dc1d(1,1,
						new_E_elecvdw[my_atom],
						old_E_elecvdw[my_atom]);
		force_elecvdw[ii1] += extrapolate_dc1d(1,1,
						new_fx_elecvdw[my_atom],
						old_fx_elecvdw[my_atom]);
		force_elecvdw[ii2] += extrapolate_dc1d(1,1,
						new_fy_elecvdw[my_atom],
						old_fy_elecvdw[my_atom]);
		force_elecvdw[ii3] += extrapolate_dc1d(1,1,
						new_fz_elecvdw[my_atom],
						old_fz_elecvdw[my_atom]);

		force_elec[ii ] += extrapolate_dc1d(1,1,
						new_E_elec[my_atom],
						old_E_elec[my_atom]);
		force_elec[ii1] += extrapolate_dc1d(1,1,
						new_fx_elec[my_atom],
						old_fx_elec[my_atom]);
		force_elec[ii2] += extrapolate_dc1d(1,1,
						new_fy_elec[my_atom],
						old_fy_elec[my_atom]);
		force_elec[ii3] += extrapolate_dc1d(1,1,
						new_fz_elec[my_atom],
						old_fz_elec[my_atom]);
	}
	
	return 0;
}

/*******************************************************************************/
/*** here the exact electrostatic forces and energies for the innermost atoms***/
/*** are calculated if partners are on the same node, we can copy the forces ***/
/*** for the partner (Newtons Law) forces not belonging to nearest ww-range, ***/
/*** will be also scanned to the new force-buffer which is needed for 	     ***/
/*** extrapolation. my_atom is the selected atom which interacts with all    ***/
/*** atoms listed in innerwwbuf!					     ***/
/*******************************************************************************/

static int
calc_exact_interact(	int my_number,int my_atom,
			LOAD_MSG_ALL_NODE *global_data,
			LOAD_MSG_ONE_NODE *my_data,
			TRAVELLING_PACKETS *packet,
			int mode)
{
	double	*xother,x = my_data->x_new[my_atom];
	double	*yother,y = my_data->y_new[my_atom];
	double	*zother,z = my_data->z_new[my_atom];
	double	*qother,Q0= 0.5 * my_data->q[my_atom] * global_data->f_elstat;
	double	*force_elec,*force_elecvdw;
	double	Q0_14= Q0 * global_data->elec14;
	double	(*cur_AB)[MXNO_ATOM_TYPES];

	int	other_atom,num_types=global_data->nr_of_atom_typs;
	int	*typother,typ1      = my_data->typ[my_atom];
	int 	my_forceout_atom    = my_data->atom_sele_flag1[my_atom];
	int	copy_flag = (my_number==packet->branch.node);
	int	ii  = 4*my_atom;
	int	ii1 = ii+1,ii2=ii+2,ii3=ii+3;
	int	*other_forceout_atoms;
	int	mapped_other_atom,res=0;
	
	xother  = packet->coordinates.x;
	yother  = packet->coordinates.y;
	zother  = packet->coordinates.z;
	qother  = packet->coordinates.q;
	typother= packet->coordinates.typ;
	force_elec    = my_data->force_elec;
	force_elecvdw = my_data->force_elecvdw;
	other_forceout_atoms = packet->coordinates.atom_sele_flag1;

	for (posinnerwwbuf=0;posinnerwwbuf<cur_size_innerwwbuf;posinnerwwbuf++) {
		int 	typ2,ex14_flag=0;
		double	q2,Q,A,B;
		double	rx,ry,rz,r_2,r_6,felecvdw=0.0,vdw_loc_energy=0.0;
		double	fx_h=0.0,fy_h=0.0,fz_h=0.0;
		double	fx_h_elec=0.0,fy_h_elec=0.0,fz_h_elec=0.0;
		double  elecstat_loc_energy=0.0,felec=0.0;	

		if ((other_atom=innerwwbuf[posinnerwwbuf])<0) {
			other_atom = -other_atom - 1;
			ex14_flag  = 1;
			cur_AB     = global_data->AB14;
			Q          = Q0_14;
		} else {
			other_atom--;
			ex14_flag  = 0;
			cur_AB     = global_data->AB;
			Q          = Q0;
		}

		typ2 = typother[other_atom];
		if (typ1 < typ2) {
			A = cur_AB[typ1][typ2];
			B = cur_AB[typ2][typ1];
		} else
		if (typ1 > typ2) {
			A = cur_AB[typ2][typ1];
			B = cur_AB[typ1][typ2];
		} else {
			A = cur_AB[typ1][num_types];
			B = cur_AB[num_types][typ1];
		}
#if DEBUG
txtout("Node %d: my %d other %d A %f B %f Q %f\n",
my_number,my_atom,other_atom,A,B,Q);
#endif	
		rx = x - xother[other_atom];
		ry = y - yother[other_atom];
		rz = z - zother[other_atom];
		r_2= 1/(rx*rx + ry*ry + rz*rz);

#if FREEENERGY
		/*** Don't calculate any forces between the two parts ***/
		if ((other_forceout_atoms[other_atom]&FLAG_PART1)&&
		    (   my_forceout_atom             &FLAG_PART2)) {
			goto energy_goon_here2;
		}
		if ((other_forceout_atoms[other_atom]&FLAG_PART2)&&
		    (   my_forceout_atom             &FLAG_PART1)) {
			goto energy_goon_here2;
		}

		/*** otherwise scale interaction between parts and the rest ***/
		if (my_forceout_atom&FLAG_PART1) {
			Q *= LAMBDA;
			A *= LAMBDA;
			B *= LAMBDA;
		} else
		if (my_forceout_atom&FLAG_PART2) {
			Q *= (1.0-LAMBDA);
			A *= (1.0-LAMBDA);
			B *= (1.0-LAMBDA);
		}
#endif

		/**************************************************/
		/*** Manage the QM interface!!!			***/
		/**************************************************/
 		if (global_data->nr_qmatoms>0) {
 			res = calculate_qmelec(	my_number,my_atom,my_unitnum,
 						other_atom,unitnumbuf[posinnerwwbuf],
 						copy_flag,
 						global_data,
 						my_data,packet);
/*** 			
 			if (res==SKIP_ELECVDW) goto energy_goon_here2;
***/
 		}

		/****************************/
		/*** electrostatic energy ***/	
		/****************************/
		q2 = qother[other_atom];
	    	if ((Q!=0.0)&&(q2!=0.0)) {
			elecstat_loc_energy = Q * q2 * sqrt(r_2);
			felec     = 2 * r_2 * elecstat_loc_energy;
	    		fx_h_elec = felec * rx;
	    		fy_h_elec = felec * ry;
	    		fz_h_elec = felec * rz;

			force_elec[ii ] += elecstat_loc_energy;
			force_elec[ii1] -= fx_h_elec;
			force_elec[ii2] -= fy_h_elec;
			force_elec[ii3] -= fz_h_elec;

			if (copy_flag) {
				int jj=4*other_atom;
				force_elec[jj++] += elecstat_loc_energy;
				force_elec[jj++] += fx_h_elec;
				force_elec[jj++] += fy_h_elec;
				force_elec[jj  ] += fz_h_elec;
			}
		}
#if DEBUG
txtout("Node %d: q2 %f\n",
my_number,q2);
#endif
		/**********************************/
		/*** electrostatic + VDW energy ***/
		/**********************************/
		if (res==SKIP_ELECVDW) {
			force_elec[ii ]    -= elecstat_loc_energy;
			force_elecvdw[ii ] -= elecstat_loc_energy;
			if (copy_flag) {
				force_elec[4*other_atom]   -= elecstat_loc_energy;
				force_elecvdw[4*other_atom]-= elecstat_loc_energy;			
			}
			felecvdw       = felec;
			vdw_loc_energy = 0.0;
		} else {
	    		r_6 = r_2 * r_2 * r_2;
			vdw_loc_energy = r_6 * (r_6*A - B) / 2;
			felecvdw       = felec + r_2 * (24*vdw_loc_energy + 6*r_6*B);
		}
	    	fx_h = felecvdw * rx;
	    	fy_h = felecvdw * ry;
	    	fz_h = felecvdw * rz;

		force_elecvdw[ii ] += elecstat_loc_energy+vdw_loc_energy;
		force_elecvdw[ii1] -= fx_h;
		force_elecvdw[ii2] -= fy_h;
		force_elecvdw[ii3] -= fz_h;
	
		if (copy_flag) {
			int jj=4*other_atom;
			force_elecvdw[jj++] += elecstat_loc_energy+vdw_loc_energy;
			force_elecvdw[jj++] += fx_h;
			force_elecvdw[jj++] += fy_h;
			force_elecvdw[jj  ] += fz_h;
		}

energy_goon_here2:

		/************************************************************/
		/*** if we have to make a new ww-list we note this list	or **/
		/*** otherwise we use the old list. In any case forces are **/
		/*** scanned, if needed to the new force buffer for        **/
		/*** extrapolation.					   **/	
		/************************************************************/
#if DEBUG
txtout("Node %d: here %lx %lx\n",
my_number,my_forceout_atom,other_forceout_atoms[other_atom]);
#endif
		mapped_other_atom  = (ex14_flag) ? -other_atom-1: other_atom+1;
		if ((mode & (int)0x01)==1) {	/*** make new ww-list ***/
			if (r_2<nearest_radius2) { /*** r_2 is inverse! ***/
				/*** my_atom and other_atom are in the ***/
				/*** second interaction shell. For these**/
				/*** we calculate the sum of elec and  ***/
				/*** vdw forces. We need it later for  ***/
				/*** for extrapolation!		       ***/
				new_E_elecvdw [my_atom] += elecstat_loc_energy+vdw_loc_energy;				
				new_fx_elecvdw[my_atom] -= fx_h;
				new_fy_elecvdw[my_atom] -= fy_h;
				new_fz_elecvdw[my_atom] -= fz_h;

				new_E_elec [my_atom] += elecstat_loc_energy;
				new_fx_elec[my_atom] -= fx_h_elec;
				new_fy_elec[my_atom] -= fy_h_elec;
				new_fz_elec[my_atom] -= fz_h_elec;

				if (copy_flag) {
					new_E_elecvdw [other_atom] += elecstat_loc_energy+vdw_loc_energy;
					new_fx_elecvdw[other_atom] += fx_h;
					new_fy_elecvdw[other_atom] += fy_h;
					new_fz_elecvdw[other_atom] += fz_h;

					new_E_elec [other_atom] += elecstat_loc_energy;
					new_fx_elec[other_atom] += fx_h_elec;
					new_fy_elec[other_atom] += fy_h_elec;
					new_fz_elec[other_atom] += fz_h_elec;
				}
			} else {
				/*** if atoms in the innermost shell 	***/
				/*** we store it here->Interaction list ***/
				nearestwwbuf[nearest_ptr++] = mapped_other_atom;
			}
		} else { 	/*** use old ww-list ***/
			if (nearestwwbuf[nearest_ptr]!=mapped_other_atom) {
				new_E_elecvdw [my_atom] += elecstat_loc_energy+vdw_loc_energy;
				new_fx_elecvdw[my_atom] -= fx_h;
				new_fy_elecvdw[my_atom] -= fy_h;
				new_fz_elecvdw[my_atom] -= fz_h;

				new_E_elec [my_atom] += elecstat_loc_energy;
				new_fx_elec[my_atom] -= fx_h_elec;
				new_fy_elec[my_atom] -= fy_h_elec;
				new_fz_elec[my_atom] -= fz_h_elec;

				if (copy_flag) {
					new_E_elecvdw [other_atom] += elecstat_loc_energy+vdw_loc_energy;
					new_fx_elecvdw[other_atom] += fx_h;
					new_fy_elecvdw[other_atom] += fy_h;
					new_fz_elecvdw[other_atom] += fz_h;

					new_E_elec [other_atom] += elecstat_loc_energy;
					new_fx_elec[other_atom] += fx_h_elec;
					new_fy_elec[other_atom] += fy_h_elec;
					new_fz_elec[other_atom] += fz_h_elec;
				}
			} else {
				nearest_ptr++;
			}
		}
		
		if (!my_forceout_atom) continue;
#if DEBUG
txtout("Node %d: there \n",
my_number);
#endif		
		/*****************************************************/
		/*** Here we eventually sort the calculated forces ***/
		/*** to the PICK_ELEC buffer which is sent to the  ***/
		/*** master and can be saved to disk		   ***/	
		/*****************************************************/
		if ((   my_forceout_atom             &PICK_ELEC)&&
		    (other_forceout_atoms[other_atom]&PICK_ELEC)){
		    	double eforce[3]; int myatom_id[2];
		    	myatom_id[0] = packet->coordinates.atom_id[other_atom];
		    	myatom_id[1] = my_data->atom_id[my_atom];		    	
		    	eforce[0] = fx_h_elec;
		    	eforce[1] = fy_h_elec;
		    	eforce[2] = fz_h_elec;		    	
		    
			sort_to_pickbond(myatom_id,
					elecstat_loc_energy,
					eforce,
					sqrt(1/r_2),
					&(my_data->pickelec));				
		}

		/*****************************************************/
		/*** Here we eventually sort the calculated forces ***/
		/*** to the PICK_VDW buffer which is sent to the   ***/
		/*** master and can be saved to disk		   ***/
		/*****************************************************/	
		if ((   my_forceout_atom             &PICK_VDW)&&
		    (other_forceout_atoms[other_atom]&PICK_VDW)){
		    	double eforce[3]; int myatom_id[2];
		    	myatom_id[0] = packet->coordinates.atom_id[other_atom];
		    	myatom_id[1] = my_data->atom_id[my_atom];		    	
		    	eforce[0] = fx_h-fx_h_elec;
		    	eforce[1] = fy_h-fy_h_elec;
		    	eforce[2] = fz_h-fz_h_elec;
		    	
			sort_to_pickbond(myatom_id,
					vdw_loc_energy,
					eforce,
					sqrt(1/r_2),
					&(my_data->pickvdw));				
		}
	}
#if DEBUG
txtout("Node %d: haha \n",
my_number);
#endif
	if (	(global_data->nr_qmatoms>0) ||
		(global_data->schwing_mode!=0) ) {
		
		/*** do nothing in this case ***/

	} else
	if (	(packet_counter==global_data->no_nodes) &&
		(nearest_old_filled==1))  {
		/*** this is due to the DC1d Method ***/

		force_elecvdw[ii1] -= new_fx_elecvdw[my_atom];
		force_elecvdw[ii2] -= new_fy_elecvdw[my_atom];
		force_elecvdw[ii3] -= new_fz_elecvdw[my_atom];

		force_elecvdw[ii1] += extrapolate_dc1d(1,0,
						new_fx_elecvdw[my_atom],
						old_fx_elecvdw[my_atom]);
		force_elecvdw[ii2] += extrapolate_dc1d(1,0,
						new_fy_elecvdw[my_atom],
						old_fy_elecvdw[my_atom]);
		force_elecvdw[ii3] += extrapolate_dc1d(1,0,
						new_fz_elecvdw[my_atom],
						old_fz_elecvdw[my_atom]);


		force_elec[ii1] -= new_fx_elec[my_atom];
		force_elec[ii2] -= new_fy_elec[my_atom];
		force_elec[ii3] -= new_fz_elec[my_atom];

		force_elec[ii1] += extrapolate_dc1d(1,0,
						new_fx_elec[my_atom],
						old_fx_elec[my_atom]);
		force_elec[ii2] += extrapolate_dc1d(1,0,
						new_fy_elec[my_atom],
						old_fy_elec[my_atom]);
		force_elec[ii3] += extrapolate_dc1d(1,0,
						new_fz_elec[my_atom],
						old_fz_elec[my_atom]);
	}

	/*** mark the end for this atom ***/
	if ((mode&(int)0x01)==1) 	nearestwwbuf[nearest_ptr++] = 0;
	else				nearest_ptr++;

	return 0;
}

/**************************************************************************/
/*** here we calculate the local expansion for my_cluster 		***/
/**************************************************************************/

static int
TREE_local_cl2cl(	struct_cluster *my_cluster,
			struct_cluster *other_cluster)
{
	double	dx,dy,dz,r2;
	double	dx2,dy2,dz2;
	int	typ = (int)other_cluster->typ;
/***
	if (my_cluster->typ==NEUTRAL)	return 0;
***/
	if (typ==NEUTRAL)		return 0;
	if ((my_cluster->status==1)&&(other_cluster->status==1)) {
		txtout("INFO: It works\n");
		return 0;
	}
	
	dx  = my_cluster->cen_refp[0] - other_cluster->cen_refp[0];
	dx2 = dx*dx;

	dy  = my_cluster->cen_refp[1] - other_cluster->cen_refp[1];
	dy2 = dy*dy;

	dz  = my_cluster->cen_refp[2] - other_cluster->cen_refp[2];
	dz2 = dz*dz;

	r2 = 1/(dx2 + dy2 + dz2);

	if (typ==POLAR) {
		double	r3 = r2*sqrt(r2);
		double	r5 = r3*r2;
		double	px = other_cluster->moment[1];
		double	py = other_cluster->moment[2];
		double	pz = other_cluster->moment[3];
		double	PdotR     = px*dx + py*dy + pz*dz;
		double	PdotR3    = 3.0*PdotR;
		double	PdotR3_r2 = PdotR3 * r2;

		my_cluster->local[0] += PdotR*r3;

		my_cluster->local[1] += (px - PdotR3_r2*dx)*r3;
		my_cluster->local[2] += (py - PdotR3_r2*dy)*r3;
		my_cluster->local[3] += (pz - PdotR3_r2*dz)*r3;

		PdotR3_r2 *= 5.0;

		my_cluster->local[4] += (dx2*PdotR3_r2 - 6.0*px*dx - PdotR3)*r5;
		my_cluster->local[5] += (dy2*PdotR3_r2 - 6.0*py*dy - PdotR3)*r5;
		my_cluster->local[6] += (dz2*PdotR3_r2 - 6.0*pz*dz - PdotR3)*r5;

		my_cluster->local[7] += (dx*dy*PdotR3_r2 - 3.0*(px*dy + py*dx))*r5;
		my_cluster->local[8] += (dx*dz*PdotR3_r2 - 3.0*(px*dz + pz*dx))*r5;
		my_cluster->local[9] += (dy*dz*PdotR3_r2 - 3.0*(py*dz + pz*dy))*r5;
	} else
	if (typ==CHARGED) {
		double	f  = other_cluster->moment[0] * sqrt(r2);
		double	f1;
	
		my_cluster->local[0] += f;

		f *= r2;		/*** f = Q/(r^3)   ****/

		my_cluster->local[1] -= f * dx;
		my_cluster->local[2] -= f * dy;
		my_cluster->local[3] -= f * dz;

		f1 = 3.0 * f * r2;	/*** f1= 3 Q/(r^5) ****/ 
		my_cluster->local[4] += dx2*f1 - f;
		my_cluster->local[5] += dy2*f1 - f;
		my_cluster->local[6] += dz2*f1 - f;

		my_cluster->local[7] += dx*dy*f1;
		my_cluster->local[8] += dx*dz*f1;
		my_cluster->local[9] += dy*dz*f1;
	}
	return 0;
}
/**********************************************************************/


