      PROGRAM SETW
C
C     Written by M. Schaefer & H. Heller; 1991.
C
C     Sets water molecules in and around a molecule read in from a pdb file;
C     water molecules are positioned on a cubic lattice wherever it is
C     allowed by considering the van der Waals radii of the atoms of the
C     molecule (to be read in from pdb file) and the water
C     molecule (RH2O = 1.4 A).
C
      IMPLICIT NONE
      INTEGER      MXNAT,MXNTEX,DOCUB
      PARAMETER    (MXNAT=120000,MXNTEX=100)
      INTEGER      NTex,NAt
      REAL         XAt(MXNAT),YAt(MXNAT),ZAt(MXNAT),R(MXNAT),V(MXNAT)
      CHARACTER*80 Text(MXNTEX)
      CHARACTER*4  ANam(MXNAT)
      CHARACTER*3  RNam(MXNAT)
      CHARACTER*1  CNam(MXNAT)
      CHARACTER*5  RId(MXNAT)
C
      REAL         R2(MXNAT),DX,DY,DZ,DIST
      INTEGER      NRes
      REAL         RH2O,VH2O,ACub,THICK
C
C     RH2O = water molecule radius; VH2O = volume of one water molecule;
C     ACub = grid constant for cubical lattice of water molecules that
C     has the density of water at standard conditions (1 H2O / 29.9 A^3).
C
      PARAMETER    (RH2O=1.4,VH2O=29.885)
C
      CHARACTER    pdbfile*132
      REAL         XS(3),Fr(6),VBox,VProt,AvVol
      INTEGER      I,J,K,NWat
      INTEGER      NN
      REAL         LGrid
      REAL         XCen(3),XL(3),XLS(3),X0(3)
      REAL         PI,alpha,beta
C
C     pick any angles alpha and beta for the arbitrary rotation of the
C     diamond lattice...
C
      PARAMETER    (PI=3.141593)
C
C     radii:
      REAL         RC,RO,RN,RH,RP,RS,RX
      PARAMETER    (RC=1.8,RO=1.6,RN=1.6,RH=1.0,RP=2.0,RS=2.0,RX=1.5)
C
      INTEGER      ISEED,IAt,IRes,IChem,IOcc
      LOGICAL      IsFree,InSide
      INTEGER      NPUp,NPLo
      REAL         PUp(3),PLo(3),PFr(6)
C-----------------------------------------------------------------------
      ACub = VH2O**(1./3.)
      WRITE(6,*) '  lattice constant of cubic lattice = ',ACub,' A'
      WRITE(6,*)
C-----------------------------------------------------------------------
      WRITE(6,*) '  the cubic lattice on which waters are placed can'
      WRITE(6,*) '  be rotated around:'
      WRITE(6,*) '  1) first the z-axis by angle alpha (0...360);'
      WRITE(6,*) '  2) then  the x-axis by angle beta (0...360).'
      WRITE(6,*)
    2 WRITE(6,*) '  alpha >'
      READ(5,*,ERR=2) alpha
      WRITE(6,*) 'alpha = ', alpha
      WRITE(6,*)
    3 WRITE(6,*) '  beta >'
      READ(5,*,ERR=3) beta
      WRITE(6,*) 'beta = ', beta
      WRITE(6,*)
C-----------------------------------------------------------------------
    5 WRITE(6,*) '  name of input pdb file >'
      READ(5,'(A)',ERR=5) pdbfile
      OPEN(71,FILE=pdbfile,STATUS='OLD',ERR=999)
      WRITE(6,*) '  reading  pdb file: '
      WRITE(6,'(A)') pdbfile
      NTex = 0
      NAt  = 0
      CALL ReProtL(71,MXNTEX,MXNAT,
     @            NTex,Text,NAt,ANam,RNam,CNam,RId,XAt,YAt,ZAt,R,V)
      CLOSE(71)
      WRITE(6,*) '  read  ',NAt,' atoms'
      WRITE(6,*)
C
      IF (NAt.GT.0) THEN
        WRITE(6,*) '  last residue Id is ',RId(NAt)
        READ(RId(NAt),*) NRes
        WRITE(6,*) '  interpreted as No  ',NRes
        WRITE(6,*)
      ELSE
        NRes = 0
      ENDIF
C
   10 WRITE(6,*) '  position of chemical symbol in atom name (1/2)'
      WRITE(6,*) '  X-PLOR uses 1, pdb-convention is 2 > '
      READ(5,*,ERR=10) IChem
      WRITE(6,*) 'IChem = ', IChem
      WRITE(6,*)
C
      WRITE(6,*) '  now setting atomic radii and volumes...'
      DO 50 I=1,NAt,1
        IF (ANam(I)(IChem:IChem).EQ.'C') THEN
	  R(I) = RC
        ELSEIF (ANam(I)(IChem:IChem).EQ.'O') THEN
	  R(I) = RO
        ELSEIF (ANam(I)(IChem:IChem).EQ.'N') THEN
	  R(I) = RN
        ELSEIF (ANam(I)(IChem:IChem).EQ.'H') THEN
	  R(I) = RH
        ELSEIF (ANam(I)(IChem:IChem).EQ.'P') THEN
	  R(I) = RP
        ELSEIF (ANam(I)(IChem:IChem).EQ.'S') THEN
	  R(I) = RS
        ELSEIF (ANam(I)(IChem:IChem).EQ.' ') THEN
C In this case we give it a second chance
	  IChem = IChem + 1
	  IF (ANam(I)(IChem:IChem).EQ.'C') THEN
	    R(I) = RC
	  ELSEIF (ANam(I)(IChem:IChem).EQ.'O') THEN
	    R(I) = RO
	  ELSEIF (ANam(I)(IChem:IChem).EQ.'N') THEN
	    R(I) = RN
	  ELSEIF (ANam(I)(IChem:IChem).EQ.'H') THEN
	    R(I) = RH
	  ELSEIF (ANam(I)(IChem:IChem).EQ.'P') THEN
	    R(I) = RP
	  ELSEIF (ANam(I)(IChem:IChem).EQ.'S') THEN
	    R(I) = RS
	  ELSE
	    WRITE(6,*) '  setw: WARNING, do not know chemical symbol of: '
	    WRITE(6,'(2X,I6,X,A4,X,A3,X,A5)') I,ANam(I),RNam(I),RId(I)
	    WRITE(6,*)
	    R(I) = RX
	  ENDIF
C Don't forget to reset IChem as otherwise things are screwed for the
C next test.
	  IChem = IChem - 1
	ELSE
	  WRITE(6,*) '  setw: WARNING, do not know chemical symbol of: '
          WRITE(6,'(2X,I6,X,A4,X,A3,X,A5)') I,ANam(I),RNam(I),RId(I)
          WRITE(6,*)
	  R(I) = RX
	ENDIF
	V(I) = 4.*PI*(R(I)**3)/3.
   50 CONTINUE
      WRITE(6,*)
C
      WRITE(6,*) '  some examples of radii:'
      DO 55 I=1,12,1
        WRITE(6,'(I5,X,3(A,X),F6.1)') I,ANam(I),RNam(I),RId(I),R(I)
   55 CONTINUE
      WRITE(6,*)
C 
      DO 100 I=1,NAt,1
        R(I) = R(I)+RH2O
        R2(I) = R(I)*R(I)
  100 CONTINUE
C
      CALL Center(1,NAt,XAt,YAt,ZAt,RNam,XS)
      WRITE(6,'(1X,A,3F6.2)') '  center of non-water atoms at ',
     @                       XS(1),XS(2),XS(3)
      WRITE(6,*)
C
      CALL FramIt(1,NAt,XAt,YAt,ZAt,RNam,Fr)
      WRITE(6,'(1X,A,6F6.1)') '  Frame (non-water)',(Fr(I),I=1,6)
      WRITE(6,*)
C
      NWat = 0
      DO 200 I=1,NAt,1
        IF (((RNam(I).EQ.'H2O').OR.(RNam(I).EQ.'TIP')).AND.
     @      ((ANam(I).EQ.' O  ').OR.(ANam(I).EQ.' OH2'))) NWat=NWat+1
  200 CONTINUE
      WRITE(6,'(1X,A,I5,A)')
     @'  input structure includes ',NWat,' water molecules'
      WRITE(6,*)
C-----------------------------------------------------------------------
  210 WRITE(6,*) '  Make cubic water box (1) or a sphere (2) >'
      READ(5,*,ERR=210) DOCUB
      IF (DOCUB.eq.1) THEN
           WRITE(6,*) '  Making a cubic water box !'
      ELSE
           WRITE(6,*) '  Making a spheric water box !'
      ENDIF	
C-----------------------------------------------------------------------
  220 WRITE(6,*) '  minimum thickness of water layer around molecule >'
      READ(5,*,ERR=220) THICK
      WRITE(6,*) '  minimum thickness is ', THICK
      WRITE(6,*)
C
C add water layer to cube:
      Fr(1) = Fr(1)-THICK
      Fr(2) = Fr(2)-THICK
      Fr(3) = Fr(3)-THICK
      Fr(4) = Fr(4)+THICK
      Fr(5) = Fr(5)+THICK
      Fr(6) = Fr(6)+THICK
C get longest dimension:
      LGrid = Fr(4)-Fr(1)
      IF (Fr(5)-Fr(2).GT.LGrid) LGrid=Fr(5)-Fr(2)
      IF (Fr(6)-Fr(3).GT.LGrid) LGrid=Fr(6)-Fr(3)
C
C because the cube including the system frame plus water layer will be
C rotated, LGrid must be multiplied by SQRT(3) to make sure no corner
C will stick out through one of the faces of the rotated cube;
C the diagonal of a cube is SQRT(3)*a, with a the cube length.
C
      LGrid = SQRT(3.0)*LGrid
      NN = INT(LGrid/ACub)+1
      XCen(1) = 0.5*(Fr(1)+Fr(4))
      XCen(2) = 0.5*(Fr(2)+Fr(5))
      XCen(3) = 0.5*(Fr(3)+Fr(6))
      WRITE(6,*) '  grid length = ',LGrid
      WRITE(6,*) '  center of grid = ',(XCen(I),I=1,3,1)
      WRITE(6,*) '  No of grid points in one dim = ',NN
      WRITE(6,*)
C
C compute the average location of the phosphorus atoms:
      CALL GetP('P','LIP',IChem,NAt,XAt,YAt,ZAt,ANam,RNam,
     @          NPUp,NPLo,PUp,PLo)
      WRITE(6,*) '  found ',NPUp,' phosphorus atoms in upper half'
      WRITE(6,*) '  found ',NPLo,' phosphorus atoms in lower half'
      WRITE(6,*) '  average position of upper P ',(PUp(I),I=1,3,1)
      WRITE(6,*) '  average position of lower P ',(PLo(I),I=1,3,1)
      WRITE(6,*)
C
C assign the phosphor frame:
	PFr(1) = Fr(1)
	PFr(2) = PLo(2)
	PFr(3) = Fr(3)
	PFr(4) = Fr(4)
	PFr(5) = PUp(2)
	PFr(6) = Fr(6)
C
      WRITE(6,*)
  505 WRITE(6,*) '  name of output water pdb file >'
      READ(5,'(A)',ERR=505) pdbfile
      OPEN(81,FILE=pdbfile,STATUS='UNKNOWN',ERR=998)
      WRITE(6,*)
C
      WRITE(6,*) '  calculating possible water locations...'
      NWat  = 0
      IOcc  = 0
      IAt   = NAt
      IRes  = 0
      ISEED = 123456
      CALL SRAND(ISEED)
      XL(1) = XCen(1)-0.5*LGrid
      DO 430 I=1,NN,1
        WRITE(6,*) '  IX=',I,' out of ',NN
        XL(2) = XCen(2)-0.5*LGrid
        DO 420 J=1,NN,1
C         WRITE(6,*) '  IY=',J,' out of ',NN
          XL(3) = XCen(3)-0.5*LGrid
          DO 410 K=1,NN,1
C           first fcc grid
            CALL Turn(XL,XCen,alpha,beta,X0)
C We place waters only inside the frame Fr which has THICK added to all
C dimensions, but we DON'T place waters below (above) the P43 line (PFr)
            IF (InSide(X0(1),X0(2),X0(3),Fr,'xyz').AND.
     @         (.NOT.InSide(X0(1),X0(2),X0(3),PFr,'y'))) THEN
               IF (IsFree(X0,NAt,XAt,YAt,ZAt,R,R2)) THEN
                  IF (DOCUB.eq.1) THEN		
                     NWat = NWat+1
                     CALL PutWat(81,IAt,IRes,X0,RH2O,VH2O,ISEED)
                  ELSE
                     DX = XS(1) - XL(1)
                     DY = XS(2) - XL(2)
                     DZ = XS(3) - XL(3)
                     DIST = SQRT(DX*DX + DY*DY + DZ*DZ)
                     IF (DIST.le.THICK) THEN
                        NWat = NWat+1
                        CALL PutWat(81,IAt,IRes,X0,RH2O,VH2O,ISEED)
                     ELSE
	                IOcc = IOcc+1
                     ENDIF
                  ENDIF
	       ELSE
	          IOcc = IOcc+1
	       ENDIF
            ENDIF
            XL(3) = XL(3)+ACub
  410     CONTINUE
          XL(2) = XL(2)+ACub
  420   CONTINUE
        XL(1) = XL(1)+ACub
  430 CONTINUE
      CLOSE(81)
      WRITE(6,'(X,A,I6,A)') '  found ',NWat,' empty water positions'
      WRITE(6,'(X,A,I6,A)') '  found ',IOcc,' occupied positions'
      WRITE(6,*)
C
      VBox = (Fr(4)-Fr(1))*(Fr(5)-Fr(2))*(Fr(6)-Fr(3))
      VProt = 0.0
      DO 500 I=1,NAt,1
        VProt = VProt+V(I)
  500 CONTINUE
      IF (NWat.GT.0) THEN
        AvVol = (VBox-VProt)/FLOAT(NWat)
        WRITE(6,'(1X,A,F12.1)') '  box volume               ',VBox
        WRITE(6,'(1X,A,F12.1)') '  protein effective volume ',VProt
        WRITE(6,'(1X,A,F12.1)') '  av. vol. occ. by 1 water ',AvVol
        WRITE(6,*)
      ENDIF
C
      STOP
C
  998 WRITE(6,*) '  setw_cub: ERROR opening output file!'
      STOP
  999 WRITE(6,*) '  setw_cub: ERROR opening input file!'
      STOP
C
      END
C
C
C***********************************************************************
      FUNCTION IsFree(X0,NAt,XAt,YAt,ZAt,R,R2)
      IMPLICIT NONE
      LOGICAL      IsFree
      INTEGER      NAt
      REAL         X0(3),XAt(*),YAt(*),ZAt(*),R(*),R2(*)
      REAL         delX,delY,delZ
      INTEGER      I
C
      IsFree = .FALSE.
      DO 100 I=1,NAt,1
        delZ = ZAt(I)-X0(3)
        IF (delZ.LT.R(I)) THEN
          delX = XAt(I)-X0(1)
          IF (delX.LT.R(I)) THEN
            delY = YAt(I)-X0(2)
            IF (delY.LT.R(I)) THEN
              IF ((delX*delX+delY*delY+delZ*delZ).LT.R2(I)) RETURN
            ENDIF
          ENDIF
        ENDIF
  100 CONTINUE
C
      IsFree = .TRUE.
C
      RETURN
      END
C
C
C***********************************************************************
      SUBROUTINE Turn(XOld,XS,alpha,beta,XNew)
      IMPLICIT NONE
      REAL XOld(3),XS(3),alpha,beta,XNew(3)
      REAL XR1(3),XR2(3),sinA,cosA,sinB,cosB
C
      sinA = SIN(alpha)
      cosA = COS(alpha)
      sinB = SIN(beta)
      cosB = COS(beta)
C
      XR1(1) = XOld(1)-XS(1)
      XR1(2) = XOld(2)-XS(2)
      XR1(3) = XOld(3)-XS(3)
C
      XR2(1) = cosA*XR1(1)-sinA*XR1(2)
      XR2(2) = sinA*XR1(1)+cosA*XR1(2)
      XR2(3) = XR1(3)
C
      XNew(1) = XS(1)+XR2(1)
      XNew(2) = XS(2)+cosB*XR2(2)-sinB*XR2(3)
      XNew(3) = XS(3)+sinB*XR2(2)+cosB*XR2(3)
C
      RETURN
      END
C
C
C***********************************************************************
      SUBROUTINE PutWat(IOUT,IAt,IRes,X,P1,P2,ISEED)
C     NOTE: in violation of PDB format we write atom numbers GT 99999
C           by using col 6..12 for the number (instead of 7..12)
      IMPLICIT NONE
      INTEGER   IOUT,IAt,IRes,ISEED
      REAL      X(3),P1,P2,XH(3)
      REAL      Axis(3),PI,alpHOH,lenOH,alp2,fac1,fac2
C changed the lenOH from 1.0A to 0.9572 A to have the same value as in 
C X-PLOR TIP3W water (otherwise problems with SHAKE)
      PARAMETER (PI=3.141593,alpHOH=105.0*PI/180.0,lenOH=0.9572)
      REAL      the,phi,len,M1(3,3),M2(3,3),xran,P(3),U(3)
      DOUBLE PRECISION RAND
C
      IRes = IRes+1
      IAt  = IAt+1
      WRITE(IOUT,'(A5,I6,1X,A4,1X,A4,A1,I5,3X,3F8.3,2F6.2)')
     @  'ATOM ',IAt,' OH2','TIP3',' ',IRes,X(1),X(2),X(3),P1,P2
C
  100 Axis(1) = -1.0+2.0*REAL(RAND())
      Axis(2) = -1.0+2.0*REAL(RAND())
      Axis(3) = -1.0+2.0*REAL(RAND())
      len = SQRT(Axis(1)*Axis(1)+Axis(2)*Axis(2)+Axis(3)*Axis(3))
      IF ((len.GT.1.0).OR.(len.LE.0.0)) GOTO 100
      Axis(1) = Axis(1)/len
      Axis(2) = Axis(2)/len
      Axis(3) = Axis(3)/len
C
      the = ACOS(Axis(3))
      IF (Axis(1).NE.0.0) THEN
        phi = ATAN(Axis(2)/Axis(1))
        IF (Axis(1).LT.0.0) phi=phi+PI
      ELSEIF (Axis(2).GE.0.0) THEN
        phi = 0.5*PI
      ELSE
        phi = -0.5*PI
      ENDIF
      CALL RotM(2,the,M1)
      CALL RotM(3,phi,M2)
C
      U(1) = 0.0
      U(2) = 0.0
      U(3) = 1.0
      CALL Rot(U,M1)
      CALL Rot(U,M2)
C     WRITE(6,*) '  ratios',U(1)/Axis(1),U(2)/Axis(2),U(3)/Axis(3)
      xran = PI*(-1.0+2.0*REAL(RAND()))
      P(1) = COS(xran)
      P(2) = SIN(xran)
      P(3) = 0.0
      CALL Rot(P,M1)
      CALL Rot(P,M2)
C     WRITE(6,*) '  scalP=',Axis(1)*P(1)+Axis(2)*P(2)+Axis(3)*P(3)
C
      alp2 = 0.5*alpHOH
      fac1 = lenOH*COS(alp2)
      fac2 = lenOH*SIN(alp2)
      XH(1) = X(1)+fac1*Axis(1)+fac2*P(1)
      XH(2) = X(2)+fac1*Axis(2)+fac2*P(2)
      XH(3) = X(3)+fac1*Axis(3)+fac2*P(3)
      IAt  = IAt+1
      WRITE(IOUT,'(A5,I6,1X,A4,1X,A4,A1,I5,3X,3F8.3,2F6.2)')
     @  'ATOM ',IAt,' H1 ','TIP3',' ',IRes,XH(1),XH(2),XH(3),0.0,0.0
      XH(1) = X(1)+fac1*Axis(1)-fac2*P(1)
      XH(2) = X(2)+fac1*Axis(2)-fac2*P(2)
      XH(3) = X(3)+fac1*Axis(3)-fac2*P(3)
      IAt  = IAt+1
      WRITE(IOUT,'(A5,I6,1X,A4,1X,A4,A1,I5,3X,3F8.3,2F6.2)')
     @  'ATOM ',IAt,' H2 ','TIP3',' ',IRes,XH(1),XH(2),XH(3),0.0,0.0
C
      RETURN
      END
C
C
C***********************************************************************
      SUBROUTINE Rot(V,M)
      IMPLICIT NONE
      REAL V(3),M(3,3)
      REAL VV(2)
C
      VV(1) = V(1)
      V(1) = M(1,1)*V(1)+M(1,2)*V(2)+M(1,3)*V(3)
C      
      VV(2) = V(2)
      V(2) = M(2,1)*VV(1)+M(2,2)*V(2)+M(2,3)*V(3)
C
      V(3) = M(3,1)*VV(1)+M(3,2)*VV(2)+M(3,3)*V(3)
C
      RETURN
      END
C
C
