C
C     Routines written by M. Schaefer; 1990 and H. Heller, thereafter.
C
C***********************************************************************        
      SUBROUTINE ReProt(IUN,MXNTEX,MXNAT,
     @                  NTex,Text,NAt,ANam,RNam,CNam,RId,
     @                  XAt,YAt,ZAt,P1,P2)
C
C     Reads a protein structure in Brookhaven format from unit IUN.
C     NOTE: parameter P2 is read in format F8.4 instead of F6.2!
C
      IMPLICIT NONE
C     input:
      INTEGER      IUN
      INTEGER      MXNTEX,MXNAT
C     output:
      INTEGER      NTex,NAt
      CHARACTER*80 Text(*)
      CHARACTER*4  ANam(*)
      CHARACTER*3  RNam(*)
      CHARACTER*1  CNam(*)
      CHARACTER*5  RId(*)
      REAL         XAt(*),YAt(*),ZAt(*),P1(*),P2(*)
C     additional parameters:
      INTEGER      IShow,NSHOW
      PARAMETER    (NSHOW=12)
      CHARACTER*80 Line                                                         
C-----------------------------------------------------------------------
C     read all atom/hetatm-lines and text lines:
      IShow = 0
  100 CONTINUE                                                                  
        READ(IUN,'(A)',END=110,ERR=999) Line
        IF ((Line(1:6).EQ.'ATOM  ').OR.(Line(1:6).EQ.'HETATM')) THEN
          IF ((Line(17:17).EQ.' ').OR.(Line(17:17).EQ.'A')) THEN
	    IF (NAt.GE.MXNAT) THEN
	      WRITE(6,*) '  ReProt: ERROR, MXNAT = ',MXNAT,' exceeded!'
	      WRITE(6,*)
     @        '  Increase MXNAT in main program and recompile!'
              WRITE(6,*)
              RETURN
	    ENDIF
            NAt = NAt+1
            READ(Line,'(12X,A4,1X,A3,1X,A1,A5,3X,3F8.3,F6.2,F8.4)',
     @           ERR=999)
     @      ANam(NAt),RNam(NAt),CNam(NAt),RId(NAt),
     @      XAt(NAt),YAt(NAt),ZAt(NAt),P1(NAt),P2(NAt)
          ENDIF
        ELSEIF ((Line(1:4).EQ.'HEAD').OR.(Line(1:4).EQ.'COMP')
     @                               .OR.(Line(1:4).EQ.'REMA')) THEN
	  IF (NTex.GE.MXNTEX) THEN
	    WRITE(6,*) '  ReProt: ERROR, MXNTEX = ',MXNTEX,' exceeded!'
	    WRITE(6,*)
     @      '  Increase MXNTEX in main program and recompile!'
            WRITE(6,*)
            RETURN
	  ENDIF
          NTex = NTex+1
          Text(NTex) = Line
        ENDIF
        IF (IShow.LE.NSHOW) THEN
          IShow = IShow+1
          WRITE(6,'(A)') Line
        ENDIF
      GOTO 100
  110 CONTINUE
      RETURN
C
  999 WRITE(6,*) '  ReProt: ERROR reading input file!'
      RETURN
      END
C
C
C***********************************************************************        
      SUBROUTINE ReProtL(IUN,MXNTEX,MXNAT,
     @                  NTex,Text,NAt,ANam,RNam,CNam,RId,
     @                  XAt,YAt,ZAt,P1,P2)
C
C     Reads a protein structure in Brookhaven format from unit IUN.
C     NOTE: parameter P2 is read in format F8.4 instead of F6.2!
C     NOTE: in violation of PDB format we read atom numbers GT 99999
C
      IMPLICIT NONE
C     input:
      INTEGER      IUN
      INTEGER      MXNTEX,MXNAT
C     output:
      INTEGER      NTex,NAt
      CHARACTER*80 Text(*)
      CHARACTER*4  ANam(*)
      CHARACTER*3  RNam(*)
      CHARACTER*1  CNam(*)
      CHARACTER*5  RId(*)
      REAL         XAt(*),YAt(*),ZAt(*),P1(*),P2(*)
C     additional parameters:
      INTEGER      IShow,NSHOW
      PARAMETER    (NSHOW=12)
      CHARACTER*80 Line                                                         
C-----------------------------------------------------------------------
C     read all atom/hetatm-lines and text lines:
      IShow = 0
  100 CONTINUE                                                                  
        READ(IUN,'(A)',END=110,ERR=999) Line
        IF ((Line(1:5).EQ.'ATOM ').OR.(Line(1:6).EQ.'HETATM')) THEN
          IF ((Line(17:17).EQ.' ').OR.(Line(17:17).EQ.'A')) THEN
	    IF (NAt.GE.MXNAT) THEN
	      WRITE(6,*) '  ReProtL: ERROR, MXNAT = ',MXNAT,' exceeded!'
	      WRITE(6,*)
     @        '  Increase MXNAT in main program and recompile!'
              WRITE(6,*)
              RETURN
	    ENDIF
            NAt = NAt+1
            READ(Line,'(12X,A4,1X,A3,1X,A1,A5,3X,3F8.3,F6.2,F8.4)',
     @           ERR=999)
     @      ANam(NAt),RNam(NAt),CNam(NAt),RId(NAt),
     @      XAt(NAt),YAt(NAt),ZAt(NAt),P1(NAt),P2(NAt)
          ENDIF
        ELSEIF ((Line(1:4).EQ.'HEAD').OR.(Line(1:4).EQ.'COMP')
     @                               .OR.(Line(1:4).EQ.'REMA')) THEN
	  IF (NTex.GE.MXNTEX) THEN
	    WRITE(6,*) '  ReProtL: ERROR, MXNTEX = ',MXNTEX,' exceeded!'
	    WRITE(6,*)
     @      '  Increase MXNTEX in main program and recompile!'
            WRITE(6,*)
            RETURN
	  ENDIF
          NTex = NTex+1
          Text(NTex) = Line
        ENDIF
        IF (IShow.LE.NSHOW) THEN
          IShow = IShow+1
          WRITE(6,'(A)') Line
        ENDIF
      GOTO 100
  110 CONTINUE
      RETURN
C
  999 WRITE(6,*) '  ReProtL: ERROR reading input file!'
      RETURN
      END
C
C
C***********************************************************************        
      SUBROUTINE WrProt(IUN,
     @                  NTex,Text,NAt,ANam,RNam,CNam,RId,
     @                  XAt,YAt,ZAt,P1,P2)
C     Writes a protein structure in Brookhaven format to unit IUN:
C     NOTE: parameter P2 is written in format F8.4 instead of F6.2!
      IMPLICIT NONE                                                             
C     input parameters:
      INTEGER      IUN
      INTEGER      NTex,NAt
      CHARACTER*80 Text(*)
      CHARACTER*4  ANam(*)
      CHARACTER*3  RNam(*)
      CHARACTER*1  CNam(*)
      CHARACTER*5  RId(*)
      REAL         XAt(*),YAt(*),ZAt(*),P1(*),P2(*)
C     additional parameters:
      INTEGER      I
C-----------------------------------------------------------------------
      DO 100 I=1,NTex,1
        WRITE(IUN,'(A)') Text(I)
  100 CONTINUE
C
      DO 200 I=1,NAt,1
        IF (RNam(I).EQ.'TIP') THEN
          WRITE(IUN,'(A6,I5,1X,A4,1X,A3,1H3,A1,A5,3X,3F8.3,F6.2,F8.4)')
     @    'ATOM  ',I,ANam(I),RNam(I),CNam(I),RId(I),
     @    XAt(I),YAt(I),ZAt(I),P1(I),P2(I)
        ELSEIF (RNam(I).EQ.'DPP') THEN
          WRITE(IUN,'(A6,I5,1X,A4,1X,A3,1HC,A1,A5,3X,3F8.3,F6.2,F8.4)')
     @    'ATOM  ',I,ANam(I),RNam(I),CNam(I),RId(I),
     @    XAt(I),YAt(I),ZAt(I),P1(I),P2(I)
        ELSEIF (RNam(I).EQ.'POP') THEN
          WRITE(IUN,'(A6,I5,1X,A4,1X,A3,1HC,A1,A5,3X,3F8.3,F6.2,F8.4)')
     @    'ATOM  ',I,ANam(I),RNam(I),CNam(I),RId(I),
     @    XAt(I),YAt(I),ZAt(I),P1(I),P2(I)
        ELSE
          WRITE(IUN,'(A6,I5,1X,A4,1X,A3,1X,A1,A5,3X,3F8.3,F6.2,F8.4)')
     @    'ATOM  ',I,ANam(I),RNam(I),CNam(I),RId(I),
     @    XAt(I),YAt(I),ZAt(I),P1(I),P2(I)
        ENDIF
  200 CONTINUE
      WRITE(IUN,'(A)') 'END   '
C
      RETURN
      END
C
C
C***********************************************************************        
      SUBROUTINE WrProtL(IUN,
     @                  NTex,Text,NAt,ANam,RNam,CNam,RId,
     @                  XAt,YAt,ZAt,P1,P2)
C     Writes a protein structure in Brookhaven format to unit IUN:
C     NOTE: parameter P2 is written in format F8.4 instead of F6.2!
C     NOTE: in violation of PDB format we write atom numbers GT 99999
C           by using col 6..12 for the number (instead of 7..12)
      IMPLICIT NONE                                                             
C     input parameters:
      INTEGER      IUN
      INTEGER      NTex,NAt
      CHARACTER*80 Text(*)
      CHARACTER*4  ANam(*)
      CHARACTER*3  RNam(*)
      CHARACTER*1  CNam(*)
      CHARACTER*5  RId(*)
      REAL         XAt(*),YAt(*),ZAt(*),P1(*),P2(*)
C     additional parameters:
      INTEGER      I
C-----------------------------------------------------------------------
      DO 100 I=1,NTex,1
        WRITE(IUN,'(A)') Text(I)
  100 CONTINUE
C
      DO 200 I=1,NAt,1
        IF (RNam(I).EQ.'TIP') THEN
          WRITE(IUN,'(A5,I6,1X,A4,1X,A3,1H3,A1,A5,3X,3F8.3,F6.2,F8.4)')
     @    'ATOM  ',I,ANam(I),RNam(I),CNam(I),RId(I),
     @    XAt(I),YAt(I),ZAt(I),P1(I),P2(I)
        ELSEIF (RNam(I).EQ.'DPP') THEN
          WRITE(IUN,'(A5,I6,1X,A4,1X,A3,1HC,A1,A5,3X,3F8.3,F6.2,F8.4)')
     @    'ATOM  ',I,ANam(I),RNam(I),CNam(I),RId(I),
     @    XAt(I),YAt(I),ZAt(I),P1(I),P2(I)
        ELSEIF (RNam(I).EQ.'POP') THEN
          WRITE(IUN,'(A5,I6,1X,A4,1X,A3,1HC,A1,A5,3X,3F8.3,F6.2,F8.4)')
     @    'ATOM  ',I,ANam(I),RNam(I),CNam(I),RId(I),
     @    XAt(I),YAt(I),ZAt(I),P1(I),P2(I)
        ELSE
          WRITE(IUN,'(A5,I6,1X,A4,1X,A3,1X,A1,A5,3X,3F8.3,F6.2,F8.4)')
     @    'ATOM  ',I,ANam(I),RNam(I),CNam(I),RId(I),
     @    XAt(I),YAt(I),ZAt(I),P1(I),P2(I)
        ENDIF
  200 CONTINUE
      WRITE(IUN,'(A)') 'END   '
C
      RETURN
      END
C
C
C***********************************************************************        
      SUBROUTINE WrProtH(IUN,
     @                   NTex,Text,NAt,ANam,RNam,CNam,RId,
     @                   XAt,YAt,ZAt,P1,P2)
C
C     Writes a protein structure in Brookhaven format to unit IUN:
C     No water hydrogen atoms are written!
C     NOTE: parameter P2 is written in format F8.4 instead of F6.2!
C
      IMPLICIT NONE
C     input parameters:
      INTEGER      IUN,NTex,NAt
      CHARACTER*80 Text(*)
      CHARACTER*4  ANam(*)
      CHARACTER*3  RNam(*)
      CHARACTER*1  CNam(*)
      CHARACTER*5  RId(*)
      REAL         XAt(*),YAt(*),ZAt(*),P1(*),P2(*)
C     additional parameters:
      INTEGER      I
C-----------------------------------------------------------------------
      DO 100 I=1,NTex,1
        WRITE(IUN,'(A)') Text(I)
  100 CONTINUE
C
      DO 200 I=1,NAt,1
        IF (RNam(I).NE.'TIP') THEN
          WRITE(IUN,'(A6,I5,1X,A4,1X,A3,1X,A1,A5,3X,3F8.3,F6.2,F8.4)')
     @    'ATOM  ',I,ANam(I),RNam(I),CNam(I),RId(I),
     @    XAt(I),YAt(I),ZAt(I),P1(I),P2(I)
        ELSEIF (ANam(I).EQ.' OH2') THEN
          WRITE(IUN,'(A6,I5,1X,A4,1X,A3,1H3,A1,A5,3X,3F8.3,F6.2,F8.4)')
     @    'ATOM  ',I,ANam(I),RNam(I),CNam(I),RId(I),
     @    XAt(I),YAt(I),ZAt(I),P1(I),P2(I)
        ENDIF
  200 CONTINUE
      WRITE(IUN,'(A)') 'END   '
C
      RETURN
      END
C
C

