C***********************************************************************
C
C     Written by Michael Schaefer, UIUC, Beckman Institute,
C                                  405 N. Mathews Ave.
C                                  Urbana, Illinois 61801
C                                  U.S.A.
C
C     This library works on SGI machines. Passing of ISEED is not
C     necessary, but kept for compatibility with implementations
C     on other machines or self made random number generators.
C
C***********************************************************************
      FUNCTION Norm(ISEED)
C
C     Returns a normal distributed real number, i.e. centered around 0
C     and with probability P(x) = Exp(-(x**2)/2.0); uses 12 steps.
C     Call srand(iseed) in the main program required for initialization!
C
      IMPLICIT NONE
      REAL             Norm
      INTEGER          ISEED
      INTEGER          I
      DOUBLE PRECISION RAND
C-----------------------------------------------------------------------
      Norm = 0.0
      DO 100 I=1,12,1
        Norm = Norm+REAL(RAND( ))
  100 CONTINUE
      Norm = Norm-6.0
      RETURN
      END
C
C
C***********************************************************************
      FUNCTION Norm1(ISEED)
C
C     Returns a normal distributed real number, i.e. centered around 0
C     and with probability P(x) = Exp(-(x**2)/2.0); uses 24 steps.
C     Call srand(iseed) in the main program required for initialization!
C
      IMPLICIT NONE
      REAL             Norm1
      INTEGER          ISEED
      REAL             SQ2m1
C     SQ2m1 = 1.0/SQRT(2.0)
      PARAMETER        (SQ2m1=0.7071068)
      INTEGER          I
      DOUBLE PRECISION RAND
C-----------------------------------------------------------------------
      Norm1 = 0.0
      DO 100 I=1,24,1
        Norm1 = Norm1+REAL(RAND( ))
  100 CONTINUE
      Norm1 = SQ2m1*(Norm1-12.0)
      RETURN
      END
C
C
C***********************************************************************
      SUBROUTINE GSS1D(X,X0,sig,ISEED)
C
C     Returns X with probability exp(-(X-X0)**2/sig**2).
C     Call srand(iseed) in the main program required for initialization!
C
      IMPLICIT NONE
      REAL    X,X0,sig
      INTEGER ISEED
      REAL    SQ2m1
C     SQ2m1 = 1.0/SQRT(2.0)
      PARAMETER(SQ2m1=0.7071068)
      REAL    Norm
C-----------------------------------------------------------------------
      X = X0+SQ2m1*sig*Norm(ISEED)
C
      RETURN
      END
C
C
C***********************************************************************
      SUBROUTINE GSS3D(X,X0,sig,ISEED)
C
C     Returns vector X with
C     probability ((X-X0)**2)*exp(-(X-X0)**2/sig**2);
C     radial distribution P(r) = (r**2)*exp(-r**2/sig**2);
C     note the factor (X-X0)**2, which results from the uniform distri-
C     bution of vectors around X0, or the volume element in spherical
C     coordinates dV = (r**2)*dr*dphi*dtheta.
C     Call srand(iseed) in the main program required for initialization!
C
      IMPLICIT NONE
      REAL    X(3),X0(3),sig
      INTEGER ISEED
      REAL    SQ2m1
C     SQ2m1 = 1.0/SQRT(2.0)
      PARAMETER(SQ2m1=0.7071068)
      REAL    Norm
C-----------------------------------------------------------------------
      X(1) = X0(1) + SQ2m1*sig*Norm(ISEED)
      X(2) = X0(2) + SQ2m1*sig*Norm(ISEED)
      X(3) = X0(3) + SQ2m1*sig*Norm(ISEED)
C
      RETURN
      END
C
C
C***********************************************************************
      SUBROUTINE GGUBFS(RANDOM)
C
C uniform (0,1) random number generator
C
C Author: Axel T. Brunger
C =======================
C
      IMPLICIT NONE
C input/output
      DOUBLE PRECISION RANDOM
C
C random number generator seed and parameters for random number
C generator.  The parameters always have to be computed as double
C precision variables.
      D O U B L E  P R E C I S I O N SEED, D2P31M, D2P31
C D2P31M=(2**31) - 1
C D2P31 =(2**31)(OR AN ADJUSTED VALUE)
      PARAMETER (D2P31M=2147483647.0D0)
      PARAMETER (D2P31=2147483711.0D0)
C
      COMMON /CCSEED/ SEED
      SAVE /CCSEED/
C begin
      IF (SEED.LT.1.0D0)  SEED=314564.D0
      SEED=MOD(16807.0D0*SEED,D2P31M)
      RANDOM=SEED/D2P31
      RETURN
      END
C
C
C
C***********************************************************************
      DOUBLE PRECISION FUNCTION RAND ()
C
C Is a dummy. Introduced to replace built-in RAND() function for f2c compiler.
C That compiler does not seem to know RAND().
C Helmut Heller
C
      IMPLICIT NONE
      CALL GGUBFS(RAND)
	  RETURN
	  END
C
C
C
C***********************************************************************
      SUBROUTINE SRAND(ISEED)
C
C Is a dummy. Introduced to replace built-in RAND() function for f2c compiler.
C That compiler does not seem to know RAND().
C Helmut Heller
C
      IMPLICIT NONE
      INTEGER ISEED
	  RETURN
	  END
	  	    
