/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: util_lib.c						**
**	Needed by: nearly every other program				**
**									**
*************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#if defined(PARIX) || defined(PARIXHAL) || defined(NEXT) || defined(LINUX) || defined(LINUX64)
#include <stdlib.h>
#endif
#include <signal.h>
#if defined(NEXT)
#include <libc.h>
#else
#include <unistd.h>
#endif

#include "ego.h"
#include "util_lib.h"
#include "egalloc.h"

#ifndef PYTHAG
static double at,bt,ct;
#define PYTHAG(a,b) ((at=fabs(a)) > (bt=fabs(b)) ? \
(ct=bt/at,at*sqrt(1.0+ct*ct)) : (bt ? (ct=at/bt,bt*sqrt(1.0+ct*ct)): 0.0))
#endif



/***********************************************************************/
/***********************************************************************/
static	char	sdum[400];

/***********************************************************************/
/*** a very good random generator from the numerical recipies **********/
/***********************************************************************/
#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

static float ran2(long *idum);

static
float ran2(long *idum)
{
	int j;
	long k;
	static long idum2=123456789;
	static long iy=0;
	static long iv[NTAB];
	float temp;
	
	if (*idum <= 0) {		
		if (-(*idum)<1) *idum=1;
		else *idum = -(*idum);
		idum2=(*idum);
		for (j=NTAB+7;j>=0;j--) {
			k=(*idum)/IQ1;
			*idum=IA1*(*idum-k*IQ1)-k*IR1;
			if (*idum<0) *idum += IM1;
			if (j<NTAB) iv[j] = (*idum);		
		}
		iy=iv[0];	
	}
	k=(*idum)/IQ1;
	*idum=IA1*(*idum-k*IQ1)-k*IR1;
	if (*idum < 0) *idum += IM1;
	k = idum2/IQ2;
	idum2 = IA2*(idum2-k*IQ2)-k*IR2;
	if (idum2<0) idum2 += IM2;
	j=iy/NDIV;
	iy=iv[j]-idum2;
	iv[j] = (*idum);
	if (iy<1) iy += IMM1;
	if ((temp=AM*iy) > RNMX) return RNMX;
	else return temp;			
}

static long myseed=-13L;

int set_rand_seed(long i)
{
	myseed=i;
	return 0;	
}
	
double ego_rand()
{
	float r=ran2(&myseed);

	return (double) r;
}
/*********************************************************************/
/*********************************************************************/

#if defined(PARIX) || defined(PARIXHAL) || defined(PARIXPPC)
void exit(int i);
#endif

#if defined(T3D) || defined(HPPA)
double copysign (double x, double y)
{
	double r;
	if (y == 0)	r = 0;
	else		r = fabs (x) * (y / fabs (y));
	return r;
}
#endif

void create_outpath(LOAD_MSG_ALL_NODE *global_data)
{
  	    if (access(global_data->outpath,W_OK)==-1) {
		int len;
	        char	sdum[MAX_STRING_SIZE];
		txtout("Creating output path <%s> !\n",global_data->outpath);
		sprintf(sdum,"mkdir %s",global_data->outpath);
		len = strlen(sdum);
		if (sdum[len-1]=='/') sdum[len-1]=0;
		system(sdum);		
	    }  		
}

int	get_my_endian(void)	/*** 0 little endean ***/
{				/*** 1 big endean    ***/		
	int	test=1;
	char	*a;
	a = (char*) &test;
	if (*a==1) return 1;
	return 0;
}

int	extract_path(char *filename,char *path)
{
	int i,len;

	len=strlen(filename);
	strcpy(path,filename);
	if (len<=0)	return -1;
	
	for (i=len-1;i>=0;i--) {
		if ((path[i]!='\\')&&(path[i]!='/')) {
			path[i]=0;
		} else break;
	}
	return 0;
}

int	extract_fname(char *filename,char *name)
{
	int i,len;

	len=strlen(filename);
	name[0] = 0;
	if (len<=0)	return -1;
	
	for (i=len-1;i>=0;i--)
		if (	(filename[i]=='\\') ||
			(filename[i]=='/') )	break;

	strcpy(name,filename+i+1);
	return 0;
}

int	strup(char *str)
{
	int	i=0;
	int	toupper(int i);

	while (str[i]!=0) {
		str[i]=(char)toupper((int)str[i]);
		i++;
	}
	return 0;
}


/**************************************************************************/
/******************** for reading files ***********************************/
/**************************************************************************/

char	read_char (FILE * f)
{
	int    i;

	fscanf (f, "%d", &i);
	return (char)i;
}

int	read_bool (FILE * f,int *error)
{
	char   str[MAX_STRING_SIZE];

	fscanf (f, "%s", str);
	if (strcmp (str, "TRUE") == 0) {
		*error = FALSE;
		return TRUE;
	} else
	if (strcmp (str, "FALSE") == 0) {
		*error = FALSE;
		return FALSE;
	} else {
		errout("Found no BOOL-Var in this line:\n");
		errout("%s\n",str);

		*error = TRUE;
		return 255;
	}
}

int	read_int (FILE * f)
{
	int    i;
	fscanf(f,"%d",&i);
	return i;
}

double	read_double (FILE * f)
{
	double x;
#if PRECISION == DOUBLE
	fscanf(f,"%le",&x);
#else
	fscanf(f,"%e",&x);
#endif
	return x;
}

char	read_char_skip (FILE * f)
{
	int    i;
	int    ch;

	fscanf (f, "%d", &i);
	while (((ch = fgetc (f)) != EOF) && (ch != ((int)'\n'))) {}
	return (char)i;
}

int	read_bool_skip(FILE *f)
{
	char   str[MAX_STRING_SIZE];
	int    i;
	int    ch;

	fscanf(f,"%s",str);
	if (strcmp(str,"TRUE")==0) {
		i = TRUE;
	} else
	if (strcmp(str,"FALSE")==0) {
		i = FALSE;
	} else {
		errout("Fatal error: Found no BOOL-Var in this line:\n");
		errout("%s\n",str);
		i = -1;
	}
	while (((ch = fgetc (f)) != EOF) && (ch != ((int)'\n'))) {}
	return i;
}

int	read_int_skip(FILE *f)
{
	int    i;
	int    ch;

	fscanf (f, "%d", &i);
	while (((ch = fgetc (f)) != EOF) && (ch != ((int)'\n'))) {}
	return i;
}

double	read_double_skip (FILE *f)
{
	double x;
	int    ch;

#if PRECISION == DOUBLE
	fscanf(f,"%le",&x);
#else
	fscanf(f,"%e",&x);
#endif
	while (((ch=fgetc (f)) != EOF) && (ch != ((int)'\n'))) {}
	return x;
}

void	skip_line (FILE *f,int n)
{
	int    i;
	int    ch;

	for (i = 0; i < n; i++) {
		while (((ch = fgetc (f)) != EOF) && (ch != ((int)'\n'))) {}
	} /* for i */
}

int	read_string(FILE *f,char *str)
{
	int	i,ch;
	i=0;	
	str[0]=0;
	do {
		ch=fgetc(f);	
		if ((ch==EOF)||(ch=='\n')) {
			str[i]=0;
			if (ch==EOF)	return -1;
			else		return 0;
		} else {
			str[i++]= (char)ch;
		}
		if (i>=MAX_STRING_SIZE-2) {
			errout("Fatal error: Line too long for string-buffer (read_string).\n");
			errout("Actual stringbuffer-size is %d.\n",MAX_STRING_SIZE);
			str[i]=0;	
			return -1;
		}
	} while (ch!=EOF);
	str[i]=0;	
	return -1;
}


/****************************************************************************/
/****************** routines for printouts and errors ***********************/
/****************************************************************************/
void set_version_string(char *version_string)
{
	strcpy(version_string,"C by M. Eichinger, H. Grubmueller and H. Heller, 1988-1999");
}

void	txtout(char *fmt,...)
{
	va_list args;
#if defined(SUNMP) 
	va_start(args);
#else
	va_start(args,fmt);
#endif
	
	vfprintf(stdout,fmt,args);
	fflush(stdout);
	
	va_end(args);
}

void	errout(char *fmt,...)
{
	va_list args;
	
#if defined(SUNMP)
	va_start(args);
#else
	va_start(args,fmt);
#endif

	vfprintf(stdout,fmt,args);
	fflush(stdout);
/*
	vfprintf(stderr,fmt,args);
	fflush(stderr);
*/
	va_end(args);
}

/************************************************************************/
/*** Just copies a file to stdout or file exept of all lines starting ***/
/*** with '!'							      ***/
/************************************************************************/
int	writeout_file(FILE *dest,char *fname)
{
	FILE *oldh;

	if ((oldh=fopen(fname,"r"))==NULL) {
		errout("Fatal error: Can not open <%s> for appending.\n",
			fname);
		return -1;	
	}
	while (read_string(oldh,sdum)!=-1) {	
		if (sdum[0]!='!') fprintf(dest,"%s\n",sdum);
	}
	fclose(oldh);
	return 0;
}

/***************************************************************************/
void write_sbound_droff(FILE *h_ctl, double *sbound_droff)
{
	int test=0;
	
	test |= fprintf(h_ctl,"%f   %f\t\tx- and x+ position SBound-planes in A.\n",
		sbound_droff[0],sbound_droff[1]);
	test |= fprintf(h_ctl,"%f   %f\t\ty- and y+ position SBound-planes in A.\n",
		sbound_droff[2],sbound_droff[3]);
	test |= fprintf(h_ctl,"%f   %f\t\tz- and z+ position SBound-planes in A.\n",
		sbound_droff[4],sbound_droff[5]);
	test |= fprintf(h_ctl,"%f   %f   %f\tCenter of spheric SBound (x,y,z).\n",
				sbound_droff[6],
				sbound_droff[7],
				sbound_droff[8]);
	test |= fprintf(h_ctl,"%f\t\t\tRadius of spheric SBound in A.\n",
				sbound_droff[9]);
	test |= fprintf(h_ctl,"%f   %f   %f\tCylindric SBound, axis || Z, (x,y,R).\n",
				sbound_droff[11],
				sbound_droff[12],
				sbound_droff[13]);
	test |= fprintf(h_ctl,"%f\tCurvature of SBound edges in A.\n",
				sbound_droff[10]);

	if (test<0) {
		errout("Fatal error: Could not write data (SBOUND) to tmp restart file.\n");
		pex(-1);
	}
	return;
}

/* default-values for lis-files */
int	fill_filename_list(LOAD_MSG_ALL_NODE *global_data)
{
	int	i;
	strcpy(global_data->ctlfile,"ctl.lis");
	strcpy(global_data->filenames[SHAKE_FILE_INDEX],		"shake.lis");
	strcpy(global_data->filenames[COORD_FILE_INDEX],		"coord.lis");
	
	for (i=1;i<=4;i++) {
		sprintf(global_data->filenames[EXCL_FILE_INDEX+i-1],"%1dexcl.lis",i);
		sprintf(global_data->filenames[EXCL_FILE_INDEX+i-1+5],"m%1dexcl.lis",i);
	}
	strcpy(global_data->filenames[EXCL_FILE_INDEX+4],	"14excl.lis");
	
	strcpy(global_data->filenames[BONDLIST_FILE_INDEX],	"bondlist.lis");
	strcpy(global_data->filenames[ANGLELIST_FILE_INDEX],	"anglist.lis");
	strcpy(global_data->filenames[DIHELIST_FILE_INDEX],	"dihelist.lis");
	strcpy(global_data->filenames[IMPRLIST_FILE_INDEX],	"imprlist.lis");
	strcpy(global_data->filenames[VDWPARA_FILE_INDEX],	"vdw.lis");
	strcpy(global_data->filenames[VDW14PARA_FILE_INDEX],	"14vdw.lis");
	strcpy(global_data->filenames[BONDPARA_FILE_INDEX],	"bondpara.lis");
	strcpy(global_data->filenames[ANGLEPARA_FILE_INDEX],	"angpara.lis");
	strcpy(global_data->filenames[DIHEPARA_FILE_INDEX],	"dihepara.lis");
	strcpy(global_data->filenames[IMPRPARA_FILE_INDEX],	"imprpara.lis");
	strcpy(global_data->filenames[HBTRAN_FILE_INDEX],	"hbtrlist.lis");
	strcpy(global_data->filenames[HBACC_FILE_INDEX],		"hbacclis.lis");
	strcpy(global_data->filenames[HBDON_FILE_INDEX],		"hbdonlis.lis");
	strcpy(global_data->filenames[HBAH_FILE_INDEX],		"hbahmat.lis");
	strcpy(global_data->filenames[HBBH_FILE_INDEX],		"hbbhmat.lis");
	strcpy(global_data->filenames[HBHT_FILE_INDEX],		"hbhtype.lis");
	strcpy(global_data->filenames[PGA_FILE_INDEX],		"load.lis");
	strcpy(global_data->filenames[TYPE_FILE_INDEX],		"typlist.lis");
	strcpy(global_data->filenames[UNIT_FILE_INDEX],		"units.lis");
	strcpy(global_data->filenames[FLOOD_FILE_INDEX],		"flooding.lis");
	strcpy(global_data->filenames[RESTART_FILE_INDEX],	"restart.lis");
	strcpy(global_data->filenames[FORCE_FILE_INDEX],		"forceout.lis");
	strcpy(global_data->filenames[BOUNDARY_FILE_INDEX],	"boundary.lis");
	strcpy(global_data->filenames[SBOUND_FILE_INDEX],	"sbound.lis");
	return 0;
}

void fill_dummy_ctl_data(LOAD_MSG_ALL_NODE *global_data)
{	int i;
	global_data->total_no_atoms	= 0;
	global_data->writeout_nr_atoms	= global_data->total_no_atoms;
	global_data->no_nodes		= 1;
	global_data->Analyse_nth	= 100;
	global_data->Restart_nth	= 10*global_data->Analyse_nth;
	global_data->step_of_energy	= 1;
	global_data->syscallstride 	= 0;
	strcpy(global_data->syscallstring,"ego2crd.sh");
	strcpy(global_data->writeout_atom_selection,"A*"); /*** All atoms ! ***/
	global_data->final_cycle_no	= (int)10000000;
	global_data->time_step		= 1.0000E-15/TIMEDIV;/**** in seconds ==> 1 fs ****/
	global_data->ex_number		= 5;
	global_data->minimisation	= 0;
	global_data->fric_coeff		= 1.0;
	global_data->max_move		= 0.08;		/**** in Angstroem ****/
	global_data->equilibration	= 0;
	global_data->target_temp	= 300.0;
	global_data->tau_T		= 1.0000E-13;	
	global_data->nr_distclasses	= 8;
	strcpy(global_data->outpath,"out/");
	global_data->elec14		= 0.4;
	global_data->stochastic		= 0;
	global_data->harmonic		= 0;
	global_data->shake		= 1;
	global_data->flooding		= 0;
	global_data->rot_correction	= 0;
	global_data->flag_adaptive_flood= 0;
	global_data->flood_energy_ini	= 0.0;
	global_data->flood_energy_final	= 0.0;
	global_data->flood_energy_inc  	= 0.0;
	global_data->flood_time_const	= 1.0000E-12;

	/*** Parameters for Pulling ***/
	global_data->flag_selimob	= 0;	/*** Default false	***/
	strcpy(global_data->selimob_str[0],"A*");	/*** Selection string 	***/
	strcpy(global_data->selimob_str[1],"A*");	/*** Selection string 	***/
	strcpy(global_data->selimob_str[2],"A*");	/*** Selection string 	***/
	global_data->selimob_start	= 0;	/*** Start integr. step ***/
	global_data->selimob_const	= 20.0;	/*** Used by Berthold	***/

	global_data->flag_pull		= 0;	/*** Default false	***/
	strcpy((char *)global_data->pull_str,"A1");	/*** The atom to pull	***/
	global_data->pull_vel		= 0.05;	/*** in A/ps		***/
	global_data->pull_dir[0]		= 0.0;	/*** x-Direction 	***/	
	global_data->pull_dir[1]		= 0.0;	/*** y-Direction 	***/	
	global_data->pull_dir[2]		= 1.0;	/*** z-Direction 	***/
	global_data->pull_const[0]	= 0.0;	/*** F.const in pull dir***/
	global_data->pull_const[1]	= 0.0;	/*** F.const in pull dir***/

	global_data->num_userints	= 0;
	global_data->num_userdoubles	= 0;	
	for (i=0;i<MXNO_USERINTS;i++)    global_data->str_userint[i][0]=0;
	for (i=0;i<MXNO_USERDOUBLES;i++) global_data->str_userdouble[i][0]=0;

	global_data->treeinfo.algo 	= 0;	/* default cluster-alg is martinez-schulten */
	global_data->treeinfo.steps	= 500;

	global_data->sbound_fbetagradient= 0.0;
	global_data->sbound_layerthick	= 0.0;
	/*** planes ***/
	global_data->sbound_droff[0] = 0.0;		
	global_data->sbound_droff[1] = 0.0;
	global_data->sbound_droff[2] = 0.0;		
	global_data->sbound_droff[3] = 0.0;
	global_data->sbound_droff[4] = 0.0;		
	global_data->sbound_droff[5] = 0.0;
	/*** sphere x,y,z,R ***/
	global_data->sbound_droff[6] = 0.0;	
	global_data->sbound_droff[7] = 0.0;	
	global_data->sbound_droff[8] = 0.0;
	global_data->sbound_droff[9] = 0.0;
	/*** cylinder x,y,R ***/
	global_data->sbound_droff[11] = 0.0;	
	global_data->sbound_droff[12] = 0.0;	
	global_data->sbound_droff[13] = 0.0;
	/*** curvature ***/
	global_data->sbound_droff[10]= 0.0;
	/*** constant pressure mode ***/
	global_data->keep_press_const = 0; /* 0=const volume, 1=const press, 2=const density */
	global_data->outside_pressure.plane_planar[0] = 1.0;
	global_data->outside_pressure.plane_planar[2] = 1.0;
	global_data->outside_pressure.plane_planar[4] = 1.0;
	
}

/***************************************************************************/
/*** This function is used by xpl2lis and by EGO (Screen, Outputfiles)   ***/
/***************************************************************************/
int	writeout_ctlfile(FILE *h_ctl,
			LOAD_MSG_ALL_NODE *global_data)
{
	int	i;
	FILE *h_force;
	static	char	boolstr[2][10];
	
	strcpy(boolstr[0],"FALSE");
	strcpy(boolstr[1],"TRUE");
	
	if (h_ctl!=stdout) {
		fprintf(h_ctl,"%d\t\tNumber of files.\n",CUR_LISFILES);
		for (i=0;i<CUR_LISFILES;i++) {
			extract_fname(global_data->filenames[i],sdum);
			strcat(sdum,"\t\t\n");	
			fprintf(h_ctl,sdum);
		}
	}

	fprintf(h_ctl,"%d\t\tRequested number of nodes.\n",
		global_data->no_nodes);
	fprintf(h_ctl,"%d\t\tNumber of atoms.\n",
		global_data->total_no_atoms);
	fprintf(h_ctl,"%d\t\tFreq. of analysis printout; next line is atom selection string:\n",
		global_data->Analyse_nth);
	fprintf(h_ctl,"%s\n",global_data->writeout_atom_selection);
	fprintf(h_ctl,"%d\t\tFrequency of writing restart file every analysis step!\n",
		global_data->Restart_nth/global_data->Analyse_nth);
	fprintf(h_ctl,"%d\t\tFrequency of system call every analysis step.\n",
		global_data->syscallstride);
	fprintf(h_ctl,"%s\n",global_data->syscallstring);
	fprintf(h_ctl,"%d\t\tFrequency of energy printout.\n",
		global_data->step_of_energy);
	fprintf(h_ctl,"%d\tNumber of integration steps.\n",
		global_data->final_cycle_no);
	fprintf(h_ctl,"%.8g\t\tIntegration time step in s.\n",
		global_data->time_step*TIMEDIV);
	fprintf(h_ctl,"%d\t\tOrder of exclusion list.\n",
		global_data->ex_number);
	fprintf(h_ctl,"%s\t\tSwitch for Minimisation.\n",
		boolstr[global_data->minimisation]);
	fprintf(h_ctl,"%.8g\t\tFriction factor (1.0=>no friction, 0.0=>no motion).\n",
		global_data->fric_coeff);
	fprintf(h_ctl,"%.8g\t\tMaximum position movement per integration step in A.\n",
		global_data->max_move);
	fprintf(h_ctl,"%s\t\tSwitch for Equilibration.\n",
		boolstr[global_data->equilibration]);
	fprintf(h_ctl,"%.8g\t\tTarget temperature in Kelvin.\n",
		global_data->target_temp);
	fprintf(h_ctl,"%.8g\t\tCoupling time constant in s.\n",
		global_data->tau_T);
	fprintf(h_ctl,"%d\t\tNumber of distance classes.\n",
		global_data->nr_distclasses);

	for (i=0;i<global_data->nr_distclasses-1;i++) {
		fprintf(h_ctl,"%f	% 2d\t\t\n",
			global_data->outer_classrad[i],
			global_data->treeinfo.clusterseq[i]);
	}
	/*********************************************************************/
	/******************* CLUSTERING **************************************/
	/*********************************************************************/
	fprintf(h_ctl,"%d\t\tUsed type of cluster algorithm (0 is default).\n",
		global_data->treeinfo.algo);
	fprintf(h_ctl,"%d\t\tNumber of hierarchy levels.\n",
		global_data->treeinfo.levels);
	fprintf(h_ctl,"%d\t\tNumber of branches on last level.\n",
		global_data->treeinfo.branches);
	fprintf(h_ctl,"%d\t\tFrequency of reclustering.\n",
		global_data->treeinfo.steps);

	for (i=strlen(global_data->outpath)-2;i>=0;i--)
		if (	(global_data->outpath[i]=='\\') ||
			(global_data->outpath[i]=='/') )	break;
	fprintf(h_ctl,"%s\t\tPath for output data.\n",
		global_data->outpath+i+1);

	fprintf(h_ctl,"%g\t\tScaling factor for special 1-4 electrostatic damping.\n",
		global_data->elec14);
	fprintf(h_ctl,"%s\t\tSwitch for stochastic boundary.\n",
		boolstr[global_data->stochastic]);
	fprintf(h_ctl,"%s\t\tSwitch for harmonic restraints.\n",
		boolstr[global_data->harmonic]);
	fprintf(h_ctl,"%s\t\tSwitch for SHAKE on hydrogens (Only bond length).\n",
			boolstr[global_data->shake]);

	/*********************************************************************/
	/*********************** SBOUND **************************************/
	/*********************************************************************/
	write_sbound_droff(h_ctl,global_data->sbound_droff);
	fprintf(h_ctl,"%d\t\tConstant volume(=0)/pressure(=1)/density(=2)/special(=3) mode.\n",
			global_data->keep_press_const);
	fprintf(h_ctl,"%f %f %f\tOutside press./ideal density: X, Y, Z. 0.0=ignore.\n",
			global_data->outside_pressure.plane_planar[0],
			global_data->outside_pressure.plane_planar[2],
			global_data->outside_pressure.plane_planar[4]);
	fprintf(h_ctl,"%f\tAdditional stochastic boundary thickness in A.\n",
			global_data->sbound_layerthick);
	fprintf(h_ctl,"%f\tMaximum friction coefficient in 1/ps.\n",
			global_data->sbound_fbetagradient);

	/*********************************************************************/
	/******************** FLOODING ***************************************/
	/*********************************************************************/
	fprintf(h_ctl,"%s\t\tSwitch for using flooding.\n",
			boolstr[global_data->flooding]);
	fprintf(h_ctl,"%s\t\tSwitch for using transl/rotation correction.\n",
			boolstr[global_data->rot_correction]);
	fprintf(h_ctl,"%s\t\tSwitch for using adaptive flooding.\n",
			boolstr[global_data->flag_adaptive_flood]);
	fprintf(h_ctl,"%.8g\t\tInitial energy for flooding in kT.\n",
			global_data->flood_energy_ini);
	fprintf(h_ctl,"%.8g\t\tFinal energy for flooding in kT.\n",
			global_data->flood_energy_final);
	fprintf(h_ctl,"%.8g\t\tFlooding energy increase in kT/ps.\n",
			global_data->flood_energy_inc);
	fprintf(h_ctl,"%.8g\t\tTime constant for adaptive flooding in s.\n",
			global_data->flood_time_const);

	/*********************************************************************/
	/************* User defined Vars *************************************/
	/*********************************************************************/
	fprintf(h_ctl,"%d %d\t\tNumber of user defined integers and doubles.\n",
			global_data->num_userints,
			global_data->num_userdoubles);

	for (i=0;i<global_data->num_userints;i++) {
		fprintf(h_ctl,"%d %s\n",global_data->userint[i],global_data->str_userint[i]);
	}
	for (i=0;i<global_data->num_userdoubles;i++) {
		fprintf(h_ctl,"%.8g %s\n",global_data->userdouble[i],global_data->str_userdouble[i]);
	}

	/*********************************************************************/
	/************* Parameters for Pulling ********************************/
	/*********************************************************************/
	fprintf(h_ctl,"%s\t\tSwitch for using immobilisation. Next line is atom sel. string:\n",
		boolstr[global_data->flag_selimob]);
	fprintf(h_ctl,"%s\n",global_data->selimob_str[0]);/* Change by HH (added [0]). Not sure if correct! */
	fprintf(h_ctl,"%d\t\tStart integration step for flooding/immobilisation/pulling.\n",
		global_data->selimob_start);
	fprintf(h_ctl,"%.8g\t\tHarmonic constant for immobilisation in kcal/(mol A^2).\n",
		global_data->selimob_const);

	fprintf(h_ctl,"%d\t\tPulling mode (0=OFF,1=Pulling,2=Stepping,3=Pulling two atoms in opposite direction). Next line is atom sel. string:\n",
		global_data->flag_pull);
	fprintf(h_ctl,"%s\n",global_data->pull_str[0]);
	fprintf(h_ctl,"%s\n",global_data->pull_str[1]);
	fprintf(h_ctl,"%.8g\t\tSpeed of pulling in A/ps.\n",global_data->pull_vel);
	fprintf(h_ctl,"%f   %f   %f\tPulling direction (dx,dy,dz).\n",
		global_data->pull_dir[0],
		global_data->pull_dir[1],
		global_data->pull_dir[2]);
	fprintf(h_ctl,"%f   %f \t\tSpring-constants (||,T) in kcal/(mol A^2).\n",
		global_data->pull_const[0],
		global_data->pull_const[1]);

	if (fprintf(h_ctl,"END\n")==-1)	goto error_ctlfile;

	/*********************************************************************/
	/************* FLAGS FOR DEBUGING ************************************/
	/*********************************************************************/
	fprintf(h_ctl,"%d\t\t%s\n",0,"DEBUG: 1 == compare with exact forces");
	fprintf(h_ctl,"%d\t\t%s\n",0,"DEBUG: only should be used by a developer.");

	if ((h_force=fopen(global_data->ctlfile,"r"))==NULL) {
		return 0;
	}
	
	while (!feof(h_force)) {
		read_string(h_force,sdum);

		if (strcmp(sdum,"END")==0) {
			read_string(h_force,sdum);
			read_string(h_force,sdum);		
			break;
		}
	}
	
	while (!feof(h_force)) {
		read_string(h_force,sdum);
		fprintf(h_ctl,"%s\n",sdum);
	}
	
	fclose(h_force);

	return 0;
error_ctlfile:
	errout("Can not write control-file data.");
	return -1;
}

int
writeout_coordfile(int nr_of_atoms,char *coordfn,STRUCT_ATOMDATA *atom_buf)
{
	FILE	*h_coord;
	int i;

	if ((h_coord=fopen(coordfn,"w"))==NULL) {
		errout("Can not open coordinate file <%s>.\n",coordfn);
		return -1;
	}

	fprintf(h_coord,"The data are in the following fixed format:\n");
	fprintf(h_coord,"AtomNr Type ResdNr  AtomS ResdS  charge   beta harmon         x          y          z          vx         vy         vz        \n");

	for (i=0;i<nr_of_atoms;i++) {
		fprintf(h_coord,"%-6d %-3d  %-6d  %-5s %-5s %# 7.4f % 6g % 6g % 10g % 10g % 10g % 10g % 10g % 10g\n",
			atom_buf[i].atomnr,
			atom_buf[i].atomtype,
			atom_buf[i].residnr,
			atom_buf[i].atomstr,
			atom_buf[i].residstr,
			atom_buf[i].charge,
			atom_buf[i].beta,
			atom_buf[i].harmon,
			atom_buf[i].ko[0],
			atom_buf[i].ko[1],
			atom_buf[i].ko[2],
			atom_buf[i].vel[0],
			atom_buf[i].vel[1],
			atom_buf[i].vel[2]);

	}

	fprintf(h_coord,"Next integration step number:  0\n");
	fprintf(h_coord,"Next file number: 0\n");
	fclose(h_coord);
	txtout("Coordinate file <%s> written.\n",coordfn);
	return 0;	
}

/***************************************************************************/

int read_coord_file(    LOAD_MSG_ALL_NODE *global_data,
		        int nr_of_atoms,
			char *coord_filen,
			STRUCT_ATOMDATA *atom_data)
{	
	FILE	*h_coordfile;
	int	i,old_num;
						/* open my coord-file */
	if ((h_coordfile=fopen(coord_filen,"r"))==NULL) {
		errout("Fatal error: Can not open coord-file <%s>\n",coord_filen);
		return -1;
	}	
						/* skip header */
	skip_line(h_coordfile,2);

	for (old_num=0;old_num<nr_of_atoms;old_num++) {
		i=old_num;

#if PRECISION == DOUBLE
		if (fscanf(h_coordfile,"%d %d %d %s %s %lf %lf %lf %lf %lf %lf %lf %lf %lf",
#else
		if (fscanf(h_coordfile,"%d %d %d %s %s %f %f %f %f %f %f %f %f %f",
#endif
			&atom_data[i].atomnr,
			&atom_data[i].atomtype,
			&atom_data[i].residnr,
			atom_data[i].atomstr,	
			atom_data[i].residstr,
			&atom_data[i].charge,
			&atom_data[i].beta,
			&atom_data[i].harmon,
			atom_data[i].ko+0,
			atom_data[i].ko+1,
			atom_data[i].ko+2,
			atom_data[i].vel+0,
			atom_data[i].vel+1,
			atom_data[i].vel+2)!=14) {
			errout("Fatal error: Illegal line (%d) in coord-file.\n",i+3);
			errout("This error may occur if you use coordinate files which have\n");
			errout("been created with older versions of xpl2lis! Use the latest\n");
			errout("version of xpl2lis included in your current distribution.\n");
			fclose(h_coordfile);
			return -1;
		}

		atom_data[i].refp[0] = atom_data[i].ko[0];
		atom_data[i].refp[1] = atom_data[i].ko[1];
		atom_data[i].refp[2] = atom_data[i].ko[2];

		if (atom_data[i].atomnr!=old_num) {
			errout("Fatal error: Numbering of lines is illegal in line %d.\n",i);
			fclose(h_coordfile);
			return -1;
		}
		if (strcmp(atom_data[i].residstr,"TIP3")==0) {
		  global_data->total_h2o_atoms++;
		} else {
		  global_data->total_protein_atoms++;
		}
	}
	fclose(h_coordfile);
	return 0;
}

void
generate_egofname(char *newname,char *path,int number)
{
	char temp[30];
	int len;

	sprintf(temp,"0000000000%d.ego",number);
	len = strlen(temp);	
	strcpy(newname,path);
	strcat(newname,temp+len-12);
}

/***************************************************************************************/
/***************************************************************************************/
/********
	This logical function matches the word in flstr agains a pattern
	containing wild cards (wcstr). Spaces at the end don't matter.
	If a positive match occurs, TRUE==1 is returned, otherwise FALSE==0

	The wildcards (a la X-PLOR) are:
	* matches any string of arbitrary length
	% matches any single character
	# matches any string of digits of arbitrary length
	+ matches any single digit

	You can specify a flag wild_cards_allowed, for switching on wild_card-compare or not.
*********/
static  int     does_it_match(char *flstr,char *wcstr);
static  int     does_singlechar_match(char c1,char c2);
static  int     singlechar_match_nedded(char c);
static  int     is_wildcard(char i);

#define ANY_STRING	'*'
#define ANY_CHAR	'%'
#define ANY_NUMBER	'#'
#define ANY_DIGIT	'+'

int	xplor_strcmp(char *flstr,char *wcstr,int wild_cards_allowed)
{
	int	i;

	for (i=0;i<(int)strlen(flstr);i++) if (flstr[i]==' ') flstr[i]=0;	/* truncate spaces */

	for (i=0;i<(int)strlen(wcstr);i++) if (wcstr[i]==' ') wcstr[i]=0;	/* truncate spaces */

	/******* some basic checks **********************/
	if (!wild_cards_allowed) {
		for (i=0;i<(int)strlen(wcstr);i++) {
			if (is_wildcard(wcstr[i]))	return 0;
		}
	}

	for (i=0;i<(int)strlen(flstr);i++) {
		if (is_wildcard(flstr[i])) {
			errout("Fatal error: Wild cards are not allowed in both strings !\n");
			return -1;
		}
	}

	return does_it_match(flstr,wcstr);
}

static
int	does_it_match(char *flstr,char *wcstr)
{
	int i;
	
	if (	(flstr[0]==0) && (wcstr[0]==0)	)	return 1;

	if (	wcstr[0]==0	)			return 0;

	if (	singlechar_match_nedded(wcstr[0])	) {

		if (does_singlechar_match(flstr[0],wcstr[0])) {
			return does_it_match(flstr+1,wcstr+1);			
		} else {
			return 0;
		}

	} else {
		for (i=0;i<=(int)strlen(flstr);i++) {
			if (wcstr[0]==ANY_NUMBER) {
				if (i==0)	continue;	/* no char skipped */
				if ((flstr[i-1]<'0')||(flstr[i-1]>'9')) break;
			}
			if (does_it_match(flstr+i,wcstr+1))	return 1;
		}
		return 0;
	}
}

static
int	does_singlechar_match(char c1,char c2)
{
	if ( c2==ANY_CHAR)	return 1;
	if ((c2==ANY_DIGIT) && (c1>='0') && (c1<='9')) return 1;
	if (c1==c2)	return 1;

	return 0;	/* does not match */
}

static
int	singlechar_match_nedded(char c)
{
	if ((c==ANY_NUMBER)||(c==ANY_STRING))	return 0;	/* no  */
	else			return 1;			/* yes */
}

static
int	is_wildcard(char i)
{
	if (i==ANY_STRING)	return 1;
	if (i==ANY_NUMBER)	return 1;
	if (i==ANY_CHAR)	return 1;
	if (i==ANY_DIGIT)	return 1;

	return 0;	
}


/*** Distance-criterion '+DA34<23.34' selects all atoms   ***/
/*** closer than 23.34 A from atom 34. Also '>' possible. ***/
/*** Distance-criterion '+DR34<23.34' selects all residues***/ 
/*** in which at least one atom is closer than            ***/
int	set_used_atoms(	char *str,int nr_of_atoms,
			STRUCT_ATOMDATA *atom,
			int *use_this_atom)
{
	int i,j;
	int starti;
	int useit_flag;
	int atom_flag=0,residue_flag=0,distance_flag=0;
	int startnumber=0,offset=1;
	int endnumber=0;
	char tempstr[500];
	char numberstr[500];
	double dist=0.0;

	starti = 0;
        while (starti<strlen(str)) {
          for (i=starti,j=0;i<strlen(str);i++) {
		if (str[i]==' ') break;
		tempstr[j++] = str[i];
          }
          tempstr[j] = 0;
          starti = starti + j+1;
	  strup(tempstr);
	  if (tempstr[0]==0) return 0;
	  if (	(tempstr[0]!='A') && (tempstr[0]!='R') && 
		(tempstr[0]!='+') && (tempstr[0]!='-') &&
		(tempstr[0]!='D')
	      ) {
		errout("Syntax error in selection string: +,-,A, R or D expected.\n");
		return -1;
	  }
	  useit_flag=1; 
          if ((tempstr[0]=='+') || (tempstr[0]=='-')) {
	    	if (tempstr[0]=='-') useit_flag=0;
	    	for (i=0;i<strlen(tempstr);i++) tempstr[i] = tempstr[i+1];
	  }
	  if (	(tempstr[0]!='A') && 
		(tempstr[0]!='R') &&
		(tempstr[0]!='D') ) {
	    	errout("Syntax error in selection string: A, R or D expected.\n");
	    	return -1;
	  }
	  atom_flag    =0;
	  residue_flag =0;
	  distance_flag=0;
	  offset=1;
	  if (tempstr[0]=='A') {   /* consider _atom_ names/numbers */
	  	  atom_flag=1;
	  }
	  if (tempstr[0]=='R') {   /* consider _residue_ names/numbers */
	  	  residue_flag=1;
	  }
	  if (tempstr[0]=='D') {   /* consider distance criterion */
		distance_flag=1;
	        offset=2;
	  	if (tempstr[1]=='A') {   /* consider _atom_ names/numbers */
	  		  atom_flag=1;
	  	} else
	  	if (tempstr[1]=='R') {   /* consider _residue_ names/numbers */
	  		  residue_flag=1;
	  	} else {
	    		errout("Syntax error in selection string: A, R expected.\n");
		}
	  }

	  for (i=0;i<strlen(tempstr);i++) tempstr[i]=tempstr[i+offset];
	  if ((tempstr[0]>='0') && (tempstr[0]<='9')) { 
		/*** consider atom/residue _numbers ***/
		startnumber=0;
		endnumber=0;
	   	for (i=0;i<strlen(tempstr);i++) {
	     		if (	(tempstr[i]=='-') ||
				(tempstr[i]=='<') ||
				(tempstr[i]=='>') ) break;
	      		numberstr[i] = tempstr[i];
	      	}
	    	numberstr[i] = 0;
            	startnumber=atoi(numberstr);
    
	       	endnumber=startnumber;
		if (distance_flag!=0) {
			if (tempstr[i]=='<') distance_flag=-1;
			else
			if (tempstr[i]=='>') distance_flag=1;
			else {
				errout("Syntax error in selection string: '<' or '>' expected.\n");
	    			return -1;
			}
			dist = -1.0;
#if PRECISION == DOUBLE
			if ((sscanf(tempstr+i+1,"%lf",&dist)!=1) ||
#else
			if ((sscanf(tempstr+i+1,"%f",&dist)!=1) ||
#endif
			    (dist<0.0) ) {
				errout("Syntax error in selection string: distance must be >0.\n");
	    			return -1;
			}	
		} else {
		    	if (tempstr[i]=='-') {
				for (i++,j=0; i<strlen(tempstr);i++) {
		      			numberstr[j++] = tempstr[i];
		        	}
		      		numberstr[j] = 0;
	              		endnumber=atoi(numberstr);
	   	        }
		}
	 	for (i=0;i<nr_of_atoms;i++) {
	     		if ((atom_flag)&&(distance_flag==0)) {
	        		if (	(atom[i].atomnr+1>=startnumber) && 
					(atom[i].atomnr+1<=endnumber)) 
  		  			use_this_atom[i] = useit_flag;
  	        	} else
  	     	 	if ((residue_flag)&&(distance_flag==0)) {
	        		if (	(atom[i].residnr>=startnumber) && 
					(atom[i].residnr<=endnumber)) 
		  			use_this_atom[i] = useit_flag;
	        	} else
			if (distance_flag!=0) {
				double dx=atom[i].ko[0]-atom[startnumber].ko[0];
				double dy=atom[i].ko[1]-atom[startnumber].ko[1];
				double dz=atom[i].ko[2]-atom[startnumber].ko[2];
				double d =sqrt(dx*dx+dy*dy+dz*dz);

				if (i==startnumber) continue;
		
				if (	(distance_flag==-1) &&
					(d<dist) ) {
					if (atom_flag) 	use_this_atom[i] = useit_flag;
					else {
						int resnr=atom[i].residnr;
						for (j=0;j<nr_of_atoms;j++)
							if (resnr==atom[j].residnr) 
								use_this_atom[j] = useit_flag; 		
					}
				} else
				if (	(distance_flag==1) &&
					(d>dist) ) {
					if (atom_flag) 	use_this_atom[i] = useit_flag;
					else {
						int resnr=atom[i].residnr;
						for (j=0;j<nr_of_atoms;j++)
							if (resnr==atom[j].residnr) 
								use_this_atom[j] = useit_flag; 	
					}
				}
			}
	      	}
	  } else {                        /* consider atom/residue _names_ */
	    	for (i=0;i<nr_of_atoms;i++) {
	      		if (atom_flag) {
                		if (xplor_strcmp(atom[i].atomstr,tempstr,1))
		  			use_this_atom[i] = useit_flag;
	        	}
	      		if (residue_flag) {
                		if (xplor_strcmp(atom[i].residstr,tempstr,1))
		  			use_this_atom[i] = useit_flag;
	        	}
	      }
	 }
    }
	return 0;
}

int set_cluster_params(int num_atoms,double *distvals,int *clusterseq)
{
	double	m=4.8+2.0,c=4.3;
	int	max_levels=1;

	distvals[0] = 6.0;	
	distvals[1] = 9.5;		
	distvals[2] = 15.0;		
	distvals[3] = 24.0;		
	distvals[4] = 38.0;		
	distvals[5] = 60.0;		
	distvals[6] = 96.0;		

	clusterseq[0] = 0;
	clusterseq[1] = 6;
	clusterseq[2] = 4;
	clusterseq[3] = 4;
	clusterseq[4] = 4;
	clusterseq[5] = 4;
	clusterseq[6] = 4;

	if (num_atoms<100) {
		clusterseq[1] = 1;
		return 1;
	} else
	if (num_atoms<500) {
		clusterseq[1] = 4;
		return 2;
	} else
	if (num_atoms<1000) {
		clusterseq[1] = 6;
		return 2;
	} else {
		max_levels = 1 + (int) (log(2.0*num_atoms/(m*m*m))/log(c));
	}
	return max_levels;
}

/************************************************************************/

int	start_bbox(double *rect)
{
	rect[0] = -9999.0;	rect[1] = -9999.0;
	rect[2] = -9999.0;	rect[3] = -9999.0;
	rect[4] = -9999.0;	rect[5] = -9999.0;
	return 0;
}

int	add_atom2bbox(double x,double y,double z,double *rect)
{

	if (rect[0]==-9999.0) {
		rect[0]=x;
		rect[1]=x;

		rect[2]=y;
		rect[3]=y;

		rect[4]=z;
		rect[5]=z;
		return 0;
	}
	
	if (rect[0]>x) rect[0]=x;
	if (rect[1]<x) rect[1]=x;

	if (rect[2]>y) rect[2]=y;
	if (rect[3]<y) rect[3]=y;

	if (rect[4]>z) rect[4]=z;
	if (rect[5]<z) rect[5]=z;
	return 0;
}

int	merge_bbox(double *box1,double *box2,double *newbox)
{
	if (box1[0]==-9999.0) {
		memcpy(newbox,box2,6*sizeof(double));
	} else
	if (box2[0]==-9999.0) {
		memcpy(newbox,box1,6*sizeof(double));
	} else {
		newbox[0]=min(box1[0],box2[0]);
		newbox[1]=max(box1[1],box2[1]);

		newbox[2]=min(box1[2],box2[2]);
		newbox[3]=max(box1[3],box2[3]);

		newbox[4]=min(box1[4],box2[4]);
		newbox[5]=max(box1[5],box2[5]);
	}

	return 0;
}

double	maxext_bbox(double *box)
{
	double dx=box[1]-box[0];
	double dy=box[3]-box[2];
	double dz=box[5]-box[4];
	
	return sqrt(dx*dx + dy*dy + dz*dz)/2.0;
}

double	size_bbox(double *box)
{
	double dx=box[1]-box[0];
	double dy=box[3]-box[2];
	double dz=box[5]-box[4];
	
	return max(dx,max(dy,dz));
}


int	bbox2cenrad(double *box,double *cenrad)
{
	double xmin,xmax,sum,dx;

	xmin = box[0];	xmax = box[1];
	cenrad[0] = (xmin+xmax)/2.0;
	dx = xmax-xmin;
	sum = dx*dx;

	xmin = box[2];	xmax = box[3];
	cenrad[1] = (xmin+xmax)/2.0;
	dx = xmax-xmin;
	sum += dx*dx;
	
	xmin = box[4];	xmax = box[5];
	cenrad[2] = (xmin+xmax)/2.0;
	dx = xmax-xmin;
	sum += dx*dx;
	cenrad[3] = (double) (sqrt(sum)/2.0);
	return 0;
}

int	is_in_bbox(double *ko,double *bbox)
{
	if (ko[0]<bbox[0]) return 0;
	if (ko[0]>bbox[1]) return 0;

	if (ko[1]<bbox[2]) return 0;
	if (ko[1]>bbox[3]) return 0;

	if (ko[2]<bbox[4]) return 0;
	if (ko[2]>bbox[5]) return 0;

	return 1;
}

/************************************************************************/

int set_signal_procs(void (*signal_handler)(int signo))
{
#if defined(SEQ_NANO) || defined(PARIX) || defined(PARIXHAL) || defined(HPPA)

#else
	/* The following signals MUST NOT BE TOUCHED or things like 
	   MPI_Abort() will no longer work properly and we end
	   up with a big mess (e.g. on the SP2).
	   Therefore the following code is all commented out!!!
	   
	if (signal(SIGTERM,signal_handler)==SIG_ERR) {
		errout("Fatal error: can't catch SIGTERM\n");
		pex(-1);
	}	
#ifdef SIGURG
	if (signal(SIGURG,signal_handler)==SIG_ERR) {
		errout("Fatal error: can't catch SIGURG\n");
		pex(-1);
	}
#endif
	if (signal(SIGXCPU,signal_handler)==SIG_ERR) {
		errout("Fatal error: can't catch SIGXCPU (CPU time limit reached)\n");
		pex(-1);
	}
	if (signal(SIGQUIT,signal_handler)==SIG_ERR) {
		errout("Fatal error: can't catch SIGQUIT\n");
		pex(-1);
	}
	if (signal(SIGINT,signal_handler)==SIG_ERR) {
		errout("Fatal error: can't catch SIGINT\n");
		pex(-1);
	}
	if (signal(SIGHUP,signal_handler)==SIG_ERR) {
		errout("Fatal error: can't catch SIGHUP\n");
		pex(-1);
	}	
	
	End of comment block of signals that we have to leave alone!!  */
	
	/******************************************************************/
	/*** this signal should produce a programm stop after the	***/
	/*** next restart file is written				***/
	/******************************************************************/
	if (signal(SIGUSR1,signal_handler)==SIG_ERR) {
		errout("Fatal error: can't catch SIGUSR1\n");
		pex(-1);
	}
	/******************************************************************/
	/*** this signal should produce a programm stop after the	***/
	/*** current integration step has finished!			***/
	/******************************************************************/
	if (signal(SIGUSR2,signal_handler)==SIG_ERR) {
		errout("Fatal error: can't catch SIGUSR2\n");
		pex(-1);
	}
#endif
	return 0;	
}


/******************************************************************/

int
crossprod(double *a,double *b,double *c)
{
	double x[3];
	
	x[0] = a[1]*b[2] - a[2]*b[1];
	x[1] =-a[0]*b[2] + a[2]*b[0];
	x[2] = a[0]*b[1] - a[1]*b[0];

	c[0]=x[0];
	c[1]=x[1];
	c[2]=x[2];
		
	return 0;
}

double
dotprod(double *a,double *b)
{
	return (double)(a[0]*b[0]+a[1]*b[1]+a[2]*b[2]);
}

double
veclen(double *a)
{
	return sqrt(a[0]*a[0] + a[1]*a[1] + a[2]*a[2]);	
}

double
normvec(double *a,double *b)
{
	double len=sqrt(a[0]*a[0] + a[1]*a[1] + a[2]*a[2]);

	if (len<=0.00000001) return 0.0;
	
	b[0]=a[0]/len;
	b[1]=a[1]/len;
	b[2]=a[2]/len;
		
	return len;
}

/*** note matrix in num.rec. notation ***/
void
matdotvec(double **mat,double *veci,double *veco) 
{
	double v[3];
			
	v[0] = mat[1][1]*veci[0] + mat[1][2]*veci[1] + mat[1][3]*veci[2];
	v[1] = mat[2][1]*veci[0] + mat[2][2]*veci[1] + mat[2][3]*veci[2];
	v[2] = mat[3][1]*veci[0] + mat[3][2]*veci[1] + mat[3][3]*veci[2];

	veco[0]=v[0];
	veco[1]=v[1];
	veco[2]=v[2];	
}



static	double	specA=1.0, specd=0.0, specs=1.0;
static	int	specmode=0,atomid1=-1,atomid2=-1;

int parse_specforce(char *keywordstr,
		char* reststr,
		STRUCT_ATOMDATA *atom_data,
		int *atom_sele_flag1,
		LOAD_MSG_ALL_NODE *global_data,
		STRUCT_FORCES *my_force,
		int *used_atoms)
{
	/* the following don't seem to be used. Commented out by HH
	int i,j,len=strlen(reststr); */

	/*** lock for the keyword ***/
	if (strcmp(keywordstr,"SPECIALFORCE")==0) {
		
		/*** parse the six parameters ***/
		if (sscanf(reststr,"%d %d %d %lf %lf %lf",
			&specmode,
			&atomid1,&atomid2,
			&specA,&specd,&specs)!=6) {
                        errout("Fatal error: Invalid format for SPECIALFORCE\n");
                        pex(-1);
                }
                
                txtout("Applying special force between atom %d and %d\n",
                	atomid1,atomid2);
 		if (specmode==0) {
			txtout("using the function %f EXP(-(d-%f)/%f) (kcal/mol A)\n",
				specA,specd,specs);
		} else {
			errout("Fatal error: Invalid mode for SPECIALFORCE\n");
                        pex(-1);
		}
		return 1;
	}
	return 0;
}

int apply_specforce(ENERGY_INFO *energy,
                LOAD_MSG_ONE_NODE *my_data,
                LOAD_MSG_ALL_NODE *global_data,
                TRAVELLING_FORCES *my_new_forces)
{
	double  rx,ry,rz,R,F=0.0;
	int id1,id2;
	
	if ((atomid1==-1)||(atomid2==-1)) return 0; /*** do not apply special force ***/

	/*** map the indices ***/
	id1=global_data->old2new_map[atomid1-1];	
	id2=global_data->old2new_map[atomid2-1];

	/*** calculate the distance between the atoms ****/
	rx = my_data->x_new[id2] - my_data->x_new[id1];
	ry = my_data->y_new[id2] - my_data->y_new[id1];
	rz = my_data->z_new[id2] - my_data->z_new[id1];
	R  = sqrt(rx*rx + ry*ry + rz*rz);

	if (specmode==0) {
		F  = specA * exp(-(R-specd)/specs);
	} 
/*** any other special force you like
	else if (specmode==1) {
		
		
	} else ....
****/

	/*** now add the force to the atoms ***/
	/*** note: the sign of "EGO-forces" is oppsite to the usual way ***/
	/*** since the integrator is programmed that way! ***/
	my_new_forces->Fx[id1] = F*rx/R;
	my_new_forces->Fy[id1] = F*ry/R; 		
	my_new_forces->Fz[id1] = F*rz/R;
	
	my_new_forces->Fx[id2] = -F*rx/R;
	my_new_forces->Fy[id2] = -F*ry/R; 		
	my_new_forces->Fz[id2] = -F*rz/R;			

	txtout("Special force acting: %f kcal/mol A    d=%f A\n",F,R);	
			
	return 0;
}

/******************************************************************/
/******* here come some functions from numerical recipes **********/
/******************************************************************/

void nrerror(char *error_text)
{
	errout("Numerical Recipes run-time error...\n");
	errout("%s\n",error_text);
	errout("...now exiting to system...\n");
	pex(-1);
}

double *dvector(int nl, int nh)
{
	double *v;

	v=(double *)alloc_doubles((unsigned) (nh-nl+1));
	if (!v) nrerror("allocation failure in dvector()");
	return v-nl;
}

int *ivector(int nl, int nh)
{
	int *v;

	v=(int *)alloc_ints((unsigned) (nh-nl+1));
	if (!v) nrerror("allocation failure in ivector()");
	return v-nl;
}

double **dmatrix(int nrl, int nrh, int ncl, int nch)
{
	int i;
	double **m;

	m=(double **) alloc_ptrs(nrh-nrl+1);
	if (!m) nrerror("allocation failure 1 in dmatrix()");
	m -= nrl;

	for(i=nrl;i<=nrh;i++) {
		m[i]=(double *) alloc_doubles((unsigned) (nch-ncl+1));
		if (!m[i]) nrerror("allocation failure 2 in dmatrix()");
		m[i] -= ncl;
	}
	return m;
}

void free_dvector(double *v, int nl, int nh)
{
	int i=nh;i++; /*** no warning ***/
	myfree((char*) (v+nl));
}

void free_ivector(int *v, int nl, int nh)
{
	int i=nh;i++; /*** no warning ***/
	myfree((char*) (v+nl));
}

void free_dmatrix(double **m, int nrl, int nrh, int ncl, int nch)
{
	int i=nch;
	for(i=nrh;i>=nrl;i--) myfree((char*) (m[i]+ncl));
	myfree((char*) (m+nrl));
}

void lubksb(double **a,int n,int *indx,double *b)
{
	int i,ii=0,ip,j;
	double sum;

	for (i=1;i<=n;i++) {
		ip=indx[i];
		sum=b[ip];
		b[ip]=b[i];
		if (ii)
			for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j];
		else if (sum) ii=i;
		b[i]=sum;
	}
	for (i=n;i>=1;i--) {
		sum=b[i];
		for (j=i+1;j<=n;j++) sum -= a[i][j]*b[j];
		b[i]=sum/a[i][i];
	}
}

#define TINY 1.0e-20;

void ludcmp(double **a,int n,int *indx,double *d)
{
	int i,imax=0,j,k;
	double big,dum,sum,temp;
	double *vv;

	vv=dvector(1,n);
	*d=1.0;
	for (i=1;i<=n;i++) {
		big=0.0;
		for (j=1;j<=n;j++)
			if ((temp=fabs(a[i][j])) > big) big=temp;
		if (big == 0.0) nrerror("Singular matrix in routine LUDCMP");
		vv[i]=1.0/big;
	}
	for (j=1;j<=n;j++) {
		for (i=1;i<j;i++) {
			sum=a[i][j];
			for (k=1;k<i;k++) sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
		}
		big=0.0;
		for (i=j;i<=n;i++) {
			sum=a[i][j];
			for (k=1;k<j;k++)
				sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
			if ( (dum=vv[i]*fabs(sum)) >= big) {
				big=dum;
				imax=i;
			}
		}
		if (j != imax) {
			for (k=1;k<=n;k++) {
				dum=a[imax][k];
				a[imax][k]=a[j][k];
				a[j][k]=dum;
			}
			*d = -(*d);
			vv[imax]=vv[j];
		}
		indx[j]=imax;
		if (a[j][j] == 0.0) a[j][j]=TINY;
		if (j != n) {
			dum=1.0/(a[j][j]);
			for (i=j+1;i<=n;i++) a[i][j] *= dum;
		}
	}
	free_dvector(vv,1,n);
}

void svdcmp(double **a,int m,int n,double *w,double **v)
{
	int flag,i,its,j,jj,k,l=0,nm=0;
	double c,f,h,s,x,y,z;
	double anorm=0.0,g=0.0,scale=0.0;
	double *rv1;

	if (m < n) nrerror("SVDCMP: You must augment A with extra zero rows");
	rv1=dvector(1,n);
	for (i=1;i<=n;i++) {
		l=i+1;
		rv1[i]=scale*g;
		g=s=scale=0.0;
		if (i <= m) {
			for (k=i;k<=m;k++) scale += fabs(a[k][i]);
			if (scale) {
				for (k=i;k<=m;k++) {
					a[k][i] /= scale;
					s += a[k][i]*a[k][i];
				}
				f=a[i][i];
				g = -SIGN(sqrt(s),f);
				h=f*g-s;
				a[i][i]=f-g;
				if (i != n) {
					for (j=l;j<=n;j++) {
						for (s=0.0,k=i;k<=m;k++) s += a[k][i]*a[k][j];
						f=s/h;
						for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
					}
				}
				for (k=i;k<=m;k++) a[k][i] *= scale;
			}
		}		w[i]=scale*g;
		g=s=scale=0.0;
		if (i <= m && i != n) {
			for (k=l;k<=n;k++) scale += fabs(a[i][k]);
			if (scale) {
				for (k=l;k<=n;k++) {
					a[i][k] /= scale;
					s += a[i][k]*a[i][k];
				}
				f=a[i][l];
				g = -SIGN(sqrt(s),f);
				h=f*g-s;
				a[i][l]=f-g;
				for (k=l;k<=n;k++) rv1[k]=a[i][k]/h;
				if (i != m) {
					for (j=l;j<=m;j++) {
						for (s=0.0,k=l;k<=n;k++) s += a[j][k]*a[i][k];
						for (k=l;k<=n;k++) a[j][k] += s*rv1[k];
					}
				}
				for (k=l;k<=n;k++) a[i][k] *= scale;
			}
		}
		anorm=MAX(anorm,(fabs(w[i])+fabs(rv1[i])));
	}
	for (i=n;i>=1;i--) {
		if (i < n) {
			if (g) {
				for (j=l;j<=n;j++)
					v[j][i]=(a[i][j]/a[i][l])/g;
				for (j=l;j<=n;j++) {
					for (s=0.0,k=l;k<=n;k++) s += a[i][k]*v[k][j];
					for (k=l;k<=n;k++) v[k][j] += s*v[k][i];
				}
			}
			for (j=l;j<=n;j++) v[i][j]=v[j][i]=0.0;
		}
		v[i][i]=1.0;
		g=rv1[i];
		l=i;
	}
	for (i=n;i>=1;i--) {
		l=i+1;
		g=w[i];
		if (i < n)
			for (j=l;j<=n;j++) a[i][j]=0.0;
		if (g) {
			g=1.0/g;
			if (i != n) {
				for (j=l;j<=n;j++) {
					for (s=0.0,k=l;k<=m;k++) s += a[k][i]*a[k][j];
					f=(s/a[i][i])*g;
					for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
				}
			}
			for (j=i;j<=m;j++) a[j][i] *= g;
		} else {
			for (j=i;j<=m;j++) a[j][i]=0.0;
		}
		++a[i][i];
	}
	for (k=n;k>=1;k--) {
		for (its=1;its<=30;its++) {
			flag=1;
			for (l=k;l>=1;l--) {
				nm=l-1;
				if (fabs(rv1[l])+anorm == anorm) {
					flag=0;
					break;
				}
				if (fabs(w[nm])+anorm == anorm) break;
			}
			if (flag) {
				c=0.0;
				s=1.0;
				for (i=l;i<=k;i++) {
					f=s*rv1[i];
					if (fabs(f)+anorm != anorm) {
						g=w[i];
						h=PYTHAG(f,g);
						w[i]=h;
						h=1.0/h;
						c=g*h;
						s=(-f*h);
						for (j=1;j<=m;j++) {
							y=a[j][nm];
							z=a[j][i];
							a[j][nm]=y*c+z*s;
							a[j][i]=z*c-y*s;
						}
					}
				}
			}
			z=w[k];
			if (l == k) {
				if (z < 0.0) {
					w[k] = -z;
					for (j=1;j<=n;j++) v[j][k]=(-v[j][k]);
				}
				break;
			}
			if (its == 30) nrerror("No convergence in 30 SVDCMP iterations");
			x=w[l];
			nm=k-1;
			y=w[nm];
			g=rv1[nm];
			h=rv1[k];
			f=((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y);
			g=PYTHAG(f,1.0);
			f=((x-z)*(x+z)+h*((y/(f+SIGN(g,f)))-h))/x;
			c=s=1.0;
			for (j=l;j<=nm;j++) {
				i=j+1;
				g=rv1[i];
				y=w[i];
				h=s*g;
				g=c*g;
				z=PYTHAG(f,h);
				rv1[j]=z;
				c=f/z;
				s=h/z;
				f=x*c+g*s;
				g=g*c-x*s;
				h=y*s;
				y=y*c;
				for (jj=1;jj<=n;jj++) {
					x=v[jj][j];
					z=v[jj][i];
					v[jj][j]=x*c+z*s;
					v[jj][i]=z*c-x*s;
				}
				z=PYTHAG(f,h);
				w[j]=z;
				if (z) {
					z=1.0/z;
					c=f*z;
					s=h*z;
				}
				f=(c*g)+(s*y);
				x=(c*y)-(s*g);
				for (jj=1;jj<=m;jj++) {
					y=a[jj][j];
					z=a[jj][i];
					a[jj][j]=y*c+z*s;
					a[jj][i]=z*c-y*s;
				}
			}
			rv1[l]=0.0;
			rv1[k]=f;
			w[k]=x;
		}
	}
	free_dvector(rv1,1,n);
}

#undef SIGN
#undef MAX
#undef PYTHAG
#undef TINY
