/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/************************************************************************/
/*  functions defined in util_lib.c ****************************************/
/************************************************************************/

int	extract_path(char *filename,char *path);
int	extract_fname(char *filename,char *name);
int	strup(char *str);
int	set_rand_seed(long i);
double	ego_rand(void);
double	copysign (double x, double y);

void 	create_outpath(LOAD_MSG_ALL_NODE *global_data);

char	read_char(FILE *f);
int	read_bool(FILE *f,int*error);
int	read_int(FILE *f);
double	read_double(FILE *f);
char	read_char_skip(FILE *f);
int	read_bool_skip(FILE *f);
int	read_int_skip(FILE *f);
double	read_double_skip (FILE * f);
void	skip_line (FILE * f,int n);
int	read_string(FILE *f,char *str);
int	fill_filename_list(LOAD_MSG_ALL_NODE	*global_data);
void 	fill_dummy_ctl_data(LOAD_MSG_ALL_NODE *global_data);

void 	set_version_string(char *version_string);
void	txtout(char *fmt,...);
void	errout(char *fmt,...);

int	writeout_file(FILE *dest,char *fname);
void 	write_sbound_droff(FILE *file, double *sbound_droff);
int	writeout_ctlfile(FILE *h_ctl,LOAD_MSG_ALL_NODE *ctl);

int	writeout_coordfile(int nr_of_atoms,char *coordfn,
			STRUCT_ATOMDATA *atom_buf);
int	read_coord_file(LOAD_MSG_ALL_NODE *global_data,
			int nr_of_atoms,
			char *coord_filen,
			STRUCT_ATOMDATA *atom_data);

void	generate_egofname(char *newname,char *path,int number);
int	get_my_endian(void);
int	xplor_strcmp(char *flstr,char *wcstr,int wild_cards_allowed);
int	set_used_atoms(	char *str,int nr_of_atoms,
			STRUCT_ATOMDATA *atom,
			int *use_this_atom);
int	set_cluster_params(int num_atoms,double *distvals,int *clusterseq);

int	start_bbox(double *rect);
int	add_atom2bbox(double x,double y,double z,double *rect);
int	merge_bbox(double *box1,double *box2,double *newbox);
double	maxext_bbox(double *box);
int	bbox2cenrad(double *box,double *cenrad);
double	size_bbox(double *box);
int	is_in_bbox(double *ko,double *bbox);
int 	set_signal_procs(void (*signal_handler)(int signo));

int	crossprod(double *a,double *b,double *c);
double	dotprod(double *a,double *b);
double	veclen(double *a);
double	normvec(double *a,double *b);
void	matdotvec(double **mat,double *veci,double *veco);

int apply_specforce(ENERGY_INFO *energy,
                LOAD_MSG_ONE_NODE *my_data,
                LOAD_MSG_ALL_NODE *global_data,
                TRAVELLING_FORCES *my_new_forces);

void nrerror(char *error_text);
double *dvector(int nl, int nh);
int *ivector(int nl, int nh);
double **dmatrix(int nrl, int nrh, int ncl, int nch);
void free_dvector(double *v, int nl, int nh);
void free_ivector(int *v, int nl, int nh);
void free_dmatrix(double **m, int nrl, int nrh, int ncl, int nch);
void ludcmp(double **a,int n,int *indx,double *d);
void lubksb(double **a,int n,int *indx,double *b);
void svdcmp(double **a,int m,int n,double *w,double **v);
int parse_specforce(char *keywordstr,
		char* reststr,
		STRUCT_ATOMDATA *atom_data,
		int *atom_sele_flag1,
		LOAD_MSG_ALL_NODE *global_data,
		STRUCT_FORCES *my_force,
		int *used_atoms);

#ifndef ABS
#define ABS(x) ((x)>=0?(x):-(x))
#endif

#ifndef MIN
#define	MIN(a,b) (((a)<(b))?(a):(b))
#endif

#ifndef MAX
#define	MAX(a,b) (((a)>(b))?(a):(b))
#endif

#ifndef SIGN
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))
#endif

