/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*****************************************************************************
**	Belongs to: adjust_sbound.c - minimize SBOUND for a given coord.lis **									**									    **
**	Module	: adjust_sbound.c	by Helmut Heller		    **
**									    **
**      Modif's:                                                            **
**      19960309.  HH first version					    **
**									    **
**								            **
*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdarg.h>

#if defined(NEXT) || defined(RS6K) || defined(MAC)
#include <stdlib.h>
#else
#include <malloc.h>
#endif

#include "../ego.h"
#include "../util_lib.h"
#include "../egbound.h"
#include "../egcolib.h"
#include "../egalloc.h"
#include "../hardw.h"

#define DEBUG 0

int	set_QQL_anglefactor(int *atomids, double factor);
int	set_bondfactor(int *atomids, double factor);

void   pex(int i)
{
	txtout("Exit by pex(%d) in adjust_sbound.c !\n",i);
	exit(i);
	return;
}

/* Here we define some dummies to avoid including a whole slew of other files */
/****************************************************************************/
int	CLU_get_all_atoms(int level,	/** level-index: 0 -> unitlevel,...**/
			  int parent,	/** parent-index ***/
			  int *nr,	/** nr of elements for which buffer is allocated */
			  int **buf,	/** the buffer for the atom-indices (OLD-EGO) */
			  struct_cluster *cluster[],
			  int *old_atom_nrs)
{
  exit(-99);
  return 0;
}

int
parse_qminterface(      char *keywordstr,char* reststr,
                        STRUCT_ATOMDATA *atom_data,
                        int *atom_sele_flag1,
                        LOAD_MSG_ALL_NODE *global_data,
                        STRUCT_FORCES *my_force,
                        int     *used_atoms)
{
return -1;
}

int
set_qmboundbox( int *atom_sele_flag1,
                STRUCT_ATOMDATA *atom_data,
                LOAD_MSG_ALL_NODE *global_data)
{
return -1;
}


/***************************************************************************/
static  int verbose = FALSE;


/****************************************************************************/
int add_ctlfile_data(FILE *h_ctl,LOAD_MSG_ALL_NODE *global_data)
{
  exit(-99);
  return 0;
}

/****************************************************************************/
static
void	print_wrong_call()
{
        txtout("This program reads the EGO control file and varies\n");
        txtout("the 6 SBOUND planes so that the total SBOUND energy is minimal.\n\n");

        txtout("Program-Usage:\n");
        txtout("adjust_sbound  [<controlfile>] [-v]\n");
        txtout("<controlfile>          Controlfile to use (default: ctl.lis)\n");
}

/****************************************************************************/
static
int	readin_vdwfile( char *vdw_filen, 
			LOAD_MSG_ALL_NODE *global_data)
{
  FILE	*fopen();
  FILE	*h_vdwfile;
  int	i,j;
  double	read_double();
  int	nr_atomtypes=0;
  char	sdum[MAX_STRING_SIZE];		/* space for dummy-strings */

					  /* open my vdw-file */
  if ((h_vdwfile=fopen(vdw_filen,"r"))==NULL) {
	  errout("Can not open vdw-file <%s>.\n",vdw_filen);
	  return -1;
  }
					  /* skip header */
  skip_line(h_vdwfile,1);
  fscanf(h_vdwfile,"%s%s%s%s%s%d",sdum,sdum,sdum,sdum,sdum,&nr_atomtypes);
  if (nr_atomtypes<1)	return -1;

  txtout("Atom types in VDW-File : %d\n",nr_atomtypes);
					  /* get atommasses in diag */
  for (i = 0; i<=nr_atomtypes; i++) {
    for (j = 0; j<=nr_atomtypes; j++) {
      global_data->AB[i][j]=read_double(h_vdwfile);
      if (i==j) 
      {
	if ((global_data->AB[i][j]<0.0) || (global_data->AB[i][j]>100.0) ) 
	{
	  errout("Error: strange mass = %f !\n",global_data->AB[i][j]);
	}
      } 
    }
  }

  fclose(h_vdwfile);
  return 0;
}

/*****************************************************************************/
void	vout(char *fmt,...)
{
  va_list args;
#if defined(SUNMP)
        va_start(args);
#else
        va_start(args,fmt);
#endif


  if (verbose) 
  {
    vfprintf(stdout,fmt,args);
    fflush(stdout);
  }
  
  va_end(args);
}

/****************************************************************************/
void calc_vir_pressure(PRESSURE *press,double Ekin,
			 double *sbound_droff,
			 double ekin_tensor[3][3])
{
  return ;
}




/********************************************************************************/
int main(int argc,char *argv[])
{
  LOAD_MSG_ALL_NODE global_data;
  LOAD_MSG_ONE_NODE my_data;
  char	ctl_file_name[MAX_STRING_SIZE];
  STRUCT_ATOMDATA *atom_data;
  int *atom_sele_flag1;
  STRUCT_FORCES my_force;
  int error_occured = FALSE;

  
  int side, direction, master_direction;
  double energy, previous_energy;
  PRESSURE pressure_dummy, density_dummy;
  
  int	i=0;
  double sbound_increment = 0.1; /* in Angstrom */
  
  double rnd_force_std_deviation, Gamma;


  txtout("==================================================================\n");
  txtout("adjust_sbound - minimize SBOUND for a given coord.lis file.\n");
  txtout("==================================================================\n");

  if ((argc > 3)) {
	  print_wrong_call();
	  return -1;
  }


  strcpy(ctl_file_name,"ctl.lis");

  if (strcmp(argv[argc-1],"-v")==0) 
  {
    verbose = TRUE;
  }

  if (argc==3 || (argc==2 && !verbose)) 
  {
    strcpy(ctl_file_name,argv[1]);
  }

  /* open my ctl-file */
  read_ctl_file(ctl_file_name, &global_data);

  if ((!any_planar_sbound_set(global_data.sbound_droff)))
  {
    errout("No planar SBOUND set, program not applicable!\n");
    error_occured = TRUE;
  }
  
  if (global_data.sbound_droff[R_SPHERE]>SMALL) 
  {
    errout("Spheric SBOUND set, program not applicable!\n");
    error_occured = TRUE;
  }

  if (global_data.sbound_droff[R_CYL]>SMALL) 
  {
    errout("Cylindric SBOUND set, program not applicable!\n");
    error_occured = TRUE;
  }

  if (error_occured)
  {
    txtout("\n======================================================\n");
    txtout("adjust_sbound stopped with a fatal error.\n");
    txtout("======================================================\n");
    return -1;
  }

  /***** buffer for holding atom-data ********************************/
  global_data.atomdata=alloc_atomdata(global_data.total_no_atoms+5);
  atom_sele_flag1=alloc_ints(global_data.total_no_atoms);

  /* work on vdw-file */
  txtout("Reading atom masses from file <%s>.\n",global_data.filenames[VDWPARA_FILE_INDEX]);
  if (readin_vdwfile(global_data.filenames[VDWPARA_FILE_INDEX],&global_data)==-1) {
	  errout("Fatal error in VDW-file!\n");
	  return -1;
  }

  txtout("Reading atom coordinates from file <%s>.\n",global_data.filenames[COORD_FILE_INDEX]);
  if (read_coord_file(&global_data,global_data.total_no_atoms,
		       global_data.filenames[COORD_FILE_INDEX],global_data.atomdata)==-1) {
	  errout("Fatal error in coordinate-file!\n");
	  return -1;
  }
#if DEBUG
  errout("x[0]=%f, typ[0]=%d\n", global_data.atomdata[i].ko[0], global_data.atomdata[0].atomtype);
#endif
  
  /* now we have to redistribute the coord info from */
  /* STRUCT_ATOMDATA to LOAD_MSG_ONE_NODE */
  my_data.no_atoms=my_data.no_more_atoms=global_data.total_no_atoms;
  alloc_LOAD_MSG_ONE_NODE (&my_data);
  my_data.force_sbound = alloc_doubles(global_data.total_no_atoms*4);
  for(i=0;i<global_data.total_no_atoms;i++)
  {
    my_data.x_new[i] = global_data.atomdata[i].ko[0];
    my_data.y_new[i] = global_data.atomdata[i].ko[1];
    my_data.z_new[i] = global_data.atomdata[i].ko[2];
    my_data.typ[i]   = global_data.atomdata[i].atomtype;
    strcpy(my_data.res_string[i], global_data.atomdata[i].residstr);
    my_data.res_id[i] = global_data.atomdata[i].residnr;
    my_data.force_sbound[i] = 0.0;
    my_data.force_sbound[i+1] = 0.0;
    my_data.force_sbound[i+2] = 0.0;
    my_data.force_sbound[i+3] = 0.0;
#if DEBUG
  errout("my_data.typ=%d, global_data.atomdata.typ=%d\n", my_data.typ[i], global_data.atomdata[i].atomtype);
#endif
  }
  myfree(global_data.atomdata); /* no longer needed */
  
  /* now we have to read the free format section of the ctl-file where the RANGE feature is set! */
  read_ctl_file_free_format_sec(	global_data.total_no_atoms,
					global_data.filenames[FORCE_FILE_INDEX],
					global_data.atomdata,
					atom_sele_flag1,
					&global_data,
					&my_force);
  
  txtout("Now varying the SBOUND planes ... \n");

  boundary (&my_data,
	    &global_data,
	    &rnd_force_std_deviation,
	    &Gamma,
	    &pressure_dummy,
	    &density_dummy);
#if DEBUG
  errout("After first boundary call\n");
#endif

  previous_energy = 0.0;
  for(i=0;i<global_data.total_no_atoms;i++)
  {
    previous_energy += my_data.force_sbound[i];
  }
  
  for(master_direction=-1;master_direction<2;master_direction+=2)
  {
    side = 0;
    while (side < 6)
    { 
      direction = (side%2)*2-1; /* +1 for positive sides, -1 for negative sides */
      direction = direction * master_direction;
      global_data.sbound_droff[side] += sbound_increment*direction;
      
      vout("%6.2f < X < %6.2f, %6.2f < Y < %6.2f, %6.2f < Z < %6.2f: ",
	    global_data.sbound_droff[0],global_data.sbound_droff[1],
	    global_data.sbound_droff[2],global_data.sbound_droff[3],
	    global_data.sbound_droff[4],global_data.sbound_droff[5]);
      
      for(i=0;i<global_data.total_no_atoms;i++)
      {
	my_data.force_sbound[i] = 0.0;
	my_data.force_sbound[i+1] = 0.0;
	my_data.force_sbound[i+2] = 0.0;
	my_data.force_sbound[i+3] = 0.0;
      }
      boundary (&my_data,
		&global_data,
		&rnd_force_std_deviation,
		&Gamma,
		&pressure_dummy,
		&density_dummy);
      energy = 0.0;
      for(i=0;i<global_data.total_no_atoms;i++)
      {
	energy += my_data.force_sbound[i];
      }
      vout("%6.1f\n",energy);
      
      if (energy > previous_energy)
      {
	/* we are done with this side */
	txtout("Done with side %d\n",side);
	/* undo last change */
	global_data.sbound_droff[side] -= sbound_increment*direction; 
	side ++; /* next side */
      }
      else
      {
	previous_energy = energy;
      }
    } /* while */
  } /* for */

  txtout("Please change the SBOUND positions to:\n");
  txtout("%6.2f < X < %6.2f, %6.2f < Y < %6.2f, %6.2f < Z < %6.2f.\n",
	  global_data.sbound_droff[0],global_data.sbound_droff[1],
	  global_data.sbound_droff[2],global_data.sbound_droff[3],
	  global_data.sbound_droff[4],global_data.sbound_droff[5]);
  
  /****
  txtout("Now writing out the new <%s>.\n", ctl_file_name);
  writeout_ctlfile(); 
  ****/
  
  myfree(my_data.force_sbound); /* no longer needed */
  free_LOAD_MSG_ONE_NODE(&my_data);
  return 0;

}


int	
set_QQL_anglefactor(int *atomids, double factor)
{
	return 0;	
}

int	
set_bondfactor(int *atomids, double factor)
{
	return 0;	
}

