/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: pdb2ego.c	by Helmut Heller, Munich		**
**	Convert PDB file to EGO format                                  **
**									**
**************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "../ego.h"
#include "../util_lib.h"
#include "../eganaly.h"
#include "../egcolib.h"
#include "../wodecls.h"
#include "../woalloc.h"
#include "../version.h"


/***************************************************************************/
static	int	atom_nr,resid_nr,nr_atoms;
static	double	ko[3],beta,harmon;
static	char	atomstr[10],residstr[10],segidstr[10];

/***************************************************************************/
double get_min_unitdist(int flag,
                int my_atom,int other_unitnum,
                LOAD_MSG_ONE_NODE *my_data,
                TRAVELLING_PACKETS *packet)
{
	/*** dummy ****/
	return 0.0;
}

int set_qmforces(LOAD_MSG_ALL_NODE *global_data,STRUCT_QM_INFO *finfo)
{
	/*** dummy ****/
	return 0.0;
}

int get_terminate_flag()
{
	/*** dummy ****/
	return 1;
}

int	get_atomtype_index(char *atomtypstr)
{
        strcpy(atomtypstr,"");
	return 0;
}

int read_pdb(int do_ini, char *old_pdb_file,LOAD_MSG_ALL_NODE *global_data)
{
	/*******************************************************************/
	/****************** reading template pdb file  *********************/
	/*******************************************************************/
	FILE	*pdb_h;
	int j, res;

        if ((pdb_h=fopen(old_pdb_file,"r"))==NULL) {
                errout("Fatal error: Can not open/find pdb-file <%s>.\n",old_pdb_file);
		return -1;
	}

	if(do_ini)
	{
	  /**** first count number of atoms ****/
	  nr_atoms=0;	
	  do {
		  res=read_next_pdbdata(pdb_h,&atom_nr,
					  ko,&beta,&harmon,
					  atomstr,residstr,&resid_nr,
					  segidstr);
		  if (res==-2) return -1;
		  nr_atoms++;
	  } while (res!=-1);
	  fclose(pdb_h);
	  nr_atoms--;
	  printf("There are %d atoms in this pdb-file\n",nr_atoms);fflush(stdout);
 
  	  /**** alloc the space *****/
	  global_data->atomdata=alloc_atomdata(nr_atoms+5);
	}
	
	/**** reading all data *****/
      	if ((pdb_h=fopen(old_pdb_file,"r"))==NULL) {
                errout("Fatal error: Can not open/find pdb-file <%s>.\n",old_pdb_file);
		return -1;
	}
	for (j=0;j<nr_atoms;j++) {
		res=read_next_pdbdata(pdb_h,&atom_nr,
					ko,&beta,&harmon,
					atomstr,residstr,&resid_nr,
					segidstr);

		global_data->atomdata[j].atomnr =atom_nr-1;
		global_data->atomdata[j].residnr=resid_nr;
		strcpy(global_data->atomdata[j].residstr,residstr);
		
		global_data->atomdata[j].ko[0] = ko[0];
		global_data->atomdata[j].ko[1] = ko[1];
		global_data->atomdata[j].ko[2] = ko[2];
		
		strcpy(global_data->atomdata[j].atomstr,atomstr);
		strcpy(global_data->atomdata[j].segidstr,segidstr);

		global_data->atomdata[j].beta=beta;
		global_data->atomdata[j].harmon=harmon;
		if (res==-2) return -1;
	}
	fclose(pdb_h);
	return 0;
}



/***************************************************************************/
int main(int argc,char *argv[])
{
	int 	status,do_ini=TRUE;
	char	version_string[MAX_STRING_SIZE];
	char	old_pdb_file[MAX_STRING_SIZE+1];
	char	new_ego_file[MAX_STRING_SIZE+1];
	int	file_start_number=0, file_count, file_end_number;
	double	time_step;
	LOAD_MSG_ALL_NODE	global_data;

	
	set_version_string(version_string);

        txtout("==================================================================\n");
        txtout("pdb2ego: convert PDB file to ego format.\n");
        txtout("(%s) %s\n",VERSION,EGO_MACHINE);
        txtout("==================================================================\n");

	if (!((argc == 3)||(argc == 4))) {
                txtout("pdb2ego <timestep> <number of pdb-file> [<last number of pdb file>]\n");
		txtout("timestep between two files in units of seconds\n");
		txtout("pdb-file has to be named <number>.pdb\n");
		return 0;
	}

	time_step=atof(argv[1]);
	file_start_number=atoi(argv[2]);
	if(argc == 4)
	  file_end_number=atoi(argv[3]);
	else
	  file_end_number = file_start_number;
	
	
	for(file_count=file_start_number;file_count<=file_end_number; file_count++)
	{
  	  initialize_analyze(); /* set global valiables for writing the ego file */
	  sprintf(old_pdb_file,"%08d.pdb", file_count);
	  sprintf(new_ego_file,"%08d.ego", file_count);
  
	  txtout("PDB file          : <%s>.\n",old_pdb_file);
	  txtout("EGO file          : <%s>.\n\n",new_ego_file);
		  
	  if((status=read_pdb(do_ini,old_pdb_file,&global_data)) != 0) 
	  {
	    errout("Error %d in reading PDB file. Aborting.\n",status);
	    return status;
	  }

	  if(do_ini)
	  { int i;
	    /* fill the global data structure. 
		Since we don't have any values, we make up dummies. */
	    fill_filename_list(&global_data);
	    fill_dummy_ctl_data(&global_data);
	    global_data.total_no_atoms = nr_atoms;
	    global_data.writeout_nr_atoms = global_data.total_no_atoms;
	    global_data.writeout_atoms = alloc_ints( global_data.writeout_nr_atoms);
            for(i=0;i<global_data.writeout_nr_atoms;i++)
	      global_data.writeout_atoms[i]= TRUE;
	    strcpy(global_data.outpath,"ego/");
	    create_outpath(&global_data);
	    global_data.Analyse_nth = 1;
	    do_ini=FALSE;
	  }
	  global_data.cycles_passed = file_count;
	  trick_cur_atoms_insortbuf(nr_atoms);
	  
	  writeout_coordinates(	&global_data,
				 time_step,	
			         file_count,
				 version_string);
	  close_output_file(NULL, 0);
	}

        txtout("======================================================\n");
        txtout("pdb2ego finished now. No fatal errors.\n");
        txtout("======================================================\n");
	return 0;
}
