/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: pdbpatch.c	by M.Eichinger, Munich			**
**	Some very useful things for manipulation of a PDB file          **
**      Added load/save, Tx=19990623. Helmut Heller                     **
**									**
**************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "../hardw.h"
#include "../ego.h"
#include "../util_lib.h"
#include "../wodecls.h"
#include "../woalloc.h"
#include "../version.h"

#define DEBUG 0

/***************************************************************************/
static	char	old_pdb_file[MAX_STRING_SIZE+1];
static	char	new_pdb_file[MAX_STRING_SIZE+1];
static	char	inputfile[MAX_STRING_SIZE+1];

static	char	sdum[MAX_STRING_SIZE+1];
static	char	cline[MAX_STRING_SIZE+1];	/*** current line ***/
static	char	lline[MAX_STRING_SIZE+1];	/*** last line    ***/
static	char	msel_str[MAX_STRING_SIZE+1];	/*** last line    ***/
static	char	wsel_str[MAX_STRING_SIZE+1];	/*** last line    ***/


static	int	atom_nr,resid_nr,nr_atoms,num_msel=0,num_wsel=0,num_tsel=0;
static	double	ko[3],beta,harmon;
static	char	atomstr[10],residstr[10],segidstr[10];
static	int	*mselected,*wselected,*traceboxselected;	
static	int	trajoutput=0;
static	int	coresort=0;
static	int	anim=0;
static	int	start_minbox=0;
static	double	wx=0.0,wy=0.0,wz=0.0,ws=5.0;

static	double	*savx,*savy,*savz;

static	double	pivotx=0.0,pivoty=0.0,pivotz=0.0;	
static	STRUCT_ATOMDATA	*atom_buf;

static	double rmatf[3][3]={ 	{1.0,0.0,0.0},
				{0.0,1.0,0.0},
				{0.0,0.0,1.0}
			    };
static	double rmatb[3][3]={ 	{1.0,0.0,0.0},
				{0.0,1.0,0.0},
				{0.0,0.0,1.0}
			    };

static	int	inputfile_open=0;
static	FILE*	fh_input;

/***************************************************************************/

int	get_atomtype_index(char *atomtypstr)
{
        strcpy(atomtypstr,"");
	return 0;
}

int read_pdb(int do_ini)
{
	/*******************************************************************/
	/****************** reading template pdb file  *********************/
	/*******************************************************************/
	FILE	*pdb_h;
	int j, res;

        if ((pdb_h=fopen(old_pdb_file,"r"))==NULL) {
                errout("Fatal error: Can not open/find pdb-file <%s>.\n",old_pdb_file);
		return -1;
	}

	/**** first count number of atoms ****/
	nr_atoms=0;	
	do {
		res=read_next_pdbdata(pdb_h,&atom_nr,
					ko,&beta,&harmon,
					atomstr,residstr,&resid_nr,
					segidstr);
		if (res==-2) return -1;
		nr_atoms++;
	} while (res!=-1);
	fclose(pdb_h);
	nr_atoms--;
	printf("There are %d atoms in this pdb-file\n",nr_atoms);

	if(do_ini)
	{
  	  /**** alloc the space *****/
	  atom_buf=alloc_atomdata(nr_atoms+5);
	  mselected=alloc_ints(nr_atoms+5);
	  wselected=alloc_ints(nr_atoms+5);
	  savx=alloc_doubles(nr_atoms+5);
	  savy=alloc_doubles(nr_atoms+5);
	  savz=alloc_doubles(nr_atoms+5);
	  traceboxselected=alloc_ints(nr_atoms+5);

	  for (j=0;j<nr_atoms;j++) {
		  mselected[j] = 0;
		  wselected[j] = 0;
	  }
	  cline[0] = 0;
	  lline[0] = 0;
	  msel_str[0] = 0;
	  wsel_str[0] = 0;
	}
	
	/**** reading all data *****/
      	if ((pdb_h=fopen(old_pdb_file,"r"))==NULL) {
                errout("Fatal error: Can not open/find pdb-file <%s>.\n",old_pdb_file);
		return -1;
	}
	for (j=0;j<nr_atoms;j++) {
		res=read_next_pdbdata(pdb_h,&atom_nr,
					ko,&beta,&harmon,
					atomstr,residstr,&resid_nr,
					segidstr);

		atom_buf[j].atomnr =atom_nr-1;
		atom_buf[j].residnr=resid_nr;
		strcpy(atom_buf[j].residstr,residstr);
		
		atom_buf[j].ko[0] = ko[0];
		atom_buf[j].ko[1] = ko[1];
		atom_buf[j].ko[2] = ko[2];
		
		strcpy(atom_buf[j].atomstr,atomstr);
		strcpy(atom_buf[j].segidstr,segidstr);

		atom_buf[j].beta=beta;
		atom_buf[j].harmon=harmon;
		if (res==-2) return -1;
	}
	fclose(pdb_h);
	return 0;
}

static int
set_qmrotmats(double wx, double wy, double wz)
{
	double sa,sb,sg,ca,cb,cg;

	if ((wx==0.0) && (wy==0.0) && (wz==0.0)) {
		rmatf[0][0]=1.0;   rmatf[0][1]=0.0;   rmatf[0][2]=0.0;
		rmatf[1][0]=0.0;   rmatf[1][1]=1.0;   rmatf[1][2]=0.0;
		rmatf[2][0]=0.0;   rmatf[2][1]=0.0;   rmatf[2][2]=1.0;	

		rmatb[0][0]=1.0;   rmatb[0][1]=0.0;   rmatb[0][2]=0.0;
		rmatb[1][0]=0.0;   rmatb[1][1]=1.0;   rmatb[1][2]=0.0;
		rmatb[2][0]=0.0;   rmatb[2][1]=0.0;   rmatb[2][2]=1.0;			
		return 0;	
	}
		
	sa = sin(wx/RADDEGREEFAC);
	sb = sin(wy/RADDEGREEFAC);
	sg = sin(wz/RADDEGREEFAC);	

	ca = cos(wx/RADDEGREEFAC);
	cb = cos(wy/RADDEGREEFAC);
	cg = cos(wz/RADDEGREEFAC);	

	rmatf[0][0]=cg*cb;   rmatf[0][1]=-sg*ca-cg*sb*sa;   rmatf[0][2]=sg*sa-cg*sb*ca;
	rmatf[1][0]=sg*cb;   rmatf[1][1]=cg*ca-sg*sb*sa;    rmatf[1][2]=-cg*sa-sg*sb*ca;
	rmatf[2][0]=sb;	     rmatf[2][1]=cb*sa;		    rmatf[2][2]=cb*ca;	

	rmatb[0][0]=cg*cb;   		rmatb[0][1]=sg*cb;	      rmatb[0][2]=sb;
	rmatb[1][0]=-sg*ca-cg*sb*sa;    rmatb[1][1]=cg*ca-sg*sb*sa;   rmatb[1][2]=cb*sa;
	rmatb[2][0]=sg*sa-cg*sb*ca;	rmatb[2][1]=-cg*sa-sg*sb*ca;  rmatb[2][2]=cb*ca;	
	return 0;
}

static int
rotforw(double *ko,double *res)
{
	double vec[3];
	vec[0] = rmatf[0][0]*ko[0] + rmatf[0][1]*ko[1] +rmatf[0][2]*ko[2];
	vec[1] = rmatf[1][0]*ko[0] + rmatf[1][1]*ko[1] +rmatf[1][2]*ko[2];
	vec[2] = rmatf[2][0]*ko[0] + rmatf[2][1]*ko[1] +rmatf[2][2]*ko[2];		

	res[0] = vec[0];
	res[1] = vec[1];
	res[2] = vec[2];
	return 0;	
}

static int
do_minbox()
{
	double oldvol=99999999.0,actvol;
	double ax=-9999.0,ay=-9999.0,az=-9999.0;
	int i;
	
	for (i=0;i<200;i++) {
		double x[2],y[2],z[2],dx,dy,dz;
		int j;

		set_qmrotmats(wx,wy,wz);
		
		x[0]=99999.0;	x[1]=-99999.0;
		y[0]=99999.0;	y[1]=-99999.0;
		z[0]=99999.0;	z[1]=-99999.0;	
		for (j=0;j<nr_atoms;j++) if (traceboxselected[j]) {

			rotforw(atom_buf[j].ko,ko);
			
			if (ko[0]<x[0]) x[0]=ko[0];
			if (ko[0]>x[1]) x[1]=ko[0];
	
			if (ko[1]<y[0]) y[0]=ko[1];
			if (ko[1]>y[1]) y[1]=ko[1];
	
			if (ko[2]<z[0]) z[0]=ko[2];
			if (ko[2]>z[1]) z[1]=ko[2];					
		}
		
		dx=x[1]-x[0];
		dy=y[1]-y[0];
		dz=z[1]-z[0];	
		actvol=dx*dy*dz;
		
		if (actvol<oldvol) {
			ax=wx;
			ay=wy;
			az=wz;
			oldvol=actvol;			
			printf("BOX: %f %f %f     %f %f %f  Vol %f\n",ax,ay,az,dx,dy,dz,oldvol);
		}
				
		wx=ax+(ego_rand()-0.5)*ws;
		wy=ay+(ego_rand()-0.5)*ws;		
		wz=az+(ego_rand()-0.5)*ws;			
	}
	return 0;
}	

void print_boxsize()
{
	double x[2],y[2],z[2],dx,dy,dz;
	int j;
	x[0]=99999.0;	x[1]=-99999.0;
	y[0]=99999.0;	y[1]=-99999.0;
	z[0]=99999.0;	z[1]=-99999.0;	
	for (j=0;j<nr_atoms;j++) if (traceboxselected[j]) {
		
		if (atom_buf[j].ko[0]<x[0]) x[0]=atom_buf[j].ko[0];
		if (atom_buf[j].ko[0]>x[1]) x[1]=atom_buf[j].ko[0];

		if (atom_buf[j].ko[1]<y[0]) y[0]=atom_buf[j].ko[1];
		if (atom_buf[j].ko[1]>y[1]) y[1]=atom_buf[j].ko[1];

		if (atom_buf[j].ko[2]<z[0]) z[0]=atom_buf[j].ko[2];
		if (atom_buf[j].ko[2]>z[1]) z[1]=atom_buf[j].ko[2];					
	}
	dx=x[1]-x[0];
	dy=y[1]-y[0];
	dz=z[1]-z[0];	
	printf("BOX: %f %f %f  Vol %f\n",dx,dy,dz,dx*dy*dz);
		
	return;	
}

static int
get_next_line(char *line)
{
	int c,i;

	if (inputfile[0]!=0) {
		if (!inputfile_open) {
			if ((fh_input=fopen(inputfile,"r"))==NULL) {
                		errout("Fatal error: Can not open/find inputfile <%s>.\n",
					inputfile);
				return -1;
			}
			inputfile_open=1;
		}
		
		if (read_string(fh_input,line)==-1) {
			txtout("Reached end of input-file\n");
			strcpy(line,"end");
			fclose(fh_input);
			inputfile_open=0;
		}

		return strlen(line);		
	} else {

		printf("\n>");
		for (	i=0;
			i<MAX_STRING_SIZE && (c=getchar())!=EOF && c!='\n';
			i++) line[i] = c;	

		if (c=='\n') line[i] = 0;
	}
	return i;
}

static int
clean_line(char *str)
{
	int i,slen=(int)strlen(str);
	int j,k;

	/*** erase trailing spaces ***/
	for (i=slen-1;i>=0;i--) if (str[i]!=32) {
		str[i+1]=0;
		break;
	}
	slen=(int)strlen(str);

	/*** erase leading spaces ***/
	for (i=0;i<slen;i++) {
		if (str[i]!=32) break;
	}
	if (str[i]==0) {
		str[0]=0;
		return 0;
	}
	if (i==0)	return strlen(str);
	for (j=i,k=0;j<slen;j++,k++) {
		str[k]=str[j];
	}
	return strlen(str);
}

static int
update_new_file()
{
	FILE	*pdb_h,*pdb_h_new;
	int j;

	if (trajoutput) {
		if ((pdb_h_new=fopen(new_pdb_file,"aw"))==NULL) {
                	errout("Fatal error: Can not append to file <%s>.\n",
				new_pdb_file);
			return -1;
		}
		
		txtout("*** Appending %d atoms to file <%s> ***\n",num_wsel,new_pdb_file);
		fprintf(pdb_h_new,"Step\n");
		for (j=0;j<nr_atoms;j++) if (wselected[j]) {
			fprintf(pdb_h_new,"%-5d % 10g % 10g % 10g\n",
					atom_buf[j].atomnr,
					atom_buf[j].ko[0],
					atom_buf[j].ko[1],
					atom_buf[j].ko[2]);
			
		}		
		fclose(pdb_h_new);
	} 
	
	if ((!trajoutput)||(anim)) {
		
		if ((pdb_h=fopen(old_pdb_file,"r"))==NULL) {
                	errout("Fatal error: Can not open/find pdb template file <%s>.\n",
				old_pdb_file);
			return -1;
		}
		strcpy(sdum,new_pdb_file);
		if (anim) strcpy(sdum,"temp.pdb");
		if ((pdb_h_new=fopen(sdum,"w"))==NULL) {
                	errout("Fatal error: Can not create new pdb-file <%s>.\n",
				new_pdb_file);
			return -1;
		}
		/* write out the patched file */
		for (j=0;j<nr_atoms;j++) {
			/* replace the next ATOM line in pdb_h with the coordinates from atom_buf[j] and write 
			   the modified line to pdb_h_new */
			patch_pdb_coords(pdb_h,pdb_h_new,
					atom_buf[j].ko[0],
					atom_buf[j].ko[1],
					atom_buf[j].ko[2]);
		}
		fclose(pdb_h);
		fclose(pdb_h_new);
	}
	
	if (anim) {
		char tstr[100];
		strcpy(tstr,"echo send rasmol script view.rasmol > wish.inp;");
		strcat(tstr,"echo exit >> wish.inp; wish wish.inp");
		txtout("Command <%s>\n",tstr);
		system(tstr);
	}
	
	return 0;
}

static int
set_tracebox(char *selstr)
{
	int j;

	num_tsel=0;
	for (j=0;j<nr_atoms;j++) traceboxselected[j] = 0;
	
	set_used_atoms(	selstr,
			nr_atoms,
			atom_buf,
			traceboxselected);
	strcpy(wsel_str,selstr);
	for (j=0;j<nr_atoms;j++) if (traceboxselected[j]) num_tsel++;

	printf("*** %d atoms selected for tracebox ***\n",num_tsel);

	return 0;
}

static int
set_mselect(char *selstr)
{
	int j;

	num_msel=0;
	for (j=0;j<nr_atoms;j++) mselected[j] = 0;
	
	set_used_atoms(	selstr,
			nr_atoms,
			atom_buf,
			mselected);
	strcpy(msel_str,selstr);
	for (j=0;j<nr_atoms;j++) if (mselected[j]) num_msel++;

	printf("*** %d atoms selected for manipulation ***\n",num_msel);

	return 0;
}

static int
set_wselect(char *selstr)
{
	int j;

	num_wsel=0;
	for (j=0;j<nr_atoms;j++) wselected[j] = 0;
	
	set_used_atoms(	selstr,
			nr_atoms,
			atom_buf,
			wselected);
	strcpy(wsel_str,selstr);
	for (j=0;j<nr_atoms;j++) if (wselected[j]) num_wsel++;

	printf("*** %d atoms selected for writing in trajectory mode ***\n",num_wsel);

	return 0;
}

static int
set_minbox(char *selstr)
{
	start_minbox=1;
	
	if (sscanf(selstr,"%lf %lf %lf %lf",&wx,&wy,&wz,&ws)!=4) return -2;

	printf("*** Starting min box with %f %f %f step %f ***\n",wx,wy,wz,ws);

	return 1;
}

int average_file(char *selstr)
{
  if(selstr[0] == 0) 
  {
    printf("Command syntax: average <filename>\n");
    printf("But <filename> was missing! Coordinates unchanged.\n");
    return -1;
  }
  else
  { /* We ignore many possible error conditions: 
       different number of atoms
       atoms not in same order */
    int j;
    
    /* Keep a copy of the current coordinates */
    for (j=0;j<nr_atoms;j++) 
    {
      savx[j]=atom_buf[j].ko[0];
      savy[j]=atom_buf[j].ko[1];
      savz[j]=atom_buf[j].ko[2];
    }
    
    strcpy(old_pdb_file, selstr);
    printf("*** Now using %s as input file ***\n", old_pdb_file);

    /* read in the new coordinates */
    if((j=read_pdb(FALSE)) == 0)
    {
      /* now do the averaging */
      for (j=0;j<nr_atoms;j++) 
      {
	atom_buf[j].ko[0]=(savx[j]+atom_buf[j].ko[0])/2.0;
	atom_buf[j].ko[1]=(savy[j]+atom_buf[j].ko[1])/2.0;
	atom_buf[j].ko[2]=(savz[j]+atom_buf[j].ko[2])/2.0;
      }
      return 0;
    }
    else
      return j;
  }
}

int load_new_file(char *selstr)
{
  if(selstr[0] == 0) 
  {
    printf("Command syntax: load <filename>\n");
    printf("But <filename> was missing! Input unchanged.\n");
    return -1;
  }
  else
  {
    strcpy(old_pdb_file, selstr);
    printf("*** Now using %s as input file ***\n", old_pdb_file);
    return read_pdb(FALSE);
  }
}

int save_new_file(char *selstr)
{
  if(selstr[0] != 0) 
  {
    strcpy(new_pdb_file, selstr);
    printf("*** Now using %s as output file ***\n", new_pdb_file);
  }
  printf("*** Writing current coordinates to %s ***\n",new_pdb_file);
  return 1;
}

static int
move_selected(char *selstr)
{
	int i;
	double dx=0.0,dy=0.0,dz=0.0;

	if (sscanf(selstr,"%lf %lf %lf",&dx,&dy,&dz)!=3) return -2;

	printf("*** moving with %f %f %f ***\n",dx,dy,dz);
	for (i=0;i<nr_atoms;i++) if (mselected[i]) {
		atom_buf[i].ko[0] += dx;
		atom_buf[i].ko[1] += dy;
		atom_buf[i].ko[2] += dz;
	}
	return 1;
}

static int
dmove_selected(char *selstr)
{
	int i;
	double dx=0.0,dy=0.0,dz=0.0,d,len;
	int a1,a2;

	if (sscanf(selstr,"%d %d %lf",&a1,&a2,&d)!=3)	return -2;
	if ((a1<=0)||(a1>nr_atoms)) 			return -2;
	if ((a2<=0)||(a2>nr_atoms)) 			return -2;
	
	dx  = atom_buf[a2-1].ko[0] - atom_buf[a1-1].ko[0];
	dy  = atom_buf[a2-1].ko[1] - atom_buf[a1-1].ko[1];
	dz  = atom_buf[a2-1].ko[2] - atom_buf[a1-1].ko[2];
	len = sqrt(dx*dx + dy*dy + dz*dz);

	dx /= len;
	dy /= len;
	dz /= len;

	printf("*** moving with %f %f %f ***\n",dx*d,dy*d,dz*d);
	for (i=0;i<nr_atoms;i++) if (mselected[i]) {
		atom_buf[i].ko[0] += dx*d;
		atom_buf[i].ko[1] += dy*d;
		atom_buf[i].ko[2] += dz*d;
	}
	return 1;
}

static int
place_selected(char *selstr)
{
	int i,atomnr=0;
	double dx=0.0,dy=0.0,dz=0.0;
	double x=0.0,y=0.0,z=0.0;

	if (sscanf(selstr,"%d %lf %lf %lf",&atomnr,&x,&y,&z)!=4) return -2;
	if ((atomnr<=0)||(atomnr>nr_atoms)) 			 return -2;

	printf("*** placing atom %d to %f %f %f ***\n",atomnr,x,y,z);
	dx = x - atom_buf[atomnr-1].ko[0];
	dy = y - atom_buf[atomnr-1].ko[1];
	dz = z - atom_buf[atomnr-1].ko[2];

	for (i=0;i<nr_atoms;i++) if (mselected[i]) {
		atom_buf[i].ko[0] += dx;
		atom_buf[i].ko[1] += dy;
		atom_buf[i].ko[2] += dz;
	}
	return 1;
}

static int
show_pos(char *selstr)
{
	int atomnr=0;
	if (sscanf(selstr,"%d",&atomnr)!=1) return -2;
	if ((atomnr<=0)||(atomnr>nr_atoms)) return -2;

	printf("*** Atom %4s %-5d  Res %4s %-4d (%4s) %8.3f %8.3f %8.3f ***\n",
		atom_buf[atomnr-1].atomstr,
		atomnr,
		atom_buf[atomnr-1].residstr,
		atom_buf[atomnr-1].residnr,
		atom_buf[atomnr-1].segidstr,
		atom_buf[atomnr-1].ko[0],
		atom_buf[atomnr-1].ko[1],
		atom_buf[atomnr-1].ko[2]);
	return 0;
}

static int
set_pivot(char *selstr)
{
	double x=0.0,y=0.0,z=0.0;
	int atomnr;
	
	if (sscanf(selstr,"%lf %lf %lf",&x,&y,&z)!=3) {
		if (sscanf(selstr,"%d",&atomnr)!=1) return -2;
		x = atom_buf[atomnr-1].ko[0];
		y = atom_buf[atomnr-1].ko[1];
		z = atom_buf[atomnr-1].ko[2];
	}
	
	printf("*** New pivot point: %8.3f %8.3f %8.3f\n",x,y,z);
	pivotx = x;
	pivoty = y;
	pivotz = z;
	return 0;
}

static int
rotate_selected(char *selstr)
{
	double	angle=0.0;
	double	ca=0.0,sa=0.0,dx,dy,dz;
	double	dxn,dyn,dzn;
	int	i;
	char	axis;

	if (sscanf(selstr,"%c %lf",&axis,&angle)!=2) return -2;
	if ((axis!='x')&&(axis!='y')&&(axis!='z')) return -2;

	printf("*** Rotating around axis %c with %f degree\n",axis,angle);

	ca = cos(angle*3.1415926/180.0);
	sa = sin(angle*3.1415926/180.0);

	for (i=0;i<nr_atoms;i++) if (mselected[i]) {
		dxn = dx = atom_buf[i].ko[0] - pivotx;
		dyn = dy = atom_buf[i].ko[1] - pivoty;
		dzn = dz = atom_buf[i].ko[2] - pivotz;

		if (axis=='x') {
			dyn = ca*dy - sa*dz;
			dzn = sa*dy + ca*dz;	
		} else
		if (axis=='y') {
			dxn = ca*dx - sa*dz;
			dzn = sa*dx + ca*dz;
		} else
		if (axis=='z') {
			dxn = ca*dx - sa*dy;
			dyn = sa*dx + ca*dy;
		}
		atom_buf[i].ko[0] = dxn + pivotx;
		atom_buf[i].ko[1] = dyn + pivoty;
		atom_buf[i].ko[2] = dzn + pivotz;
	}
	return 1;
}

/*      currently unused: 
static int
rotback(double *ko,double *res)
{
	double vec[3];
	vec[0] = rmatb[0][0]*ko[0] + rmatb[0][1]*ko[1] +rmatb[0][2]*ko[2];
	vec[1] = rmatb[1][0]*ko[0] + rmatb[1][1]*ko[1] +rmatb[1][2]*ko[2];
	vec[2] = rmatb[2][0]*ko[0] + rmatb[2][1]*ko[1] +rmatb[2][2]*ko[2];		

	res[0] = vec[0];
	res[1] = vec[1];
	res[2] = vec[2];
	return 0;	
}
*/

static int
show_help()
{
	printf("help or ?		     Help information on defined commands.\n");
	printf("status			     Information about current status.\n");
	printf("info <atomnr>		     Shows info about atom with <atomnr>.\n");
	printf("mselect <selstr>	     Select atoms for manipulation.\n");
	printf("wselect <selstr>	     Select atoms for writing in trajectory mode.\n");
	printf("move <dx> <dy> <dz>	     Move selected atoms.\n");
	printf("dmove <atomnr> <atomnr> d    Move selected atoms.\n");
	printf("place <atomnr> <x> <y> <z>   Shifts selected atoms, so that atom with\n");
	printf("                             <atomnr> is placed at position x,y,z.\n");
	printf("pivot <x> <y> <z>	     Set pivot point of rotation.\n");
	printf("pivot <atomnr>	             Set pivot point of rotation.\n");
	printf("rotate <x|y|z> <angle>       Rotate selected atoms around axis.\n");
	printf("tracebox <selstr>            Select atoms for box-calc.\n");
	printf("minbox wx wy wz sw           Minimize volume.\n");
	printf("append                       Only in trajectory-mode: append coordinates to output file.\n");
	printf("load <filename>              Set a new input filename.\n");
	printf("average <filename>           Compute the average of the current coordinates and those in <filename>.\n");
	printf("save [<filename>]            Write current coordinates to file.\n");
	printf("                             If <filename> is given it is set and used as the new <outputfile>.\n");
	printf("end			     Ends the program.\n");	
        return 0;    
}

static int
show_status()
{
	printf("Current status:\n");
	printf("Input file    : <%s>\n",old_pdb_file);
	printf("Output file   : <%s>\n",new_pdb_file);
	printf("Selected %d atoms out off %d with <%s> for manipulation.\n",num_msel,nr_atoms,msel_str);
	printf("Selected %d atoms out off %d with <%s> for writing in trajectory mode.\n",num_wsel,nr_atoms,wsel_str);
	printf("Pivot point   : %8.3f %8.3f %8.3f\n",pivotx,pivoty,pivotz);
        return 0;
}

static int
parse_line(char *cline)
{
	int	i,len;
	char	cstr[200];

do_it_again:
	len=strlen(cline);
	strcpy(cstr,cline);
	for (i=0;i<len;i++) {
		if (cstr[i]==32) {
			cstr[i] = 0;
			break;
		}	
	}

	if (strcmp(cstr,"l")==0) {
		strcpy(cline,lline);
		len=strlen(cline);
		printf(">%s\n",cline);
		goto do_it_again;
	}
	if (strcmp(cstr,"help")==0)	return show_help();
	if (strcmp(cstr,"?")==0)	return show_help();
	if (strcmp(cstr,"status")==0)	return show_status();
	if (strcmp(cstr,"info")==0)	return show_pos(cline+i+1);
	if (strcmp(cstr,"mselect")==0)	return set_mselect(cline+i+1);
	if (strcmp(cstr,"wselect")==0)	return set_wselect(cline+i+1);
	if (strcmp(cstr,"move")==0)	return move_selected(cline+i+1);
	if (strcmp(cstr,"dmove")==0)	return dmove_selected(cline+i+1);
	if (strcmp(cstr,"place")==0)	return place_selected(cline+i+1);
	if (strcmp(cstr,"pivot")==0)	return set_pivot(cline+i+1);	
	if (strcmp(cstr,"rotate")==0)	return rotate_selected(cline+i+1);
	if (strcmp(cstr,"tracebox")==0)	return set_tracebox(cline+i+1);
	if (strcmp(cstr,"minbox")==0)   return set_minbox(cline+i+1);
	if (strcmp(cstr,"append")==0)	return 2;
	if (strcmp(cstr,"load")==0)	return load_new_file(cline+i+1);
	if (strcmp(cstr,"average")==0)	return average_file(cline+i+1);
	if (strcmp(cstr,"save")==0)	return save_new_file(cline+i+1);
	if (strcmp(cstr,"end")==0)	return 0;

	printf("Unkown Option!\n");
	return -1;
}



/***************************************************************************/
int main(int argc,char *argv[])
{
	int	res,j;

        txtout("==================================================================\n");
        txtout("patchpdb - Patching a PDB file.\n");
        txtout("(%s) %s\n",VERSION,EGO_MACHINE);
        txtout("==================================================================\n");

	if (argc<3) {
                txtout("patchpdb <pdb-file-template> <outputfile> [-coresort] [-traj] [-f <inputfile>] [-anim]\n");
		txtout(" -coresort resorts the coordinates in the PDB file so that the C-O atoms are listed last.\n");
		txtout(" -traj     appends the selected atoms to the outputfile (in PDB format).\n");
		txtout(" -f        if specified, commands are taken from this file, otherwise stdin is read.\n");
		txtout(" -anim     generates a RasMol script and sends it to RasMol, using wish.\n");
		txtout("           interpolated frames if you make a movie.\n");
		return 0;
	}

	trajoutput=0;
	inputfile[0] = 0;
	strcpy(old_pdb_file,argv[1]);
	strcpy(new_pdb_file,argv[2]);
	if (argc>3) {
		for (j=3;j<argc;j++) {
			if (strcmp(argv[j],"-coresort")==0) coresort=1;
			if (strcmp(argv[j],"-anim")==0) anim=1;
			if (strcmp(argv[j],"-traj")==0) trajoutput=1;
			if (strcmp(argv[j],"-f")==0) {
				if (j+1>=argc) {
					errout("Wrong number of arguments\n");	
					exit(-1);
				}
				strcpy(inputfile,argv[j+1]);
			}
			
		}	
	}

        txtout("Template PDB file          : <%s>.\n",old_pdb_file);
        txtout("New      PDB file          : <%s>.\n",new_pdb_file);
	if (trajoutput) txtout("Writing trajectory format\n");
	if(inputfile[0]==0)
	  txtout("Taking command input from <stdin>.\n");
	else
	  txtout("Taking command input from <%s>.\n",inputfile);

	if (anim) {
		strcpy(sdum,"(rasmol ");
		strcat(sdum,old_pdb_file);
		strcat(sdum,")&");
		system(sdum);
		getchar();
	}
		
        if((j=read_pdb(TRUE)) != 0) return j;
	
	/*** here we resort C-O in PDB files to end of each residue ***/
	if (coresort) {
		FILE *pdb_h_new;
		int s1,s2,curres,i;
		s1=nr_atoms+1;
		s2=nr_atoms+2;
		for (i=0;i<nr_atoms-2;i++) {
			if (	(strcmp(atom_buf[i  ].atomstr,"CA  ")==0) &&
				(strcmp(atom_buf[i+1].atomstr,"C   ")==0)  &&
				(strcmp(atom_buf[i+2].atomstr,"O   ")==0)) {
				curres=atom_buf[i+1].residnr;
				
				atom_buf[s1].atomnr = atom_buf[i+1].atomnr;
				atom_buf[s1].residnr= atom_buf[i+1].residnr;				
				atom_buf[s1].beta   = atom_buf[i+1].beta;
				atom_buf[s1].harmon = atom_buf[i+1].harmon;
				atom_buf[s1].ko[0]  = atom_buf[i+1].ko[0];				
				atom_buf[s1].ko[1]  = atom_buf[i+1].ko[1];
				atom_buf[s1].ko[2]  = atom_buf[i+1].ko[2];
				strcpy(atom_buf[s1].atomstr,atom_buf[i+1].atomstr);
				strcpy(atom_buf[s1].residstr,atom_buf[i+1].residstr);								
				strcpy(atom_buf[s1].segidstr,atom_buf[i+1].segidstr);

				atom_buf[s2].atomnr = atom_buf[i+2].atomnr;
				atom_buf[s2].residnr= atom_buf[i+2].residnr;				
				atom_buf[s2].beta   = atom_buf[i+2].beta;
				atom_buf[s2].harmon = atom_buf[i+2].harmon;
				atom_buf[s2].ko[0]  = atom_buf[i+2].ko[0];				
				atom_buf[s2].ko[1]  = atom_buf[i+2].ko[1];
				atom_buf[s2].ko[2]  = atom_buf[i+2].ko[2];
				strcpy(atom_buf[s2].atomstr,atom_buf[i+2].atomstr);
				strcpy(atom_buf[s2].residstr,atom_buf[i+2].residstr);								
				strcpy(atom_buf[s2].segidstr,atom_buf[i+2].segidstr);
				
				for (j=i+3;j<nr_atoms;j++) {
					if (atom_buf[j].residnr==curres) {
						atom_buf[j-2].atomnr = atom_buf[j].atomnr;
						atom_buf[j-2].residnr= atom_buf[j].residnr;				
						atom_buf[j-2].beta   = atom_buf[j].beta;
						atom_buf[j-2].harmon = atom_buf[j].harmon;
						atom_buf[j-2].ko[0]  = atom_buf[j].ko[0];				
						atom_buf[j-2].ko[1]  = atom_buf[j].ko[1];
						atom_buf[j-2].ko[2]  = atom_buf[j].ko[2];
						strcpy(atom_buf[j-2].atomstr,atom_buf[j].atomstr);
						strcpy(atom_buf[j-2].residstr,atom_buf[j].residstr);								
						strcpy(atom_buf[j-2].segidstr,atom_buf[j].segidstr);
					} 
					if ((atom_buf[j].residnr!=curres)||(j==nr_atoms-1)) {
						if (j==nr_atoms-1) j++;
						atom_buf[j-2].atomnr = atom_buf[s1].atomnr;
						atom_buf[j-2].residnr= atom_buf[s1].residnr;				
						atom_buf[j-2].beta   = atom_buf[s1].beta;
						atom_buf[j-2].harmon = atom_buf[s1].harmon;
						atom_buf[j-2].ko[0]  = atom_buf[s1].ko[0];				
						atom_buf[j-2].ko[1]  = atom_buf[s1].ko[1];
						atom_buf[j-2].ko[2]  = atom_buf[s1].ko[2];
						strcpy(atom_buf[j-2].atomstr,atom_buf[s1].atomstr);
						strcpy(atom_buf[j-2].residstr,atom_buf[s1].residstr);								
						strcpy(atom_buf[j-2].segidstr,atom_buf[s1].segidstr);						

						atom_buf[j-1].atomnr = atom_buf[s2].atomnr;
						atom_buf[j-1].residnr= atom_buf[s2].residnr;				
						atom_buf[j-1].beta   = atom_buf[s2].beta;
						atom_buf[j-1].harmon = atom_buf[s2].harmon;
						atom_buf[j-1].ko[0]  = atom_buf[s2].ko[0];				
						atom_buf[j-1].ko[1]  = atom_buf[s2].ko[1];
						atom_buf[j-1].ko[2]  = atom_buf[s2].ko[2];
						strcpy(atom_buf[j-1].atomstr,atom_buf[s2].atomstr);
						strcpy(atom_buf[j-1].residstr,atom_buf[s2].residstr);								
						strcpy(atom_buf[j-1].segidstr,atom_buf[s2].segidstr);						
						i=j-1;
						break;	
					}
				}	
			}
	
		}

		if ((pdb_h_new=fopen(new_pdb_file,"w"))==NULL) {
                	errout("Fatal error: Can not create file <%s>.\n",
				new_pdb_file);
			return -1;
		}
		for (i=0;i<nr_atoms;i++) {
                        char mystr[10];

			atom_buf[i].atomnr=i+1;
              
		        if (strlen(atom_buf[i].atomstr)==4) {
				strcpy(mystr,atom_buf[i].atomstr);
			} else {
				mystr[0]=' '; strcpy(mystr+1,atom_buf[i].atomstr);
			}
	
			fprintf(pdb_h_new,"ATOM   %4d %4s %4s %4d    % 8.3f% 8.3f% 8.3f  1.00  0.00              \n",
				atom_buf[i].atomnr,
				mystr,
				atom_buf[i].residstr,
				atom_buf[i].residnr,
				atom_buf[i].ko[0],atom_buf[i].ko[1],atom_buf[i].ko[2]
				);
		}
		fclose(pdb_h_new);
		return 0;
	}

	/************************************************************************/
	/*** here starts the loop for parsing the input file (from -f option) ***/
	/************************************************************************/
	do {
		cline[0] = 0;

		get_next_line(cline);
		if (clean_line(cline)==0) continue;
/**
printf("<%s>\n",cline);
**/
		res = parse_line(cline);
		if (res==-2) {
			printf("Illegal input!\n");
			continue;
		} else 
		if (res==-1) continue;

		if ((res==1)&&(!trajoutput)) {
			update_new_file();
			strcpy(lline,cline); /* remember the command line in "last line" (lline) */
		}
		
		if ((res==2)&&(trajoutput)) {
			update_new_file();			
		}
		if (num_tsel) {
			print_boxsize();
		}
		if (start_minbox) {
			do_minbox();
			start_minbox=0;	
		}

	} while (strcmp(cline,"end")!=0);

        txtout("\n======================================================\n");
        txtout("patchpdb finished now. No fatal errors.\n");
        txtout("======================================================\n");
	return 0;
}
