/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

#define DEBUG 0


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#if defined(NEXT) || defined(RS6K) || defined(MAC)
#include <stdlib.h>
#else
#include <malloc.h>
#endif

#include "../ego.h"
#include "../util_lib.h"
#include "../egbound.h"
#include "../egcolib.h"
#include "../egmach.h"
#include "../egalloc.h"

#define INC_SIZE 1000

#define sign(a) (a>=0 ? 1 : -1)

int	set_QQL_anglefactor(int *atomids, double factor);
int	set_bondfactor(int *atomids, double factor);

/* Here we define some dummies to avoid including a whole slew of other files */
/**************************************************************************/
void   pex(int i)
{
	txtout("Exit by pex(%d) in round_corners.c !\n",i);
	exit(i);
	return;
}

/***************************************************************************/
int
parse_qminterface(      char *keywordstr,char* reststr,
                        STRUCT_ATOMDATA *atom_data,
                        int *atom_sele_flag1,
                        LOAD_MSG_ALL_NODE *global_data,
                        STRUCT_FORCES *my_force,
                        int     *used_atoms)
{
return -1;
}

int	CLU_get_all_atoms(int level,	/** level-index: 0 -> unitlevel,...**/
			int parent,	/** parent-index ***/
			int *nr,	/** nr of elements for which buffer is allocated */
			int **buf,	/** the buffer for the atom-indices (OLD-EGO) */
			struct_cluster *cluster[],
			int *old_atom_nrs)
{ /* we have to access all parameters or the PARIX ancc will complain! */
  int dummy;
  dummy = level;
  dummy = parent;
  dummy = *nr;
  dummy = **buf;
  dummy = *old_atom_nrs;
  dummy = cluster[0][0].status;
  return 0;
}

/***************************************************************************/
int add_ctlfile_data(FILE *h_ctl,LOAD_MSG_ALL_NODE *global_data)
{ FILE dummy;
  int idummy;
  dummy = *h_ctl;
  idummy = global_data->shake;
  return 0;
}

/************************************************************************/
void calc_vir_pressure(PRESSURE *press,double Ekin,
			 double *sbound_droff,
			 double ekin_tensor[3][3])
{
  return ;
}

int
set_qmboundbox( int *atom_sele_flag1,
		STRUCT_ATOMDATA *atom_data,
		LOAD_MSG_ALL_NODE *global_data)
{
return -1;
}

/********************************************************************************/
int main(int argc,char *argv[])
{
  char	pdb_in_file[MAX_STRING_SIZE];
  char	pdb_out_file[MAX_STRING_SIZE];
  char	ctl_file[MAX_STRING_SIZE];
  LOAD_MSG_ALL_NODE global_data;
  FILE	*pdb_out_file_ptr,*pdb_in_file_ptr;
  int *atom_sele_flag1;
  STRUCT_FORCES my_force;
  STRUCT_ATOMDATA *atom_data;
  
  
  txtout("==================================================================\n");
  txtout("round_corners - cut corners from a PDB file to fit SBOUND.\n");
  txtout("==================================================================\n");

  /* It was a bad decision to use column 9 for the sbound energy. But other programs
     depend on it now. Too late to change history. */
  if (argc!=3 && argc!=4) {
	  txtout("round_corners <pdb in> <pdb out> [<control-file>]\n");
	  txtout("The default for <control-file> is ./ctl.lis.\n");
	  txtout("The atoms which are to be cut away will get a positive\n");
	  txtout("value in the 9th column (friction coefficient, X-PLOR Q value).\n");
	  txtout("Their value is simply the truncated (+-99.99) SBOUND energy.\n");
	  txtout("To prevent numerical underflow, 0.01 is used as the smallest value.\n");
	  txtout("ATTENTION: hydrogen atoms will be deselected by the X_PLOR\n");
	  txtout("script; here they are assigned an SBOUND energy, although\n");
	  txtout("in EGO they do not feel the SBOUND.\n");
	  txtout("Use the X-PLOR script round_corners.inp to delete the marked waters.\n");
	  txtout("In addition, we write into the 10th column (X-PLOR B value) the\n");
	  txtout("friction coefficient in 1/ps.\n");
	  return 0;
  }
  /***************************************************************************/
  /*** Manage INPUT **********************************************************/
  /***************************************************************************/
  strcpy(pdb_in_file,argv[1]);
  strcpy(pdb_out_file,argv[2]);
  if (argc == 4)
  {
    strcpy(ctl_file,argv[3]);
  }
  else
  {
    strcpy(ctl_file,"ctl.lis");
  }

  txtout("Input PDB file          : <%s>.\n",pdb_in_file);
  txtout("Output PDB file         : <%s>.\n",pdb_out_file);
  txtout("Control file            : <%s>.\n",ctl_file);

  /* first we read in the control file */
  read_ctl_file(ctl_file,&global_data);

#if DEBUG == 1
printf("Read ctl.lis file.\n");
#endif

  read_sbound_file(global_data.filenames[SBOUND_FILE_INDEX],&global_data);

  if ((!any_planar_sbound_set(global_data.sbound_droff)) || (global_data.sbound_droff[R_CURVE] < SMALL))
  {
    errout("No planar SBOUND with round corners set, program not applicable!\n");
  /* We relax (comment out) the following restriction: now we assign the SBOUND energy in any case */
  /*
    goto error_occured;
  */
  }
  
  if ((pdb_out_file_ptr=fopen(pdb_out_file,"w"))==NULL) {
	  errout("Fatal error: Can not create pdb output file <%s>.\n", pdb_out_file);
	  goto error_occured;
  }
  
  if ((pdb_in_file_ptr=fopen(pdb_in_file,"r"))==NULL) {
	  errout("Fatal error: Can not open/find pdb input file <%s>.\n",pdb_in_file);
	  goto error_occured;
  }

  atom_data=alloc_atomdata(global_data.total_no_atoms+5);
  atom_sele_flag1=alloc_ints(global_data.total_no_atoms);

  /* now we have to read the free format section of the ctl file where the RANGE feature is set! */
  read_ctl_file_free_format_sec(	global_data.total_no_atoms,
			ctl_file,
			atom_data,
			atom_sele_flag1,
			&global_data,
			&my_force);
  global_data.restarting = TRUE; /* we have to set this to enable the friction computation */
  global_data.harmonic = FALSE; /* we have to disable this because we fake a lot here */
 
  /* now we go through the pdb file and make changes on the fly */
  {
    int done;
    char   line_str[MAX_STRING_SIZE];
    char   tmpstr[MAX_STRING_SIZE];
#if DEBUG == 1
    int a_num; /* atom number */
#endif
    
    double x,y,z;
    double rnd_force_std_deviation, Gamma;
    int dummy=0; /* instead of allocating since we need only 1 integer! */

    LOAD_MSG_ONE_NODE my_data;
    PRESSURE pressure, density;

    global_data.AB[0][0] = 10.0; /* mass > 4.0 for testing in SBOUND routine:
    				    here all atoms, even hydrogens, participate in the
				    SBOUND. Hydrogens are then flagged out later by an awk
				    script. */

    my_data.typ=&dummy; /* the type of atom 0 (the only one here) is 0 */
    my_data.res_string=(char (*) [LEN_RES_TYPE+1]) alloc_chars((LEN_RES_TYPE+1)+10);
    my_data.no_atoms=1;
    my_data.x_new=&x;
    my_data.y_new=&y;
    my_data.z_new=&z;
    my_data.force_sbound = alloc_doubles(4);
      
    done = read_string(pdb_in_file_ptr, line_str);
    while (done != EOF)
    {
      strncpy(tmpstr, line_str,4);
      tmpstr[4]=0;
      if (strcmp(tmpstr,"ATOM")==0)   
      { /* found one */
#if DEBUG == 1
	strncpy(tmpstr, line_str+6,5);  /* atom number (usefull for debugging) */
	tmpstr[5]=0;
	sscanf(tmpstr,"%d",&a_num);
#endif
	strncpy(tmpstr, line_str+30,8);  /* X */
	tmpstr[8]=0;
	sscanf(tmpstr,"%lf",&x);
	strncpy(tmpstr, line_str+38,8);  /* Y */
	tmpstr[8]=0;
	sscanf(tmpstr,"%lf",&y);
	strncpy(tmpstr, line_str+46,8);  /* Z */
	tmpstr[8]=0;
	sscanf(tmpstr,"%lf",&z);
        strncpy(tmpstr, line_str+17, 4); /* residue name */
	tmpstr[4] = 0;
	sscanf(tmpstr,"%s",(my_data.res_string[0]));

	my_data.force_sbound[0] = 0.0;
	zero_pressure(&pressure);

	boundary (&my_data,
		  &global_data,
		  &rnd_force_std_deviation,
		  &Gamma,
		  &pressure,
		  &density);

#if DEBUG == 1
        if(a_num == 32040)
        printf("%d, =>%s<=, %f\n",a_num,my_data.res_string[0], my_data.force_sbound[0]);
#endif

        /* prevent problems when printing to the PDB column */
        if (my_data.force_sbound[0] > 99.0)  my_data.force_sbound[0] =  99.99;
        if (my_data.force_sbound[0] < -99.0) my_data.force_sbound[0] = -99.99;
        if ((my_data.force_sbound[0] != 0.0) && 
	    fabs(my_data.force_sbound[0]) < 0.01) 
	      my_data.force_sbound[0] = 0.01 * sign(my_data.force_sbound[0]);
	
	/* now we compute the friction back into 1/ps units: */
	Gamma = Gamma / global_data.time_step * 2.0 / TIMEFAC;
	
        strncpy(tmpstr,line_str,54);
        tmpstr[54]=0;
        sprintf(tmpstr+54,"%6.2f",my_data.force_sbound[0]);
        sprintf(tmpstr+60,"%6.2f",Gamma);
        strncpy(tmpstr+66,line_str+66,strlen(line_str)-66);
        fprintf(pdb_out_file_ptr,"%s\n", tmpstr);
      }
      else
      {
        fprintf(pdb_out_file_ptr,"%s\n", line_str);
      }
      done = read_string(pdb_in_file_ptr,line_str);
    }
    myfree(my_data.res_string);
    myfree(my_data.force_sbound);
  }
  txtout("\n");

  fprintf(pdb_in_file_ptr,"END\n");
  fclose(pdb_out_file_ptr);
  fclose(pdb_in_file_ptr);

  txtout("\n======================================================\n");
  txtout("round_corners finished. No fatal errors.\n");
  txtout("======================================================\n");
  return 0;

error_occured:
  txtout("\n======================================================\n");
  txtout("round_corners stopped with a fatal error.\n");
  txtout("======================================================\n");
  return -1;
}



int	
set_QQL_anglefactor(int *atomids, double factor)
{
	return 0;	
}

int	
set_bondfactor(int *atomids, double factor)
{
	return 0;	
}
