/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

#include "ego.h"
#include "woalloc.h"

static void	oom (void *ptr,int size);
	
/***********************************************************************/
void   pex (int i)
{
	printf("Exit by pex(%d) in woalloc.c !\n",i);
	fflush(stdout);

	exit(i);
	return;
}

static void oom (void *ptr,int news)
{				/* Out Of Memory */
	if (ptr==NULL) {
		printf("Out of memory (%d)! Program stops.\n",news);
		pex (-1);
	}
	return;
}

STRUCT_ATOMDATA *alloc_atomdata (int len)
{
	STRUCT_ATOMDATA   *ptr;
	size_t	nobytes;

	if (len < 1)	len = 1;
	nobytes = (size_t) (len * sizeof(STRUCT_ATOMDATA));

	ptr = (STRUCT_ATOMDATA *)malloc(nobytes);
	oom(ptr,nobytes);
	return (ptr);
}

/***********************************************************************/
int   *alloc_ints(int len)
{
	int   *ptr;
	size_t	nobytes;

	if (len < 1)	len = 1;
	nobytes = (size_t) (len * sizeof(int));

	ptr = (int *)malloc(nobytes);
	oom(ptr,nobytes);
	return (ptr);
}

int   **alloc_ptrs(int len)
{
        int   **ptr;
        size_t  nobytes;

        if (len < 1)    len = 1;
        nobytes = (size_t) (len * sizeof(int*));
#if INFO
        txtout("Node(ALC) %2d: alloc_ptrs [%-10s] for %d x %d = %d ...",
                my_number,allocinfostr,len,sizeof(int*),nobytes);

#endif
        ptr = (int **)malloc(nobytes);
        oom(ptr,nobytes);
#if INFO
        txtout("  newp %p\n",ptr);
#endif
        return (ptr);
}

/***********************************************************************/
double *alloc_doubles (int len)
{
	double *ptr;
	size_t	nobytes;

	if (len < 1) len = 1;
	nobytes = (size_t) (len * sizeof(double));

	ptr = (double *)malloc(nobytes);
	oom (ptr,nobytes);
	return (ptr);
}

float *alloc_floats (int len)
{
	float *ptr;
	size_t	nobytes;

	if (len < 1) len = 1;
	nobytes = (size_t) (len * sizeof(float));

	ptr = (float *)malloc(nobytes);
	oom (ptr,nobytes);
	return (ptr);
}

void   myfree (void *ptr)
{
	free(ptr);
	return;
}

