/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

#ifdef WODECLSINCL
#else
#define WODECLSINCL

#define MAX_COMMENT_LINES 10


/*******************************************************/
/************ global functions used in xpl2lis *********/
/*******************************************************/

/**** womain.c *******/

int	get_atomtype_index(char	*newstr);
int	get_atti_by_atomnr(int atom_nr);
int	goto_psf_atomdata(FILE *h_psf,int *nr_of_atoms);
double	get_atom_mass(int att);
int	get_atomtype_str(int att_index,char *str);
int	get_pdb_atomdata(int atom_nr,STRUCT_ATOMDATA *data);
int	isit_a_hydrogen(double mass);

/**** woangle.c ******/

int	readin_all_anglepara(int argc,char *argv[], char atom_typestr_list[][LEN_ATOM_TYPE+1]);
int	writeout_anglelist(char *psffn,char *anglelistfn,int nr_of_atoms);
int	writeout_anglepara(char *angleparafn);
int	free_anglebuffers(void);

/***** wobond.c ******/

int	readin_bondlist(char *psffn,int nr_of_atoms);
int	writeout_bondlist(char *bondlistfn);
int	free_bondlist(void);
int	get_atom_bondlist(int at_nr,int *bondl);

/***** woexcl.c *******/

int	buildup_1excl_list(char *psffn,int nr_of_atoms);
int	free_1excl_buffer(void);

int	writeout_m1excl(char *mexcl1fn,int nr_of_atoms);
int	writeout_1excl(char *excl1fn,int nor_of_atoms);
int	writeout_2excl(char *excl2fn,char *exclm2fn,int nr_of_atoms);
int	writeout_3excl(char *excl3fn,char *exclm3fn,int nr_of_atoms);
int	writeout_4excl(char *excl4fn,char *exclm4fn,int nr_of_atoms);
int	writeout_14excl(char *excl14fn,int nr_of_atoms);

/****** woshake *******/

int	writeout_shakelist(int nr_of_atoms,char *shakelfn);

/****** wovdw.c *********/

int	readin_vdw_params(int argc,char *argv[],int nr_atomtypes);
int	writeout_vdwparams(char *vdwparafn,char *vdw14parafn);

/******* wodihe.c **********/

int	readin_dihepara(int argc,char *argv[]);
int	writeout_dihelist(char *psffn,char *dihelistfn,int nr_of_atoms);
int	writeout_dihepara(char *diheparafn);
int	free_dihebuffers(void);

int	xplor_strcmp(char *fullstr,char *wcstr,int wild_cards_allowed);

/********* woimpr.c *********/

int	readin_imprpara(int argc,char *argv[]);
int	writeout_imprlist(char *psffn,char *imprlistfn,int nr_of_atoms);
int	writeout_imprpara(char *imprparafn);
int	free_imprbuffers(void);

/********* wohbond.c ***********/

int	writeout_hbacclis(char *psffn,char *hbacclistfn,
			int nr_of_atoms,int nr_atomtypes);
int	writeout_hbdonlis(char *psffn,char *hbdonlistfn,int nr_of_atoms);
int	writeout_hbhtype(char *hbtypefn,char *hbtrfn);
int	buildup_hbmats(int argc,char *argv[]);
int	writeout_hbmats(char *hbahmatfn, char *hbbhmatfn);

/********* wounits.c ***********/

int	scan4_units(int nr_of_atoms,int *nr_of_neutrals);
int	writeout_units(char *unitfn);
int	print_restinfo(void);

/********** wodata.c ********/

int	read_next_pdbdata(FILE *fh,	int *atom_nr,
					double *ko,	/* [0][1][2] x y z */
					double *beta,
					double *harmon,
					char *atomstr,
					char *residstr,
					int *residnr,
					char *segidstr);

int	patch_pdb_coords(FILE *fh,FILE *fh1,double x, double y, double z);

double	limit_coords(double coord);

int	read_next_psfatom(FILE *fh,	int *atom_nr,
					char *atom_typestr,
					double *charge,
					double *mass,
					char *atomstr,
					char *residstr,
					int *residnr,
					char *segidstr);

int	read_next_bondpara(FILE *fh,	int *att1,int *att2,
					double *bond_kb,double *bond_r0);

int	read_next_anglepara(FILE *fh,	int *att1,int *att2,int *att3,
					double *angle_kt,double *angle_theta0,
					double *ubk,double *ubd);

int	read_next_vdwpara(FILE *fh,	int *att,
					double *sigma,double *eps,
					double *sigma14,double *eps14);

int	read_next_dihepara(FILE *fh,	char *at1str,char *at2str,char *at3str,char *at4str,
					double *dihe_kd,double *dihe_d,int *dihe_n);

int	read_next_imprpara(FILE *fh,	char *at1str,char *at2str,char *at3str,char *at4str,
					double *impr_ki,double *impr_omega);

int	read_next_nbfixpara(FILE *fh,	int *att1,int *att2,
					double *a,double *b,
					double *a14,double *b14);

int	read_next_hbondpara(FILE *fh,	char *rat1str,char *rat2str,
					double *e_min,double *r_min);

int	read_trjego_header(FILE *fh,int *numcoords,int *framenum,int *framestep,double *timestep);
int	read_eneego_header(FILE *fh,char str[MAX_COMMENT_LINES][MAX_STRING_SIZE], int *rows);
int	read_ego_coords(FILE *fh,double *coords);
int	skip_ego_header(FILE *fh);

/***************************************************************************************/

#endif
