/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: woexcl.c	by M.Eichinger, Munich			**
**	Belongs2: xpl2lis	Converter Xplor-data to EGO-data (*.lis)**
**	Contents: scan in and write out the exclusion list.		**
**									**
**									**
**  	Modif's:							**
**		05.06.94						**
**		07.12.94 impruved error-output				**
**									**
*************************************************************************/
#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <math.h>

#include "hardw.h"
#include "ego.h"
#include "util_lib.h"
#include "wodecls.h"
#include "woalloc.h"

/***** maybe should be added to ego.h ********/
#define	MXNO_SHORT_EXCL		4000
#define LONG_EXCLLIST_FACTOR	5
/***************************************/

#define TEST_SHORTEXCL	0

/******************************************************************************/
/**************** functions in this modul *************************************/
/******************************************************************************/
	int	buildup_1excl_list(char *psffn,int nr_of_atoms);

	int	free_1excl_buffer(void);

	int	writeout_2excl(char *excl2fn,char *exclm2fn,int nr_of_atoms);
	int	writeout_3excl(char *excl3fn,char *exclm3fn,int nr_of_atoms);
	int	writeout_4excl(char *excl4fn,char *exclm4fn,int nr_of_atoms);
	int	writeout_14excl(char *excl14fn,int nr_of_atoms);
	int	writeout_m1excl(char *mexcl1fn,int nr_of_atoms);
	int	writeout_1excl(char *excl1fn,int nr_of_atoms);

static	int	parse_element(int next,int at_nr);
static	int	merge_to_totallist(int nr_to_merge,int *atomlist,int own_atnr);
static	int	goto_psf_specexcl(FILE *h_psf,int *size_short_list);
static	int	write_exclusion_line(FILE *h_excl,int at_nr);
static	int	write_14exclusion_line(FILE *h_excl,int at_nr);
static	int	start_scan_1excl(void);
static	int	get_1excl_list(int at_nr,int *excl1list);

/******************************************************************************/

static	int	maxsize_totl = MXNO_FIELDS_BOND*MXNO_FIELDS_EXCL*MXNO_FIELDS_EXCL;
static	int	cursize_totl = 0;
static	int	totallist[MXNO_FIELDS_EXCL*MXNO_FIELDS_EXCL*MXNO_FIELDS_EXCL+1];

static	int	nr_of_expl_ex=0;

static	int	*expl_ex_list;		/* buffer for exclusion list we want to write*/
static	int	ex_list_len=0;		/* allocated space of that buffer */
static	int	expl_ex_ptr=-1;		/* points to last element */
static	int	expl_ex_list_here=0;

static	int	expl_ex_short[MXNO_SHORT_EXCL+1];
static	int	ex_short_len=MXNO_SHORT_EXCL;
static	int	last_index;

#define true	1
#define false	0

/*************************************************************************************/
/******* the minus-exclusion-lists are corrspondig to the (normal=postiv) ************/
/******* exclusion-lists, but without the explicit exclusion from 1excl.   ***********/
/******* The m1excl-list excludes nothing (-1) ***************************************/

/*****************************************************************************
****************************************************************************** 
  Format of expl_ex_list : 
	<atomnr>, <exclusion1>, <exclusion2>, ..., <END_OF_LIST>, <atomnr>,... 

  Nach dem Keyword !NNB folgen 2 Listen; Die Kurze und die Lange_;
  Die Kurze Liste hat soviele Elemente, wie die Zahl vor !NNB angibt;
  (Zahl der expliciten Ausschluesse). Sie kann u_ U_ auch ganz fehlen_;
  Ihr folgt direkt die Lange Liste, die von fester Laenge ist (fuer jedes;
  Atom im Molekuel ein Eintrag in der Reihenfolge der Atome).;
  Die Eintraege in der Langen Liste sind nun Pointer in die Kurze Liste;
  Der Index der Kurzen Liste beginnt bei 0.;

  Beispiel:
      Lange Liste   (Element)                     Kurze Liste  (Element)
        0             0              -------- >>        5          0
        0             1             /   ------>>       12          1
        1             2         -- /   / ----->>        7          2
        3             3         ------/
        3             4

      Dies bedeutet:
      Atom 1 hat keine expl. Ausschluesse (Element 0 in Langer Liste).
      Atom 2 hat keine expl. Ausschluesse (Element 1 in Langer Liste).
      Atom 3 (Element 2 in LL) schliesst Atom 5 (Element 0 in KL) explicit aus.
      Atom 4 (Element 3 in LL) schliesst Atom 12 und 7 (Element 1 und 2 in KL)
      explicit aus.
      Atom 5 hat keine expl. Ausschluesse (Element 4 in Langer Liste).

   Dabei sind alle CHARMM Listen als ansonsten ungeordnet anzunehmen.;

  Bei mir ist die EXPLICITE EXCLUSION LIST folgendermasen aufgebaut:
  Format: <atom>, <ausschluss1>, <auschluss2>, ..., <end.list>, <atom>,...
  Dabei zeigt expl.ex.ptr auf das letzte belegte Element.
  Die <atom>e sind aufsteigend geordnet; ebenso die jeweilige
  Ausschlussliste.
**********************************************************************************/

/*********************************************************************************/
/** this function has to be called before all other functions in this modul ******/
/**** because they need the special exclusion list *******************************/
/*********************************************************************************/

int	buildup_1excl_list(char *psffn,int nr_of_atoms)
{
	FILE	*h_psf;
	int	i,j;
	int	index,next;

	/********** open PSF file *********************************************/	        
	if ((h_psf=fopen(psffn,"r"))==NULL) {
		errout("Fatal error: Can not read PSF-file.\n");
		return -1;
	}

	if (goto_psf_specexcl(h_psf,&nr_of_expl_ex)==-1) {
		errout("Fatal error: No !NNB section found in PSF-file!\n");
		return -1;
	}

	if (nr_of_expl_ex>ex_short_len) {
		errout("Fatal error: Overflow of the short explicit exclusion list ! %d %d\n",
			nr_of_expl_ex,ex_short_len);
		fclose(h_psf);
		return -1;
  	} 
	/************ alloc buffers *********************************************/
	ex_list_len=LONG_EXCLLIST_FACTOR*nr_of_atoms+100;
	expl_ex_list=alloc_ints(ex_list_len);
	expl_ex_list_here=1;

	/**************** readin short exclusion list ****************************/
#if TEST_SHORTEXCL
	printf("The short exclusion list.\n");
#endif
	for (i = 0; i < nr_of_expl_ex; i++) {
		expl_ex_short[i]= read_int(h_psf);
#if TEST_SHORTEXCL
		if ((i+1)%8==0)	printf("%4d\n",expl_ex_short[i]);
		else		printf("%4d  ",expl_ex_short[i]);
#endif
	}

	expl_ex_ptr=-1;					/* points to last element */
	last_index =0;
	/************** readin long eclusion list and parse **********************/    
	for (i=0; i<nr_of_atoms; i++) {		/* for each atom in long list */
		index=read_int(h_psf);		/* LongList[i],index auf kurze liste */            

		if (index>last_index) {			/* we have new exclusions */
			for (j=last_index;j<index;j++)	{	/* ptr in short list */
				next=expl_ex_short[j]-1; 

				if (parse_element(next,i)==-1)	return -1;
				if (parse_element(i,next)==-1)	return -1;
            		}
			last_index=index;
		}
	}

	txtout("Read special exclusionatoms: %d\n",nr_of_expl_ex);

	return 0;
}

/************************************************************************************/
/****** call this function after all exclusion list has been written ****************/
/************************************************************************************/

int	free_1excl_buffer()
{
	if (expl_ex_list_here) {
		myfree(expl_ex_list);
		expl_ex_list_here=0;
	}
	return 0;
}

/***********************************************************************************/

int	writeout_1excl(char *excl1fn,int nr_of_atoms)
{
    	int	i;
	FILE	*h_1excl;

	if ((h_1excl=fopen(excl1fn,"w"))==NULL) {
		errout("Fatal error: Can not open 1. exclusion-list-file.\n");
		return -1;
	}

	start_scan_1excl();
    
    	for (i=0;i<nr_of_atoms;i++) {

		if ((cursize_totl=get_1excl_list(i,totallist))==-1)	goto error_1excl;

		if (write_exclusion_line(h_1excl,i)==-1)		goto error_1excl;
	}

	fclose(h_1excl);
	txtout(" 1. exclusion-list-file written.\n");
	return 0;

error_1excl:
	fclose(h_1excl);
	errout("Error while building 1. exclusion-list.\n");
	return -1;
}

/************************************************************************************/
/************ the 2. order exclusion-list is excludes all directly bonded atoms *****/
/************ so it is identic with the bonded-list ! *******************************/

int	writeout_2excl(char *excl2fn,char *exclm2fn,int nr_of_atoms)
{
	int	i;
static	FILE	*h_excl,*h_mexcl;
	int	nr_expl_excl;
	int	expl_excl_list[MXNO_FIELDS_EXCL+1];

	if (!expl_ex_list_here)	goto error_2excl;	/* were is the buffer ? */
	
	if ((h_excl=fopen(excl2fn,"w"))==NULL) {
		errout("Can not open 2. exclusion-list-file.\n");
		return -1;
	}

	if ((h_mexcl=fopen(exclm2fn,"w"))==NULL) {
		errout("Can not open -2. exclusion-list-file.\n");
		fclose(h_excl);
		return -1;
	}

	start_scan_1excl();

	for (i=0;i<nr_of_atoms;i++) {			/***** for each atom *****/

		cursize_totl=get_atom_bondlist(i,totallist);
		if (write_exclusion_line(h_mexcl,i)==-1)	goto error_2excl;

						/* add explicit exclusion list ***/
		nr_expl_excl=get_1excl_list(i,expl_excl_list);
		if (merge_to_totallist(nr_expl_excl,expl_excl_list,i)==-1) 
							goto error_2excl;
		if (write_exclusion_line(h_excl,i)==-1)	goto error_2excl;
	}

	txtout(" 2. exclusion-list-file written.\n");
	txtout("-2. exclusion-list-file written.\n");

	fclose(h_excl);
	fclose(h_mexcl);	
	return 0;

error_2excl:
	errout("Error while building +-2. exclusion-list.\n");
	fclose(h_excl);
	fclose(h_mexcl);
	return -1;
}

/***********************************************************************************/
/******* the 3. order exclusion-list are all atoms, which are bonded directly or ***/
/******* via one intermediate atom ! ***********************************************/

int	writeout_3excl(char *excl3fn,char *exclm3fn,int nr_of_atoms)
{
	int	i,j;
	int	nr_bonds1,nr_bonds2,bond_atnr;
static	FILE	*h_excl,*h_mexcl;
	int	nr_expl_excl;
	static	int	bondl1[MXNO_FIELDS_EXCL+1];
	static	int	bondl2[MXNO_FIELDS_EXCL+1];
	static	int	expl_excl_list[MXNO_FIELDS_EXCL+1];

	if (!expl_ex_list_here)	goto error_3excl;	/* were is the buffer ? */

	if ((h_excl=fopen(excl3fn,"w"))==NULL) {
		errout("Fatal error: Can not open 3. exclusion-list-file.\n");
		return -1;
	}
	if ((h_mexcl=fopen(exclm3fn,"w"))==NULL) {
		errout("Fatal error: Can not open -3. exclusion-list-file.\n");
		fclose(h_excl);
		return -1;
	}

	start_scan_1excl();

	for (i=0;i<nr_of_atoms;i++) {			/***** for each atom *****/

		nr_bonds1=get_atom_bondlist(i,bondl1); /*get the directly bonded atoms*/

						/*ini the total bond-list with it*/
		memcpy(totallist,bondl1,sizeof(int)*(nr_bonds1+1));
		cursize_totl=nr_bonds1;
	
		for (j=0;j<nr_bonds1;j++) {
						/* get the via one atom bonded atoms */
			bond_atnr=bondl1[j];		
			nr_bonds2=get_atom_bondlist(bond_atnr,bondl2);
			
			if (merge_to_totallist(nr_bonds2,bondl2,i)==-1) goto error_3excl;
		}
		
		if (write_exclusion_line(h_mexcl,i)==-1)		goto error_3excl;

						/* add the expl exclusion list */	
		nr_expl_excl=get_1excl_list(i,expl_excl_list);
		if (merge_to_totallist(nr_expl_excl,expl_excl_list,i)==-1) 
							goto error_3excl;
		if (write_exclusion_line(h_excl,i)==-1)	goto error_3excl;
	}

	txtout(" 3. exclusion-list-file written.\n");
	txtout("-3. exclusion-list-file written.\n");

	fclose(h_excl);
	fclose(h_mexcl);
	return 0;

error_3excl:
	fclose(h_excl);
	fclose(h_mexcl);
	errout("Error while building +-3. exclusion-list.\n");
	return -1;
}

/**********************************************************************************/
/******* the 4. order exclusion-list are all atoms, which are bonded directly or **/
/******* via one or two intermediate atoms ! **************************************/

int	writeout_4excl(char *excl4fn,char *exclm4fn,int nr_of_atoms)
{
	int	i,j,k;
	int	nr_bonds1,nr_bonds2,nr_bonds3;
	int	bond_atnr;
static	FILE	*h_excl,*h_mexcl;
	static	int	bondl1[MXNO_FIELDS_EXCL+1];
	static	int	bondl2[MXNO_FIELDS_EXCL+1];
	static	int	bondl3[MXNO_FIELDS_EXCL+1];
	static	int	expl_excl_list[MXNO_FIELDS_EXCL+1];
	static	int	nr_expl_excl;

	if (!expl_ex_list_here)	goto error_4excl;	/* were is the buffer ? */
	
	if ((h_excl=fopen(excl4fn,"w"))==NULL) {
		errout("Fatal error: Can not open 4. exclusion-list-file.\n");
		return -1;
	}

	if ((h_mexcl=fopen(exclm4fn,"w"))==NULL) {
		errout("Fatal error: Can not open -4. exclusion-list-file.\n");
		fclose(h_excl);
		return -1;
	}

	start_scan_1excl();

	for (i=0;i<nr_of_atoms;i++) {			/***** for each atom *****/

		nr_bonds1=get_atom_bondlist(i,bondl1);/*get the directly bonded atoms*/

						/*ini the total bond-list with it*/
		memcpy(totallist,bondl1,sizeof(int)*(nr_bonds1+1));
		cursize_totl=nr_bonds1;
	
		for (j=0;j<nr_bonds1;j++) {
						/* get the via one atom bonded atoms*/
			bond_atnr=bondl1[j];		
			nr_bonds2=get_atom_bondlist(bond_atnr,bondl2);
			
			if (merge_to_totallist(nr_bonds2,bondl2,i)==-1) goto error_4excl;
					
			for (k=0;k<nr_bonds2;k++) {
						/*get the via two atom bonded atoms*/
				bond_atnr=bondl2[k];
				nr_bonds3=get_atom_bondlist(bond_atnr,bondl3);
				
				if (merge_to_totallist(nr_bonds3,bondl3,i)==-1) 
							goto error_4excl;
			}
		}

		if (write_exclusion_line(h_mexcl,i)==-1) goto error_4excl;

							/* add the expl exclusion list */	
		nr_expl_excl=get_1excl_list(i,expl_excl_list);
		if (merge_to_totallist(nr_expl_excl,expl_excl_list,i)==-1)
							goto error_4excl;
		if (write_exclusion_line(h_excl,i)==-1) goto error_4excl;
	}

	txtout(" 4. exclusion-list-file written.\n");
	txtout("-4. exclusion-list-file written.\n");

	fclose(h_excl);
	fclose(h_mexcl);
	return 0;

error_4excl:
	errout("Error while building +-4. exclusion-list.\n");
	fclose(h_excl);	
	fclose(h_mexcl);
	return -1;
}

/*********************************************************************************/
/******* the special 1-4 list are all atoms, which are bonded exactly ************/
/******* via two intermediate atoms ! ********************************************/

int	writeout_14excl(char *excl14fn,int nr_of_atoms)
{
	int	i,j,k;
	int	nr_bonds1,nr_bonds2,nr_bonds3,nr_bondvia1;
	int	bond_atnr,l,s;
	FILE	*h_excl;
	static	int	bondl1[MXNO_FIELDS_EXCL+1];
	static	int	bondl2[MXNO_FIELDS_EXCL+1];
	static	int	bondl3[MXNO_FIELDS_EXCL+1];
	static	int	bondvia1[MXNO_FIELDS_EXCL*MXNO_FIELDS_EXCL+1];
	
	if ((h_excl=fopen(excl14fn,"w"))==NULL) {
		errout("Fatal error: Can not write special 1-4 list-file.\n");
		return -1;
	}

	for (i=0;i<nr_of_atoms;i++) {			/***** for each atom *****/

		nr_bonds1=get_atom_bondlist(i,bondl1);	/*get the directly bonded atoms*/

					/* ini the total bond-list with END_OF_LIST */
		totallist[0]=END_OF_LIST;
		cursize_totl=0;
			
		for (j=0;j<nr_bonds1;j++) {
					/* get all via one to atom i bonded atoms */
			bondvia1[0]=-1;
			nr_bondvia1=0;
			for (k=0;k<nr_bonds1;k++) {
				if (k==j)	continue;		
				bond_atnr=bondl1[k];		
				nr_bonds2=get_atom_bondlist(bond_atnr,bondl2);
				for (l=0;l<nr_bonds2;l++) {
					bondvia1[nr_bondvia1+l]=bondl2[l];
					bondvia1[nr_bondvia1+l+1]=-1;
				}
				nr_bondvia1+=nr_bonds2;
			}			
							/* take that branch */
			bond_atnr=bondl1[j];		
			nr_bonds2=get_atom_bondlist(bond_atnr,bondl2);

			for (k=0;k<nr_bonds2;k++) {
				int	in_list=0;
						/* get the via two atom bonded atoms */
				bond_atnr=bondl2[k];
				if (bond_atnr==i)	continue; /* skip that case */

				/****** delete all via 1 bonded atoms out of bondl3 */
				nr_bonds3=get_atom_bondlist(bond_atnr,bondl3);
				for (s=0;s<nr_bondvia1;s++) {
					in_list=0;
					for (l=0;l<nr_bonds3;l++) {
						if (bondvia1[s]==bondl3[l]) {
							in_list=1;
							break;
						}
					}
					if (in_list) {
						int m;
						for (m=l;m<nr_bonds3;m++) {
							bondl3[m]=bondl3[m+1];
						}
						nr_bonds3--;
					}					

				}

				/******************* skip this atom-nr ---\ ! ********/
				if (merge_to_totallist(nr_bonds3,bondl3,bondl1[j])==-1)
								goto error_14excl;
			}
		}

		if (write_14exclusion_line(h_excl,i)==-1)	goto error_14excl;
	}

	txtout("Special 1-4 list-file written.\n");
	fclose(h_excl);
	return 0;

error_14excl:
	errout("Error while building 1-4-exclusion-list.\n");
	fclose(h_excl);
	return -1;
}

/*************************************************************************************/

int	writeout_m1excl(char *mexcl1fn,int nr_of_atoms)
{
	int	i;
	FILE	*h_excl;

	if ((h_excl=fopen(mexcl1fn,"w"))==NULL) {
		errout("Fatal error: Can not open -1. exclusion-list-file.");
		return -1;
	}

	for (i=0;i<nr_of_atoms;i++) {			/***** for each atom *****/
		fprintf(h_excl,"Atom  %4d  excludes    %4d\n",i,END_OF_LIST);
	}

	txtout("-1. exclusion-list-file written.\n");

	fclose(h_excl);
	return 0;
}


/************************************************************************************/
/****** all atoms in atomlist are sorted to the totallist. You can specifie one atom_nr *****/
/****** which must not be added to the totallist, even if this number occures in atomlist ***/
static
int	merge_to_totallist(int nr_to_merge,int *atomlist,int skip_atnr)
{
	int	i,j,tmp;
	int	already_in_list;
	
	if (nr_to_merge==0)	return 0;

	for (i=0;i<nr_to_merge;i++) {
							/* don't need this atom nr */
		if (atomlist[i]==skip_atnr)	continue;
					
		already_in_list=0;			/* test for new atom */
		for (j=0;j<cursize_totl;j++) {
			if (totallist[j]==atomlist[i]) {
				already_in_list=1;
				break;
			}
		}		
		if (already_in_list)	continue;	/* try next atom in list */

		/********** new atom found --> add it to totallist *****************/
		totallist[cursize_totl]  =atomlist[i];	/*add to end of sorted list*/
		totallist[cursize_totl+1]=END_OF_LIST;
		
		for (j=cursize_totl;j>0;j--) {		/* bubble-sort */
			if (totallist[j]<totallist[j-1]) {
				tmp=totallist[j-1];
				totallist[j-1]=totallist[j];
				totallist[j]=tmp;
			} else	break;			/* sorted to right place */
		}

		if (++cursize_totl>=maxsize_totl) {
			errout("Buffer to small in merge_to_totallist.\n");
			return -1;
		}
	}
	return 0;
}

/***********************************************************************************/

static
int	goto_psf_specexcl(FILE *h_psf,int *size_short_list)
{
	static	char	lstr[MAX_STRING_SIZE];
	static	char	tmpstr[MAX_STRING_SIZE];
	int 	res;

	while (read_string(h_psf,lstr)!=-1) {			/* read a psf-line */
		strup(lstr);
		if (sscanf(lstr,"%d %s",&res,tmpstr)==2) {	/* check it */
			tmpstr[4]=0;
			if (strcmp(tmpstr,"!NNB")==0) {
				*size_short_list=res;
				return 0;
			}
		}	
	}
        *size_short_list=0;
	return -1;
}

/**************** write it out to file ********************************/	

static
int	write_14exclusion_line(FILE *h_excl,int at_nr)
{
	int	j,tmp;
		
	fprintf(h_excl,"Atom  %4d   interacts with  ",at_nr);

	for (j=0;j<=cursize_totl;j++) {		/***** for each exclusion *****/
		tmp=totallist[j];

		if (tmp!=END_OF_LIST)	fprintf(h_excl," %4d ",tmp);
		else			fprintf(h_excl," %4d\n",tmp);
	}
	return 0;
}

/**************************************************************************/

static
int	write_exclusion_line(FILE *h_excl,int at_nr)
{
	int	j,tmp;
		
	fprintf(h_excl,"Atom  %4d  excludes   ",at_nr);
		
	for (j=0;j<=cursize_totl;j++) {		/***** for each exclusion *****/
		tmp=totallist[j];

		if (tmp!=END_OF_LIST)	fprintf(h_excl," %4d ",tmp);
		else			fprintf(h_excl," %4d\n",tmp);
	}
	return 0;
}

/*********************************************************************************/

static	int	expl_atom,expl_ptr;

static	
int	start_scan_1excl(void)
{
	if (expl_ex_ptr==-1) {			/* no explicit exclusion list at all*/
		expl_atom = -1; 		/*-1 means, no exclusions for this atom*/
	} else {				/* there are exclusions */
		expl_atom=expl_ex_list[0];	/* at-nr for for first atom with exclusions */
		expl_ptr=1;			/* ptr to first element */
	}
	return 0;
}

static
int	get_1excl_list(int at_nr,int *excl1list)
{
	int	ex_len;

	if (at_nr!=expl_atom) {		/* no exclusion list for this atom-nr */
		excl1list[0]=END_OF_LIST;	
		return 0;
	} else {			/* copy exclusion list to excl1list-buffer */
		ex_len=0;
		while ((excl1list[ex_len]=expl_ex_list[expl_ptr+ex_len])!=END_OF_LIST) {
			ex_len++;
       		}
		expl_ptr+=ex_len+1;

		if (ex_len>=MXNO_FIELDS_EXCL) {
			errout("Buffer to small in get_1excl_list. Increase MXNO_FIELDS_EXCL in ego.h.\n");	
			return -1;
		}

		if (expl_ex_ptr<=expl_ptr) {	/* no more explicit exclusions */
          		expl_atom=END_OF_LIST; 	
        	} else {			/* get next at-nr with exclusions */
          		expl_atom=expl_ex_list[expl_ptr];
          		expl_ptr++;
       		}
		return ex_len;
	}
}

/**********************************************************************************/
/****** that is the function that expands the nasty comppressed exclusion list ****/
/**********************************************************************************/

static
int	parse_element(int next,int at_nr)
{	
	int	h_ptr=0;	/* help-pointer */
	int	searching;
	int	og;
	int	i,jj;

	if (next==at_nr) {
  				/* EGO excludes this automatically */
	} else {		

		while (	(expl_ex_ptr-h_ptr>=0) && 
			(expl_ex_list[h_ptr]<next) ) {
			while (expl_ex_list[h_ptr]!=END_OF_LIST)	h_ptr++;
			h_ptr++;			/* points to new root */
            	}

		if (expl_ex_ptr-h_ptr<0) {		/* add to end of list */
			if (expl_ex_ptr+3>=ex_list_len)	goto error_expl_list;
 			expl_ex_list[++expl_ex_ptr]=next;
			expl_ex_list[++expl_ex_ptr]=at_nr;
			expl_ex_list[++expl_ex_ptr]=END_OF_LIST;
		} else 
		if (expl_ex_list[h_ptr]==next) {	/* insert to list */
			h_ptr++;	/* h_ptr points now to first elemnt to move */
			searching=true;
			while (searching) {
							/* skip it */
				if (expl_ex_list[h_ptr]==at_nr)  searching = false; 		else
				if (	(expl_ex_list[h_ptr]>at_nr) ||
					(expl_ex_list[h_ptr]==END_OF_LIST) ) {
					if (expl_ex_ptr+1>=ex_list_len) 
							goto error_expl_list;
         				og=expl_ex_ptr+1;
					for (i=1;i<1+og-h_ptr;i++) {
						jj=og-i;	/* step back */
						expl_ex_list[jj+1]=expl_ex_list[jj];	
								/* shift 1 place */
					} 																		
					expl_ex_list[h_ptr]=at_nr;/* ins new entry */
					expl_ex_ptr++;
	  				searching=false;
				} else
				if (expl_ex_list[h_ptr]<at_nr) h_ptr++;
			}
		} else
		if (expl_ex_list[h_ptr]>next) {
       					/*h_ptr points to first element to move */
			if (expl_ex_ptr+3>=ex_list_len) goto error_expl_list;
			og=expl_ex_ptr+1;
			for (i=1;i<1+og-h_ptr;i++) {
				jj=og-i;			/* step back */
				expl_ex_list[jj+3]=expl_ex_list[jj]; /* shift 3 places */
			}
			expl_ex_list[h_ptr++]=next;		/* ins new entrys */
			expl_ex_list[h_ptr++]=at_nr;
			expl_ex_list[h_ptr]  =END_OF_LIST;
			expl_ex_ptr+=3;
		}
	}
	return 0;

error_expl_list:
	errout("Fatal error: Not enough space for special explicit exclusion list !\n");
	errout("Increase LONG_EXCLLIST_FACTOR in woexcl.c !\n");
	return -1;
}
