/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: wohbond.c	by M.Eichinger, Munich  		**
**	Belongs2: xpl2lis	Converter Xplor-data to EGO-data (*.lis)**
**	Contents: scan in and write out H-Bond-Acceptor-list.		**
**		  scan in and write out H-Bond-Donator-list.		**
**		  finds and write out the different H-Types, that make 	**
**		  H-Bonds. find and check the H-Bond-Parameters 	**
**		  scan in the H-Bond parameters and check for exponents ok**
** 									**
**  	Modif's:							**
**		26.06.94						**
**		11.07.94 Rest of the H-bond-stuff			**
**									**
*************************************************************************/

#include <math.h>

#include "hardw.h"
#include "ego.h"
#include "woalloc.h"
#include "util_lib.h"
#include "wodecls.h"


#define	TEST_HBONBPARA		0
#define TEST_HBOND_COMPARE	0

/******************************************************************************/
/**************** functions in this modul *************************************/
/******************************************************************************/

/** do not change the calling order, because they all manipulate translist[..] ****/

	int	writeout_hbacclis(char *psffn,char *hbacclistfn,int nr_of_atoms,
				int nr_atomtypes);
	int	writeout_hbdonlis(char *psffn,char *hbdonlistfn,int nr_of_atoms);
	int	buildup_hbmats(int argc,char *argv[]);
	int	writeout_hbhtype(char *hbtypefn,char *hbtrfn);
	int	writeout_hbmats(char *hbahmatfn, char *hbbhmatfn);

static	int	readin_hbond_params(FILE *h_para);
static	int	check_hbond_exponent(FILE *fh);
static 	int	get_hbondpara_index(int i, int j);

static	int	add_to_hbacclist(int *hbacclist,int *hbaccdubs,int nr_read_hbacc);
static	int	get_next_hbaccdubs(FILE *h_psf,int *hbaccdubs);
static	int	goto_psf_hbaccdata(FILE *h_psf,int *nr_of_hbacc);

static	int	add_to_hbdonlist(int *hbdonlist,int *hbdondubs,int nr_read_hbdon);
static	int	get_next_hbdondubs(FILE *h_psf,int *hbdondubs);
static	int	goto_psf_hbdondata(FILE *h_psf,int *nr_of_hbdon);

/*******************************************************************************/

typedef struct {
	char	wcstr[2][2*LEN_ATOM_TYPE+1];	/*holds the wild-cards.may be slightly */
						/* more then LEN_ATOM_TYPE. I think */ 
						/* factor 2 is enough for it.*/
	double	e_min;
	double	r_min;
} hbondpara_type;	
						/* holds the h-bond parameters */	
static	hbondpara_type	hbpara[MXNO_HBOND_PARAMS+1];
static	int	nr_read_hbpara=0;


static	int	translist[MXNO_ATOM_TYPES];/* 255 if not making H-Bond, otherwise */
					   /* index to the h-bond-matrix with range */
					   /* 0,...,nr_hbond_atoms-1 */ 
static	int	nr_hbond_atoms=0;

static	int	nr_atomtypes=0;		/* needed often, set by writeout_hbacclis */

static	double	*amat;
static	double	*bmat;
static	int	abmat_here=0;

static	int	good_exponents=0;	/* counts, wheather we got 4 exponents */

/***************************************************************************************
      -- The A and B paramters are based on the hbond function below:
      --
      --                        A         B
      --             E (r) =  -----  -  -----  .
      --                         6         4
      --                        r         r
      --
      -- A and B have to be calculated from the Emin, Rmin given in
      -- the paramter file. Emin is the well depth, Rmin the distance
      -- of the minimum. Derivation dE/dr = 0 and subsequent calculation
      -- of E yields:
      --                        A = -2 Emin Rmin^6
      --                        B = -3 Emin Rmin^4.
      --
      -- These formulas are used for the calculation of the parameter
      -- matrices.

      -- Since we don't know in advance, WHICH atom types will really H-bond
      -- later on, we have to provide complete matrices.
      -- By selecting only the possible candidates (nr_hbond_atoms), we already 
	 narrowed down the choice a little bit.
*************************************************************************************/

int	writeout_hbacclis(char *psffn,char *hbacclistfn,int nr_of_atoms,int nr_att)
{
	int	i,res1,off=0;
static	int	*hbacclist;
	int	atom1,atom2;
	int	nr_of_hbacc=0,nr_read_hbacc=0;
	int	hbaccdubs[2*4],nr_converted=0;
	int	hbacclistbuf_here=0;
	FILE	*h_psf,*h_hbacc;
	
	nr_atomtypes=nr_att;
	if ((h_psf=fopen(psffn,"r"))==NULL) {
		errout("Can not open specified psf-file.\n");
		return -1;
	}

	if (goto_psf_hbaccdata(h_psf,&nr_of_hbacc)==-1) {
		goto write_hbacc_list;		/* no hbacclist-section ? */
	}

	hbacclist=alloc_ints((nr_of_hbacc+10)*2);
	hbacclistbuf_here=1;

	/******************** build up sorted hbacclist ******************************/

	while ((nr_converted=get_next_hbaccdubs(h_psf,hbaccdubs))>0) {
		for (i=0;i<nr_converted;i+=2) {
	
			atom1=hbaccdubs[i  ];		/* just for better reading */
			atom2=hbaccdubs[i+1];
							/* some tests 	*/		
			if (	(atom1<-1) || (atom1>=nr_of_atoms) ||
				(atom2<-1) || (atom2>=nr_of_atoms) ) {
				errout("Fatal error: Illegal atomnumber in PSF-file.\n");
				goto error_in_hbacclist;	
			}		
		
			res1=add_to_hbacclist(hbacclist,hbaccdubs+i,nr_read_hbacc);
						
			if (res1==-1) goto error_in_hbacclist;
				
			if (++nr_read_hbacc>nr_of_hbacc) {
					errout("Fatal error: Incompatible number (to much !) of hb-acceptors in PSF-file.\n");
					goto error_in_hbacclist;
			}
		} /******* for next: for each dublett in line */

	} /******* while: for each line in hbacc-pair-section */

	if (nr_converted<0)	goto error_in_hbacclist;

	if (nr_read_hbacc!=nr_of_hbacc) {
		errout("Fatal error: Incompatible number (to less !) of hb-acceptors in PSF-file\n");
		goto error_in_hbacclist;		
	}
		
	/******************** write out hb-acc to file ****************************/
write_hbacc_list:	
	if ((h_hbacc=fopen(hbacclistfn,"w"))==NULL) {
		errout("Can not open hb-acceptor-list-file.\n");
		goto error_in_hbacclist;
	}
	
	fprintf(h_hbacc,"Number of acceptors:  %4d .\n",nr_of_hbacc);
	fprintf(h_hbacc,"Format of the acceptor list:\n");
	fprintf(h_hbacc,"Acceptor   Acceptor.h\n");

	for (i=0,off=0;i<nr_of_hbacc;i++,off+=2) {		/*** for each hbacc **/
		fprintf(h_hbacc,"%4d         %4d\n",hbacclist[off],hbacclist[off+1]);
	}

	txtout("Read hb-acceptors in PSF-file: %d\n",nr_of_hbacc);

	if (hbacclistbuf_here)	myfree(hbacclist);
	fclose(h_psf);
	fclose(h_hbacc);

	return 0;

error_in_hbacclist:
	errout("Error after %d read hb-acceptors.\n",nr_read_hbacc);

	if (hbacclistbuf_here) myfree(hbacclist);
	fclose(h_psf);

	return -1;
}

/**************************************************************************************/

int	writeout_hbdonlis(char *psffn,char *hbdonlistfn,int nr_of_atoms)
{
	int	i,res1,off=0;
static	int	*hbdonlist;
	int	atom1,atom2;
	int	nr_of_hbdon=0,nr_read_hbdon=0;
	int	hbdondubs[2*4],nr_converted=0;
	int	hbdonlistbuf_here=0;
	FILE	*h_psf,*h_hbdon;
	
	/******************  open PSF-file and alloc space for hbdonlist ***************/

	if ((h_psf=fopen(psffn,"r"))==NULL) {
		errout("Can not open specified psf-file.\n");
		return -1;
	}

	if (goto_psf_hbdondata(h_psf,&nr_of_hbdon)==-1) {
		goto write_hbdon_list;		/* no hbdonlist-section ? */
	}

	hbdonlist=alloc_ints((nr_of_hbdon+10)*2);
	hbdonlistbuf_here=1;

	/******************** build up sorted hbdonlist ******************************/

	while ((nr_converted=get_next_hbdondubs(h_psf,hbdondubs))>0) {
		for (i=0;i<nr_converted;i+=2) {
	
			atom1=hbdondubs[i  ];		/* just for better reading */
			atom2=hbdondubs[i+1];
							/* some tests 	*/		
			if (	(atom1<0) || (atom1>=nr_of_atoms) ||
				(atom2<0) || (atom2>=nr_of_atoms) ) {
				errout("Fatal error: Illegal atomnumber in PSF-file.\n");
				goto error_in_hbdonlist;	
			}		
		
			res1=add_to_hbdonlist(hbdonlist,hbdondubs+i,nr_read_hbdon);
						
			if (res1==-1) goto error_in_hbdonlist;

			if (++nr_read_hbdon>nr_of_hbdon) {
				errout("Fatal error: Incompatible number (to much !) of hb-donators in PSF-file\n");
				goto error_in_hbdonlist;
			}
		} /******* for next: for each dublett in line */

	} /******* while: for each line in hbdon-pair-section */

	if (nr_converted<0)	goto error_in_hbdonlist;

	if (nr_read_hbdon!=nr_of_hbdon) {
		errout("Fatal error: Incompatible number (to less !) of hb-donators in PSF-file\n");
		goto error_in_hbdonlist;		
	}
		
	/******************** write out hb-don-list to file *************************/
write_hbdon_list:	
	if ((h_hbdon=fopen(hbdonlistfn,"w"))==NULL) {
		errout("Can not open hb-donator-list-file.\n");
		goto error_in_hbdonlist;
	}
	
	fprintf(h_hbdon,"Number of donators:  %4d .\n",nr_of_hbdon);
	fprintf(h_hbdon,"Format of the donator list:\n");
	fprintf(h_hbdon,"Donator   Donator.h\n");

	for (i=0,off=0;i<nr_of_hbdon;i++,off+=2) {		/** for each hbdon *****/
		fprintf(h_hbdon,"%4d         %4d\n",hbdonlist[off+0],hbdonlist[off+1]);
	}

	txtout("Read hb-donators  in PSF-file: %d\n",nr_of_hbdon);

	if (hbdonlistbuf_here)	myfree(hbdonlist);
	fclose(h_psf);
	fclose(h_hbdon);

	return 0;

error_in_hbdonlist:
	errout("Error after %d read hb-donators.\n",nr_read_hbdon);

	if (hbdonlistbuf_here) myfree(hbdonlist);
	fclose(h_psf);

	return -1;
}

/***************************************************************************************/

int	writeout_hbhtype(char *hbtypefn,char *hbtrfn)
{
	FILE	*h_hbht,*h_hbtr;
	int	i;
	int	nr_diff_h=0;	/* counts the number of different hydrogen-atoms */
	
	/************* write out the translation list *********************************/
	if ((h_hbtr=fopen(hbtrfn,"w"))==NULL) {
		errout("Can not open HBond-translation-list-file.\n");
		return -1;
	}

	fprintf(h_hbtr,"This is the HBOND translation list.\n");	
	fprintf(h_hbtr,"The number of different atom types is:  %4d\n",nr_atomtypes);

	for (i=0;i<nr_atomtypes;i++) {
		if ((i+1)%MXNO_FIELDS_BYTE==0)	fprintf(h_hbtr,"%4d\n",translist[i]);
		else				fprintf(h_hbtr,"%4d  ",translist[i]);
	}

	fprintf(h_hbtr,"\n");		

	txtout("H-Bond-translation-list written.\n");
	fclose(h_hbtr);

	/********************* write hydrogen type file ********************************/
	for (i=0;i<nr_atomtypes;i++) {	/* here we just count the hydrogens */
		double mass=get_atom_mass(i);

		if (isit_a_hydrogen(mass)) {
			nr_diff_h++;
		}
	}

	if ((h_hbht=fopen(hbtypefn,"w"))==NULL) {
		errout("Can not open hydrogen-type-file.\n");
		return -1;
	}

	fprintf(h_hbht,"These are the different hydrogen types. Total number on first line.\n");	
	fprintf(h_hbht,"%4d\n",nr_diff_h);

	for (i=0;i<nr_atomtypes;i++) {
		double mass=get_atom_mass(i);
		if (isit_a_hydrogen(mass)) {
			fprintf(h_hbht,"%4d   ",i);
		}
	}
	fprintf(h_hbht,"\n");	

	txtout("Different hydrogens types    : %d\n",nr_diff_h);

	fclose(h_hbht);

	return 0;
}

/**********************************************************************************/

int	buildup_hbmats(int argc,char *argv[])
{
		
	int	i,j,file_ind,res1=0,res2=0;
	FILE	*h_para;
	int	para_ind=-1;
	double	r4,r6;
	static	char	att1str[LEN_ATOM_TYPE+1];

	nr_read_hbpara=0;
	/******************* scan each parmater-file *********************************/
	/******************* for h-bond-parameters ***********************************/
	for (file_ind=3;file_ind<argc;file_ind++) {
		if ((h_para=fopen(argv[file_ind],"r"))==NULL) {
			errout("Can not read parameter-file: %s\n",argv[file_ind]);
			return -1;
		}

		if (readin_hbond_params(h_para)==-1) {
			fclose(h_para);
			errout("Error occured in parameter-file: %s\n",argv[file_ind]);
			return -1;
		}
		
		fclose(h_para);
	}

	txtout("Found H-bond-parameters      : %d\n",nr_read_hbpara);

	/******** now find all atom-types used in our data-file, for which ********/
	/******** we have some h-bond-parameter-datas *****************************/
	nr_hbond_atoms=0;
	for (i=0;i<nr_atomtypes;i++) {	/* for all atom-types */
		translist[i]=255;
		get_atomtype_str(i,att1str);

		for (j=0;j<nr_read_hbpara;j++) {	/* for all h-bond-datas */
			res1=xplor_strcmp(att1str,hbpara[j].wcstr[0],1);
			res2=xplor_strcmp(att1str,hbpara[j].wcstr[1],1);
			if ((res1==-1)||(res2==-1)) {
				return -1;		/* error */
			}
			
			if ((res1==1)||(res2==1)) {	/* fill translation-list */
				translist[i]=nr_hbond_atoms;
				nr_hbond_atoms++;
				break;
			}
		}

	}

	txtout("Found H-bond-atomtypes       : %d\n",nr_hbond_atoms);

	/************* allocate and ini A-B-matrix with zero *************************/
	amat=alloc_doubles(nr_hbond_atoms*nr_hbond_atoms+10);
	bmat=alloc_doubles(nr_hbond_atoms*nr_hbond_atoms+10);
	abmat_here=1;

	for (i=0;i<nr_hbond_atoms;i++) {
		for (j=0;j<nr_hbond_atoms;j++) {
			amat[i*nr_hbond_atoms+j]=0.0;
			bmat[i*nr_hbond_atoms+j]=0.0;
		}
	}

	/******************* Now fill the A-B-Matrix with the values wo got ************/
	for (i=0;i<nr_hbond_atoms;i++) {
		for (j=0;j<nr_hbond_atoms;j++) {
			if (i==j) {			/* diagonal-elements are zero */
				bmat[i*nr_hbond_atoms+j]= (double) 0.0;
				amat[i*nr_hbond_atoms+j]= (double) 0.0;
			} else {
				para_ind=get_hbondpara_index(i,j);
				if (para_ind==-1) continue;

				r4=hbpara[para_ind].r_min;
				r4 *= r4;	r6 = r4;
				r4 *= r4;
				r6 *= r4;
				bmat[i*nr_hbond_atoms+j]= -3.0*r4*hbpara[para_ind].e_min;
				amat[i*nr_hbond_atoms+j]= -2.0*r6*hbpara[para_ind].e_min;
			}
		}
	}
	if (nr_read_hbpara==0) return 0;	/* no need for check of exponents */

	/*** here we read again all parameters-files for the HBond Exponents **********/
	/*** In EGO only certain exponents are allowed. Here we check for it. *********/
	/*** Exponents are defined in ego.h (AEXP,REXP,HAEX,HAEX) *********************/
	/******************* scan each parmater-file **********************************/
	good_exponents=0;
	for (file_ind=3;file_ind<argc;file_ind++) {
		if ((h_para=fopen(argv[file_ind],"r"))==NULL) {
			errout("Can not read parameter-file: %s\n",argv[file_ind]);
			myfree(amat);	myfree(bmat);
			abmat_here=0;
			return -1;
		}

		if (check_hbond_exponent(h_para)==-1) {
			fclose(h_para);
			myfree(amat);	myfree(bmat);
			abmat_here=0;
			errout("Error occured in parameter-file: %s\n",argv[file_ind]);
			return -1;
		}
		
		fclose(h_para);
	}
	if (good_exponents==4) {
		txtout("H-Bond-exponents are OK for EGO.\n");
	} else {
		errout("Warning error: Not all H-Bond-exponents found in parameter-files.\n");
		errout("Taking default-values for EGO.\n");
	}
	
	/**** all went ok *****/	
	return 0;
}

static 
int	get_hbondpara_index(int i, int j)	/* i,j are indices to the Matrix */
{						/* of possibly making H-Bonds */
	int 	k=0,l=0,res1=0,res2=0;		/* indices to the Atom-Type-List */
	int	c=0,index=-1;
	static	char	att1str[LEN_ATOM_TYPE+1];
	static	char	att2str[LEN_ATOM_TYPE+1];
						/* OK,OK,... not fast, but easy ! */
			/* find atom-type-index (k,l) to given matrix-element (i,j) */ 	
	for (k=0;k<nr_atomtypes;k++) if (translist[k]==i) break;
	for (l=0;l<nr_atomtypes;l++) if (translist[l]==j) break;
	
	if ((k>=nr_atomtypes)||(l>=nr_atomtypes)) {
		errout("Fatal error: No matches found in get_hbondpara_index.\n");
		errout("That can not be ! Check the Program.\n");
		return -1;
	}

	get_atomtype_str(k,att1str);
	get_atomtype_str(l,att2str);

	/********** try with no wildcards allowed ************************/
	for (c=0;c<nr_read_hbpara;c++) {	/* for each h-bond-parameter */ 
		res1=xplor_strcmp(att1str,hbpara[c].wcstr[0],0);
		res2=xplor_strcmp(att2str,hbpara[c].wcstr[1],0);

		if ((res1==-1)||(res2==-1))	return -1;	/* thats an error */
		
		if ((res1==1)&&(res2==1)) {			/* both match */
			if (index==-1) {			/* first time */
				index = c;
			} 				/* more params */
			else				/* ==> check of consistency */
			if ((fabs(hbpara[c].e_min-hbpara[index].e_min)>SMALL) ||
			    (fabs(hbpara[c].r_min-hbpara[index].r_min)>SMALL) ) {
				errout("Warning error: Different H-Bond-Params for <%s> <%s>\n",
					att1str,att2str);
				errout("Old vals(Emin,Rmin): (%f,%f) Skipped: (%f,%f).\n",
				hbpara[index].e_min,hbpara[index].r_min,
				hbpara[c].e_min,hbpara[c].r_min);
			}
		}
	}
	if (index!=-1)	return index;

	/********** try with one wildcard allowed ************************/
	for (c=0;c<nr_read_hbpara;c++) {	/* for each h-bond-parameter */ 
		res1=xplor_strcmp(att1str,hbpara[c].wcstr[0],1);
		res2=xplor_strcmp(att2str,hbpara[c].wcstr[1],0);

		if ((res1==-1)||(res2==-1))	return -1;	/* thats an error */
		
		if ((res1==1)&&(res2==1)) {			/* both match */
			if (index==-1) {			/* first time */
				index = c;
			} 				/* more params */
			else				/* ==> check of consistency */
			if ((fabs(hbpara[c].e_min-hbpara[index].e_min)>SMALL) ||
			    (fabs(hbpara[c].r_min-hbpara[index].r_min)>SMALL) ) {
				errout("Warning error: Different H-Bond-Params for <%s> <%s>\n",
					att1str,att2str);
				errout("Old vals(Emin,Rmin): (%f,%f) Skipped: (%f,%f).\n",
				hbpara[index].e_min,hbpara[index].r_min,
				hbpara[c].e_min,hbpara[c].r_min);
			}
		}
	}
	if (index!=-1)	return index;

	/********** try with one wildcard allowed ************************/
	for (c=0;c<nr_read_hbpara;c++) {	/* for each h-bond-parameter */ 
		res1=xplor_strcmp(att1str,hbpara[c].wcstr[0],0);
		res2=xplor_strcmp(att2str,hbpara[c].wcstr[1],1);

		if ((res1==-1)||(res2==-1))	return -1;	/* thats an error */
		
		if ((res1==1)&&(res2==1)) {			/* both match */
			if (index==-1) {			/* first time */
				index = c;
			} 				/* more params */
			else				/* ==> check of consistency */
			if ((fabs(hbpara[c].e_min-hbpara[index].e_min)>SMALL) ||
			    (fabs(hbpara[c].r_min-hbpara[index].r_min)>SMALL) ) {
				errout("Warning error: Different H-Bond-Params for <%s> <%s>\n",
					att1str,att2str);
				errout("Old vals(Emin,Rmin): (%f,%f) Skipped: (%f,%f).\n",
				hbpara[index].e_min,hbpara[index].r_min,
				hbpara[c].e_min,hbpara[c].r_min);
			}
		}
	}
	if (index!=-1)	return index;

	/********** try with all wildcards allowed ************************/
	for (c=0;c<nr_read_hbpara;c++) {	/* for each h-bond-parameter */ 
		res1=xplor_strcmp(att1str,hbpara[c].wcstr[0],1);
		res2=xplor_strcmp(att2str,hbpara[c].wcstr[1],1);

		if ((res1==-1)||(res2==-1))	return -1;	/* thats an error */
		
		if ((res1==1)&&(res2==1)) {			/* both match */
			if (index==-1) {			/* first time */
				index = c;
			} 				/* more params */
			else				/* ==> check of consistency */
			if ((fabs(hbpara[c].e_min-hbpara[index].e_min)>SMALL) ||
			    (fabs(hbpara[c].r_min-hbpara[index].r_min)>SMALL) ) {
				errout("Warning error: Different H-Bond-Params for <%s> <%s>\n",
					att1str,att2str);
				errout("Old vals(Emin,Rmin): (%f,%f) Skipped: (%f,%f).\n",
					hbpara[index].e_min,hbpara[index].r_min,
					hbpara[c].e_min,hbpara[c].r_min);
			}
		}
	}

	return index;		/* index to the H-Bond-Parameters-list */
}


static
int	check_hbond_exponent(FILE *fh)
{
	static	char	lstr[MAX_STRING_SIZE];
	static	char	tmpstr[MAX_STRING_SIZE];
	double	exponent;
	int	converted=0;

	while (read_string(fh,lstr)!=-1) {
		tmpstr[0]=0;
		exponent =0.0;
		strup(lstr);

		converted=sscanf(lstr,"%s %lf",tmpstr,&exponent);
		tmpstr[4]=0;	/* truncate keyword */

		if (strcmp(tmpstr,"AEXP")==0) {		/* found one */
			if (converted!=2)	goto illegal_exppara_line;
			if (fabs(exponent-(double)AEXP)<SMALL) good_exponents++;
			else {
				errout("Fatal error: Illegal H-Bond-Exponents found.\n");
				errout("AEXP is %f, but should be %f for EGO.\n",
					exponent,(double)AEXP);
				return -1;
			}
		} else
		if (strcmp(tmpstr,"REXP")==0) {		/* found one */
			if (converted!=2)	goto illegal_exppara_line;
			if (fabs(exponent-(double)REXP)<SMALL) good_exponents++;
			else {
				errout("Fatal error: Illegal H-Bond-Exponents found.\n");
				errout("REXP is %f, but should be %f for EGO.\n",
					exponent,(double)REXP);
				return -1;
			}
		} else 
		if (strcmp(tmpstr,"HAEX")==0) {		/* found one */
			if (converted!=2)	goto illegal_exppara_line;
			if (fabs(exponent-(double)HAEX)<SMALL) good_exponents++;
			else {
				errout("Fatal error: Illegal H-Bond-Exponents found.\n");
				errout("HAEX is %f, but should be %f for EGO.\n",
					exponent,(double)HAEX);
				return -1;
			}
		} else
		if (strcmp(tmpstr,"AAEX")==0) {		/* found one */
			if (converted!=2)	goto illegal_exppara_line;
			if (fabs(exponent-(double)AAEX)<SMALL) good_exponents++;
			else {
				errout("Fatal error: Illegal H-Bond-Exponents found.\n");
				errout("AAEX is %f, but should be %f for EGO.\n",
					exponent,(double)AAEX);
				return -1;
			}
		}
	}
	return 0;	/* reached end of file */

illegal_exppara_line:
	errout("Fatal error: Illegal parameterline for for H-bond-exponents.\n");
	errout(lstr);
	return 0;
}

/**************************************************************************************/

static
int	readin_hbond_params(FILE *h_para)
{
	double	e_min,r_min;
	static	char	at1str[2*LEN_ATOM_TYPE+1];
	static	char	at2str[2*LEN_ATOM_TYPE+1];
	int	res;

	while ((res=read_next_hbondpara(h_para,at1str,at2str,&e_min,&r_min))>=0) {

		strcpy(hbpara[nr_read_hbpara].wcstr[0],at1str);
		strcpy(hbpara[nr_read_hbpara].wcstr[1],at2str);

		hbpara[nr_read_hbpara].e_min=e_min;
		hbpara[nr_read_hbpara].r_min=r_min;
#if TEST_HBONBPARA
		printf("HBond-Params: <%s> <%s>  e_min %f  r_min %f\n",
			hbpara[nr_read_hbpara].wcstr[0],
			hbpara[nr_read_hbpara].wcstr[1],
			hbpara[nr_read_hbpara].e_min,
			hbpara[nr_read_hbpara].r_min);
#endif
		if (++nr_read_hbpara>=MXNO_HBOND_PARAMS) {
			errout("Fatal error: Too many HBond-parameters (MXNO_HBOND_PARAMS is %d).\n",MXNO_HBOND_PARAMS);
			return -1;
		}
	}
	if (res==-2)	return -1;		/* thats an error */

	return 0;
}

/**************************************************************************************/

int	writeout_hbmats(char *hbahmatfn, char *hbbhmatfn)
{
	FILE	*h_hbah,*h_hbbh;
	int	i,j;

	if ((h_hbah=fopen(hbahmatfn,"w"))==NULL) {
		errout("Can not open HBond-ah-file.\n");
		return -1;
	}

	/**************** write it out *************************************************/
	fprintf(h_hbah,"This is the HBOND parameter matrix Ah.\n");	
	fprintf(h_hbah,"Number of different parameters:   %4d\n",nr_hbond_atoms);

	for (i=0;i<nr_hbond_atoms;i++) {
		for (j=0;j<nr_hbond_atoms;j++) {
			if (((j+1)%MXNO_FIELDS_BOND==0) ||
			    (j+1==nr_hbond_atoms) 
			   ) {
				fprintf(h_hbah,"%8.4f\n",amat[i*nr_hbond_atoms+j]);
			} else {
				fprintf(h_hbah,"%8.4f  ",amat[i*nr_hbond_atoms+j]);
			}
		}
	}
	
	errout("H-Bond-ah matrix written     : %d\n",nr_hbond_atoms);
	fclose(h_hbah);

	/******************************************************************************/

	if ((h_hbbh=fopen(hbbhmatfn,"w"))==NULL) {
		errout("Can not open HBond-ah-file.\n");
		return -1;
	}

	/**************** write it out ************************************************/
	fprintf(h_hbbh,"This is the HBOND parameter matrix Bh.\n");	
	fprintf(h_hbbh,"Number of different parameters:    %4d\n",nr_hbond_atoms);

	for (i=0;i<nr_hbond_atoms;i++) {
		for (j=0;j<nr_hbond_atoms;j++) {
			if (((j+1)%MXNO_FIELDS_BOND==0) ||
			    (j+1==nr_hbond_atoms) 
			   ) {
				fprintf(h_hbbh,"%8.4f\n",bmat[i*nr_hbond_atoms+j]);
			} else {
				fprintf(h_hbbh,"%8.4f  ",bmat[i*nr_hbond_atoms+j]);
			}
		}
	}

	errout("H-Bond-bh matrix written     : %d\n",nr_hbond_atoms);
	fclose(h_hbbh);

	return 0;
}

/***************************************************************************************/

static
int	add_to_hbacclist(int *hbacclist,int *hbaccdubs,int nr_read_hbacc)
{
	int	i;
	int 	off;
	int	swapbuf[4];

	off=2*nr_read_hbacc;	
						/* add to end of (sorted) list */
	hbacclist[off+0]=hbaccdubs[0];
	hbacclist[off+1]=hbaccdubs[1];

	if (nr_read_hbacc==0)	return 1;	/* first dublett, no sorting */

						/* sort new entry in ascending order */
						/* of the first atom */
	for (i=nr_read_hbacc,off=2*nr_read_hbacc;i>0;i--,off-=2) {
		if (hbacclist[off]<hbacclist[off-2]) {
			memcpy(swapbuf,		hbacclist+off-2,2*sizeof(int));
			memcpy(hbacclist+off-2,	hbacclist+off,	2*sizeof(int));
			memcpy(hbacclist+off,	swapbuf,	2*sizeof(int));
		} else	break;			/* now sorted to right place */
	}	
	return 1;
}

/************************************************************************************/

static
int	get_next_hbaccdubs(FILE *h_psf,int *hbaccdubs)	/* returns number of integers */
{							/* in bondpair ==> even       */
	static	char	linestr[MAX_STRING_SIZE];
	int 	i;
	int	nr_converted=0;

	if (read_string(h_psf,linestr)==-1)	return -1; /* error:can not read at all*/

	nr_converted=sscanf(linestr,"%d %d %d %d %d %d %d %d",
				hbaccdubs+0,hbaccdubs+1,
				hbaccdubs+2,hbaccdubs+3,
				hbaccdubs+4,hbaccdubs+5,
				hbaccdubs+6,hbaccdubs+7				
				);
	
	if (nr_converted<=0) {		/* nothing converted, may be the end */
		return 0;
	} else
	if (nr_converted%2!=0) {
		errout("Fatal error: odd converted numbers in PSF-hb-acc-list.\n");
		return -1;
	} else {
		for (i=0;i<nr_converted;i++) {
			hbaccdubs[i]--;				/* atomnumbers in ego */
								/* start with 0 */
		}
		return nr_converted;			
	}
}

/**********************************************************************************/

static
int	goto_psf_hbaccdata(FILE *h_psf,int *nr_of_hbacc)
{
	static	char	lstr[MAX_STRING_SIZE];
	static	char	tmpstr[MAX_STRING_SIZE];
	int 	res;

	while (read_string(h_psf,lstr)!=-1) {			/* read a psf-line */
		strup(lstr);
		if (sscanf(lstr,"%d %s",&res,tmpstr)==2) {	/* check it */
			tmpstr[5]=0;
			if (strcmp(tmpstr,"!NACC")==0) {
				*nr_of_hbacc=res;
				return 0;
			}
		}	
	}

	*nr_of_hbacc=0;
	errout("Warning error: No hb-acceptor-data-section (!NACC) in PSF-file.\n");
	return -1;
}

/*************************************************************************************/

static
int	add_to_hbdonlist(int *hbdonlist,int *hbdondubs,int nr_read_hbdon)
{
	int	i;
	int 	off;
	int	swapbuf[4];

	off=2*nr_read_hbdon;	
						/* add to end of (sorted) list */
	hbdonlist[off+0]=hbdondubs[0];
	hbdonlist[off+1]=hbdondubs[1];

	if (nr_read_hbdon==0)	return 1;	/* first dublett, no sorting */

						/* sort new entry in ascending order */
						/* of the first atom */
	for (i=nr_read_hbdon,off=2*nr_read_hbdon;i>0;i--,off-=2) {
		/* sort in first AND second row !!! other then in acc-list */
		if ( (hbdonlist[off]<hbdonlist[off-2]) ||
		    ((hbdonlist[off]==hbdonlist[off-2])&&
		     (hbdonlist[off+1]>hbdonlist[off-2+1])
		    )
		   ) {
			memcpy(swapbuf,		hbdonlist+off-2,2*sizeof(int));
			memcpy(hbdonlist+off-2,	hbdonlist+off,	2*sizeof(int));
			memcpy(hbdonlist+off,	swapbuf,	2*sizeof(int));
		} else	break;				/* now sorted to right place */
	}	
	return 1;
}

/*********************************************************************************/

static
int	get_next_hbdondubs(FILE *h_psf,int *hbdondubs)	/* returns number of integers */
{							/* in hb-don   ==> even       */
	static	char	linestr[MAX_STRING_SIZE];
	int 	i;
	int	nr_converted=0;

	if (read_string(h_psf,linestr)==-1)	return -1; /* error:can not read at all*/

	nr_converted=sscanf(linestr,"%d %d %d %d %d %d %d %d",
				hbdondubs+0,hbdondubs+1,
				hbdondubs+2,hbdondubs+3,
				hbdondubs+4,hbdondubs+5,
				hbdondubs+6,hbdondubs+7				
				);
	
	if (nr_converted<=0) {		/* nothing converted, may be the end */
		return 0;
	} else
	if (nr_converted%2!=0) {
		errout("Fatal Error: odd converted numbers in PSF-hb-don-list.\n");
		return -1;
	} else {
		for (i=0;i<nr_converted;i++) {
			hbdondubs[i]--;				/* atomnumbers in ego */
								/* start with 0 */
		}
		return nr_converted;			
	}
}

/*************************************************************************************/

static
int	goto_psf_hbdondata(FILE *h_psf,int *nr_of_hbdon)
{
	static	char	lstr[MAX_STRING_SIZE];
	static	char	tmpstr[MAX_STRING_SIZE];
	int 	res;

	while (read_string(h_psf,lstr)!=-1) {	
		strup(lstr);					/* read a psf-line */
		if (sscanf(lstr,"%d %s",&res,tmpstr)==2) {	/* check it */
			tmpstr[5]=0;
			if (strcmp(tmpstr,"!NDON")==0) {
				*nr_of_hbdon=res;
				return 0;
			}
		}	
	}

	*nr_of_hbdon=0;
	errout("Warning error: No hb-donator-data-section (!NDON) in PSF-file.\n");

	return -1;
}

/*************************************************************************************/

