/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*********************************************************************************
**										**
**	Module	: woshake.c	by M.Eichinger, Munich				**
**	Belongs2: xpl2lis	Converter Xplor-data to EGO-data (*.lis)	**
**	Contents: scan in and write out the SHAKE-list				**
**										**
**  	Modif's:								**
**		05.06.94							**
**		07.10.94 Improved coment in case of errors 			**
**										**
*********************************************************************************/
#include <stdio.h>

#include "hardw.h"
#include "ego.h"
#include "util_lib.h"
#include "wodecls.h"

/******************************************************************************/
/**************** functions in this modul *************************************/
/******************************************************************************/

	int	writeout_shakelist(int nr_of_atoms,char *shakelfn);

/******************************************************************************/

int	writeout_shakelist(int nr_of_atoms,char *shakelfn)
{
	FILE	*h_shake;
	int	res=0,warning=0;
	int	nr_of_hydrogens=0;
	int	atom_index=0;
	double	atom_mass=0.0;
static	int	bondlist[MXNO_FIELDS_BOND+1],nr_of_bonds;

	if ((h_shake=fopen(shakelfn,"w"))==NULL) {
		errout("Can not open shake-list-file <%s>.\n",shakelfn);
		return -1;
	}
	fprintf(h_shake,"Format of the SHAKE list:\n");
	fprintf(h_shake,"Hydrogen  Antezedent\n");
	
	for (atom_index=0;atom_index<nr_of_atoms;atom_index++) {

		res = get_atti_by_atomnr(atom_index);
		atom_mass = get_atom_mass(res);

		if (isit_a_hydrogen(atom_mass)) {
			nr_of_bonds=get_atom_bondlist(atom_index,bondlist);			
			if ((warning<10)&&(nr_of_bonds!=1)) {	/* a H-Atom can only make one bond ! */
				warning++;
				errout("Found a hydrogen with nr_of_bonds(%d)!=1. This may be an error!\n",nr_of_bonds);
				errout("This warning is printed at maximum 10 times.\n");
			}
			nr_of_hydrogens++;
		}
	}

	txtout("Number of hydrogens (SHAKE-List): %d\n",nr_of_hydrogens);

	fprintf(h_shake," %4d\n",nr_of_hydrogens);
	for (atom_index=0;atom_index<nr_of_atoms;atom_index++) {

		res = get_atti_by_atomnr(atom_index);
		atom_mass = get_atom_mass(res);

		if (isit_a_hydrogen(atom_mass)) {
			nr_of_bonds=get_atom_bondlist(atom_index,bondlist);			
			if ((warning<10)&&(nr_of_bonds!=1)) {   /* a H-Atom can only make one bond ! */
                                warning++;
                                errout("Found a hydrogen with nr_of_bonds(%d)!=1. This may be an error!\n",nr_of_bonds);
                                errout("This warning is printed at maximum 10 times.\n");
                        }

			fprintf(h_shake," %4d     %4d\n",atom_index,bondlist[0]);		
		}
	}

	fclose(h_shake);
	return 0;
}

