/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*********************************************************************************
**										**
**	Module	: wounitid.c	by Chris Brandt				**
**	Belongs2: xpl2lis	Converter Xplor-data to EGO-data (*.lis)	**
**	Contents:  								**
**		code to read residue descriptions from unit.def	**
**  	Modif's:								**
**		19.10.1995 mcb created	
**		12.02.1996 mcb fixed bug: didn't test for exact equality between
			       residue and alternatives in identify_alternative()**										**										**
*********************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "ego.h"
#include "wounitid.h"
#include "wodecls.h"

/******************************************************************************************
**	add_residue()
**	adds the datastructure for the description of another residue to the array descriptions
*******************************************************************************************/

res_description_t ** add_residue(res_description_t ** descriptions, char res_name[LEN_RES_TYPE+1])
{
	res_description_t * res_des;

	/* add room for one more element to descriptions */
	descriptions = (res_description_t **) realloc(descriptions, SIZE_PLUS_ONE(descriptions));
	LAST_PLUS_ONE_ELEMENT(descriptions) = NULL; 

	/* create element */
	res_des = (res_description_t *) malloc(sizeof(res_description_t));
	strcpy(res_des->name, res_name);
	res_des->alternatives = (res_template_t **) malloc(sizeof(res_template_t *));
	res_des->alternatives[0] = NULL;
	
	/* add element to descriptions */
	LAST_ELEMENT(descriptions) = res_des; 
	
			
	/* return (possibly changed) descriptions */
	
	return 	descriptions;
}


/******************************************************************************************
**	add_alternative()
**	adds the datastructure for the description of another alternative to residue
*******************************************************************************************/

int add_alternative(res_description_t *residue)
{
	res_template_t * res_template;
	
	/* add room for one more element to residue */
	residue->alternatives = (res_template_t **) realloc(residue->alternatives, SIZE_PLUS_ONE(residue->alternatives));
	LAST_PLUS_ONE_ELEMENT(residue->alternatives) = NULL;

	/* create element */
	res_template = (res_template_t *) malloc(sizeof(res_template_t));
	res_template->units = (unit_template_t **) malloc(sizeof(unit_template_t *));
	res_template->units[0] = NULL;


	/* add element to residue */
	LAST_ELEMENT(residue->alternatives) = res_template; 
	      	

	return 	0;
}


/******************************************************************************************
**	add_unit()
**	adds the datastructure for the description of another unit to alternative
*******************************************************************************************/

int add_unit(res_template_t *alternative)
{
	unit_template_t 	*unit_template;
	
	/* add room for one more element to alternative */
	alternative->units = (unit_template_t **) realloc(alternative->units, SIZE_PLUS_ONE(alternative->units));
	LAST_PLUS_ONE_ELEMENT(alternative->units) = NULL;
	
	/* create element */
	unit_template = (unit_template_t *) malloc(sizeof(unit_template_t));
	unit_template->atoms = (char **) malloc(sizeof(char *));
	unit_template->atoms[0] = NULL;
	
	/* add element to alternative */
	LAST_ELEMENT(alternative->units) = unit_template; 
	
	return 	0;
}


/******************************************************************************************
**	add_atom()
**	adds the datastructure for the description of another atom to unit
*******************************************************************************************/

int add_atom(unit_template_t *unit, char atom_name[LEN_ATOM_TYPE+1])
{
	char	*name;
	
	/* add room for one more element to unit */
	unit->atoms = (char **) realloc(unit->atoms, SIZE_PLUS_ONE(unit->atoms));
	LAST_PLUS_ONE_ELEMENT(unit->atoms) = NULL;

	/* create element */
	name = (char *) malloc(LEN_ATOM_TYPE+1);
	strcpy(name, atom_name);
	
	/* add element to unit */
	LAST_ELEMENT(unit->atoms) = name; 
	

	return 	0;
}


/******************************************************************************************
**	strapchr()
**	appends c to to, we assume that to is long enough 
*******************************************************************************************/

char * strapchr(char *to, char c)
{
	int	len = strlen(to);
	
	to[len] = c;
	to[len+1] = '\0';
	
	return to;
}


/******************************************************************************************
**	newest_index()
**	returns the second to last index into array, array elementsize must be the size of a pointer 
*******************************************************************************************/

int last_index(void **array)
{
	int	i;
	
	for(i=0; array[i] != NULL; i++)
		;
		
	return i;
}


int newest_index(void **array){
	int	i;
	
	for(i=0; array[i] != NULL; i++)
		;
	i--;
		
	return i;
}



/******************************************************************************************
**	add_to_buffer()
**	adds the atom to the working buffer
*******************************************************************************************/

abuf_entry_t ** add_to_buffer(abuf_entry_t **buffer, STRUCT_ATOMDATA *atom)
{
	abuf_entry_t	*entry;
	
	/* add room for one more entry to buffer */
	buffer = (abuf_entry_t **) realloc(buffer, SIZE_PLUS_ONE(buffer));
	LAST_PLUS_ONE_ELEMENT(buffer) = NULL;

	/* create entry */
	entry = (abuf_entry_t *) malloc(sizeof(abuf_entry_t));
	entry->atom_id = atom->atomnr;
	strcpy(entry->atom_name, atom->atomstr);
	strcpy(entry->res_name, atom->residstr);
	entry->assigned = 0;

	/* add entry to buffer */
	LAST_ELEMENT(buffer) = entry; 
	      	

	return 	buffer;
}


/******************************************************************************************
**	free_buffer()
**	frees the working buffer
*******************************************************************************************/

abuf_entry_t **free_buffer(abuf_entry_t **buffer)
{	
	int	i;
	
	for (i = 0; buffer[i] != NULL; i++) 
		free(buffer[i]);
		
	buffer = realloc(buffer, sizeof(abuf_entry_t *));
	buffer[0] = NULL;
	
	return buffer;
}


/******************************************************************************************
**	reset_buffer()
**	resets the assigned flag of all entrys in the buffer
*******************************************************************************************/

void reset_buffer(abuf_entry_t **buffer)
{	
	int	i;
	
	for (i = 0; buffer[i] != NULL; i++) 
		buffer[i]->assigned = 0;
		
}


/******************************************************************************************
**	get_and_mark()
**	clears the working buffer
*******************************************************************************************/

int get_and_mark(abuf_entry_t **buffer, char *name)
{	
	int	i;
	
	for (i = 0; buffer[i] != NULL; i++) {
		if ((buffer[i]->assigned == 0) && (strcmp(buffer[i]->atom_name, name) == 0)) {
			buffer[i]->assigned = 1;
			return buffer[i]->atom_id;
		}
	}
/**
	txtout("No match for atomname <%s>\n",name);
**/ 
	return -1;			
}


/******************************************************************************************
**	identify_residue()
**	returns the index into descriptions for the residue of the atoms in buffer
**	
*******************************************************************************************/

int identify_residue(res_description_t ** descriptions, abuf_entry_t **buffer)
{
	int	i;
	
	if (buffer[0] == NULL) {
		printf("Fatal error in identify_residue(): recieved empty buffer.\n");
		exit(-1); /* empty buffer */
	}
	
	for (i = 0; descriptions[i] != NULL; i++)
		if (strcmp(descriptions[i]->name, buffer[0]->res_name) == 0)
			return i;
	
	printf("Fatal error in identify_residue(): No definition for residue name <%s>.\n",
		buffer[0]->res_name);
	exit(-1); /* no match */
	return -1;
}


/******************************************************************************************
**	identify_alternative()
**	returns the index into alternatives for the residue of the atoms in buffer
**	
*******************************************************************************************/

int identify_alternative(res_description_t *residue, abuf_entry_t **buffer)
{
	int	i;
	int	x;	/* if < 0: atom in alternative was not found in buffer, no match with this alternative */
	int	alt_idx, unit_idx, atom_idx;
	
	if (buffer[0] == NULL) {
		printf("Fatal error in identify_alternative(): received empty buffer.\n");
		exit(-1); /* empty buffer */
	}
	
	for (alt_idx = 0, x = 0; residue->alternatives[alt_idx] != NULL; alt_idx++, x = 0) {
		for (unit_idx = 0; (residue->alternatives[alt_idx]->units[unit_idx] != NULL) && (x >= 0); unit_idx++) {
			for (atom_idx = 0; (residue->alternatives[alt_idx]->units[unit_idx]->atoms[atom_idx] != NULL) && (x >= 0); atom_idx++) {
				x = get_and_mark(buffer, residue->alternatives[alt_idx]->units[unit_idx]->atoms[atom_idx]);					
			}
		}
		
		/* Now we test, if all atoms have been marked. ***/
		/* If so, we have identified the alternative ***/
		
		for (i = 0; buffer[i] != NULL; i++)
			if (	(buffer[i]->assigned == 0) && 
				(!is_hydrogen(buffer[i])))
				break;
			
		reset_buffer(buffer);
		
		if (buffer[i] == NULL)	return alt_idx;	
				/* the loop went all the way through, ***/
				/* meaning we have covered all atoms in the buffer */
	}
	
	printf("Fatal error in identify_alternative(): no (%d) valid alternative definition\n");
	printf("found for atom %d, residue name <%s>, atom name <%s>.\n",buffer[0]->assigned, buffer[0]->atom_id, buffer[0]->res_name, buffer[0]->atom_name);
	exit(-1); /* no match */
	return -1;
}

/******************************************************************************************
**	is_hydrogen()
**	returns !=0 if atom is hydrogen, else returns 0
**	
*******************************************************************************************/

int is_hydrogen(abuf_entry_t *atom)
{
	return isit_a_hydrogen(get_atom_mass(get_atti_by_atomnr(atom->atom_id)));
}



/******************************************************************************************
**	getToken()
**	reads from theFile until a relevant character for the unit definition is read and 
**	returns that character
*******************************************************************************************/

int getToken(FILE *theFile)
{
	int		inComment = 0;
	int	theChar;
	
	while( (theChar = getc(theFile)) != EOF){
		switch (inComment) {
			case 1:
				if (theChar == '\n')
					inComment = 0;
				break;
		
			case 0:
				if (theChar == '!'){
					inComment = 1;
					break;
				} else if (isspace(theChar))
					break;
					
			default:
				return theChar;
		}
	}
	return theChar;
}

/******************************************************************************************
**	add_padding()
**	adds padding blanks to the end of str 
**	
*******************************************************************************************/

void add_padding(char *str)
{
	int i;
	
	for (i = 0; i < LEN_RES_TYPE; i++)		/* for now we asume, that Residue and Atom strings have the same length */
		if (str[i] == '\0')
			break;
			
	for( ;i <LEN_RES_TYPE; i++)	
		str[i] = ' ';	/* blank */

	str[i] = '\0';
}
/******************************************************************************************
**	clear_string()
**	sets the entirety of str to '\0' 
**	
*******************************************************************************************/

void clear_string(char *str)
{
	int i;
	
	for (i = 0; i <= LEN_RES_TYPE; i++)		/* for now we asume, that Residue and Atom strings have the same length */
		str[i] = '\0';
}

			
/******************************************************************************************
**	read_res_descriptions()
**	reads token after token and builds up the residue/unit description data in descriptions 
**	and returns it
*******************************************************************************************/

res_description_t ** read_res_descriptions(char *def_file_name)
{
	FILE				*theFile;
	int					state = cStateInName;
	int					result;
	int				theChar;
	char				*name;
	res_description_t	**descriptions;
	
	name = (char*) malloc(128);
	clear_string(name);
	
	descriptions = (res_description_t **) malloc(sizeof(res_description_t *));
	descriptions[0] = NULL;
	
	theFile = fopen(def_file_name, "r");
	if (theFile == NULL) {
		printf("Fatal error in read_res_descriptions(): could not open units.def file.\n");
		exit(-1); /* no match */
	}
	
	while ((theChar = getToken(theFile)) != EOF) {
		switch (state) {
		
			case cStateInName:
				if ( 	isalnum(theChar) || 
					(theChar=='*') || 
					(theChar==39) /*** the prime ' ***/
				) {
					name = strapchr(name, theChar);
					break;
					
				} else if (theChar == '{') {
					add_padding(name);
					descriptions = add_residue(descriptions, name);
					clear_string(name);
					state = cStateInAlternatives;
					break;
					
				} else 
					goto parse_error;
					
					
			case cStateInAlternatives:
				if (theChar == '{') {
					result = add_alternative(NEWEST_ELEMENT(descriptions));
					state = cStateInUnits;
					break;
					
				} else if (theChar == '}') {
					state = cStateInName;
					break;
					
				} else
					goto parse_error;
						
			case cStateInUnits:
				if (theChar == '{') {
					result = add_unit(NEWEST_ELEMENT(NEWEST_ELEMENT(descriptions)->alternatives));
					state = cStateInAtoms;
					break;
					
				} else if (theChar == '}') {
					state = cStateInAlternatives;
					break;
					
				} else
					goto parse_error;
			
			
			case cStateInAtoms:
				if ( 	isalnum(theChar) ||
					(theChar=='*') || 
					(theChar==39) /*** the prime ***/
				) {
					name = strapchr(name, theChar);
					break;
					
				} else if (theChar == ',') {
					add_padding(name);
					result = add_atom(NEWEST_ELEMENT(NEWEST_ELEMENT(NEWEST_ELEMENT(descriptions)->alternatives)->units), name);
					clear_string(name);
					break;
					
				} else if (theChar == '}') {
					if (strlen(name)>0) {
						add_padding(name);
						result = add_atom(NEWEST_ELEMENT(NEWEST_ELEMENT(NEWEST_ELEMENT(descriptions)->alternatives)->units), name);
						clear_string(name);
					}
					state = cStateInUnits;
					break;
				} else
					goto parse_error;
			}
	}
	
	fclose(theFile);
	return descriptions;

parse_error:
	printf("Fatal error in read_res_descriptions(): error parsing units.def file.\n");
	printf("Last read character (%c,%d)\n",theChar,(int)theChar);
	printf("Please check if file has correct format.\n");
	exit(-1);
	return descriptions; /*** to avoid compiler warning ***/
}
		
						

/******************************************************************************************
**	write_res_descriptions()
**	writes out the unit descriptions 
**	usefull for debugging and possibly later for toolsupport of units.def file creation
*******************************************************************************************/

int write_res_descriptions(res_description_t **descriptions, char *filen)
{
	FILE	*wf;
	int		result;
	int		ire, ial, iun, iat;	/* index variables for the corresponding arrays */
	
	wf = fopen(filen, "w");
	if (wf == NULL) {
		printf("Fatal error in write_res_descriptions(): could not open output file.\n");
		exit(-1);
	}
		
	for( ire = 0; descriptions[ire] != NULL; ire++) {
		fprintf(wf, "%s {", descriptions[ire]->name);
		for( ial = 0; descriptions[ire]->alternatives[ial] != NULL; ial++){
			fprintf(wf, "\n\t{\n\t\t");
			for(iun = 0; descriptions[ire]->alternatives[ial]->units[iun]!= NULL; iun++) {
				fprintf(wf, "{");
				for( iat = 0; descriptions[ire]->alternatives[ial]->units[iun]->atoms[iat] != NULL; iat++){
					fprintf(wf, "%s, ", descriptions[ire]->alternatives[ial]->units[iun]->atoms[iat]);
				}
				fprintf(wf,"}");
			}
			fprintf(wf, "\n\t}");
		}
		fprintf(wf, "\n}\n\n");
	
	}
	result = fclose(wf);
	if (result) {
		printf("Fatal error in write_res_descriptions(): could not close output file.\n");
		exit(-1);
	}
	
	return 0;
}	










