/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*********************************************************************************
**										**
**	Module	: wounitid.h	by Chris Brandt				**
**	Belongs2: xpl2lis	Converter Xplor-data to EGO-data (*.lis)	**
**	Contents:  								**
**		defines the datastructures needed for unit identification based on unit.def	**
**  	Modif's:								**
**		19.10.1995 mcb created								**										**										**										**
*********************************************************************************/


/*****************************************************************
**						Constants								**
*****************************************************************/

enum states {cStateInName = 1, cStateInAlternatives, cStateInUnits, cStateInAtoms};


/*****************************************************************
**						Macros									**
*****************************************************************/

#define	SIZE_PLUS_ONE(a)	((last_index((void**)a) + 2) * sizeof(a[0]))
#define LAST_ELEMENT(a) (a[last_index((void**)a)])
#define NEWEST_ELEMENT(a) (a[newest_index((void**)a)])
#define LAST_PLUS_ONE_ELEMENT(a) (a[last_index((void**)a) + 1])
#define CHECK(a) if (malloc_size(a) != 4) printf("error\n");


/*****************************************************************
**						Types									**
*****************************************************************/

/****** Unit template, holds all atoms for one unit of a residue */

typedef struct unit_temp {
	char	**atoms;		/* NULL terminated array of atom strings */
} unit_template_t;


/****** Residue template, holds all units for one alternative representation of a residue */

typedef struct res_temp {
	unit_template_t	**units;		/* NULL terminated array of pointer to units */
} res_template_t;


/****** Residue description, holds all units for all alternatives */

typedef struct res_des {
	char			name[LEN_RES_TYPE+1];
	res_template_t	**alternatives;		/* NULL terminated array of pointer to alternatives */
} res_description_t;


/****** atom buffer entry */

typedef struct abuf_entry {
	char			res_name[LEN_RES_TYPE+1];
	char			atom_name[LEN_ATOM_TYPE+1];
	int				atom_id;
	int				assigned;	/* 0=not yet assigned, 1 = has been assigned a unit */
} abuf_entry_t;


/*****************************************************************
**						Prototypes								**
*****************************************************************/

/****** utilities */
char * strapchr(char *to, char c);
int newest_index(void **array);
int	last_index(void **array);

/****** Create Datastructures */
res_description_t ** add_residue(res_description_t ** descriptions, char res_name[LEN_RES_TYPE+1]);
int add_alternative(res_description_t *residue);
int add_unit(res_template_t *alternative);
int add_atom(unit_template_t *unit, char atom_name[LEN_ATOM_TYPE+1]);

/****** Read Data */
res_description_t **read_res_descriptions(char *def_file_name);
int getToken(FILE *theFile);
void add_padding(char *str);
void clear_string(char *str);

/****** Write Data */
int write_res_descriptions(res_description_t **descriptions, char *filen);

/****** atom buffer management ******/
abuf_entry_t ** add_to_buffer(abuf_entry_t **buffer, STRUCT_ATOMDATA *atom);
abuf_entry_t **free_buffer(abuf_entry_t **buffer);
void reset_buffer(abuf_entry_t **buffer);
int get_and_mark(abuf_entry_t **buffer, char *name);

/****** finding data in descriptions */
int identify_residue(res_description_t ** descriptions, abuf_entry_t **buffer);
int identify_alternative(res_description_t *residue, abuf_entry_t **buffer);

/******* handling of hydrogen */
int is_hydrogen(abuf_entry_t *atom);

